SystemOrganization addCategory: #'Qwaq-Widgets-Annotations'!
SystemOrganization addCategory: #'Qwaq-Widgets-Basic'!
SystemOrganization addCategory: #'Qwaq-Widgets-Desktop'!
SystemOrganization addCategory: #'Qwaq-Widgets-Editing'!
SystemOrganization addCategory: #'Qwaq-Widgets-General'!
SystemOrganization addCategory: #'Qwaq-Widgets-MovieApp'!
SystemOrganization addCategory: #'Qwaq-Widgets-Portals'!
SystemOrganization addCategory: #'Qwaq-Widgets-Project'!
SystemOrganization addCategory: #'Qwaq-Widgets-SlideShow'!
SystemOrganization addCategory: #'Qwaq-Widgets-TextApp'!

QLabel subclass: #QLabelButton
	instanceVariableNames: 'handler clientData target action'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QLabelButton commentStamp: '<historical>' prior: 0!
A Q-Label that is also a clickable button.  
Calls the action selector on its target, passing optionally
	first argument: CroquetEvent	
	second argument: (clientData, or if that is nil, the buttonlabel itself)

Detects double-clicks, and raises them as
	 target signal: #buttonDoubleClick with: event with ( clientData ifNil self )
(The single-click action -does- fire before this.)

Cannot be persisted (the clientData could be anything at all).!

!QLabelButton class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLabelButton class methodsFor: 'defaults' stamp: 'bgf 8/20/2009 21:57'!
defaultBackground
	^ (QColors controlNormalColor * 0.8) asColor! !

!QLabelButton class methodsFor: 'defaults' stamp: 'bgf 1/19/2010 16:04'!
defaultDoesClickSelect
	^ false! !

!QLabelButton class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QLabelButton class methodsFor: 'defaults' stamp: 'bgf 8/20/2009 20:43'!
defaultMargin
	^ 1! !

!QLabelButton class methodsFor: 'defaults' stamp: 'bgf 8/20/2009 20:43'!
defaultStyle
	"The default TextStyle to use for rendering me."
	^ QStyles find:#SignStyle
! !

!QLabelButton class methodsFor: 'defaults' stamp: 'bgf 8/20/2009 20:32'!
defaultTextColor
	^ Color white! !

!QLabelButton methodsFor: 'accessing' stamp: 'bgf 8/20/2009 21:14'!
clientData
	^ clientData! !

!QLabelButton methodsFor: 'accessing' stamp: 'bgf 8/20/2009 21:21'!
clientData: anObject
	"If supplied, will be returned as the second argument when the action is performed.
	 (The first argument is the CroquetEvent that fires the button.)"
	clientData := anObject! !

!QLabelButton methodsFor: 'initializing' stamp: 'bgf 11/10/2009 16:33'!
initialize
	super initialize.

	handler := QButtonHandler body: self target: self action: #noAction:.

	handler overColor: (QColors qwaqOrange asVectorColor alpha:0.7).
	handler downColor: (Color black asVectorColor alpha: 0.3).
	handler fireColor: (Color white asVectorColor alpha: 0.5).

	self runScript: #relayDoubleClick: when: { self . #doubleClick }.! !

!QLabelButton methodsFor: 'events' stamp: 'bgf 8/20/2009 22:15'!
relayDoubleClick: event
	"Raise buttonDoubleClick  on the target - relayed from my buttonHander"
	target signal: #buttonDoubleClick with: event with: (clientData ifNil: [ self ])! !

!QLabelButton methodsFor: 'events' stamp: 'bgf 8/20/2009 21:25'!
runAction: event
	target ifNotNil: [ 
		action numArgs caseOf: {
			[0] -> [ target perform: action ].
			[1] -> [ target perform: action with: event ].
			[2] -> [ target perform: action with: event with: (clientData ifNil: [ self ]  ) ].
		} otherwise: [
			QLogger logWarning: 'Button action ', action, ' has more than 2 arguments.'.
		].
	].! !

!QLabelButton methodsFor: 'persist' stamp: 'bgf 8/20/2009 21:24'!
shouldPersist
	^ false! !

!QLabelButton methodsFor: 'initializing' stamp: 'bgf 8/20/2009 21:59'!
target: aFrame action: aSelector
	target := aFrame.
	action := aSelector.
	
	"The handler disables itself if I don't understand the action I tell it...
	 so I tell it an action I don't understand, when my target tells me one it doesn't..."
	(target respondsTo: aSelector) ifTrue: [
		handler action: #runAction:
	] ifFalse: [
		handler action: #noAction:
	].! !

TOutlineGroup subclass: #QEditBoxOutline
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QEditBoxOutline class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEditBoxOutline methodsFor: 'scale' stamp: 'bgf 2/11/2010 00:53'!
useMatrixScaling
	^ true! !

Object subclass: #QControlFader
	instanceVariableNames: 'assembly persist animateTime stepping baseAlpha'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QControlFader commentStamp: '<historical>' prior: 0!
Implements on-demand auto-fadeout behavior
for a TFrame assembly.

When the ControlFader is enabled (#enable),
its assembly is made visible and its transparency is set to 'baseAlpha'.
If the fader is persistent, it remains that way.
If not, it begins to slowly fade into transparency.

Instances are created with QControlFader for: aTFrame!

!QControlFader class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QControlFader class methodsFor: 'instance creation' stamp: 'bgf 8/25/2006 21:01'!
for: aTFrame
	"Create a ControlFader assigned to fade the given tframe."

	^ (self new) assembly: aTFrame; yourself! !

!QControlFader methodsFor: 'accessing' stamp: 'bgf 8/25/2006 20:42'!
assembly
	"The TFrame whose visibility I control"
	^ assembly! !

!QControlFader methodsFor: 'initialize' stamp: 'bgf 8/25/2006 20:46'!
assembly: aTFrame
	"Assign the frame over which the fader has control;
	 only do this at initialization time, before any interaction."

	assembly := aTFrame.
	assembly solidVisibleTree: false.! !

!QControlFader methodsFor: 'accessing' stamp: 'bgf 8/25/2006 21:35'!
baseAlpha

	"The alpha (transparency) value used when the assembly is to be fully visible;
	  fading takes the frame slowly from the given transparency to fully transparent."
	^ baseAlpha! !

!QControlFader methodsFor: 'accessing' stamp: 'bgf 8/25/2006 20:44'!
baseAlpha: aScalar
	"The alpha (transparency) value used when the assembly is to be fully visible;
	  fading takes the frame slowly from the given transparency to fully transparent."
	baseAlpha := aScalar min: 1.0 max: 0.0! !

!QControlFader methodsFor: 'action' stamp: 'bgf 8/25/2006 21:04'!
disable
	
	"Fade the frame out of sight (slowly), unless it is persistent."
	persist ifFalse:[
		animateTime := assembly now + self fadeDelay.
		self startStepping.
	].
! !

!QControlFader methodsFor: 'action' stamp: 'bgf 8/25/2006 20:47'!
enable

	"Make the frame visible, ready for resizing."

	assembly solidVisibleTree: true. 
	assembly transparency: baseAlpha.

	animateTime := nil.
	persist ifFalse:[
		self startStepping
	].! !

!QControlFader methodsFor: 'accessing' stamp: 'bgf 8/25/2006 20:38'!
fadeDelay

	"The period of time over which the frame fades to invisible, 
	 when it is disabled."

	^1500.0  "1.5 seconds"! !

!QControlFader methodsFor: 'initialize' stamp: 'bgf 9/6/2006 15:20'!
initialize
	super initialize.
	
	stepping := false.
	persist := false.
	animateTime := nil.
	baseAlpha := 0.6.
! !

!QControlFader methodsFor: 'accessing' stamp: 'bgf 8/25/2006 20:38'!
persist

	^ persist.! !

!QControlFader methodsFor: 'accessing' stamp: 'bgf 8/25/2006 20:38'!
persist: bool

	persist := bool.
	persist ifTrue:[ 
		self enable
	] ifFalse: [
		self disable
	].! !

!QControlFader methodsFor: 'stepping' stamp: 'bgf 8/25/2006 20:38'!
startStepping

	stepping ifFalse:[
		stepping := true.
		self step
	].! !

!QControlFader methodsFor: 'stepping' stamp: 'bgf 8/25/2006 21:05'!
step

	"interpolate transparency"
	| timeLeft |
	persist ifFalse:[
		animateTime ifNotNil: [
			timeLeft := animateTime - assembly now.
			timeLeft <= 0 ifTrue: [
				stepping := false.
				animateTime := nil. 
				assembly solidVisibleTree: false. 
			] ifFalse: [
				assembly transparency: ((timeLeft asFloat / self fadeDelay)*baseAlpha).
			].
		].
	] ifTrue:[
		stepping := false
	].
	stepping ifTrue:[(self future: 50) step].! !

!QControlFader methodsFor: 'stepping' stamp: 'bgf 8/25/2006 20:38'!
stepping
	^stepping! !

Object subclass: #QCreateObjectCallback
	instanceVariableNames: 'event avatar pointer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QCreateObjectCallback commentStamp: '<historical>' prior: 0!
See QFileCard dropInSpace:pointer: .. this is a pseudo-closure
that allows a file card to launch asynchronous jobs
to instantiate objects from template documents.

I remember the event and pointer that triggered a drop action,
so that once the harness creates the dropped object,
I can send those back to the dragAdapter to have the new object
positioned appropriately in the space.!

!QCreateObjectCallback class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCreateObjectCallback methodsFor: 'initialize-release' stamp: 'bgf 6/2/2007 02:30'!
event: ev avatar: av pointer: pt

	"I exist only to record this triplet until
	 my receipt of the #objectCreated: callback."
	event := ev.
	avatar := av.
	pointer := pt.! !

!QCreateObjectCallback methodsFor: 'initialize-release' stamp: 'bgf 3/26/2010 11:24'!
objectCreated: anObject
	"The callback we've been waiting for."
	(avatar notNil and: [ avatar root notNil] and: [anObject notNil]) ifTrue: [
		QDragAdapter drop: anObject inSpace: (avatar root) forEvent: event pointer: pointer.
		QSoundEffects play:#tingle for: anObject.
		anObject root ifNotNilDo: [ : space | 
			avatar selectFrame: anObject.
			(space getsEditBoxOnCreation: anObject) ifTrue: [ 
				"Do editor in future so shapeChanges and root: handlers can settle."
				avatar future editFrame: anObject ].
			].

		].
	avatar := nil.
	event := nil.
	pointer := nil.! !

Object subclass: #QRange
	instanceVariableNames: 'min max'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QRange commentStamp: 'das 11/26/2007 08:42' prior: 0!
QRange is used to generate a random float between two values. It is designed to work deterministically, hence the #select: method takes a Random as an argument.!

!QRange class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRange methodsFor: 'persist' stamp: 'das 11/30/2007 13:41'!
constructFrom: pStream

	self min: (pStream readNumber: 'min' default: self min)
		max:(pStream readNumber: 'max' default: self max).! !

!QRange methodsFor: 'accessing' stamp: 'das 11/20/2007 09:13'!
max

	^ max.! !

!QRange methodsFor: 'accessing' stamp: 'das 11/20/2007 09:13'!
max: mx

	max := mx.! !

!QRange methodsFor: 'accessing' stamp: 'das 11/20/2007 09:12'!
min

	^ min! !

!QRange methodsFor: 'accessing' stamp: 'das 11/20/2007 09:12'!
min: mn

	min := mn.! !

!QRange methodsFor: 'accessing' stamp: 'das 11/20/2007 09:13'!
min: mn max: mx

	min := mn.
	max := mx.! !

!QRange methodsFor: 'persist' stamp: 'das 11/30/2007 13:40'!
persistOn: pStream
	pStream storeNumber: 'min' value: self min.
	pStream storeNumber: 'max' value: self max.! !

!QRange methodsFor: 'action' stamp: 'das 11/20/2007 09:26'!
select: rand

	^ rand next * (max-min) + min.! !

QFixedSurface subclass: #QNoteBoard
	instanceVariableNames: 'surface resize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QNoteBoard commentStamp: '<historical>' prior: 0!
A specialized fixed surface for use as a notes+topics work area.
Body is a textured rectangle (whose material is mine, and editable, and persistent.)
Always note-collecting.

!

QNoteBoard subclass: #QBrainstormingBoard
	instanceVariableNames: 'logo'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QBrainstormingBoard commentStamp: '<historical>' prior: 0!
A noteboard extended with
	- custom graphics meant to be useful at large sizes
	- a brainstorming logo indicator in top left
	- the ability to persist children

This is the Brainstorming Wall class used in project forums
and other more recent templates.
!

!QBrainstormingBoard class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QBrainstormingBoard class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 17:30'!
defaultBorderWidth
	"Most templates cover a BB wall with a handbuilt whole texture."
	^ 0! !

!QBrainstormingBoard class methodsFor: 'defaults' stamp: 'bgf 7/27/2008 13:55'!
defaultMaterial
	^ (QMaterials brainstormBoardMaterial)
		cullFace: false;
		yourself! !

!QBrainstormingBoard class methodsFor: 'defaults' stamp: 'bgf 7/25/2008 11:51'!
logoMaterial
	"Material for the logo marker in the upper right of brainstorming boards."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(1 1 1) asVectorColor;
			diffuseColor: #(0.7 0.7 0.7 0.8) asVectorColor;
			specularColor: #(0.1 0.1 0.1) asVectorColor;
			texture: (QTextures find: 'Content/Qwaq/Textures/brainstorm.jpg');
			transparency: 0.9.
	^ material.
! !

!QBrainstormingBoard methodsFor: 'accessing' stamp: 'bgf 8/27/2009 20:21'!
childrenForBrowser
	"Allow modifying decorations, frames, etc."
	^ self childrenForPersistence! !

!QBrainstormingBoard methodsFor: 'accessing' stamp: 'bgf 8/27/2009 20:20'!
childrenForPersistence
	"Allow modifying decorations, frames, etc."
	^ self children copyWithoutAll: { surface . logo . (resize grip) }.! !

!QBrainstormingBoard methodsFor: 'persist' stamp: 'bgf 8/27/2009 19:44'!
constructFrom: pStream
	"Recreate the receiver from storage"
	super constructFrom: pStream.

	self class == QBrainstormingBoard ifTrue:[
		(pStream readSequence: 'children' default: #()) do: [ :otherKid |
			"From the past, a lot of these are carrying around a duplicate logo child.
		       This guard can be deleted in the late fall of 2009"
			((otherKid isKindOf: TSimpleRectangle )
			 and: [ otherKid material notNil ] 
			 and: [ otherKid material texture = logo material texture 
					or: [ otherKid material texture = (QMaterials resizeCornerMaterial texture) ]]) 
			ifFalse: [
				self addChild: otherKid
			].
		].
	].
! !

!QBrainstormingBoard methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:08'!
defaultIsAtomic
	^ false! !

!QBrainstormingBoard methodsFor: 'accessing' stamp: 'bgf 7/26/2009 20:10'!
destinationExtent
	^ (self extent) max: (5@3).! !

!QBrainstormingBoard methodsFor: 'initialize' stamp: 'bgf 8/27/2009 19:43'!
initialize
	super initialize.

	logo := TSimpleRectangle new.
	logo material: self class logoMaterial.
	logo pickable: false; solid: false.
	logo shouldPersist: false.
	self addChild: logo.! !

!QBrainstormingBoard methodsFor: 'testing' stamp: 'bgf 7/28/2008 18:32'!
isEditableGroup
	"User-editing protocol; return true if this group's content should
	 be available for interactive reparenting (in and out.)"
	^ (self class == QBrainstormingBoard)! !

!QBrainstormingBoard methodsFor: 'accessing' stamp: 'bgf 8/9/2009 00:02'!
labelForBrowser
	^ (self objectName beginsWith: self class defaultObjectName) ifTrue: [
		'Brainstorming'
	] ifFalse: [
		'Brainstorming: ', (self objectName)
	].! !

!QBrainstormingBoard methodsFor: 'menu' stamp: 'bgf 8/31/2008 16:08'!
menuTitle
	^ 'Brainstorming'! !

!QBrainstormingBoard methodsFor: 'persist' stamp: 'bgf 8/15/2009 14:03'!
persistOn: pStream
	"Persist the receiver"

	super persistOn: pStream.

	self class == QBrainstormingBoard ifTrue:[
		pStream storeSequence: 'children' value: (self childrenForPersistence) default: #().
	].

! !

!QBrainstormingBoard methodsFor: 'initialize' stamp: 'bgf 8/15/2009 14:02'!
updateShape
	"Place the logo-thingy"
	| fraction basis half |
	super updateShape.
	fraction := 7.	
	basis := (self extent / fraction asFloat) minDimension.
	half := self extent / 2.0.
	logo extent: (basis @ basis);
	       translation: (half x negated + basis) @ (half y - basis)  @ 0.01.! !

QNoteBoard subclass: #QMultiBoard
	instanceVariableNames: 'tabs cursor tabNames'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QMultiBoard commentStamp: '<historical>' prior: 0!
A tabbed noteboard (tabbed in the sense used for a web browser - switchable named pages.)
Divides its attached content into named groups, displaying one at a time.
The others remain extant in the island but are removed from the visible scene.!

!QMultiBoard class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMultiBoard class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:01'!
board: mbRef newTab: tabIndex name: aString oldCursor: oldCursor

	^ QBlockAction
		undoBlock: [
			mbRef future removeTab: tabIndex+1.  
			mbRef future cursor: oldCursor ]
		redoBlock: [
			mbRef future addTab: tabIndex withLabel: aString.
			mbRef future cursor: (tabIndex + 1)  ]
		description: 'Add Tab'
		requiredFrames: { mbRef }! !

!QMultiBoard class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:01'!
board: boardRef removeTab: index label: label notes: capList

	^ QBlockAction
		undoBlock: [ :avatar|
			boardRef future restoreTab: index label: label notes: capList forAvatar: avatar ]
		redoBlock: [ 
			boardRef future removeTab: index ifLabeled: label]
		description: 'Remove Tab'
		requiredFrames: { boardRef }
! !

!QMultiBoard class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:01'!
board: boardRef tab: cursor direction: offset

	^QBlockAction 
		undoBlock: [ boardRef future moveTab: cursor+offset direction: offset negated ]
		redoBlock: [  boardRef future moveTab: cursor direction: offset  ]
		description: 'Move Tab'
		requiredFrames: { boardRef }! !

!QMultiBoard class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 10:44'!
defaultExtent
	^ 14.0 @ 7.0! !

!QMultiBoard class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 16:52'!
defaultMaterial
	^ (QMaterials flatFeltMaterial)
		uvScale: (5@2);
		yourself! !

!QMultiBoard class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 23:42'!
maxTabs
	^ 16! !

!QMultiBoard methodsFor: 'annotations' stamp: 'bgf 8/11/2009 15:43'!
addNote: aNote
	
	aNote ifNil: [ ^ self ].

	super addNote: aNote.

	"We only accept visible notes in this call, as our means to detect notes
	 that get added by the attachment manager at space-construction
	 (off-tab notes are invisible when saved.
	 If the note is visible, then it was visible at save time (so belongs on my current-tab at load),
	 or else we're back running live, and new attachments always come to the current tab."

	(cursor > 0 and: [ aNote visible ])  ifTrue: [
		((tabs at: cursor) includes: aNote) ifFalse: [
			(tabs at: cursor) add: aNote
		]].! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/5/2009 04:58'!
addTab
	"Add a tab at the end of the list."
	self addTab: (tabs size).! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/11/2010 14:27'!
addTab: index
	self addTab: index withLabel: 'Untitled'! !

!QMultiBoard methodsFor: 'menu' stamp: 'bgf 3/13/2010 08:52'!
addTab: tabNum forEvent: cEvent

	| oldCursor |
	oldCursor := self cursor.

	self addTab: tabNum.
	self cursor: tabNum + 1.

	cEvent avatar ifNotNil: [ : avatar |
		avatar recordAction: #board:newTab:name:oldCursor:
			    class: QMultiBoard
			    arguments: { self . tabNum . oldCursor . 'Untitled' }.
	].! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/11/2010 14:26'!
addTab: index withLabel: aString
	"Add a tab after the given index."
	tabs size < (self class maxTabs) ifTrue: [
		tabs add: (self newTab) afterIndex: index.
		tabNames add: aString afterIndex: index.
	] ifFalse: [
		QLogger logWarning: 'Cannot hold more than ', (self class maxTabs), ' tabs'.
	].
	self signal: #notesChanged.! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/20/2009 15:57'!
anyUser
	"We need a volunteer sometimes to launch or save enclosed content.
	 This is a placeholder under Croquet gets proper Principal-awareness in event dispatch."
	^ self root ifNotNil: [ 
		[ self root allUsers anyOne ] on: Error do: [nil].
	]! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/20/2009 15:22'!
checkHideItem: someItem forCursor: idx
	cursor ~= idx ifTrue: [
		someItem visible ifFalse: [
			self anyUser ifNotNilDo: [ : avatar | 	
				someItem showOrHide: false forPrincipal: avatar .
			]]].



		! !

!QMultiBoard methodsFor: 'menu' stamp: 'bgf 3/12/2010 14:57'!
checkRemoveTab: anEvent

	anEvent avatar ifNil: [ ^ self ].

	self numTabs > 1 ifFalse: [
		anEvent avatar signal: #alertMessage with: 'Cannot remove the last tab.'.
		^ self.
	].

	"With undo in place, no need to confirm."
	^ self removeCurrentTab: anEvent
	! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/20/2009 15:55'!
checkShowItem: someItem forCursor: idx
	"Latent reveal-starter for applications.
	 If the given item is still on this page and visible, start it."
	cursor = idx ifTrue: [
		someItem visible ifTrue: [
			self anyUser ifNotNilDo: [ : avatar | 	
				someItem showOrHide: true forPrincipal: avatar .
			]]].



		! !

!QMultiBoard methodsFor: 'menu' stamp: 'bgf 2/9/2010 23:47'!
clearCurrentTab: event
	event avatar deleteFrames: ((tabs at: cursor) select: [ : item | item notNil ]) event: event. 
	tabs at: cursor put: OrderedCollection new.

	self checkStaleNotes.
	self signal: #notesChanged.! !

!QMultiBoard methodsFor: 'persist' stamp: 'bgf 8/11/2009 15:43'!
constructFrom: pStream

	| pns |
	cursor := 0.
	super constructFrom: pStream.
	self extent: (pStream readPoint: 'extent' default: (self extent)).
	self numTabs: (pStream readNumber: 'nTabs' default: (self numTabs)).
	"Tab contents..."
	1 to: (self numTabs) do: [ :i |
		|items|
		items := pStream readSequence: ('tab-',i) default: #().
		items do: [ : item |
			item ifNotNil: [
				((tabs at: i) includes: item) ifFalse: [
					(tabs at: i) add: item
				]]]].

	pns := (pStream readStringSequence: 'tabNames' default: nil).
	pns ifNotNil: [
		pns doWithIndex: [ : pn : i | 
			tabNames size >= i ifTrue: [
				tabNames at: i put: pn
			]]].

	self cursor: (pStream readNumber: 'cursor' default: 1).
	self signal: #notesChanged.! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/5/2009 02:48'!
currentTabName
	^ tabNames at: cursor! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/2/2009 17:49'!
cursor
	"Which of my tabs is showing"
	^ cursor! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/20/2009 15:18'!
cursor:  aScalar
	"Which of my tabs is showing"
	| nc |
	nc := (aScalar min: (tabs size) max: 1).
	(cursor > 0 and:[ cursor ~= nc]) ifTrue: [ self hideTab: (cursor) ].
	cursor := nc.
	self showTab: (cursor).

	self signal: #notesChanged! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/20/2009 15:37'!
hideTab: tabNum
	"Suppress the content of the given tab."
	tabNum = cursor ifFalse: [
		QLogger logWarning: 'Hiding a tab that is not the current one. '.
		^ nil
	].

	(tabs at: cursor) do: [ : item |
		item visible ifTrue: [
			"If we don't tab back within 3 seconds, turn whatever it is off."
			(self future: 3000) checkHideItem: item forCursor: cursor.			
			item visible: false.
		]].! !

!QMultiBoard methodsFor: 'initialize' stamp: 'bgf 8/5/2009 05:00'!
initialize
	super initialize.

	tabs := OrderedCollection new.
	tabNames := OrderedCollection new.	"These two arrays are always parallel."

	self numTabs: 1.
	cursor := 0.
	self cursor: 1.! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/5/2009 02:17'!
labelForBrowser
	((self tabNameAt: cursor) = 'Untitled') ifTrue: [
		^ (self objectName beginsWith: self class defaultObjectName) ifTrue: [
			'Multi-Board'
		] ifFalse: [
			self objectName
		]	
	] ifFalse: [
		^ (self tabNameAt: cursor)
	]! !

!QMultiBoard methodsFor: 'menu' stamp: 'bgf 8/5/2009 18:14'!
menuTitle
	((self tabNameAt: cursor) = 'Untitled') ifTrue: [
		^ 'Untitled Tab'
	] ifFalse: [
		^ self tabNameAt: cursor
	]! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/12/2010 12:35'!
moveTab: index direction: offset

	"Swap the given tab with the one just before or after it in the list."
	| tmp |

	offset > 0 ifTrue: [
		(index < 1 or: [ index > (self numTabs-1) ]) ifTrue: [ 
			^ self ]].
	offset < 0 ifTrue: [
		(index < 2 or: [ index > self numTabs ]) ifTrue: [
			^ self ]].

	tmp := tabs at: index.
	tabs at: index put: (tabs at: index+offset).
	tabs at: index+offset put: tmp.

	tmp := tabNames at: index.
	tabNames at: index put: (tabNames at: index+offset).
	tabNames at: index+offset put: tmp.

	index = cursor ifTrue: [
		self cursor: index+offset
	].! !

!QMultiBoard methodsFor: 'drag-drop' stamp: 'bgf 8/18/2009 00:38'!
newPanelFor: delivery
	"Use fixed-displays to receive most content,
	 except that python documents require being surrounded in QDisplayPanels."
	(delivery document notNil and: [ delivery document isPythonDoc ]) ifTrue: [
		^ QDisplayPanel withSimpleFrame
	].

	^ QFixedDisplay new.
	! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/5/2009 05:00'!
newTab
	"A fresh new content-holder for my tabs collection.
	 The representation is not complex :-).
	 A tab is just a collection of (usally space-rooted) TFrames
	 that happen to be attached to me.
	 The attachments get assigned to a specific tab
	 simply because only one tab is in play at any time,
	 including when things are attached to the board. "

	^ OrderedCollection new! !

!QMultiBoard methodsFor: 'annotations' stamp: 'bgf 8/5/2009 05:02'!
notes
	"I override QFixedSurface with a notes array that is specific to the current tab.
	 This way, all the tools and queries for browsing topics, show the current set."
	^ (tabs at: cursor) asArray! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/4/2009 23:41'!
numTabs
	"How many tabs of content do I contain?"
	^ tabs size! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/4/2009 23:42'!
numTabs: aScalar
	"Set my number of tabs to the given capacity, adding and removing as needed;
	 contents are unceremoniously discarded."

	| nTabs |
	nTabs := aScalar min: (self class maxTabs) max: 1.
	
	[tabs size > nTabs] whileTrue: [
		self removeTab: (tabs size).
	].

	[tabs size < nTabs] whileTrue: [
		self addTab.
	].

! !

!QMultiBoard methodsFor: 'persist' stamp: 'bgf 8/4/2009 23:44'!
persistOn: pStream
	super persistOn: pStream.
	pStream storePoint: 'extent' value: self extent default: (self class defaultExtent).
	pStream storeNumber: 'nTabs' value: (self numTabs) default: 1.
	pStream storeNumber: 'cursor' value: (self cursor) default: 1.

	"tabs..."
	tabs doWithIndex: [ : items : i | 
		pStream storeSequence: ('tab-', i) value: items default: #() 
	].
	pStream storeStringSequence: 'tabNames' value: tabNames default: #().! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/13/2010 08:52'!
removeCurrentTab: event
	| tc cap |
	self numTabs > 1 ifFalse: [ ^ self ].

	tc := self cursor.
	(tabs at: tc) ifNotNil: [ : notes | 
		notes ifNotEmpty: [
			cap := event avatar captureFrames: notes.
			notes copy do: [ : note |
				note reallyDelete: event.
				noteCollection remove: note.
			]]].

	event avatar 
		    recordAction: #board:removeTab:label:notes:
		    class: QMultiBoard
		    arguments: { self . tc . (self tabNameAt: tc) . cap }.

	self removeTab: tc.! !

!QMultiBoard methodsFor: 'annotations' stamp: 'bgf 8/5/2009 05:03'!
removeNote: aNote
	"Attachment notification.   
	 Take this note away from the master allNotes list,
	 and remove it from any tab we might have recorded it in."
	super removeNote: aNote.
	tabs do: [ : tab |
		(tab includes: aNote) ifTrue: [ tab remove: aNote ].
	].
! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 1/29/2010 11:49'!
removeTab: index
	"Remove the given tab, unceremoniously destroying its attachments."
	(index > 0 and: [ index <= tabs size]) ifFalse: [
		QLogger logWarning: 'Cannot remove tab ', index,  '(currently ', tabs size, ')'.
		^ nil
	].
	QLogger logInfo: 'Removing tab ', index.

	"Scrub the attachments thoroughly so we're not carrying the freight of deferred-destroys."
	self hideTab: (index).
	(tabs at: index) clone  do: [ : note |
		note ifNotNil: [ 
			self root ifNotNilDo: [ : root | root detach: note ].
			note obliterateLater.
			noteCollection remove: note.
		]].
	tabs removeAt: index.
	tabNames removeAt: index.
	cursor > tabs size 
		ifTrue: [ cursor := tabs size ]
		ifFalse: [ 
			index > 1 ifTrue: [ cursor := cursor - 1].	
		].
	self checkStaleNotes.
	self showTab: (cursor).
	self signal: #notesChanged! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/12/2010 15:11'!
removeTab: index ifLabeled: label
	"Undo support"
	(index > 0 and: [ index <= tabs size]) ifFalse: [
		QLogger logWarning: 'Cannot remove tab ', index,  '(currently ', tabs size, ')'.
		^ nil ].
	(self tabNameAt: index) = label ifFalse: [ 
		QLogger logWarning: 'Cannot remove tab ', index,  '(label has changed)'.
		^ nil ].
	self removeTab: index.! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/12/2010 14:31'!
restoreTab: index label: label notes: frameCap

	self addTab: index-1 withLabel: label.
	self cursor: index.

! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 3/12/2010 14:41'!
restoreTab: index label: label notes: frameCap forAvatar: avatar
	"Undo support."
	self addTab: index-1 withLabel: label.
	self cursor: index.
	frameCap ifNotNil: [
		frameCap restoreForAvatar: avatar.
	]! !

!QMultiBoard methodsFor: 'private' stamp: 'bgf 8/26/2009 16:25'!
showTab: tabNum
	tabNum = cursor ifFalse: [
		QLogger logWarning: 'Showing a tab that is not the current one. '.
		^ nil
	].

	self root ifNotNilDo: [ : space |
		self checkStaleNotes.
		(tabs at: tabNum) do: [ : item |
			item visible ifFalse: [
				"If we don't tab away again soon, we'll turn on the apps."
				(self future: 1500) checkShowItem: item forCursor: tabNum.
				item visible: true.
			]]].! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/4/2009 23:43'!
tabNameAt: index
	^ tabNames at: index! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/5/2009 02:58'!
tabNameAt: index put: string
	tabNames at: index put: (string withBlanksTrimmed).
	self signal: #notesChanged.! !

!QMultiBoard methodsFor: 'accessing' stamp: 'bgf 8/4/2009 23:43'!
tabNames
	^ tabNames asArray! !

!QMultiBoard methodsFor: 'menu' stamp: 'bgf 3/12/2010 12:31'!
tradeTabBackward: index forEvent: anEvent
	"Swap the given tab with the one just before it in the list."

	self moveTab: index direction: -1! !

!QMultiBoard methodsFor: 'menu' stamp: 'bgf 3/12/2010 12:32'!
tradeTabForward: index forEvent: anEvent
	self moveTab: index direction: 1! !

!QNoteBoard class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QNoteBoard class methodsFor: 'properties' stamp: 'bgf 5/25/2010 20:26'!
addPropertyRows: propertyDialog forFrame: nbref

	self addNameRowIfNeeded: propertyDialog forFrame: nbref.
	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ : f | f get:#extent ];
			acceptBlock: [ :pt :f | f future extent: pt ];
			balloonHelp: 'Dimensions of the surface.';
		yourself
	). 

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Border Width';
			fetchBlock: [ :f | f get:#borderWidth ];
			acceptBlock: [ :pt :f | f future borderWidth: pt ];
			balloonHelp: 'Width of the border strip.';
		yourself
	). 

	super addPropertyRows: propertyDialog forFrame: nbref.

	TGeometry addTextureUVRows: propertyDialog forFrame: nbref.

	"A Noteboard Board is always note-collecting."
	propertyDialog removeRow: 'Noteboard (Manages Notes and Topics)'.! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 16:58'!
defaultBorderMaterial
	^ QMaterials mediumWoodMaterial! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 13:54'!
defaultBorderWidth
	^ 0.1! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 7/24/2008 13:08'!
defaultExtent
	^ 5@3! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 8/15/2009 14:20'!
defaultIsFromTemplate
	"Prefer not to bury these in the template so they show up in normal editing."
	^false! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 8/15/2009 14:19'!
defaultIsPlacementLocked
	"Eases interactions if these are unlikely to be unlocked."
	^true! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 8/15/2009 14:19'!
defaultIsTemplateExclusion
	"Don't hide thise in the template, so they show up in normal editing."
	^true! !

!QNoteBoard class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:08'!
defaultMaterial
	^ (QMaterials feltMaterial)
		uvScale: (1.5@1);
		diffuseColor: (QColors pureWhite);
		ambientColor: (VectorColor r: 1 g: 0.86 b: 0.8);
		emissiveColor: (VectorColor r: 0.1 g: 0.08 b: 0.05);
		yourself! !

!QNoteBoard methodsFor: 'editing' stamp: 'bgf 8/29/2009 13:07'!
allEditableMaterials
	^ surface allEditableMaterials! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/29/2009 13:06'!
borderMaterial
	^ surface borderMaterial! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/29/2009 13:06'!
borderMaterial: aMaterial
	surface borderMaterial: aMaterial! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/29/2009 13:04'!
borderWidth
	^ surface borderWidth! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/29/2009 13:04'!
borderWidth: aScalar
	"The border width is a dimension, like extent (not a factor or ratio).
	 It extends --outside-- the basic #extent of the bordered rectangle."

	surface borderWidth = aScalar ifFalse: [
		surface borderWidth: aScalar.
		self shapeChanged
	].! !

!QNoteBoard methodsFor: 'testing' stamp: 'bgf 2/18/2009 21:46'!
canEditMaterial
	^ true! !

!QNoteBoard methodsFor: 'persist' stamp: 'bgf 8/29/2009 13:12'!
constructFrom: pStream
	"Persist the receiver"
	| uvs uvo uva |
	super constructFrom: pStream.

	self extent: (pStream readPoint: 'extent' default: self extent).
	uvs := (pStream readPoint: 'uvScale' default: self uvScale).
	uvo := (pStream readPoint: 'uvOffset' default: self uvOffset).
	uva :=(pStream readNumber: 'uvAngle' default: self uvAngle).

	surface position: (pStream readVector3: 'surfacePosition' default: surface position).
	surface rotation: (pStream readQuaternion: 'surfaceRotation' default: surface rotation).
	surface material: (pStream readObject: 'surfaceMaterial' default: (self defaultMaterial)).
	surface borderMaterial: (pStream readObject: 'borderMaterial' default: (self borderMaterial)).
	surface borderWidth: (pStream readNumber: 'borderWidth' default: (self borderWidth)).
	"uv params need to be set after material"
	self uvScale: uvs.
	self uvOffset: uvo.
	self uvAngle: uva.

	"This can get lost if the material editor gets into play; restore it next-time in."
	surface material cullFace: false.

	self isPlacementLocked: (self isPlacementLocked).! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/15/2009 14:07'!
defaultMaterial
	"So the material editor gets the surface's main treatment."
	^ surface defaultMaterial! !

!QNoteBoard methodsFor: 'transform' stamp: 'bgf 7/26/2009 20:09'!
destinationExtent
	^ (self extent * 1.2) max: (5@3).! !

!QNoteBoard methodsFor: 'menu' stamp: 'bgf 7/27/2009 00:12'!
destinationStanceFor: avatar
	^ QGotoButton 
			stanceFor: avatar
			toView: self 
			inDestination: self
			inAngle: 30
			atDistance: ((self extent maxDimension) min: 40 max: (avatar avatarHeight/3.0))
! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
extent
	"Extent of the body rectangle."
	^ surface extent! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/1/2009 18:59'!
extent:  aPt
	"Extent of the surface rect"
	surface extent: aPt.
	self shapeChanged.! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/4/2009 10:25'!
height
	^ self extent y! !

!QNoteBoard methodsFor: 'initialize' stamp: 'bgf 8/29/2009 13:12'!
initialize
	super initialize.
	self isNoteCollector: true.

	surface := TBorderRectangle new.
	surface material: self class defaultMaterial.
	surface borderMaterial: self class defaultBorderMaterial.
	surface borderWidth: (self class defaultBorderWidth).
	surface extent: self class defaultExtent.
	self addChild: surface.

	resize := QCorner2DResizer forLuggage: self.
	(self pickable and: [ self isPlacementLocked not ])  ifTrue: [
		resize enable
	] ifFalse: [
		resize disable
	].
! !

!QNoteBoard methodsFor: 'testing' stamp: 'bgf 6/24/2009 15:20'!
isFlatRectangle
	"So drops can be clamped vertically."
	^ true! !

!QNoteBoard methodsFor: 'initialize' stamp: 'bgf 8/1/2009 18:51'!
isPlacementLocked: aBool
	super isPlacementLocked: aBool.
	"The resizer has to be suppressed when locked or non-pickable."
	resize ifNotNil: [
		((aBool not) & (self pickable)) ifTrue: [
			resize enable
		] ifFalse: [
			resize disable
		]].! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/2/2009 19:30'!
labelForBrowser
	^ (self objectName beginsWith: self class defaultObjectName) ifTrue: [
		'Noteboard'
	] ifFalse: [
		'Noteboard: ', (self objectName)
	].! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 6/30/2010 14:02'!
material
	^ surface ifNotNil: [ surface material ]! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/29/2009 14:07'!
material: m
	surface material: m! !

!QNoteBoard methodsFor: 'menu' stamp: 'bgf 8/31/2008 16:35'!
menuTitle
	^ 'Noteboard'! !

!QNoteBoard methodsFor: 'persist' stamp: 'bgf 8/29/2009 13:06'!
persistOn: pStream
	"Persist the receiver"

	super persistOn: pStream.

	"The basic geometry and appearance"
	pStream storePoint: 'extent' value: self extent default: (self class defaultExtent).
	pStream storePoint: 'uvScale' value: self uvScale default: 1@1.
	pStream storePoint: 'uvOffset' value: self uvOffset default: 0@0.
	pStream storeNumber: 'uvAngle' value: self uvAngle default: 0.

	"Just in case a tunneled prop editor is used to reposition the interior surface within child frames."
	pStream storeVector3: 'surfacePosition' value: surface position default: 0@0@0.
	pStream storeQuaternion: 'surfaceRotation' value: surface rotation default: Quaternion identity.
	pStream storeObject: 'surfaceMaterial' value: (surface material) default: (self class defaultMaterial).
	pStream storeObject: 'borderMaterial' 
			value: (self borderMaterial) 
			default: (self class defaultBorderMaterial).
	pStream storeNumber: 'borderWidth' value: self borderWidth default: self class defaultBorderWidth.

! !

!QNoteBoard methodsFor: 'menu' stamp: 'bgf 8/1/2009 17:52'!
photographerClass
	^ QDirectPhotoJob! !

!QNoteBoard methodsFor: 'hierarchy' stamp: 'bgf 8/1/2009 18:52'!
updateShape
	super updateShape.
	resize updateShape! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
uvAngle
	^ surface uvAngle! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
uvAngle: angle
	surface uvAngle: angle! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
uvOffset
	^ surface uvOffset! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
uvOffset: offset
	surface uvOffset: offset! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
uvScale 
	^ surface uvScale! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:05'!
uvScale: scale
	surface uvScale: scale! !

!QNoteBoard methodsFor: 'accessing' stamp: 'bgf 8/4/2009 10:25'!
width
	^ self extent x! !

QFixedSurface subclass: #QWall
	instanceVariableNames: 'mesh extent thickness'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QWall commentStamp: '<historical>' prior: 0!
A wall (a textured opaque TCube) that can be raised and lowered
by clicking on a footing control in the lower left corner.

Walls are fixed surfaces -- they accept dropped QFileCards 
and create QDisplayPanels displaying the 2-D content,
and QGates opening onto portals.

(The intention is to use the lowered state, in conjunction
with a smart floor, to allow rearranging the room.)

!

!QWall class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QWall class methodsFor: 'Inspector' stamp: 'bgf 1/26/2010 13:11'!
addPropertyRows: propertyDialog forFrame: wallRef

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ :frame  | frame get:#extent ];
			acceptBlock: [ :e :frame | frame future extent: e ];
			validateBlock: [ : pt | 
				(pt isNil or: [(pt x <= 0) | (pt y <= 0)]) ifTrue: [
					self inform: 'Extent must be positive.'.
					false
				] ifFalse: [
					true
				]];
			balloonHelp: 'Length and height of the wall.';
	 	yourself	).
	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'UV Scale';
			fetchBlock: [ :frame | frame get:#uvScale ];
			acceptBlock: [ :pt :frame | frame future uvScale: pt  ];
			balloonHelp: 'Scale/repetition of texture on the surface.';
	yourself). 
	super addPropertyRows: propertyDialog forFrame: wallRef.! !

!QWall class methodsFor: 'defaults' stamp: 'bgf 9/21/2008 16:00'!
defaultDoContextMenu
	"Fewer unintended menus this way.  (Means you need the inspector for editing.)"
	^ false! !

!QWall class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:24'!
defaultEventMask
	^ EventPointerDown! !

!QWall class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:13'!
defaultIsFromTemplate
	"Reduces xml volume since these are most always scenery."
	^ true! !

!QWall class methodsFor: 'defaults' stamp: 'bgf 6/5/2009 22:41'!
defaultIsPlacementLocked
	"Reduces xml volume since these are most always scenery."
	^ true! !

!QWall class methodsFor: 'defaults' stamp: 'bgf 4/9/2007 00:01'!
defaultThickness

	"The thickness of a wall."

	^ 0.08! !

!QWall methodsFor: 'testing' stamp: 'bgf 2/18/2009 21:49'!
canEditMaterial
	^ true! !

!QWall methodsFor: 'persist' stamp: 'bgf 3/3/2010 12:04'!
constructFrom: pStream
	| uvs |
	super constructFrom: pStream.

	self extent: (pStream readPoint: 'extent' default: (self extent)).
	uvs := (pStream readPoint: 'uvScale' default: (self uvScale)).
	mesh material: (pStream readObject: 'material' default: (mesh material)).
	self uvScale: uvs.


	"CONVERSION HACK for the Library - the terrace glass wall
	 must be made non-solid to allow motion to get through the wall
	 if obstacle-detection driving is in use.   
	 The texture SHA below is the gray+glass window+doors 
	 wall texture used in Library V1 between room and terrace."

	(mesh material texture notNil and: [
		mesh material texture sha = '8A3EWOHJSSUJE143Y0B8VPOBDZ931L0'])
	ifTrue: [
		self solid: false
	].

! !

!QWall methodsFor: 'accessing' stamp: 'bgf 7/20/2008 15:49'!
defaultMaterial
	"So the material editor edits the wall, not the buttons,etc."
	^ mesh ifNotNil: [ mesh defaultMaterial ]! !

!QWall methodsFor: 'accessing' stamp: 'bgf 10/19/2006 21:30'!
extent

	"The (non-collapsed) extent of the wall."
	^ extent ifNil: [
		(mesh extent x@mesh extent y)
	].
! !

!QWall methodsFor: 'accessing' stamp: 'bgf 9/27/2009 20:51'!
extent: aPoint

	"Change the (non-collapsed) extent of the wall."

	(extent = aPoint) ifTrue: [ ^ self ].	"Cut circular recursions."
	extent := aPoint.
	mesh extentX: (aPoint x) y:(aPoint y) z:thickness;
		translationX: 0 y:(aPoint y/2) z: 0.

! !

!QWall methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:39'!
initialize
	super initialize.
	thickness := (self class defaultThickness).
	mesh := TCube new.
	self addChild: mesh.
	mesh material color: (Color white).
! !

!QWall methodsFor: 'accessing' stamp: 'bgf 3/27/2009 09:48'!
isComponent
	^ true! !

!QWall methodsFor: 'accessing' stamp: 'bgf 9/27/2009 20:46'!
isPlacementLocked: bool
	super isPlacementLocked: bool.
! !

!QWall methodsFor: 'accessing' stamp: 'bgf 9/12/2006 22:51'!
isRaised
	"Is the wall at full height?"

	^ mesh extent y >= (extent y - Float epsilon).
! !

!QWall methodsFor: 'accessing' stamp: 'bgf 4/8/2007 14:25'!
material
	^ mesh material! !

!QWall methodsFor: 'accessing' stamp: 'bgf 4/8/2007 14:25'!
material: aTMaterial
	mesh material: aTMaterial
! !

!QWall methodsFor: 'accessing' stamp: 'bgf 4/8/2008 23:11'!
menuTitle
	^ 'Wall'! !

!QWall methodsFor: 'persist' stamp: 'bgf 3/3/2010 12:05'!
persistOn: pStream
	||
	super persistOn: pStream.

	pStream storePoint: 'extent' value: (self extent).
	pStream storePoint: 'uvScale' value: (mesh uvScale) default: 1@1.
	pStream storeObject: 'material' value: (mesh material).


! !

!QWall methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:59'!
privateScale: aFactor

	nomScale := nomScale * aFactor.
	self extent: (self extent * aFactor).! !

!QWall methodsFor: 'events' stamp: 'bgf 6/28/2010 11:10'!
replaceWithRectangle: event

	"Private content-preparation utility;
	 replace this QWall with a simple rectangle of the same appearance,
	 that is inert and pick-transparent.
	 Richard has a habit of using walls for fixed-placement texture decorations;
	 this simplifies those items."

	| parent trans ext r |

	parent := self parent.
	trans := self translation.
	ext := self extent x @ self extent y.
	trans := trans + (0@(self extent y * 0.5) @ (thickness * -0.52)).

	r := (TRectangle new).
	parent addChild: r.
	r material: (self material copy).
	r extent: ext.
	r translation: trans.
	r orientation: (self orientation).
	r objectName: self objectName.

	r material cullFace: false.
	"Since the normal use I encounter is walls made for graphics in front of other walls..."
	r pickable: false; solid: false; doContextMenu: false.
	
	self future reallyDelete: event.! !

!QWall methodsFor: 'accessing' stamp: 'bgf 12/6/2006 13:47'!
texture
	^ (mesh material) texture

! !

!QWall methodsFor: 'accessing' stamp: 'bgf 12/6/2006 13:47'!
texture: aTexture 
	"Modify the wall to use the given texture"
	(mesh material) texture: aTexture.

! !

!QWall methodsFor: 'drag-drop' stamp: 'bgf 3/3/2010 12:05'!
translationForPanel: aPanel droppedAt: aTSelection

	"The position for a new display panel.
	 Wall extends this to keep the top of the panel within the height of the wall.
	 We want the panel top to be visible because it is the motion grip."

	| result upperY fudge lowerY |

	fudge := 0.8.	" A little leeway; ceilings often come a bit lower than wall height."
	
	result := super translationForPanel: aPanel droppedAt: aTSelection.

	"Keep it in view, giving priority to keeping it below the top of the wall."
	lowerY := self globalTransform translation y.
	(result y - (aPanel height / 2)) < (lowerY + fudge) ifTrue: [
		result := result x @ (lowerY + fudge + (aPanel height / 2.0)) @ result z.
	].
	upperY := (self globalTransform translation y) + (self extent y).
	((result y + (aPanel height / 2.0)) > (upperY - fudge)) ifTrue: [
		result := result x @ (upperY - fudge- (aPanel height /2.0)) @ result z.
	].
	^ result! !

!QWall methodsFor: 'accessing' stamp: 'bgf 12/6/2006 13:48'!
uvScale
	^ mesh uvScale
! !

!QWall methodsFor: 'accessing' stamp: 'bgf 12/6/2006 13:47'!
uvScale: aScale
	mesh uvScale: aScale
! !

QSeat subclass: #QStool
	instanceVariableNames: 'pad post base shadow triangle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QStool commentStamp: '<historical>' prior: 0!
A simple stool.  Upon which you can sit.!

!QStool class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStool class methodsFor: 'defaults' stamp: 'bgf 8/8/2009 13:37'!
defaultSeatMaterial
	^ QMaterials redFabricMaterial! !

!QStool class methodsFor: 'defaults' stamp: 'bgf 8/7/2009 16:00'!
defaultSeatOffset
	^ (0@0.1@-0.2).! !

!QStool methodsFor: 'persist' stamp: 'bgf 8/8/2009 13:36'!
canEditMaterial
	^ true! !

!QStool methodsFor: 'persist' stamp: 'bgf 8/8/2009 13:50'!
constructFrom: pStream
	super constructFrom: pStream.
	pad material: (pStream readObject: 'seatMaterial' default: (pad defaultMaterial)).
	self isPlacementLocked: (self isPlacementLocked).! !

!QStool methodsFor: 'initialize' stamp: 'bgf 8/6/2009 20:51'!
defaultIsAtomic
	^ true! !

!QStool methodsFor: 'persist' stamp: 'bgf 8/8/2009 13:36'!
defaultMaterial
	"So the material editor gets the seat."
	^ pad defaultMaterial! !

!QStool methodsFor: 'initialize' stamp: 'bgf 8/29/2009 11:37'!
initialize
	super initialize.

	pad := TCylinder new.
	pad objectName: 'Stool'.
	pad material: (self class defaultSeatMaterial).
	self addChild: pad.

	triangle := ((TTriangle new)
				objectName: 'direction';
				v1: (Vector3 x: (0) y: 0 z: -0.48)
				v2: (Vector3 x: (-0.2) y: 0 z: 0.36)
				v3: (Vector3 x: (0.2) y: 0 z: 0.36);
				material: ((QMaterials darkGreyMaterial) opacity: 0.3; yourself);
				pickable: false;
				checkAlpha;
				yourself).
	self addChild: triangle.
				
	post := TCylinder new.
	post objectName: 'post'.
	post material: (QMaterials darkGreyMaterial).
	self addChild: post.
	
	base := TCylinder new.
	base objectName: 'base'.
	base material: (QMaterials lightWoodMaterial).
	self addChild: base.

	shadow := QMaterials newRadialShadow.
	shadow material diffuseColor alpha: 0.85.
	shadow addRotationAroundX:-90.
	self addChild: shadow.

	self shapeChanged.
! !

!QStool methodsFor: 'testing' stamp: 'bgf 8/9/2009 00:16'!
isComponent
	^ self shouldPersist! !

!QStool methodsFor: 'testing' stamp: 'bgf 8/9/2009 00:15'!
isGroundBound
	^ true! !

!QStool methodsFor: 'persist' stamp: 'bgf 8/8/2009 13:39'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeObject: 'seatMaterial' value: (pad material) default: (self class defaultSeatMaterial).
! !

!QStool methodsFor: 'hierarchy' stamp: 'bgf 2/12/2010 14:18'!
updateShape
	"This is a rigid shape; scaling works fine to resize it."
	| scl |
	scl := (self scaleMemo).
	post radius: 0.05* scl ; 
		height: 1.8 * scl.
	base radius: 0.38 * scl ;
		height: 0.1 * scl ; 		
		slices: 12.
	pad radius: 0.65 * scl ; 	
		height: 0.2 * scl ; 
		translation: (0@(post height)@0); slices: 16.
	triangle translation: (0@(post height + (pad height *1.025)) @ 0).
	shadow extent: (1.07@1.07) * scl; 
			translation: (0@0.01@0).! !

!QStool methodsFor: 'hierarchy' stamp: 'bgf 8/8/2009 14:20'!
warpThresholdAdjustment
	"Slight increase so these aren't too grabby."
	^ 2! !

TCube subclass: #QFloor
	instanceVariableNames: 'subfloor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QFloor commentStamp: '<historical>' prior: 0!
A cube with drag/drop handling ... dropping an image card
onto the floor creates a QDisplayFrame facing the dropper,
floating above the floor at the drop point.

A floor can be polished by setting isReflecting:true;
this is expensive to render, but gives the floor a mirror finish.

Floors can be overlaid cheaply with simple fixed carpet areas,
using #addRug: aRectangle material:aTMaterial.

Work to do: the floor and wall together will present editable spaces,
probably by having the floor implement drawing-like behavior.!

!QFloor class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFloor class methodsFor: 'inspector' stamp: 'sn 3/24/2013 13:24'!
addPropertyRows: propertyDialog forFrame: floorref

	super addPropertyRows: propertyDialog forFrame: floorref.
	propertyDialog addRow: (
		(QBooleanPropertyRow new) 
			titleString: 'Mirror Finish'translateMe;
			fetchBlock: [ : f | f get: #isReflecting ];
			acceptBlock: [ : b : f | f future isReflecting: b ];
			balloonHelp: 'Use a reflecting surface.\(May impair performance.)'translateMe;
		yourself).

! !

!QFloor class methodsFor: 'defaults' stamp: 'bgf 9/21/2008 16:00'!
defaultDoContextMenu
	"Fewer unintended menus this way.  (Means you need the inspector for editing.)"
	^ false! !

!QFloor class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:12'!
defaultIsFromTemplate
	"Most always scenery"
	^ true! !

!QFloor class methodsFor: 'defaults' stamp: 'bgf 6/5/2009 22:41'!
defaultIsPlacementLocked
	"Reduces xml volume since these are most always scenery."
	^ true! !

!QFloor class methodsFor: 'accessing' stamp: 'bgf 8/30/2006 18:20'!
defaultThickness

	"The thickness of a floor."

	^ 0.2! !

!QFloor methodsFor: 'drag-drop' stamp: 'bgf 3/13/2010 08:52'!
acceptDisplayDrop: delivery avatar: avatar selection: aTXSel
	| panel cap cardCap |
	(QDisplaySheet canAcceptDrop: delivery) ifFalse: [
		^ false
	].
	"Accept a drop of 2D content and display it in a display stand."
	panel := QDisplayStand withSimpleFrame.
	panel setupForAvatar: avatar atFloorPoint: (aTXSel globalPoint).
	panel displayCard: delivery fromUser: avatar.
	self root addChild: panel.

	cap := avatar captureFrames: { panel }.
	cardCap := avatar captureFrames: { delivery duplicate }. 
	avatar recordAction: #surface:newDisplay:capture:cardCapture:
		    class: QFixedSurface
		    arguments: { self . panel . cap . cardCap }.

	^ false
! !

!QFloor methodsFor: 'drag-drop' stamp: 'bgf 3/13/2010 08:52'!
acceptLinkDrop: delivery avatar: avatar selection: aTXSel
	"If the delivery is a postcard, open a gate to its destination."
	| dropPt gate cap |
	((delivery isKindOf:QFileCard) and: [delivery kind = #postcardLink]) ifFalse: [
		^ false
	].
	dropPt := aTXSel globalPoint.
	gate := QGate new.
	gate label: (delivery contentLink shortDisplayString);
		 isFromTemplate: false;
		 rotationAroundY: (avatar yaw);
		"Adjusting to ground won't be correct in all cases, but is necessary in most."
		 translation: (Vector3 x: dropPt x y: (self root groundHeightAt: dropPt) z: dropPt z). 
	self root addChild: gate.
	gate establishLink: delivery contentLink forUser: avatar.
	cap := avatar captureFrames: { gate }.
	avatar recordAction: #surface:newDisplay:capture:cardCapture:
		    class: QFixedSurface
		    arguments: { self . gate . cap . nil }.
	^ true.! !

!QFloor methodsFor: 'initialize' stamp: 'bgf 9/7/2007 13:30'!
addQuarterRugs

	"Make a rug in each outer quadrant, 
	 using four different hues derived from the base material;
	 used to give our generic forums some terms of reference for calling people around."

	| ext cent mat center |
	ext := ( self extentX @ self extentZ) * 0.333.
	center := ( self extentX @ self extentZ ) * 0.333 - (4@4).
	
	mat := self material copy.
	mat color: (mat diffuseColor asColor darker mixed: 0.9 with: (Color r:0.05 g:0.7 b:0.5)).
	cent := center.
	(self addRugAt: cent extent: ext material: mat)
		uvScale: self uvScale / 4.0;	
		objectName: 'southeast-green-rug' copy. 

	mat := self material copy.
	mat color: (mat diffuseColor asColor mixed: 0.85 with: (Color brown lighter)).
	cent := center * (-1@1).
	(self addRugAt: cent extent: ext material: mat)
		uvScale: self uvScale / 4.0;
		objectName: 'southwest-brown-rug' copy. 

	mat := self material copy.
	mat color: (mat diffuseColor asColor darker mixed: 0.9 with: (Color r:1 g:0.15 b:0.1)).
	cent := center * (1@-1).
	(self addRugAt: cent extent: ext material: mat)
		uvScale: self uvScale / 4.0;
		objectName: 'northeast-red-rug' copy. 


	mat := self material copy.
	mat color: (mat diffuseColor asColor darker mixed: 0.85 with: (Color r:0.1 g:0.4 b:0.9)).
	cent := center*-1.
	(self addRugAt: cent extent: ext material: mat)
		uvScale: self uvScale / 4.0;
		objectName: 'northwest-blue-rug' copy.! !

!QFloor methodsFor: 'initialize' stamp: 'bgf 5/31/2009 10:54'!
addRugAt: pos extent: ext  material: aTMaterial
	"overlay a region (Rectangle-defined) of the floor with a rug (textured rectangle)
	using the given material.   The overlay keeps all the drag-drop capacity of the floor itself."
	| rug |
	rug := TRectangle new.
	rug material: aTMaterial.
	rug extent: ext.
	rug divisions: (4@4).
	rug translation: (pos x)@(self extentY/2.0 + 0.02)@(pos y).
	rug pitch: -90.
	rug doContextMenu: false.
	self addChild: rug.
	^ rug! !

!QFloor methodsFor: 'persist' stamp: 'bgf 9/1/2007 19:24'!
children
	"for persistence"
	^(self frameChildren ifNil:[#()]) asArray! !

!QFloor methodsFor: 'persist' stamp: 'bgf 9/1/2007 19:24'!
children: aSequence
	"for persistence"
	self children do:[:each| self removeChild: each].
	aSequence do:[:each| self addChild: each].! !

!QFloor methodsFor: 'accessing' stamp: 'bgf 9/4/2007 09:38'!
childrenForBrowser
	^ self children copyWithout: subfloor.! !

!QFloor methodsFor: 'persist' stamp: 'bgf 2/27/2009 18:15'!
constructFrom: pStream

	super constructFrom:pStream.

	self class == QFloor ifTrue:[
		self children: (pStream readSequence:'children' default: self children).
		"Clean up after older implementations that seemed to leave unused reflection-portals around."
		self children 
			select: [ : f | f isKindOf: TPortal ]
			thenDo: [ :f |
				self removeChild: f.
				f destroy.
			]].
	self isReflecting: (pStream readBool: 'isReflecting' default: (self isReflecting)).! !

!QFloor methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:14'!
consumeDrop: delivery avatar: avatar selection: aTXSel
	"Is it a postcard link ? if so, open a gate."
	(self acceptLinkDrop:delivery avatar:avatar selection:aTXSel) ifTrue: [
		^ false "Dont consume the card."
	].

	"Can I display this content?  If so, put it in a display panel."

	(QDisplaySheet canAcceptDrop: delivery)  ifTrue: [
		^ self acceptDisplayDrop: delivery avatar: avatar selection: aTXSel.
	].

	^ false! !

!QFloor methodsFor: 'initialize' stamp: 'bgf 6/2/2009 19:52'!
extent: aPoint
	"Initialize the floor to the given x@z extent."

	| e |
	e := (aPoint isKindOf:Vector3) ifFalse: [
		(aPoint x)@(self class defaultThickness)@(aPoint y).
	] ifTrue: [
		aPoint.
	].
	super extent:e.

! !

!QFloor methodsFor: 'initialize' stamp: 'bgf 4/7/2007 21:55'!
extent: aPoint texture: aTexture repeat: aUVScale

	"Initialize the wall to the given x@y extent,
       using the given texture at the given uvScale."
	self extent: aPoint.
	self material color: (Color gray:0.5).
	self material texture: aTexture.
	self uvScale: aUVScale.
! !

!QFloor methodsFor: 'accessing' stamp: 'bgf 9/4/2007 09:38'!
isAtomic
	^ false! !

!QFloor methodsFor: 'accessing' stamp: 'bgf 6/1/2009 16:09'!
isComponent
	^ true! !

!QFloor methodsFor: 'drag-drop' stamp: 'bgf 9/6/2006 14:01'!
isDropTarget: payload

	(payload class includesBehavior: QWall) ifTrue: [
		^ true.
	].

	^ QDisplaySheet canAcceptDrop: payload.! !

!QFloor methodsFor: 'accessing' stamp: 'bgf 11/12/2009 15:49'!
isEditableGroup
	"Although not a TGroup, I permit children for historical reasons (rugs in conf rooms)"
	^ true! !

!QFloor methodsFor: 'accessing' stamp: 'bgf 3/31/2007 21:11'!
isReflecting
	"the floor is reflecting when it has a portal subfloor."
	^ subfloor notNil! !

!QFloor methodsFor: 'accessing' stamp: 'bgf 3/31/2007 21:49'!
isReflecting: reflect

	"Set the surface, reflective or plain; 
	 this is done by embedding a mirror portal within the surface of the cube.
	 The 'subfloor' portal only exists when we are set to be reflecting,
	 so all we do here is suit the state to the argument."

	(reflect and: [subfloor notNil]) ifTrue: [ ^self].

	(reflect not and: [subfloor notNil]) ifTrue: [
		"Take away reflection"
		subfloor destroy.
		subfloor := nil.
		self material opacity: 1.0.
		self checkAlpha.
		^self.
	].

	reflect ifTrue: [
		"Add reflection"
		subfloor := TPortal new.
		subfloor rotationAroundX: -90.
		subfloor translation: (0@(self extent y/2.0 - 0.01)@0);
				 extent: (self extent x@self extent z)+0.1.
		self addChild: subfloor.
	
		"Become slightly translucent so the reflection shows through;
	 	the portal is embedded just under my top surface."
		self material opacity: 0.88.
		self checkAlpha.
		^ self
	].
! !

!QFloor methodsFor: 'persist' stamp: 'bgf 2/26/2009 18:53'!
persistOn: pStream

	super persistOn:pStream.
	pStream storeBool: 'isReflecting' value: (self isReflecting) default: false.
	(self class == QFloor) ifTrue: [
		pStream storeSequence: 'children' value: (self childrenForBrowser) default: #().
	].! !

!QFloor methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 17:33'!
redoDrop: frameCapture card: cardCap fromUser: avatar
	"Undo-redo support.   Repeat the creation of a drop-display."
	| disp cd |
	disp := (frameCapture restoreForAvatar: avatar) first.
	cardCap ifNotNil: [
		cd := (cardCap restoreForAvatar: avatar) first.
		disp displayCard: cd fromUser: avatar.
		cd obliterateLater.
	].
	avatar selectFrames: { disp }.
! !

!QFloor methodsFor: 'render' stamp: 'das 6/30/2008 14:21'!
setupAlpha: ogl

	self root addAlphaObject: self 
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance:  100000 "arbitrarily large number"
		subDistance: (frameParent childSubdistance:self)
		parent: frameParent
		transparency: ogl transparency
		hilite: ogl forceHilite.
! !

TEmbeddedApp subclass: #QEmbeddedApp
	instanceVariableNames: 'document displayPanel'
	classVariableNames: 'ClickStartPreview'
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QEmbeddedApp commentStamp: '<historical>' prior: 0!
(c) 2007 Qwaq, Inc.

Adds to TEmbeddedApp
	- multi-user pointer tracking
	- use of preview images as content for non-running apps.!

!QEmbeddedApp class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 5/24/2008 13:22'!
asTonedGray: aForm

	"Modify the form to a washed-out (grayed) version of itself,
	 with just a suggestion of the original color.
	 
	 May replace the form bits in-place, or may produce a new form.

	 This is used wherever we're showing an app preview.
	 This is a variation of the ParticipantUI class method;
	 this version keeps the image lighter so (mostly-white) applications don't 	
	 look overly darkened, just a little gray and muted."

	| result base |
	aForm ifNotNil: [
		"This is a bit costly, but less so than taking a gray scale of a deep Form.
	       The fast-and cheap as8BitColorForm uses a predefined palette
	       so overall, this is the fastest path to sloppy grayscale."
		base := aForm as8BitColorForm asGrayScale.
		"Flatten the contrast"
		base colors: (base colors collect: [:c | 
						| v | 
						v := c brightness.
						(Color r: (0.2 + (0.75 * v)) 
							   g: (0.2 + (0.75 * v)) 
							   b: (0.2 + (0.75 * v)))]).
		result := aForm asFormOfDepth: 32.
		"The above ensures we have a copy before going inside it to modify"
		result := result copyBits: base at: 0@0 translucent: 0.85.
	].
	^ result! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 11/27/2009 15:51'!
emptyFormForThisIsland
	"The nasty green-arrow we show before there's a thum... this is the on-island preview for an app.
	 To save space we make one per island."

	^ Processor activeIsland 
			at: #AppEmptyIconForm
			ifAbsentPut: [ 
				(Form extent: 1@1 depth: 32)
						form fillColor: (Color transparent);
						yourself ].
			


	! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 3/5/2009 17:12'!
fileCardTexture
	^ QTextures find: 'Content/Qwaq/Textures/Cards2/', (self fileCardBacking ifNil:[ 'AppCard.png'])! !

!QEmbeddedApp class methodsFor: 'class initialization' stamp: 'bgf 1/23/2009 15:59'!
initialize

	"QEmbeddedApp initialize."
	[
		ClickStartPreview := ImageReadWriter formFromFileNamed:
			(FileDirectory default pathFromURI: 'icons/Qwaq/StartBackground.gif').
	] on: Error do: [
		Transcript cr; show: 'Update SVN and evaluate QEmbeddedApp initialize'.
	].! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 11/27/2009 15:47'!
makeIconTexture: hashedName

	"A TFileTexture holding the file-card icon derived from the app-preview.
	 The card hashedName (which is also given to the embedded-app as its hashedName) 
	 is used directly as the texture SHA
	 (giving a sort of casual dynamic binding - the preview date
	 can vary depending on the last save,  but the SHA does not;
	 the sha is based on a the file-card full-Summary (kind, full name or url.)
	 The hashed name is also the basis for the server-storage of the preview.

	 The fallback thumbnail image when there is no preview
 	 is simply transparent, so the card-itself is all that is seen."

	| thumbForm previewName |

	"The in-island thumb - nothing, until filled with a preview, so only the base card image is visible."
	thumbForm := (self emptyFormForThisIsland).
	previewName := (hashedName
						ifNil: [nil] 
						ifNotNil: [QEmbeddedApp previewPath, hashedName]).
	^ ((TFileTexture new) 
		initializeSHA: hashedName form: thumbForm mipmap: true shrinkFit: true extension: nil;
		fileName: previewName;
		extent: 4@3;		"For a typical aspect ratio, when I appear on file-cards."
		bThumb: true;
		yourself).


	! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 11/27/2009 15:42'!
makePreviewTexture: hashedName

	"A TFileTexture holding the preview for an app of the given hashed name.
	 The hashed name is used directly as the SHA for local lookup
	 (giving a sort of casual dynamic binding - the preview content
	 might vary depending on the last save,  but the SHA does not.)
	 The hashed name is also the basis for the service-provider storage of the preview.

	 Because we use this SHA-matching, the fallback image when there is no preview
 	 is the in-island thumbnail (my ClickStartPreview, a small right-triangle 'start' image)
	 which is coarse and unattractive, but hopefully not often seen."

	| thumbForm previewName |

	"The in-island thumb - a ghastly start arrow..."
	thumbForm := (self previewFormForThisIsland).
	previewName := (hashedName
						ifNil: [nil] 
						ifNotNil: [ QEmbeddedApp previewPath, hashedName]).
	^ ((TFileTexture new) 
		initializeSHA: hashedName form: thumbForm mipmap: true shrinkFit: true extension: nil;
		fileName: previewName;
		bThumb: true;
		yourself).


	! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 11/27/2009 16:11'!
previewFormForThisIsland
	"The nasty green-arrow we show before there's a preview... 
	 this is the on-island preview for an app.
	 To save space we make only one per island.   64x64 textures aren't huge but they add up."

	^ Processor activeIsland 
			at: #AppPreviewForm
			ifAbsentPut: [ 
				ClickStartPreview 
				ifNotNilDo: [: f | f deepCopy]  
				ifNil: [ (Form extent: 1@1 depth: 32)
						form fillColor: (Color gray:0.95);
						yourself ].
			].


	! !

!QEmbeddedApp class methodsFor: 'defaults' stamp: 'bgf 1/8/2008 11:52'!
previewPath

	"The document-manager path for storing server-side preview images for applications.
	 This is a per-forum service-manager file path (relative, not absolute) 
	 that is nested inside 'resources/' so that 'convertToLazyTextures' is kept happy."

	^ 'resources/screens/'! !

!QEmbeddedApp class methodsFor: 'previews' stamp: 'bgf 5/17/2010 11:09'!
savePreviewForApp: eApp hashedName: sha inSession: aQCSession

	"Generate and upload a standard preview image
	 for the given TEmbeddableApp.  
	 Return the Form containing that preview.
	 (The form upload is backgrounded; the preview is returned locally first."

	| byteStream data form |

	QLogger logInfo: 'Saving app preview: ', (eApp).  "A bit noisy, for dirty-bit testing."

	sha ifNil: [ ^ nil ].
	eApp displayForm isNil ifTrue: [^ nil ].
	eApp app island isValidFarRef ifFalse: [ ^nil ].
	aQCSession owner serviceMgr ifNil: [ ^ nil ].
	aQCSession owner signal: #statusMessage with: 'Saving application preview'.

	"GIF readwriter really dislikes encoding empty forms."
	eApp display area <= 0 ifTrue: [ ^ nil ].

	"Generate the image - GIF is usually much smaller than PNG for typical application snaps. "
	byteStream := (ByteArray new: 32768) writeStream.
	form := eApp displayForm contentsOfArea:((0@0) extent: (eApp display extent)).
	(form extent area > (512*256)) ifTrue: [
		"Arbitrary threshold."
		form := form magnify: form boundingBox by: 0.5 smoothing: 2.
	].
	form border: (form boundingBox) width:2 fillColor: (Color gray:0.25).
	form := self asTonedGray: form.
	
	GIFReadWriter putForm: form onStream: byteStream.
	data := byteStream contents.

	"Install it as texture data, and share it on the server, which informs peers."
	aQCSession updateAppPreview: form data: data sha: sha island: (eApp app island).

	^ form! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'jcg 5/28/2008 18:55'!
appExtent: aPoint
	"Answer the application extent for this display"
	super appExtent: aPoint.
	"If in a panel, force the panel to adjust me to the new size / aspect ratio. "
	displayPanel ifNotNil:[displayPanel shapeChanged].
! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'jcg 5/28/2008 16:19'!
appTitle: aString
	"Set the application title"
	displayPanel ifNotNil:[displayPanel title: aString].! !

!QEmbeddedApp methodsFor: 'protocols' stamp: 'bgf 3/19/2010 14:28'!
canSendControlToSave
	"Protocol for VNC OpenOffice documents.
	 Return true if a Control-S should be sent to open office when it is desired to save this document.
	 Needed because some document types cause OO to open unwanted save-name dialogs
	 when a Ctl-S is received."

	self document ifNotNil: [ : doc | doc docType ifNotNil: [ : dt | ^ dt isMicrosoftXML not ] ].
	
	"Assume  yes  unless the document type is recognized as an exclusion."
	^ true! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'bgf 12/21/2008 18:09'!
displayLabelForBrowser
	"Most apps appear inside a display sheet that defines its title
	 using the content file-card to describe the application and document"
	(self parent isKindOf: QDisplaySheet) ifTrue:[
		"Some apps change the display panel title if they have one, so consult it first."
		(self parent parent isKindOf: QDisplayPanel) ifTrue:[
			^ self parent parent title.
		].
		^ self parent title
	].
	^ self objectName.! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'jcg 5/28/2008 16:29'!
displayPanel
	"The display panel this app is contained in"
	^displayPanel! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'jcg 5/28/2008 16:29'!
displayPanel: aQDisplayPanel
	"The display panel this app is contained in"
	displayPanel := aQDisplayPanel! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'bgf 2/27/2009 16:41'!
document
	"The document for which this app was launched."
	^ document! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'bgf 2/27/2009 16:41'!
document: anAbstractDocument
	"The document for which this app was launched."
	(document == anAbstractDocument) ifFalse: [
		document := anAbstractDocument.
		self signal: #documentChanged with: document
	].! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'bgf 5/27/2010 09:32'!
documentName
	"The sharedName of the document for which this app was launched, if any."
	^ document ifNotNil: [ document sharedName ifNil: [ document shortName ] ]! !

!QEmbeddedApp methodsFor: 'protocols' stamp: 'bgf 3/5/2009 15:01'!
fileCardBacking
	"Return the name of one of the Qwaq/Cards/<something.png> 
	 file-card icon graphics.   E.g. 'DocumentCard.png'   
	 A custom -Embeddable- app may specify its card-backing name by overriding its class method.
	 A custom -Embedded- app may specify its card with an instance-side method overriding this one.

	 (Must return the name of a backing that is in the product package)"

	^	(self class appNamed: self appName) ifNotNilDo: [ : appClass | appClass fileCardBacking ]! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'das 1/5/2009 10:45'!
hashedName: aString

	"When a hashedName is assigned,
	 it allows us to identify the preview texture to use for the running app.
	 For many of our apps, the preview is always just the run-arrow;
	 only for some do we use a per-app generated preview.
	 (Those correspond to lazy-started VNC applications.)"

	QLogger panelLog.
	super hashedName: aString.
	self appIcon: (self class makePreviewTexture: 
					(self usePreviewInAppIcon ifTrue: [aString] ifFalse: [nil])).

! !

!QEmbeddedApp methodsFor: 'protocols' stamp: 'bgf 3/12/2009 10:17'!
markCleanFor: userLogin utcSeconds: utcSeconds
	
	"If I'm dirty, note this as document last-change update."
	self isDirty ifTrue: [
		self root ifNotNilDo: [ : space |
			self document ifNotNilDo: [ : doc |
				doc mtime who: userLogin  what: 'Save' utc:  utcSeconds
			]]].

	super markCleanFor: userLogin utcSeconds: utcSeconds! !

!QEmbeddedApp methodsFor: 'events' stamp: 'bgf 12/21/2007 23:31'!
pointerDown: event
	"This can arrive only when the app is not running (normally)
	 although we check to be sure.   Provide some feedback;
	 the inherited code starts the application."

	self isRunning ifFalse: [
		QSoundEffects play: #switch for: self.
	].
	super pointerDown: event.! !

!QEmbeddedApp methodsFor: 'events' stamp: 'ar 4/20/2007 00:07'!
pointerEnter: event
	"Handle a pointer enter event"
	QArrowCursor setArrowCursor: event frame: self.
	super pointerEnter: event.! !

!QEmbeddedApp methodsFor: 'events' stamp: 'bgf 11/13/2008 13:11'!
pointerLeave: event
	"Handle a pointer over event"

	event avatar ifNotNil: [
		self isRunning ifFalse: [
			event avatar clearFloatMessageForEvent: event. 
		]].
	super pointerLeave: event
! !

!QEmbeddedApp methodsFor: 'events' stamp: 'sn 3/4/2013 21:22'!
pointerOver: event
	"Handle a pointer over event"
	event avatar ifNotNil: [
		QArrowCursor setArrowCursor: event frame: self.
		self isRunning ifFalse: [
			event avatar showFloatMessage: 'Click to start 'translateMe, (self shortName)  forEvent: event. 
		]].
	super pointerOver: event.! !

!QEmbeddedApp methodsFor: 'protocols' stamp: 'hrs 9/23/2009 14:50'!
recordAction: aString userID: userID
	"Record the specified action by userID, in the text chat log."
	| avatar sheet |
	displayPanel ifNotNil: [sheet := displayPanel sheet] ifNil: [
		frameParent ifNil:[^nil].
		(frameParent isKindOf: QDisplaySheet) ifFalse: [^nil].
		sheet := frameParent.
	].
	
	self root ifNotNilDo: [:root |
		avatar := root find: [:f | (f isKindOf: TAvatarReplica) and: [f sessionID = userID]].
		sheet recordAction: aString
			avatar: (avatar ifNotEmpty: [avatar anyOne] ifEmpty: []).
		].! !

!QEmbeddedApp methodsFor: 'events' stamp: 'bgf 7/19/2009 14:11'!
removeUser: user selection: selection

	super removeUser: user selection: selection.
	selection userInfo ifNotNil:["remove the frame that is in this slot"
		selection userInfo delete. 
		selection userInfo: nil.
	].! !

!QEmbeddedApp methodsFor: 'protocols' stamp: 'bgf 3/13/2009 18:17'!
shortName
	"This is a fallback tag used to label app views in the early stages.
	 E.g the startup says 'Starting <this> ..."

	^ document
		ifNotNil: [ document kindSummary ]
		ifNil: [  super shortName ].! !

!QEmbeddedApp methodsFor: 'protocols' stamp: 'hrs 1/15/2009 21:42'!
startRunning: userID

	QLogger panelLog.
	self recordAction: 'started' userID: userID.
	(running not and: [userID notNil]) ifTrue: [
		QActivityCapture action: #StartApplication 
						 details: (self shortName) 
					       forSession: userID
	].
	super startRunning: userID.! !

!QEmbeddedApp methodsFor: 'accessing' stamp: 'bgf 1/24/2008 23:10'!
usePreviewInAppIcon

	"Should this app instance use the hashedName as the basis
	 for a texture file for its app icon (its texture when not running) ?   
	 Currently, all our apps are wired to take previews in #saveSharedContent,
	 and so this is generally true."

	^ true! !

QEmbeddedApp subclass: #QEmbeddedMpegPlayer
	instanceVariableNames: 'playState startTime stopTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-MovieApp'!

!QEmbeddedMpegPlayer commentStamp: '<historical>' prior: 0!
An embedded movie player.!

!QEmbeddedMpegPlayer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEmbeddedMpegPlayer class methodsFor: 'instance creation' stamp: 'bgf 3/5/2009 14:35'!
fileCardBacking
	"Stub to allow this class to fit where the tembeddable app class is expected."
	^ 'MovieCard.png'! !

!QEmbeddedMpegPlayer class methodsFor: 'class initialization' stamp: 'ar 5/26/2007 20:14'!
initialize
	"self initialize"
	TEmbeddedApp registerApp: self.! !

!QEmbeddedMpegPlayer class methodsFor: 'instance creation' stamp: 'bgf 12/3/2008 16:44'!
newForDocument: aSharedDoc extent: aPixelExtent

	"Generator for activating a shared-document MPG viewer;
	I am the application class for QSharedDocuments that are typed as mpg files."

	^self new document: aSharedDoc! !

!QEmbeddedMpegPlayer methodsFor: 'persist' stamp: 'bgf 2/27/2009 16:39'!
constructFrom: pStream
	"Restore only the document"
	super constructFrom: pStream.
	self document: (pStream readObject: 'document' default: document).
	self startTime: (pStream readNumber: 'startTime' default: startTime).
	self stopTime: (pStream readNumber: 'stopTime' default: stopTime).
	self playState: (pStream readString: 'playState' default: playState) asSymbol.
! !

!QEmbeddedMpegPlayer methodsFor: 'initialize' stamp: 'bgf 3/9/2009 02:11'!
initialize
	"Initialize the receiver"
	super initialize.
	self name: #TweakProjectWorld extent: 320@240 data: #(makeWorld: QMoviePlayer).
	playState := #paused.
	startTime := stopTime := self nowTime.
! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'bgf 2/27/2009 16:40'!
localFileUri
	"Movies are always kept locally in the cache, or are bundled."
	^document asLocalFileUri! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:26'!
nowTime
	"Shorthand"
	^Processor activeIsland time! !

!QEmbeddedMpegPlayer methodsFor: 'persist' stamp: 'bgf 2/27/2009 16:39'!
persistOn: pStream
	"Persist only the document"
	super persistOn: pStream.
	pStream storeObject: 'document' value: document.
	pStream storeNumber: 'startTime' value: startTime.
	pStream storeNumber: 'stopTime' value: stopTime.
	pStream storeString: 'playState' value: playState.
! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:01'!
playState
	"Answer the current play state of the mpeg player"
	^playState! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 21:05'!
playState: aSymbol
	"Set the current play state of the mpeg player"
	playState := aSymbol.
	self signal: #playStateChanged with: aSymbol.! !

!QEmbeddedMpegPlayer methodsFor: 'render' stamp: 'ar 5/29/2007 10:28'!
render: ogl
	playState = #playing ifTrue:[ogl resetIdle].
	^super render: ogl! !

!QEmbeddedMpegPlayer methodsFor: 'render' stamp: 'ar 5/29/2007 10:31'!
renderAlpha: ogl
	playState = #playing ifTrue:[ogl resetIdle].
	^super renderAlpha: ogl! !

!QEmbeddedMpegPlayer methodsFor: 'playing' stamp: 'ar 5/26/2007 20:34'!
resumePlaying
	"Resume playing at the current position"
	playState == #playing ifFalse:[
		self startPlayingAt: self nowTime - (stopTime - startTime).
	].! !

!QEmbeddedMpegPlayer methodsFor: 'playing' stamp: 'ar 5/26/2007 20:56'!
rewind
	"Rewind; e.g., set start and stop time to the same values and halt"
	startTime := stopTime := self nowTime.
	self signal: #playStateChanged.! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'bgf 12/20/2007 11:34'!
shortName
	^ 'Player'! !

!QEmbeddedMpegPlayer methodsFor: 'playing' stamp: 'ar 5/26/2007 19:26'!
startPlaying
	"Start playing"
	self startPlayingAt: self nowTime.! !

!QEmbeddedMpegPlayer methodsFor: 'playing' stamp: 'ar 5/26/2007 21:16'!
startPlayingAt: time
	"Start playing"
	startTime := time.
	self playState: #playing.! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:20'!
startTime
	"Answer the island time when the last start command was issued"
	^startTime! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 21:21'!
startTime: aNumber
	"Set the island time when the last start command was issued"
	startTime := aNumber! !

!QEmbeddedMpegPlayer methodsFor: 'playing' stamp: 'ar 6/2/2007 22:58'!
stopPlaying
	"Start playing"
	playState == #playing ifTrue:[
		self stopPlayingAt: self nowTime.
	].! !

!QEmbeddedMpegPlayer methodsFor: 'playing' stamp: 'ar 5/26/2007 21:16'!
stopPlayingAt: time
	"Start playing"
	stopTime := time.
	self playState: #paused.! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:20'!
stopTime
	"Answer the island time when the last stop command was issued"
	^stopTime! !

!QEmbeddedMpegPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 21:22'!
stopTime: aNumber
	"Set the island time when the last stop command was issued"
	stopTime := aNumber! !

QEmbeddedApp subclass: #QEmbeddedSlidePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-SlideShow'!

!QEmbeddedSlidePlayer commentStamp: '<historical>' prior: 0!
Embedded-App class for in-island presentation of the slide-show player;
creates a TweakProjectWorld holding a QSlideShowPlayer as world.

Satisfies the protocols for document-based launch;
this is the registered application for slide-show documents.!

!QEmbeddedSlidePlayer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEmbeddedSlidePlayer class methodsFor: 'protocol' stamp: 'bgf 3/5/2009 14:32'!
fileCardBacking
	"Stub to allow this class to fit where the tembeddable app class is expected,
	 as the embeddableAppClass of a QSlideShowDocument."
	^'SlideShowCard.png'! !

!QEmbeddedSlidePlayer class methodsFor: 'protocol' stamp: 'bgf 12/10/2008 15:59'!
newForDocument: ssDoc extent: anExtent
	"Protocol for launching in the panel/filecard system"
	^ (self new)	
		document: ssDoc
		yourself.! !

!QEmbeddedSlidePlayer methodsFor: 'accessing' stamp: 'bgf 4/1/2010 13:53'!
document: slideShowDoc
	super document: slideShowDoc.
	slideShowDoc ifNotNil: [
		"Updating the desired extent here prevents create+resize flashes during launch."
		self appExtent: (QSlideShowPlayer totalExtentForExtent: (document maxDimensions)).

		slideShowDoc alwaysLooping ifTrue: [
			slideShowDoc startPlaying
		]].
	! !

!QEmbeddedSlidePlayer methodsFor: 'hierarchy' stamp: 'bgf 1/31/2009 19:06'!
fileTextures
	"So that various things like inspector texture-locate can work properly."
	^ document
		ifNotNil: [ (document textureArray, (super fileTextures)) asSet asArray ]
		ifNil: [ super fileTextures ].! !

!QEmbeddedSlidePlayer methodsFor: 'initialize' stamp: 'bgf 3/9/2009 02:11'!
initialize
	"Initialize the receiver.... The initial app extent matches the default extent of slide-show-player."
	super initialize.
	self name: #TweakProjectWorld 
		extent: ((256 @ (256 + (QSlideShowPlayer controlHeightForExtent: 256@256))) )
		data: #(makeWorld: QSlideShowPlayer).

	self objectName: 'Slide Player'.
! !

!QEmbeddedSlidePlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 14:15'!
resize: extent forSHA: sha

	"If the indicated sha is indeed the image our document wants displayed,
	 carry out the resizing for it.  We confirm that the sha whose size is being sent
	 is still the current sha, to avoid dueling resizes if two start requesting conflicting updates
	 (a safety measure to prevent runaway future/signal/future loops;
	 since the harness-side handling of appExtent changes is pretty intensive and run-scripted.)"

	(sha = document currentImageSHA) ifTrue: [
		self appExtent = extent ifFalse: [
			self appExtent: extent
		] ifTrue: [
			document synchronize.
		].
	].! !

!QEmbeddedSlidePlayer methodsFor: 'accessing' stamp: 'bgf 12/14/2008 12:40'!
shortName
	^ 'Slide Show'! !

QEmbeddedApp subclass: #QEmbeddedTextApp
	instanceVariableNames: 'text markIndex pointIndex scrollOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-TextApp'!

!QEmbeddedTextApp commentStamp: '<historical>' prior: 0!
QEmbeddedTextApp is a a simple replicated text editor model. It contains the text representation, the current selection and scroll offset.!

!QEmbeddedTextApp class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEmbeddedTextApp class methodsFor: 'accessing' stamp: 'bgf 3/5/2009 14:32'!
fileCardBacking
	"Stub to allow this class to fit where the tembeddable app class is expected,
	 as the embeddableAppClass of a QTextDocument."
	^ 'TextCard.png'! !

!QEmbeddedTextApp class methodsFor: 'instance creation' stamp: 'bgf 12/3/2008 17:27'!
newForDocument: txtDocument extent: anExtent

	"Instantiate the embeddeded application class for this document;
	 we refer this to the EmbeddableApp class registered in #documentTypes."

	"Size in pixels for a text editor in my current display."
	| extent |
	extent := (anExtent x > anExtent y)  
				ifTrue: [
					512@(512 * (anExtent y / anExtent x)) rounded
				] ifFalse: [
					(512 * anExtent x/anExtent y) rounded @ 512
				].
	^ self new extent: extent document: txtDocument .! !

!QEmbeddedTextApp methodsFor: 'editing' stamp: 'bgf 2/27/2009 16:35'!
accept
	"Accept current text."
	document text: text asString.

		! !

!QEmbeddedTextApp methodsFor: 'editing' stamp: 'bgf 3/11/2009 23:03'!
acceptForUser: aUserID
	"Accept current text."
	document text: text asString.

	"If the avatar and its current time can be identified,
	 mark the document as having been changed by that person (now)"
	self isDirty ifTrue: [
		self root ifNotNilDo: [ : space | 
			(space avatarForSession: aUserID) ifNotNilDo: [ :avatar|
				document mtime 	
					who: avatar userLogin 
					what: 'Accept' 
					utc: (avatar lastAliveUTC).
			]]]! !

!QEmbeddedTextApp methodsFor: 'editing' stamp: 'ar 9/17/2008 18:14'!
backspace
	"Replace the current selection in the receiver with aText"
	markIndex = pointIndex ifTrue:[self selectFrom: markIndex to: markIndex-1].
	self replaceSelectionWith: ''.! !

!QEmbeddedTextApp methodsFor: 'persist' stamp: 'ar 4/19/2007 18:31'!
constructFrom: pStream
	"Reconstruct the receiver"
	super constructFrom: pStream.
	text := (pStream readString: 'text' default: '') asText.
	markIndex := pStream readNumber: 'markIndex' default: 1.
	markIndex := pStream readNumber: 'pointIndex' default: 1.
	scrollOffset := pStream readPoint: 'scrollOffset' default: 0@0.! !

!QEmbeddedTextApp methodsFor: 'initialize' stamp: 'bgf 2/27/2009 16:35'!
extent: aPoint document: doc
	"Initialize the receiver"
	self appExtent: aPoint.
	self document: doc.
	text := doc text asText.! !

!QEmbeddedTextApp methodsFor: 'initialize' stamp: 'bgf 3/9/2009 02:11'!
initialize
	"Initialize the receiver"
	super initialize.
	appName := #TweakProjectWorld.
	appData := #(makeWorld: QTextAppEditor).
	text := Text new.
	markIndex := pointIndex := 1.
	scrollOffset := 0@0.! !

!QEmbeddedTextApp methodsFor: 'testing' stamp: 'ar 12/30/2007 10:39'!
isUserOver: userID
	"Answer whether the user with the given ID is currently mousing over the app or not"
	currentUser ifNil:[^false].
	^currentUser users includesKey: userID.! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/20/2007 00:00'!
markIndex
	"Answer the mark index, e.g., the origin of the selection"
	^markIndex! !

!QEmbeddedTextApp methodsFor: 'persist' stamp: 'ar 4/19/2007 18:29'!
persistOn: pStream
	"Persist the receiver"
	super persistOn: pStream.
	pStream storeString: 'text' value: self string default: ''.
	pStream storeNumber: 'markIndex' value: markIndex default: 1.
	pStream storeNumber: 'pointIndex' value: pointIndex default: 1.
	pStream storePoint: 'scrollOffset' value: scrollOffset default: 0@0.! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/20/2007 00:00'!
pointIndex
	"Answer the point index, e.g., the current end point of the selection"
	^pointIndex! !

!QEmbeddedTextApp methodsFor: 'events' stamp: 'bgf 2/12/2009 15:14'!
pointerLeave: event
	"This is a workaround  for now because QTextDocument triggers a major update cycle when we change its text. If not for this we could just set the text after each text change or even share the text with the document."
	self acceptForUser: (event avatar ifNotNilDo: [ : av | av  userLogin ]).
	^super pointerLeave: event! !

!QEmbeddedTextApp methodsFor: 'editing' stamp: 'ar 4/19/2007 17:02'!
replaceSelectionWith: aText
	"Replace the current selection in the receiver with aText"
	| start stop |
	(aText isEmpty and:[markIndex = pointIndex]) ifTrue:[^self].
	start := self selectionStart.
	stop := self selectionStop-1.
	aText isString ifTrue:[
		text replaceFrom: start to: stop with: (Text string: aText emphasis: #()).
	] ifFalse:[
		text replaceFrom: start to: stop with: aText.
	].
	markIndex := pointIndex := self selectionStart + aText size.
	self signal: #updateText with: {
		#replaceSelection.
		aText asString.
	}.! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/20/2007 00:00'!
scrollOffset
	"Answer the scroll offset for the text"
	^scrollOffset! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/20/2007 00:01'!
scrollOffset: aPoint
	"Change the scroll offset for the text"
	scrollOffset := aPoint.
	self signal: #updateText with: {#scrollOffset. aPoint}.! !

!QEmbeddedTextApp methodsFor: 'editing' stamp: 'ar 4/25/2007 15:33'!
selectFrom: selStart to: selEnd
	"Select the region between selStart and selEnd"
	selStart ifNotNil:[markIndex := selStart].
	selEnd ifNotNil:[pointIndex := selEnd].
	"Ensure current selection is valid"
	markIndex := markIndex min: text size+1 max: 1.
	pointIndex := pointIndex min: text size+1 max: 1.
	self signal: #updateText with: {
		#updateSelection.
		markIndex.
		pointIndex.
	}.! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/19/2007 17:00'!
selectionStart
	"Answer the start of the current selection"
	^markIndex min: pointIndex! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/19/2007 17:00'!
selectionStop
	"Answer the end of the current selection"
	^markIndex max: pointIndex! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'bgf 12/20/2007 11:33'!
shortName
	^'Text Editor'! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/20/2007 00:01'!
string
	"Answer the string representation of the text"
	^text string! !

!QEmbeddedTextApp methodsFor: 'accessing' stamp: 'ar 4/19/2007 16:17'!
text
	"Answer the replicated text"
	^text! !

TConic subclass: #QTree
	instanceVariableNames: 'base splitRange heightRange angleRange baseAngleRange rotationRange branchesRange scaleRange randomSeed constructDepth renderDepth random parameters barkTexture leafTexture leafExtent leafRotation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QTree commentStamp: 'das 11/26/2007 11:55' prior: 0!
QTree is a recursively defined tree with lots of knobs to control it. It is designed to be used primarily for procedural testing of rendering performance, but it can also be used as decoration. The controls are:
#constructDepth: how deeply recursive is the tree defined
#randomSeed: seed used for random
#angleMin:max: when a new branch is added, this determines the angle from the trunk.
#baseAngleMin:max: the first branch may have a separate angle from others (pine trees)
#branchesMin:max: #branches per split
#heightMin:max: the scaling height of successive splits
#rotationMin:max: rotation around the axis of the branch or trunk
#splitMin:max: where along the branch or trunk we split

!

!QTree class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 08:49'!
angleMin: mn max: mx

	angleRange := QRange new min:mn max:mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:52'!
angleRange

	^ angleRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:57'!
angleRange: ar

	angleRange := ar.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 13:05'!
barkTexture

	^ barkTexture.! !

!QTree methodsFor: 'accessing' stamp: 'bgf 10/19/2008 22:32'!
barkTexture: textureFileName

	barkTexture := QTextures find: textureFileName.
! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 11:22'!
baseAngleMin: mn max: mx

	baseAngleRange := QRange new min:mn max:mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 11:22'!
baseAngleRange

	^ baseAngleRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:58'!
baseAngleRange: bar

	baseAngleRange := bar.! !

!QTree methodsFor: 'construct' stamp: 'das 12/3/2007 11:17'!
branch

	| splitHeight branchCount branch rand |
	rand := parameters random.
	splitHeight :=  (parameters splitRange select: rand).
	base := TConic new.
	base globalTransform: self globalTransform.
	base baseRadius: baseRadius.
	base topRadius: baseRadius * (1-splitHeight).
	splitHeight := height * splitHeight.
	base height: splitHeight*1.07.
	base capTop: false.
	base capBase: false.
	base material texture: parameters barkTexture.
	base uvScale: parameters uvScale.

	branchCount := (parameters branchesRange select: rand)roundTo:1.
	branchCount timesRepeat:[
		branch := QTree new.
		branch material texture: parameters barkTexture.
		branch uvScale: parameters uvScale.
		branch parameters: parameters.
		branch addRotationAroundY:(parameters rotationRange select: rand).
		branch translation: 0@splitHeight@0.
		self frameChildren ifNil:[
			branch height: self height-splitHeight.
			branch addRotationAroundX:(parameters baseAngleRange select: rand).
			branch baseRadius: base topRadius.
		]ifNotNil:[
			branch height: 
				(self height-splitHeight)*(parameters heightRange select: rand).
			branch addRotationAroundX:(parameters angleRange select: rand).
			branch baseRadius: base topRadius*(parameters scaleRange select: rand).
		].
		self addChild: branch.
	].
		
	
! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 08:50'!
branchesMin: mn max: mx

	branchesRange := QRange new min:mn max:mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:53'!
branchesRange

	^ branchesRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:58'!
branchesRange: br

	branchesRange := br.! !

!QTree methodsFor: 'accessing' stamp: 'bgf 2/18/2009 21:50'!
canEditMaterial
	^ false! !

!QTree methodsFor: 'construct' stamp: 'bgf 8/27/2009 17:03'!
construct

	"reinit random"
	| leaf |
	random := Random new.
	random seed: randomSeed.
	self removeAll.
	constructDepth timesRepeat:[ self constructPass.].
	self allLeaves do:[:f |
		leaf := TSimpleRectangle new.
		leaf extent: parameters leafExtent.
		leaf translation:0@f height@0.
		leaf material texture: leafTexture.
		leaf material cullFace: false.
		leaf material diffuseColor:1@1@1@0.99.
		leaf checkAlpha.
		leaf addRotationAroundZ:leafRotation.
		f addChild: leaf.
	].! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:47'!
constructDepth

	^ constructDepth.
! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:33'!
constructDepth: dp

	constructDepth := dp.
	renderDepth := dp.! !

!QTree methodsFor: 'persist' stamp: 'das 12/3/2007 11:53'!
constructFrom: pStream
	"Persist the receiver"
	super constructFrom: pStream.
	self randomSeed: (pStream readNumber: 'randomSeed' default: self randomSeed).
	self constructDepth: (pStream readNumber: 'constructDepth' default: self constructDepth).
	self splitRange: (pStream readObject: 'splitRange' default: self splitRange type: QRange).
	self heightRange: (pStream readObject: 'heightRange' default: self heightRange type: QRange).
	self angleRange: (pStream readObject: 'angleRange' default: self angleRange type: QRange).
	self baseAngleRange: (pStream readObject: 'baseAngleRange' default: self baseAngleRange type: QRange).
	self rotationRange: (pStream readObject: 'rotationRange' default: self rotationRange type: QRange).
	self branchesRange: (pStream readObject: 'branchesRange' default: self branchesRange type: QRange).
	self scaleRange: (pStream readObject: 'scaleRange' default: self scaleRange type: QRange).
	barkTexture:= (pStream readObject: 'barkTexture' default: self barkTexture type: TFileTexture).
	leafTexture:= (pStream readObject: 'leafTexture' default: self leafTexture type: TFileTexture).
	self leafExtent: (pStream readPoint: 'leafExtent' default: self leafExtent).
	self leafRotation: (pStream readNumber: 'leafRotation' default: self leafRotation).
	self construct.
! !

!QTree methodsFor: 'construct' stamp: 'das 11/26/2007 11:20'!
constructPass

	frameChildren ifNil:[
		self branch.
	] ifNotNil:[
		self allLeaves do:[:b | b branch].
	].
! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:36'!
heightMin: mn max: mx

	heightRange := QRange new min: mn max: mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:53'!
heightRange

	^ heightRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:58'!
heightRange: hr

	heightRange := hr.! !

!QTree methodsFor: 'initialize' stamp: 'das 12/3/2007 11:52'!
initialize

	super initialize.
	randomSeed := 0.5.
	self topRadius: 0.
	self baseRadius: 1.0.
	self height: 20.
	self constructDepth: 5.
	self splitMin: 0.2 max: 0.5.
	self heightMin: 0.75 max: 1.25.
	self branchesMin: 2 max: 3.
	self baseAngleMin:10 max: 30.
	self angleMin: 30 max: 45.
	self rotationMin: -180 max: 180.
	self scaleMin: 0.5 max: 0.9.
	self leafExtent: 4@4.
	self capTop: false.
	self capBase: false.
	self leafRotation:0.
	parameters := self.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 15:10'!
leafExtent

	^ leafExtent.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 15:10'!
leafExtent: le

	leafExtent := le.! !

!QTree methodsFor: 'accessing' stamp: 'das 12/3/2007 11:22'!
leafRotation

	^ leafRotation.! !

!QTree methodsFor: 'accessing' stamp: 'das 12/3/2007 11:22'!
leafRotation: lr

	leafRotation:= lr! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 14:30'!
leafTexture

	^ leafTexture.! !

!QTree methodsFor: 'accessing' stamp: 'bgf 10/19/2008 23:53'!
leafTexture: textureFileName

	leafTexture := QTextures find: textureFileName.
! !

!QTree methodsFor: 'samples' stamp: 'das 12/3/2007 11:26'!
makeEuropean: seed

	self constructDepth: 4.
	self randomSeed: seed.
	self heightMin:0.5 max:1.2.
	self splitMin:0.20 max:0.5.
	self branchesMin:2 max: 4.
	self uvScale: 3@3.
	leafRotation := 45.
	self barkTexture:'Content/Qwaq/Models/Plants/trees_bark_dark.png'.
	self leafTexture:'Content/Qwaq/Models/Plants/Branch_Euro_8.png'.! !

!QTree methodsFor: 'samples' stamp: 'das 11/26/2007 14:56'!
makeOak1

	self constructDepth: 5.
	self randomSeed: 0.6.
	self heightMin:0.5 max:1.2.
	self splitMin:0.20 max:0.5.
	self branchesMin:2 max: 4.
	self barkTexture:'Content/DungeonDemo/textures/BarkRoughV2.BMP'.
	self leafTexture:'Content/DungeonDemo/textures/BranchOakLow.BMP'.! !

!QTree methodsFor: 'samples' stamp: 'bgf 1/28/2008 13:46'!
makeOak: seed

	self constructDepth: 4.
	self randomSeed: seed.
	self heightMin:0.5 max:1.2.
	self splitMin:0.20 max:0.5.
	self branchesMin:2 max: 4.
	self uvScale: 3@3.
	self leafRotation: -90.
	self barkTexture:'Content/Qwaq/Models/Plants/trees_bark_dark.png'.
	self leafTexture:'Content/Qwaq/Models/Plants/leaf.png'.	
! !

!QTree methodsFor: 'samples' stamp: 'das 11/26/2007 15:23'!
makePine1

	self baseRadius: 0.75.
	self constructDepth: 4.
	self randomSeed: 0.2.
	self angleMin:45 max:60.
	self baseAngleMin: 0 max:10.
	self branchesMin:2 max: 3.
	self heightMin:0.5 max:0.75.
	self scaleMin:0.5 max: 0.75.
	self splitMin:0.30 max:0.5.
	self leafExtent: 2@2.
	self barkTexture:'Content/DungeonDemo/textures/BarkRoughV2.BMP'.
	self leafTexture:'Content/DungeonDemo/textures/BranchFirLow.BMP'.! !

!QTree methodsFor: 'samples' stamp: 'das 12/3/2007 11:36'!
makePine: seed

	self baseRadius: 0.75.
	self constructDepth: 5.
	self randomSeed:seed.
	self height: 40.
	self angleMin:60 max:85.
	self baseAngleMin: 0 max:10.
	self branchesMin:2 max: 3.
	self heightMin:0.5 max:0.75.
	self scaleMin:0.5 max: 0.75.
	self splitMin:0.40 max:0.6.
	self rotationMin: 90 max: 180.
	self leafExtent: 4@4.
	self uvScale: 3@3.
	self leafRotation: -90.

	self barkTexture:'Content/DungeonDemo/textures/BarkRoughV2.BMP'.
	self leafTexture:'Content/DungeonDemo/textures/BranchFirLow.BMP'.! !

!QTree methodsFor: 'construct' stamp: 'das 11/26/2007 09:51'!
parameters: param

	parameters:= param.! !

!QTree methodsFor: 'persist' stamp: 'das 12/3/2007 11:52'!
persistOn: pStream

	super persistOn: pStream.
	pStream storeNumber: 'randomSeed' value: (self randomSeed).
	pStream storeNumber: 'constructDepth' value: (self constructDepth).
	pStream storeTyped: 'splitRange' value: splitRange.
	pStream storeTyped: 'heightRange' value: heightRange.
	pStream storeTyped: 'angleRange' value: angleRange.
	pStream storeTyped: 'baseAngleRange' value: baseAngleRange.
	pStream storeTyped: 'rotationRange' value: rotationRange.
	pStream storeTyped: 'branchesRange' value: branchesRange.
	pStream storeTyped: 'scaleRange' value: scaleRange.
	barkTexture ifNotNil:[pStream storeTyped: 'barkTexture' value: barkTexture].
	leafTexture ifNotNil:[pStream storeTyped: 'leafTexture' value: leafTexture].
	pStream storePoint: 'leafExtent' value: self leafExtent default: self leafExtent.
	pStream storeNumber: 'leafRotation' value: (self leafRotation).
! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:56'!
random

	^ random.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:47'!
randomSeed

	^ randomSeed.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:39'!
randomSeed: seed

	randomSeed := seed.! !

!QTree methodsFor: 'render' stamp: 'das 11/26/2007 11:11'!
render: ogl

	base ifNotNil:[^base render: ogl].
	super render: ogl.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:34'!
renderDepth: rd

	renderDepth := rd.! !

!QTree methodsFor: 'construct' stamp: 'das 11/26/2007 10:20'!
reset

	base removeAll.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 08:50'!
rotationMin: mn max: mx

	rotationRange := QRange new min:mn max:mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:54'!
rotationRange

	^ rotationRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:59'!
rotationRange: rr

	rotationRange := rr.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 11:32'!
scaleMin: mn max: mx

	scaleRange := QRange new min:mn max:mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 11:32'!
scaleRange

	^ scaleRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:59'!
scaleRange: sr

	scaleRange := sr.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 08:49'!
splitMin: mn max: mx

	splitRange := QRange new min:mn max:mx.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/26/2007 09:54'!
splitRange

	^ splitRange.! !

!QTree methodsFor: 'accessing' stamp: 'das 11/30/2007 13:59'!
splitRange: sr

	splitRange := sr.! !

CPlayer subclass: #QMovieControls
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-MovieApp'!

!QMovieControls commentStamp: '<historical>' prior: 0!
Movie controls for the player. Still early in their inception.!

QMovieControls defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="rewindButton" toSet="rewindButton:" changeEvent="rewindButtonChanged">rewindButton</field>
	<field toGet="playButton" toSet="playButton:" changeEvent="playButtonChanged">playButton</field>
	<field toGet="pauseButton" toSet="pauseButton:" changeEvent="pauseButtonChanged">pauseButton</field>
</fields>'!

!QMovieControls class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMovieControls methodsFor: 'initialize' stamp: 'ar 5/26/2007 20:43'!
initialize
	super initialize.
	self define: #rewindButton as: CPushButton new.
	self define: #pauseButton as: CPushButton new.
	self define: #playButton as: CPushButton new.! !

!QMovieControls methodsFor: 'events' stamp: 'ar 5/26/2007 21:07'!
onPauseButtonFire
	<on: fire in: pauseButton>
	self signal: #pauseRequest.! !

!QMovieControls methodsFor: 'events' stamp: 'ar 5/26/2007 21:07'!
onPlayButtonFire
	<on: fire in: playButton>
	self signal: #playRequest.! !

!QMovieControls methodsFor: 'events' stamp: 'ar 5/26/2007 21:07'!
onRewindButtonFire
	<on: fire in: rewindButton>
	self signal: #rewindRequest.! !

!QMovieControls methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:42'!
pauseButton
	"Answer the pauseButton of the receiver"
	<bewareOf: #pauseButtonChanged>
	^self propertyValueAt: #pauseButton! !

!QMovieControls methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:42'!
pauseButton: aValue
	"Modify the receiver's pauseButton"
	^self propertyValueAt: #pauseButton put: aValue with: #pauseButtonChanged! !

!QMovieControls methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:42'!
playButton
	"Answer the playButton of the receiver"
	<bewareOf: #playButtonChanged>
	^self propertyValueAt: #playButton! !

!QMovieControls methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:42'!
playButton: aValue
	"Modify the receiver's playButton"
	^self propertyValueAt: #playButton put: aValue with: #playButtonChanged! !

!QMovieControls methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:42'!
rewindButton
	"Answer the rewindButton of the receiver"
	<bewareOf: #rewindButtonChanged>
	^self propertyValueAt: #rewindButton! !

!QMovieControls methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:42'!
rewindButton: aValue
	"Modify the receiver's rewindButton"
	^self propertyValueAt: #rewindButton put: aValue with: #rewindButtonChanged! !

!QMovieControls methodsFor: 'initialize' stamp: 'ar 5/26/2007 20:53'!
setupCostume
	super setupCostume.
	self extent: 75@25.
	rewindButton label: '<'.
	rewindButton bounds: (0@0 extent: 25@25).
	self add: rewindButton.
	pauseButton label: '||'.
	pauseButton bounds: (25@0 extent: 25@25).
	self add: pauseButton.
	playButton label: '>'.
	playButton bounds: (50@0 extent: 25@25).
	self add: playButton.
! !

CPlayer subclass: #QMoviePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-MovieApp'!

!QMoviePlayer commentStamp: '<historical>' prior: 0!
A non-replicated view for a replicated QEmbeddedMpegPlayer.!

QMoviePlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="embeddedApp" toSet="embeddedApp:" changeEvent="embeddedAppChanged">embeddedApp</field>
	<field toGet="croquetSession" toSet="croquetSession:" changeEvent="croquetSessionChanged">croquetSession</field>
	<field toGet="docRequest" toSet="docRequest:" changeEvent="docRequestChanged">docRequest</field>
	<field toGet="controls" toSet="controls:" changeEvent="controlsChanged">controls</field>
	<field toGet="fileName" toSet="fileName:" changeEvent="fileNameChanged">fileName</field>
	<field toGet="mpegFile" toSet="mpegFile:" changeEvent="mpegFileChanged">mpegFile</field>
	<field toGet="soundTrack" toSet="soundTrack:" changeEvent="soundTrackChanged">soundTrack</field>
	<field toGet="frameRate" toSet="frameRate:" changeEvent="frameRateChanged">frameRate</field>
	<field toGet="frameBuffer" toSet="frameBuffer:" changeEvent="frameBufferChanged">frameBuffer</field>
	<field toGet="running" toSet="running:" changeEvent="runningChanged">running</field>
	<field toGet="startTime" toSet="startTime:" changeEvent="startTimeChanged">startTime</field>
</fields>'!

!QMoviePlayer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 5/29/2007 10:35'!
advanceToFrame: msecs
	"Advance to the next frame if it is time to do so, skipping frames if necessary."
	| currentFrame desiredFrame framesToAdvance maxFrames |
	mpegFile hasVideo ifFalse: [^ self].
	maxFrames := mpegFile videoFrames: 0.
	desiredFrame := ((msecs * frameRate) // 1000) + 1.
	desiredFrame := desiredFrame min: maxFrames.
	currentFrame := mpegFile videoGetFrame: 0.
	framesToAdvance := desiredFrame - currentFrame.
	framesToAdvance <= 0 ifTrue: [^self].
	(framesToAdvance > 1) 
		ifTrue:[mpegFile videoDropFrames: framesToAdvance - 1 stream: 0].
	self nextFrame.
	desiredFrame >= maxFrames 
		ifTrue:[embeddedApp future stopPlaying].! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 6/20/2007 15:36'!
atEnd
	"Answer whether this movie is at its end"
	mpegFile ifNil:[^true].
	^(mpegFile videoGetFrame: 0) >= (mpegFile videoFrames: 0)! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 5/26/2007 21:59'!
closeFile
	"Close the mpeg file if it is open"
	soundTrack ifNotNil:[
		soundTrack pause.
		soundTrack := nil.
	].
	mpegFile ifNotNil:[
		mpegFile closeFile.
		mpegFile := nil.
	].
! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:46'!
controls
	"Answer the controls of the receiver"
	<bewareOf: #controlsChanged>
	^self propertyValueAt: #controls! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 20:46'!
controls: aValue
	"Modify the receiver's controls"
	^self propertyValueAt: #controls put: aValue with: #controlsChanged! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/27/2007 18:14'!
croquetSession
	"Answer the croquetSession of the receiver"
	<bewareOf: #croquetSessionChanged>
	^self propertyValueAt: #croquetSession! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/27/2007 18:14'!
croquetSession: aValue
	"Modify the receiver's croquetSession"
	^self propertyValueAt: #croquetSession put: aValue with: #croquetSessionChanged! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/28/2007 13:45'!
docRequest
	"Answer the docRequest of the receiver"
	<bewareOf: #docRequestChanged>
	^self propertyValueAt: #docRequest! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/28/2007 13:45'!
docRequest: aValue
	"Modify the receiver's docRequest"
	^self propertyValueAt: #docRequest put: aValue with: #docRequestChanged! !

!QMoviePlayer methodsFor: 'drawing' stamp: 'ar 5/28/2007 13:52'!
drawOn: aCanvas in: drawingBounds
	"Draw myself"
	| box ofs |
	frameBuffer ifNil:[
		"Note: We could do a lot more here, like displaying a download bar
		if the movie hasn't been downloaded yet, or displaying the good 'ole
		histogram for the sound being played"
		^super drawOn: aCanvas in: drawingBounds.
	].
	scale := (self width / drawingBounds width)
		max: (self height / drawingBounds height).
	box := 0@0 extent: (self extent / scale) truncated.
	ofs := (drawingBounds extent - box extent) // 2.
	box := drawingBounds origin + ofs extent: box extent.
	(drawingBounds areasOutside: box) 
		do:[:rect| aCanvas fillRectangle: rect color: self color].
	frameBuffer drawOn: aCanvas in: box sourceRect: (frameBuffer boundingBox).! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:02'!
embeddedApp
	"Answer the embeddedApp of the receiver"
	<bewareOf: #embeddedAppChanged>
	^self propertyValueAt: #embeddedApp! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:02'!
embeddedApp: aValue
	"Modify the receiver's embeddedApp"
	^self propertyValueAt: #embeddedApp put: aValue with: #embeddedAppChanged! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:06'!
fileName
	"Answer the fileName of the receiver"
	<bewareOf: #fileNameChanged>
	^self propertyValueAt: #fileName! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:06'!
fileName: aValue
	"Modify the receiver's fileName"
	^self propertyValueAt: #fileName put: aValue with: #fileNameChanged! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:12'!
frameBuffer
	"Answer the frameBuffer of the receiver"
	<bewareOf: #frameBufferChanged>
	^self propertyValueAt: #frameBuffer! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:12'!
frameBuffer: aValue
	"Modify the receiver's frameBuffer"
	^self propertyValueAt: #frameBuffer put: aValue with: #frameBufferChanged! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:11'!
frameRate
	"Answer the frameRate of the receiver"
	<bewareOf: #frameRateChanged>
	^self propertyValueAt: #frameRate! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:11'!
frameRate: aValue
	"Modify the receiver's frameRate"
	^self propertyValueAt: #frameRate put: aValue with: #frameRateChanged! !

!QMoviePlayer methodsFor: 'initialize' stamp: 'ar 5/26/2007 20:46'!
initialize
	super initialize.
	self define: #controls as: QMovieControls new.
! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:08'!
mpegFile
	"Answer the mpegFile of the receiver"
	<bewareOf: #mpegFileChanged>
	^self propertyValueAt: #mpegFile! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:08'!
mpegFile: aValue
	"Modify the receiver's mpegFile"
	^self propertyValueAt: #mpegFile put: aValue with: #mpegFileChanged! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 5/26/2007 19:49'!
nextFrame
	"Fetch the next frame into the frame buffer."
	mpegFile ifNil: [^ self].
	mpegFile videoReadFrameInto: frameBuffer stream: 0.
	self invalidate.
! !

!QMoviePlayer methodsFor: 'events' stamp: 'bgf 3/16/2008 18:35'!
onDestroyWorld
	"My world is being destroyed. Clean up."
	<on: destroyWorld>
	self closeFile.
	docRequest := nil.! !

!QMoviePlayer methodsFor: 'events' stamp: 'bgf 3/16/2008 18:49'!
onDocumentChanged
	"Update movie player state"
	| path doc |
	<on: documentChanged in: embeddedApp>
	doc := embeddedApp get: #document.
	path := embeddedApp get: #localFileUri.
	fileName := FileDirectory default pathFromURI: path.
	"Request download from the document manager"
	self openFile ifFalse:[
		docRequest := croquetSession documentManager fetchDocument: doc.
	] ifTrue: [
		docRequest := nil.	"Just clean living."
	].
	self onPlayStateChanged.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 19:05'!
onEmbeddedAppChanged
	"Update movie player state"
	<on: embeddedAppChanged>
	embeddedApp ifNotNil:[
		self onDocumentChanged.
	].! !

!QMoviePlayer methodsFor: 'events' stamp: 'bgf 3/16/2008 18:35'!
onFetchComplete: data
	"We are making progress"
	<on: fetchComplete in: docRequest>

	data ifNil:[self text: 'Download failed (click to retry)'].
	self openFile.
	self onPlayStateChanged.
	self invalidate.
	docRequest := nil.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/28/2007 13:53'!
onFetchProgress: amount
	"We are making progress"
	<on: fetchProgress in: docRequest>
	self text: 'Downloading ... (', ((amount * 100) truncateTo: 0.1),'%)'.
! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/28/2007 14:02'!
onMouseDown
	"Handle click on the player itself"
	| state |
	<on: mouseDown>
	mpegFile ifNil:[^self onDocumentChanged]. "request doc again if needed"
	"If we're at end, always just start playing"
	self atEnd ifTrue:[^embeddedApp future startPlaying].
	state := embeddedApp get: #playState.
	state caseOf: {
		[#playing] -> [embeddedApp future stopPlaying].
		[#paused] -> [embeddedApp future resumePlaying].
	} otherwise:[].! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 6/2/2007 22:55'!
onPauseRequest
	"Pause/Resume request"
	<on: pauseRequest in: controls>
	embeddedApp future stopPlaying.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 6/2/2007 22:55'!
onPlayRequest
	"Does nothing right now"
	<on: playRequest in: controls>
	self atEnd 
		ifTrue:[embeddedApp future startPlaying]
		ifFalse:[embeddedApp future resumePlaying].! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 20:25'!
onPlayStateChanged
	"Update movie player state"
	| state |
	<on: playStateChanged in: embeddedApp>
	state := embeddedApp get: #playState.
	state caseOf: {
		[#playing] -> [self startPlayingAfter: 
			(embeddedApp get: #nowTime) - (embeddedApp get: #startTime)].
		[#paused] -> [self stopPlayingAfter: 
			(embeddedApp get: #stopTime) - (embeddedApp get: #startTime)].
	} otherwise:[].! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 21:11'!
onPointerDown
	"Argh!! For some reason we get a pointerLeave before a pointerDown. 
	Must fake this so that the controls stay up..."
	<on: pointerDown in: embeddedApp>
	self onPointerEnter.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 21:10'!
onPointerEnter
	"Show controls"
	<on: pointerEnter in: embeddedApp>
	controls topLeft: (self bounds bottomCenter - ((controls extent x // 2) @ controls extent y)).
	controls visible: true.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 21:10'!
onPointerLeave
	"Hide controls"
	<on: pointerLeave in: embeddedApp>
	controls visible: false.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 21:10'!
onRewindRequest
	"Rewind the movie"
	<on: rewindRequest in: controls>
	embeddedApp future rewind.! !

!QMoviePlayer methodsFor: 'events' stamp: 'ar 5/26/2007 20:08'!
onTick
	"Go to the next frame"
	| msecs |
	<ticking: 20>
	startTime ifNotNil:[
		soundTrack
			ifNil: [msecs := Time millisecondClockValue - startTime]
			ifNotNil: [msecs := soundTrack millisecondsSinceStart - SoundPlayer bufferMSecs].
		self advanceToFrame: msecs
	].! !

!QMoviePlayer methodsFor: 'operations' stamp: 'bgf 10/12/2009 13:17'!
openFile
	"Try to open the MPEG file with the given name."
	| e msg |
	self closeFile.
	[
		(MPEGFile isFileValidMPEG: fileName) ifFalse:[^false].
		Smalltalk platformName = 'Win32' ifTrue: [
			"At the moment (Oct 12 2009)
			 the file-based mpeg causes unpleasantness - total death VM crashes in CRT."
			mpegFile := MPEGFile openFileUseBuffer: fileName.
		] ifFalse: [
			mpegFile := MPEGFile openFile: fileName.
		].
		mpegFile fileHandle ifNil: [self closeFile. ^false].

		"initialize soundTrack"
		mpegFile hasAudio
			ifTrue: [soundTrack := mpegFile audioPlayerForChannel: 1]
			ifFalse: [soundTrack := nil].

		mpegFile hasVideo ifTrue:[ 
			"set screen size and display first frame"
			frameRate := mpegFile videoFrameRate: 0.
			soundTrack ifNotNil: [  "compute frame rate from length of audio track"
				frameRate := (mpegFile videoFrames: 0) / soundTrack duration].
			e := (mpegFile videoFrameWidth: 0)@(mpegFile videoFrameHeight: 0).
			frameBuffer := Form extent: e depth: 32.
			self nextFrame.
		].
	] on: Error do:[:ex|
		msg := 'Error opening mpeg file: ', fileName,'\' withCRs, ex description.
		self text: msg.
		QLogger logWarning: msg.
		[mpegFile ifNotNil:[self closeFile]] on: Error do:[:nix].
		mpegFile := nil.
		^false
	].
	^true! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:15'!
running
	"Answer the running of the receiver"
	<bewareOf: #runningChanged>
	^self propertyValueAt: #running! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:15'!
running: aValue
	"Modify the receiver's running"
	^self propertyValueAt: #running put: aValue with: #runningChanged! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 5/26/2007 20:03'!
seekToFrame: msecs
	"Seek msecs milliseconds into the stream"
	| desiredFrame |
	mpegFile hasVideo ifFalse: [^ self].
	desiredFrame := ((msecs * frameRate) // 1000) + 1.
	desiredFrame := desiredFrame min: (mpegFile videoFrames: 0).
	mpegFile videoSetFrame: desiredFrame stream: 0.
	self nextFrame.
! !

!QMoviePlayer methodsFor: 'initialize' stamp: 'ar 5/28/2007 13:54'!
setupCostume
	super setupCostume.
	self userDraw: true. "let me draw myself please"
	controls visible: false.
	self add: controls.
	self textEnabled: true.
	self textFont: QStyles controlFont.
	self textAnchorPoint: #center.! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:11'!
soundTrack
	"Answer the soundTrack of the receiver"
	<bewareOf: #soundTrackChanged>
	^self propertyValueAt: #soundTrack! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:11'!
soundTrack: aValue
	"Modify the receiver's soundTrack"
	^self propertyValueAt: #soundTrack put: aValue with: #soundTrackChanged! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 5/26/2007 20:06'!
startPlayingAfter: deltaTime
	"Start playing, deltaTime msecs into the movie stream"
	mpegFile ifNil:[^nil].
	mpegFile hasVideo ifTrue:[
		"Set start time and video position according to deltaTime"
		startTime := Time millisecondClockValue - deltaTime truncated.
		self seekToFrame: deltaTime truncated.
		mpegFile hasAudio ifTrue:[
			soundTrack reset.  "ensure file is open before positioning"
			soundTrack soundPosition: deltaTime / (soundTrack duration * 1000).
			soundTrack resumePlaying.
		].
	].
! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:54'!
startTime
	"Answer the startTime of the receiver"
	<bewareOf: #startTimeChanged>
	^self propertyValueAt: #startTime! !

!QMoviePlayer methodsFor: 'accessing' stamp: 'ar 5/26/2007 19:54'!
startTime: aValue
	"Modify the receiver's startTime"
	^self propertyValueAt: #startTime put: aValue with: #startTimeChanged! !

!QMoviePlayer methodsFor: 'operations' stamp: 'ar 5/26/2007 20:04'!
stopPlayingAfter: deltaTime
	"Stop playing, deltaTime msecs into the movie stream"
	startTime := nil.
	mpegFile ifNil:[^nil].
	soundTrack ifNotNil:[soundTrack pause].
	self seekToFrame: deltaTime truncated.
! !

CPlayer subclass: #QSlideShowPlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-SlideShow'!

!QSlideShowPlayer commentStamp: '<historical>' prior: 0!
A simplistic player for on-island SlideShow documents.

This is a tweak player that is equipped for use as the harness-side world
of a QEmbeddedSlidePlayer, using the TweakProjectWorld machinery. !

QSlideShowPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="curSlideTexture" toSet="curSlideTexture:" changeEvent="curSlideTextureChanged">curSlideTexture</field>
	<field toGet="curSlideForm" toSet="curSlideForm:" changeEvent="curSlideFormChanged">curSlideForm</field>
	<field toGet="slideShowDoc" toSet="slideShowDoc:" changeEvent="slideShowDocChanged">slideShowDoc</field>
	<field toGet="croquetSession" toSet="croquetSession:" changeEvent="croquetSessionChanged">croquetSession</field>
	<field toGet="embeddedApp" toSet="embeddedApp:" changeEvent="embeddedAppChanged">embeddedApp</field>
	<field toGet="embeddableApp" toSet="embeddableApp:" changeEvent="embeddableAppChanged">embeddableApp</field>
	<field toGet="slideLabel" toSet="slideLabel:" changeEvent="slideLabelChanged">slideLabel</field>
	<field toGet="firstButton" toSet="firstButton:" changeEvent="firstButtonChanged">firstButton</field>
	<field toGet="prevButton" toSet="prevButton:" changeEvent="prevButtonChanged">prevButton</field>
	<field toGet="nextButton" toSet="nextButton:" changeEvent="nextButtonChanged">nextButton</field>
	<field toGet="lastButton" toSet="lastButton:" changeEvent="lastButtonChanged">lastButton</field>
	<field toGet="numSlides" toSet="numSlides:" changeEvent="numSlidesChanged">numSlides</field>
	<field toGet="curSlideNum" toSet="curSlideNum:" changeEvent="curSlideNumChanged">curSlideNum</field>
	<field toGet="maxSlideDimensions" toSet="maxSlideDimensions:" changeEvent="maxSlideDimensionsChanged">maxSlideDimensions</field>
</fields>'!

!QSlideShowPlayer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSlideShowPlayer class methodsFor: 'defaults' stamp: 'bgf 12/13/2008 10:36'!
controlHeightForExtent: ext
	"This is the height across the bottom of the player reserved
	 for controls... has to be based on the overall image size 
	 since that in turn is scaled relentlessly into the embedded-app display.

	The limits suit the clamping on overall size done in the player,
	 which has an image field min 256@256, and max 1024@1024."

	^  (ext y / 16) rounded min: 42 max: 18
	! !

!QSlideShowPlayer class methodsFor: 'defaults' stamp: 'bgf 12/14/2008 11:24'!
totalExtentForExtent: maxSlideDimensions

	"Compute the total size needed for the current slide-show + the control strip."
	| baseExtent scl |

	"We set both lower and upper limits on the image-output size,
	 so that we aren't forced to make a tiny control strip below a tiny image
	 (the  control strip is sized to be a fixed proportion of the total height.)
	 On the high end, we don't want to overrun texture capacity,
	 so we keep the image field within 1024@1024."

	baseExtent := maxSlideDimensions.
	scl := nil.
	baseExtent maxDimension > 1024 ifTrue: [
		scl := 1024.0 / baseExtent maxDimension	
	] ifFalse: [
		baseExtent maxDimension < 256 ifTrue: [ 
			scl := 256.0 / (baseExtent maxDimension max: 1) 
		].
	].
	scl ifNotNil: [
		baseExtent := (baseExtent * scl) rounded.
	].

	^baseExtent + (0 @ (QSlideShowPlayer controlHeightForExtent: maxSlideDimensions))! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 4/1/2010 14:31'!
alwaysLooping
	^ slideShowDoc isValidFarRef and: [  (slideShowDoc get: #alwaysLooping) = true ]! !

!QSlideShowPlayer methodsFor: 'drawing' stamp: 'bgf 12/20/2008 23:41'!
clear
	curSlideTexture := nil.
	curSlideForm := nil.
	maxSlideDimensions := 256@256.
	self message: ''.
	self invalidate.! !

!QSlideShowPlayer methodsFor: 'drawing' stamp: 'bgf 12/20/2008 23:41'!
clearWithMessage: aString
	curSlideTexture := nil.
	curSlideForm := nil.
	self message: aString; 
		invalidate.! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/10/2008 16:20'!
croquetSession
	"Answer the croquetSession of the receiver"
	<bewareOf: #croquetSessionChanged>
	^self propertyValueAt: #croquetSession! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/10/2008 16:20'!
croquetSession: aValue
	"Modify the receiver's croquetSession"
	^self propertyValueAt: #croquetSession put: aValue with: #croquetSessionChanged! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/20/2008 23:15'!
curSlideForm
	"Answer the curSlideForm of the receiver"
	<bewareOf: #curSlideFormChanged>
	^self propertyValueAt: #curSlideForm! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/20/2008 23:15'!
curSlideForm: aValue
	"Modify the receiver's curSlideForm"
	^self propertyValueAt: #curSlideForm put: aValue with: #curSlideFormChanged! !

!QSlideShowPlayer methodsFor: 'private' stamp: 'bgf 12/18/2008 12:45'!
curSlideLabel

	"Assemble the label-strip text."
	curSlideTexture ifNil: [
		^ 'No Image'
	].
	curSlideTexture url ifNotNilDo: [ :uri |
		| nm |
		nm := QStyles simpleFileName: uri.
		^ slideLabel text: ( nm, ' (', curSlideNum, ' / ', numSlides, ')' ).
	].
	^ 'Slide ', curSlideNum, ' / ', numSlides! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 21:29'!
curSlideNum
	"Answer the curSlideNum of the receiver"
	<bewareOf: #curSlideNumChanged>
	^self propertyValueAt: #curSlideNum! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 21:29'!
curSlideNum: aValue
	"Modify the receiver's curSlideNum"
	^self propertyValueAt: #curSlideNum put: aValue with: #curSlideNumChanged! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 13:57'!
curSlideTexture
	"Answer the curSlideTexture of the receiver;
	 this is either nil, or the TContentManager full texture corresponding to the current slide."
	<bewareOf: #curSlideTextureChanged>
	^self propertyValueAt: #curSlideTexture! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 13:54'!
curSlideTexture: aValue
	"Modify the receiver's curSlideTexture"
	^self propertyValueAt: #curSlideTexture put: aValue with: #curSlideTextureChanged! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 12/13/2008 00:28'!
displayCurrentImage
	<on: fire in: slideLabel>

	"Sync myself to the current slide in the slide-show;
	 this is the single-harness equivalent of having the slide show do #synchronize. "
	slideShowDoc isValidFarRef ifTrue: [
		self displayImage: (slideShowDoc get: #currentImageSHA) 
			position: (slideShowDoc get:#position)
			total: (slideShowDoc get: #size)
			maxDimensions: (slideShowDoc get: #maxDimensions).
	].! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 12/12/2008 08:41'!
displayFirstImage
	<on: fire in: firstButton>
	"Send the show back to the first."
	slideShowDoc isValidFarRef ifTrue: [
		slideShowDoc future toStart.
	].! !

!QSlideShowPlayer methodsFor: 'drawing' stamp: 'bgf 8/11/2009 15:23'!
displayImage: sha position: slideNum total: sz maxDimensions: md

	"Display the current slide image. 
	 This handles slide-show #syncrhonize, which fires on all changes."
	| dtxtr properExtent |

	sha ifNil: [ ^ self clear ].

	maxSlideDimensions := md max: (256@256).	
	self setSlide: slideNum total: sz.
	properExtent := self totalExtent.

	"Short cut if this is merely a refresh."
	(sha notNil and: [curSlideTexture notNil] 
	 and: [ curSlideTexture sha = sha ]
	 and: [ properExtent = extent ])
	ifTrue: [
		self invalidate.
		^ self ].

	"Get the actual texture, if we've got it available in content..."
	dtxtr := croquetSession contentManager resolveSHA: (sha).
	dtxtr ifNil: [
		self clearWithMessage: 'Fetching...'.
		croquetSession contentManager prioritizeSHA: (sha).
		slideShowDoc 
			isValidFarRef ifTrue: [ slideShowDoc 
								future synchronizeAfter: 5000 for: (croquetSession sessionID) ]
			ifFalse: [ QLogger logWarning: 'No slideshow, but handing sync.' ].
		^ self ].
	curSlideTexture := dtxtr.
	curSlideForm := nil.
	"For best results,  if the form texture has been POT-scaled,
	 try to get the original image-form so we're not unscaling in drawOn:..."
	(curSlideTexture form extent = curSlideTexture extent) ifFalse: [
		[	curSlideForm :=
				 (croquetSession cacheManager find: sha)
					ifNotNil: [ : data |
							ImageReadWriter formFromStream: (ReadStream on: data) ]
		] ifError: [ :msg|  QLogger logWarning: 'Cannot read cached form ', sha, ': ', msg ] ].
	
	slideLabel text: (self curSlideLabel).

	"This is just a prod to make the next texture appear more quickly
	 (for better response if rendered-textures are loaded lazily.)"
	TContentManager default resolveSHA: (slideShowDoc get: #nextImageSHA).

	"If the current app extent isn't appropriate, 
	 relay the update through the embedded app,
	 since it needs to know about appExtent changes.  
	 It will update the extent and signal a new synchronize."
	(extent = properExtent) ifFalse: [
		embeddedApp isValidFarRef ifTrue: [
			embeddedApp future resize: properExtent forSHA: sha.
			^ self
		]].

	self invalidate.

	
	! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 12/12/2008 08:41'!
displayLastImage
	<on: fire in: lastButton>
	"Send the show to the last slide"
	slideShowDoc isValidFarRef ifTrue: [
		slideShowDoc future toEnd.
	].! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 12/12/2008 08:41'!
displayNextImage
	<on: fire in: nextButton>
	"Send the show ahead one slide."
	slideShowDoc isValidFarRef ifTrue: [
		slideShowDoc future forwardOne.
	].! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 12/12/2008 08:41'!
displayPrevImage
	<on: fire in: prevButton>
	"Send the show ahead one slide."
	slideShowDoc isValidFarRef ifTrue: [
		slideShowDoc future backOne.
	].! !

!QSlideShowPlayer methodsFor: 'drawing' stamp: 'bgf 12/20/2008 23:42'!
drawOn: aCanvas in: drawingBounds
	"Draw myself - blit the slide texture into the display canvas."

	| destRect ch ext field scl box ofs form |

	super drawOn: aCanvas in: drawingBounds.
	(curSlideTexture isNil or: [ curSlideTexture form isNil ]) ifTrue: [
		"Could do more here, like displaying a download bar
		if the resource hasn't been downloaded yet."
		^ self ].

	"We use the direct-from-disk curSlideForm if we have it, else fall back to the scaled one."
	form := curSlideForm ifNil: [ curSlideTexture form ].
	ch :=  (QSlideShowPlayer controlHeightForExtent: (maxSlideDimensions)).

	"Adjust for curSlideTextures that are smaller or larger 
	 than the min/max dimensions... see totalExtent"
	ext := curSlideTexture extent.
	field :=  drawingBounds origin extent: (drawingBounds extent - (0@ch)).
	destRect := field.

	(ext x > destRect width or: [ ext y > destRect height ]) ifTrue: [
		scl := (destRect width / ext x) min: (destRect height / ext y). 
	] ifFalse: [
		scl := 1.
	].

	"Often the texture will just fit; but sometimes we'll need to center it,
	 and we cap the output size so some large textures will get downscaled."
	box := 0@0 extent: (ext * scl) truncated.
	ofs := (field extent - (box extent)) // 2.
	box := (field origin + ofs) extent: box extent.
	(field areasOutside: box) do: [ : r | aCanvas fillRectangle: r color: (self color)].

	(ext = form extent) ifTrue: [
		form
			drawOn: aCanvas 
			in: box 
			sourceRect: (form boundingBox).
	] ifFalse: [
		"We want high-quality scaling, for powerpoint slides bearing text,
		 so we use the more complex API that allows specifying cellSize.
		 We're (usually) restoring the slide-texture to its native image extent
	  	 after it has been POT-enlarged by the texture system."
		aCanvas
				warpImage: form
				transform: (CMatrixTransform
								localRect:  form boundingBox 
								toGlobal: box)
				at: 0@0
				sourceRect:  form boundingBox
				cellSize: 3.
	].! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 08:49'!
embeddableApp
	"Answer the embeddableApp of the receiver"
	<bewareOf: #embeddableAppChanged>
	^self propertyValueAt: #embeddableApp! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 08:49'!
embeddableApp: aValue
	"Modify the receiver's embeddableApp"
	^self propertyValueAt: #embeddableApp put: aValue with: #embeddableAppChanged! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/10/2008 16:20'!
embeddedApp
	"Answer the embeddedApp of the receiver"
	<bewareOf: #embeddedAppChanged>
	^self propertyValueAt: #embeddedApp! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/10/2008 16:20'!
embeddedApp: aValue
	"Modify the receiver's embeddedApp"
	^self propertyValueAt: #embeddedApp put: aValue with: #embeddedAppChanged! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
firstButton
	"Answer the firstButton of the receiver"
	<bewareOf: #firstButtonChanged>
	^self propertyValueAt: #firstButton! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
firstButton: aValue
	"Modify the receiver's firstButton"
	^self propertyValueAt: #firstButton put: aValue with: #firstButtonChanged! !

!QSlideShowPlayer methodsFor: 'private' stamp: 'bgf 4/1/2010 14:29'!
indicatePlaying: bPlay

	| buttons |
	buttons := { prevButton . nextButton . firstButton . lastButton . slideLabel }.
	buttons do: [ : b | b visible: bPlay not ].
! !

!QSlideShowPlayer methodsFor: 'initialize' stamp: 'bgf 12/12/2008 20:31'!
initialize
	super initialize.
	self define: #slideLabel as: CPushButton new.
	self define: #firstButton as: CPushButton new.
	self define: #lastButton as: CPushButton new.
	self define: #nextButton as: CPushButton new.
	self define: #prevButton as: CPushButton new.

	maxSlideDimensions := 256@256.! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 4/1/2010 13:55'!
isPlaying
	^ slideShowDoc isValidFarRef and: [  (slideShowDoc get: #isPlaying) = true ]! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
lastButton
	"Answer the lastButton of the receiver"
	<bewareOf: #lastButtonChanged>
	^self propertyValueAt: #lastButton! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
lastButton: aValue
	"Modify the receiver's lastButton"
	^self propertyValueAt: #lastButton put: aValue with: #lastButtonChanged! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 8/22/2009 04:10'!
layoutControls

	"We update the control sizes to suit the new overall extent...
	 trying for something close to a fixed-height control strip inside the rendered world,
	 while letting the texture extent mostly control the app height."
	| lowerMargin font bwid slideFont |
	lowerMargin := QSlideShowPlayer controlHeightForExtent: (maxSlideDimensions).

	"We use bitstream here because we need many size variants up to 36 point."
	font := ((TextStyle named: #BitstreamVeraSans) fontOfSize: ((lowerMargin - 5) min: 36)).
	"Making the <> font to suit gives us a big font...buy some room by using a smaller one for the title."
	slideFont := ((TextStyle named: #BitstreamVeraSans) fontOfSize: (((lowerMargin - 2) min: 30) max: 6)).

	bwid := (lowerMargin * 1.75) rounded.
	{ firstButton . lastButton . prevButton . nextButton } do: [ : b | 
		b textFont: font; 
		extent: bwid @ lowerMargin ].
	
	firstButton bottomLeft: self bottomLeft; fullBounds.
	lastButton bottomRight: self bottomRight; fullBounds.
	prevButton bottomLeft: firstButton bottomRight.
	nextButton bottomRight: lastButton bottomLeft.

	slideLabel 
		textFont: slideFont;
		extent: (self extent x - (4*bwid) @ lowerMargin);
		bottomLeft: (prevButton bottomRight).! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 20:30'!
maxSlideDimensions
	"Answer the maxSlideDimensions of the receiver"
	<bewareOf: #maxSlideDimensionsChanged>
	^self propertyValueAt: #maxSlideDimensions! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/13/2008 00:18'!
maxSlideDimensions: aValue
	"Modify the receiver's maxSlideDimensions... using equals as update-check"
	(aValue = maxSlideDimensions) ifFalse: [
		self propertyValueAt: #maxSlideDimensions put: aValue with: #maxSlideDimensionsChanged
	].
! !

!QSlideShowPlayer methodsFor: 'private' stamp: 'bgf 12/11/2008 21:24'!
message: aString
	"Set the control-strip label to a given fixed text."
	slideLabel text: aString! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
nextButton
	"Answer the nextButton of the receiver"
	<bewareOf: #nextButtonChanged>
	^self propertyValueAt: #nextButton! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
nextButton: aValue
	"Modify the receiver's nextButton"
	^self propertyValueAt: #nextButton put: aValue with: #nextButtonChanged! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 21:29'!
numSlides
	"Answer the numSlides of the receiver"
	<bewareOf: #numSlidesChanged>
	^self propertyValueAt: #numSlides! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 21:29'!
numSlides: aValue
	"Modify the receiver's numSlides"
	^self propertyValueAt: #numSlides put: aValue with: #numSlidesChanged! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 4/1/2010 13:59'!
onAlwaysLoopingChanged: bLooping
	<on: alwaysLoopingChanged in: slideShowDoc>
	"Playing is entirely in the document."! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 4/1/2010 14:41'!
onAutoStepIntervalChanged:interval
	<on: autoStepIntervalChanged in: slideShowDoc>
	
	"Playing is entirely in the document."! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 4/8/2010 07:38'!
onContextMenu: event
	| menu loop item uid |
	<on: yellowButtonDown>

	uid := embeddedApp get: #controlUser.
	loop := ((slideShowDoc get: #alwaysLooping) = true).
	menu := CMenu new.

	self isPlaying ifFalse: [
		item := menu add: 'Play' target: self action: #startPlayingForUser: argument: uid.
		item enabled: (loop not)
	] ifTrue: [
		item := menu add: 'Stop' target: self action: #stopPlayingForUser: argument: uid.
		item enabled: (loop not)
	].

	menu addSeparator.
	item := menu add:  'Continuous Play'  target: self action: #toggleLoopingForUser: argument: uid.
	item checked: loop.
	
	QStyles installMenuFontInMenu: menu.
	menu openAt: hand position.
! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 4/1/2010 14:14'!
onEmbeddedAppChanged
	<on: embeddedAppChanged>

	embeddedApp isValidFarRef ifFalse: [
		slideShowDoc := nil.
		^ self
	].
	slideShowDoc := embeddedApp get: #document.
	slideShowDoc isValidFarRef ifTrue: [
		"This is the event we use for slide show synchronization"
		self startScript: #displayImage:position:total:maxDimensions: 
			when: {slideShowDoc . #imageChanged}.

		"This provokes a followup image check for one client."
		self startScript: #session:displayImage:position:total:maxDimensions: 
			when: {slideShowDoc . #sessionSync }.

		self indicatePlaying: ((slideShowDoc get: #isPlaying) = true).
		slideShowDoc future synchronizeSession: (self croquetSession sessionID).
	].! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 12/13/2008 00:25'!
onExtentChanged

	<on: extentChanged>

	self displayCurrentImage.	"I.e. get currentSlideForm from the slide show."
	self layoutControls.! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 12/11/2008 10:15'!
onKeyDown: event
	<on: keyDown>

	(CPlatform current standardEditingKeys lookupEvent: event)
	 caseOf: {
		[ #movePageDown ] -> [ self displayNextImage ] .
		[ #moveCursorRight ] -> [ self displayNextImage ] .
		[ #moveLineDown ] -> [ self displayNextImage ] .
		[ #movePageUp ] -> [ self displayPrevImage ].
		[ #moveCursorLeft ] -> [ self displayPrevImage ] .
		[ #moveLineUp ] -> [ self displayPrevImage ] .
		[ #moveLineStart ] -> [ self displayFirstImage ].	"Home"
		[ #moveLineEnd ] -> [ self displayLastImage ].	"End"
	} otherwise: [].

! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 12/14/2008 10:52'!
onKeyStroke: event
	<on: keyStroke>

	"This is a stroke rather than down since we want the 'real' character..."
	event keyCharacter = (Character space) ifTrue: [
		self displayNextImage.
	].
! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 12/13/2008 00:27'!
onMaxSlideDimensionsChanged

	<on: maxSlideDimensionsChanged>

	self layoutControls.

	"But not the main display,
	 which is changed in the same places that update  maxSlideDimensions."
! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 4/2/2010 07:08'!
onMouseDown: event
	"Just induce a refresh..."
	<on: redButtonDown>
	self onExtentChanged.

	(self isPlaying and: [ self alwaysLooping not ])
		ifTrue: [ self stopPlayingForUser: (embeddedApp controlUser) ].
! !

!QSlideShowPlayer methodsFor: 'events' stamp: 'bgf 4/1/2010 14:14'!
onPlayingChanged: bPlay
	<on: isPlayingChanged in: slideShowDoc>

	self indicatePlaying: (bPlay = true).! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
prevButton
	"Answer the prevButton of the receiver"
	<bewareOf: #prevButtonChanged>
	^self propertyValueAt: #prevButton! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/12/2008 08:23'!
prevButton: aValue
	"Modify the receiver's prevButton"
	^self propertyValueAt: #prevButton put: aValue with: #prevButtonChanged! !

!QSlideShowPlayer methodsFor: 'drawing' stamp: 'bgf 8/11/2009 15:59'!
session: sessionID displayImage: sha position: slideNum total: sz maxDimensions: md
	"Display the current slide image. 
	 This handles slide-show #syncrhonizeSession, 
	 which fires a followup check for one client."

	croquetSession ifNotNil: [
		croquetSession sessionID = sessionID ifTrue: [
			self displayImage: sha position: slideNum total: sz maxDimensions: md
		]].! !

!QSlideShowPlayer methodsFor: 'private' stamp: 'bgf 12/13/2008 00:37'!
setSlide: pos total: count
	"Remember where we are, and adjust the controls accordingly"
	| haveSome |
	curSlideNum := pos.
	numSlides := count.

	haveSome := (slideShowDoc isValidFarRef and: [ count > 0 ]).
	firstButton enabled: haveSome & (curSlideNum > 1).
	lastButton enabled: haveSome & (curSlideNum ~= numSlides).
	prevButton enabled: firstButton enabled.
	nextButton enabled: lastButton enabled.! !

!QSlideShowPlayer methodsFor: 'initialize' stamp: 'bgf 12/14/2008 13:59'!
setupCostume
	| gradient |
	super setupCostume.
	self userDraw: true. "let me draw myself please"
	self textEnabled: true.
	self textFont: QStyles controlFont.
	self textAnchorPoint: #center.
	self color: (CWidgetLibrary default gradientColors second adjustSaturation: 0 brightness: -0.1).

	gradient := (CWidgetLibrary default gradientColors).
	slideLabel fillColors: gradient;
		 hResizing: #rigid; vResizing: #rigid;
		 extent: (256@16);
		 textFont: QStyles controlFont;
		 textAnchorPoint: #center.
	self add: slideLabel.

	firstButton label: '|<'.
	lastButton label: '>|'.
	nextButton label: '>'.
	prevButton label: '<'.

	{ firstButton . lastButton . nextButton  . prevButton } do: [ : b |
		b fillColors: gradient;
			 hResizing: #rigid; vResizing: #rigid;
			 extent: (256@16);
			 font: QStyles controlFont.
		self add: b.
	].

	self onExtentChanged.		! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 17:42'!
slideLabel
	"Answer the slideLabel of the receiver"
	<bewareOf: #slideLabelChanged>
	^self propertyValueAt: #slideLabel! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/11/2008 17:42'!
slideLabel: aValue
	"Modify the receiver's slideLabel"
	^self propertyValueAt: #slideLabel put: aValue with: #slideLabelChanged! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/10/2008 16:24'!
slideShowDoc
	"Answer the slideShowDoc of the receiver"
	<bewareOf: #slideShowDocChanged>
	^self propertyValueAt: #slideShowDoc! !

!QSlideShowPlayer methodsFor: 'accessing' stamp: 'bgf 12/10/2008 16:24'!
slideShowDoc: aValue
	"Modify the receiver's slideShowDoc"
	^self propertyValueAt: #slideShowDoc put: aValue with: #slideShowDocChanged! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 4/2/2010 06:56'!
startPlayingForUser: userID
	slideShowDoc isValidFarRef ifTrue: [
		slideShowDoc future startPlaying.
	].! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 4/2/2010 06:56'!
stopPlayingForUser: userID
	slideShowDoc isValidFarRef ifTrue: [	
		slideShowDoc future stopPlaying.
	]! !

!QSlideShowPlayer methodsFor: 'slide-show actions' stamp: 'bgf 4/2/2010 07:28'!
toggleLoopingForUser: userID
	slideShowDoc isValidFarRef ifTrue: [
		slideShowDoc future
				alwaysLooping: ((slideShowDoc get: #alwaysLooping) ~= true)
				inApplication: embeddedApp
				forUser: userID.
	].! !

!QSlideShowPlayer methodsFor: 'private' stamp: 'bgf 12/14/2008 11:25'!
totalExtent
	
	"The total app-extent desired to contain the current slide show and the control strip."
	^ self class totalExtentForExtent: maxSlideDimensions! !

CPlayer subclass: #QTextAppPane
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-TextApp'!

!QTextAppPane commentStamp: '<historical>' prior: 0!
Captures and forwards scroll requests to the embedded app.!

QTextAppPane defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="embeddedProxy" toSet="embeddedProxy:" changeEvent="embeddedProxyChanged">embeddedProxy</field>
</fields>'!

!QTextAppPane class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextAppPane methodsFor: 'accessing' stamp: 'ar 6/2/2008 20:09'!
embeddedProxy
	"Answer the embeddedProxy of the receiver"
	<bewareOf: #embeddedProxyChanged>
	^self propertyValueAt: #embeddedProxy! !

!QTextAppPane methodsFor: 'accessing' stamp: 'ar 6/2/2008 20:09'!
embeddedProxy: aValue
	"Modify the receiver's embeddedProxy"
	^self propertyValueAt: #embeddedProxy put: aValue with: #embeddedProxyChanged! !

!QTextAppPane methodsFor: 'scrolling' stamp: 'das 12/7/2009 12:02'!
justScrollOffset: aPoint
	"Don't replicate the scroll offset."
	^super scrollOffset: aPoint! !

!QTextAppPane methodsFor: 'scrolling' stamp: 'ar 6/2/2008 20:57'!
scrollOffset: aPoint
	"Capture the scroll offset and forward it. Since the scroll offset is idempotent, also set it locally for faster feedback."
	embeddedProxy ifNotNil:[
		(embeddedProxy sendScrollOffset: aPoint truncated) ifTrue:[^self].
	].
	^super scrollOffset: aPoint! !

!QTextAppPane methodsFor: 'initialize' stamp: 'ar 4/19/2007 18:09'!
setupCostume
	"Set up my costume"
	super setupCostume.
	self text: ''.
	self textAnchorPoint: #topLeft.
	self textWrap: true.! !

TGeometry subclass: #QArrowHead
	instanceVariableNames: 'extent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QArrowHead commentStamp: 'das 3/16/2007 12:36' prior: 0!
A triangular prism.  An extruded isosceles triangle pointing up at positive y,
extruded along the z axis.   Angular edges.
Currently does not have intelligent texture coordinates.

Though this looks like a procedural TGeometry, it isn't. It has to maintain a local variant of the TMeshX for picking purposes anyway.!

!QArrowHead class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QArrowHead class methodsFor: 'mesh' stamp: 'ar 12/30/2007 06:11'!
constructMesh: geometry

	| za temp sin60 cos60 groups xa ya vertices normals aMesh extent |

	temp := Vector3Array new: 8.
	extent := geometry get: #extent.
	xa := extent x*0.5.
	ya := extent y*0.5.
	za := extent z*0.5.

	cos60 := ya / (extent x).
	sin60 := cos60 arcCosClamped sin.

	temp at: 1 put: (Vector3 x: 0 y: ya z: za).
	temp at: 2 put: (Vector3 x: xa y: ya negated z: za).
	temp at: 3 put: (Vector3 x: xa negated y: ya negated z: za).
	temp at: 4 put: (Vector3 x: 0  y: ya z: za negated).
	temp at: 5 put: (Vector3 x: xa y: ya negated z: za negated).
	temp at: 6 put: (Vector3 x: xa negated y: ya negated z: za negated).

	groups := #((1 3 2) (4 5 6) (1 4 3) (4 6 3) (2 6 5) (6 2 3) (1 2 4) (2 5 4)).

	vertices := Vector3Array ofSize: 24.
	groups doWithIndex: [:g :i |
		1 to: 3 do: [ :j| vertices at: (i-1)*3 + j put: (temp at: (g at: j))]
	].

	normals := Vector3Array ofSize: 24.
	1 to:3 do: [: i | normals at: i put: (Vector3 x: 0 y: 0 z: 1) ].
	4 to:6 do: [: i | normals at: i put: (Vector3 x: 0 y: 0 z: -1) ].
	7 to: 12 do: [ :i | normals at: i put: (Vector3 x: sin60 negated y: cos60  z: 0) ].
	13 to: 18 do: [ :i | normals at: i put: (Vector3 x: 0 y: -1 z: 0) ].	
	19 to: 24 do: [ :i | normals at: i put: (Vector3 x: sin60 y: cos60 z: 0) ].
	
	aMesh := TMeshX new initialize.
	aMesh triangleFaces: #(0 1 2  3 4 5  6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23) asIntegerArray.
	aMesh vertices: vertices; normals: normals; materialIndex:1.
	^ aMesh.! !

!QArrowHead class methodsFor: 'defaults' stamp: 'bgf 2/6/2007 13:50'!
defaultExtent
	^ 1@1@1! !

!QArrowHead methodsFor: 'persist' stamp: 'bgf 2/6/2007 13:50'!
constructFrom: pStream
	"Construct the receiver"
	super constructFrom: pStream.
	self extent: (pStream readVector3: 'extent' default: self class defaultExtent).! !

!QArrowHead methodsFor: 'accessing' stamp: 'bgf 2/6/2007 10:07'!
extent
	^ extent.
! !

!QArrowHead methodsFor: 'accessing' stamp: 'bgf 2/6/2007 10:06'!
extent: vect3
	extent = vect3 ifFalse:[
		extent := vect3.
		self shapeChanged.
	].! !

!QArrowHead methodsFor: 'accessing' stamp: 'bgf 10/4/2007 11:44'!
frameBox
	"My bounding box."
	^ TBox min: (extent/2 negated) max: (extent/2)! !

!QArrowHead methodsFor: 'initialize' stamp: 'das 3/16/2007 12:20'!
initialize
	super initialize.
	
	extent := self class defaultExtent.
! !

!QArrowHead methodsFor: 'testing' stamp: 'das 3/21/2007 20:18'!
isProcedural

	^ true.! !

!QArrowHead methodsFor: 'persist' stamp: 'bgf 2/6/2007 10:06'!
persistOn: pStream
	"Persist the receiver"
	super persistOn: pStream.
	pStream storeVector3: 'extent' value: self extent default: 1@1@1.
! !

!QArrowHead methodsFor: 'render' stamp: 'das 3/21/2007 20:15'!
pick: pointer


	^ (pointer pickBox: extent location: 0@0@0).
! !

!QArrowHead methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:55'!
privateScale: scl
	extent := extent * scl.
	super privateScale: scl.
! !

!QArrowHead methodsFor: 'initialize' stamp: 'bgf 9/20/2007 11:20'!
updateBounds

	super updateBounds.
	boundSphere := TBoundSphere localPosition: 0@0@0 radiusSquared: ((extent length/2.0) squared).
	boundSphere frame: self.

! !

TGeometry subclass: #QLozenge
	instanceVariableNames: 'segments radius extent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QLozenge commentStamp: '<historical>' prior: 0!
A flat lozenge shape, like a TWindowFrame with flush front and back faces
rather than an opening.

Extent runs xy; radius is the radius of the rounded boundary,
and also half the z-thickness of the whole.
!

!QLozenge class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLozenge class methodsFor: 'inspector' stamp: 'bgf 4/9/2010 10:24'!
addPropertyRows: propertyDialog forFrame: lref

	self addNameRowIfNeeded: propertyDialog forFrame: lref.

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ :frame | frame get:#extent ];
			acceptBlock: [ :v2 :frame | frame future extent: v2 ];
			balloonHelp: 'Extent of the flat surface of the Lozenge.';
	 	yourself	).

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Radius';
			fetchBlock: [ :frame | frame get:#radius ];
			acceptBlock: [ : r :frame | frame future radius: r ];
			validateBlock: [ : r : f  | 
				(r isNil or: [ r <= 0 ]) ifTrue: [
					self inform: 'Radius must be positive.'.
					false
				] ifFalse: [ true ]];
			balloonHelp: 'Radius of the rounded edge.';
	 	yourself	).

	super addPropertyRows: propertyDialog forFrame: lref.! !

!QLozenge class methodsFor: 'mesh' stamp: 'bgf 10/17/2007 21:46'!
constructMesh: geometry

	"Sloppily adapted from TWindowFrame,
	 with the front and back faces spliced in longhand at the end."

	| ringSin ringCos segs halfSegs pi2 ax vert norm texCoord dx dy count savedVert savedNorm stream rowCount totalCount verts norms txtrCoords strips aMesh segments extent radius faceCount |

	segments := geometry get: #segments.
	extent := geometry get: #extent.
	radius := geometry get: #radius.

	aMesh := TMeshX new initialize.
	vert := Vector3 new.
	norm := Vector3 new.
	texCoord := Vector2 new.
	segs := 4 * segments.
	halfSegs := 2 * segments.
	ringSin := FloatArray ofSize: segs+1.
	ringCos := FloatArray ofSize: segs+1.
	pi2 := Float pi * 2.0.

	"Calculate the number of vertices in a 'row', and the total number of vertices."
	rowCount := 4*(segments+1) + 5.
	totalCount := (halfSegs+1)*rowCount.

	faceCount := 8.
	verts := Vector3Array ofSize: totalCount + faceCount.
	norms := Vector3Array ofSize: totalCount + faceCount.
	txtrCoords := Vector2Array ofSize: totalCount + faceCount.

	1 to: segs do: [:index |
		ax := ((index-1) * pi2) / segs.
		ringSin at: index put: ax sin.
		ringCos at: index put: ax cos.].

	ringSin at: segs+1 put: (ringSin at: 1).
	ringCos at: segs+1 put: (ringCos at: 1).
	dx := extent x/2.0.
	dy := extent y/2.0.

	texCoord x:0 y:0.
	count := 1.

	1 to: halfSegs+1 do:[ :iv |
" ---- top right ---- "
		1 to: segments+1 do:
			[:ih |
			norm
				x: (ringSin at: iv) * (ringSin at: ih)
				y: (ringSin at: iv) * (ringCos at: ih)
				z: (ringCos at: iv).
			vert := norm * radius.
			vert x: vert x + dx.
			vert y: vert y + dy.
			verts at: count put: vert.
			norms at: count put: norm.
			txtrCoords at: count put: texCoord.
			count := count + 1.

			"Need to repeat the first one at the end."
			ih = 1 ifTrue:
				[savedVert := vert copy. 
				savedNorm := norm copy.]
			].
		vert y: 0.
		verts at: count put: vert.
		norms at: count put: norm.
		txtrCoords at: count put: texCoord.
		count := count + 1.

" ---- bottom right ----- "
		segments+1 to: (2*segments)+1 do:
			[:ih |
			norm
				x: (ringSin at: iv) * (ringSin at: ih)
				y: (ringSin at: iv) * (ringCos at: ih)
				z: (ringCos at: iv).
			vert := norm * radius.
			vert x: vert x + dx.
			vert y: vert y - dy.
			verts at: count put: vert.
			norms at: count put: norm.
			txtrCoords at: count put: texCoord.
			count := count + 1.
			].
		vert x: 0.
		verts at: count put: vert.
		norms at: count put: norm.
		txtrCoords at: count put: texCoord.
		count := count + 1.

" ---- bottom left ---- "
		(2*segments)+1 to: (3*segments)+1 do:
			[:ih |
			norm 
				x: (ringSin at: iv) * (ringSin at: ih)
				y: (ringSin at: iv) * (ringCos at: ih)
				z: (ringCos at: iv).
			vert := norm * radius.
			vert x: vert x - dx.
			vert y: vert y-  dy.
			verts at: count put: vert.
			norms at: count put: norm.
			txtrCoords at: count put: texCoord.
			count := count + 1.
			].
		vert y: 0.
		verts at: count put: vert.
		norms at: count put: norm.
		txtrCoords at: count put: texCoord.
		count := count + 1.

" ---- top left ----- "
		(3*segments)+1 to: (4*segments)+1 do:
			[:ih |
			norm 
				x: (ringSin at: iv) * (ringSin at: ih)
				y: (ringSin at: iv) * (ringCos at: ih)
				z: (ringCos at: iv).		
			vert := norm * radius.
			vert x: vert x - dx.
			vert y: vert y +  dy.
			verts at: count put: vert.
			norms at: count put: norm.
			txtrCoords at: count put: texCoord.
			count := count + 1.
			].
		vert x: 0.
		verts at: count put: vert.
		norms at: count put: norm.
		txtrCoords at: count put: texCoord.
		count := count + 1.

		verts at: count put: savedVert.
		norms at: count put: savedNorm.
		txtrCoords at: count put: texCoord.
		count := count + 1.
	].

	(count-1) = 	(totalCount) ifFalse: [self error: 'I Miscalculated the number of vertices!!!!'].

	"Compute face groups."
	strips := Array new: halfSegs + 2.
	0 to: halfSegs-1 do:[:iv |
		stream := WriteStream on: IntegerArray new.
		0 to: rowCount-1 do: [:ih | stream
									nextPut: ((iv+1)*rowCount) + ih;
									nextPut: (iv*rowCount) + ih].
		strips at: iv+1 put: (TStripData mode: GLTriangleStrip data: stream contents materialIndex:1).
	].
	
	"Faces"
	norm x:0 y:0 z: 1.0.
	vert x: 0-dx y: dy z: radius.
	texCoord u: 0.0 v: 0.0.
	verts at: totalCount+1 put: vert.
	txtrCoords at: totalCount+1 put: texCoord.
	norms at: totalCount+1 put: norm.

	vert x: 0-dx y: 0-dy z: radius.
	texCoord u: 0.0 v: 1.0.
	verts at: totalCount+2 put: vert.
	txtrCoords at: totalCount+2 put: texCoord.
	norms at: totalCount+2 put: norm.

	vert x: dx y: 0-dy z: radius.
	texCoord u: 1.0 v: 1.0.
	verts at: totalCount+3 put: vert.
	txtrCoords at: totalCount+3 put: texCoord.
	norms at: totalCount+3 put: norm.

	vert x: dx y: dy z: radius.
	texCoord u: 1.0 v: 0.
	verts at: totalCount+4 put: vert.
	txtrCoords at: totalCount+4 put: texCoord.
	norms at: totalCount+4 put: norm.

	"Back face"
	norm x:0 y:0 z:-1.0.
	vert x: 0-dx y: dy z: 0-radius.
	texCoord u: 0.0 v: 0.0.
	verts at: totalCount+5 put: vert.
	txtrCoords at: totalCount+5 put: texCoord.
	norms at: totalCount+5 put: norm.

	vert x: 0-dx y: 0-dy z: 0-radius.
	texCoord u: 0.0 v: 1.0.
	verts at: totalCount+6 put: vert.
	txtrCoords at: totalCount+6 put: texCoord.
	norms at: totalCount+6 put: norm.

	vert x: dx y: 0-dy z: 0-radius.
	texCoord u: 1.0 v: 1.0.
	verts at: totalCount+7 put: vert.
	txtrCoords at: totalCount+7 put: texCoord.
	norms at: totalCount+7 put: norm.

	vert x: dx y: dy z: 0-radius.
	texCoord u: 1.0 v: 0.
	verts at: totalCount+8 put: vert.
	txtrCoords at: totalCount+8 put: texCoord.
	norms at: totalCount+8 put: norm.

	strips at: halfSegs+1 
		put: (TStripData mode: GLQuads data: {totalCount+0 . totalCount+1 . totalCount+2 . totalCount+3} materialIndex:1).

	strips at: halfSegs+2 
		put: (TStripData mode: GLQuads data: {totalCount+7 . totalCount+6 . totalCount+5 . totalCount+4} materialIndex:1).
	
	aMesh vertices: verts normals: norms colors: nil texCoords: txtrCoords strips: strips.
	^ aMesh.! !

!QLozenge class methodsFor: 'defaults' stamp: 'bgf 1/28/2009 17:48'!
defaultExtent
	"The inner extent, of the flat rectangular bit."
	^ 4@4! !

!QLozenge class methodsFor: 'defaults' stamp: 'bgf 1/28/2009 17:47'!
defaultRadius
	"Half the thickness of the lozenge; the radius of the edge-round."
	^ 0.25! !

!QLozenge class methodsFor: 'defaults' stamp: 'bgf 1/28/2009 17:47'!
defaultSegments
	"Number of segments/corner for the rounding-over."
	^ 3! !

!QLozenge methodsFor: 'persist' stamp: 'bgf 7/12/2008 23:23'!
constructFrom: pStream

	super constructFrom: pStream.
	self extent: (pStream readPoint: 'extent' default: extent).
	self radius: (pStream readNumber: 'radius' default: radius).
	self segments: (pStream readNumber: 'segments' default: segments).

! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 17:51'!
extent

	"X-y extent of the lozenge's flat portion
	 (this is the dimension of the slab;
	 the whole is surrounded by a half-cylinder frame of radius 'radius'."
	^ extent.! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 22:11'!
extent: aPointOrV3

	"Set the x-y extent of the lozenge. 
	 Also, if the argument is a V3, set its radius to half the given z,
	 which makes the thickness in z equal to that supplied.
	 (The V3 case gives compatibility with TCube.)"

	(aPointOrV3 isKindOf: Vector3)  ifTrue: [
		self extentX: (aPointOrV3 x) y: (aPointOrV3 y).
		self radius: (aPointOrV3 z / 2).
	] ifFalse: [
		extent := aPointOrV3.
	].
	self shapeChanged.
! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 14:14'!
extentX: x y: y

	extent setX: x setY: y.
	self shapeChanged.
! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 14:14'!
frameBox

	| extent2 |
	extent2 := extent/2.0.
	^TBox origin: (extent2 x + radius) negated @ (extent2 y + radius) negated @ radius negated corner: (extent2 x + radius) @ (extent2 y + radius) @ radius.! !

!QLozenge methodsFor: 'initialize' stamp: 'bgf 1/28/2009 17:50'!
initialize

	super initialize.
	segments := self class defaultSegments. 	" number of segments/corner"
	radius := self class defaultRadius. 		" also 1/2 the thickness of the slab"
	extent := self class defaultExtent. 		" x-y extent of the slab."

! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 14:14'!
isProcedural

	^ true.! !

!QLozenge methodsFor: 'persist' stamp: 'bgf 1/28/2009 12:34'!
persistOn: pStream

	super persistOn: pStream.
	pStream storePoint: 'extent' value: extent default: (self class defaultExtent).
	pStream storeNumber: 'radius' value: radius default: (self class defaultRadius).
	pStream storeNumber: 'segments' value: segments default: (self class defaultSegments).! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 7/13/2008 01:23'!
pick: pointer

	"Adapted from TWindowFrame pick:, adapted to the filled-body structure of the lozenge."

	| dx dy dxn dyn rs sel justDoBody |

	rs := radius * radius.
	dx := extent x/2.0.
	dy := extent y/2.0.
	dxn := dx negated.
	dyn := dy negated.
	sel := pointer copiedSelection.
	
	justDoBody := false.

	"The justDoBody := true cases below are all because 
	 we picked a portion of one of the boundary cylinders that is inside the lozenge body.
	 To get a true pick with the proper selection-point, etc,  we re-pick those cases against 
	 the body cube instead.  And we skip right to that, for a healthy speedup."

	" Top bar "
	(pointer pickRightCylinderFrom: (Vector3 x: dxn y: dy z: 0) to:
							    (Vector3 x: dx y: dy z: 0) radius: radius) ifTrue: [
			pointer selectedPoint y < dy 
				ifTrue:[ justDoBody := true ]
				ifFalse:[ ^ true ]].

	" Bottom bar "
	justDoBody ifFalse: [
		(pointer pickRightCylinderFrom: (Vector3 x: dxn y: dyn z: 0) to:
							    (Vector3 x: dx y: dyn z: 0) radius: radius) ifTrue: [
			pointer selectedPoint y> (dy negated)
				ifTrue:[ justDoBody := true ]
				ifFalse:[ ^ true ]]].

	" Left bar "
	justDoBody ifFalse: [
		(pointer pickRightCylinderFrom: (Vector3 x: dxn y: dy z: 0) to:
							    (Vector3 x: dxn y: dyn z: 0) radius: radius) ifTrue: [
			pointer selectedPoint x> (dx negated)
					ifTrue:[ justDoBody := true ]
					ifFalse:[ ^ true ]]].

	" Right bar "
	justDoBody ifFalse: [
		(pointer pickRightCylinderFrom: (Vector3 x: dx y: dy z: 0) to:
							    (Vector3 x: dx y: dyn z: 0) radius: radius) ifTrue: [
			pointer selectedPoint x< dx
				ifTrue:[ justDoBody := true ]
				ifFalse:[ ^ true ]]].

	justDoBody ifFalse: [
		" Top Left corner"
		(pointer pickSphere: (Vector3 x: dxn y: dy z: 0) radiusSquared: rs) ifTrue: [ ^ true].

		" Top Right corner "
		(pointer pickSphere: (Vector3 x: dx y: dy z: 0) radiusSquared: rs) ifTrue: [^ true].

		"Bottom Right corner "
		(pointer pickSphere: (Vector3 x: dx y: dyn z: 0) radiusSquared: rs) ifTrue: [^ true].

		"Bottom Left corner"
		(pointer pickSphere: (Vector3 x: dxn y: dyn z: 0) radiusSquared: rs) ifTrue: [^ true].
	].

	"Body"
	justDoBody ifTrue: [
		pointer selected: sel		"I.e. forget whichever cylinder pick we found inside."
	].
	(pointer pickBox: (Vector3 x: extent x y: extent y z: (radius*2)) location: 0@0@0) ifTrue: [
		^ true].

	^ false.! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 2/11/2010 17:19'!
privateScale: scale

	super privateScale: scale.
	extent := extent * scale.
	radius := radius * scale.
! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 14:14'!
radius
	
	^ radius! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 4/9/2010 10:22'!
radius: rad

	radius := (rad max: (Float epsilon)).
	self shapeChanged.! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 14:14'!
segments
	
	^ segments.! !

!QLozenge methodsFor: 'accessing' stamp: 'bgf 10/17/2007 14:14'!
segments: sg
	
	segments := sg.! !

!QLozenge methodsFor: 'initialize' stamp: 'bgf 10/17/2007 15:26'!
updateBounds
	super updateBounds.
	boundSphere := TBoundSphere 
						localPosition: 0@0@0 
						radiusSquared: (extent asPoint r / 2.0 + (2*radius)) squared.
	boundSphere frame: self! !

TFrame subclass: #QArrowCursor
	instanceVariableNames: 'arrowHead shaft height userID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QArrowCursor commentStamp: 'das 4/16/2007 14:20' prior: 0!
(c) 2007 Qwaq Inc. 
This is a 3D arrow cursor - shows where the end of the users pointer is, whether or not we have a mouse down event. This is driven by the target object.To use it you must define the following instance method:

#removeUser: user selection: selection

	selection userInfo ifNotNil:["remove the frame that is in this slot"
		selection userInfo delete. 
	].

and inside of #pointerOver: you need to send:

	QArrowCursor setArrowCursor: event frame: self.
!

!QArrowCursor class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QArrowCursor class methodsFor: 'use' stamp: 'bgf 9/3/2008 14:59'!
setArrowCursor: event frame: target

	| lastSelection myIcon avatar myFrame myPoint trans |

	avatar := event avatar.
	avatar ifNil:[^ self ]. "don't bother without a local avatar"

	lastSelection := target currentUser at: event userID.
	lastSelection ifNotNil:[
		myIcon := lastSelection userInfo.
	].
	myIcon ifNil:[ 
		myIcon := avatar arrowCursor.
	].
	event selection userInfo: myIcon.
	myFrame := event selection frame.
	myPoint := event selection point.

	myFrame addChild: myIcon.
	"Adjust for distance and my view size... 
	heuristic, gives stable size, about twice the size of the black pointer."
	event selection distance isNaN ifFalse: [
		myIcon height: ( ((event selection distance * 50) / (avatar viewExtent length)) min: 5.0 max: 0.01).
	] ifTrue: [
		myIcon height: 0.4
	].

	trans := target globalTransform inverseByTranspose * (avatar laser globalTransform orientation addRotationAroundX:-45).
	trans translation:myPoint + (0@0@0.01).
	myIcon localTransform: trans. 
! !

!QArrowCursor methodsFor: 'accessing' stamp: 'das 4/15/2007 19:29'!
color: col

	| myColor |
	col ifNil:[ myColor := VectorColor red:0.5 green:0.5 blue:0.5 alpha: 0.5.]
	ifNotNil:[
		myColor := col asVectorColor copy.
		myColor alpha:0.5.
	].
	arrowHead colorize: myColor.
	shaft colorize: myColor.! !

!QArrowCursor methodsFor: 'render' stamp: 'das 4/15/2007 19:50'!
hasAlpha

	^ true.! !

!QArrowCursor methodsFor: 'render' stamp: 'das 4/15/2007 19:50'!
hasOpaque

	^ false.! !

!QArrowCursor methodsFor: 'accessing' stamp: 'bgf 2/21/2008 12:39'!
height
	^ height! !

!QArrowCursor methodsFor: 'accessing' stamp: 'bgf 2/21/2008 12:39'!
height: aScalar
	height ~= aScalar ifTrue: [
		height := aScalar.
		self shapeChanged.
	].! !

!QArrowCursor methodsFor: 'initialize' stamp: 'bgf 2/21/2008 12:40'!
initialize

	super initialize.
	arrowHead := TConic new.
	shaft := TCylinder new.
	self height: 0.5.! !

!QArrowCursor methodsFor: 'render' stamp: 'bgf 2/21/2008 22:46'!
setupAlpha: ogl

	| current distance subDistance |

	"Should I appear at all? Some users choose not to see their own."
	(ogl view show3DPointerFor: userID) ifFalse: [ ^ self ].

	current := ogl currentScale. 
	distance :=  (frameParent child: self setDistance: ogl).
	subDistance := (frameParent childSubdistance:self).
	self root addAlphaObject: arrowHead
		transform: self
		offset: ogl offsetShift
		scale: current
		distance: distance
		subDistance: subDistance
		parent: frameParent
		transparency: 0.5		"We know what we want."
		hilite: nil.				"We are already a sort of hilite."

	self root addAlphaObject: shaft
		transform: self
		offset: ogl offsetShift
		scale: current
		distance: distance
		subDistance: subDistance
		parent: frameParent
		transparency: 0.5		"We know what we want."
		hilite: nil.				"We are already a sort of hilite."
! !

!QArrowCursor methodsFor: 'initialize' stamp: 'bgf 2/21/2008 12:39'!
updateShape
	| h1 h2 |

	h1 := height *3.0/5.0.
	h2 := height - h1/3.0.
	arrowHead height: h1.
	arrowHead topRadius:0.0.
	arrowHead baseRadius: height/6.0.
	arrowHead location: 0@(h1 negated)@0.
	shaft height: h2.
	shaft translation: 0@(height/-2.0)@0.
	shaft radius:height/10.0.
	shaft location:0@(height negated)@0.! !

!QArrowCursor methodsFor: 'accessing' stamp: 'bgf 2/21/2008 22:45'!
userID: aString
	"Session id of the avatar whose pointer this... used to choose which ones to render."
	userID := aString! !

TFrame subclass: #QEditDragIcon
	instanceVariableNames: 'drag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QEditDragIcon commentStamp: 'das 12/11/2006 00:52' prior: 0!
QEditDrag is just some sugar to show the user that selecting on this surface will allow him to drag the object in that plane.!

!QEditDragIcon class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEditDragIcon methodsFor: 'accessing' stamp: 'das 12/11/2006 00:46'!
extent: ext

	drag extent: ext.
! !

!QEditDragIcon methodsFor: 'render' stamp: 'das 12/11/2006 00:45'!
hasAlpha

	^ true.! !

!QEditDragIcon methodsFor: 'render' stamp: 'das 12/11/2006 00:45'!
hasOpaque

	^ false.! !

!QEditDragIcon methodsFor: 'initialize' stamp: 'bgf 2/21/2010 16:55'!
initialize

	| mat |
	super initialize.

	mat :=  (TMaterial new)
		texture: (QTextures find: 'Content/Qwaq/Textures/dragArrows2.png');
		ambientColor: (QColors editboxIndicatorColor);
		diffuseColor: (QColors editboxIndicatorColor);
		fullBright: true;
		cullFace: false.
	drag := TRectangle new.
	drag material: mat.! !

!QEditDragIcon methodsFor: 'accessing' stamp: 'das 12/11/2006 00:46'!
material: mat

	drag material: mat.! !

!QEditDragIcon methodsFor: 'render' stamp: 'das 1/11/2008 17:00'!
setupAlpha: ogl

	self root addAlphaObject: drag
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance:  (frameParent child: self setDistance: ogl)
		subDistance: (frameParent childSubdistance:self)
		parent: frameParent
		transparency: 0.6			"We know what we want."
		hilite: nil.					"We are already a sort of hilite."
! !

TFrame subclass: #QEditRingIcon
	instanceVariableNames: 'frontRing backRing'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QEditRingIcon commentStamp: 'das 12/11/2006 00:52' prior: 0!
QEditRing is some UI sugar to show the user that a selection here will allow the user to spin the object around this axis.!

!QEditRingIcon class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEditRingIcon class methodsFor: 'instance creation' stamp: 'bgf 10/19/2008 22:26'!
defaultMaterial

	"the default material for the ring"

	^ QEditBox indicatorMaterial
 		texture: (QTextures find: 'Content/Qwaq/Textures/treads.png');
		yourself! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 19:30'!
capped: bool

	frontRing capped: bool.
	backRing capped: bool.! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 19:25'!
extent: ext

	frontRing extent: ext.
	backRing extent: ext.! !

!QEditRingIcon methodsFor: 'render' stamp: 'das 12/10/2006 19:32'!
hasAlpha

	^ true.! !

!QEditRingIcon methodsFor: 'render' stamp: 'das 12/10/2006 19:37'!
hasOpaque

	^ false.! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 22:21'!
height: ht

	ht ~= frontRing height ifTrue:[
		frontRing height: ht.
		backRing height: ht.
	].! !

!QEditRingIcon methodsFor: 'initialize' stamp: 'bgf 11/1/2009 14:49'!
initialize

	super initialize.
	frontRing := TCylinder new.
	backRing := TCylinder new.
	self capped: false.
	self slices:32.
	self material:QEditRingIcon defaultMaterial.
	self uvScale: 48.0@1.0.
! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'bgf 1/21/2007 15:25'!
material: mat

	frontRing material: mat.
"set the back of the ring to be a lighter alpha"
	backRing material: mat copy. 
	backRing material flipFace: true. 		
	backRing material ambientColor: (mat ambientColor copy alpha:0.4).
	backRing material diffuseColor: (mat diffuseColor copy alpha:0.3).
! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:56'!
privateScale: scl

	frontRing scale: scl.
	backRing scale: scl.! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/11/2006 01:05'!
radius

	^ frontRing radius.! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 22:21'!
radius: rad

	rad ~= frontRing radius ifTrue:[
		frontRing radius: rad.
		backRing radius: rad.
	].! !

!QEditRingIcon methodsFor: 'render' stamp: 'das 1/11/2008 17:00'!
setupAlpha: ogl

	| distance subDistance |
	distance :=  (frameParent child: self setDistance: ogl).
	subDistance := (frameParent childSubdistance:self).
	self root addAlphaObject: frontRing
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance: (distance sqrt - (frontRing radius*1.1 * ogl camera frameScaleSquared))squared
		subDistance: subDistance
		parent: frameParent
		transparency: ogl transparency
		hilite: ogl forceHilite.

"this is not a great distance measure, but it is better than the default"
	self root addAlphaObject: backRing 
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance: (distance sqrt + (backRing radius * 1.1 *ogl camera frameScaleSquared))squared
		subDistance: subDistance
		parent: frameParent
		transparency: ogl transparency
		hilite: ogl forceHilite.
! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 22:22'!
slices: sl

	sl ~= frontRing slices ifTrue:[
		frontRing slices: sl.
		backRing slices: sl.
	].! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 19:29'!
stacks: st

	frontRing stacks: st.
	backRing stacks: st.! !

!QEditRingIcon methodsFor: 'accessing' stamp: 'das 12/10/2006 19:30'!
uvScale: uv

	frontRing uvScale: uv.
	backRing uvScale: uv.! !

TFrame subclass: #QEditScaleIcon
	instanceVariableNames: 'frontBox centerBox backBox extent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QEditScaleIcon class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEditScaleIcon methodsFor: 'accessing' stamp: 'das 12/15/2006 13:57'!
hasAlpha

	^ true.! !

!QEditScaleIcon methodsFor: 'accessing' stamp: 'das 12/15/2006 13:57'!
hasOpaque

	^ false.! !

!QEditScaleIcon methodsFor: 'initialize' stamp: 'bgf 11/7/2009 20:35'!
initialize

	super initialize.
	frontBox := TCube new.
	centerBox := TCube new.
	backBox := TCube new.
	extent := 0.2 @ 0.2 @ 0.2.
	self material: (QEditBox indicatorMaterial).! !

!QEditScaleIcon methodsFor: 'accessing' stamp: 'das 12/15/2006 13:58'!
material: mat

	frontBox material: mat.	
	centerBox material: mat.
	backBox material: mat.! !

!QEditScaleIcon methodsFor: 'render' stamp: 'das 1/11/2008 17:02'!
setupAlpha: ogl 

	| distance distance2 subDistance |
	distance :=  (frameParent child: self setDistance: ogl).
	subDistance := (frameParent childSubdistance:self).
	self root addAlphaObject: frontBox
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance: distance - 0.1
		subDistance: distance2
		parent: frameParent
		transparency: ogl transparency
		hilite: ogl forceHilite.

	self root addAlphaObject: centerBox
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance: distance+0.01
		subDistance: subDistance
		parent: frameParent
		transparency: ogl transparency
		hilite: ogl forceHilite.

	self root addAlphaObject: backBox
		transform: self
		offset: ogl offsetShift
		scale: ogl currentScale
		distance: distance + 0.1
		subDistance: subDistance
		parent: frameParent
		transparency: ogl transparency
		hilite: ogl forceHilite.
! !

!QEditScaleIcon methodsFor: 'accessing' stamp: 'bgf 3/23/2010 12:07'!
translation: trans
	| n |
	"Generate the cascade of 3 boxes.  
	 The middle box superimposes
	 on the original corner box of the editor."

	super translation: trans.
	(trans squaredLength > Float epsilon) ifTrue: [
		n := trans safelyNormalized.
		centerBox ifNotNil: [
			centerBox extent: extent *1.05.
			frontBox extent: extent *1.5.
			frontBox location: (n*extent*1.4).
			backBox extent: extent* 0.65.
			backBox location:(n*extent*0.85) negated.
		].
	].
! !

!QEditScaleIcon methodsFor: 'initialize' stamp: 'bgf 11/7/2009 16:44'!
translation: trans extent: ext

	"Generate the cascade of 3 boxes.  
	 The middle box superimposes
	 on the original corner box of the editor."

	extent := ext clone.
	self translation: trans.
! !

TFrame subclass: #QPickable2D
	instanceVariableNames: 'sphere oglForPick pixelRadius'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QPickable2D commentStamp: '<historical>' prior: 0!
A base class for 2D (bitmap-rendered) elements that allows them to be picked
(approximately) by approximating their bounds based on camera distance.

The problem is that a raster 2D glyph has no meaningful extent inside the space;
its apparent size depends on the camera distance to the viewer who's trying to pick it.
If the user picks in the region of screen occupied by
the 2D glyph, he expects that to pick the object.   

But inside the space, the screen region's size relative to the TFrame tree
depends on how far the pickable is from the point of view;
since the screen size is constant, the apparent size of the pickable increases
as the distance from viewer to pickable increases.

So here, we fake out the rendering system's coarse grained pick filter
by creating a fixed very large bound sphere, 
and override the fine-grained #pick: with a method 
that estimates in-space bounds to try and give sensible behavior.


!

QPickable2D subclass: #QImage2D
	instanceVariableNames: 'form align disableZ xyOffset material'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QImage2D commentStamp: '<historical>' prior: 0!
A 2-B billboarded form; always displays at a the underlying form's 
pixel size, regardless of orientation or distance from viewer.!

!QImage2D class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QImage2D methodsFor: 'accessing' stamp: 'bgf 10/30/2006 09:15'!
align
	^ align! !

!QImage2D methodsFor: 'accessing' stamp: 'bgf 1/9/2007 23:45'!
align: aSymbol
	align :=  aSymbol.
	self shapeChanged.! !

!QImage2D methodsFor: 'accessing' stamp: 'ar 7/7/2006 11:29'!
disableZ
	^disableZ! !

!QImage2D methodsFor: 'accessing' stamp: 'ar 7/7/2006 11:29'!
disableZ: aBool
	disableZ := aBool! !

!QImage2D methodsFor: 'accessing' stamp: 'ar 7/7/2006 11:28'!
form
	^form! !

!QImage2D methodsFor: 'accessing' stamp: 'bgf 12/9/2006 15:42'!
form: aForm
	form := aForm asFormOfDepth: 32.
	self pixelRadius: ((form width max: form height) min: 16).
! !

!QImage2D methodsFor: 'accessing' stamp: 'ar 7/7/2006 15:29'!
hasAlpha
	^true! !

!QImage2D methodsFor: 'accessing' stamp: 'ar 7/7/2006 15:30'!
hasOpaque
	^false! !

!QImage2D methodsFor: 'initialize' stamp: 'bgf 6/28/2010 10:23'!
initialize

	super initialize.
	align := #center.
	disableZ := false.
	xyOffset := 0@0.
	material := (TMaterial new)
					color: (VectorColor r: 1 g: 1 b: 1);
					fullBright: true.
! !

!QImage2D methodsFor: 'accessing' stamp: 'bgf 6/28/2010 10:40'!
material
	^ material! !

!QImage2D methodsFor: 'render' stamp: 'das 8/31/2008 13:55'!
renderAlpha: ogl
	| enableZ offset rect |
	form ifNil:[^nil].
	ogl isMiniScene ifTrue: [ ^ self ].

	enableZ := disableZ ifTrue:[ogl glIsEnabled: GLDepthTest] ifFalse:[false].
	disableZ ifTrue:[ogl glDisable: GLDepthTest].
	offset := 0@0.
	align ifNotNil: [
		rect := 0@0 extent: form extent.
		offset := (rect perform: align).
	].
	material enable: ogl.
	"This is a fake glBitmap call to get the form centered"
	ogl glRasterPos3f: 0  with: 0 with: 0.0.
	ogl glBitmap: 0 with: 0 with: 0 with: 0
			with: xyOffset x - offset x
			with: offset y - xyOffset y - (form height)
			with: nil.
	ogl glDrawPixels: form width with: form height 
		with: ogl imagePixelFormat32 with: ogl imagePixelType32 with: form bits.
	enableZ ifTrue:[ogl glEnable: GLDepthTest].
	material disable: ogl.
! !

!QImage2D methodsFor: 'accessing' stamp: 'bgf 10/30/2006 08:35'!
xyOffset
	"Used to offset the label from its frame translation in screen raster coordinates
	 rather than in-space transformed coordinates."

	^ xyOffset! !

!QImage2D methodsFor: 'accessing' stamp: 'bgf 1/10/2007 01:23'!
xyOffset: aPoint
	"Used to offset the label from its frame translation in screen raster coordinates
	 rather than in-space transformed coordinates."

	xyOffset := aPoint.
	sphere localPosition: (0@(xyOffset y negated/20.0)@0).  "Approximate."! !

QImage2D subclass: #QLabel2D
	instanceVariableNames: 'contents styleName backgroundColor textColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QLabel2D commentStamp: '<historical>' prior: 0!
A simple 2D billboarded label.!

!QLabel2D class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLabel2D methodsFor: 'accessing' stamp: 'ar 2/23/2006 17:17'!
align
	^align! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 1/9/2007 23:35'!
align: aSymbol
	align := aSymbol.
	self shapeChanged.! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:31'!
backgroundColor
	^ backgroundColor! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:32'!
backgroundColor: aColor
	backgroundColor := aColor! !

!QLabel2D methodsFor: 'persist' stamp: 'das 10/2/2007 15:49'!
constructFrom:pStream
	super constructFrom:pStream.
	self contents: (pStream readString: 'contents' default: ' - ').! !

!QLabel2D methodsFor: 'accessing' stamp: 'ar 2/23/2006 17:16'!
contents
	^contents! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 1/10/2007 00:07'!
contents: aString
	contents := aString.
	self pixelRadius: ((contents size * 4) max: 18).
	self shapeChanged.

! !

!QLabel2D methodsFor: 'accessing' stamp: 'das 8/31/2008 14:04'!
getBackgroundColor
	^ backgroundColor! !

!QLabel2D methodsFor: 'accessing' stamp: 'das 8/29/2008 16:59'!
getText

	^ self contents.
! !

!QLabel2D methodsFor: 'accessing' stamp: 'das 8/31/2008 14:03'!
getTextColor
	^textColor! !

!QLabel2D methodsFor: 'initialize' stamp: 'das 8/31/2008 14:01'!
initialize
	super initialize.
	contents := 'Label'.
	textColor := Color black.
	backgroundColor := Color yellow alpha:0.8.
	styleName := #AnnotationStyle.
	^self! !

!QLabel2D methodsFor: 'persist' stamp: 'das 10/2/2007 15:48'!
persistOn:pStream
	super persistOn:pStream.
	pStream storeString: 'contents' value: contents default: ' - '.
	! !

!QLabel2D methodsFor: 'render' stamp: 'bgf 11/2/2007 23:09'!
renderAlpha:ogl

	"Mode appropriate for TTextLine-composited Form."
	ogl pushBlend: GLSrcAlpha with: GLOneMinusSrcAlpha.
	super renderAlpha:ogl.
	ogl popBlend.! !

!QLabel2D methodsFor: 'accessing' stamp: 'das 8/31/2008 14:04'!
setBackgroundColor: aColor
	backgroundColor := aColor! !

!QLabel2D methodsFor: 'accessing' stamp: 'das 8/29/2008 16:59'!
setText: aString

	self contents: aString.
! !

!QLabel2D methodsFor: 'accessing' stamp: 'das 8/31/2008 14:03'!
setTextColor: aColor
	textColor := aColor.
	self shapeChanged.! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 1/16/2007 18:43'!
styleName
	"Style names are defined by QwaqStyle.  See QwaqStyle #findTextStyle:"
	^styleName! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 7/7/2007 17:08'!
styleName: aQwaqTextStyleName	

	"Style names are defined by QwaqStyle.  See QwaqStyle #findTextStyle:"
	styleName := aQwaqTextStyleName.
	self shapeChanged.! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:32'!
textColor
	^textColor! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:32'!
textColor: aColor
	textColor := aColor.
	self shapeChanged.! !

!QLabel2D methodsFor: 'accessing' stamp: 'bgf 1/27/2008 09:07'!
updateShape

	| style |

	style := (QStyles find: styleName) ifNil: [QStyles find: #AnnotationStyle].

	self form: ((QTextLine new)
					foreground: (self textColor);
					background: (self backgroundColor);
					style: style;
					message: (contents ifNil: ['QLabel2D']);
					redraw;
				form).
	super updateShape.! !

!QPickable2D class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPickable2D methodsFor: 'accessing' stamp: 'bgf 12/9/2006 15:36'!
boundSphere

	"We preset an enormous bound sphere to defeat the preflight check
	 in the pick system... so that my pick: method is called and I can 
	 do a test based on the camera position of the picking avatar."
	^sphere! !

!QPickable2D methodsFor: 'initialize' stamp: 'bgf 12/9/2006 16:05'!
initialize
	super initialize.
	pixelRadius := 32.
	sphere := TBoundSphere localPosition: 0@0@0 radiusSquared: 100000000.0. 
		"Huge so preflight checks always succeed, and my pick: method gets to decide."

	sphere frame: self.
! !

!QPickable2D methodsFor: 'render' stamp: 'bgf 12/9/2006 15:54'!
pick: ptr
	| rs |
	oglForPick ifNotNil:[
		rs := (oglForPick camera globalPosition - self globalPosition) length * (pixelRadius/1000.0).
		^ptr pickSphere: (sphere localPosition) radiusSquared: rs*rs
	] ifNil: [
		"For software-driven picking without a distance, assume I'm very small."
		^ ptr pickSphere: (sphere localPosition) radiusSquared: (pixelRadius/1000.0).
	].
	^false! !

!QPickable2D methodsFor: 'accessing' stamp: 'bgf 12/9/2006 15:50'!
pixelRadius
	"The approximate 2D dimension from which to simulate a bound sphere...
	 specified in pixels, to match the appearance of the 2D glyph."
	^ pixelRadius! !

!QPickable2D methodsFor: 'accessing' stamp: 'bgf 12/9/2006 16:06'!
pixelRadius: pr
	"The approximate 2D dimension from which to simulate a bound sphere...
	 specified in pixels, to match the appearance of the 2D glyph."
	pixelRadius := pr ! !

!QPickable2D methodsFor: 'render' stamp: 'bgf 12/9/2006 15:52'!
renderFrame: ogl

	"To pick properly, I need access to the camera position,
	 which depends on the avatar/event giving rise to the pick;
	 so I save the ogl here, temporarily, being sure to release it immediately."
	oglForPick := ogl.
	^[super renderFrame: ogl] ensure:[oglForPick := nil].
! !

TFrame subclass: #QTracer
	instanceVariableNames: 'path maxSize bounds minStep color stroke boundingBox'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QTracer commentStamp: '<historical>' prior: 0!
QTracer provides user line-drawing as a form of annotation.!

QTracer subclass: #QRibbonTracer
	instanceVariableNames: 'width delta material normal n dxLast'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QRibbonTracer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 21:57'!
appendPath:pos

	| sz dx p1 p2 |
	path ifNil:[
		path := Vector3Array new. 
		delta := Vector3Array new.
		normal := Vector3Array new.
		boundingBox := TBox new.
	].
	path size > 0 ifTrue:[
		pos = path last ifTrue:[^self].
	].
	path := path copyWith: pos.
	delta := delta copyWith: (0@0@1).
	normal := normal copyWith:(0@1@0).
	delta size > 2 ifTrue:[
		sz := delta size.
		p1 := ((path at:sz)-(path at:sz-1)) normalized.
		p2 := ((path at: sz-2)-(path at:sz-1)) normalized.
		dx := p1 cross: p2.
		dx length > 0 ifTrue:[
			dx normalize.
		] ifFalse:[
			dx := dxLast ifNil:[0@0@1].
		].
		(p1 dot: p2) abs >0.999 ifTrue:[
			n := 0@1@0.
		] ifFalse:[
			n := (p1+p2)normalize negated.
		].
		dxLast ifNotNil:[(dxLast dot:dx)<0 ifTrue:[dx := dx negated].].
		dxLast := dx.
		normal at: sz put: n.
		normal at: sz-1 put: n.
		delta at: sz put: dx.
		delta at: sz-1 put: dx.
		delta size = 3 ifTrue:[delta at:1 put: dx.].
	].
	path size > maxSize ifTrue:[
		path := path allButFirst.
		delta := delta allButFirst.
		normal := normal allButFirst.
	].
	boundingBox growVertex: pos.
	self shapeChanged.! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 19:01'!
color: col

	material color: col.! !

!QRibbonTracer methodsFor: 'persist' stamp: 'das 8/17/2008 21:00'!
constructFrom: pStream

	super constructFrom: pStream.
	delta := (pStream readVector3Array: 'delta' default: nil).
	normal := (pStream readVector3Array: 'normal' default: nil).
	material := (pStream readObject: 'material' default: material).
	material ifNil:[material := TMaterial new].
! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 21:17'!
getColor

	^ material color.! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 18:05'!
getWidth

	^ width*2.0! !

!QRibbonTracer methodsFor: 'initialize' stamp: 'das 8/17/2008 19:02'!
initialize

	super initialize.
	material := TMaterial new.
	material cullFace: false.
	width := 0.5.! !

!QRibbonTracer methodsFor: 'persist' stamp: 'das 8/17/2008 20:59'!
persistOn: pStream

	super persistOn: pStream.
	pStream storeVector3Array: 'delta' value: delta default: nil.
	pStream storeVector3Array: 'normal' value: normal default: nil.
	pStream storeObject: 'material' value: material default: nil.
! !

!QRibbonTracer methodsFor: 'render' stamp: 'das 8/17/2008 21:19'!
render: ogl
	|  pt dx |

	path ifNotNil:[
		material enable: ogl.
		ogl glBegin: GLTriangleStrip.
		1 to: path size do:[:index |
			pt := path at:index.
			dx := (delta at: index)*width.
			n := normal at:index.
			ogl glNormal3fv: n.
			ogl glVertex3fv: pt-dx.
			ogl glNormal3fv: n.
			ogl glVertex3fv: pt+dx.
		].
		ogl glEnd.
		material disable: ogl.
	].! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 21:00'!
reset

	path := nil.
	delta := nil.
	normal := nil.! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 21:01'!
resetPath

	path := nil.
	delta := nil.
	normal := nil.! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 21:15'!
setColor: col

	material color: col.! !

!QRibbonTracer methodsFor: 'accessing' stamp: 'das 8/17/2008 18:05'!
setWidth: wd

	width := wd/2.0.! !

!QTracer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/19/2008 21:59'!
appendPath:pos

	path ifNil:[path := Vector3Array new. boundingBox := TBox new.].
	path size = 0 ifTrue:[ path := path copyWith: pos. ^ self.].

	(path last - pos) length > minStep ifTrue:[
		path := path copyWith: pos.
		path size > maxSize ifTrue:[path := path allButFirst].
	].
	boundingBox growVertex: pos.
	self shapeChanged.! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/6/2008 22:18'!
appendPathx: x y:y z:z
	
	self appendPath:(x@y@z).
! !

!QTracer methodsFor: 'accessing' stamp: 'das 12/30/2003 15:31'!
boundSphere

	^ bounds.! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/6/2008 22:12'!
color

	^ color! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/25/2006 10:50'!
color: col

	color := col.! !

!QTracer methodsFor: 'persist' stamp: 'das 4/6/2008 22:26'!
constructFrom: pStream
	"Persist the receiver"
	super constructFrom: pStream.
	maxSize := (pStream readNumber: 'maxSize' default: 1000).
	minStep := (pStream readNumber: 'minStep' default: 0.1).
	stroke := (pStream readNumber: 'stroke' default: 2.0).
	self setColor: (pStream readColor: 'color' default: self getColor).
	path := (pStream readVector3Array:'path' default: nil).


! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/6/2008 22:12'!
getColor

	^ color! !

!QTracer methodsFor: 'accessing' stamp: 'das 7/28/2008 19:16'!
getMinStep

	^ minStep.! !

!QTracer methodsFor: 'accessing' stamp: 'das 3/20/2008 19:09'!
getWidth
	"Set the line width.  Default is 4."
	^ stroke! !

!QTracer methodsFor: 'initialize' stamp: 'das 7/28/2008 19:15'!
initialize

	super initialize.
	path := nil.
	maxSize := 1000.
	stroke := 2.0.
	boundingBox := TBox min:0@0@0 max:0@0@0.
	bounds := TBoundSphere localPosition:(0@0@0) radiusSquared: 1.
	bounds frame: self.
	bounds box: boundingBox.
	minStep := 0.0.
	color := (1.0@0.0@0.0@1.0)asVectorColor.
! !

!QTracer methodsFor: 'accessing' stamp: 'das 11/17/2003 22:31'!
maxSize: ms

	maxSize := ms.! !

!QTracer methodsFor: 'accessing' stamp: 'das 11/17/2003 11:35'!
path 

	^ path.! !

!QTracer methodsFor: 'persist' stamp: 'das 4/6/2008 22:15'!
persistOn: pStream
	"Persist the receiver"

	super persistOn: pStream.
	pStream storeNumber: 'maxSize' value: maxSize default: 1000.
	pStream storeNumber: 'minStep' value: minStep default: 0.1.
	pStream storeNumber: 'stroke' value: stroke default: 2.0.
	pStream storeColor: 'color' value: self getColor default: ( (1.0@0.0@0.0@1.0)asVectorColor).
	pStream storeVector3Array:'path' value: path default: nil.
	! !

!QTracer methodsFor: 'render' stamp: 'das 4/6/2008 22:16'!
render: ogl

	| lights |

	path ifNotNil:[
		lights := ogl glIsEnabled: GLLighting.
		ogl glDisable: GLLighting.
		ogl glDisable: GLCullFace.
		ogl glColor4fv: color.
		ogl glLineWidth: stroke.
		ogl glBegin: GLLineStrip.
		path do:[ :gp | ogl glVertex3fv: gp.].
 		ogl glEnd.
		ogl glLineWidth: 0.25.
		lights ifTrue:[ogl glEnable: GLLighting].
		ogl glEnable: GLCullFace.
	].
! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/6/2008 22:16'!
reset

	path := nil.! !

!QTracer methodsFor: 'accessing' stamp: 'das 4/6/2008 22:16'!
resetPath

	path := nil.! !

!QTracer methodsFor: 'accessing' stamp: 'das 7/28/2008 18:51'!
setColor: col

	| acol |
	acol := col.
	acol class = Color ifTrue:[acol := acol asVectorColor].
	acol class = TranslucentColor ifTrue:[ acol := acol asVectorColor].
	acol size = 3 ifTrue:[acol := (acol,1)asVectorColor].
	color := acol asVectorColor.	! !

!QTracer methodsFor: 'accessing' stamp: 'das 3/20/2008 19:10'!
setMaxSize: ms

	maxSize := ms.! !

!QTracer methodsFor: 'accessing' stamp: 'das 7/28/2008 19:16'!
setMinStep: ms

	minStep := ms.! !

!QTracer methodsFor: 'accessing' stamp: 'das 3/20/2008 19:07'!
setWidth: aWidth
	"Set the line width.  Default is 4."
	stroke := aWidth.! !

!QTracer methodsFor: 'accessing' stamp: 'bgf 5/28/2006 00:07'!
stroke: aWidth
	"Set the line width.  Default is 4."
	stroke := aWidth.! !

!QTracer methodsFor: 'initialize' stamp: 'das 4/19/2008 21:59'!
updateBounds
	super updateBounds.	"In case I have children."
	bounds localPosition: boundingBox center.
	bounds radius: boundingBox radius. 
	bounds transform: self globalTransform.
! !

!QTracer methodsFor: 'initialize' stamp: 'das 4/19/2008 21:56'!
updateShape

	self updateBounds.! !

!TFrame methodsFor: '*Qwaq-Widgets-Annotations' stamp: 'bgf 2/9/2010 16:32'!
attachToFrame: aFrame atPoint: aPoint normal: aNormal
	"This is called to introduce a new note into the space (e.g. from the harness.)
	 Take the given frame as my reference frame,
	 and attach myself to it at the given point& normal,
	 which are presented in the frame's coordinate space."

	| gp gn |
	gp := aFrame localToGlobal: aPoint.
	gn := (aFrame globalOrientation transformVector:aNormal) safelyNormalized.

	self turnToFaceNormal: gn.
	self translateIntoContactWith: gp alongNormal: gn.

	self root attach: self to: (self outerFrameFor: aFrame).
	! !

!TFrame methodsFor: '*Qwaq-Widgets-Annotations' stamp: 'bgf 8/18/2009 00:04'!
isNoteCollector
	"For deciding when a note collector needs to be informed of attach/detach...
	 True if this object attracts/remembers annotations,
	 or is attached to something that does.
	 Base note collectors should be persisted components,
	 not private sub-parts of other objects, since references to them are persisted. "

	^ self surfaceAttachment
			ifNotNilDo: [ :sfc | sfc isNoteCollector ]
			ifNil: [ false ]! !

!TFrame methodsFor: '*Qwaq-Widgets-Annotations' stamp: 'bgf 2/9/2010 16:32'!
outerFrameFor:aFrame
	"Find the container (attachment) frame around the given TFrame;
	 this is the frame that an attachment  will 'follow' as it moves.
	 Currently this is the nearest note-collecting ancestor,
	 or the outermost non-space ancestor if no note collector is involved."
	| outer |
	(aFrame isNil or: [ aFrame root isNil ]) ifTrue: [
		^ nil
	].
	outer := aFrame.
	[outer isNoteCollector or: [outer parent = outer root]] whileFalse: [
		(outer surfaceAttachment) ifNotNilDo: [ :sa |
			outer := sa
		] ifNil: [
			outer := outer parent
		]].

	"We do not keep reference-frame attachments to non-persistent objects."
	outer shouldPersist ifFalse: [ ^ nil ].
	^outer! !

!TFrame methodsFor: '*Qwaq-Widgets-Annotations' stamp: 'bgf 8/20/2009 15:28'!
showOrHide: bShow forPrincipal: evOrAv
	"Frame notification that a container has changed or will change my visible state.
	 This is for things like applications which should continue to consume resources,
	 and/or count against limit totals, when they are out of sight inside a collection manager.
	 The evOrAv argument is either a Croquet Event (which answers avatar and userID)
	 or a TAvatarReplica, which does the same."! !

!TFrame methodsFor: '*Qwaq-Widgets-Annotations' stamp: 'bgf 2/3/2010 10:35'!
stickers
	"All TFrames that the space attachment manager lists as being attached to me."
	^ self root ifNil: [ #() ] ifNotNilDo: [ : space | space stickersAttachedTo: self ]! !

QAbstractDocument subclass: #QSlideShowDocument
	instanceVariableNames: 'textures title position hname alwaysLooping isPlaying autoStepInterval'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-SlideShow'!

!QSlideShowDocument commentStamp: '<historical>' prior: 0!
-Draft-
An ordered collection of images (TFileTextures) 
packaged as a document (for manipulation with file cards) 
and tied to a viewer application (QSlideShowApp) for embedded display.

Maintains the position of the show as an on-island member,
with signalling of updates ... so participants are always coordinated
and the show's position is preserved across save / load.!

!QSlideShowDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSlideShowDocument class methodsFor: 'inspector' stamp: 'bgf 4/1/2010 14:47'!
addPropertyRows: dialog forFrame: ssdref
	
	"A means to change the slide-show name."
	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Title';
			fetchBlock: [ :docref | docref get: #title ];
			acceptBlock: [ : v :docref | docref future title: v ];
			balloonHelp: 'Title of the slide show';
			extraLong;
			yourself).

	dialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Interval Seconds' ;
			validateBlock: [ : v : f  | 
				(v isNil or: [ v <= 0 ]) ifTrue: [
					self inform: 'Interval must be positive.'.
					false
				] ifFalse: [
					true
				]];
			fetchBlock: [ : docref | docref get: #autoStepInterval ];
			acceptBlock: [ : v : docref | docref future autoStepInterval: v ];
			balloonHelp: 'When advancing automatically,\how many seconds between changes.' withCRs;
			yourself).
			
	dialog addRow: (
		(QBooleanPropertyRow new)
			titleString: 'Always Looping' ;
			fetchBlock: [ : docref | docref get: #alwaysLooping ];
			acceptBlock: [ : v : docref | docref future alwaysLooping: v ];
			balloonHelp: 'Should the slide show advance continually?';
			yourself).

	super addPropertyRows: dialog forFrame: ssdref.! !

!QSlideShowDocument class methodsFor: 'services' stamp: 'bgf 3/12/2009 21:56'!
checkForSlideShowDrop: droppedFiles

	"The argument is a list of FileStreams, from a user drop.
	 Answer: 
		#oneByOne - it the files should be imported individually
		#slides		- if the files should be imported as a slide show.
		#cancel		- if nothing further should be done. 
	"

	| nimages |
	nimages := droppedFiles count: [ : strm | ( ImageReadWriter readerClass: strm ) notNil ].
	nimages < 2 ifTrue: [ ^ #oneByOne ].
	(nimages = droppedFiles size) ifFalse: [
		| msg ans |
		msg := 'Of the ', (droppedFiles size), ' files dropped, ', nimages, ' are images.\\',
				'Make a slide show from the images?'.
		ans := CDialog confirm: msg withCRs
				yes: 'Slide Show'
				no: 'Separate'
				cancel: 'Cancel'
				label: 'Importing Images...'
				centerAt: nil.
		ans ifNil: [ ^ #cancel ].			"Cancel: Do nothing... but claim to have processed the files."
		ans ifFalse: [ ^ #oneByOne ].
	].

	QLogger logInfo: 'Dropped ', nimages, ' images out of ', (droppedFiles size), ' for slide show'.
	^ #slides ! !

!QSlideShowDocument class methodsFor: 'inventory-protocol' stamp: 'bgf 2/7/2010 12:40'!
copy: docRef toActiveForum: participant
	"The in-island slide show is a list of tfile textures...
	 we just have to get those copied."
	| textures |
	(textures := docRef get: #textureArray) ifNil: [ #() ].
	textures do: [ : txtrRef |
		txtrRef isValidRef ifTrue: [
			TFileTexture copy: (txtrRef) toActiveForum: participant
		] ifFalse: [
			self error: 'Connection loss prevented retrieval.'
		]].		
	docRef isValidFarRef ifTrue: [
		docRef future serviceURL: (participant currentForumURL).
	] ifFalse: [
		docRef isValidRef ifTrue: [ docRef serviceURL: (participant currentForumURL)] 
	].! !

!QSlideShowDocument class methodsFor: 'inventory-protocol' stamp: 'bgf 2/7/2010 12:41'!
copy: docRef toUserInventory: participant
	"The in-island slide show is a list of tfile textures...
	 we just have to get those copied."
	| textures |
	textures := docRef get: #textureArray.
	textures do: [ : txtrRef |
		txtrRef isValidFarRef ifTrue: [
			((txtrRef get: #isDocumentInForumStorage) = true) ifTrue: [
				TFileTexture copy: (txtrRef) toUserInventory: participant
			].
		] ifFalse: [
			self error: 'Connection loss prevented retrieval.'
		]].! !

!QSlideShowDocument class methodsFor: 'defaults' stamp: 'bgf 4/1/2010 09:20'!
defaultAlwaysLooping
	^ false! !

!QSlideShowDocument class methodsFor: 'defaults' stamp: 'bgf 4/1/2010 14:41'!
defaultAutoStepInterval
	^ 10   "Seconds"! !

!QSlideShowDocument class methodsFor: 'instance creation' stamp: 'bgf 12/4/2008 19:21'!
defaultSuffix
	"The save-out document format is a c3z including the collection of textures, and the the texture resources."
	^ 'qsz'! !

!QSlideShowDocument class methodsFor: 'instance creation' stamp: 'bgf 12/4/2008 22:22'!
defaultTitle
	"The save-out document format is a c3z including the collection of textures, and the the texture resources."
	^ 'Untitled'! !

!QSlideShowDocument class methodsFor: 'services' stamp: 'bgf 6/15/2009 16:42'!
filenameCompare: n1 with: n2

	"For sorting powerpoint directories, which sometimes
		use 'Slide1 ... Slide9 ... Slide 10 ... Slide15 etc,
	 i.e. omittiing the 0's that would let alpha sorting work."

	| b1 b2 haveDots split1 split2 |
	haveDots := (n1 includes: $.) and: [ n2 includes: $. ].
	haveDots ifTrue: [
	 	b1 := n1 copyUpToLast: $.. 
		b2 := n2 copyUpToLast: $..
	] ifFalse: [
		b1 := n1.
		b2 := n2.
	].

	split1 := b1 stemAndNumericSuffix.
	split2 := b2 stemAndNumericSuffix.
	"If the word-part matches, compare numeric value of suffix."
	split1 first = split2 first ifTrue: [
		^ split1 second <= split2 second
	].

	"Otherwise always fall back to full alphabetic."
	^ b1 lessEqual: b2 caseSensitive: false
		! !

!QSlideShowDocument class methodsFor: 'instance creation' stamp: 'bgf 6/5/2009 11:22'!
forLocalFile: file suffix: ext  sharedName: sharedName inClient: owner
	"Create a slide show by importing all the images from the (image zip archive) file.
	 importing the slides as sharedDocuments/ texture-files for the active forum."
	| archive tmpUnpackDir title |

	title :=  ((file localName includes: $.) 
					ifTrue: [ file localName copyUpToLast: $. ]
					ifFalse: [ file localName ] ).
	[
		file binary.
		archive := ZipArchive new readFrom: file.
	] on: Error do: [ :ex|
		QLogger logWarning: 'Slide show file not a zipped slide show: ', ex description.
		^ nil
	].

	[
		"The entire texture-importer regime assumes it recieves file-streams,
	       (uses the local name, etc, etc..)  so don't fight it, give it what it wants."
		tmpUnpackDir := (FileDirectory tempDirectory
							directoryNamed: (TObjectID new hex)) 
								directoryNamed: title.
		tmpUnpackDir assureExistence.
		tmpUnpackDir recursiveDeleteContents.
		archive extractAllTo: tmpUnpackDir.
	] on: Error do: [ : ex |
		QLogger logWarning: 'Cannot unpack slide show into temp: ', ex.
		^ nil
	].

	[
		"The Zip may contain the files, or the eponymous directory full of files,
	 	 depending on what sort of desktop zipper the user used."
		(tmpUnpackDir directoryExists: title) ifTrue: [
			^ self fromDirectory: (tmpUnpackDir directoryNamed: title) inSession: owner session.
		] ifFalse: [
			"The processing may have induced a new name to make the new show unique.
			 So its inner directory won't match.  See if there is a single directory in there."
			| entries | 
			entries := tmpUnpackDir entries select: [ : e | e name first isAlphaNumeric ].
				"That select filters out __MACOSX__ and other such 'hidden' things that zippers 
				  tend to stuff into archives made from desktop folders."
			((entries size = 1) and: [entries first isDirectory ])
			ifTrue: [
				^ self fromDirectory: (tmpUnpackDir directoryNamed: (entries first name)) 
						inSession: owner session
			].
			"If no other case makes it, assume the zip itself is the file collection."
			^ self fromDirectory: tmpUnpackDir inSession: owner session.
		].
	] ensure: [
		[  tmpUnpackDir deleteLocalFiles. ] on: Error do: [ : ex |
			QLogger logWarning: 'Import cleanup: ', ex description.
		].
	].! !

!QSlideShowDocument class methodsFor: 'instance creation' stamp: 'bgf 3/20/2010 10:08'!
fromDirectory: dir inSession: aForumsSession
	"Create a slide show by importing all the images from the given directory;
	 importing the slides as sharedDocuments/ texture-files for the active forum."
	| doc |
	doc := QSlideShowDocument new.
	doc title: dir localName.
	dir fileNames 
		select: [ : n | n notEmpty and: [ n first isAlphaNumeric ] ]
		thenDo: [ : n |
			[ doc add: (aForumsSession importImageAsTexture: (dir readOnlyFileNamed: n))]
		 	on: Error
		 	do: [ : ex |
				(ex description includesSubString: 'OutOfMemory') ifTrue: [ ex pass ].
				QLogger logWarning: 'Slide-show item ignored, not a picture: ', n.
				]
			].
	^ doc
! !

!QSlideShowDocument class methodsFor: 'instance creation' stamp: 'bgf 3/13/2009 01:16'!
getContent: docRef forClient: owner
	"Mostly to support upload-to-forum."

	| textures archive  |
	archive := ZipArchive new.

	(textures := docRef get: #textureArray) ifNil: [ #() ].
	textures do: [ : txtrRef |
		| img |
		img := TFileTexture getContent: txtrRef forClient: owner.
		img ifNil: [ ^ false ].		"GetContent has done the inform."
		archive addString: img as: (txtrRef get: #shortName).
	].
	^ ByteArray streamContents:[:s| archive writeTo: s]! !

!QSlideShowDocument class methodsFor: 'services' stamp: 'bgf 3/20/2010 10:00'!
isSlideShowZip: file
	| archive extensions dirs |
	"The argument is a FileStream, probably from a user drop.
	 Answer true if the zip matches what we expect in a slide show - a simple collection of images."

	extensions := (ImageReadWriter allTypicalFileExtensions).
	[
		file binary.
		archive := ZipArchive new readFrom: file.
	] on: Error do: [ :ex|
		^ false
	].
	dirs := 0.

	^ (archive members allSatisfy: [ : mbr |
		(mbr fileName isEmpty or: [ mbr fileName first isAlphaNumeric not])
		or: [ mbr isDirectory and: [ (dirs := dirs + 1)  < 2 ] ]
		or: [ mbr fileName includesSubString: '__MACOSX']
		or: [ mbr fileName includesSubString: '.DS_Store' ]
		or: [ extensions anySatisfy: [ :ext | mbr fileName asLowercase endsWith: '.', ext ]]])

				! !

!QSlideShowDocument class methodsFor: 'files' stamp: 'hrs 8/17/2009 16:34'!
openDocumentWebPage: docRef forClient: aUI
	"Open the Web page for the document."
	^aUI openQrlForDocument: docRef! !

!QSlideShowDocument class methodsFor: 'customs' stamp: 'jcg 6/2/2009 12:44'!
processCustomsEntryData: entryData customs: customsPane type: docType path: path
	"We need the whole file; bounce the responsibility back to the 'customsPane'."
	^customsPane
		postProcessEntryData: entryData
		fromFile: path
		sharedName: entryData internedServicePath
		extension: entryData fileExtension asUppercase! !

!QSlideShowDocument class methodsFor: 'services' stamp: 'bgf 3/3/2009 23:15'!
saveDocument: document asFile: newName forClient: owner
	"Slide shows are saved as directories full of pictures."
	| textures outer local origName localDir shortNames |

	"Preflight checks before we clobber pre-existing files.
	 Make sure textures have sha and name, and gather all the names for use below."
	
	document isValidRef ifFalse: [ self error: 'Connection lost or document destroyed.' ].
	(textures := document get: #textureArray) isEmptyOrNil ifTrue: [
		self error: 'Cannot get the image list from slide show.' ].
	shortNames := textures collect: [ : tft |
		tft isValidRef ifFalse: [ self error: 'Connection lost or document destroyed.' ].
		(tft get: #sha) ifNil: [ self error: 'Cannot get image information.' ].
		(tft get: #shortName) ifNil: [ self error: 'Cannot get image information (filename)' ] 	
		].

	"Verify the directory, and empty it of files, since it is likely a previous save of me."
	[
		origName := document get: #title.
		outer := FileDirectory forFileName: newName.
		local := FileDirectory default localNameFor: newName.

		"Here,  a temporary workaround for the lack of a directory option
	 	 in the file directory primitives.  
	       Windows in particular forces choosing within an existing directory,
	 	 so detect if we've done that and step up once before building the subdirectory. "
		(local = origName and: [outer localName = local] and: [ outer exists ])
		  ifTrue: [
			QLogger logInfo: 'Eliding doubled slide show subdirectory: ', local.
			local := outer localName.
			outer := outer containingDirectory.
		].
		"--- end temporary workaround. "

		(outer fileExists: local) ifTrue: [
			outer deleteFileNamed: local.
		].
		(outer directoryExists: local) ifTrue: [
			"Delete files since they aren't part of the new slide show save."
			(outer directoryNamed: local) deleteLocalFiles. 
		].
		localDir := (outer directoryNamed: local) assureExistence.
 		
	] on: Error do: [ :ex|
		self error: 'Cannot create the slide directory.', (String cr), (ex messageText).
	].

	textures doWithIndex: [ : tft : idx |
		tft isValidRef ifFalse: [ self error: 'Connection lost or document destroyed.' ].
		TFileTexture saveDocument: tft asFile: ( localDir fullNameFor: (shortNames at: idx) ) forClient: owner.
	].
	"Protocal - return the local basic name of the directory."
	^  local

	
		
	! !

!QSlideShowDocument class methodsFor: 'undo-redo' stamp: 'bgf 4/2/2010 07:40'!
show: slideShowRef application: eApp toggleLooping: bFlag
	"Undoable action for flipping the always-looping setting of a slide show in an application."

	^ QBlockAction 
		undoBlock: [ slideShowRef future alwaysLooping: (bFlag not) ]
		redoBlock: [ slideShowRef future alwaysLooping: (bFlag) ]
		description: (bFlag ifTrue: [ 'Start Looping' ] ifFalse: [ 'Stop Looping' ])
		requiredFrames: { eApp }
! !

!QSlideShowDocument class methodsFor: 'services' stamp: 'bgf 3/13/2009 14:26'!
uploadForImport: aFile docType: type session: qSession 
	"We reach here when a .qsz or .zip-of-images is dropped."

	^ self forLocalFile: aFile 
		   suffix: (self defaultSuffix) 
		   sharedName: ('sharedDocuments/slideshows/', (aFile localName))
		   inClient: (qSession owner).! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 09:31'!
add: aCardOrTexture

	"Append the image to the end of the show."
	hname := nil.

	(aCardOrTexture isKindOf: QFileCard) ifTrue: [
		aCardOrTexture kind = #image ifTrue: [
			^ textures add: aCardOrTexture fileTexture
		]].

	(aCardOrTexture isKindOf: TFileTexture) ifFalse: [
		self error: 'Only textures may be added to a SlideShow'.
	].

	textures add: aCardOrTexture
! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/1/2010 09:21'!
alwaysLooping
	^ alwaysLooping! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/1/2010 13:49'!
alwaysLooping: bAlwaysLooping
	bAlwaysLooping = alwaysLooping 
	 ifFalse: [
		alwaysLooping := bAlwaysLooping.
		self signal: #alwaysLoopingChanged with: bAlwaysLooping.

		alwaysLooping 
			ifTrue: [ self startPlaying ]
			ifFalse: [ self stopPlaying ].
	].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/2/2010 07:38'!
alwaysLooping: bAlwaysLooping inApplication: eApp forUser: userID
	"Interaction to toggle the looping flag. Since this is persistent state, make it undoable."

	| avatar |
	bAlwaysLooping = alwaysLooping ifTrue: [ ^ self ].

	eApp root ifNotNil: [ : space | 
		avatar := space allUsers detect: [ : av | av userID = userID ] ifNone: [ nil ].
	].
	self alwaysLooping: bAlwaysLooping.
	avatar ifNotNil: [	
		avatar recordAction: #show:application:toggleLooping:
				class: QSlideShowDocument
				arguments: { self . eApp . bAlwaysLooping } .
	].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/1/2010 14:40'!
autoStepInterval
	^ autoStepInterval ! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/1/2010 14:47'!
autoStepInterval: gapSeconds
	gapSeconds = autoStepInterval 
	 ifFalse: [
		autoStepInterval := gapSeconds.
		self signal: #autoStepIntervalChanged with: autoStepInterval.
		self isPlaying ifTrue: [  
			self stopPlaying; startPlaying
		]].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/11/2008 10:16'!
backOne
	"Move the position indicator to the previous slide."

	self position: ((position - 1) max: 1).! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 12/7/2008 14:11'!
canBeCopiedToInventory
	^ true! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 02:23'!
canBeOpenedPrivately
	"Can users do 'Open Private Copy...?' to invoke a QOpenDocExternalJob ?
	 This is offered for most documents.
	 See QOpenDocExternalJob"
	^ self textureArray notEmpty! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 02:27'!
canBeSavedLocally
	"Can users do 'Save Local Copy...?' 
	 Only if we have the information."
	
	^ self textureArray notEmpty 	 ! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'ar 4/16/2011 17:45'!
canCopyContentsOnly
	"Slide shows have special handling associated for copying between Forums"

	^true
! !

!QSlideShowDocument methodsFor: 'persist' stamp: 'bgf 4/1/2010 14:44'!
constructFrom: pStream
	super constructFrom: pStream.
	self title: (pStream readString: 'title' default: nil).
	textures := pStream readObject: 'textures' default: (self textures).
	position := pStream readNumber: 'position' default: 1.
	hname := nil.

	self alwaysLooping: (pStream readBool: 'alwaysLooping' default: (self alwaysLooping)).
	self autoStepInterval: (pStream readNumber: 'autoStepInterval' default: (self autoStepInterval)).! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 16:44'!
contentSummary
	"Shared-doc protocol - short user-meaningful description of this particular document"
	^ self title! !

!QSlideShowDocument methodsFor: 'copying' stamp: 'bgf 2/16/2009 10:00'!
copyFrom: aShow

	"This supports transfers to and from inventory,
	 during which there may be modifications to texture file paths,
	 so we copy each TFileTexture."

	super copyFrom: aShow.
	textures := (QCollection empty).
	aShow textures do: [ :txtr |
		textures add: ( (TFileTexture new) copyFrom: txtr ).
		].
	title := aShow title.
	hname := nil.! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 7/19/2009 15:09'!
currentImageSHA
	"Sha of the current slide, or nil."
	^ self currentTexture ifNotNil: [ : txtr | txtr sha ]
! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 7/19/2009 15:09'!
currentTexture
	"The TFileTexture I mean currently to display."
	^ (textures size < position) 
			ifTrue: [ nil ]
			ifFalse: [ (textures at: position)  ].
! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 14:34'!
fileCardBacking

	^ 'SlideShowCard.png'! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 2/16/2009 09:36'!
fileName
	"Completely fake, so we show up in manifests."
	^ 'sharedDocuments/slideshows/', (self title)! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 1/31/2009 19:10'!
fileTextures
	"Expose the actual slides as my texture-dependencies."
	^ self textureArray! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/11/2008 10:17'!
forwardOne
	"Move the position indicator to the next slide, wrapping at the end."
	| pos |
	pos := (position + 1) min: (textures count).

	self position: pos.! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 12:26'!
hashableName
	"I cache my hashable name as it is not quite trivial to construct..."
	^ hname ifNil: [
		hname := (String streamContents: [ : str |
					self textures do: [ : txtr | str nextPutAll: txtr sha ].
					str nextPutAll: (self title ifNil: ['']).
					str nextPutAll: (self serviceURL ifNil: ['']) ] ).
		].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'hrs 8/11/2009 14:10'!
identifier
	"Used in opening Document Page"
	^'slideshows/', self title! !

!QSlideShowDocument methodsFor: 'initialize-release' stamp: 'bgf 4/1/2010 14:41'!
initialize
	super initialize.
	title := self class defaultTitle.
	suffix := self class defaultSuffix.
	textures := QCollection empty.
	position := 1.
	isPlaying := false.
	alwaysLooping := self class defaultAlwaysLooping.
	autoStepInterval := self class defaultAutoStepInterval.! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 10:24'!
isDocumentInForum: forumURL
	"Is the document in the shared area of the given forum's server-side storage?"
	^  (self serviceURL = forumURL and: [ self isInInventory not ]) 
		or: [
		  self textureArray allSatisfy: [ : t | t isDocumentInForum: forumURL ]]! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 12/7/2008 14:08'!
isDocumentInForumStorage
	"Do I represent resources stored in this forum's doc-storage?"
	^ self textureArray anySatisfy: [ : t | t isDocumentInForumStorage ]! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 12/7/2008 14:09'!
isDocumentInInventory
	"Do I represent resources stored in this forum's doc-storage?"
	^ self textureArray anySatisfy: [ : t | t isDocumentInInventory ]! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 4/1/2010 13:41'!
isPlaying
	^ isPlaying! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 16:44'!
kindSummary
	^ 'Slide Show'! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/12/2008 17:32'!
maxDimensions
	"A point with the largest width and largest height in the texture collection;
	 the minimum returned is 256@256, for stability in the player."
	| result |
	result := 256@256.
	self textures do: [ : txtr |
		result := result max: (txtr extent)
	].
	^ result! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 3/25/2009 00:52'!
nextImageSHA
	"Sha of the next forward slide after the one we're seeing; or nil."
	^ (textures size <= position) 
			ifTrue: [ nil ]
			ifFalse: [ (textures at: position+1) sha ].
! !

!QSlideShowDocument methodsFor: 'persist' stamp: 'bgf 4/1/2010 14:42'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeString: 'title' value: (title) default: nil.
	pStream storeObject: 'textures' value: textures default: nil.

	"A bit iffy - should the current-slide state be part of the saved document?"
	false ifTrue: [	pStream storeNumber: 'position' value: position default: 1 ].

	pStream storeNumber: 'autoStepInterval' value: autoStepInterval default: (self class defaultAutoStepInterval).
	pStream storeBool: 'alwaysLooping' value: alwaysLooping default: (self class defaultAlwaysLooping). ! !

!QSlideShowDocument methodsFor: 'playing' stamp: 'bgf 4/1/2010 14:45'!
playOneStep
	isPlaying ifTrue: [
		self position >= textures count ifTrue: [
			self alwaysLooping ifTrue: [
				self toStart.
			] ifFalse: [
				self stopPlaying.
			]
		] ifFalse: [
			self forwardOne
		]].

	isPlaying ifTrue: [ (self future: (self autoStepInterval * 1000)) playOneStep ].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 22:41'!
position
	^position! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/11/2008 17:29'!
position: anIndex
	"Set the current slide position.  Doesn't wrap, clamps at the top.  Signals imageChanged"

	anIndex ~= position ifTrue: [
		position := (anIndex min: textures count) max: 1.
		self signal: #shortNameChanged.
		self synchronize.
	].
! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 03:17'!
serviceURL: aString
	"Also pass to my sub-documents, 
	 which allows retrieval of slide shows from the lobby,
	 when there are no island properties from which to query the service URL
	 in the various resource-path support protocols."
	super serviceURL: aString.
	self textureArray do: [ : txtr | txtr serviceURL: aString ].! !

!QSlideShowDocument methodsFor: 'protocol' stamp: 'bgf 2/28/2009 17:11'!
shortName
	^ title 	"Not really a file name, but we ain't got one for now."! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/11/2008 10:02'!
size
	"The number of slides"
	^ textures size! !

!QSlideShowDocument methodsFor: 'playing' stamp: 'bgf 4/1/2010 14:45'!
startPlaying
	isPlaying ifFalse: [
		isPlaying := true.
		self signal: #isPlayingChanged with: true.
		(self future: (self autoStepInterval * 1000)) playOneStep
	].! !

!QSlideShowDocument methodsFor: 'playing' stamp: 'bgf 4/1/2010 14:38'!
stopPlaying
	isPlaying ifTrue: [
		isPlaying := false.
		self signal: #isPlayingChanged with: false
	].
	(self island isKindOf: TIsland) ifTrue: [ 
		self island unscheduleMessageMatching: [ : msg |
			msg receiver = self and: [ msg selector == #playOneStep ]
		]].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 8/11/2009 10:56'!
synchronize

	"Can be sent (especially future-sent) to ensure all viewers are live with the current page."

	self signal: #imageChanged 
		 with: self currentImageSHA
		 with: position
		 with: self size
		 with: self maxDimensions.! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 8/11/2009 15:23'!
synchronizeAfter: millis for:sessionID
	"Ask for a refresh-display signal after the given time."
	(self future: millis) synchronizeSession: sessionID! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 8/11/2009 15:25'!
synchronizeSession: sessionID

	"Can be sent (especially future-sent) to ensure all viewers are live with the current page."

	self signal: #sessionSync 
		withArguments: {
			sessionID .
		 	(self currentImageSHA) .
		 	(position) .
		 	(self size) .
		 	(self maxDimensions) }.! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 20:20'!
textureArray 
	"The array of textures comprising the slide show."
	^ textures items! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 16:41'!
textures
	"The QCollection of textures comprising the slide show."
	^ textures! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 2/16/2009 09:58'!
textures: aQCollection

	(textures ~= aQCollection) ifTrue: [
		textures := (QCollection empty) addAll: aQCollection; yourself.
		self signal: #texturesChanged.
		self position: 1.
		hname := nil.
	].! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 16:43'!
title
	"The user-visible title for this collection."
	^ title! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 7/31/2009 01:44'!
title: aName
	"The user-visible title for this show."
	aName ~= title ifTrue: [
		title := aName.
		self signal: #shortNameChanged.	"So QDisplaySheet notices."
		self signal: #frameChanged with: self.	 
			"Generic prop-tool update, nothing to do with video frames!!"
	].
! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/11/2008 09:57'!
toEnd
	"Move the position indicator to the next slide, wrapping at the end."
	self position: (textures size).! !

!QSlideShowDocument methodsFor: 'accessing' stamp: 'bgf 12/11/2008 09:57'!
toStart
	"Move the position indicator to the next slide, wrapping at the end."
	self position: 1.! !

!TEmbeddableApp class methodsFor: '*Qwaq-Widgets-Desktop' stamp: 'bgf 3/5/2009 15:41'!
fileCardBacking
	"Return the name of one of the Qwaq/Cards/<something.png> 
	 file-card icon graphics.   E.g. 'DocumentCard.png'   
	 A custom -Embeddable- app may specify its card-backing name by overriding this."

	^ nil! !

!TEmbeddableApp class methodsFor: '*Qwaq-Widgets-Desktop' stamp: 'bgf 12/3/2008 22:50'!
newForDocument: aDocument extent: rectExtent

	"Protocol backstop for document-bound application launch;
       see QAbstractDocument.   rectExtent is the extent of the displaying Rectangle
	 in TSpace dimensions (i.e. 3-space dimensions, -NOT- pixels);
	 it can be used to guide the aspect ratio of the display texture for the embedded app."

	^ self subclassResponsibility! !

!TEmbeddableApp methodsFor: '*Qwaq-Widgets' stamp: 'bgf 5/27/2010 09:03'!
documentName
	"For activity recording... specifically Merck-requested tracing of document focus activities."
	app isValidFarRef ifTrue: [
		^ app get: #documentName
	].
	^ nil! !

!TEmbeddableApp methodsFor: '*Qwaq-Widgets' stamp: 'bgf 3/13/2009 17:41'!
hasPreview
	"Is there a preview texture for this app? "

	^ session notNil ifTrue: [
		(session contentManager contentWithSHA: (self hashedName)) notNil
	] ifFalse: [ 
		false
	]! !

!TEmbeddableApp methodsFor: '*Qwaq-Widgets' stamp: 'bgf 3/13/2009 17:39'!
savePreview
	"Standard Qwaq grayscale preview generator,
	 writes the preview to the server as a reduced b/w gif,
	 using the app-preview name based on <hashedName> "

	(session notNil and: [display notNil])  ifTrue: [
		(app isValidFarRef 
		 and: [ ((app get: #isDirty) == true) or: [ self hasPreview not]  ]) 
		ifTrue: [
			^ QEmbeddedApp 
					savePreviewForApp: self 
					hashedName: self hashedName 
					inSession: session ]].
	^ nil! !

TBorderRectangle subclass: #Q2DNoteSweeper
	instanceVariableNames: 'resizeNotes pinPoint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!Q2DNoteSweeper commentStamp: '<historical>' prior: 0!
A 2D manipulation tool that implements marquis-selection for a flat note-collecting surface.
See QFixedSurface, especially the code involved with its pointerMove:   event handling.!

!Q2DNoteSweeper class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!Q2DNoteSweeper class methodsFor: 'defaults' stamp: 'bgf 2/9/2010 16:44'!
defaultBorderMaterial
	"The border for an initial marquis-sweep, before we've decided to make a permanent group."
	^ (TMaterial new)
		diffuseColor: #(0.025 0.15 0.4 0.8);
		fullBright: true;
		cullFace: false;
		yourself
! !

!Q2DNoteSweeper class methodsFor: 'defaults' stamp: 'bgf 2/9/2010 14:12'!
defaultBorderWidth
	"The absolute width of the backing border."
	^ 0.02! !

!Q2DNoteSweeper class methodsFor: 'defaults' stamp: 'bgf 2/9/2010 14:12'!
defaultExtent
	"Default size for an empty new group."
	^ 2@2! !

!Q2DNoteSweeper class methodsFor: 'defaults' stamp: 'bgf 2/9/2010 16:45'!
defaultMaterial
	"The material to use for an initial marquis-sweep background."
	^ (TMaterial new)
		diffuseColor: #(0.03 0.15 0.55 0.05);
		fullBright: true;
		cullFace: false;
		yourself
! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 14:12'!
contains: note sweepingFrom: topLeft to: bottomRight orient: orient
	"Private support for marquis rubberbanding.
	 We don't require full overlap-inclusion, just that the center be included.	
         But nonetheless the group will not enclose anything larger than itself."
	| trd lbd |

	(note extent x < self extent x and: [ note extent y < self extent y ])
		ifFalse: [ ^ false ].

	trd := orient transformVector: (note translation - topLeft).
	(trd x > Float epsilon and: [ trd y < Float epsilon ]) 
		ifFalse: [ ^ false ].

	lbd := orient transformVector: (note translation - bottomRight).
	(lbd x < Float epsilon    and: [ lbd y >  Float epsilon ] )
		ifFalse: [ ^ false ].

	^ true! !

!Q2DNoteSweeper methodsFor: 'transform' stamp: 'bgf 2/9/2010 14:57'!
destinationExtent
	"For destination distance-finding."
	^ super destinationExtent *  1.5! !

!Q2DNoteSweeper methodsFor: 'transform' stamp: 'bgf 2/9/2010 14:12'!
destinationStanceFor: anAvatar

	"The stance to apply to the avatar to bring it to me."
	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self.

! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 15:23'!
drag: cornerPt corner: compassPt delta: dcorner
	"Progressive drag, repeatedly called when sweeping out the note group."
	| newSize dt pos topLeft bottomRight origin |
	
	pos := self translation clone.
	origin := self globalToLocal: pinPoint.
	topLeft := origin.
	bottomRight := (cornerPt x) @ (cornerPt y).
	topLeft x > bottomRight x ifTrue:  [
		topLeft := (cornerPt x) @ ( topLeft y ).
		bottomRight := (origin x) @ (cornerPt y)
	].
	topLeft y < bottomRight y ifTrue: [
		topLeft := (topLeft x) @ (cornerPt y).
		bottomRight := (bottomRight x) @ (origin y) 
	].
	newSize := (bottomRight x - topLeft x) @ (topLeft y - bottomRight y).
	dt :=  Vector3 x: (bottomRight x + topLeft x)
		  		  y: (bottomRight y + topLeft y) 
				  z: 0.00001.		"Encouragement to keep from moving away."
	self  translation: (pos + (self localTransform transformVector: dt * 0.5 ));
		extent: newSize;
		shapeChanged.

	self resizeSweep: topLeft to: bottomRight.! !

!Q2DNoteSweeper methodsFor: 'accessing' stamp: 'bgf 2/9/2010 14:12'!
height
	^ self extent y! !

!Q2DNoteSweeper methodsFor: 'initialize' stamp: 'bgf 2/9/2010 14:58'!
initialize
	super initialize.
	self objectName: 'Note Sweep'.
	self
		cullFace: false;
		yourself.
	self solid: false.
! !

!Q2DNoteSweeper methodsFor: 'accessing' stamp: 'bgf 2/9/2010 14:12'!
labelForBrowser
	^ self menuTitle! !

!Q2DNoteSweeper methodsFor: 'menu' stamp: 'bgf 2/9/2010 14:57'!
menuTitle
	^ (self genericName)! !

!Q2DNoteSweeper methodsFor: 'accessing' stamp: 'bgf 2/9/2010 14:49'!
minExtent
	"Corner-resize protocol - stay out of trouble."
	^ 0.35@0.35! !

!Q2DNoteSweeper methodsFor: 'events' stamp: 'bgf 2/9/2010 14:12'!
pointerClick: event
	"No click response just yet."! !

!Q2DNoteSweeper methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:54'!
privateScale: aFactor
	"Keep my layout consistent."
	nomScale := aFactor * nomScale.
	self extent: (self extent * aFactor).! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 14:12'!
resizeBegin: event
	"My resize process is actually a marquis-sweep-selection,
	 Save my initial topLeft as origin, since that's what the marquis sweeps out from."
	self resizeBegin: event 
		forInitialPoint: (self localToGlobal: (self extent x * -0.5) @ (self extent y * 0.5) @ 0).! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 14:55'!
resizeBegin: event forInitialPoint: pp

	pinPoint := pp.
	resizeNotes := nil.! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 14:55'!
resizeCanceled

	resizeNotes := nil.
	pinPoint := nil.! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/10/2010 09:19'!
resizeComplete: event
	"We've done a rubberband-resize. 
	 The discovered items become the avatar's selection."
	event avatar ifNotNil: [ :avatar|
		avatar selectFrames: (resizeNotes ifNil: [#()]).
	].
	resizeNotes := nil.
	pinPoint := nil.
	super resizeComplete: event.! !

!Q2DNoteSweeper methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 16:19'!
resizeSweep: ltl to: lbr
	"Sweep the attachment set the of the surface underneath me, looking for ins and outs."
	| sa sao topLeft bottomRight candidates orient |

	sa := self surfaceAttachment.
	sa ifNotNil: [ 
		candidates := ((sa stickers) copyWithout: self) 
						select: [ : n | n isListedTopic and: [ n visible ]]].
	candidates isEmptyOrNil ifTrue: [ 
		resizeNotes := #().
		^ self ].
	orient := self globalOrientation.
	sao := orient inverse.
	topLeft := self translation + ( sao transformVector: (ltl @ 0) ).
	bottomRight := self translation + ( sao transformVector: (lbr @ 0) ).
	resizeNotes := candidates select: [ : note |
		self contains: note sweepingFrom: (topLeft) to: (bottomRight) orient: orient ].
! !

!Q2DNoteSweeper methodsFor: 'accessing' stamp: 'bgf 2/9/2010 17:22'!
shouldPersist
	^ false! !

!Q2DNoteSweeper methodsFor: 'transform' stamp: 'bgf 2/9/2010 16:34'!
warpThresholdAdjustment
	"We want to drag/arrange them from afar."
	^ 6! !

!Q2DNoteSweeper methodsFor: 'accessing' stamp: 'bgf 2/9/2010 14:12'!
width
	^ self extent x! !

TObject subclass: #QChartData
	instanceVariableNames: 'values labels colors descriptions zeroValue topValue minRange'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QChartData commentStamp: '<historical>' prior: 0!
Chart data is an ordered collection of values, each with corresponding label and text,
and a color favored for that item when drawing charts.  
I.e. this is data to drive simple pie charts or bar charts.!

!QChartData class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:14'!
colorAt: n

	colors size >= n ifTrue: [
		(colors at: n) ifNotNilDo: [ : s | ^ s asVectorColor ]].
	
	^ VectorColor r: 1 g: 0.1 b: 0.02.
	! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 07:53'!
colors
	^ colors asArray! !

!QChartData methodsFor: 'persist' stamp: 'bgf 8/18/2009 17:51'!
constructFrom: pStream
	super constructFrom: pStream.

	self setValues: (pStream readFloatArray: 'values' default: values).
	self setLabels: (pStream readStringSequence: 'labels' default: labels).
	self setDescriptions: (pStream readStringSequence: 'descriptions' default: descriptions).
	self setColors: (pStream readColor4Array: 'colors' default: colors).

	self zeroValue: (pStream readNumber: 'minValue' default: nil ).
	self topValue: (pStream readNumber: 'topValue' default: nil ).
	self minRange: (pStream readNumber: 'minRange' default: nil ).! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/28/2009 15:45'!
count
	^ values size! !

!QChartData methodsFor: 'private' stamp: 'bgf 8/15/2009 08:06'!
dataChanged
	"shapeChanged is signaled when a value, color, or label in the existing set changes."
	self signal: #dataChanged! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:15'!
descriptionAt: barNum

	descriptions size >= barNum ifTrue: [
		(descriptions at: barNum) ifNotNilDo: [ : s | ^ s asString ]].
	
	^ self labelAt: barNum! !

!QChartData methodsFor: 'initialize' stamp: 'bgf 8/15/2009 08:52'!
initialize
	super initialize.

	values := FloatArray new.
	labels := #().
	descriptions := #().
	colors := VectorColor4Array new.
! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:13'!
labelAt: barNum
	labels size >= barNum ifTrue: [
		(labels at: barNum) ifNotNilDo: [ : s | ^ s asString ]].
	
	^ 'Value ', barNum! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:05'!
maxValue 
	"The greatest value in the data array."
	^ values 
		ifEmpty:[ topValue ifNil: [ self zeroValue + 1] ] 
		ifNotEmpty: [ values max ]! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/18/2009 17:48'!
minRange
	"If not nil, the minumum value-range to allow for in a display.
	 (Prevents, e.g., a single vote from looking like a huge tower."
	^ minRange ! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/18/2009 17:49'!
minRange: mrScalar
	"If not nil, the minumum value-range to allow for in a display.
	 (Prevents, e.g., a single vote from looking like a huge tower."
	minRange := mrScalar! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:05'!
minValue 
	"The least value in the data array."
	^  values 
		ifEmpty:[ self zeroValue ] 
		ifNotEmpty: [ values min ]! !

!QChartData methodsFor: 'persist' stamp: 'bgf 8/18/2009 17:49'!
persistOn: pStream
	super persistOn: pStream.

	pStream storeFloatArray: 'values' value: values asFloatArray default: #().
	pStream storeStringSequence: 'labels' value: labels default: #().
	pStream storeStringSequence: 'descriptions' value: descriptions default: #().
	pStream storeColor4Array: 'colors' value: colors default: #().

	pStream storeNumber: 'zeroValue' value: zeroValue default: nil.
	pStream storeNumber: 'topValue' value:  topValue default: nil.
	pStream storeNumber: 'minRange' value: minRange default: nil. 
! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:12'!
setColors: arrayOfVectorColors
	| safe |
	safe :=  (arrayOfVectorColors ifNil: [ #() ]) asArray.
	(colors hasEqualElements: safe) ifFalse: [
		colors := VectorColor4Array new: (safe size).
		colors copyFrom: safe.
		self dataChanged.
	].! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:19'!
setDescription: string atIndex: i
	"Set the label for the value at the given index.  
	 Grows the descriptions array to a sparse array of that length if necessary."
	descriptions size < i ifTrue: [
		descriptions := descriptions forceTo: i paddingWith: nil.
	].
	descriptions at: i put: string.
	self dataChanged.! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:16'!
setDescriptions: arrayOfShortStrings
	| safe |
	safe :=  (arrayOfShortStrings  ifNil: [ #() ]) asArray.
	(descriptions hasEqualElements: safe) ifFalse: [
		descriptions := safe.
		self dataChanged.
	].! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:19'!
setLabel: string atIndex: i
	"Set the label of an -existing- bar"
	labels size < i ifTrue: [
		labels := labels forceTo: i paddingWith: nil.
	].
	labels at: i put: string.
	self dataChanged.! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 07:53'!
setLabels: arrayOfShortStrings
	| safe |
	safe :=  (arrayOfShortStrings  ifNil: [ #() ]) asArray.
	(labels hasEqualElements: safe) ifFalse: [
		labels := safe.
		self shapeChanged.
	].! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:22'!
setValue: value atIndex: i
	"Set the value for the sample at i.  
	 The sample set will grow if needed,  but inefficently."
	| grown |
	grown := false.
	values size < i ifTrue: [
		values := values forceTo: i paddingWith: (self zeroValue).
		grown := true.
	].
	values at: i put: value.
	grown ifTrue: [
		self shapeChanged
	] ifFalse: [
		self dataChanged.
	].! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:05'!
setValues: arrayOfNumbers
	| safe |
	safe :=  (arrayOfNumbers ifNil: [ #() ]) asArray.
	(values hasEqualElements: safe) ifFalse: [
		values := safe.
		self shapeChanged.
	].! !

!QChartData methodsFor: 'private' stamp: 'bgf 8/15/2009 08:06'!
shapeChanged
	"shapeChanged is signaled when the size of the value set changes."
	self signal: #shapeChanged! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:03'!
size
	"Number of samples."
	^ values size! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:40'!
topValue 
	"The value at the top of any display ; the largest value shown "
	^ topValue ifNil: [ self maxValue ]! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:04'!
topValue: aValue
	"The value at the top of any display.
	 May be nil, in which case the maximum current value is shown. "
	topValue := aValue.
	self dataChanged.! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:07'!
valueAt: index
	^ values at: index! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 6/15/2010 11:56'!
values
	"Raw values array"
	^ (values ifNil: [#()]) asArray! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:35'!
zeroValue 
	"If not nil, defines the bottom-line of a chart display."
	^ zeroValue ifNil: [ 0 ]! !

!QChartData methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:41'!
zeroValue: aScalar
	"If not nil, defines the bottom-line of a chart display."
	 zeroValue := aScalar.
	self dataChanged.! !

TPortal3D subclass: #QPortal3D
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Portals'!

!QPortal3D commentStamp: '<historical>' prior: 0!
Qwaq extended portal 3D - 
when a QPortal3D is unresolved,
it attempts to use a  preview-graphic from the session
as the basis for the cube's unresolved-material texture.
!

!QPortal3D class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPortal3D class methodsFor: 'defaults' stamp: 'bgf 9/2/2007 15:03'!
defaultUnresolvedMaterial

	"Underlay for our preview graphics."

	^ (TMaterial new)
		opacity: 0.8;
		texture: (QTextures find: 'portal-void.gif');
		ambientColor: #(0 0 0 0) asVectorColor;
		diffuseColor: #(0.85 0.95 0.95 0.65) asVectorColor;
		emissiveColor: #(0.85 0.95 0.95 0.3) asVectorColor;
		cullFace: true;
		yourself! !

!QPortal3D methodsFor: 'accessing' stamp: 'bgf 9/2/2007 15:05'!
billboardTexture

	"If a billboard texture is set, it is rendered in the portal's place
	 when the portal's destination viewpoint is not resolved.
	 Also, if the session has a cached preview of the target space,
	 that is substituted as the billboard (privately and locally) during render.
	 Default is the 'void' blur texture." 

	^ unresolvedMaterial texture! !

!QPortal3D methodsFor: 'accessing' stamp: 'bgf 9/2/2007 15:04'!
billboardTexture: txtr

	"If a billboard texture is provided, it is rendered in the portal's stead
	 when the portal's destination viewpoint is not resolved.
	 (The billboard is invisible -except- for within the rendering pass of
	 replicants whose local session does not have the portal's viewpoint.)" 

	unresolvedMaterial texture: txtr! !

!QPortal3D methodsFor: 'accessing' stamp: 'bgf 9/2/2007 14:15'!
postcardUrl
	^ self postcardLink ifNotNilDo: [ : pc | pc urlString ]! !

!QPortal3D methodsFor: 'accessing' stamp: 'bgf 9/2/2007 22:28'!
renderAlpha: ogl

	"Extend to provide preview-billboarding for unresolved 3D portals."

	| txtr svcTexture mat uvs uvo |

	resolved ifFalse: [
		txtr := unresolvedMaterial texture.
		mat := cube material.
		uvs := cube uvScale.
		uvo := cube uvOffset.

		"We swap to the unresolvedMaterial if needed.
		 In any case, we take care to undo any changes we make,
		 so that these render-time choices do not affect replicated island state."
		[
			self postcardUrl ifNotNil: [
				svcTexture := ogl view session previewForUrl: (self postcardUrl).
				svcTexture ifNotNil: [unresolvedMaterial texture: svcTexture].
			].
			cube material: unresolvedMaterial.
			cube uvScale: 1@0.7.		"Compensate for non-square previews."
			cube uvOffset: 0@0.15.
			cube renderAlpha: ogl.
		] ensure: [
			cube material: mat.
			unresolvedMaterial texture: txtr.
			cube uvScale: uvs.
			cube uvOffset: uvo.
		].
	] ifTrue: [
		super renderAlpha: ogl.
	].
! !

QPortableObject subclass: #QSimpleToy
	instanceVariableNames: 'doStep mesh meshOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QSimpleToy commentStamp: '<historical>' prior: 0!
Movable, reusable toys that can be kept in toychests.

I describe a movable object whose appearance is a TMesh.
Since imported meshes may not have convenient geometry,
I allow declaration of an origin offset 'meshOffset:'
and will maintain the relationship that my 'mesh' is always
translated that amount from my origin.      This allows toys
to be handled using simple layout assumptions (their origin is at their base.)

Provides a restartable 100ms stepping protocol (start/stop/toggleStepping)
which can be exploited by a subclass implementations of #singleStep

	!

!QSimpleToy class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSimpleToy methodsFor: 'persist' stamp: 'bgf 2/11/2010 11:34'!
constructFrom: pStream

	super constructFrom: pStream.

	"The only direct instance of toy is the treefort wandering shuttle;
	 commenting out the code below stops us from reloading its mesh, 
	 and keeps Treefort testPersistence from crashing."

	(self class = QSimpleToy) ifTrue: [
		self meshOffset: (pStream readVector3: 'meshOffset' default: (self meshOffset)).
		self mesh: (pStream readObject: 'mesh' default: nil).
	].

	"If my initialize/construction process has set up my mesh,
	 then it was hardcoded in initialize.  So re-apply the scale to it."
	self mesh ifNotNil: [self mesh scaleBy: (self scale) ].

	doStep := (pStream readBool: 'isStepping' default: doStep).
	"Will be honored by 'start' when I'm added into the frame tree."! !

!QSimpleToy methodsFor: 'initialize' stamp: 'bgf 1/8/2007 08:25'!
initialize

	|  |
	super initialize.
	self isGroundBound: true.
	self isAttractedToSurfaces: true.
	meshOffset := 0@0@0.
	doStep  := false.! !

!QSimpleToy methodsFor: 'events' stamp: 'bgf 3/28/2009 22:20'!
keyStroke: event

	"Toys have some extra operations... +/- to enlarge and shrink,
	 and 's' to start and stop the basic toy animation steps."

	self motionAdapter isEnabled ifFalse: [
		^ self
	].

	event keyCharacter caseOf: {
		[$r]	 -> [self recoverTranslation].
	} otherwise: [
		super keyStroke: event.
	].! !

!QSimpleToy methodsFor: 'accessing' stamp: 'bgf 7/26/2006 14:08'!
mesh

	"The TMesh I present as my visible model."
	^mesh! !

!QSimpleToy methodsFor: 'accessing' stamp: 'bgf 12/7/2006 13:56'!
mesh: aTFrame

	"The mesh loaded by the ASE loader."

	| |
	self addChild: (mesh := aTFrame).
! !

!QSimpleToy methodsFor: 'initialize' stamp: 'bgf 7/26/2006 12:27'!
meshOffset

	"Translation applied to mesh at creation time, to make the TreefortToy's center
	 correspond with the mesh's useful center.  This service is used so that 
	 imported models (which may have their own positioning conventions)
	 can be handled in a uniform way."

	^ meshOffset! !

!QSimpleToy methodsFor: 'initialize' stamp: 'bgf 2/12/2010 14:31'!
meshOffset: offset
	"Translation applied to mesh at creation time, to make the TreefortToy's center
	 correspond with the mesh's useful center.  This service is used so that 
	 imported models (which may have their own positioning conventions)
	 can be handled in a uniform way."

	meshOffset := offset.
	mesh ifNotNil: [
		mesh translation: (offset * (self scaleMemo))
	].! !

!QSimpleToy methodsFor: 'persist' stamp: 'bgf 3/20/2009 00:55'!
persistOn: pStream

	super persistOn: pStream.

	(self class = QSimpleToy) ifTrue: [
		pStream storeVector3: 'meshOffset' value: meshOffset default: (0@0@0).
		pStream storeObject: 'mesh' value: mesh default: nil.
	].
	pStream storeBool: 'isStepping' value: doStep default: true.! !

!QSimpleToy methodsFor: 'step' stamp: 'bgf 2/12/2010 14:36'!
recoverTranslation
	"Transfer the mesh's internal translation to the toy, recovering draggability;
	Needed until the draggable/movable code can deal with changing internal offsets.
	The physics engine pushes around the mesh, not the container."

	true ifTrue: [ ^self ].  "We have no physics engine."

	mesh ifNil: [ ^ self].
	self translation: 
		(self localTransform transformPoint: (mesh translation - (self meshOffset * (self scaleMemo)))).
	mesh translation: (self meshOffset * (self scaleMemo)).! !

!QSimpleToy methodsFor: 'render' stamp: 'bgf 3/28/2009 23:04'!
renderFrame: ogl
	"This is a cheap performance hack - ignore portable objects through 3D portals."
	(ogl inPortal3D) ifTrue: [
		^ 0
	].
	^ super renderFrame: ogl.! !

!QSimpleToy methodsFor: 'accessing' stamp: 'bgf 7/16/2006 13:15'!
scaleAsToy

	"Reduce the toy to its appropriate 'toy' scale, for indoors."
	^ 0.2! !

!QSimpleToy methodsFor: 'step' stamp: 'bgf 12/14/2006 11:36'!
singleStep
	"Do a single step of my basic animation."

	self recoverTranslation.	"Preserve the mesh offset constraint. 
							 (Counteracts the physics engine)"
! !

!QSimpleToy methodsFor: 'step' stamp: 'bgf 1/8/2007 08:21'!
start

	doStep ifFalse: [
		doStep := true.
		self step.
	].! !

!QSimpleToy methodsFor: 'step' stamp: 'bgf 1/8/2007 08:20'!
step

	doStep ifTrue:[
		self singleStep.
		(self future:100) step.
	].! !

!QSimpleToy methodsFor: 'step' stamp: 'bgf 1/8/2007 08:22'!
stop

	super stop.

	"Stop stepping the default simulation of this toy."
	doStep := false.! !

!QSimpleToy methodsFor: 'step' stamp: 'bgf 1/8/2007 08:22'!
toggleStepping

	doStep ifFalse:[
		self start.
	] ifTrue: [
		self stop.
	].
	! !

QDraggable subclass: #QPlaceMarker
	instanceVariableNames: 'sortString category'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QPlaceMarker commentStamp: '<historical>' prior: 0!
A marker denoting a point of interest,
that contributes to the 'Places' menu of landmarks when it appears in a QSpace
that is visited with the QwaqParticipant UI.

(See also QSpace, and QwaqParticipantUI's 'landmarks' methods.)

In the space, the marker is a translucent arrow;
it is movable by drag and drop, and has a click menu.

Place markers with nonzero pitchYawRoll make the destination 'sticky'.

The 'sortString' property can be used to control the order of place marks in the menu,
so that predefined marks can be added in software to create a 'tab' tour of a demo space.
Placemarks with a nil sortString appear in creation order, after all the sorted entries.

Use marks sparingly, since the menu can grow unwieldy.
!

!QPlaceMarker class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlaceMarker class methodsFor: 'inspector' stamp: 'sn 3/9/2013 13:29'!
addPropertyRows: propertyDialog forFrame: pmref

	self addNameRowIfNeeded: propertyDialog forFrame: pmref.

	propertyDialog addRow: (
		(QChoicesPropertyRow new)
			titleString: 'Places Category'translateMe;
			setChoices: (QDestination categoryChoices);
			fetchBlock: [ : f | f get: #category ];
			acceptBlock: [ : t :f | f future category: t ];
			balloonHelp: 'Where the entry appears in Places'translateMe withCRs;
			inputEnabled: (true);
		yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new )
			titleString: 'Sort String'translateMe;
			fetchBlock: [ : f | (f get: #sortString) ifNil: [''] ];
			acceptBlock: [ :v : f | f future sortString: ((v = '') ifTrue: [nil] ifFalse: [v]) ];
			balloonHelp: 'Override sort key for Place order'translateMe;
		yourself).

	super addPropertyRows: propertyDialog forFrame: pmref
! !

!QPlaceMarker class methodsFor: 'defaults' stamp: 'bgf 9/3/2009 15:06'!
defaultCategory
	^ 'Named Places'! !

!QPlaceMarker class methodsFor: 'as yet unclassified' stamp: 'bgf 1/19/2010 16:04'!
defaultDoesClickSelect
	^ false! !

!QPlaceMarker class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QPlaceMarker class methodsFor: 'defaults' stamp: 'bgf 9/26/2007 02:27'!
defaultExtent
	^ (0.55@0.75@0.12)! !

!QPlaceMarker class methodsFor: 'defaults' stamp: 'bgf 9/8/2009 21:52'!
defaultLabel
	^ 'New Landmark'! !

!QPlaceMarker class methodsFor: 'defaults' stamp: 'bgf 4/22/2007 10:07'!
menuStyle
	"TMenuStyle for editing tool menus -
	 use this style to tie things together visually."

	| c |
	c := QColors placemarkColor asColor.

	^ (QStyles menuStyle)
		borderColor: c duller darker;
		separatorColor: c duller lighter;
		disabledTextColor: c duller darker;
		selectedItemColor: c;
		yourself.! !

!QPlaceMarker methodsFor: 'initialize' stamp: 'bgf 4/14/2009 14:46'!
adoptViewPosition: anAvatar

	self translation: anAvatar translation 
						- (0@(anAvatar footToEyes)@0)
						+ (0@(self extent y/2.0)@0).
	self pitchYawRoll: anAvatar pitchYawRoll.
	anAvatar parent addChild: self.! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 9/3/2009 15:10'!
asPlace
	"The QPlace descriptor for this destination."
	^	QPlace 
			category: (self category)
		  	label:  self label 
		  	frame: self
		 	sticky: (self globalTransform pitchYawRoll x abs > 0) 	"Sticky if at an odd angle" .! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 10/25/2009 15:01'!
canEditMaterial
	^ false! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 9/3/2009 15:07'!
category
	^ category ifNil: [ self class defaultCategory ]
! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 9/3/2009 15:07'!
category: placeCategory
	category := placeCategory ifNil: [ self class defaultCategory ].
! !

!QPlaceMarker methodsFor: 'menus' stamp: 'bgf 7/18/2009 11:09'!
comeHere: event

	"Interactive pull-me-to-the-placemark menu action."
	event avatar ifNil: [ ^ self ].
	"Menu action."

	QActivityCapture action: #GoToPlaceMark details: (self label) forSession: (event avatar sessionHex).
	event avatar allowedToMove ifTrue: [
		event avatar goToPlace: (self asPlace)
	]! !

!QPlaceMarker methodsFor: 'persist' stamp: 'bgf 9/3/2009 15:31'!
constructFrom: pStream
	super constructFrom: pStream.
	self sortString:(pStream readString: 'sortString' default: nil).
	self category: (pStream readString: 'category' default: (self category)).! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 3/11/2010 16:44'!
contentsFromEditor: aString
	| tmp |
	self hiliteBriefly.

	"Strip leading and trailing space."
	tmp := aString.
	[tmp isEmpty not and: [tmp last isSeparator]] whileTrue: [
		tmp := tmp allButLast.
	].
	[tmp isEmpty not and: [tmp first isSeparator]] whileTrue: [
		tmp := tmp allButFirst.
	].
	self label: tmp
! !

!QPlaceMarker methodsFor: 'menus' stamp: 'bgf 10/25/2009 15:11'!
contextMenuMaskForAvatar: avatar
	"Remove the system 'Attach Note' command from the menu."
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	mask remove: #attachNote ifAbsent: [].
	mask remove: #approach ifAbsent: [].		"Reordered to top in custom commands"
	mask remove: #copyToInventory ifAbsent: [].	"Meaningless."
	^ mask! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 7/23/2009 13:05'!
destinationStanceFor: anAvatar
	"So right-click on a place marker takes you there."
	| xform pyr |

	xform := self globalTransformReal clone.
	xform a24: xform a24 + (anAvatar footToEyes).		" Y "

	pyr := xform pitchYawRoll.
	pyr z: 0.
	xform pitchYawRoll: pyr.		"No roll... we only rock."

	^Stance here: (self root freePlaceFor: anAvatar near: xform translation)
		       there: (xform translation - (xform lookAt* 100.0))! !

!QPlaceMarker methodsFor: 'events' stamp: 'bgf 10/25/2009 15:08'!
destroySoon
	"See TThumbtack #destroySoon."
	self parent ifNotNil: [
		self parent removeChild:self.
	].
	(self future:30000) destroy.! !

!QPlaceMarker methodsFor: 'menus' stamp: 'bgf 9/8/2009 21:51'!
editLabel: anAvatar

	QTextEditJob edit: (self label) 
				title: 'Landmark Name' 
				forUser: anAvatar 
				recipient: self 
				selector: #contentsFromEditor:.

	"Flicker the cube to reinforce the association."
	self hiliteBriefly.
! !

!QPlaceMarker methodsFor: 'menus' stamp: 'bgf 2/6/2007 15:40'!
editLabelForMenuEvent:event
	"Given a croquet event, have the ui of the event avatar open a text editor."
	self editLabel:event avatar
! !

!QPlaceMarker methodsFor: 'menus' stamp: 'sn 12/27/2012 16:04'!
fillContextMenu:  menu forEvent: event
	menu addItemSpec:  {'Approach' translateMe. #comeHere: } target: self.
	self isPlacementLocked ifFalse: [
		menu addItemSpec: {'Edit Name...' translateMe . #editLabelForMenuEvent: }  target: self.
	].! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'jjl 8/7/2008 15:00'!
freeRange

	^true! !

!QPlaceMarker methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:34'!
initialize

	super initialize.

	self addChild: ((QArrowHead new)
			extent: (self class defaultExtent);
			material: QMaterials placemarkMaterial;
			rotationAroundX:-80;
			solid: false;
			yourself).

	self category: self class defaultCategory.
	self solid: false.
	adapter isGroundBound: true.
	adapter isAttractedToSurfaces: true.
	adapter isMovable: true.
	adapter downColor: (QGotoButton downHilite).
	adapter overColor: (QGotoButton overHilite).
	self label: self class defaultLabel.
! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 3/15/2008 17:57'!
isListedLandmark
	"Place markers are put into the places menu 
	 using a different search than other self-proclaimed landmarks,
	 to allow custom tab-orderings."
	^ false! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 7/21/2008 13:35'!
label
	"The place marker label - also its TFrame object name,
	 except we use a special marker to represent nil... 
	 which TFrame doesn't return as an object name."
	^ (self objectName = 'anUnlabeledPlaceMarker') ifTrue: [
		nil
	] ifFalse: [ 
		self objectName
	].! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 7/21/2008 13:23'!
label: aString
	"The menu item label."
	self objectName: (aString ifNil: ['anUnlabeledPlaceMarker']).! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 9/8/2009 22:16'!
labelForBrowser
	^ 'Landmark: ', (self label ifNil: [ 'Unnamed' ])! !

!QPlaceMarker methodsFor: 'menus' stamp: 'bgf 9/8/2009 22:16'!
menuTitle	
	 ^'Landmark'! !

!QPlaceMarker methodsFor: 'persist' stamp: 'bgf 9/3/2009 15:09'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeString: 'sortString' value: self sortString default: nil.
	pStream storeString: 'category' value: self category default: self class defaultCategory.
! !

!QPlaceMarker methodsFor: 'events' stamp: 'bgf 10/25/2009 15:06'!
pointerClick: event
	"Up-event from the drag adapter when no drag was performed."
	event avatar allowedToMove ifTrue: [
		event avatar goToPlace: (self asPlace).
	]
! !

!QPlaceMarker methodsFor: 'events' stamp: 'bgf 1/28/2008 21:39'!
pointerEnter: event
	self pointerOver: event! !

!QPlaceMarker methodsFor: 'events' stamp: 'bgf 10/25/2009 15:06'!
pointerOver: event
	event avatar ifNil: [ ^ self ].
	event avatar showFloatMessage: ('Landmark: ', self label) forEvent: event.! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 2/8/2007 22:09'!
sortString
	^ sortString! !

!QPlaceMarker methodsFor: 'accessing' stamp: 'bgf 2/8/2007 22:11'!
sortString: aString

	"Provided so that hardcoded named places can take precedence
	 over interactive ones, and so that sort order can be preserved for demo tabbing.
	 When the places menu is assembled for a space, items with a non-null
	 sort string are placed first, sorted alphabetically by their sort string."

	sortString := aString! !

!QPlaceMarker methodsFor: 'events' stamp: 'bgf 12/25/2008 12:52'!
warpThresholdAdjustment
	"Place markers are very small, and offer go-to as part of clicking them...
	 so greatly increase this, otherwise it is difficult to get a normal click menu."
	^ 4.0! !

QPlaceMarker subclass: #QPlaceToken
	instanceVariableNames: 'place'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QPlaceToken commentStamp: 'hrs 3/11/2009 21:54' prior: 0!
Any object can carry one (or more) of these (as a child) in order to show up in the places list.

This is indepent of the isListedLandmark/asPlace protocol, and therefore useful for things like a Dropbox panel that should appear in places as whatever document is being displayed (using the isListedLandmark/asPlace protocol), and also appear in the listing with a fixed name.!

!QPlaceToken class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlaceToken class methodsFor: 'instance creation' stamp: 'bgf 6/22/2009 21:50'!
consistingOf: aTFrame named: aString
	"Answer a place token named aString for the given frame.
	 (Caution - the token is added to the children list of the frame,
	 but will not persist itself properly.  
	 Should only be used by components
	 tha recreate it when unpickled, 
	 and that exclude it when persisting their children.) "

	| token place |
		
	"Clone the target's native place so if it has custom flags or stance methods,
	 the token-place acts the same way as the original."
	place := aTFrame asPlace 
				 ifNotNilDo: [ : origPlace | origPlace clone ]
				 ifNil: [ (QPlace new) frame: aTFrame; yourself ].
	place category: 'Named Places';
		   label: aString.
	token := self new.
	token place: place.
	token visible: false.
	aTFrame addChild: token.
	^token! !

!QPlaceToken methodsFor: 'accessing' stamp: 'hrs 3/11/2009 17:34'!
asPlace
	"Answer the place for the object holding this token."
	^place! !

!QPlaceToken methodsFor: 'accessing' stamp: 'hrs 3/11/2009 18:14'!
place: aQPlace
	place := aQPlace! !

TPortal subclass: #QPortal
	instanceVariableNames: 'billboard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Portals'!

!QPortal commentStamp: '<historical>' prior: 0!
Portal extended to use session-cached preview textures
when its destination island is unresolved.
All that decision-making happens privately at render time.

We add an always-present TRectangle child (#billboard)
that is invisible -except- for a private moment during render.
When the portal is unresolved,
the billboard is declared visible during renderFrame:
and set to use a cached preview texture,
and then immediately reset to its replicated normal state.

The net result is a QGate portal that behaves identically 
in all replicated matters, but shows previews for those clients
that have cached them and have not resolved that space yet.
!

!QPortal class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPortal methodsFor: 'accessing' stamp: 'bgf 5/29/2009 11:21'!
beMirrorWhenUnresolved
	"We use the billboard child when unresolved,
	 so we don't want the TPortal mirror rendering.
	 Except for the legacy uses of display sheets as mirrors,
	 when they carried a Mirror url."

	^ self postcardUrl ifNotNilDo: [ : url | (url beginsWith: '//') not   "Relative" ]
					   ifNil: [ false ]! !

!QPortal methodsFor: 'initialize' stamp: 'das 12/7/2007 17:31'!
initialize
	super initialize.
	billboard := QPortalBillboard new.
	billboard portal: self.
	self addChild: billboard.! !

!QPortal methodsFor: 'accessing' stamp: 'bgf 4/27/2007 22:23'!
postcardUrl
	^ self postcardLink ifNotNilDo: [ : pc | pc urlString ]! !

!QPortal methodsFor: 'accessing' stamp: 'bgf 3/29/2007 04:26'!
updateShape
	super updateShape.
	billboard ifNotNilDo: [:bb|bb extent: self extent ]! !

TTaskMain subclass: #QDrawTask
	instanceVariableNames: 'startFrame startPosition path island color stroke'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QDrawTask commentStamp: 'das 4/24/2006 17:26' prior: 0!
TDrawTask is used to create a line drawing over the surface of a world or an object.!

!QDrawTask class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDrawTask methodsFor: 'accessing' stamp: 'das 4/25/2006 10:46'!
color: col

	color := col.! !

!QDrawTask methodsFor: 'accessing' stamp: 'bgf 2/11/2007 10:39'!
cursor
	^ Cursor marker! !

!QDrawTask methodsFor: 'events' stamp: 'das 3/24/2008 10:56'!
doPointerDown: event

	| selection target norm |
	selection := event selection.
	target := selection target.
	target ifNotNil:[
		(target get:#isInAvatar) ifFalse:[
			startFrame := target future getOwnedObject: selection frameID.
			startPosition := selection point ifNil:[self halt].
			norm := selection normal.
			norm ifNotNil:[norm:= norm*0.1] ifNil:[norm:=0@0@0].
			island := target island.
			path := island future new: QTracer.
			path future appendPath: startPosition+norm.
			path future color: color.
			path future stroke: stroke.
			startFrame future addChild: path.
		].
	].
! !

!QDrawTask methodsFor: 'events' stamp: 'das 4/24/2006 21:20'!
doPointerEnter: event

! !

!QDrawTask methodsFor: 'events' stamp: 'bgf 12/8/2006 12:16'!
doPointerLeave: event

	"Clear this task."
	self doPointerUp: event.! !

!QDrawTask methodsFor: 'events' stamp: 'bgf 8/21/2008 12:06'!
doPointerMove: event

	| selection target  nextFrame nextPosition n1 n2 norm |
	selection := event selection.
	target := selection target.
	target ifNotNil:[ island = target island ifTrue:[
		(target get:#isInAvatar) ifFalse:[
			nextFrame := target future getOwnedObject: selection frameID.
			nextPosition := selection point.
			norm := selection normal.
			norm ifNotNil:[norm:= norm*0.1] ifNil:[norm:=0@0@0].

			nextPosition := nextPosition+norm.
			n1 := nextFrame future localToGlobal: nextPosition.
			n2 := startFrame future globalToLocal: n1. 
			path future appendPath: n2. "add to the path"
		].
	].].! !

!QDrawTask methodsFor: 'events' stamp: 'bgf 11/16/2006 17:53'!
doPointerOver: event

	"We should only remain in place as long as the mouse is down."
	(view notNil and: [view task = self]) ifTrue: [ 
		view popTask.
	].! !

!QDrawTask methodsFor: 'events' stamp: 'bgf 12/8/2006 12:17'!
doPointerUp: event

	self doPointerMove: event.

	(view notNil and: [view task = self]) ifTrue: [ 
		view popTask.
	].! !

!QDrawTask methodsFor: 'initialize-release' stamp: 'bgf 10/6/2006 14:54'!
initialize

	super initialize.

	color := Color white.
	stroke := 4.0.! !

!QDrawTask methodsFor: 'events' stamp: 'das 4/24/2006 14:06'!
noPickWhileDown

	^ false.! !

!QDrawTask methodsFor: 'accessing' stamp: 'bgf 5/28/2006 00:09'!
stroke: aWidth

	"Set the line stroke for the traced line. Default is 4.0"
	stroke := aWidth.! !

TTaskMain subclass: #QEditTask
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QEditTask class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEditTask methodsFor: 'controls' stamp: 'bgf 2/11/2007 10:39'!
cursor
	^ Cursor crossHair! !

!QEditTask methodsFor: 'events' stamp: 'bgf 6/1/2009 09:55'!
doPointerDown: event

	| selection target |
	view owner playSoundEffect:#softClick.
	selection := event selection.
	target := selection target.
	target ifNil:[
		^ view owner statusAlert: 'There is nothing there to edit.'.
	].
	
	view readyForEvents ifTrue: [
		view avatar replica future
			editClick: target 
			subframeID: (selection frameID) 
			forcingDirect:  (event shiftPressed or:[event controlKeyPressed])
		].! !

!QEditTask methodsFor: 'events' stamp: 'das 12/23/2006 01:08'!
doPointerUp: event

	(view notNil and: [view task = self]) ifTrue: [ 
		view popTask.
	].! !

!QEditTask methodsFor: 'initialize' stamp: 'das 12/11/2006 10:53'!
initialize

	super initialize.

! !

CTextParagraphEditor subclass: #QTextAppParagraphEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-TextApp'!

!QTextAppParagraphEditor commentStamp: '<historical>' prior: 0!
Captures and forwards editing requests to the embedded app.!

QTextAppParagraphEditor defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="embeddedProxy" toSet="embeddedProxy:" changeEvent="embeddedProxyChanged">embeddedProxy</field>
</fields>'!

!QTextAppParagraphEditor class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'ar 9/17/2008 18:11'!
backspace
	"Handle backspace in one future send instead of two"
	embeddedProxy ifNotNil:[
		(embeddedProxy sendBackspace) ifTrue:[^self].
	].
	^super backspace! !

!QTextAppParagraphEditor methodsFor: 'accessing' stamp: 'ar 6/2/2008 20:09'!
embeddedProxy
	"Answer the embeddedProxy of the receiver"
	<bewareOf: #embeddedProxyChanged>
	^self propertyValueAt: #embeddedProxy! !

!QTextAppParagraphEditor methodsFor: 'accessing' stamp: 'ar 6/2/2008 20:09'!
embeddedProxy: aValue
	"Modify the receiver's embeddedProxy"
	^self propertyValueAt: #embeddedProxy put: aValue with: #embeddedProxyChanged! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'das 12/7/2009 12:11'!
justSelectAll
	"don't replicate"
	^super selectFrom: 1 to: paragraph text size+1! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'das 12/7/2009 12:01'!
justSelectFrom: markIndex to: pointIndex
	"Just select the text yet but post it as a future message"
	^super selectFrom: markIndex to: pointIndex! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'das 12/7/2009 11:57'!
justZapSelectionWith: aText

	"don't replicate or undo"
	^super zapSelectionWith: aText! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'ar 6/2/2008 20:57'!
selectFrom: markIndex to: pointIndex
	"Do not actually select the text yet but post it as a future message"
	embeddedProxy ifNotNil:[
		(embeddedProxy sendSelectFrom: markIndex to: pointIndex) ifTrue:[^self].
	].
	^super selectFrom: markIndex to: pointIndex! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'ar 7/2/2007 11:16'!
selectWordAt: characterIndex
	"Select the word at the given character index. Do not select entire paragraphs or text since this can cause problems with multiple people interacting and potentially deleting each others work."
	| interval |
	interval := self selectWordIn: paragraph string at: characterIndex.
	self selectFrom: interval first to: interval last+1.! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'ar 7/2/2007 20:32'!
selectWordIn: aString at: characterIndex
	"Select the word at the given character index. Do not select entire paragraphs or text since this can cause problems with multiple people interacting and potentially deleting each others work."
	| here start direction stop level hereChar |
	here := characterIndex.
	stop := aString size.
	"Never select when at first or last text position"
	here <= 1 ifTrue:[^1 to: 0].
	here >= stop ifTrue:[^stop+1 to: stop].

	"no delimiters -- select a token"
	direction := -1.
	level := 1.
	[level > 0 and: [direction > 0
			ifTrue: [here < aString size]
			ifFalse: [here > 1]]] whileTrue:[
		hereChar := aString at: (here := here + direction).
		"token scan goes left, then right"
		hereChar tokenish ifTrue: [
			here = 1 ifTrue:[
				start := 1.
				"go right if hit string start"
				direction := 1.
			].
		] ifFalse: [
			direction < 0 ifTrue:[
				start := here + 1.
				"go right if hit non-token"
				direction := 1.
			] ifFalse: [level := 0].
		]
	].
	level > 0 ifTrue: ["in case ran off string end"	here := here + direction].
	direction > 0
		ifTrue: [^start to: here - 1]
		ifFalse: [^here + 1 to: stop].! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'ar 12/30/2007 10:46'!
showCaret
	markBlock ifNil:[^false].
	pointBlock ifNil:[^false].
	markBlock = pointBlock ifFalse:[^true]. "always display if selected"
	^costume wantsCaretShowingIn: self! !

!QTextAppParagraphEditor methodsFor: 'editing' stamp: 'ar 6/2/2008 20:56'!
zapSelectionWith: aText
	"Lowest level entry point - bounce it back to the embedded app from here"
	embeddedProxy ifNotNil:[
		(embeddedProxy sendReplaceSelectionWith: aText) ifTrue:[^self].
	].
	^super zapSelectionWith: aText! !

Object subclass: #QEmbeddedOverlay
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Qwaq-Widgets-TextApp'!

!QEmbeddedOverlay commentStamp: 'das 9/13/2007 21:51' prior: 0!
Render a TEmbeddedApp in a nice viewable location on top of everything else. The cursor positions of the overlay and underlying TEmbeddedApp match up. To use this, simply put your cursor over a TEmbeddedApp in the 3D world and hold down the shift key.!

!QEmbeddedOverlay class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEmbeddedOverlay methodsFor: 'rendering' stamp: 'das 9/13/2007 19:07'!
beginOverlay: ogl
	"Set up gl to provide overlay drawing - this is done in pixel coordinates"

	ogl glMatrixMode: GLProjection.
	ogl glPushMatrix.
	ogl glLoadIdentity.
	ogl glMatrixMode: GLModelview.
	ogl glPushMatrix.
	ogl glLoadIdentity.
	ogl glTranslated: -1 with: 1 with: 0.0.
	ogl glDisable: GLDepthTest.
	"ogl pushBlend: GLOne  with: GLOneMinusSrcAlpha."
	ogl pushBlend: GLSrcAlpha with: GLOneMinusSrcAlpha.
 
! !

!QEmbeddedOverlay methodsFor: 'rendering' stamp: 'das 9/13/2007 19:06'!
endOverlay: ogl
	"Set up gl to provide overlay drawing - this is done in pixel coordinates"
	ogl popBlend.
	ogl glMatrixMode: GLModelview.
	ogl glPopMatrix.
	ogl glMatrixMode: GLProjection.
	ogl glPopMatrix.
	ogl glMatrixMode: GLModelview.
! !

!QEmbeddedOverlay methodsFor: 'rendering' stamp: 'das 8/27/2008 20:16'!
renderOverlay: ogl

	| frame tw th box sha txtr extent formExtent frameExtent point dx dy ep |
false ifTrue:[^self].
ep := ogl view eventPointer.
ep ifNotNil:[ "no event pointer?"
	ogl view event controlKeyPressed  ifFalse:[^ self]."is shift pressed?"
	ep selected frame ifNotNil:[ "is there a selected frame?"
		(ep selected frame get: #isEmbeddedApp) ifTrue:[ "is it an embedded app?"
			frame := ep selected frame.
			sha := frame get:#defaultTextureSHA.
			sha ifNil:[^ self ]. "no sha found to dereference"
			self beginOverlay: ogl.
			txtr := ogl installTextureSHA: sha. "set up the texture"
			extent := frame get:#appExtent.
			frameExtent := frame get:#extent.
			formExtent := txtr form extent.
			ogl glScaled: (2.0 / ogl extent x) with: (-2.0 /ogl extent y) with: 1.0.
			ogl glEnable: GLTexture2d.
		"this is where we map the 2D overlay position of the embedded app to the 3D position in the scene" 
			point := ep selected point.
			dx :=   ogl view eventPosition x - ((point x asFloat/frameExtent x +0.5)*extent x).
			dy :=ogl view eventPosition y - ((0.5 - (point y asFloat/frameExtent y))*extent y).
			ogl glDisable: GLLighting.
			ogl glColor4f: 1.0 with: 1.0 with: 1.0 with: 0.5.

			tw := extent x asFloat/formExtent x.
			th := extent y asFloat/formExtent y.
			box := (dx@dy) corner: (extent+(dx@dy)).

			ogl glBegin: GLQuads;
				glTexCoord2f: 0.0 with: th;
				glVertex2f: box left with: box bottom;

				glTexCoord2f: tw with: th;
				glVertex2f: box right with: box bottom;

				glTexCoord2f: tw with: 0.0;
				glVertex2f: box right with: box top;

				glTexCoord2f: 0.0 with: 0.0;
				glVertex2f: box left with: box top;

				glEnd.
			self endOverlay: ogl.
			ogl glEnable: GLLighting.
		].
	].
].
! !

QDestination subclass: #QFixedDisplay
	instanceVariableNames: 'box sheet floatButton menuButton floatHandler motionAdapter resize menuButtonHandler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QFixedDisplay commentStamp: '<historical>' prior: 0!
This is a simplified display-sheet holder without many controls, 
for installing formal preset displays
at special places in forum templates.

The appearance is a simple brightly-colored slab 
with the sheet on the front.
(The base material of the slab can be edited;
the default material is the glowing-blue slab used in the welcome forums.)

The display is a destination (yellow goto-triangle)
with a built-in 'floating window' button.
Basic document actions are offered in a click menu on the frame itself.!

!QFixedDisplay class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFixedDisplay class methodsFor: 'inspector' stamp: 'sn 3/10/2013 21:51'!
addPropertyRows: propertyDialog forFrame: dispref

	self addNameRowIfNeeded: propertyDialog forFrame: dispref.
	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ : f | f get:#extent3D ];
			acceptBlock: [ :v3 : f | f future extent3D: v3 ];
			balloonHelp: 'Dimensions of the fixed frame.'translateMe;
	 	yourself	).

	propertyDialog addRow: (
		(QBoolean3PropertyRow newForFlag: 'Drop Enabled' flag: 'Listed' flag: 'Empty')
			fetchBlock: [ :f | { (f get: #isDropEnabled) = true . 
						   (f get: #isListed) = true .
						   (f get: #contents) value isNil } ];
			acceptBlock: [ : b :f | 
					f future isDropEnabled: b first.
					f future isListed: b second. 
					b third ifTrue: [ f future clear ]];
			balloonHelp: 'Can content be dropped in?\Am I listed in Places?\\Checking Empty clears out my content.'translateMe withCRs;
		yourself).

	super addPropertyRows: propertyDialog forFrame: dispref.

	"Always a viewing destination."
	propertyDialog removeRow: 'Destination Type'.
	propertyDialog removeRow: 'Seating Range'.
	propertyDialog removeRow: 'One-Way Invisible Direction'.
	propertyDialog removeRow: 'Places Category'.
	propertyDialog removeRow: 'Go-To Button Position'.! !

!QFixedDisplay class methodsFor: 'undo-redo' stamp: 'bgf 4/2/2010 21:40'!
clearDisplay: dref oldCard: pickle
	"Record the undoable clear of a fixed-display."
	^ QBlockAction
		undoBlock: [ : avatar | dref future restoreCard: pickle forAvatar: avatar ]
		redoBlock: [ : avatar | dref future clear ]
		description: 'Clear'
		requiredFrames: { dref }.! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 00:55'!
defaultButtonPosition
	^ Vector3 x: 0 y: 2.75 z: -0.125! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 8/17/2009 23:20'!
defaultCategory
	^ 'Panels'! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 12/31/2008 13:17'!
defaultDestinationMode
	^ #front! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 8/17/2009 23:38'!
defaultExtent
	^ Vector3 x: 5 y: 4 z: 0.125! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 8/17/2009 23:07'!
defaultIsPlacementLocked
	^ true! !

!QFixedDisplay class methodsFor: 'defaults' stamp: 'bgf 12/31/2008 15:13'!
defaultMaterial
	^ QMaterials fixedDisplayMaterial! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 4/2/2010 21:28'!
addMenuActions: aMenu forEvent: event

	"Build the display's left-click menu."

	sheet contentCard content ifNotNil: [
		sheet contentCard addMenuActions: aMenu forEvent: event.
		aMenu menuItems size > 0 ifTrue: [
			aMenu menuItems last separator: true.
		].
	].

	"Drop-Enabled really means can-be-modified, 	
	 so use to control access to Clear also."
	(aMenu add: 'Clear' target: self action: #clear: )	
			enabled: (self isDropEnabled and: [sheet contentCard content notNil]).
	! !

!QFixedDisplay methodsFor: 'transform' stamp: 'bgf 7/22/2009 13:50'!
asPlace
	"Extended to update the hasApp: flag."
	^ (super asPlace)
		hasApp: (sheet notNil and: [ sheet hasEmbeddedApp] );
		cameraSelector: #cameraStanceFor:;
		yourself! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 2/3/2010 10:42'!
asTopicRecord	
	"The record presenting me as a line-item in a topics-tool."
	self document ifNotNilDo: [ : doc |
		^ (QTopicRecord new)
			cardRef: self;
			author: (self document ctime who);
			color: (Color black);
			creationTime: (doc ctime time);
			lastModifiedTime: (doc mtime time);
			topic: ('Doc: ', doc shortName);
			nComments: (self stickers) size;
		yourself
	].

	^ (QTopicRecord new)
			cardRef: self;
			author: '';
			color: Color black;
			creationTime: 0;
			lastModifiedTime: 0;
			topic: 'Empty Display';
			nComments: (self stickers) size;
		yourself
! !

!QFixedDisplay methodsFor: 'hierarchy' stamp: 'bgf 7/26/2009 13:13'!
bestSeatWithin: tbox nearest: avatarPosition
	"Backstop."
	^nil! !

!QFixedDisplay methodsFor: 'testing' stamp: 'bgf 1/1/2009 12:58'!
childrenForBrowser
	"Present as a single object; the parts are not user-relevant."
	^ #().! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 12/31/2008 16:08'!
clear
	"Remove contents."
	sheet clear! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 4/2/2010 21:50'!
clear: event
	"Remove contents."
	| writer |

	sheet isEmpty ifTrue: [ ^ self ].

	writer := TSceneWriter new.
	writer preserveObjectIDs: true.

	event avatar recordAction: #clearDisplay:oldCard:
				class: QFixedDisplay
				arguments: { self . writer storeSceneAsZip: (sheet contentCard) }.

	self clear.! !

!QFixedDisplay methodsFor: 'persist' stamp: 'bgf 3/15/2010 16:38'!
constructFrom: pStream
	| tmp |
	super constructFrom: pStream.
	self extent3D: (pStream readVector3: 'extent' default: (self extent3D)).
	box material: (pStream readObject: 'material' default: (self material)).
	box checkAlpha.
	self isDropEnabled: (pStream readBool: 'isDropEnabled' default: (self isDropEnabled)).
	tmp := (QFileCard new).
	tmp constructContentFrom: pStream.
	tmp isRetiredDemoContent ifTrue: [
		tmp clear
	].
	sheet displayCard: tmp from: (pStream sessionID).
	pStream preserveObjectIDs ifTrue: [
		(pStream readString: 'sheetOID' default: nil) ifNotNil: [ : hex | 
			sheet objectID: (TObjectID readHexFrom: hex)
		]].
! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 8/17/2009 22:52'!
consumeDrop: delivery avatar: avatar selection: aTXSel
	"Drops pass through to the display sheet"
	^ sheet consumeDrop: delivery avatar: avatar selection: aTXSel! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 1/1/2009 13:24'!
contents
	"The image rectangle, embedded app, or what-have-you,
	 currently displaying."
	^ sheet contents! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 4/3/2010 09:20'!
convertToPanel: event
	"See also QPoster, which has the same method."
	| panel prePair |

	sheet contentCard ifNotNil: [	
		sheet contentCard saveForEvent: event.
	].

	prePair := { self objectID . (event avatar captureFrames: { self }) }.

	panel := QDisplayPanel withSimpleFrame.
	panel localTransform: (self globalTransform);
		extent: (self extent + (0.1@0.2));
		isPlacementLocked: false;
		isFromTemplate: false.

	panel translation: (panel translation + (self lookAt * -0.05)).
	panel displayCard: (sheet contentCard) fromUser: event avatar.
	self root 
		addChild: panel;
		attach: panel to: (self surfaceAttachment).

	panel future recordConversion: prePair forAvatar: event avatar.
	self obliterateLater.! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 2/5/2010 17:12'!
copyToInventory: event
	self document ifNotNilDo: [ : doc |
		event avatar ifNotNilDo: [ :av | 
			av preCapture: { self }.
			av addContentToInventory: doc. 
		]
	] ifNil: [
		super copyToInventory: event
	]! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 1/1/2009 12:59'!
defaultMaterial
	"So the material editor changes me properly."
	^ box defaultMaterial! !

!QFixedDisplay methodsFor: 'transform' stamp: 'bgf 8/18/2009 01:42'!
destinationExtent
	"So we get consistent placement
 	 between cameraStanceFor: and destinationStanceFor: "

	^ (self extent) * (QGotoButton defaultSlopFactor)! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 3/19/2010 11:43'!
displayCard: aCard fromUser: anAvatar
	"Take the content of the given QFileCard as my display, 
	 on behalf of the given user.   The avatar parameter identifies the principal,
 	 and this form also indicates that the action originated from a live user request."
	sheet displayCard: aCard fromUser: anAvatar! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/17/2009 22:36'!
document
	"The QAbstractDocument I currently present, if any, or nil if I'm not showing one."
	^ sheet document 
! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/17/2009 22:35'!
dragAdapter
	^ motionAdapter! !

!QFixedDisplay methodsFor: 'annotations' stamp: 'bgf 8/17/2009 22:35'!
editByUser: replica
	"Protocol infill for treating panels as attachments to topic collectors;
	 this is the response to the Topic Tool 'Open' (and topic-list double-click) command."
	sheet hasEmbeddedApp ifTrue: [
		 replica signal: #floatForEmbeddedApp with: (sheet childContent)
	] ifFalse: [
		replica goToPlace: (self asPlace)
	].
! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/18/2009 01:41'!
extent
	^ box extentX @ box extentY! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/18/2009 01:38'!
extent3D
	"My extent including thickness"
	^ box extent! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/18/2009 01:38'!
extent3D: extent3

	box extent ~= extent3 ifTrue: [
		box extent: extent3.
		self shapeChanged.
	].! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/18/2009 01:54'!
extent: extent2or3
	"Either my width+height, or my entire 3d box extent."
	extent2or3 isArray ifTrue:  [
		^ self extent3D: extent2or3
	].
	box extentX: extent2or3 x y: (extent2or3 y) z: (box extentZ).
	self shapeChanged.
	! !

!QFixedDisplay methodsFor: 'menu' stamp: 'sn 12/28/2012 20:29'!
fillContextMenu: aMenu forEvent: event
	"The menu commands for a poster; 
	 available only if you have modify-access."

	event avatar canModifyForum ifFalse: [ ^ self ].
	self isPlacementLocked ifTrue: [ ^ self ].
	self contents ifNotNil: [
		aMenu addItemSpec: {'Convert to Panel' translateMe . #convertToPanel: } target: self.
	].

	! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/17/2009 22:53'!
height
	^ box extentY! !

!QFixedDisplay methodsFor: 'hierarchy' stamp: 'bgf 1/31/2009 19:07'!
holdsDocument: aSharedDocument
	"Locate-support for the inspector... return true if I'm showing the document."
	^ (sheet contentCard content = aSharedDocument)! !

!QFixedDisplay methodsFor: 'initialize' stamp: 'bgf 11/27/2009 20:34'!
initialize
	super initialize.
	self objectName: 'Fixed Display'.

	box := TCube new.
	box objectName: 'FixedDisplayFrame'.
	box material: (self class defaultMaterial).
	self addChild: box.

	sheet := QDisplaySheet new.
	sheet objectName: 'FixedDisplaySheet'.
	self addChild: sheet.
	self addChild: sheet contentCard.	" To root it. "
	sheet contentCard visible: false.
	
	floatButton := TSimpleRectangle new.
	floatButton material: (QMaterials expandButtonMaterial).
	self addChild: floatButton.
	floatButton checkAlpha.

	floatHandler := QButtonHandler
		body: floatButton
		target: sheet
		action: #showFloatingAppPane:.
	floatHandler overColor: (QColors controlOverColor).
	floatHandler fireColor: (QColors controlGoColor).
	floatHandler label: 'Floating Window...'.

	menuButton := TRectangle new.
	menuButton material: (QMaterials menuButtonRectangleMaterial clone).
	menuButton material diffuseColor alpha: (0.85).
	self addChild: menuButton.
	menuButtonHandler := QButtonHandler
			body: menuButton
			target: self
			action: #showMenu:
			label: 'Menu'.

	self extent3D: (self class defaultExtent).
	self startScript: #onContentChanged when: {sheet . #contentChanged}.
	motionAdapter := QMotionAdapter forLuggage: self grip: box.
	motionAdapter
		isGroundBound: false;
		isAttractedToSurfaces: true;
		attractionRange: 3.0;	" Factor based on original distance "
		openSpacePolicy: #viewer;
		surfacePolicy: #upright;
		overColor: nil;
	 	downColor: nil.

	resize := QCorner2DResizer forLuggage: self.
	self category: 'Panels'.
	self isPlacementLocked: (self isPlacementLocked).! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 12/31/2008 16:09'!
isDropEnabled
	"Turn off to lock in the content."
	^ sheet isDropEnabled! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 12/31/2008 16:09'!
isDropEnabled: bDE
	"Turn off to lock in the content."
	sheet isDropEnabled: bDE! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 8/18/2009 00:30'!
isDropTarget: payload
	"We can accept anything the display sheet can present."
	^ sheet isDropTarget: payload! !

!QFixedDisplay methodsFor: 'testing' stamp: 'bgf 12/31/2008 12:58'!
isEditableGroup
	^false! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 9/3/2009 14:05'!
isListedLandmark
	"Controllable by a property."
	^ super isListedLandmark and: [ sheet isEmpty not ]! !

!QFixedDisplay methodsFor: 'annotations' stamp: 'bgf 8/17/2009 22:35'!
isListedTopic
	^ true! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 9/2/2009 11:32'!
isPlacementLocked: bLocked
	super isPlacementLocked: bLocked.
	motionAdapter ifNotNil: [
		bLocked ifTrue: [
			motionAdapter enableClickOnly
		] ifFalse: [
			motionAdapter enable.
		]].
 	resize ifNotNil: [
		resize isEnabled: bLocked not.
	]! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 12/31/2008 13:31'!
label
	^ super buttonLabel! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 12/31/2008 13:12'!
label: aString
	super buttonLabel: aString.! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/17/2009 23:16'!
labelForBrowser
	^ 'Display: ', (self buttonLabel)! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/17/2009 22:39'!
lastModifiedTime
	"The topic tool sorts summaries using this."
	^ self document ifNotNil: [ : doc | doc mtime time ] ifNil: [ 0 ]
! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/18/2009 01:46'!
margin
	"Basis for updateShape layout calculations."
	^ (box extent max) * 0.0235! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 1/1/2009 13:02'!
material
	"For the material editor, and persistence.  
	 The box frame material can be changed."
	^ box material! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 1/1/2009 13:02'!
material: m
	"For the material editor, and persistence.  
	 The box frame material can be changed."
	box material: m.
	box checkAlpha.! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 1/11/2009 17:45'!
menuTitle
	"Emphasis the content if any."
	^ (sheet contentCard content) ifNotNilDo: [ :ct| 
		(sheet contentCard shortContentSummary).
	] ifNil: [
		'Fixed Display'
	].! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 9/22/2009 13:13'!
onContentChanged
	"Update to suit the current displayed content."
	
	"Only applications can appear in floating-embedded-app windows."
	floatButton visible: (sheet contentCard isApplicationContent).

	self isDropEnabled ifTrue: [
		self document ifNotNilDo: [ : doc | 
			doc shortName ifNotNilDo: [ :short| self buttonLabel: short ]
				ifNil: [ self buttonLabel: 'Document' ]
		] ifNil: [
			'Empty Display'
		]].

	"So if empty, you can drag anywhere in the blue middle..."
	sheet pickable: (sheet isEmpty not).
	self root ifNotNil: [
		sheet contentCard registerContent.
	].

	"Provoke a refresh in any tools watching the collectors I'm attached to..."
	self signal: #contentsChanged.
	self shapeChanged.! !

!QFixedDisplay methodsFor: 'persist' stamp: 'bgf 3/15/2010 16:42'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeVector3: 'extent' value: self extent3D default: (self class defaultExtent).
	pStream storeObject: 'material' value: self material default: self class defaultMaterial.
	pStream storeBool: 'isDropEnabled' value: self isDropEnabled default: true.
	sheet contentCard persistContentOn: pStream.	
	pStream preserveObjectIDs ifTrue: [  
		"Undo-redo of drops is much easier if sheet ID's survive panel undo/redo"
		pStream storeString: 'sheetOID' value: (sheet objectID hex).
	].! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 8/17/2009 22:40'!
photographerClass
	^ QDirectPhotoJob! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 1/1/2009 13:04'!
placeMenuActions: anAvatar
	"Label-symbol pairs for the Places context menu."
	| list |
	list := OrderedCollection new.
	sheet hasEmbeddedApp ifTrue: [
		list add: { 'Floating Window...' . #showFloatingAppPane: } .
	].
	sheet contentCard content ifNotNil: [ 
		list add: { 'Zoom to Window' . #zoomToWindow100: } .
	].
	^ list asArray.
! !

!QFixedDisplay methodsFor: 'transform' stamp: 'bgf 8/18/2009 01:26'!
positioningBox
	"For materialization, we want to see the display even if we're at exact wall-distance 
	 for it... to be sure, use only the inner bounds to figure distances so we err too close.
	 Helps prevent panel-behind-wall problems."
	^ box boundingBox! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 8/17/2009 22:40'!
reallyDelete: event
	"Forget attachment links ASAP (analogous to un-rooting early)"
	self root ifNotNil: [ self root detach: self ].
	super reallyDelete: event.
! !

!QFixedDisplay methodsFor: 'initialize' stamp: 'bgf 4/2/2010 21:36'!
restoreCard: pickle forAvatar: aUser
	sheet replayDrop: pickle forAvatar: aUser! !

!QFixedDisplay methodsFor: 'hierarchy' stamp: 'hrs 9/30/2009 16:16'!
root: aSpaceOrNil
	super root: aSpaceOrNil.

	"Honor document registration requirements."
	sheet ifNotNil: [
		aSpaceOrNil 
			ifNil: [ sheet contentCard unregisterContent  ]
			ifNotNil: [ sheet contentCard registerContent ].
	].
	aSpaceOrNil ifNil: [^self].
	aSpaceOrNil name: self uniquelyAmongClass: QFixedDisplay.! !

!QFixedDisplay methodsFor: 'menu' stamp: 'bgf 10/25/2009 16:15'!
showContextMenu: event
	(event frame = menuButton) ifTrue: [ 
		"Make right-clicking the menu button itself do the same as left-clcking the MB;
	  	 users can right-click the panel body or title for the 'pure' context menu."
		self showMenu: event
	] ifFalse: [
		super showContextMenu: event
	].! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 1/1/2009 12:54'!
showFloatingAppPane: anEvent
	"Raise the content into  a floating window."
	sheet showFloatingAppPane: anEvent! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 8/20/2009 15:50'!
showOrHide: bShow forPrincipal: evOrAv
	"Frame notification that a collection manager has changed or will change my visible state.
	 We stop 'expensive' apps when hidden,
	 and make sure that all but expensive apps are started when revealed. "
	bShow ifFalse: [
		sheet embeddedApp ifNotNilDo: [ : app | 
			app shouldRunUponLoad ifFalse: [
				self stopRunning: evOrAv action: #Hide
			]]
	] ifTrue: [
		sheet embeddedApp ifNotNilDo: [ : app | 
			app shouldRunUponLoad ifTrue: [
				self startRunning: evOrAv
			]]].! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 8/20/2009 15:39'!
startRunning: eventOrAv
	"Sometimes we start running in response to a gesture, sometimes during load/startup.
	If during load-startup, there's no principal, so eventOrAv will be nil."
	sheet hasEmbeddedApp ifTrue: [
		sheet childContent lastEvent: (
				(eventOrAv class = CroquetEvent) ifTrue: [ eventOrAv ]ifFalse: [nil]).
		sheet childContent startRunning: (eventOrAv ifNotNil: [eventOrAv userID]).
	].! !

!QFixedDisplay methodsFor: 'events' stamp: 'hrs 9/23/2009 14:55'!
stopRunning: eventOrAv action: aSymbol
	"The stop-button has been pressed."
	sheet hasEmbeddedApp ifTrue: [
		sheet embeddedApp isDirty ifTrue: [
			sheet onSaveForUser: eventOrAv userID.
		].
		sheet onQuitForUser: eventOrAv userID action: aSymbol.
		"Take enough time to let the save control-s find its way upstream."
		(sheet childContent future: 1200) stopRunning: (eventOrAv userID).

		sheet recordAction: 'paused' avatar: (eventOrAv ifNotNil: [eventOrAv avatar]).
	].! !

!QFixedDisplay methodsFor: 'hierarchy' stamp: 'bgf 10/25/2009 15:57'!
updateShape
	super updateShape.

	box translationX: 0 y: 0 z: (box extentZ * -0.5).
	sheet extent: ((box extentX) @ (box extentY)) - ((2 * self margin) @ (5 * self margin));
		translationX:0 y: (self margin*0.5)  z: (box extentZ * 0.02).

	floatButton extent: (self margin * 2.7) @ (self margin*2.3);
		translationX: (box extentX negated/2.0  + (1.9 * self margin)) 
				   y: (box extentY negated/2.0 + (1.65 * self margin))
				   z: (box extentZ * 0.02).

	menuButton	extent: (self margin * 4.2) @ (self margin* 1.7);
		translationX: (box extentX negated/2.0  + (2.55 * self margin)) 
				   y: (box extentY /2.0 - (1.25 * self margin))
				   z: (box extentZ * 0.02).

	button arrow extent: (
				QGotoButton defaultExtent 
					* 0.8
					* ((self extent maxDimension / (QFixedDisplay defaultExtent max)) 
						atLeast: 0.1 atMost: 1.8)).
	button translationX:0 
				 y: (self extent y*0.5)+(2*self margin)+(button extent y * 0.66) 
				 z: 0.
	button shapeChanged.

	resize updateShape.
	self updateBounds! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/18/2009 01:32'!
warpThresholdAdjustment
	"Slight increase so these aren't too grabby."

	^ (self isPlacementLocked ifTrue: [ 1.1 ] ifFalse: [1.6])! !

!QFixedDisplay methodsFor: 'accessing' stamp: 'bgf 8/17/2009 22:53'!
width
	^ box extentX! !

!QFixedDisplay methodsFor: 'events' stamp: 'bgf 1/1/2009 11:06'!
zoomToWindow100: evOrAvatar
	"Ask the UI to zoom my display-sheet to be shown as the entire content of the main window."
	sheet zoomToWindow100: evOrAvatar! !

QDestination subclass: #QShelfStack
	instanceVariableNames: 'extent case shelves slats inside shadow'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QShelfStack commentStamp: '<historical>' prior: 0!
A bookshelf, comprising 5 shelf tiers each of which hold a transparent 
document-shelf presenting file cards.    
A builtin update for the old mesh-based fake shelf stack used in library and Project-*forums.
!

!QShelfStack class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QShelfStack class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:35'!
addPropertyRows: dialog forFrame: stackref

	dialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ :stack | stack get:#extent ];
			acceptBlock: [ :v3 :stack | stack future extent: v3 ];
			balloonHelp: 'Dimensions of the shelf case.';
	 	yourself	).

	dialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Number of Shelves';
			fetchBlock: [ :stack | stack get: #nShelves ];
			acceptBlock: [ :n :stack | stack future nShelves: (n asInteger atLeast:1 atMost: 12) ];
			balloonHelp: 'Number of shelves.\At least 1.\At most 12.' withCRs;
		yourself).

	super addPropertyRows: dialog forFrame: stackref.

	"We fix some properties."
	dialog removeRow: 'Destination Type'.
	dialog removeRow: 'Go-To Label'.
	dialog removeRow: 'Go-To Button Position'.
	dialog removeRow: 'Seating Range'.! !

!QShelfStack class methodsFor: 'defaults' stamp: 'bgf 6/24/2009 13:55'!
defaultCaseMaterial
	"Cedar woodgrain material."
	^(TMaterial new)
		texture: (QTextures find:'Content/Qwaq/TexturePalette/woodGradient.jpg');
		ambientColor: #(0.44 0.46 0.34) asVectorColor;
		diffuseColor: #(0.87 0.92 0.76) asVectorColor;
		specularColor: #(0.42 0.43 0.3) asVectorColor;
		yourself
	! !

!QShelfStack class methodsFor: 'defaults' stamp: 'bgf 6/23/2009 16:13'!
defaultDestinationMode
	^ #front! !

!QShelfStack class methodsFor: 'defaults' stamp: 'bgf 6/26/2009 23:15'!
defaultExtent
	^ Vector3 x: 6 y: 6 z: 0.8! !

!QShelfStack class methodsFor: 'defaults' stamp: 'bgf 6/23/2009 18:45'!
defaultNumShelves
	^ 5! !

!QShelfStack methodsFor: 'private' stamp: 'bgf 6/23/2009 19:43'!
addNShelves: numToAdd
	
	| newShelves newSlats |

	"The document-shelves themselves.  These are the invisible card-managers."
	newShelves := Array new: numToAdd.
	1 to: numToAdd do: [ :i|
		newShelves at: i put: (self newShelf) ].

	"The slats are the intermediate shelf surfaces"
	newSlats := Array new: numToAdd.
	1 to: numToAdd do: [ : i |
		newSlats at: i put: (self newSlat) ].
	
	shelves := shelves, newShelves.
	slats := slats, newSlats.
	self shapeChanged.! !

!QShelfStack methodsFor: 'hierarchy' stamp: 'bgf 2/8/2010 13:44'!
addPastedChild: aChild
	"Protocol for <paste> when selected.  Not used, as this is not an editable group."
	(aChild isKindOf: QFileCard) ifTrue: [
		^ (self shelves first) insertCard: aChild
	].
	^ super addPastedChild: aChild! !

!QShelfStack methodsFor: 'testing' stamp: 'bgf 6/23/2009 16:39'!
canEditMaterial
	^ true! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 11/16/2009 13:37'!
childrenForBrowser
	^ shelves asArray! !

!QShelfStack methodsFor: 'persist' stamp: 'bgf 6/23/2009 19:01'!
constructFrom: pStream

	super constructFrom: pStream.
	
	self material: (pStream readObject: 'material'default: self material).
	self extent: (pStream readVector3: 'extent' default: self extent).
	self nShelves: (pStream readNumber: 'nShelves' default: (self nShelves)).
	
	shelves doWithIndex: [ : shelf : i | 
		| cards | 
		cards := pStream readSequence: 'cards-', i default: #().
		cards do: [ :  cd | shelf insertCard: cd ].
	].
! !

!QShelfStack methodsFor: 'hierarchy' stamp: 'bgf 2/14/2010 16:57'!
consumeDrop: delivery avatar: avatar selection: aTXSel
	^ shelves notEmpty and: [
		shelves first consumeDrop: delivery avatar: avatar selection: aTXSel
		]! !

!QShelfStack methodsFor: 'initialize' stamp: 'bgf 11/16/2009 13:37'!
defaultIsAtomic
	^ false! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 15:44'!
defaultMaterial
	^ case first defaultMaterial! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 15:44'!
defaultMaterial: m
	self material: m! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 15:44'!
extent
	^ extent! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 15:40'!
extent: aV3
	extent := aV3.
	self shapeChanged.! !

!QShelfStack methodsFor: 'initialize' stamp: 'sn 1/2/2013 14:41'!
initialize

	super initialize.
	self objectName: 'Shelf Stack'.
	
	"The inside back."
	inside := TSimpleRectangle new.
	inside objectName: 'Shelf Stack'.
	self addChild: inside.

	"The case includes the 5 primary slabs top/bottom/left/right/rear"
	case := Array new: 5.
	case at: 1 put: TRectangle new.

	self addChild: (case first).
	2 to: (case size) do: [ :i | 
			case at: i put: TCube new.
			self addChild: (case at: i) ].

	slats := #().
	shelves := {self newShelf}.

	shadow := QMaterials newRectangularShadow.
	shadow addRotationAroundX: -90.
	self addChild: shadow.

	self buttonLabel: 'Shelf Stack' translateMe.
	self isSticky: false.
	self material: self class defaultCaseMaterial.
	self nShelves: (self class defaultNumShelves).
	self extent: self class defaultExtent.! !

!QShelfStack methodsFor: 'testing' stamp: 'bgf 2/14/2010 16:58'!
isDropTarget: aTFrame
	^ (aTFrame isKindOf: QFileCard)! !

!QShelfStack methodsFor: 'testing' stamp: 'bgf 11/16/2009 13:37'!
isEditableGroup
	^ false! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 22:18'!
isGroundBound
	"So I'm on the floor when created from the Objects menu."
	^ true! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/24/2009 11:17'!
labelForBrowser
	^ (self objectName), ' (Stack)'! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 15:43'!
material
	^ case first material! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/24/2009 11:33'!
material: aMaterial
	
	case do: [ : item | item material: aMaterial ].

	inside material: (
		(aMaterial clone)
			diffuseColor: ((aMaterial diffuseColor * 0.6) alpha: (aMaterial diffuseColor alpha));
			ambientColor: (aMaterial ambientColor * 0.6);
			specularColor: (aMaterial specularColor * 0.5);
			emissiveColor: (aMaterial emissiveColor *0.2);
			yourself).

	slats do: [ : item | item material: (aMaterial clone uvAngle: 90; yourself) ].! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 18:48'!
nShelves
	"The number of shelves I hold (usually 5, but can be altered by property.)"
	^ shelves size! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/23/2009 20:03'!
nShelves: n
	"Change the number of shelves in my body.
	 Cards are moved if needed to avoid losing any.
	 For user-safety, n must be from 1 to 12. shelves."
	| numShelves |
	numShelves := n min: 12 max: 1.
	numShelves > (self nShelves) ifTrue: [
		self addNShelves: (numShelves - (self nShelves)).
	] ifFalse: [
		numShelves < (self nShelves) ifTrue: [
			self removeNShelves: (self nShelves - numShelves).
		]].! !

!QShelfStack methodsFor: 'private' stamp: 'bgf 6/28/2010 11:39'!
newShelf
	|shelf|
	shelf := QDocumentShelf new.
	self addChild: shelf.

	"The material is invisible but with a very faint ghost hilite when moused over."
	shelf baseMaterial: (QMaterials portalGlassMaterial transparency: 0.02; yourself);
		  borderMaterial: (QMaterials clearMaterial);
		  isPlacementLocked: true;
		  objectName: 'Shelf'.
	^ shelf! !

!QShelfStack methodsFor: 'private' stamp: 'bgf 6/23/2009 22:40'!
newSlat
	|slat|
	slat := TCube new.
	self addChild: slat.
	slat material: (self material clone).
	slat uvAngle: 90.
	^ slat! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/24/2009 11:19'!
objectName: aName
	"The destination label is based on this name, so propagate it."
	super objectName: aName.
	inside ifNotNil: [ inside objectName: aName ].
	self buttonLabel: aName.! !

!QShelfStack methodsFor: 'persist' stamp: 'bgf 6/23/2009 19:01'!
persistOn: pStream

	super persistOn: pStream.
	
	pStream storeObject: 'material' value: self material default: (self class defaultCaseMaterial).
	pStream storeVector3: 'extent' value: (self extent) default: (self class defaultExtent).

	"No default for nShelves so there is no risk of content loss if we change the default;
	 we have to know how many shelves to read the card-arrays of."
	pStream storeNumber: 'nShelves' value: (self nShelves).

	shelves doWithIndex: [ : shelf : i | 
		pStream storeSequence: 'cards-', i value: (shelf cards) default: #().
	].
! !

!QShelfStack methodsFor: 'private' stamp: 'bgf 2/11/2010 15:58'!
privateScale: aFactor
	nomScale := nomScale * aFactor.
	self extent: (extent * aFactor).
! !

!QShelfStack methodsFor: 'private' stamp: 'bgf 6/23/2009 20:02'!
removeNShelves: numToRemove
	"Discard a number of shelves, moving their cards into the remaining shelves.
	Private.  There must be shelves remaining!!"

	numToRemove >= self nShelves ifTrue: [
		QLogger logError: 'Attempt to remove all shelves from a ShelfStack.'.
		^ nil 
	].

	"Save the cards into earlier shelves."
	1 to: numToRemove do: [ : i |
		| oldShelf receiver |
		oldShelf := shelves at: (shelves size + 1 - i).
		receiver := shelves at: ((shelves size - numToRemove + 1 - i) max: 1).

		oldShelf cards do: [ : cd |  receiver insertCard: (cd) ].
		oldShelf destroy.
		(slats at: (slats size + 1 - i)) destroy.
	].
	shelves := shelves copyFrom: 1 to: (shelves size - numToRemove).
	shelves size > 1 ifTrue: [
		slats := slats copyFrom: 1 to: (shelves size -1).
	] ifFalse: [
		slats := #()	
	].
	self shapeChanged.

! !

!QShelfStack methodsFor: 'accessing' stamp: 'bgf 6/24/2009 10:51'!
shelves
	"My QDocumentShelf list, which allows examing all my cards."
	^ shelves clone! !

!QShelfStack methodsFor: 'hierarchy' stamp: 'bgf 8/29/2009 18:37'!
updateShape
	| thickness halft halfe spacing cardExtent slots |

	"Heuristics for the shelf-case thickness, don't overpower any of the basic dimensions."
	thickness := (extent x min: (extent y / (self nShelves max: 2))) / 8.5.
	thickness := thickness min: (extent z / 6.0).

	halft := thickness / 2.0.
	halfe := extent/2.0.

	inside extent: (extent x @ extent y) - (thickness*0.1);
		   translation: (0@ (halfe y) @ ((halfe z - (thickness)) negated)).

	case first	 	"Back"
			rotationAroundY: 180;
			extent: (inside extent);
			translation: (0@(halfe y) @ ((halfe z - (thickness*0.03)) negated)).
	
	case second 	"Side"
			extent: (thickness @ (extent y) @ (extent z));
			translation: (halfe x - halft) @ (halfe y) @ 0.
	case third		"Side" 
			extent: (case second extent clone);
			translation: (halfe x - halft) negated @ (halfe y) @ 0.
	case fourth 		"Bottom"
		extent: (extent x - (thickness*1.2)) @ (thickness*2) @ (extent z - (thickness*0.01));
		translation: (0@thickness@0).
	case fifth		"Top"
		extent: (case fourth extent);
		translation: (case fourth translation + (0@(extent y-(2*thickness))@0)).
	
	spacing := (extent y - (3*thickness)) / (slats size+1).
	slats doWithIndex: [ :slat :i |
		slat extent: (extent x - (thickness*1.2)) @ (thickness) @ (extent z - (thickness*0.1));
			translation: (0 @ ((i*spacing) + (thickness + halft)) @ 0).
	].

	cardExtent := (spacing*0.95)@(spacing*0.75).
	slots :=  ((extent x - (2*thickness)) / (cardExtent x)) truncated max: 1.
	slots > 1 ifTrue: [
		cardExtent := (extent x - (4*thickness)) / slots @ (spacing*0.75).
	].

	1 to: shelves size do: [ : i |
		| tier shelf |
		shelf := shelves at: i.
		tier := i - 1.
		shelf basicExtent: cardExtent; numVisibleSlots: slots.
		shelf translation: 0@((tier * spacing) + thickness + (spacing*0.44)) @ (extent z *0.23).
		shelf rotationAroundX: -18.
	].

	shadow extent: (extent x @ extent z);	
		      borderWidth: (thickness*1.25);
		      translationX: 0 y: (thickness*0.075) z: 0.

	super updateShape.
	self buttonPosition: (0@ (extent y + (extent z * 3)) @ 0).
	self updateBounds.! !

QDestination subclass: #QTabbedWall
	instanceVariableNames: 'slab board pole foot cap resizers shadow tabs controls banner title'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QTabbedWall commentStamp: '<historical>' prior: 0!
High-level component comprising a mutliboard and its chrome.
Body is the multiboard.
Chrome is a large slab on the left to contain a list of tabs to choose from
and buttons to add/remove them, and an outer frame to contain the whole
( three long-skinny cubes, top, bottom, and far side of board.)

Movable and resizable by manipulating frame pieces
(bottom to move, top to increase height, right edge to stretch.)
No direct rotation control as yet; that would be a stubby portion at the bottom of the right-hand edge slat.
!

!QTabbedWall class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTabbedWall class methodsFor: 'inspector' stamp: 'sn 3/8/2013 17:57'!
addPropertyRows: dialog forFrame: twref

	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Title'translateMe;
			fetchBlock: [ :tw | tw get:#buttonLabel ];
			acceptBlock: [ :pt :tw | tw future editedButtonLabel: pt ];
			balloonHelp: 'Title of the tabbed wall.'translateMe;
		yourself
	). 
	dialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :tw | tw get:#extent ];
			acceptBlock: [ :pt :tw | tw future extent: pt ];
			balloonHelp: 'Dimensions of the noteboard surface.'translateMe;
		yourself
	). 

	super addPropertyRows: dialog forFrame: twref.

	"We fix some properties."
	dialog removeRow: 'Destination Type'.
	dialog removeRow: 'Go-To Label'.
	dialog removeRow: 'Go-To Button Position'.
	dialog removeRow: 'Seating Range'.
	dialog removeRow: 'Scale Factor'.! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 11:07'!
defaultCapExtent
	| slab |
	slab := self defaultSlabExtent.
	^ ( slab x @ (slab z*0.8) @ (slab z * 1.5)) * 1.08.! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 17:35'!
defaultCategory
	"Maybe Destinations, but these seems to make more sense since I'm a content aggregation."
	^ 'Panels'! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 10:39'!
defaultDestinationMode
	^ #front! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 17:34'!
defaultFootExtent
	^ self defaultCapExtent! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 17:34'!
defaultIsSticky
	"Yes, this is a focal destination"
	^ true! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 23:03'!
defaultSlabExtent
	| height |
	height := QMultiBoard defaultExtent y.
	^ (height * 0.28) @ height @ ( height * 0.02) ! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 3/30/2010 18:00'!
minBoardHeight
	"We impose a min so that we can use the height as a divisor when tracking resizes.
	 Also prevents nonsense proportions after laying out the framework."
	^ 2.0! !

!QTabbedWall class methodsFor: 'defaults' stamp: 'bgf 3/30/2010 18:00'!
minBoardWidth
	"We impose a min so that we can use the width as a divisor when tracking resizes.
	 Also prevents nonsense proportions after laying out the framework."
	^ 1.0! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 8/18/2009 00:00'!
addNote: aNote
	super addNote: aNote.
	board addNote: aNote.! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 3/30/2010 18:01'!
applyResizerState: stateVec
	"Protocol for resizing via QAbstractResizer, with undo.
	 The resizer state vector is a local transform, a surface to be attached to, and a 2D extent.
	 Tabbed-wall recomputes from the delta, so that the attached items are moved proportionally."

	self dragHeightDelta: (stateVec third y - (self extent y)).
	self dragLengthDelta: (stateVec third x - (self extent x)).

	self localTransform: (stateVec first).
	self root ifNotNil: [ : space | space attach: self to: (stateVec second) ].! !

!QTabbedWall methodsFor: 'persist' stamp: 'bgf 8/4/2009 13:47'!
childrenForBrowser
	^ self childrenForPersistence! !

!QTabbedWall methodsFor: 'persist' stamp: 'bgf 8/5/2009 17:14'!
childrenForPersistence
	"We do persist the board...
	  since it carries all the attachments
	  (and must be shouldPersist to even receive attachments."

	^ super childrenForPersistence 
			copyWithoutAll: { slab . cap . foot .pole . tabs . controls . shadow  . banner}.! !

!QTabbedWall methodsFor: 'persist' stamp: 'bgf 8/29/2009 16:54'!
constructFrom: pStream
	| bd |
	super constructFrom:pStream.
	self class == QTabbedWall ifTrue:[
		(pStream readSequence:'children' default: #()) do: [ :child | self addChild: child ].
	].
	bd := pStream readObject: 'board'  default: board.
	bd == board ifFalse: [
		self removeChild: board.
		board obliterate.
		self addChild: (board := bd).
		board isPlacementLocked: true; doContextMenu: false.
		tabs board: board.
		controls board: board.
	].
	self setupBoardEvents! !

!QTabbedWall methodsFor: 'initialize' stamp: 'bgf 8/4/2009 10:27'!
defaultIsAtomic
	^ true! !

!QTabbedWall methodsFor: 'menu' stamp: 'bgf 11/22/2009 12:28'!
deletionLabel
	"String for the delete-confirm menu."
	^ (board allNotes size > 0) 
		ifTrue: [ self menuTitle, ' and Contents' ]
		ifFalse: [ self menuTitle ].! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/4/2009 10:42'!
destination
	^ self! !

!QTabbedWall methodsFor: 'transform' stamp: 'bgf 8/18/2009 11:12'!
destinationExtent
	"In case we're really wide, concentrate on the starting area."
	^  (( (board width + slab extentX ) atLeast: (slab extentY) atMost: (16) )
		@ slab extentY) * 1.25! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:49'!
dragAdapter
	^ resizers third! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 3/30/2010 17:57'!
dragHeightDelta: inDeltaY
	|  bo bh noteYs bp bio ndt deltaY |

	"When drag-resizing, we don't want the attachments to go astray,
	 so we make an effort to reposition them relative to the edges and center of the board."
	inDeltaY = 0 ifTrue: [ ^ self ].
	bh := board height.
	deltaY := inDeltaY max: ((self class minBoardHeight * (self scaleInMatrix)) - bh).
	deltaY = 0 ifTrue: [ ^ self ].

	bp := board globalPosition.
	bo := board globalTransform orientation.
	bio := board globalTransform inverseByTranspose orientation.
	noteYs := board allNotes collect: [ : note | (bio transformVector:  (note globalPosition - bp))  y ].

	self extent: (self extent + (0 @ deltaY)).
	self translation: (self translation  +  (self globalOrientation transformVector: (0@deltaY*0.5@0))).

	board allNotes with: noteYs do: [ : note  : oldY |
		|  newy sa |
		sa := note surfaceAttachment.
		bh > Float epsilon ifTrue: [ 
			newy := (oldY * (bh + deltaY) / bh).	
		] ifFalse: [
			newy := oldY
		].
		ndt := (bo transformVector: (0@(newy - oldY)@0)).
		note translation: (note translation + ndt).
		self root attach: note to: sa.
	].

	self allUpdateShape.
! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 3/30/2010 17:55'!
dragLengthDelta: inDeltaX
	| bw bp  bo bio noteXs deltaX |
	"When drag-resizing, we don't want the attachments to go astray,
	 so we make an effort to reposition them relative to the edges and center of the board."
	inDeltaX = 0 ifTrue: [ ^ self ].
	bw := board width.
	deltaX := inDeltaX max: ((self class minBoardWidth * (self scaleInMatrix)) - bw).
	deltaX = 0 ifTrue: [ ^ self ].

	bp := board globalPosition.
	bo := board globalTransform orientation.
	bio := board globalTransform inverseByTranspose orientation.

	noteXs := board allNotes collect: [ : note | (bio transformVector:  (note globalPosition - bp))  x ].

	self extent: ( (self extent + (deltaX @ 0)) max: (0.1@0)).
	self translation: (self translation  +  (self globalOrientation transformVector: (deltaX*0.5@0@0))).

	board allNotes with: noteXs do: [ : note  : oldX |
		|  newx sa |
		sa := note surfaceAttachment.
		bw > Float epsilon ifTrue: [ 
			newx := (oldX * (bw + deltaX) / bw).	
		] ifFalse: [
			newx := oldX
		].
		note translation: (note translation + (bo transformVector: ((newx - oldX)@0@0))).
		self root attach: note to: sa.
	].

	self allUpdateShape.
! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/5/2009 17:55'!
editToolClass
	"These text edit jobs (renaming tabs) always want one-liners."
	^ QTextLinePane! !

!QTabbedWall methodsFor: 'menu' stamp: 'bgf 8/5/2009 17:54'!
editedButtonLabel: string
	self buttonLabel: string;
		updateTitle.! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/4/2009 13:09'!
extent
	"The main body extent of my content board (xy)"
	^ board extent! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/4/2009 13:14'!
extent: xy
	"The main body extent of my content board (xy)"
	board extent: xy.
	self shapeChanged.! !

!QTabbedWall methodsFor: 'menu' stamp: 'bgf 3/8/2010 14:13'!
fillContextMenu: tmenu forEvent: anEvent
	| cursor |
	"We reinforce the colum-controls with this set of tab control commands."
	board ifNil: [ ^ self ].
	board checkStaleNotes.
	cursor := board cursor.

	cursor < board numTabs ifTrue: [ 
		tmenu add: 'Next Tab' target: board action: #cursor: arguments:  { (cursor+1) }
	].
	cursor > 1 ifTrue: [
		tmenu add: 'Previous Tab' target: board action: #cursor: arguments:  { (cursor-1) }
	].
	tmenu addSeparator.
	board numTabs < board class maxTabs ifTrue: [
		tmenu add: 'Add Tab' target: board action: #addTab:forEvent:
										arguments: { cursor . anEvent }
	].
	board numTabs > 1 ifTrue: [
		tmenu add: 'Remove Tab' target: board action: #checkRemoveTab: arguments: { anEvent }
	].


! !

!QTabbedWall methodsFor: 'initialize' stamp: 'bgf 8/31/2009 12:04'!
initialize
	super initialize.
	self objectName: 'Tabbed Work Area'.
	board := QMultiBoard new.
	self addChild: board.
	slab := TCube new.
	self addChild: slab.
	foot := TCube new.
	self addChild: foot.
	cap := TCube new.
	self addChild: cap.
	pole := TCube new.
	self addChild: pole.
	banner := TCube new.
	self addChild: banner.
	title := QLabel new.
	banner addChild: title.

	tabs := QMultiBoardTabButtons new.
	tabs board: board.
	slab addChild: tabs.
	controls := QMultiBoardEditButtons new.
	controls board: board.
	slab addChild: controls.

	shadow := QMaterials newRectangularShadow.
	shadow addRotationAroundX:-90.
	self addChild: shadow.

	{ foot . cap . slab . pole. tabs . controls . shadow . banner } do: [ : kid |
		kid shouldPersist: false; isPlacementLocked: true.
	].
	board isPlacementLocked: true; doContextMenu: false.

	slab material: (QMaterials fixedDisplayMaterial).
	banner material: (QMaterials fixedDisplayMaterial).

	foot material: (QMaterials lightWoodMaterial).
	cap material: (QMaterials lightWoodMaterial).
	pole material: (QMaterials lightWoodMaterial).
	self extent: (QMultiBoard defaultExtent).

	resizers := OrderedCollection new.
	resizers add: (QLengthResizer forLuggage: self grip: pole).
	resizers add: (QHeightResizer forLuggage: self grip: cap).
	resizers add: (QMotionAdapter forLuggage: self grip: foot).  	
	resizers last attractionRange: 2; 		"Walls shouldn't be eager to leap."
		openSpacePolicy: #viewer;
		surfacePolicy: #uprightOnly.

	self setupBoardEvents.
	self setPlacementLocked: self isPlacementLocked.! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/4/2009 10:46'!
isComponent
	^ self shouldPersist! !

!QTabbedWall methodsFor: 'persist' stamp: 'bgf 8/4/2009 10:27'!
isEditableGroup
	"C.F. TGroup; plain destination instances are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QTabbedWall) and: [ self isAtomic not ]! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 1/20/2010 18:50'!
isFromTemplate: bool
	"The multiboard is persisted and a component, 
	 but needs to inherit my from-Template setting regardless."
	super isFromTemplate: bool.
	board ifNotNil: [ board isFromTemplate: bool ].! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 8/4/2009 10:18'!
isGroundBound
	^ true! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/4/2009 13:38'!
isPlacementLocked: aBool

	"When placement is locked, direct-manipulation for move/resize should be disabled.
	 Subclasses should extend setPlacementLocked, rather than this accessor!! "

	(aBool ~= self isPlacementLocked) ifTrue: [
		self setPlacementLocked: aBool
	].
! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/5/2009 12:48'!
labelForBrowser
	^ 'Tabbed Wall: ', (self buttonLabel)! !

!QTabbedWall methodsFor: 'transform' stamp: 'bgf 8/18/2009 11:17'!
localCenter
	"See destinationExtent - for gotos, we keep the control slab in view."
	| x |

	x := ((board width * 0.5) + (slab extentX)) - (self destinationExtent x * 0.5).
	^ self translation -
		(self orientation 
			transformVector: (x @ 0 @ 0))! !

!QTabbedWall methodsFor: 'accessing' stamp: 'bgf 8/4/2009 13:17'!
minExtent
	^ 0.1@0.1! !

!QTabbedWall methodsFor: 'initialize' stamp: 'bgf 8/5/2009 18:18'!
onNotesChanged
 	self updateTitle.
	self unionSphereChanged.! !

!QTabbedWall methodsFor: 'persist' stamp: 'bgf 8/4/2009 14:41'!
persistOn: pStream
	super persistOn: pStream.
	self class == QTabbedWall ifTrue:[
		pStream storeSequence: 'children' value: (self childrenForPersistence) default: #().
	].
	pStream storeObject: 'board' value: board.! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 8/5/2009 10:41'!
photographerClass
	^ QDirectPhotoJob! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 8/5/2009 17:59'!
pointerDown: event
	event selection ifNil: [ ^ self ].
	event selection frame ifNotNilDo: [  : sel |
		(title = sel and: [event nrClicks = 2]) ifTrue: [
				QTextEditJob 
					edit: (self buttonLabel) 
					title: 'Edit Title' 
					forUser: event avatar 
					recipient: self 
					selector: #editedButtonLabel:
			]].
! !

!QTabbedWall methodsFor: 'menu' stamp: 'bgf 1/29/2010 11:42'!
reallyDelete: event
	"Interactive delete - includes cleaning up all that content "
	board allNotes do: [  : item | 
		item ifNotNil: [  item reallyDelete: event ].
	].
	super reallyDelete: event! !

!QTabbedWall methodsFor: 'protocol' stamp: 'bgf 8/17/2009 23:58'!
removeNote: aNote
	super removeNote: aNote.
	board removeNote: aNote.! !

!QTabbedWall methodsFor: 'persist' stamp: 'bgf 8/4/2009 13:39'!
setPlacementLocked: bLocked
	super isPlacementLocked: bLocked.
	resizers isEmptyOrNil ifFalse: [ 
		resizers do: [ : r |
			bLocked ifTrue: [ r disable ] ifFalse: [ r enable ]
		]].! !

!QTabbedWall methodsFor: 'initialize' stamp: 'bgf 1/19/2010 17:08'!
setupBoardEvents
	self stopScript: #onNotesChanged.
	self startScript: #onNotesChanged  when: { board . #notesChanged }.
	board ifNotNil: [ board isFromTemplate: (self isFromTemplate) ].! !

!QTabbedWall methodsFor: 'hierarchy' stamp: 'bgf 10/25/2009 14:31'!
updateShape
	| ratio slabDims |
	super updateShape.
	board translation: (0 @ 0 @ 0).
	ratio := board extent y / self class defaultSlabExtent y.

	slabDims := (self class defaultSlabExtent * ratio) + ((board width * 0.025)@0@0).
	slab extent: slabDims.
	slab translation: ((board width + slabDims x) *-0.5  @ 0 @ 0).

	tabs translation: (0 @ (slabDims y * 0.2) @ (slabDims z * 0.51));
		extent: (slabDims x  @ (slabDims y  * 0.6)).
	controls	
		translation: (0@ (slabDims y * -0.45) @ (slabDims z * 0.51));	
		extent: (slabDims x @ (slabDims y * 0.1)).

	cap extent: (self class defaultCapExtent * ratio + (board width * 1.025 @0@0)).
	foot extent: (self class defaultFootExtent * ratio) + (board width * 1.025 @0@0).
	pole extent: ((cap extentY)@ (slabDims y ) @ (cap extentZ * 0.75)).
	banner extent: (slabDims x @ (3.6 * slabDims z ) @ (slabDims z)) * 1.03.

	cap translation:  (slabDims x * -0.495) @ (slabDims y - (cap extentY * 0.5) * 0.5) @ 0.
	foot translation: ((slabDims x * -0.495) @ (slabDims y - (cap extentY * 0.5) * -0.5) @ 0).
	pole translation: ((board width * 0.5) + (pole extentX * 0.1) @ (0) @ 0).
	banner translation: (slab translation + (0@(slabDims y * 0.5 + (banner extentY * 0.6))@0)).

	self buttonPosition: ((board width + slab extent x * -0.5)
						 @ ((board height * 0.5) + (slabDims x * 0.7))
						 @ 0).
	shadow extent: (foot extentX @ (foot extentZ)).
	shadow translation: (slabDims x * -0.495) @ (board height * -0.5025) @ 0.
	shadow borderWidth: (slabDims z * 1.5).

	self updateTitle.
	self unionSphereChanged.! !

!QTabbedWall methodsFor: 'hierarchy' stamp: 'bgf 8/10/2009 10:34'!
updateTitle
	"The title is the label attached to the banner above the slab."
 	title contents: (' ', (self buttonLabel), ' ').
	title translation: (0@(banner extentY * 0.05)@ (banner extentZ * 0.51)).
	title extent: (banner extentX @ banner extentY) * 0.99.
	title textColor: (QMultiBoardTabButtons defaultTextColor).
	title textAnchorPoint: #centerLeft.
	title backgroundColor: (Color transparent).
! !

QDestination subclass: #QTable
	instanceVariableNames: 'table'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QTable commentStamp: '<historical>' prior: 0!
An abstract simple table; a raised horizontal surface, on legs.

Can be rotated (by dragging the top) and moved (by dragging a leg).

Can be the base of a larger structure; assign '#lugguge:' to the
assembly container, and the top and legs become rotation and
motion grips for the assembly.

The coordinate system is ground-based;
the top of the table is at 'self extent y'.
!

QTable subclass: #QPedestal
	instanceVariableNames: 'portal scrollbox isBlocked doStepping matte isMatteVisible menuButtonHandler menuButton knob initialOffset initialScale downState'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Portals'!

!QPedestal commentStamp: '<historical>' prior: 0!
A scrollable 3D portal on a table; the table can be moved and rotated
to position the portal in space.   The portal can be opened and closed 
using a button that appears on the table.!

!QPedestal class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPedestal class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QPedestal class methodsFor: 'defaults' stamp: 'bgf 4/22/2009 13:52'!
defaultSeatingRange
	"Unlike tables, pedestals don't do seating."
	^ 0! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 06:58'!
addMenuActions: menu forEvent: event
	"Does not have full canModifyForum protections ... tbd."
	| atSavedViewpoint box |
	atSavedViewpoint :=  (portal offset = initialOffset and: [ portal scale = initialScale ] ).
	box := scrollbox ifNil: [ portal ].

	(menu addItemSpec: {'Reset Rotation' . #resetRotation:} target: self)
		disabled: (box rotation = Quaternion identity).
	(menu addItemSpec: {'Reset Viewpoint' . #resetPortal:} target: self)
		disabled: atSavedViewpoint.

	event avatar canUseInventory ifTrue: [	
		menu addSeparator.
		(menu addItemSpec: {'Copy Link to Inventory' . #copyDestToInventory: }
			target: self) disabled: (self postcardLink isNil).
	].
	menu addSeparator.

	(menu addItemSpec: {'Show Matte' . #toggleMatte:} target: self)
		checked: self isMatteVisible.
	(menu addItemSpec: {'Show Scroll Box' . #toggleScrollbox:} target: self)
		checked: self isScrollboxed.

	menu addSeparator.

	(menu addItemSpec: {'Allow Passage' . #toggleBlocked:} target: self)
		disabled: (self postcardLink isNil);
		checked: (self isBlocked not).
	(menu addItemSpec: {'Disconnect Link' . #confirmClear: } target: self)
		disabled: (self postcardLink isNil).
	(menu addItemSpec: {'Edit Label' . #editLabel:} target: self).
	(menu addItemSpec: {'Save Viewpoint for Reset' . #updateSavedViewpoint} target: self)
	 	disabled: (box visible not | atSavedViewpoint).

	menu addSeparator.
	menu addItemSpec: {'Delete Pedestal' . #confirmDelete:} target: self.
! !

!QPedestal methodsFor: 'private' stamp: 'bgf 3/14/2010 17:50'!
applyStateVector: stateVec
	"See #stateVector - this reapplies a captured state."	self postcardLink: (stateVec first).
	self label: (stateVec second).
	portal transform: stateVec third.
	self portalScale: stateVec fourth.
	self portalOffset: stateVec fifth.
	self isPortalVisible: stateVec sixth.
	self isScrollboxed: stateVec seventh .
	self isMatteVisible: stateVec eighth .
	self isBlocked: (stateVec ninth).
! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/11/2010 17:36'!
confirmClear: event
	"Ask the user to confirm clearing the portal link. "
	event avatar ifNil: [ ^ self ].
	self root ifNil: [ ^ self ].
	event avatar confirm: 'Disconnect Link?'  
				selector:  #reallyDisconnect:
				target: self.
! !

!QPedestal methodsFor: 'persist' stamp: 'bgf 4/13/2008 11:34'!
constructFrom: pStream

	super constructFrom: pStream.
	"Upgrade 1.0.7-1.0.8 pedestals - if both 1.0.7 and .10.8 were used to load/save
	 the same forum, 
	 so the pedestal would end up with doubled-up portals;
	 the newer form recovers the postcard link and uses a hardcoded portal."
	table contents copy do: [ :ch |
		(ch isKindOf: TScrollBox3D) ifTrue: [
			"I know my initialize made the real one, and it's not 'contents',
	 		 so this is a leftover from the older structure 
			 where the table was a child of the pedestal.)"
			self removeContents:ch.
			ch destroy.
		]
	].
	self extent:(pStream readVector3: 'extent' default: self extent).
	self isBlocked: (pStream readBool: 'isBlocked' default: self isBlocked).
	self isPortalVisible: (pStream readBool: 'isPortalVisible' default: self isPortalVisible).
	self isMatteVisible: (pStream readBool: 'isMatteVisible' default: self isMatteVisible).

	"InitialScale and InitialOffset were not persisted pre 1.1 versions;
	 however, we want to use the 'best' normal value, so if no persistent value is set,
	 use whatever actual offset/scale was set at save time.  
	 Newer pedestals will have non-default values persisted as needed.
	 This is needed so that the 'Reset Viewpoint' command on a legacy demo pedestal
	 doesn't take you back to the 0.0.0, 0.1 viewpoint (which often leaves the box empty.)"

	self portalOffset:(pStream readVector3: 'portalOffset' default: self portalOffset).
	initialOffset := (pStream readVector3: 'initialOffset' default: self portalOffset).
	self portalScale: (pStream readNumber: 'portalScale' default: self portalScale).
	initialScale := (pStream readNumber: 'initialScale' default: self portalScale).
	self postcardLink: (pStream readObject: 'postcardLink' default: nil).
	scrollbox rotation: (pStream readQuaternion: 'portalRotation' default: Quaternion identity).
	self isScrollboxed: (pStream readBool: 'isScrollboxed' default: self isScrollboxed).

	"Reassert control over destination-labeling."
	self label: (self label).! !

!QPedestal methodsFor: 'events' stamp: 'bgf 3/12/2010 07:15'!
consumeDrop: payload avatar: avatar selection: aTXSel
	"Can I display this content?  It must be linkable to do so."
	| old |

	(self isDropTarget: payload) ifFalse: [ ^ false ].

	old := self stateVector.

		"Reset portal to origin"
		portal transform: (Matrix4x4 new).
		self portalScale: 0.1. "TP3D default"

		"Hook it in"
		self postcardLink: (payload contentLink).
		self label: (payload shortSummary).
		self isPortalVisible: true.

		"Request resolution in my harness."
		avatar signal: #activateLinkForFrameAndGo 
					with: payload contentLink
					with: self 
					with: false.

	self maybeUndo: old description: 'Link' avatar: avatar.

	^ false	"the payload is unaffected"
! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 1/2/2008 23:51'!
copyDestToInventory: event

	"Copy the destination URL this portal looks at, to the user's inventory."
	(QFileCard forServiceURL: (self postcardLink urlString)) copyToInventory: event.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/9/2009 22:29'!
destinationStanceFor: anAvatar
	"The transform to apply to the avatar to bring it to me.
	 Override inherited distance - we need to be further from pedestals
	 which are lower; the user needs to see the button."

	| distance opp |

	"Want to be far enough away to see the button & label"
	opp := ((self extent y + self extent x) / 2.0) + 1.		

	"Table height + cube dimension of portal"
	distance := opp * 2.55.		"folding in the standard trig:  1/ (45 degreesToRadians / 2.0)  "
	distance := distance + (self extent z / 2.0).

	^ QGotoButton groundAdjust: (
		QGotoButton
			stanceFor: anAvatar
			toView: (self destination)
			inDestination: self
			inAngle: 180
			atDistance: distance)
		forAvatar: anAvatar! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 8/20/2009 22:08'!
editLabel: event
	"Change the pedestal's front Label"
	table labelTag editByAvatar: event avatar.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 2/11/2007 10:13'!
extent: aVector3

	"Specify the extent of the table upon which the portal appears.
	 The portal is always a cube of dimension (aVector3 x - 1),
	 floating slightly above the table top."

	super extent: aVector3.
! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 01:39'!
initialOffset
	^ initialOffset! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 02:56'!
initialOffset: aVector3
	"Set the offset of the 3D scrollbox, and save it for use in Reset.
	 This is useful in forum template world-building code,
	 since most preset 3D portals need offsets other than 0.0.0."
	initialOffset := aVector3 copy.
	self portalOffset: aVector3.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 01:39'!
initialScale
	^ initialScale! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 02:55'!
initialScale: aNumber
	"Set the scale of the 3D portal, and save it for use in Reset.
	 This is useful in forum template world-building code,
	 since the original default scale of the portal is rarely appropriate."
	self portalScale: aNumber.
	initialScale := aNumber.! !

!QPedestal methodsFor: 'initialize' stamp: 'bgf 4/8/2010 08:01'!
initialize

	super initialize.
	doStepping := false.

	"Extent refers to the table extent."
	self extent: 3@1@3.

	menuButton := (TRectangle new).
	menuButton material: (QMaterials menuButtonRectangleMaterial).
	self addChild: menuButton.
	menuButtonHandler := QButtonHandler
			body: menuButton
			target: self
			action: #showMenu:
			label: 'Pedestal Menu'.
	
	knob := (QLozenge new).
	knob material: QMaterials controlOnMaterial.
	knob addRotationAroundX: 90.
	knob doesClickSelect: false.
	self addChild: knob.

	portal := QPortal3D new.
	portal blocked: false.
	portal billboardTexture: (QTextures find: 'portal-void.gif').
	table addChild: portal.	"Temporary (scrollbox may replace it)"

	"The matte coincides with the scroll box
	  and serves to visually isolate the 3D content from the background.
	  It's a cube whose faces are only visible from the interior side,
	  so it appears open to the viewer from all viewpoints."
	matte := TCube new.
	matte material: ((QMaterials matteMaterial)
						flipFace:true;			"Only visible from 'within'"
						cullFace:false;			"Show the back faces"
						yourself).
	matte solid: false.
	matte objectName: 'matte' copy.
	portal addChild: matte.
	matte checkAlpha.

	portal doesClickSelect: false.
	matte doesClickSelect: false.

	self objectName: 'Pedestal'.
	self label: '3D Projection'.
	self suggestedPOV: 'First Person'.
	self category: 'Pedestals'.

	"Defaults"
	self initialScale: 0.1.
	self initialOffset: 0@0@0.
	self isBlocked: true.
	self isScrollboxed: true.
	self isPortalVisible: true.
	self isMatteVisible: true.
	self startScripts.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/20/2006 13:37'!
isBlocked
	^ isBlocked
! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 12/23/2006 20:42'!
isBlocked: bFlag
	isBlocked := bFlag.
	portal blocked: bFlag.
! !

!QPedestal methodsFor: 'events' stamp: 'bgf 2/20/2007 19:26'!
isDropTarget: payload 

	"Can I display this content?  It must be a postcard"

	^(payload respondsTo:#kind) and: [payload kind = #postcardLink]! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 3/15/2008 17:57'!
isListedLandmark

	"Should I appear in a landmarks list?"

	^ self postcardLink notNil! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/24/2006 21:42'!
isMatteVisible
	"If true, the portal is set inside an inside-out white background cube
	 so that it is visually isolated from the background."
	^ isMatteVisible! !

!QPedestal methodsFor: 'accessing' stamp: 'das 12/13/2007 13:52'!
isMatteVisible: aBool
	"Surround the portal with a white background matte?"
	isMatteVisible := aBool.
	matte material opacity: (aBool ifTrue: [0.99] ifFalse: [ 0.0 ]).
	matte checkAlpha.
! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/28/2007 19:24'!
isPortalVisible
	^ scrollbox ifNotNil: [ scrollbox visible ] ifNil: [ portal visible]! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/20/2006 15:10'!
isPortalVisible: aBool
	(self isPortalVisible ~= aBool) ifTrue: [
		self togglePortal.
	].! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/28/2007 19:20'!
isScrollboxed
	^ scrollbox notNil! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 4/8/2010 08:01'!
isScrollboxed: aBool
	"Is the 3D portal in a scroll box?"

	(self isScrollboxed = aBool) ifTrue: [ ^ self ].

	aBool ifFalse: [
		scrollbox ifNotNil: [		"belt&braces safety"
			scrollbox outOfBox.
			(scrollbox future: 5000) obliterate.
			scrollbox := nil.
		]
	] ifTrue: [
		|vis|
		vis := self isPortalVisible.
		scrollbox := TScrollBox3D new.
		scrollbox sideScale: 18.0.
		scrollbox initializeWithContents: portal.
		scrollbox solid: false.
		table addChild: scrollbox.
		scrollbox doesClickSelect: false.
		self isPortalVisible: vis.
	].

	self startScripts.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 4/13/2008 11:41'!
label: aString
	super label: aString.
	self buttonLabel: ('Pedestal: ',(self label ifNil: ['3D Projection'])).! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 9/4/2007 12:49'!
labelForBrowser
	self postcardLink ifNotNil: [
		^ 'Pedestal - ', (self postcardLink shortDisplayString)
	] ifNil: [
		^ 'Pedestal'
	].! !

!QPedestal methodsFor: 'private' stamp: 'bgf 3/13/2010 08:53'!
maybeUndo: oldState description: desc avatar: avatar
	"If the state has changed from oldState, register an undoable state change."
	| newState |

	avatar ifNil: [ ^ self ].
	newState := self stateVector.
	(newState hasEqualElements: oldState) ifFalse: [
		avatar recordAction: #frame:oldState:newState:description:
				class: QStateVectorAction
				arguments: {  self . oldState. newState .  desc }.
	].! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 9/14/2007 21:24'!
menuTitle
	| lbl |

	lbl := 'Pedestal'.
	self postcardLink 
		ifNil: [ lbl := lbl, ' (Not Linked)']
		ifNotNil: [ self isBlocked ifFalse: [ lbl := lbl, ' (3D Portal)'] ].
	^ lbl! !

!QPedestal methodsFor: 'persist' stamp: 'bgf 4/13/2008 11:14'!
persistOn: pStream

	super persistOn: pStream.
	
	pStream storeVector3: 'extent' value: self extent.
	pStream storeBool: 'isBlocked' value: self isBlocked default: true.
	pStream storeBool: 'isPortalVisible' value: self isPortalVisible default: true.
	pStream storeBool: 'isMatteVisible' value: self isMatteVisible default: true.
	pStream storeBool: 'isScrollboxed' value: self isScrollboxed default: true.

	pStream storeVector3: 'portalOffset' value: self portalOffset default: 0@0@0.
	pStream storeNumber: 'portalScale' value: self portalScale default: 0.1.
	pStream storeObject: 'postcardLink' value: self postcardLink default: nil.
	pStream storeVector3: 'initialOffset' value: self initialOffset default: 0@0@0.
	pStream storeNumber: 'initialScale' value: self initialScale default: 0.1.
	pStream storeQuaternion: 'portalRotation' value: (scrollbox ifNil: [portal]) orientation asQuaternion default: (Quaternion identity).! !

!QPedestal methodsFor: 'events' stamp: 'bgf 3/12/2010 07:15'!
pointerDown: event 

	downState := nil.
	(scrollbox notNil and: [ scrollbox isOrContains: (event selection frame) ])
	ifTrue: [
		downState := self stateVector.
	].! !

!QPedestal methodsFor: 'events' stamp: 'bgf 3/12/2010 07:03'!
pointerEnter: event
	self pointerOver: event.! !

!QPedestal methodsFor: 'events' stamp: 'bgf 3/12/2010 07:03'!
pointerLeave:event
	downState := nil.
	knob unhilite.
	super pointerLeave: event.! !

!QPedestal methodsFor: 'events' stamp: 'ar 4/16/2011 17:14'!
pointerOver: event

	downState := nil.
	(self button isOrContains: event selection frame) ifTrue: [ ^ self ].

	super pointerOver: event.

	self postcardLink ifNil: [
		^ event avatar 
			signal: #statusMessage with: 'To connect the pedestal, drag a Forum link onto it.' 
	].
	event selection frame = knob ifTrue: [
		^ self pointerOverKnob: event
	].

	self isPortalVisible ifTrue: [
		"Note that this message goes only to the local user !!"
		portal resolved ifFalse: [
			event avatar signal: #statusMessage 
				with: 'Click in the projection to log in to: ', self postcardLink shortDisplayString
		] ifTrue: [
			event avatar signal: #statusMessage 
				with: 'Projection of ', self postcardLink shortDisplayString.
		].
	] ifFalse: [
		event avatar signal: #statusMessage 
				with: 'Projection of: ', self postcardLink shortDisplayString, ' (Closed)'.
	].

! !

!QPedestal methodsFor: 'events' stamp: 'bgf 11/13/2008 12:53'!
pointerOverKnob: event

	event avatar ifNil: [ ^ self ].
	self postcardLink ifNotNil: [
		knob hilite.
		self isPortalVisible ifTrue: [
			event avatar showFloatMessage: 'Click knob to hide projection' forEvent: event.
		] ifFalse: [
			event avatar showFloatMessage: 'Click knob to show projection' forEvent: event.
		].
	].
! !

!QPedestal methodsFor: 'events' stamp: 'bgf 3/12/2010 07:02'!
pointerUp: event

	(scrollbox notNil and: [ scrollbox isOrContains: (event selection frame) ]) ifTrue: [
		downState ifNotNil: [ 
			self maybeUndo: downState description: 'Adjust Portal' avatar: event avatar 
			].
	].
	downState := nil.

	"Clicking on an open, unresolved portal triggers login."
	(((scrollbox ifNil: [ portal ]) isOrContains: (event selection frame)) and: [self isPortalVisible]) ifTrue:[
		 event avatar signal: #activateLinkForFrameAndGo 
					with: self postcardLink
					with: self 
					with: false.
	].



		
	(event frame = knob) ifTrue: [
		self postcardLink ifNotNil: [
			^ self togglePortal: event.
		]
	].! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/20/2006 13:58'!
portal
	"The 3D portal itself."
	^ portal
! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/20/2006 13:43'!
portalOffset
	^ portal offset! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 01:35'!
portalOffset: aVector3
	"The offset shift from the postcardLink's point of reference to the point
	 placed at the center of the 3D portal."
	initialOffset := aVector3 copy.
	portal offset: aVector3! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/20/2006 13:42'!
portalScale

	"The scaling factor used in the 3d portal."
	^ portal scale! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 01:34'!
portalScale: aNumber

	"The scaling factor used in the 3d portal."
	initialScale := aNumber.
	portal scale: aNumber.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 10/20/2006 13:39'!
postcardLink

	^ portal postcardLink! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 9/3/2007 05:27'!
postcardLink: pc

	portal postcardLink: pc.
	pc ifNil: [
		self isPortalVisible: false.
	].
	self updateKnob.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 4/13/2008 10:03'!
povForDestinationMode: aMode
	"Pedestals are always things to look at, so POV is first-person"
	^ 'First Person'! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 07:16'!
reallyDisconnect: event

	| old |
	old := self stateVector.

	self postcardLink: nil.
	self label: '3D Projection'.

	self maybeUndo: old description: 'Disconnect' avatar: event avatar! !

!QPedestal methodsFor: 'render' stamp: 'bgf 4/8/2010 07:48'!
renderSelection: ogl box: box color: color thickness: lineWid
	"Render selection-boxing indicators that this object is selected.
	 For pedestals, outline the waist only, for simplicity."
		
	ogl glDisable: GLFog.

	"Show the hidden part in a stippled thin outline."

	(OpenGL maxRendererFlags anyMask: 8  "B3D-Antialiasing " ) ifTrue: [
		"If we're not antialiasing, don't do stipples either; they cause trouble with very-lame cards."
		ogl glDisable: GLDepthTest.
		ogl glLineStipple: 1 with: 4r03030303.
		ogl glEnable: GLLineStipple.
		box outlineXZ: ogl color: (color clone alpha: (color alpha * 0.6)) thickness: 1.
		ogl glEnable: GLDepthTest.
		ogl glDisable: GLLineStipple.
	].

	box outlineXZ: ogl color:color thickness: lineWid.
	ogl glEnable: GLFog.

	^ box! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 07:16'!
resetPortal: event

	"Reset the scrollbox 3D viewpoint and scale factor;
	 these are the settings changed by zooming and scrolling.
	 The viewpoint we return to is the one established by self initialScale: and self initialOffset:,
	 which are available through the menu operation 'updateResetValues',
	 and also through software."

	| old |
	old := self stateVector.

	self portalScale: self initialScale.
	self portalOffset: self initialOffset.

	self maybeUndo: old description: 'Reset' avatar: event avatar! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 07:16'!
resetRotation: event
	"Reset the rotation of the portal this is a convenience for tidying-up."

	| old |
	old := self stateVector.
	(scrollbox ifNil: [portal])  rotation: (Quaternion identity).

	self maybeUndo: old description: 'Rotation' avatar: event avatar.
! !

!QPedestal methodsFor: 'menu' stamp: 'bgf 10/25/2009 16:15'!
showContextMenu: event
	(event frame = menuButton) ifTrue: [ 
		"Make right-clicking the menu button itself do the same as left-clcking the MB;
	  	 users can right-click the panel body or title for the 'pure' context menu."
		self showMenu: event
	] ifFalse: [
		super showContextMenu: event
	].! !

!QPedestal methodsFor: 'events' stamp: 'bgf 10/20/2006 13:48'!
start

	self stepTurntable.! !

!QPedestal methodsFor: 'initialize' stamp: 'bgf 3/12/2010 07:00'!
startScripts

	"Start the pointer-up listeners we use to trigger login..."
	self stopScript: #pointerUp:.
	self stopScript: #pointerDown:.
	self startScript: #pointerUp: when: { matte . #pointerUp }.
	scrollbox ifNotNil: [ 
		self startScript: #pointerUp: when: { scrollbox . #pointerUp }.
		self startScript: #pointerDown: when: { scrollbox . #pointerDown }.
	].! !

!QPedestal methodsFor: 'events' stamp: 'bgf 10/20/2006 13:47'!
startTurntable

	doStepping ifFalse: [
		doStepping := true.
		self stepTurntable.
	].! !

!QPedestal methodsFor: 'private' stamp: 'bgf 3/14/2010 17:51'!
stateVector
	"Capture the state of  the portal setup, for undo-redo of link drops and portal adjustments."
	^ { 		
		self postcardLink .	"First"
		self label . 			"Second"
		portal transform . 
		self portalScale .
		self portalOffset .		"Fifth"
		self isPortalVisible .
		self isScrollboxed .
		self isMatteVisible .
		self isBlocked .		"Ninth"
	 }! !

!QPedestal methodsFor: 'events' stamp: 'bgf 10/20/2006 15:59'!
stepTurntable

	doStepping ifTrue: [
		portal transform addRotationAroundY: -0.05.
		(self future:150) stepTurntable.
	].! !

!QPedestal methodsFor: 'events' stamp: 'bgf 10/20/2006 13:49'!
stopTurntable
	
	doStepping := false.
! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 07:16'!
toggleBlocked: event
	"Whether or not walking into the portal transports you to the target space."
	| old |
	old := self stateVector.
	self isBlocked: self isBlocked not.
	self maybeUndo: old description: 'Toggle Block' avatar: event avatar.! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 07:16'!
toggleMatte: event
	"Show or hide the matte - the faint gray translucent surround."
	| old |
	old := self stateVector.
	self isMatteVisible: (self isMatteVisible not).
	self maybeUndo: old description: 'Toggle Matte' avatar: event avatar.! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 8/23/2007 02:27'!
togglePortal

	"Toggle the portal from software.
	 As no avatar is available, this will not trigger harness-side login to the target space."
	self togglePortal:nil! !

!QPedestal methodsFor: 'accessing' stamp: 'bgf 4/13/2008 13:29'!
togglePortal: event

	"Show or hide the scroll box containing the portal;
	 if a live CroquetEvent is supplied, and the toggle opens the portal,
	 and the portal is configured with a link, then login will be prompted."

	| box |
	box := scrollbox ifNil: [portal].	"May have a scrollbox, may not."
	box visible: box visible not.
	portal visible: box visible.	"So we can overlap on-offs of visible and scrollbox."
	box visible ifTrue: [
		table addChild: box.
		"Request resolution in my harness."
		(self postcardLink notNil and: [event notNil and: [event avatar notNil]]) ifTrue: [
			event avatar signal: #activateLinkForFrameAndGo 
					with: self postcardLink
					with: self 
					with: false.
		].

	] ifFalse: [
		table removeChild: box.
	].
	self updateKnob.
	QSoundEffects play: #switch for: self.! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 3/12/2010 07:16'!
toggleScrollbox: event

	"Whether or not the scrollbox appears around the portal3D"
	| old |
	old := self stateVector.
	self isScrollboxed: self isScrollboxed not.
	self maybeUndo: old description: 'Toggle Scroll Box' avatar: event avatar.! !

!QPedestal methodsFor: 'private' stamp: 'bgf 9/3/2007 05:29'!
updateKnob

	knob material: (
		self postcardLink ifNil: [
			QMaterials controlDisabledMaterial.
		] ifNotNil: [
			self isPortalVisible ifTrue: [
				QMaterials controlOnMaterial.
			] ifFalse: [
				QMaterials controlMaterial.
			]]).
		! !

!QPedestal methodsFor: 'private' stamp: 'bgf 3/15/2007 01:41'!
updatePostcard: newPostcard originalUrl: orig
	"Callback after postcard resolution."

	self postcardLink ifNotNilDo: [:pc |
		"for side effects, in case the postcard is shared with any far-away rolodexes."
		pc urlString = orig ifTrue: [ pc urlString: newPostcard urlString].
	].
	self postcardLink: newPostcard.
	self label: newPostcard shortDisplayString.! !

!QPedestal methodsFor: 'menus' stamp: 'bgf 8/23/2007 03:03'!
updateSavedViewpoint
	"Save the current 3D scrolling position as the one we 'reset' to"
	self initialOffset: self portalOffset.
	self initialScale: self portalScale.! !

!QPedestal methodsFor: 'initialize' stamp: 'bgf 9/6/2009 12:37'!
updateShape

	"Geometry changes are written into the Table, and then propagated."
	| inset |

	super updateShape.

	inset := (table barHorizontal*2).

	"The whole portal floats a wee bit above the table top."
	scrollbox ifNotNil: [
		scrollbox extent: (self extent - inset);
			 updateBox;
		   	 translationX: 0 y: (((self extent x - inset)/2) + (inset) + self extent y) z:0.
		matte extent: (scrollbox extent).
	] ifNil: [
		portal extent: (self extent - (1@1@1));
			translationX: 0 y: (((self extent x - inset)/2) + (inset) + self extent y) z:0.
		matte extent: (portal extent).
	].

	menuButton
		translationX: (self extent x/-2+(table barHorizontal) + (inset*1.5))
				   y: (self extent y - (table barVertical/2.4)) 
				   z: (self extent z/2.0 + (table barHorizontal * 0.505));
		extent: ((inset*2)@(table barVertical*0.6))  + (table barVertical *0.15).

	knob
		extent: (inset@inset);
		translation: (self extent x/2- inset)@(self extent y+(inset*0.2))@(self extent z / 2 - inset);
		radius: (inset *0.25).

	self updateBounds.! !

!QTable class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTable class methodsFor: 'inspector' stamp: 'sn 3/9/2013 00:17'!
addPropertyRows: propertyDialog forFrame: stref

	self addNameRowIfNeeded: propertyDialog forFrame: stref.

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :frame | frame get:#extent ];
			acceptBlock: [ :v3 :frame | frame future extent: v3 ];
			balloonHelp: 'Dimensions of the table.'translateMe;
	 	yourself	).

	propertyDialog addRow: (
		(QBooleanPropertyRow new) 
			titleString: 'Mirror Top'translateMe;
			fetchBlock: [ :frame| frame get: #isReflecting ];
			acceptBlock: [ : b :frame  | frame future isReflecting: b ];
			balloonHelp: 'A shiny reflective top.\Adds to hardware load.'translateMe;
		yourself).

	super addPropertyRows: propertyDialog forFrame: stref.

	"Always a gathering destination."
	propertyDialog removeRow: 'Destination Type'.
	"Unlisted"
	propertyDialog removeRow: 'Places Category'.
! !

!QTable class methodsFor: 'defaults' stamp: 'bgf 4/22/2009 14:12'!
defaultSeatingRange
	"Tables by default do seating"
	^ 2.5! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:18'!
addContents: aTFrame

	"Put something on the table.  Doesn't handle positioning, just ensures that
	 the frame moves when the table moves, and is  saved with it persistently."
	table addContents: aTFrame.! !

!QTable methodsFor: 'testing' stamp: 'bgf 4/13/2008 14:30'!
childrenForBrowser
	"The table-mesh itself is not shown as a separate element, 
	 so that it cannot be removed via ungroup (which would break me.)"
	^ super childrenForBrowser copyWithout: table! !

!QTable methodsFor: 'testing' stamp: 'bgf 6/22/2009 13:23'!
childrenForPersistence
	"The table-mesh itself is not shown as a separate element, 
	 so that it cannot be removed via ungroup (which would break me.)"
	^ super childrenForPersistence copyWithout: table! !

!QTable methodsFor: 'persist' stamp: 'bgf 4/13/2008 14:52'!
constructFrom: pStream
	| kids newTable |
	super constructFrom: pStream.

	"Possible in future to use the QTable wrapper around other models.
	 This won't work just yet (and isn't done by current software)
	 as there are assumptions about the API satistfied by <table>.
 	So for now, newTable is always nil."

	newTable := pStream readObject: 'table' default: nil.
	newTable ifNil: [
		self thickness: (pStream readNumber: 'thickness' default: (self thickness)).
		self extent: (pStream readVector3: 'extent' default: (self extent)).
		self isReflecting: (pStream readBool: 'isReflecting' default: (self isReflecting)).
		self label: (pStream readString: 'label' default:nil).

		kids := (pStream readSequence: 'contents' default: #()).
		kids  do: [ :kid | self addContents: kid ].
	] ifNotNil:  [
		table := newTable.
	].

	"If we saved any extra bits from user-drag-drop actions, recover those.
	 This is done after recovering the <table> so that it is first child, hence destination."
	self class == QTable ifTrue:[
		(pStream readSequence:'children' default: #()) do: [ :child | self addChild: child ].
	].
! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:22'!
contents
	"The user-supplied TFrames being carried by this table."
	^ table contents! !

!QTable methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^  (self class ~~ QTable)! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:22'!
extent
	"The cube occupied by the table.
	 Internally, the table's elements are arranged to grow upward from y=0,
	 so the top surface is the plane through extent y."
	^ table extent! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:29'!
extent: anExtent

	"Change the extent of the table;
       the value is a cube corresponding to the volume of the table."
	table extent: anExtent.
	self shapeChanged.! !

!QTable methodsFor: 'initialize' stamp: 'bgf 8/20/2009 22:05'!
initialize

	super initialize.
	self objectName: 'Table'.
	self destinationMode: #circle.
	self suggestedPOV: 'Normal'.
	table := QSimpleTable new.
	self addChild: table.
	self buttonLabel: 'Table'.
	table luggage: self.
! !

!QTable methodsFor: 'testing' stamp: 'bgf 6/1/2009 11:42'!
isEditableGroup
	"C.F. TGroup; plain destination instances are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QTable) and: [ self isAtomic not ]! !

!QTable methodsFor: 'accessing' stamp: 'bgf 1/28/2007 18:46'!
isGroundBound
	^ true! !

!QTable methodsFor: 'testing' stamp: 'bgf 4/13/2008 11:29'!
isListedLandmark
	^ false! !

!QTable methodsFor: 'accessing' stamp: 'bgf 6/2/2009 00:00'!
isPlacementLocked: bFlag
	super isPlacementLocked: bFlag.
	"In case the table is a QSimpleTable... it needs it drag-handles suppressed."
	table ifNotNil: [ table isPlacementLocked: bFlag ].! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:22'!
isReflecting
	^ table isReflecting! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:21'!
isReflecting: bool

	table isReflecting: bool! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 11:53'!
label
	"The string appearing as a label for this table."
	^ table label ! !

!QTable methodsFor: 'accessing' stamp: 'bgf 8/20/2009 22:08'!
label: aString
	"Hang a label under the front face of the table.
	 Note a bit of historical confusion... the label instance variable is a QLabel,
	 but my label/label: accessors refer to its content string."

	table label: (aString).
	self buttonLabel: (aString ifNil: [ 'Table' ]).
! !

!QTable methodsFor: 'accessing' stamp: 'bgf 9/14/2007 21:26'!
menuStyle 
	^ QStyles menuStyle! !

!QTable methodsFor: 'persist' stamp: 'bgf 8/20/2009 22:31'!
persistOn: pStream
	super persistOn: pStream.

	"I'm an editable group so users can attach things to (place on) the table;
	 have to persist those, as they are my children and otherwise lost;
	 note that this excludes the table-mesh and gotoButton."
	(self class == QTable) ifTrue: [
		pStream storeSequence: 'children' value: (self childrenForPersistence) default: #().
	].

	"Historical - table and simpletable used to be the same thing,
	  se we persist it inline."
	(table isKindOf: QSimpleTable) ifTrue: [
		pStream storeNumber: 'thickness' value: (self thickness)
				   default: (QSimpleTable defaultThickness).
		pStream storeVector3: 'extent' value: (self extent).
		pStream storeString: 'label' value: self label default: nil.
		pStream storeSequence: 'contents' value: self contents default:#().
		pStream storeBool: 'isReflecting' value: self isReflecting default: false.	
	] ifFalse: [
		pStream storeObject: 'table' value: table.
	].! !

!QTable methodsFor: 'accessing' stamp: 'bgf 2/11/2010 17:13'!
privateScale: aFactor
	table scaleBy: aFactor.
	self shapeChanged.! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:26'!
removeContents: aTFrame

	"Take something off the table.  Doesn't handle positioning;
	 table-contents presence just ensures that
	 the frame moves when the table moves, and is saved with it persistently."
	table removeContents: aTFrame.! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:24'!
thickness
	^table thickness! !

!QTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:24'!
thickness: aScalar
	table thickness: aScalar! !

!QTable methodsFor: 'hierarchy' stamp: 'bgf 4/13/2008 09:49'!
updateShape

	"When the geometry changes, the button height may also..."
	button translation: (0@( 9 max: (table extent y + 4))@ 0).! !

TGroup subclass: #QAnnotation
	instanceVariableNames: 'adapter lastModifiedTime creatorLogin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QAnnotation commentStamp: '<historical>' prior: 0!
Abstract base class for annotations: 
small widgets (like the QNote post-it style note)
that users attach to existing content in the space.

Annotation rendering can be turned on and off
using a global control in the participant-UI.

This base class provides no visuals or content;
implements the drag-drop and reference-frame attachment logic
for keeping notes attached to frames.

Annotations are draggable items
that attach to the object they are dragged over.
Attachment is done by listening to the reference frame's transform updates,
and repositioning whenever one occurs.

(Adding the note into the parent on the fly as a child
would be simpler, but raises problems w/respect to persistence,
since the note would not be persisted by most compound components.)

Single-clicking opens an editor;
the subclass responsibility is #editForUser:anAvatar.!

!QAnnotation class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAnnotation class methodsFor: 'as yet unclassified' stamp: 'bgf 5/25/2010 19:03'!
addGroupMenuItems: tmenu mask: maskSet frames: frames event: cEvent

	"As the common base of all ballot candidate classes..."
	(maskSet includes: #createPoll) ifTrue: [ 
		| ballots |
		"Ensure sort-order consistent across participants."
		ballots := QPollBarChart sortBallots: frames.
		tmenu add: 'Create a Poll...'
			target: cEvent avatar
			action: #generatePoll:ballots:
			arguments: { cEvent . ballots }.
	].

	super addGroupMenuItems: tmenu mask: maskSet frames: frames event: cEvent.! !

!QAnnotation class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:19'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QAnnotation class methodsFor: 'defaults' stamp: 'das 8/30/2009 12:01'!
defaultSerialNumbered

	"These objects have serial numbers by default."
	^ true! !

!QAnnotation class methodsFor: 'defaults' stamp: 'bgf 8/27/2007 14:23'!
floatingMessage
	^'Click note to review or edit.'! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 2/10/2007 22:23'!
attachmentOffset
	"The offset from an attachment point to the origin of the annotation,
	 along the normal to the attachment surface."
	^ 0! !

!QAnnotation methodsFor: 'menu' stamp: 'bgf 4/1/2010 08:56'!
canBeCopiedBy: anAvatar
	"Notes by a specific user cannot be taken into the inventory of other users."

	^ (super canBeCopiedBy: anAvatar) 
	  and: [ self creatorLogin 
				ifNil: [ true ]	
				ifNotNil:  [ : cl | anAvatar userLogin = cl ] ]! !

!QAnnotation methodsFor: 'persist' stamp: 'bgf 8/1/2009 11:06'!
checkReferenceFrame: rf
	"For upconverting older persistence, predating the attachment manager"
	self root ifNotNilDo: [ : space | space attach: self to: rf ].! !

!QAnnotation methodsFor: 'persist' stamp: 'bgf 8/10/2009 01:03'!
constructFrom:pStream
	super constructFrom:pStream.

	self creatorLogin: (pStream readString: 'creatorLogin' default: (self creatorLogin)).
	self lastModifiedTime: (pStream readNumber: 'lastModifiedTime' default: (self lastModifiedTime)).
	(pStream readObject: 'referenceFrame' default: nil)  ifNotNilDo: [ : rf |
		"Recover a pre-attachment-manager persisted reference frame."
		self future checkReferenceFrame: rf
	].

	self setPlacementLocked: (self isPlacementLocked).! !

!QAnnotation methodsFor: 'menu' stamp: 'bgf 2/9/2010 15:40'!
contextMenuMaskForAvatar: avatar
	"Remove the copyToInventory command if this is an attributed note by someone else."
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	(self creatorLogin notNil and: [avatar userLogin ~= self creatorLogin]) ifTrue: [
		mask remove: #copyToInventory ifAbsent: []. 
	].

	^ mask! !

!QAnnotation methodsFor: 'persist' stamp: 'bgf 8/15/2009 23:46'!
copyToInventory: event
	"Check ownership, so I cannot copy/restore annotations that
 	claim to be the work of someone else."

	event avatar ifNil: [ ^ self ].
	(self creatorLogin notNil and: [
		event avatar userLogin ~= self creatorLogin
	]) ifTrue: [
		^ event avatar signal: #alertMessage with: 'You cannot take another user''s notes.'.
	].

	"This has to be done without the ref frame held, lest we copy the entire space.  
	 We write the reference frame for backward-compatibility in persistOn:"
	event avatar addObjectToInventory: self ! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 7/17/2008 17:37'!
creatorLogin
	"The identification of the person adding the annotation, if supplied; default is nil."
	^ creatorLogin! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 7/17/2008 17:37'!
creatorLogin: aString
	"The identification of the person adding the annotation, if supplied."
	creatorLogin := aString.! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 9/1/2008 17:33'!
creatorName
	| nm |
	"A prettified variant of the creatorLogin, without the email gubbage.
	 In-space, without the folks around, we don't have easy access to screen name sadly,
	 but that should be used to replace this if a nice mechanism for fetching absent-user screen names
	 can be made available in-island (or we could remember at creation time."

	nm := self creatorLogin ifNil: [ ^ nil ].
	^ QStyles simplerLoginName: nm.
! !

!QAnnotation methodsFor: 'initialize' stamp: 'bgf 10/25/2009 15:08'!
destroySoon
	"Remove the thumbtack from the frame tree,
	 and arrange to destroy it after a decent interval;
	 I.e. after the current event timeslice.

	XXXX -
	This is a workaround that illustrates a problem with synchronous #destroy. 
	Thumbtacks are deleted by editing them, erasing their label, and hitting return.
	 (See QwaqParticipantUI #onAcceptInput, the sender of this message.)

	 That <accept> handler used to do a 'future destroy' of the thumbtack.
	 However, <accept> is a keyDown event;
	 the participant UI goes on to handle an immediate keyStroke on the return key.
	 That is treated as normal keyboard event by the participant and view,
	 and so it gets delivered  to the space as a future keyStroke: call.
	 The harness doesn't do a pick between these calls; it retains the keyboard
	 event selection frame that it last identified based on the pointer... the Thumbtack.
	 So the keyStroke event is dispatched with a TXSelection that contains
	 a far ref to the thumbtack, but the the thumbtack has just been destroyed.
	 The decodeScheduleAndExecute decoding fails with exceptions.

	 To avoid this for now 
	  we destroy the thumbtack in a 2-phase call;
	 the harness calls 'future destroySoon, and here (in the island,
	 where can do timed futures) we further defer the destruction;
	 this allows the keyStroke and Up events to pass unnoticed."

	self root ifNotNil: [
		self root detach: self.
	].

	self parent ifNotNil: [
		self parent removeChild: self.
	].

	(self future: 30000) destroy.! !

!QAnnotation methodsFor: 'events' stamp: 'bgf 2/15/2010 17:54'!
doubleClick: event
	"Generic d-c response - annotation classes decide what editByUser means."
	event shiftPressed ifFalse: [
		self editByUser: (event avatar)
	].! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:51'!
dragAdapter
	^ adapter! !

!QAnnotation methodsFor: 'updating' stamp: 'bgf 7/14/2008 09:52'!
editByUser:anAvatar
	"Placeholder for basic pointerClick: response;
       will also be called after an annotate-tool is used to add an annotation."

! !

!QAnnotation methodsFor: 'updating' stamp: 'bgf 8/15/2009 21:56'!
editByUser: anAvatar atCreation: bIsNew
	"Called after an annotate-tool is used to add an annotation."
	self editByUser: anAvatar

! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 7/17/2008 14:19'!
floatingMessageFor: anAvatar
	"The mouse-over message to display to the given avatar's session
	 when that user mouses-over  me"
	^ self class floatingMessage! !

!QAnnotation methodsFor: 'initialize' stamp: 'bgf 2/15/2010 17:55'!
initialize
	super initialize.
	adapter := (QDragAdapter new)
		grip: self;
		luggage: self;
		isMovable: true;
		isDroppable: false;
		isAttractedToSurfaces: true;
		attractionRange: 50.0; 	"Very sticky"
		isGroundBound: true;
		openSpacePolicy: #viewer;
		surfacePolicy: #surface;
		excludeAvatars: true.

	self runScript: #doubleClick: when: { self . #doubleClick }.	"Hook up to graspAdapter dc signal"! !

!QAnnotation methodsFor: 'testing' stamp: 'bgf 8/17/2009 23:57'!
isListedTopic
	"Protocol - identifies annotations that are managed by NoteCollecting surfaces,
	 and are visible in the QTopicsTool.
	 Such annotations are added/removed from the note-collector notes list 
	 when the annotation is attached/detached from such a note-collector surface.

	 Classes for which this is true must implement 
		#asTopicRecord
	 returning a QTopicRecord describing their essentials.    
	 (And should have a native 'editByUser: editor regime.)"

	^ false! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 6/1/2009 12:40'!
isPlacementLocked: aBool

	"When placement is locked, direct-manipulation for move/resize should be disabled.
	 Subclasses should extend setPlacementLocked, rather than this accessor!! "

	(aBool ~= self isPlacementLocked) ifTrue: [
		self setPlacementLocked: aBool
	].
! !

!QAnnotation methodsFor: 'testing' stamp: 'bgf 8/23/2009 17:56'!
isPollBallot
	"Include me when making a poll from a group of attachments?
	 If this answers yes, the object must also offer 
		ballotColor - a VectorColor 
		contents	     - a String "
	^ false! !

!QAnnotation methodsFor: 'testing' stamp: 'bgf 2/28/2008 17:48'!
isReparentable
	"Annotations stick to other objects by cross-references in global space."
	^ false! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 7/17/2008 17:28'!
lastModifiedTime
	"UTC seconds of my creation or most recent edit, or nil if unknown"

	^ lastModifiedTime! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 9/1/2008 19:50'!
lastModifiedTime: utcSeconds
	"UTC seconds of my creation or most recent edit, or nil if unknown"

	lastModifiedTime = utcSeconds ifFalse: [ 	
		lastModifiedTime := utcSeconds.
		self signal: #lastModifiedTimeChanged
	].! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 9/14/2007 21:29'!
menuStyle
	^ (QStyles menuStyle)
! !

!QAnnotation methodsFor: 'persist' stamp: 'bgf 8/15/2009 23:47'!
persistOn:pStream
	super persistOn:pStream.

	pStream storeString: 'creatorLogin' value: (self creatorLogin) default: nil.
	pStream storeNumber: 'lastModifiedTime' value: (self lastModifiedTime) default: nil.
! !

!QAnnotation methodsFor: 'menu' stamp: 'bgf 8/1/2009 18:09'!
photographerClass
	^ QDirectPhotoJob! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 8/3/2008 19:26'!
pickable: aBool

	super pickable: aBool.

	"Update drag adapter availability; pickable-false should suppress everything.
	 (Drag adapaters use their own event-handling scripts.)"

	self setPlacementLocked: (self isPlacementLocked).

! !

!QAnnotation methodsFor: 'events' stamp: 'bgf 2/15/2010 17:57'!
pointerClick: event
	
	"The default annotation click behavior is inaction."
! !

!QAnnotation methodsFor: 'events' stamp: 'bgf 1/28/2008 21:34'!
pointerEnter: event
	self pointerOver: event.! !

!QAnnotation methodsFor: 'events' stamp: 'bgf 11/13/2008 13:07'!
pointerLeave: event
	event avatar ifNil: [ ^ self ].
	self pickable ifFalse: [ ^ self ].
	event avatar clearFloatMessageForEvent: event.

! !

!QAnnotation methodsFor: 'events' stamp: 'bgf 11/13/2008 13:07'!
pointerOver: event
	event avatar ifNil: [ ^ self ].
	self pickable ifFalse: [ ^ self ].

	event avatar showFloatMessage: (self floatingMessageFor: event avatar) forEvent: event.
! !

!QAnnotation methodsFor: 'updating' stamp: 'bgf 8/20/2009 11:14'!
resizeComplete: event
	"General catchall for objects with surface resizers;
	 make sure the attachment-transform is updated when done."
	self root attach: self to: (self surfaceAttachment).! !

!QAnnotation methodsFor: 'updating' stamp: 'bgf 6/1/2009 12:40'!
setPlacementLocked: aBool
	"For subclass extension; enable/disable direct move-resize controls.
	 All such affordance should be disabled if aBool is false
	 or if pickable is false (i.e. pickable should override everything.) "
	super isPlacementLocked: aBool.

	adapter ifNil: [ ^ self ].			"Initializing base"
	self pickable ifTrue: [
		aBool ifTrue: [
			adapter enableClickOnly
		] ifFalse: [
			adapter enable.
		].
	] ifFalse: [
		adapter disable
	].
! !

!QAnnotation methodsFor: 'accessing' stamp: 'bgf 8/28/2008 00:40'!
titleString
	"The composed string for my title line."

	^ self creatorName ifNil: [
		'Note'
	] ifNotNilDo: [ :nm|
		'Note (', nm, ')'
	].! !

!QAnnotation methodsFor: 'updating' stamp: 'bgf 8/15/2009 17:56'!
translateIntoContactWith: aPoint alongNormal: aNormal
	"Extended so we stand off from the attachment point by our attachmentOffset"
	
	super translateIntoContactWith: aPoint alongNormal:aNormal.
	self translation: (self translation + (aNormal * self attachmentOffset)).! !

QAnnotation subclass: #QGroupOfNotes
	instanceVariableNames: 'back resize pendingAttachments preResizeNotes resizeNotes pinPoint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QGroupOfNotes commentStamp: '<historical>' prior: 0!
This is a rectangular annotation (note-collector content)
whose role is to group other notes and topics.
!

!QGroupOfNotes class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QGroupOfNotes class methodsFor: 'inspector' stamp: 'sn 3/8/2013 12:01'!
addPropertyRows: propertyDialog forFrame: gnref

	self addNameRowIfNeeded: propertyDialog forFrame: gnref.

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [: f | f get: #extent ];
			acceptBlock: [ : ext : f| f future extent: ext ];
			balloonHelp: 'Dimensions of the rectangle'translateMe;
		yourself).

	super addPropertyRows: propertyDialog forFrame: gnref.! !

!QGroupOfNotes class methodsFor: 'drag-drop' stamp: 'bgf 8/15/2009 13:41'!
canAcceptDrop: payload

	^ false! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:12'!
defaultBacking
	"The material of the group's body sheet.
	 Fullbright to preserve the exact hue as a UI signature."
	^ (QMaterials fullWhiteMaterial)
		diffuseColor: (QColors groupOfNotesColor);
		texture: (QTextures find: 'paleStipple.jpg');
		uvScale: 2@2;
		fullBright: true;
		yourself! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 19:11'!
defaultBorderMaterial
	"The material with which to border the grouping."
	| color |
	color := VectorColor r:0.82 g: 0.75 b: 0.7.
	^ (TMaterial new)
		diffuseColor: (color * 0.95 alpha: 0.99);
		ambientColor: (color * 0.5);
		fullBright: false;
		yourself! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 8/28/2009 00:39'!
defaultBorderWidth
	"The absolute width of the backing border."
	^ 0.02! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 8/15/2009 12:50'!
defaultExtent
	"Default size for an empty new group."
	^ 2@2! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 8/28/2009 01:36'!
defaultSweepoutBacking
	"The material to use for an initial marquis-sweep, before we've decided to make a permanent group."
	^ (TMaterial new)
		diffuseColor: #(0.03 0.25 0.55 0.2);
		fullBright: true;
		cullFace: false;
		yourself
! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 8/28/2009 01:36'!
defaultSweepoutBorderMaterial
	"The border for an initial marquis-sweep, before we've decided to make a permanent group."
	^ (TMaterial new)
		diffuseColor: #(0.1 0.3 0.6 0.75);
		fullBright: true;
		cullFace: false;
		yourself
! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 8/15/2009 17:51'!
floatingMessage
	^ nil! !

!QGroupOfNotes class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 08:47'!
hiliteColor
	^VectorColor r: 1.0 g: 0.05 b: 0.05 a: 0.5! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'sn 12/27/2012 15:17'!
addMenuActions: tmenu forEvent: event
	"Custom context menu."

	| pop sa editProtect |
	editProtect := (self fromTemplate and: [ event avatar isEditingTemplate not ])
				  or: [ event avatar canModifyForum not ].
	pop := self population.

	event avatar allowedToMove ifTrue: [
		tmenu addItemSpec: { 'Approach' translateMe. #bringAvatarHere: } target: self.
		tmenu addSeparator
		].

	editProtect ifFalse: [
		tmenu add: 'Sticky Note...' translateMe
		    target: self
		    action: #createStickyNote:
	    	    arguments: { event }.
		tmenu add: 'Topic Card...' translateMe
		    target: self
		    action: #createTopicCard:
		    arguments: { event }.
		tmenu addSeparator.
		self nBallots > 1 ifTrue: [
			tmenu add: 'Create a Poll...' translateMe
				target: self
				action: #generatePoll:
				arguments: { event }.
			tmenu addSeparator.
		]
	].

	"The note browser covers the outer manager I live in."
	sa := self surfaceAttachment.
	(sa isNil or: [ sa isNoteCollector not ])  ifTrue: [ sa := self ].
	tmenu add: 'Browse Topics...' translateMe
			    target: sa 	
			    action: #reviewTopics:
			    arguments: { event }.
	"But saving as text is scoped to me..."		
	(tmenu add: 'Save As Text...' translateMe
	   	 	target: self
	    		action: #saveTopicSummary:
	   		arguments: { event avatar })
			enabled: (self notes isEmptyOrNil not).
	tmenu addSeparator.

	((self canBeCopiedBy: event avatar) and: [ event avatar canUseInventory ])  ifTrue: [
		tmenu addItemSpec: { 'Copy to Inventory' translateMe . #copyToInventory: } target: self
	].

	tmenu addSeparator.
	editProtect ifFalse: [
		tmenu add: 'Material...' translateMe target: (event avatar) action: #editFrameMaterial:  argument: self.
		tmenu addItemSpec: { 'Properties...' translateMe. #editProperties: } target: self.
		self isPlacementLocked ifFalse: [
			tmenu addSeparator.
			tmenu addItemSpec: {
				(pop > 0 ifTrue: ['Ungroup' translateMe] ifFalse: ['Discard' translateMe]) . #menuUngroup: } 
			target: self.
			(pop > 0) ifTrue: [	
				tmenu addSeparator.
				tmenu addItemSpec: {'Delete All' translateMe. #confirmDeleteAll: } target: self.
			]]].

! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/16/2009 18:15'!
addNote: aNote
	super addNote: aNote.
	self signal: #notesChanged.! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 18:24'!
asPlace
	"The QPlace descriptor for this destination."
	^	(QPlace 
				category: 'Panels'
			  	label:  'Group of Topics' 
			  	frame: self)
			cameraSelector: #cameraStanceFor:;
			yourself! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 2/3/2010 10:51'!
asTopicRecord
	| n |
	n := self stickers size.
	^ (QTopicRecord new)
			cardRef: self;
			author: (self creatorLogin);
			color: (Color gray:0.2);
			creationTime: 0;
			lastModifiedTime: ((self  lastModifiedTime) ifNil: [ 0 ]);
			topic: (self labelForBrowser);
			nComments: n;
		yourself! !

!QGroupOfNotes methodsFor: 'hierarchy' stamp: 'bgf 8/16/2009 18:18'!
attachedTo: aSurface
	super attachedTo: aSurface.
	self signal: #notesChanged.! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 15:56'!
attachmentOffset
	"For z-buffer relief... but very litte, since we want to be under everything else."
	^ 0.01! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 11:44'!
backingMaterial: aMaterial
	back material: aMaterial.
! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 5/25/2010 19:02'!
ballots
	"Return all my ballot-candidate attachments."
	^ QPollBarChart sortBallots:  (self notes select: [ : n | QPollBarChart isBallotCandidate: n ]) 
	! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/28/2009 00:58'!
borderMaterial
	^ back borderMaterial! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/28/2009 00:58'!
borderMaterial: bm
	^ back borderMaterial: bm! !

!QGroupOfNotes methodsFor: 'transform' stamp: 'bgf 8/15/2009 14:50'!
cameraStanceFor: anAvatar

	^ QGotoButton frame: self cameraStanceFor: anAvatar! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 4/1/2010 09:02'!
canBeCopiedBy: anAvatar
	^ (super canBeCopiedBy: anAvatar) 
	  and: [ self notes allSatisfy: [ : n | n canBeCopiedBy: anAvatar ] ]! !

!QGroupOfNotes methodsFor: 'testing' stamp: 'bgf 10/25/2009 14:53'!
canEditInBox
	"The edit-box is inappropriate for these.  Distorts the surface-attachment."
	^ false! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 11:44'!
canEditMaterial
	^ true! !

!QGroupOfNotes methodsFor: 'render' stamp: 'bgf 8/15/2009 10:31'!
child: c setDistance: ogl
	"Treat children as equidistant so we can control subdistance.
	 The resize grip needs a negative subdistance so it is rendered as if closer than the bottom strip."
	frameParent ifNotNil: [
		^ (frameParent child: self setDistance: ogl) - Float epsilon.
	].
	^ super child: c setDistance: ogl
! !

!QGroupOfNotes methodsFor: 'render' stamp: 'bgf 8/15/2009 10:31'!
childSubdistance: c
	"The resize grip needs a negative subdistance so it is visible at all angles."
	^ c == (resize grip) ifTrue: [ -0.1 ] ifFalse: [ 0 ]
! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 11/17/2009 13:57'!
confirmDeleteAll: event

	event avatar ifNil: [ ^ self ].
	self root ifNil:  [ ^ self ].	"Safety check... in case this is already happening."

	event avatar confirm: 'Delete Entire Group?'		
			 	selector:  #menuDeleteAll:
				target: self

! !

!QGroupOfNotes methodsFor: 'persist' stamp: 'bgf 8/28/2009 00:59'!
constructFrom:pStream

	super constructFrom:pStream.
	self extent: (pStream readPoint: 'extent' default: self class defaultExtent).
	self backingMaterial: (pStream readObject: 'backing' default: back material).
	self borderMaterial: (pStream readObject: 'border' default: back borderMaterial).

	"See #root:.  Works in conspiracy with the surface manager to persist and recover notes."
	pendingAttachments := pStream readSequence: 'attachments' default: #().! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 8/19/2009 15:54'!
contains: note sweepingFrom: topLeft to: bottomRight orient: orient
	"Private support for marquis rubberbanding.
	 We don't require full overlap-inclusion, just that the center be included.	
         But nonetheless the group will not enclose anything larger than itself."
	| trd lbd |

	(note extent x < self extent x and: [ note extent y < self extent y ])
		ifFalse: [ ^ false ].

	trd := orient transformVector: (note translation - topLeft).
	(trd x > Float epsilon and: [ trd y < Float epsilon ]) 
		ifFalse: [ ^ false ].

	lbd := orient transformVector: (note translation - bottomRight).
	(lbd x < Float epsilon    and: [ lbd y >  Float epsilon ] )
		ifFalse: [ ^ false ].

	^ true! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 23:48'!
copyToInventory: event

	event avatar addObjectToInventory: self ! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 12:22'!
createStickyNote: event
	"Make a sticky note at the event-selection point."
	event avatar ifNotNil: [
		event avatar addAnnotation: QStickyNote toFrame: self selection: event selection
	].! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 12:22'!
createTopicCard: event
	"Make a topic card (QTopicCard) at the event-selection point."
	event avatar ifNotNil: [
		event avatar addAnnotation: QTopicCard toFrame: self selection: event selection
	].! !

!QGroupOfNotes methodsFor: 'initialize' stamp: 'bgf 8/15/2009 10:54'!
creator: anAvatar
	"Have the note remember who made it."

	anAvatar ifNotNil: [
		self creatorLogin: (anAvatar userLogin).
	] ifNil: [
		self creatorLogin ifNotNil: [
			self creatorLogin: nil.		"Decouple further updates."
		].
	].! !

!QGroupOfNotes methodsFor: 'transform' stamp: 'bgf 8/15/2009 18:49'!
destinationExtent
	"For destination distance-finding."
	^ back destinationExtent *  (self population > 0 ifTrue: [ 1.25 ] ifFalse: [ 4.0 ]).! !

!QGroupOfNotes methodsFor: 'transform' stamp: 'bgf 8/15/2009 14:56'!
destinationStanceFor: anAvatar

	"The stance to apply to the avatar to bring it to me."
	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self.

! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 8/21/2009 12:11'!
drag: cornerPt corner: compassPt delta: dcorner
	"Progressive drag, repeatedly called when sweeping out the note group."
	| newSize dt pos topLeft bottomRight origin |
	
	pos := self translation clone.
	origin := self globalToLocal: pinPoint.
	topLeft := origin.
	bottomRight := (cornerPt x) @ (cornerPt y).
	topLeft x > bottomRight x ifTrue:  [
		topLeft := (cornerPt x) @ ( topLeft y ).
		bottomRight := (origin x) @ (cornerPt y)
	].
	topLeft y < bottomRight y ifTrue: [
		topLeft := (topLeft x) @ (cornerPt y).
		bottomRight := (bottomRight x) @ (origin y) 
	].
	newSize := (bottomRight x - topLeft x) @ (topLeft y - bottomRight y).
	dt :=  Vector3 x: (bottomRight x + topLeft x)
		  		  y: (bottomRight y + topLeft y) 
				  z: 0.00001.		"Encouragement to keep from moving away."
	self  translation: (pos + (self localTransform transformVector: dt * 0.5 ));
		extent: newSize.

	self resizeSweep: topLeft to: bottomRight.! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 8/15/2009 10:31'!
dropOn: aTarget avatar: avatar selection: aTXSel

	"XXX -draft code, inactive until #initialize is changed to make my adapter Droppable."

	"Special case... if I am dragged into a display sheet or panel,
	 I destroy myself - the panel takes up the picture."
	((aTarget isKindOf: QDisplayPanel) or: [aTarget isKindOf: QDisplaySheet]) ifTrue: [
		aTarget consumeDrop: (self payload) avatar: avatar selection: aTXSel.
		self future destroy.
		^ true
	].

	"Don't allow auto-open of displays on walls and floors... posters must be just movables to them."
	((aTarget isKindOf: QFixedSurface) or: [ aTarget isKindOf: QFloor]) ifFalse: [

		^ aTarget consumeDrop: (self payload) avatar: avatar selection: aTXSel.
	].
		
	^ true	"Don't snap back".! !

!QGroupOfNotes methodsFor: 'updating' stamp: 'bgf 8/15/2009 15:33'!
editByUser: aUser
	"Used by the topics tool... so give a response. "

	aUser goToPlace: self asPlace
! !

!QGroupOfNotes methodsFor: 'updating' stamp: 'bgf 8/15/2009 21:56'!
editByUser: anAvatar atCreation: bIsNew
	"Called after an annotate-tool is used to add an annotation."
	"Do nothing, since my default editByUser: action is a go-to,
	 which is disruptive when just making one."! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 13:14'!
editToolClass
	"Only on a reachable flow when I am a 'NoteCollector;
	 the edit action is to open the topics browser."
	^ QTopicsTool! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 10:31'!
extent
	^ back extent! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/21/2009 12:15'!
extent: xy
	"Idempotence check is essential for cooperating with the resize handler."
	back extent: xy.
	self updateShape.
	
! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 2/9/2010 13:42'!
generatePoll: event
	"Make a poll using my ballot-candidate notes,
	 and plop it right here."

	event avatar generatePoll: event ballots: (self ballots)
! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 10:31'!
height
	^ self extent y! !

!QGroupOfNotes methodsFor: 'updating' stamp: 'bgf 8/15/2009 10:31'!
hilite
	self hilite: self class hiliteColor! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 12:45'!
hilite: x
	"Don't bother... too busy."! !

!QGroupOfNotes methodsFor: 'initialize' stamp: 'bgf 11/8/2009 09:34'!
initialize
	super initialize.
	self objectName: 'Topic Group'.
	pendingAttachments := #().
	adapter overColor: ((Color gray: 0.7) alpha: 0.2).
	adapter downColor: ((Color gray: 0.7) alpha: 0.3).
	false ifTrue: [ adapter isDroppable: true ]. 	"Experimental."

	back := TBorderRectangle new.
	back material: (self class defaultBacking);
		borderMaterial: (self class defaultBorderMaterial);
		extent: (self class defaultExtent * 0.75);	"So later finishing sees a change"
		borderWidth: (self class defaultBorderWidth);
		cullFace: false;
		yourself.
	self addChild: back.
	self solid: false.

	resize := QCorner2DResizer forLuggage: self.
	resize transient: true.

	self extent: self class defaultExtent. 	"Ensure margins and resize border are all in harmony."
	self setPlacementLocked: (self isPlacementLocked).		"Configure the resizer."! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 8/15/2009 10:31'!
isDropTarget: payload

	^ self class canAcceptDrop: payload! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 10:31'!
isListedTopic
	"We list posters in the topics card as one of the items you'd use in a stack."
	^ true! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 23:38'!
isNoteCollector
	^ true! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 23:34'!
labelForBrowser
	^ self menuTitle! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 11:44'!
material
	^ back material! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 11:44'!
material: tMaterial
	^ back material: tMaterial! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 2/9/2010 18:57'!
menuDeleteAll: event
	"Explicitly destroy all the attachments on me as well as myself,
	 so they're all selected when we recover the item in undo."
	event avatar ifNotNilDo: [ : av |
		av deleteFrames: (self root stickersOn: self),  { self }  event: event.
	].
! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 23:33'!
menuTitle
	^ (self objectName),  
	   ((self population = 0) ifTrue: [ ' (Empty)' ] ifFalse: [ ' (', (self population), ')' ])! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 2/4/2010 12:20'!
menuUngroup: event

	"For future undo... the command will need to record these attachments
	 so that when undone, the notes return to the group."
	self surfaceAttachment ifNotNilDo: [ : surface |
		( self stickers) do: [ : sticker |
			self root attach: sticker to: surface
		]].
	
	self performDelete: event. event.! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 10:31'!
minExtent
	"Corner-resize protocol - stay out of trouble."
	^ 0.4@0.4! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 2/9/2010 14:22'!
nBallots
	"How many ballot-candidates are attached to me?"
	^ self notes count: [ : n | QPollBarChart isBallotCandidate: n ].! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 13:29'!
notes
	"All the  Listed-Topics (notes for topics tools)  attached eventually to me."
	^ (self root ifNil: [ #() ] ifNotNil: [  : space | space stickersOn: self ])
		select: [ : sticker | sticker isListedTopic ]! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/16/2009 22:58'!
objectName: aString
	super objectName: aString.
	self signal: #contentsChanged.		"for topic tool updates."! !

!QGroupOfNotes methodsFor: 'persist' stamp: 'bgf 8/28/2009 01:35'!
persistOn:pStream
	super persistOn:pStream.

	pStream storePoint: 'extent' value: self extent default: self class defaultExtent.	
	pStream storeObject: 'backing' value: back material default: self class defaultBacking.
	pStream storeObject: 'border' value: self borderMaterial default: (self class defaultBorderMaterial).

	"If we're coming out of a rooted context, store the notes-relations as they stand here;	
	 otherwise transcribe what we remember (the transcribe case applies to inventory-import,
 	 where the harness unpacks and then injects island-bits.)"
	self root ifNotNilDo: [ : space | 
		space attachments 
			persistNotes: ( self notes )  for: self on: pStream as: 'attachments'.
	] ifNil: [
		pStream storeSequence: 'attachments' value: pendingAttachments default: #()
	].			
! !

!QGroupOfNotes methodsFor: 'events' stamp: 'bgf 10/25/2009 12:58'!
pointerClick: event
	"No click response just yet."! !

!QGroupOfNotes methodsFor: 'events' stamp: 'bgf 8/15/2009 10:31'!
pointerDown: event
	super pointerDown: event.
	self isPlacementLocked ifFalse: [ resize grip visible: true ].! !

!QGroupOfNotes methodsFor: 'events' stamp: 'bgf 8/15/2009 12:47'!
pointerEnter: event

	self isPlacementLocked ifFalse: [ resize grip visible: true ].! !

!QGroupOfNotes methodsFor: 'events' stamp: 'bgf 8/15/2009 10:31'!
pointerLeave: event
	super pointerLeave: event.
	resize grip visible: false! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 2/3/2010 10:47'!
population
	^ self stickers size! !

!QGroupOfNotes methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:56'!
privateScale: aFactor

	"Keep my layout consistent."
	nomScale := aFactor * nomScale.
	self extent: (self extent * aFactor).! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 2/9/2010 18:58'!
reallyDelete: event
	"Since the persistent capture of something with stickers captures the stickers,
	 the default behavior here is to delete the stickers explicitly;
	 so <Del> by keyboard is the same as Delete-All, rather than ungroup."

	(self root stickersOn: self) do: [ : sticker | sticker reallyDelete: event ].
	super reallyDelete: event.! !

!QGroupOfNotes methodsFor: 'persist' stamp: 'bgf 8/15/2009 15:18'!
recoverPendingAttachments
	"If we're holding attachments waiting to be re-instated into a space, now is the time."
	self root ifNotNil: [ 
		self root attachments recoverNotesFrom: pendingAttachments for: self.
		pendingAttachments := #().
	].! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/16/2009 18:16'!
removeNote: aNote
	super removeNote: aNote.
	self signal: #notesChanged.! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 2/9/2010 14:07'!
resizeBegin: event
	"My resize process is actually a marquis-sweep-selection,
	 Save my initial topLeft as origin, since that's what the marquis sweeps out from."
	self resizeBegin: event 
		forInitialPoint: (self localToGlobal: (self extent x * -0.5) @ (self extent y * 0.5) @ 0).! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 2/3/2010 10:49'!
resizeBegin: event forInitialPoint: pp
	preResizeNotes ifNil: [
		preResizeNotes := self stickers.
		preResizeNotes do: [ : note |
			"Stick'em to what I'm stuck to; I'll catch them in sweep, or not..."
			self root attach: note to: (self surfaceAttachment).
		]].
	pinPoint := pp.
	resizeNotes := nil.! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 8/20/2009 12:29'!
resizeCanceled
	preResizeNotes ifNotNil: [
		preResizeNotes do: [ : note |
			self root attach: note to: self
		]].
	preResizeNotes := nil.
	resizeNotes := nil.
	pinPoint := nil.! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 8/20/2009 13:26'!
resizeComplete: event
	"We've done a rubberband-resize. 
	 Sweep up all the discovered items and take over their attachments."
	resizeNotes ifNil: [
		self resizeCanceled.	"Not even one sweep-state was made."
	] ifNotNil: [
		| normal global invGlobal |
		normal := self lookAt.
		global := self globalTransform.
		invGlobal := global inverse.
		resizeNotes do: [ : note |
			| pos nm |
			pos := (note globalPosition 
					+ (note globalTransform transformVector:(note positioningBox center))).
			pos := (invGlobal transformPoint: pos).
			pos := global transformPoint: (pos x @ pos y @ 0.001).
			note parent = note root ifFalse: [ 
				pos := note parent globalTransform inverse transformPoint: pos.
				nm := note parent globalOrientation transformVector: normal.
			].
			"Now reinforce the attachment offsets, so they appear in front of me."
			note globalPosition: (pos).
			note allUpdateShape.
			note translateIntoContactWith:  pos alongNormal: (nm ifNil: [ normal ]).
			self root attach: note to: self
		].
	] .
	preResizeNotes := nil.
	resizeNotes := nil.
	pinPoint := nil.

	super resizeComplete: event.! !

!QGroupOfNotes methodsFor: 'drag-drop' stamp: 'bgf 2/3/2010 10:49'!
resizeSweep: ltl to: lbr
	"Sweep the attachment set the of the surface underneath me, looking for ins and outs."
	| sa sao topLeft bottomRight candidates orient |

	sa := self surfaceAttachment.
	sa ifNotNil: [ 
		candidates := ((sa stickers) copyWithout: self) 
						select: [ : n | n isListedTopic and: [ n visible ]]].
	candidates isEmptyOrNil ifTrue: [ 
		resizeNotes := #().
		^ self ].
	orient := self globalOrientation.
	sao := orient inverse.
	topLeft := self translation + ( sao transformVector: (ltl @ 0) ).
	bottomRight := self translation + ( sao transformVector: (lbr @ 0) ).
	resizeNotes := candidates select: [ : note |
		self contains: note sweepingFrom: (topLeft) to: (bottomRight) orient: orient ].
! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 18:46'!
reviewTopics: anEvent
	"Open the topic-review browser.  
	 This is only reachable when I am a NoteCollector,
	 and not attached within a larger one."
	QTextEditJob
		edit: (self objectName asString)
		title: ('Topic Group' )
		forUser: (anEvent avatar)
		recipient: self
		selector: nil.! !

!QGroupOfNotes methodsFor: 'hierarchy' stamp: 'bgf 8/15/2009 14:47'!
root: aSpace

	"After recovery from persistence, recover any internally-persisted notes I carry.
	 (The mechanism will ignore attachments that are already recovered normally,
	 since it only re-instates attachments that are still unrooted at that time.
	 So recovery applies when importing a subscene, but not when loading a forum.)"
	self future recoverPendingAttachments.
	super root: aSpace.! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 2/21/2010 08:35'!
saveTopicSummary: anAvatar
	"We do the save-summary by generating a text blob,
 	 and using save-local-copy of a temporary TextDocument"
	| txt doc |
	txt := self topicSummaryText.
	doc :=	(QTextDocument forString: txt).
	doc shortName: 'notes.txt'.
	QSaveDocumentJob save: doc  forUser: anAvatar.! !

!QGroupOfNotes methodsFor: 'updating' stamp: 'bgf 8/15/2009 10:31'!
setPlacementLocked: aBool
	super setPlacementLocked: aBool.
	"The resizer has to be suppressed when locked or non-pickable."
	resize ifNotNil: [ resize isEnabled: (adapter isEnabled)].! !

!QGroupOfNotes methodsFor: 'updating' stamp: 'bgf 8/28/2009 01:32'!
setupForPermanance
	"Color the group as a noteboard element (rather than a marquis sweep)."
	back material: (self class defaultBacking);
		  borderMaterial: (self class defaultBorderMaterial).! !

!QGroupOfNotes methodsFor: 'updating' stamp: 'bgf 8/28/2009 01:31'!
setupForSweep
	"Color the group for the marquis-sweeping gesture."
	back material: (self class defaultSweepoutBacking);
		  borderMaterial: (self class defaultSweepoutBorderMaterial).! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 10/25/2009 16:12'!
showContextMenu: event
	"Context menu is customized. (Many things added, so some tactical omissions.)"
	self showMenu: event! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 8/15/2009 13:26'!
topicRecords
	^ self notes collect: [ : n | n asTopicRecord ]! !

!QGroupOfNotes methodsFor: 'menu' stamp: 'bgf 2/21/2010 08:59'!
topicSummaryText
	"A string comprising a text report of all the topics and notes in this group."
	| sorted |
	^ String streamContents: [ : stream |
		stream nextPutAll: self objectName; cr.
		stream nextPutAll: 'Notes and Topics ('.
		DateAndTime now printYMDOn: stream.
		stream space.
		DateAndTime now printHMSOn: stream.
		stream nextPut: $).
		stream cr; cr.
		"Sort by date - most recent first - to match the browser"
		sorted := self notes
					sort: [ : n1 : n2 | 
							(n1 lastModifiedTime ifNil: [0]) > (n2 lastModifiedTime ifNil: [0]) ].
		sorted do: [ : note |
			(note respondsTo: #writeTranscriptOn:) ifTrue: [
				note writeTranscriptOn: stream.
				stream cr.
			]].
	].
! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 10:31'!
unhilite: x
	"Don't bother."! !

!QGroupOfNotes methodsFor: 'hierarchy' stamp: 'bgf 8/28/2009 01:34'!
updateShape
	"Fit my grip into the standard place at the corner."

	| ext gext z |
	ext := self extent.
	gext := (ext asPoint minDimension * 0.09) @ (ext asPoint minDimension * 0.09).
	gext := gext max: 0.1@0.1.
	gext := gext min: (0.25@0.25).
	z := 0.003.
	resize grip
		extent: gext;
		translation: ((ext x * 0.5 + (back borderWidth)) - (gext x * 0.5)) 
					@ ((ext y * -0.5 - (back borderWidth)) + (gext y * 0.5)) 
					@ z.

	self updateBounds.! !

!QGroupOfNotes methodsFor: 'transform' stamp: 'bgf 8/15/2009 15:04'!
warpThresholdAdjustment
	"We want to drag/arrange them from afar."
	^ 4! !

!QGroupOfNotes methodsFor: 'accessing' stamp: 'bgf 8/15/2009 10:31'!
width
	^ self extent x! !

QAnnotation subclass: #QMirror
	instanceVariableNames: 'back resize portal glass gotoButton isListedLandmark'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QMirror commentStamp: '<historical>' prior: 0!
A draggable, resizeable mirror.!

!QMirror class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMirror class methodsFor: 'inspector' stamp: 'sn 3/8/2013 12:01'!
addPropertyRows: propertyDialog forFrame: mref

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :f | f get: #extent ];
			acceptBlock: [ : ext :f | f  future extent: ext ];
			balloonHelp: 'Dimensions of the rectangle.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QBooleanPropertyRow new)	
			titleString: ' Show in Places List';
			fetchBlock: [ :f | f  get: #isListedLandmark ];
			acceptBlock: [ : v :f | f  future isListedLandmark: v ];
			balloonHelp: 'If checked, the mirror''s name will be listed in Places.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QBooleanPropertyRow new)	
			titleString: ' Vampire Mode';
			fetchBlock: [ :f | f  get: #vampireMode ];
			acceptBlock: [ : v :f | f  future vampireMode: v ];
			balloonHelp: 'If checked, avatars will not be visible in the mirror.'translateMe;
		yourself).

	super addPropertyRows: propertyDialog forFrame: mref.! !

!QMirror class methodsFor: 'defaults' stamp: 'bgf 6/1/2009 17:05'!
defaultBacking
	"The material with which to back the poster."
	^ (QMaterials greyMaterial)
		cullFace:false;
		opacity: 0.8;
		yourself.! !

!QMirror class methodsFor: 'defaults' stamp: 'bgf 10/4/2007 10:19'!
defaultExtent
	"Default size for a poster."
	^ 3@4! !

!QMirror class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:12'!
defaultIsFromTemplate
	"Reduces xml volume since these are often premade."
	^true ! !

!QMirror class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 08:47'!
hiliteColor
	^VectorColor r: 0.24 g: 0.24 b: 0.3 a: 0.2! !

!QMirror methodsFor: 'transform' stamp: 'bgf 4/26/2009 11:58'!
asPlace
	^ (gotoButton asPlace)
		cameraSelector: #cameraStanceFor:;
		yourself! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 10/4/2007 10:56'!
attachmentOffset
	"For z-buffer relief - stand off from surfaces just enough."
	^ 0.05! !

!QMirror methodsFor: 'transform' stamp: 'bgf 4/26/2009 14:09'!
cameraStanceFor: anAvatar
	
	^ QGotoButton frame: glass cameraStanceFor: anAvatar! !

!QMirror methodsFor: 'render' stamp: 'bgf 9/5/2008 14:37'!
child: c setDistance: ogl
	"Treat children as equidistant so we can control subdistance.
	 The resize grip needs a negative subdistance so it is rendered as if closer than the bottom strip."
	frameParent ifNotNil: [
		^ (frameParent child: self setDistance: ogl) - Float epsilon.
	].
	^ super child: c setDistance: ogl
! !

!QMirror methodsFor: 'render' stamp: 'bgf 9/5/2008 14:37'!
childSubdistance: c
	"The resize grip needs a negative subdistance so it is visible at all angles."
	^ c == (resize grip) ifTrue: [ -0.1 ] ifFalse: [ 0 ]
! !

!QMirror methodsFor: 'persist' stamp: 'bgf 8/4/2008 01:10'!
constructFrom:pStream
	super constructFrom:pStream.
	self extent: (pStream readPoint: 'extent' default: self class defaultExtent).
	self isListedLandmark: (pStream readBool: 'isListedLandmark' default: (self isListedLandmark)).! !

!QMirror methodsFor: 'transform' stamp: 'bgf 4/26/2009 14:59'!
destinationExtent
	"For destination distance-finding."
	^ glass destinationExtent! !

!QMirror methodsFor: 'transform' stamp: 'bgf 4/26/2009 11:58'!
destinationStanceFor: anAvatar
	"The stance to apply to the avatar to bring it to me."
	^ QGotoButton groundedStanceFor: anAvatar toView: glass inDestination: self.
! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 10/4/2007 10:19'!
extent
	^ back extent! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 1/20/2009 19:12'!
extent: xy

	"Idempotence check is essential for cooperating with the resize handler."
	| rt |
	(xy ~= self extent) ifTrue: [
		portal extent: xy.
		glass extent: xy *1.01.
		back extent: xy.
		back translation: (Vector3 x: 0 y: 0 z: ((-0.005*(xy maxDimension)) min: -0.005)).
	].
	resize updateShape.
	rt := resize grip translation.
	resize grip translationX: rt x y: rt y z: (0.035).		"Above glass"
	gotoButton translationX: 0 y: (xy y + gotoButton extent y)*0.5 + 0.5 z: 0.! !

!QMirror methodsFor: 'updating' stamp: 'bgf 1/27/2008 03:09'!
hilite
	"Do nothing - mirror milk on pointer-over just distracts."! !

!QMirror methodsFor: 'updating' stamp: 'bgf 1/27/2008 03:08'!
hilite: aColor
	"Do nothing - mirror milk on pointer-over just distracts."! !

!QMirror methodsFor: 'initialize' stamp: 'bgf 1/2/2009 12:45'!
initialize
	|  |
	super initialize.
	
	self objectName: 'Mirror'.
	adapter overColor: self class hiliteColor.
	adapter downColor: self class hiliteColor.
	adapter attractionRange: 4.0. 	"Moderately sticky... we don't want to track the ground too aggressively"
	adapter isGroundBound: false.

	portal := (TPortal new)
		extent: (self class defaultExtent).
	portal blocked: true.
	self addChild: portal.

	glass := TRectangle new.
	glass material: QMaterials mirrorGlassMaterial.
	glass translation: (0@0@0.025).
	self addChild: glass.

	back := TRectangle new.
	back material: (self class defaultBacking);
		extent: self class defaultExtent;
		translation:(0@0@-0.01);
		normal: nil;
		cullFace: false;
		yourself.
	self addChild: back.

	gotoButton :=  QGotoButton new.
	gotoButton destination: self.
	gotoButton label: 'Mirror'.
	self addChild: gotoButton.
	gotoButton disable.

	self solid: false.

	resize := QCorner2DResizer forLuggage: self.
	self pickable ifTrue: [ resize enable ].	
	resize transient: true.

	self extent: self class defaultExtent. 	"Ensure margins, and resize border, are all in harmony."! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 8/4/2008 01:06'!
isListedLandmark
	^ isListedLandmark ifNil: [ false ]! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 8/4/2008 01:08'!
isListedLandmark: aBool

	"If I'm a listed landmark (for the Places list) then I need a listed category."
	(isListedLandmark ~= aBool) ifTrue: [
		gotoButton ifNotNil: [
			gotoButton category: (aBool ifTrue: [ 'Named Places' ] ifFalse: [ 'Mirrors' ] ).
		].
		isListedLandmark := aBool.
	].! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 8/1/2009 12:06'!
isListedTopic
	^ false! !

!QMirror methodsFor: 'menu' stamp: 'bgf 10/4/2007 10:20'!
menuTitle
	^  'Mirror'! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 9/5/2008 14:37'!
minExtent
	"Corner-resize protocol - stay out of trouble."
	^ 0.4@0.4
! !

!QMirror methodsFor: 'initialize' stamp: 'bgf 3/5/2010 18:28'!
objectID: oid
	super objectID: oid.
	gotoButton ifNotNil: [ gotoButton destination: self ].! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 8/4/2008 01:06'!
objectName: aString
	super objectName: aString.
	gotoButton ifNotNil: [
		gotoButton label: aString
	].! !

!QMirror methodsFor: 'persist' stamp: 'bgf 8/4/2008 01:10'!
persistOn:pStream
	super persistOn:pStream.
	pStream storePoint: 'extent' value: self extent default: self class defaultExtent.
	pStream storeBool: 'isListedLandmark' value: (self isListedLandmark) default: false.! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 9/5/2008 14:41'!
pickable: aBool
	"Disable the resizer if we're considered non-pickable."
	super pickable: aBool.
	aBool ifTrue: [
		resize ifNotNil: [ resize enable ].
	] ifFalse: [
		resize ifNotNil: [ resize disable ].
	].! !

!QMirror methodsFor: 'events' stamp: 'bgf 10/25/2009 12:59'!
pointerClick: event
	"From the adapter.  No handling as yet."! !

!QMirror methodsFor: 'events' stamp: 'bgf 9/5/2008 14:37'!
pointerDown: event
	super pointerDown: event.
	self isPlacementLocked ifFalse: [ resize grip visible: true ].
! !

!QMirror methodsFor: 'events' stamp: 'bgf 9/5/2008 14:38'!
pointerEnter: event
	super pointerEnter: event.
	self isPlacementLocked ifFalse: [ resize grip visible: true ].
! !

!QMirror methodsFor: 'events' stamp: 'bgf 9/5/2008 14:38'!
pointerLeave: event
	super pointerLeave: event.
	resize grip visible: false
! !

!QMirror methodsFor: 'events' stamp: 'bgf 2/10/2008 12:34'!
pointerOver: event
	"Overridden to suppress the annotation mouseover message."! !

!QMirror methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:57'!
privateScale: aFactor

	"Keep my layout consistent."
	nomScale := aFactor * nomScale.
	self extent: (self extent * aFactor).! !

!QMirror methodsFor: 'updating' stamp: 'bgf 6/1/2009 12:41'!
setPlacementLocked: aBool
	super setPlacementLocked: aBool.
	"The resizer has to be suppressed when locked or non-pickable."
	resize ifNotNil: [
		((aBool not) & (self pickable)) ifTrue: [
			resize enable
		] ifFalse: [
			resize disable
		].
	].! !

!QMirror methodsFor: 'updating' stamp: 'bgf 1/27/2008 03:08'!
unhilite
	"Do nothing - mirror milk on pointer-over just distracts."! !

!QMirror methodsFor: 'accessing' stamp: 'das 7/23/2008 20:02'!
vampireMode

	^ portal vampireMode.! !

!QMirror methodsFor: 'accessing' stamp: 'das 7/23/2008 20:01'!
vampireMode: vm

	portal vampireMode: vm.! !

!QMirror methodsFor: 'accessing' stamp: 'bgf 11/17/2009 16:50'!
visible: aBool
	"Portals are rendered without regard to the visibility of their containers."
	portal ifNotNil: [ portal visible: aBool ].
	super visible: aBool! !

QAnnotation subclass: #QPoster
	instanceVariableNames: 'picture back resize gotoButton margin card'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QPoster commentStamp: '<historical>' prior: 0!
A lightweight, movable, resizeable picture display.
Accepts a picture drop (from a file card)  and displays it in a simple white surround.
Can be dragged - gives sticky-note style surface-attraction.
Click for menu to delete.!

!QPoster class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPoster class methodsFor: 'inspector' stamp: 'sn 3/8/2013 12:00'!
addPropertyRows: dialog forFrame: posterref
	| docs |
	dialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :frame | frame get: #extent ];
			acceptBlock: [ : ext :frame | frame future extent: ext ];
			balloonHelp: 'Dimensions of the rectangle'translateMe;
		yourself).
	dialog addRow: (
		(QScalarPropertyRow new )
			titleString: 'Margin'translateMe;
			fetchBlock: [ :frame | frame get: #margin ];
			acceptBlock: [ :v :frame | frame future margin: v ];
			balloonHelp: 'Portion of the poster used as border matte.';
		yourself).

	docs := (dialog targets collect: [ : cd | cd get: #document ]).
	(docs allSatisfy: [ : d | d isValidFarRef ]) 
	 ifTrue: [
		(QPropertyDialog commonRefClass: docs) ifNotNilDo: [ : klass |
			(klass includesBehavior: QAbstractDocument) ifTrue: [
				| row |
				row := dialog rows size.
				klass addPropertyRows: dialog forFrame: (docs first).
				(row+1)  to: dialog rows size do: [ : i |
					(dialog rows at: i) targets: docs
				].
			] ifFalse: [
				QLogger logInfo: 'Common document class base problem: ', klass
			]]].

	super addPropertyRows: dialog forFrame: posterref.! !

!QPoster class methodsFor: 'drag-drop' stamp: 'bgf 9/9/2007 15:14'!
canAcceptDrop: payload

	"posters can accept only images as drops..."
	^(payload respondsTo:#kind) and: [payload kind = #image]! !

!QPoster class methodsFor: 'defaults' stamp: 'bgf 9/3/2007 23:02'!
defaultBacking
	"The material with which to back the poster."
	^ (QMaterials fullWhiteMaterial)
		cullFace:false;
		yourself.! !

!QPoster class methodsFor: 'defaults' stamp: 'bgf 9/3/2007 13:14'!
defaultContents
	^ QTextures find: 'no.png'! !

!QPoster class methodsFor: 'defaults' stamp: 'bgf 9/3/2007 13:18'!
defaultExtent
	"Default size for a poster."
	^ 3@4! !

!QPoster class methodsFor: 'defaults' stamp: 'bgf 6/22/2008 02:32'!
defaultMargin
	^ 0.09! !

!QPoster class methodsFor: 'defaults' stamp: 'bgf 9/3/2007 13:21'!
floatingMessage
	^ 'To display an image, drop the image card onto the poster.'! !

!QPoster class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:09'!
hiliteColor
	^ QColors redWeakHilite! !

!QPoster methodsFor: 'private' stamp: 'bgf 3/14/2010 17:50'!
applyStateVector: stateVec
	"See #stateVector - this reapplies a captured state."
	self card: (
		stateVec first ifNotNil: [ : pickle |   
			(TSceneReader new)  preserveObjectIDs: true; readScene: pickle 
		] ifNil: [
			nil
		]).
! !

!QPoster methodsFor: 'transform' stamp: 'bgf 4/24/2009 15:48'!
asPlace
	^ (gotoButton asPlace)
		cameraSelector: #cameraStanceFor:;
		yourself! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 2/3/2010 10:49'!
asTopicRecord
	self document ifNotNilDo: [ : doc |
		^ (QTopicRecord new)
			cardRef: self;
			author: (self document ctime who);
			color: (Color black);
			creationTime: (doc ctime time);
			lastModifiedTime: (doc mtime time);
			topic: ('Image: ', doc shortName);
			nComments: (self stickers) size;
		yourself
	].

	^ (QTopicRecord new)
			cardRef: self;
			author: '';
			color: Color black;
			creationTime: 0;
			lastModifiedTime: 0;
			topic: 'Empty Poster';
			nComments: (self stickers) size;
		yourself! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 8/23/2009 17:33'!
attachmentOffset
	"For z-buffer relief"
	^ 0.075! !

!QPoster methodsFor: 'transform' stamp: 'bgf 4/26/2009 14:08'!
cameraStanceFor: anAvatar

	^ QGotoButton frame: back cameraStanceFor: anAvatar! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 4/29/2009 17:12'!
card
	"My image file card, or nil if I have nonesuch"
	^ card! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 4/29/2009 21:36'!
card: imgFileCard
	"This is the bottleneck accessor by which the poster content is set;
	 the card is a QFileCard of type image."

	| txtr |
	card = imgFileCard ifFalse: [
		card ifNotNil: [card unregisterContent; destroy ].
	].
	card := imgFileCard.
	card ifNotNil: [ 
		self root ifNotNil: [ card registerContent ].
		txtr := card fileTexture. ].

	picture material texture: txtr.
	picture extent: (self pictureSize) 
		    	aspect: (txtr 
					ifNotNil: [ txtr originalAspect] 
					ifNil: [ self extent y / self extent x ]).! !

!QPoster methodsFor: 'render' stamp: 'bgf 9/5/2008 13:54'!
child: c setDistance: ogl
	"Treat children as equidistant so we can control subdistance.
	 The resize grip needs a negative subdistance so it is rendered as if closer than the bottom strip."
	frameParent ifNotNil: [
		^ (frameParent child: self setDistance: ogl) - Float epsilon.
	].
	^ super child: c setDistance: ogl
! !

!QPoster methodsFor: 'render' stamp: 'bgf 9/5/2008 13:54'!
childSubdistance: c
	"The resize grip needs a negative subdistance so it is visible at all angles."
	^ c == (resize grip) ifTrue: [ -0.1 ] ifFalse: [ 0 ]
! !

!QPoster methodsFor: 'persist' stamp: 'bgf 4/30/2009 12:34'!
constructFrom:pStream
	| txtr |
	super constructFrom:pStream.

	"We write slightly fat persistence to keep things loadable from pre-image-document versions.
	 To ensure that pStream sees the texture object,
	 we read it first, even if there is also a card entry that refers to it.
	 (If we overlook reading the contents, then the persist-reader 
	 never sees the original definition of that object-tag and returns an unitialized tft.)"

	txtr := pStream readObject: 'contents' default: self class defaultContents.
	(pStream readObject: 'card' default: nil) ifNotNilDo: [ : cd | 
		self card: cd
	] ifNil: [
		self contents: txtr.
	].
	self margin: (pStream readNumber: 'margin' default: self class defaultMargin).
	self extent: (pStream readPoint: 'extent' default: self class defaultExtent).

! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 07:46'!
consumeDrop: aFileCard avatar: anAvatar selection: aTXSel

	| tmp |

	aFileCard fileTexture = self contents ifTrue: [ ^ false ].

	tmp := QFileCard new copyFrom: aFileCard.
	(tmp transferToCurrentForum: anAvatar) ifTrue: [
		self startScript: #contentsFromCard:fromUser:
			withArguments: {tmp . anAvatar} 
			when: {tmp . #transferComplete}.
	] ifFalse: [
		self contentsFromCard: tmp fromUser: anAvatar
	].
	
	^ true! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 9/3/2007 13:02'!
contents
	^ picture material texture ! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 8/11/2009 14:27'!
contents: aTextureOrImgDoc
	(aTextureOrImgDoc isKindOf: QImageDocument) ifTrue: [
		self card: (QFileCard forDocument: aTextureOrImgDoc).
	] ifFalse: [
		self card: (QFileCard forTexture: aTextureOrImgDoc).
	].

	self signal: #contentsChanged.! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:53'!
contentsFromCard: aFileCard
	(aFileCard kind = #image) ifTrue: [
		self contents: aFileCard content
	].
! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 3/13/2010 08:53'!
contentsFromCard: aFileCard fromUser: avatar
	"Drop-copy support, called when a dropped-in image has completed any needed transfers,	
	 (i.e. after the document's shared name has been corrected for inventory/forum transfers.)"
	
	| old new |
	old := self stateVector.
	self contentsFromCard: aFileCard.
	new := self stateVector.

	(old hasEqualElements: new) ifFalse: [	
		avatar recordAction: #frame:oldState:newState:description:
			class: QStateVectorAction
			arguments: {  self . old. new .  'Drop Image' }.
	].! !

!QPoster methodsFor: 'menu' stamp: 'bgf 4/3/2010 09:23'!
convertToPanel: event
	"See also QFixedDisplay, which has the same method."
	| panel prePair |

	prePair := { self objectID . (event avatar captureFrames: { self }) }.

	panel := QDisplayPanel withSimpleFrame.
	panel localTransform: (self globalTransform);
		extent: (self extent + (0.1@0.25));
		isPlacementLocked: false;
		isFromTemplate: false.

	panel translation: (panel translation + (self lookAt * 0.05)).
	panel displayCard: (card) fromUser: event avatar.
	self root 
		addChild: panel;
		attach: panel to: (self surfaceAttachment).

	panel future recordConversion: prePair forAvatar: event avatar.
	self obliterateLater.! !

!QPoster methodsFor: 'menu' stamp: 'bgf 3/8/2008 22:28'!
copyToInventory: event

	self contents ifNotNilDo: [ : cts |
		event avatar ifNotNilDo: [ :av | av addContentToInventory: cts ]
	]! !

!QPoster methodsFor: 'transform' stamp: 'bgf 4/26/2009 14:59'!
destinationExtent
	"For destination distance-finding."
	^ back destinationExtent! !

!QPoster methodsFor: 'transform' stamp: 'bgf 4/26/2009 10:43'!
destinationStanceFor: anAvatar

	"The stance to apply to the avatar to bring it to me."
	^ QGotoButton groundedStanceFor: anAvatar toView: back inDestination: self.

! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:50'!
displayCard: aFileCard fromUser: anAvatar
	| tmp |
	tmp := QFileCard new copyFrom: aFileCard.
	(tmp transferToCurrentForum: anAvatar) ifTrue: [
		self startScript: #contentsFromCard:
			withArguments: {tmp } 
			when: {tmp . #transferComplete}.
	] ifFalse: [
		self contentsFromCard: tmp
	].
	^ true! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 4/30/2009 09:49'!
document
	"My ImageDocument"
	^ card ifNotNil: [ card content ]! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 9/9/2007 21:34'!
dropOn: aTarget avatar: avatar selection: aTXSel

	"XXX -draft code, inactive until #initialize is changed to make my adapter Droppable."

	"Special case... if I am dragged into a display sheet or panel,
	 I destroy myself - the panel takes up the picture."
	((aTarget isKindOf: QDisplayPanel) or: [aTarget isKindOf: QDisplaySheet]) ifTrue: [
		aTarget consumeDrop: (self payload) avatar: avatar selection: aTXSel.
		self future destroy.
		^ true
	].

	"Don't allow auto-open of displays on walls and floors... posters must be just movables to them."
	((aTarget isKindOf: QFixedSurface) or: [ aTarget isKindOf: QFloor]) ifFalse: [

		^ aTarget consumeDrop: (self payload) avatar: avatar selection: aTXSel.
	].
		
	^ true	"Don't snap back".! !

!QPoster methodsFor: 'updating' stamp: 'bgf 8/1/2009 12:59'!
editByUser: aUser
	"Used by the topics tool... so give a response. "
	aUser goToPlace: self asPlace! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 9/9/2007 21:00'!
endDrop: event
	"Nothing to do here."! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 9/3/2007 13:49'!
extent
	^ back extent! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 6/11/2009 07:36'!
extent: xy

	"Idempotence check is essential for cooperating with the resize handler."
	(xy ~= self extent) ifTrue: [
		back extent: xy.
		self updateShape.	"I.e. in the foreground, so shapeToFit works immediately."
		back translation: (Vector3 x: 0 y: 0 z: ((-0.005*(self extent maxDimension)) min: -0.005)).
	].

	gotoButton translationX: 0 y: ((xy y + gotoButton extent y)*0.5) + 0.5 z: 0.! !

!QPoster methodsFor: 'menu' stamp: 'bgf 11/20/2009 23:25'!
fillContextMenu: aMenu forEvent: event
	"The menu commands for a poster; 
	 available only if you have modify-access."

	event avatar canModifyForum ifFalse: [ ^ self ].
	self isPlacementLocked ifTrue: [ ^ self ].
	self contents ifNotNil: [
		aMenu addItemSpec: {'Shape to Fit' . #shapeToFit } target: self.
		aMenu addSeparator.
		aMenu addItemSpec: {'Convert to Panel' . #convertToPanel: } target: self.
	].

	! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 9/9/2007 14:37'!
height
	^ self extent y! !

!QPoster methodsFor: 'updating' stamp: 'bgf 9/3/2007 12:59'!
hilite
	self hilite: self class hiliteColor! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 1/26/2008 13:42'!
hilite: x
	"Don't bother... messes the goto button"! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 4/29/2009 17:52'!
holdsDocument: aDocument
	^ card notNil and: [  card content = aDocument ]! !

!QPoster methodsFor: 'initialize' stamp: 'bgf 11/8/2009 08:55'!
initialize
	super initialize.
	margin := self class defaultMargin.
	adapter overColor: ((Color gray:0.4) alpha: 0.1).
	adapter downColor: ((Color gray: 0.45) alpha: 0.3).
	false ifTrue: [ adapter isDroppable: true ]. 	"Experimental."

	picture := (TSimpleRectangle new)
		material: (QMaterials fullWhiteAlphaMaterial);
		extent: (self pictureSizeFor: (self class defaultExtent)).
	picture material fullBright: true.	"Immune to space lighting."
	self addChild: picture.

	back := TSimpleRectangle new.
	back material: (self class defaultBacking);
		extent: self class defaultExtent;
		translation:(0@0@-0.01);
		cullFace: false;
		yourself.
	self addChild: back.

	gotoButton := QGotoButton new.
	gotoButton destination: self.
	gotoButton label: 'Approach the Poster'.
	self addChild: gotoButton.
	gotoButton disable.

	self solid: false.
	self contents: (self class defaultContents).

	resize := QCorner2DResizer forLuggage: self.
	resize transient: true.

	self extent: self class defaultExtent. 	"Ensure margins and resize border are all in harmony."
	self setPlacementLocked: (self isPlacementLocked).		"Configure the resizer."! !

!QPoster methodsFor: 'drag-drop' stamp: 'bgf 9/9/2007 14:30'!
isDropTarget: payload

	^ self class canAcceptDrop: payload! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 8/1/2009 15:43'!
isListedTopic
	"We list posters in the topics card as one of the items you'd use in a stack."
	^ true! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 12/18/2008 12:43'!
labelForBrowser

	self contents ifNotNil: [
		self contents fileName ifNotNil: [ : uri |
			^ 'Poster - ', (QStyles simpleFileName: uri)
		]].

	^ super labelForBrowser
! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 6/22/2008 02:33'!
margin
	"Margin is a fraction (e.g. 0.09 is default) applied to the overall size."
	^ margin! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 9/9/2009 12:09'!
margin: aScalar
	"Margin is a fraction (e.g. 0.09 is default) applied to the overall size."
	margin := aScalar min: 0.4999 max: 0.0.
	self shapeChanged.! !

!QPoster methodsFor: 'menu' stamp: 'bgf 9/14/2007 20:00'!
menuTitle
	^  'Poster'! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 9/5/2008 14:32'!
minExtent
	"Corner-resize protocol - stay out of trouble."
	^ 0.4@0.4! !

!QPoster methodsFor: 'initialize' stamp: 'bgf 3/5/2010 18:28'!
objectID: oid
	super objectID: oid.
	gotoButton ifNotNil: [ gotoButton destination: self ].! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 4/29/2009 19:02'!
payload
	"The payload I would offer as a draggable,
 	 so it is possible to drag posters into rolodexes, etc."
	^ card! !

!QPoster methodsFor: 'persist' stamp: 'bgf 4/30/2009 12:32'!
persistOn:pStream
	super persistOn:pStream.
	"We store both ways... contents, 
	 the texture itself, to ensure readability from pre-QImageDocument (e.g. 2.0.10),
	 and the file card holding the image document, for updated saves."

	pStream storeObject: 'contents' value: self contents default: (self class defaultContents).
	pStream storeObject: 'card' value: (self card) default: nil.
	pStream storePoint: 'extent' value: self extent default: self class defaultExtent.	
	pStream storeNumber: 'margin' value: self margin default: self class defaultMargin.! !

!QPoster methodsFor: 'private' stamp: 'bgf 9/3/2007 13:59'!
pictureSize
	"the rectangle for the picture in my current overall extent"
	^ self pictureSizeFor: self extent! !

!QPoster methodsFor: 'private' stamp: 'bgf 6/22/2008 02:40'!
pictureSizeFor: anExtent

	"The rectangle for the picture, inset from the  overall extent using the margin."
	^ anExtent - (anExtent maxDimension * margin).
! !

!QPoster methodsFor: 'events' stamp: 'bgf 9/5/2008 14:09'!
pointerDown: event
	super pointerDown: event.
	self isPlacementLocked ifFalse: [ resize grip visible: true ].! !

!QPoster methodsFor: 'events' stamp: 'bgf 9/5/2008 14:01'!
pointerEnter: event

	self pointerOver: event.
	self isPlacementLocked ifFalse: [ resize grip visible: true ].! !

!QPoster methodsFor: 'events' stamp: 'bgf 9/5/2008 14:22'!
pointerLeave: event
	super pointerLeave: event.
	resize grip visible: false! !

!QPoster methodsFor: 'events' stamp: 'bgf 11/13/2008 12:54'!
pointerOver: event
	| fname |
	event avatar ifNil: [ ^ self ].
	self pickable ifFalse: [ ^ self ].
	self contents ifNil: [ ^ self ].

	(self contents = self class defaultContents) ifTrue: [
		event avatar showFloatMessage: (self floatingMessageFor: event avatar) forEvent: event.
	] ifFalse: [
		fname := (self contents fileName) ifNil: [ 'Source Unknown'].
		(fname beginsWith: 'resources') ifTrue: [
			fname := 'From Imported Model'.
		].
		(fname includes: $/) ifTrue: [
			"Strip the 'shared documents' part from the standard doc filename"
			fname := fname copyAfter: $/
		].
		event avatar signal: #infoMessage with: ('Image: ', fname)
	].! !

!QPoster methodsFor: 'transform' stamp: 'bgf 8/15/2009 18:41'!
positioningBox
	"For materialization, we want to see the display even if we're at exact wall-distance 
	 for it... to be sure, use only the inner bounds to figure distances so we err too close.
	 Helps prevent panel-behind-wall problems."
	| box |
	box := back boundingBox.	
	box min: (box min x - 0.2) @ (box min y -0.1) @ (-0.1).
	box max: (box max x + 0.2) @ (box max y) + (0.1)  @ 0.2.
	^ box! !

!QPoster methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:57'!
privateScale: aFactor

	"Keep my layout consistent."
	nomScale := aFactor * nomScale.
	self extent: (self extent * aFactor).! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 4/30/2009 12:31'!
root: aSpace
	"We have to manage the registration of the card document resource,
	 since we don't actually add the card to the scene."
	card ifNotNil: [
		(aSpace ~= self root) ifTrue: [ 
			aSpace ifNil: [ card unregisterContent ]
					ifNotNil: [ card registerContent ] 
		]].

	super root: aSpace.! !

!QPoster methodsFor: 'updating' stamp: 'bgf 8/10/2009 01:06'!
setPlacementLocked: aBool
	super setPlacementLocked: aBool.
	"The resizer has to be suppressed when locked or non-pickable."
	resize ifNotNil: [ resize isEnabled: (adapter isEnabled)].! !

!QPoster methodsFor: 'menu' stamp: 'bgf 12/25/2008 00:30'!
shapeToFit

	| m w |
	picture ifNotNil: [
		w := picture extent x.
		m := (picture extent maxDimension * margin) * (1 + margin).
								"Since margin is subtracted from total extent later."
		self extent: (w @ (w * picture originalAspect)) + m.
	].
! !

!QPoster methodsFor: 'private' stamp: 'bgf 3/12/2010 07:15'!
stateVector
	"Capture the state of my image-card, for undo-redo "
	| pickle |
	card ifNotNil: [
		pickle := (TSceneWriter new)
					preserveObjectIDs: true;
					storeScene: card
	].	
	^ { pickle }! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 1/26/2008 13:42'!
unhilite: x
	"Don't bother."! !

!QPoster methodsFor: 'hierarchy' stamp: 'bgf 6/11/2009 07:54'!
updateShape

	| xy |
	xy := self extent.
	self contents ifNotNil: [
		picture extent: (self pictureSizeFor: xy) aspect: (self contents originalAspect).
	] ifNil: [
		picture extent: (self pictureSizeFor: xy) aspect: (xy y / xy x).
	].
	margin > Float epsilon ifTrue: [
		back visible: true.
	] ifFalse: [
		"When there is no margin, we have to wrap the back to the picture
	       so the overall extent is correct for corner-resizer placement."
		back visible: false; extent: picture extent.
	].
	resize updateShape.! !

!QPoster methodsFor: 'accessing' stamp: 'bgf 9/9/2007 14:37'!
width
	^ self extent x! !

QAnnotation subclass: #QStickyNote
	instanceVariableNames: 'label back backingColor sizer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QStickyNote commentStamp: '<historical>' prior: 0!
A note is a sticky annotation, like a Post-It, that displays as
a small rectangle.   Editing is done in a floating editor.
The paper color can be specified.  In normal use,
the paper color is tied to the creator's avatar color.

The note orients itself to appear 
flat against the surface where it is attached.

When attached to a frame, it arranges to follow that frame
by maintaining its transform relative to the frame's transform;
it does not add itself to the frame, since that is problematic for persistence.!

QStickyNote subclass: #QDisplaySign
	instanceVariableNames: 'destination customImage setback'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QDisplaySign commentStamp: '<historical>' prior: 0!
Like a sticky note, but configured larger and horizontal;
advertises as a place in 'Signs'  in the places list.

The destination is ordinarily a setback from the sign itself,
but can be another destination frame.!

!QDisplaySign class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDisplaySign class methodsFor: 'inspection' stamp: 'sn 3/10/2013 21:44'!
addPropertyRows: propertyDialog forFrame: signRef

	"No super - signs have a custom sheet."
	propertyDialog addRow: (
		(QTextPropertyRow new)
			titleString: 'Wording'translateMe;
			fetchBlock: [ :f | (f get: #contents) ifNil: [''] ];
			acceptBlock: [ :v : f| f future contents: (v ifNil: [''] ifNotNil: [v]) ];
			balloonHelp: 'Display Text in the Sign.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QTFramePropertyRow new)
			titleString: 'Goes To'translateMe;
			nameSelector: #placeName;
			fetchBlock: [ :f | ( f get: #destinationOrNil) ];
			validateBlock: [ : frame : sign : row |
				(frame isValidFarRef and: [ (frame get:#asPlace) isNil]) ifTrue: [
					self inform: ('The go-to destination frame\\	', 
								(frame get: #labelForBrowser), 
								'\\does not define a named place.') withCRs.
					false
				] ifFalse: [
					true
				]];
			acceptBlock: [: v :f | f future destination: v ];
			balloonHelp: 'Alternate destination.  You can drag a Place here from the Inspector.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [:f | f get: #extent ];
			acceptBlock: [ : ext :f | f future extent: ext ];
			balloonHelp: 'Dimensions of the rectangle'translateMe;
		yourself).

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Setback for Arrival'translateMe;
			fetchBlock: [:f | f get: #setback ];
			validateBlock: [ :v | 
					v > 0
						ifFalse: [ self inform: 'Setback must be positive.'translateMe. false ]
						ifTrue: [ true ]];
			acceptBlock: [ : ext :f | f future setback: ext ];
			balloonHelp: 'A factor applied to the sign size to decide where to put arriving avatars.'translateMe;
		yourself).


	propertyDialog addRow: (
		(QBooleanPropertyRow new)
			titleString: 'Display as Custom Image'translateMe;
			fetchBlock: [ :f :row | |v| v := (row get: #useCustomImage) ifNil: [ false ].
							    row inputEnabled: v.
							    v ];
			acceptBlock: [ : v :f | f future useCustomImage: v ];
			balloonHelp: 'If a custom image has been installed, uncheck to discard it.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Position (in 'translateMe, ((signRef get:#parentForInspector) get:#genericName), ')';
			fetchBlock: [ :f | f get:#positionForInspector ];
			acceptBlock: [ :v3 :f | f future positionFromInspector: v3 ];
			balloonHelp: 'Center of the object relative to its container.'translateMe;
			inputEnabled: false;
	 	yourself	).

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Orientation (Pitch/Yaw/Roll)'translateMe;
			fetchBlock: [ :f | f get:#pyrForInspector ];
			acceptBlock: [ :v3 :f | f future pyrFromInspector: v3 ];
			balloonHelp: 'Rotation of the sign relative to its container.'translateMe;
			inputEnabled: false;
		yourself).

	propertyDialog addRow: (
		(QFrameHintsPropertyRow newForFrame: signRef)).
	propertyDialog addRow: (
		self templatePropertyRow: signRef).


! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 8/11/2009 22:01'!
defaultBackground
	^ Color r: 0.96 g: 0.94 b: 0.88
	! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:14'!
defaultBackingColor
	"The color for the underlay backing material.
	 This is where the outline around the sign comes from."

	^ QColors displaySignColor! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 8/27/2007 13:58'!
defaultContents
	^'<Sign>'! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 8/27/2007 14:10'!
defaultExtent
	"Default size for a display sign."
	^ 5@1.5! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:12'!
defaultIsFromTemplate
	"Reduces xml volume since these are often premade."
	^true ! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 7/25/2009 15:47'!
defaultMargin
	^ 8! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 12/18/2008 15:23'!
defaultSetback
	"The default multiple of the sign's size that we stand back from the sign asPlace "
	^ 3.0
	! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 8/27/2007 14:17'!
defaultStyleName
	^ #SignStyle! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 8/11/2009 21:57'!
defaultTextColor
	^ Color r:  0.05 g:0.03 b:0.02! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 8/27/2007 14:26'!
editorTitle
	^ 'Display Sign Text'! !

!QDisplaySign class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 09:35'!
hiliteColor
	^VectorColor r: 0.6 g:0.61 b:0.57 a: 0.3! !

!QDisplaySign methodsFor: 'updating' stamp: 'bgf 3/19/2010 11:31'!
applyStateVector: pickle
	"See #stateVector - this reapplies a captured state."	

	pickle ifNotNil: [
		self installCustomImage: (TSceneReader new readScene: pickle )
	] ifNil: [
		self discardCustomImage.
	].
! !

!QDisplaySign methodsFor: 'transform' stamp: 'bgf 12/17/2008 17:12'!
asPlace
	"The QPlace descriptor for this sign as a destination."
	
	"Signs that are go-to shortcuts are not places in themselves,
	 since any clicking on them should take the user elsewhere.
	 (We don't want warp-placement to bring us to this sign, for instance.)"

	self destination = self ifFalse: [ ^ nil ].

	^	QPlace 
			category:  'Signs' 
		  	label:  self contents 
		  	frame: self
		 	sticky:  false ! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 2/3/2010 10:41'!
asTopicRecord
	^ (QTopicRecord new)
			cardRef: self;
			author: '';
			color: (self class defaultTextColor);
			creationTime: 0;
			lastModifiedTime: 0;
			topic: 'Sign: ', (self contents);
			nComments: (self stickers) size;
		yourself! !

!QDisplaySign methodsFor: 'persist' stamp: 'bgf 12/18/2008 15:39'!
constructFrom: pStream

	super constructFrom: pStream.
	self destination: (pStream readObject: 'destination' default: self ).
	self setback: (pStream readNumber: 'setback' default: self class defaultSetback).
	(pStream readObject: 'customImage' default: nil) ifNotNil: [ : txtr |
		self installCustomImage: txtr
	].! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 3/11/2010 10:45'!
consumeDrop: payload avatar: avatar selection: aTXSel
	| tmp |

	"Accept drops of image-file-cards,
	 for the sole purpose of becoming a custom-image sign."

	(self isDropTarget: payload) ifFalse: [ ^ false ].

	"Now we the card is an #image card whose content is a texture."

	"Usual rigamorole - we're creating a dependency on this resource..."
	tmp := QFileCard new copyFrom: payload.
	(tmp transferToCurrentForum: avatar) ifTrue: [
		self startScript: #consumeTransferredCard:fromUser:
			withArguments: {tmp . avatar} 
			when: {tmp . #transferComplete}.
		^ true.
	].
	self installCustomImage: tmp fileTexture fromUser:  avatar.
	^ false! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 3/11/2010 10:46'!
consumeTransferredCard: aQFileCard fromUser:  anAvatar
	"Post data-copy callback... install the texture referred to by the card,
	 which may have been pulled into the forum from inventory."
	(self isDropTarget: aQFileCard) ifFalse: [ ^ false ].
	self installCustomImage: aQFileCard fileTexture fromUser: anAvatar.
	^ false! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 7/17/2008 12:54'!
creator: anAvatar
	"Do nothing - display signs do not remember their creator,
	 or adapt their coloration."! !

!QDisplaySign methodsFor: 'updating' stamp: 'bgf 1/10/2010 12:31'!
deletionLabel
	^ 'Sign'
! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 12/17/2008 16:39'!
destination
	"If I have been given a separate frame as a destination, use that, else I'm my own"

	destination ifNotNil: [
		"This prevents persisting or using destinations that have been deleted."
		destination root ifNil: [ destination := nil ].
	].

	^ destination ifNil: [ self ]! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 7/31/2009 01:44'!
destination: aFrame
	"If I have been given a separate frame as a destination, use that, else I'm my own"

	aFrame = destination ifTrue: [ ^ self ].
	aFrame ifNotNil: [
		destination :=  (self isOrContains: aFrame) ifTrue: [ nil ] ifFalse: [ aFrame ].
	] ifNil: [
		destination := nil.
	].
	self signal: #frameChanged with: self.	 
		! !

!QDisplaySign methodsFor: 'transform' stamp: 'bgf 7/27/2009 00:08'!
destinationExtent
	^ self extent ! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 12/17/2008 15:49'!
destinationOrNil
	"If I have been given a separate frame as a destination, return it."
	^ destination! !

!QDisplaySign methodsFor: 'transform' stamp: 'bgf 7/27/2009 00:09'!
destinationStanceFor: anAvatar
	| stance |
	stance :=  QGotoButton 
				stanceFor: anAvatar
				toView: self 
				inDestination: self	
				inAngle: 30
				atDistance: (self extent maxDimension * (self setback)).

	"We go to signs to be in an area, not to see the sign; 
	 so fix the result to keep the avatar afoot, and looking levelly ahead."
	stance yHere: (self root groundHeightAt: (stance here)) + (anAvatar footToEyes).
	stance yThere: (stance yHere).
	^ stance! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 8/10/2009 01:08'!
discardCustomImage
	"If I have a customImage installed, chuck it and go back to normal label/backing display."
	customImage ifNotNil: [
		self removeChild: customImage.
		(customImage future: 60000) obliterate.
		customImage := nil.
	].
	self addChild: label.
	self addChild: back.
	sizer disable.
	self isPlacementLocked ifFalse: [ sizer enable ].! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 9/5/2008 15:58'!
extent
	^ back extent! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 8/23/2009 17:36'!
extent: xy
	"Fixed narrow margin."
	back extent: xy.
	label extent: xy - (xy maxDimension * 0.025).
	label translation: (Vector3 x: 0 y: 0 z: ((0.005*(xy maxDimension)) min: 0.005)).
	customImage ifNotNil: [ 
		customImage extent: back extent 
					  aspect: (customImage originalAspect).
		back extent: customImage extent.
	].
	sizer updateShape.! !

!QDisplaySign methodsFor: 'events' stamp: 'bgf 3/2/2010 11:09'!
floatingMessageFor: anAvatar

	"Override the inherited fancy sticky-note message."
	^ 'Double-click sign to edit.'! !

!QDisplaySign methodsFor: 'initialize' stamp: 'bgf 8/11/2009 22:09'!
initialize
	super initialize.
	adapter attractionRange: 25.0.		"Less than a sticky note, as these are bigger."
	adapter	isGroundBound: false.	
	adapter overColor: (self class hiliteColor).
	adapter downColor: ((self class hiliteColor * 0.9) alpha: 1).

	setback := self class defaultSetback.
	sizer grip material opacity: 0.5.
	self extent: self class defaultExtent. 	"Ensure margins, and resize border, are all in harmony."! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 12/18/2008 11:49'!
installCustomImage: aTexture

	aTexture ifNil: [ ^ self discardCustomImage ].
	(customImage notNil and: [ aTexture = customImage material texture ])
	 ifTrue: [ ^ self ].
	
	self discardCustomImage.
	self removeChild: label.
	self removeChild: back.

	self addChild: (customImage := (aTexture asTFrame)).
	(customImage material)
		diffuseColor:  (Color white alpha: 0.99) asVectorColor;
		ambientColor: (Color gray: 0.8) asVectorColor;
		emissiveColor: (Color gray: 0.2) asVectorColor.

	customImage
		aspectFixed: true;
		checkAlpha.
	self extent: (self extent).
! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 7/15/2010 14:16'!
installCustomImage: aTexture fromUser: anAvatar
	| old new |
	
	old := self stateVector.
	self installCustomImage: aTexture.
	new := self stateVector.

	(old notNil and: [ old hasEqualElements: new ]) ifFalse: [ 
		anAvatar
			recordAction: #frame:oldState:newState:description:
			class: QStateVectorAction
			arguments: {  self . old. new .  'Drop Image' }.
	].
		

! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 4/30/2009 13:23'!
isDropTarget: aPayload

	"Accept drops of image file-cards,
	 for the sole purpose of becoming a custom-image sign."

	^ (aPayload isKindOf: QFileCard) 
		and: [ aPayload kind = #image ] 
		and: [ aPayload fileTexture notNil ]! !

!QDisplaySign methodsFor: 'transform' stamp: 'bgf 12/17/2008 15:50'!
isListedLandmark
	"Signs always go into the places list once configured;
	 unless I'm a go-to-somewhere-else sign."
	^ (self contents ~= self class defaultContents) and: [ destination isNil ]
! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 8/18/2009 00:06'!
isListedTopic
	"Yes, for completeness."
	^ true! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 12/20/2008 16:48'!
labelForBrowser
	self asPlace ifNotNilDo: [ : place | ^ 'Sign: ', (place label) ].
	"Some signs are not places..."
	^ 'Sign: ', (self contents)! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 8/15/2008 17:10'!
labelWrapForContents: aString
	"No autowrap in signs - fully manual"
	^ nil! !

!QDisplaySign methodsFor: 'persist' stamp: 'bgf 12/18/2008 15:36'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeObject: 'destination' value: (self destination) default: self.
	pStream storeNumber: 'setback' value: (self setback) default: (self class defaultSetback).
	customImage ifNotNil: [
		pStream storeObject: 'customImage' value: (customImage material texture) default: nil 
	].! !

!QDisplaySign methodsFor: 'events' stamp: 'bgf 6/1/2009 12:18'!
pointerClick: event
	"Clicking on a sign...
	 - If it is a sign-post (points to other destination) go there.
	 - If it is not locked, edit the wording. "
	self destination ~= self ifTrue: [
		event avatar takeMeToFrame: self destination.
		^self
	].
	self isPlacementLocked ifFalse: [
		super pointerClick: event
	] ifTrue: [
		event avatar takeMeToFrame: self destination
	].! !

!QDisplaySign methodsFor: 'events' stamp: 'bgf 1/28/2008 21:36'!
pointerEnter: event
	self pointerOver: event! !

!QDisplaySign methodsFor: 'events' stamp: 'bgf 11/7/2009 18:25'!
pointerOver: event
	event avatar ifNil: [ ^ self ].
	self pickable ifFalse: [ ^ self ].

	"If a signpost, do feedback..."
	(destination notNil or: [self isPlacementLocked]) ifTrue: [
		event avatar signal: #overWarpPlace with: (self destination asPlace).
	] ifFalse: [
		"If we're not yet filled-in, encourage filling-in."
		(self contents isNil or: [self contents isEmpty] or: [ self contents first = $<]) ifTrue: [
			event avatar floatingInfo: (self floatingMessageFor: event avatar).
		]
	].! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 12/18/2008 15:39'!
setback
	"Scalar - multiple applied to the sign maxDimension to compute the approach distance."
	^ setback ifNil: [ self class defaultSetback ].! !

!QDisplaySign methodsFor: 'accessing' stamp: 'bgf 12/18/2008 15:39'!
setback: aScalar
	"Scalar - multiple applied to the sign maxDimension to compute the approach distance."
	setback := aScalar max: 0.! !

!QDisplaySign methodsFor: 'updating' stamp: 'bgf 3/12/2010 07:19'!
stateVector
	"Undo support"
	^ customImage ifNotNil: [
		(TSceneWriter new) storeScene: (customImage material texture)
		].
	! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 12/18/2008 13:15'!
useCustomImage
	"Psuedo-property for the inspector;
	 appears as true when we have a custom image."
	^ customImage notNil
! !

!QDisplaySign methodsFor: 'custom-image' stamp: 'bgf 12/18/2008 13:15'!
useCustomImage: aBool
	"Pseudo-property for the inspector.
	 Attempting to set to true has no effect;
	 if set to false, will discard any custom image in use."
	aBool ifFalse: [
		self discardCustomImage.
	].! !

!QStickyNote class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStickyNote class methodsFor: 'inspector' stamp: 'sn 3/9/2013 14:50'!
addPropertyRows: propertyDialog forFrame: noteref

	| cRow aRow |

	"No Super - for sticky notes we override the property sheet."

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :frame | frame get: #extent ];
			acceptBlock: [ : ext :frame | frame future extent: ext ];
			balloonHelp: 'Dimensions of the rectangle'translateMe;
		yourself).

	propertyDialog addRow: (
		cRow := (QChoicesPropertyRow new)
			titleString: 'Color'translateMe;
			setChoices: (self colorChoices);
			fetchBlock: [ :frame |(frame get: #backingColor) ];
			acceptBlock: [ : c : frame | frame future backingColor: c ];
			balloonHelp: 'The background color of the item.\(Automatic if the author is recorded.)'translateMe withCRs;
			comparisonBlock: [ :v1 : v2 | QColors sameColor: v1 as: v2 ];
			inputEnabled: (propertyDialog targets allSatisfy: [ : ref |  (ref get: #creatorLogin) isNil ]);
		yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Last Modified'translateMe;
			fetchBlock: [  :frame |(frame get: #lastModifiedTime) 
							ifNil: [''] 
							ifNotNilDo: [ :seconds| QStyles dateStringFromUTC:  seconds ] ];
			balloonHelp: 'When last edited, if known.'translateMe;
			inputEnabled: false;
		yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Author'translateMe;
			fetchBlock: [  :frame |(frame get: #creatorLogin) ifNil: [''] ];
			balloonHelp: 'The original author, if known.'translateMe;
			inputEnabled: false;
		yourself).

	propertyDialog addRow: (
		aRow := (QBooleanPropertyRow new)	
			titleString: 'Attributed to Original Author'translateMe;
			fetchBlock: [  :frame | (frame get: #creatorLogin) notNil ];
			acceptBlock: [ : v :frame | v ifFalse: [ frame future creator: nil ]];
			balloonHelp: 'When checked, the author is recorded, and only that person can edit.\Cannot be restored if unchecked.'translateMe withCRs;
			inputEnabled: (propertyDialog targets anySatisfy: [ : ref | (ref get: #creatorLogin) notNil ]);
		yourself).

	propertyDialog addRow: (
		(QFrameHintsPropertyRow newForFrame: noteref)).
	propertyDialog addRow: (
		(self templatePropertyRow: noteref)).

	"Keep consistency about what is editable..."		
	propertyDialog  startScript: [
			| notes | 
			notes := propertyDialog targets.
			notes ifNotNil: [ 
				cRow inputEnabled: (notes allSatisfy: [ : ref |  (ref get: #creatorLogin) isNil ]).
				aRow inputEnabled: (notes anySatisfy: [ : ref | (ref get: #creatorLogin) notNil ]).
			]]
		when: { propertyDialog . #accepted }.
		
		! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 7/18/2008 04:01'!
colorChoices
	
	"A dictionary of value names->choices for the paper color.
	 The names are the avatar-color names, plus 'Default' and 'Paper';
	 the values are VectorColors ... the paper-pastel variants of avatar colors."

	| result |
	result := Dictionary new.

	QUserPreferences colorTable keysAndValuesDo: [ : key : vecColor |
		key = '(Random)' ifFalse: [
			result at: key put: (QColors paperColorForAvatarColor: (vecColor)) asVectorColor
		]].
	result at: 'Default' put: nil.
	result at: 'Paper' put: (Color paleBuff) asVectorColor.

	^ result.
! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 22:04'!
defaultBackground
	"the default label background is transparent
	 so that the textured background label appears."
	^ Color transparent! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 10/19/2008 16:20'!
defaultBacking
	"The material with which to back the label."
	^ (QMaterials flatWhiteMaterial)
		texture: (QTextures find: 'Content/Qwaq/Textures/annotationGray.png');
		ambientColor: (self defaultBackingColor);
		diffuseColor: (self defaultBackingColor);
		cullFace:false;
		fullBright:true;
		yourself.! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:10'!
defaultBackingColor

	"The (post-processed, i.e. lightened) backing color to use
	  when no creatorColor is set.
	 Post-processed color so that persistOn: can detect equality."

	^ QColors stickyNoteColor! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 11/24/2009 19:56'!
defaultCanTakeAnnotations
	^ false! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 20:31'!
defaultContents
	^ '<Note>'! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 9/4/2008 00:35'!
defaultExtent
	"Default size for a post-it note.  Huge in world-scale, more of a sheet-o-paper.."
	^ 1.25@0.95! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 6/22/2008 22:09'!
defaultMargin
	^ 4! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 8/27/2007 14:17'!
defaultStyleName
	^ #NoteStyle! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 7/25/2009 15:27'!
defaultTextColor
	^ Color gray: 0.05	"Renders slightly better than black."! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 8/15/2008 16:54'!
defaultWrap
	"Pixel length of a line before it autowraps - relates to default style"
	^ 240! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 7/18/2008 10:44'!
editorTitle
	^ 'Note'! !

!QStickyNote class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 10:29'!
hiliteColor
	^VectorColor r: 0.75 g: 0.73 b: 0.61 a:0.25! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/28/2008 00:25'!
asPlace
	^ QPlace
		category: 'Notes'
		label: (self titleString)
		frame: self
		sticky: true! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/15/2009 22:35'!
asTopicRecord
	"The QTopicRecord describing me for topic review.
	  This must be side-effect free. 
	  QTopicRecord is a pass-by-copy simple data record."

	^ (QTopicRecord new)
			cardRef: self;
			author: (self creatorLogin);
			color: (self backingColor);
			creationTime: 0;
			lastModifiedTime: ((self  lastModifiedTime) ifNil: [ 0 ]);
			topic: ((self topicFirstLineMax: 34) ifNil: ['']);
			nComments: 0;
		yourself! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/23/2009 17:33'!
attachmentOffset
	"For z-buffer relief"
	^ 0.035! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 7/18/2008 01:48'!
backingColor
	"The 'paper color' of the note.
 	 Note that if the creatorLogin is set,
	 then the color will be updated to match the current appearance of that user,
	 if they are present in a forum session,
	 and the persisted color will be the last such color.
	 If creatorLogin is cleared to nil, that updating is disabled."

	^ backingColor! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 3/16/2010 11:56'!
backingColor: aColor
	"The backing color for the note...
	 often, the personal color of the person introducing the note.
	 Default is a yellow post-it style note.

	 To respect compact persistence for the default,
	 we record <nil> as-is, applying the default to the material in that case."

	aColor ifNotNil: [
		backingColor := aColor asVectorColor.
	] ifNil: [
		backingColor := nil.
	].
	back defaultMaterial 
		  ambientColor: (backingColor ifNil: [ self class defaultBackingColor ]);
		  diffuseColor: (backingColor ifNil: [ self class defaultBackingColor ]).! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 6/22/2008 22:10'!
backingScale
	"How much bigger"
	^ 1.1! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/23/2009 19:24'!
ballotColor
	^ ((self backingColor ifNil:[ self class defaultBackingColor ])
		asColor adjustSaturation: 0.2 brightness: -0.28) 
			asVectorColor
		! !

!QStickyNote methodsFor: 'testing' stamp: 'bgf 10/25/2009 14:51'!
canEditInBox
	^ false! !

!QStickyNote methodsFor: 'render' stamp: 'bgf 9/3/2008 23:41'!
child: c setDistance: ogl

	"Treat children as equidistant so we can control subdistance.
	 The resize grip needs a negative subdistance so it is rendered as if closer than the bottom strip."
	frameParent ifNotNil: [
		^ (frameParent child: self setDistance: ogl) - Float epsilon.
	].
	^ super child: c setDistance: ogl
! !

!QStickyNote methodsFor: 'render' stamp: 'bgf 9/4/2008 00:33'!
childSubdistance: c
	"The resize grip needs a negative subdistance so it is visible at all angles."
	^ c == (sizer grip) ifTrue: [ -0.1 ] ifFalse: [ 0 ]
! !

!QStickyNote methodsFor: 'persist' stamp: 'bgf 9/3/2008 23:55'!
constructFrom:pStream
	super constructFrom:pStream.
	self extent: (pStream readPoint: 'extent' default: self extent).
	self contents: (pStream readString: 'contents' default: self class defaultContents).
	self backingColor: (pStream readColor: 'backingColor' default: (self backingColor)).

	sizer updateShape.! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 2/10/2007 22:34'!
contents
	^ label contents ! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 6/30/2009 22:17'!
contents: aString
	"The string is the note. 
 	 Cap the contents at a large (for a note) but manageable length."
	| txt |

	txt := ((aString ifNil: ['']) truncateWithElipsisTo: 5000)  asText.
	txt = (label contents) ifFalse: [
		label wrap: (self labelWrapForContents: aString).
		label contents: txt.
		self signal: #contentsChanged.
	].! !

!QStickyNote methodsFor: 'events' stamp: 'bgf 2/9/2010 13:49'!
contextMenuMaskForAvatar: avatar
	"Remove the system 'Attach Note' command from the menu."
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	mask remove: #attachNote ifAbsent: [].
	mask add: #createPoll.
	^ mask! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 7/18/2008 02:05'!
creator: anAvatar

	"Have the note remember who made it.
	 Also set the backing color to the current paperColor for that avatar...
	 (And note that so long as the creatorLogin is remembered, 
	  subsequent avatar-preference color updates will be reflected in my color.)"

	| color |
	anAvatar ifNotNil: [
		color := anAvatar avatarColor.
		color ifNotNil: [
			self backingColor: (QColors paperColorForAvatarColor: color).
		] ifNil: [
			self backingColor: nil.
		].
		self creatorLogin: (anAvatar userLogin).
	] ifNil: [
		self creatorLogin ifNotNil: [
			self creatorLogin: nil.		"Decouple further updates."
			self backingColor: nil.		"And forget the old association."
		].
	].! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 8/30/2009 10:43'!
deletionLabel
	^ 'Note'
! !

!QStickyNote methodsFor: 'transform' stamp: 'bgf 1/10/2010 12:14'!
destinationExtent
	"Don't approach full-on; notes matter in context."
	^ (self extent * 8) min: 12@12 max: 4@2! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 7/27/2009 00:03'!
destinationStanceFor: anAvatar

	^ QGotoButton 
			stanceFor: anAvatar
			toView: self
			inDestination: self		! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 3/8/2010 11:15'!
editByUser: anAvatar

	"Actually 'view or edit by user...'.
	 We open the dialog in either case,  but only allow editing/updating for 
 	 those allowed to edit (usually the owner.)"

	self hiliteBriefly.

	QTextEditJob
		edit: (label contents asString)
		title: (self editorTitleFor: anAvatar)
		forUser: anAvatar
		recipient: self
		selector: ((self editableByUser:anAvatar)
					ifTrue: [ #textFromEditor:utcSeconds:avatar:]
					ifFalse: [ nil ] ).
! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 7/17/2008 19:06'!
editableByUser: anAvatar

	"A note is editable by its creator only.
 	 If the note has no remembered creator, it is editable by anyone."

	^ (creatorLogin isNil) 
			or: [ 
				(anAvatar notNil 
				 and: [ anAvatar userLogin = creatorLogin ]) ].
! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/28/2008 01:15'!
editorTitleFor: anAvatar

	"How to caption the editor?  Say who owns it, if we know."
	(anAvatar userLogin = creatorLogin) ifTrue: [
		^ (self class editorTitle), ' by You'
	].

	creatorLogin ifNotNil: [
		^ (self class editorTitle), ' by ', (self creatorName)
	].
	^ self class editorTitle. ! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 9/3/2008 23:58'!
extent
	^back extent
! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/23/2009 17:36'!
extent: xy
	"Subclasses may override."
	back extent: xy.
	label translation: (Vector3 x: 0 y: 0 z: ((0.003*(xy maxDimension)) min: 0.005)).
	label extent: xy * (self labelScale).
	sizer ifNotNil: [ sizer updateShape ].
		! !

!QStickyNote methodsFor: 'accessing' stamp: 'sn 3/10/2013 21:20'!
floatingMessageFor: anAvatar

	"Adjust the floating message according to whether I can edit or not,
	 whether there's a persisted creator and/or last-mod date, etc."

	| date tail msg front |
	self lastModifiedTime ifNotNil: [
		date := QStyles dateStringFromUTC: self lastModifiedTime.
	].

	tail := (self editableByUser: anAvatar)
			ifTrue: [ 'Double-click to edit.' translateMe ]
			ifFalse: [ 'Double-click to review.' translateMe].

	front := self creatorLogin ifNotNil: [
					(self creatorLogin = (anAvatar userLogin)) ifTrue: [
						'Note by You' 
					] ifFalse: [
						'Note by' , ' ', (self creatorName)
					].
				] ifNil: [
					'Note' .
				].

	msg :=	self creatorLogin ifNotNil: [
				date ifNotNil: [
					front, '\', date, '\', tail
				] ifNil: [
					front, '\', tail
				]
			] ifNil: [
				date ifNotNil: [
					front, '\', date, '\', tail
				] ifNil: [
					front, ' - ', tail
				]
			].

	^ msg
! !

!QStickyNote methodsFor: 'accessing' stamp: 'das 10/18/2009 15:59'!
getColor

	self backingColor.! !

!QStickyNote methodsFor: 'accessing' stamp: 'das 8/5/2009 13:28'!
getLog
	"For assembling text reports of note-collections."
	^ '<stickyNote>', String cr,
	String tab, '<title>', self titleString asString, '</title>',String cr,
	String tab, '<time>', (QStyles dateStringFromUTC: self lastModifiedTime), '</time>', String cr,
	String tab, '<message>', self contents asString, '</message>', String cr,
	'</stickyNote>'.
! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 2/11/2007 09:05'!
hilite
	self hilite: self class hiliteColor! !

!QStickyNote methodsFor: 'initialize' stamp: 'bgf 11/8/2009 10:48'!
initialize
	
	super initialize.

	"The avatar-coloring can make the default drag-adapter hilites hard to see."
	adapter overColor: (QColors whiteOverColor alpha: 0.15).

	label := (QLabel new)
		styleName:  self class defaultStyleName;
		backgroundColor: self class defaultBackground;
		textColor: self class defaultTextColor;
		margin: self class defaultMargin;
		extent: self class defaultExtent * (self labelScale).
	self addChild: label.

	back := TSimpleRectangle new.
	back material: (self class defaultBacking);
		extent: self class defaultExtent;
		translation:(0@0@-0.01);
		yourself.
	self addChild: back.

	sizer := QCorner2DResizer forLuggage: self.

	self solid: false.
	self contents: (self class defaultContents).

	sizer updateShape.! !

!QStickyNote methodsFor: 'testing' stamp: 'bgf 9/1/2008 19:40'!
isListedTopic
	^ true
	"Yes, this is a managed item visible in the topics tool"! !

!QStickyNote methodsFor: 'testing' stamp: 'bgf 8/23/2009 16:48'!
isPollBallot
	^ true! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/28/2008 01:16'!
labelForBrowser
	^ 'Note - ', (self creatorName ifNil: [ 'Unattributed' ]).! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 9/3/2008 23:57'!
labelScale
	"How much is the label inset"
	^ 0.925! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/16/2008 03:09'!
labelWrapForContents: aString

	"Returns desired pixel width of label compositing box.
	 Semismart - if there aren't more than a couple of CR's and the string is small
	 use a default small width.  If the string is long, use a wider limit.
	 If there are already lots of CRs in the source string, don't autowrap."

	^ (aString occurrencesOf: (Character cr)) > 3  "The user seems to want control"
		ifTrue: [ nil ]
		ifFalse: [ 
			(aString size < 200)
			ifTrue: [ (aString size > 120)
						ifTrue: [ self class defaultWrap ]
						ifFalse: [ self class defaultWrap // 1.25 ]	"less width for shorter notes" ]
			ifFalse: [ (self class defaultWrap * 1.5) rounded ]  "extra width for longer notes" ].

! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 8/30/2009 10:40'!
menuTitle
	^ (QStyles shorten: (self contents) toLabelOfLength: 12), ' ...'! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 9/3/2008 23:42'!
minExtent
	"Corner-resize protocol."
	^ 0.3@0.2
! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 9/3/2008 23:52'!
normal
	"So the resizer can use me as if I were a rectangle."
	^ Vector3 x: 0 y: 0 z: -1! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 7/18/2008 12:55'!
onAvatarColorChanged: anAvatar

	"Try always to appear in the color of my avatar."
	(anAvatar notNil 
	 and: [anAvatar userLogin = (self creatorLogin)])
	 ifTrue: [
		"Since we know the creator matches, this just refreshes the color"
		self creator: anAvatar
	].! !

!QStickyNote methodsFor: 'persist' stamp: 'bgf 7/17/2008 17:31'!
persistOn:pStream
	super persistOn:pStream.
	pStream storePoint: 'extent' value: self extent default: self class defaultExtent.
	pStream storeString: 'contents' value: label contents default: (self class defaultContents).
	pStream storeColor: 'backingColor' value: (self backingColor) default: nil.
! !

!QStickyNote methodsFor: 'transform' stamp: 'bgf 1/10/2010 12:19'!
portraitExtent
	^ self extent! !

!QStickyNote methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:59'!
privateScale: aFactor

	"Keep my layout consistent."
	nomScale := aFactor * nomScale.
	self extent: (self extent * aFactor).! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 7/20/2008 12:27'!
root: aSpace

	"Extended to watch for avatar color changes."
	self stopScript: #onAvatarColorChanged:.
	super root: aSpace.
	aSpace ifNotNil: [
		self startScript: #onAvatarColorChanged: when: { aSpace . #avatarColorChanged }.
	].! !

!QStickyNote methodsFor: 'accessing' stamp: 'das 10/18/2009 15:58'!
setColor: col

	self backingColor: col.! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 9/3/2008 23:50'!
setPlacementLocked: aBool

	super setPlacementLocked: aBool.
	sizer ifNotNil: [ sizer isEnabled: (adapter isEnabled)].
! !

!QStickyNote methodsFor: 'updating' stamp: 'bgf 3/27/2010 08:50'!
textFromEditor: aString utcSeconds: seconds avatar: anAvatar

	"This is the callback for updating the text from a client-side edit job.
	 Labels are touchy about leading and trailing linefeeds,
	 so we do some surgery to remove leading and trailing feeds."

	| tmp |

	self root ifNil: [ ^ self ].	"I've been removed"

	tmp := aString.
	[tmp isEmpty not and: [tmp last = (Character cr)]] whileTrue: [
		tmp := tmp allButLast.
	].
	[tmp isEmpty not and: [tmp first = (Character cr)]] whileTrue: [
		tmp := tmp allButFirst.
	].

	"Nil indicates a non-change, so don't update the modified time."
	(seconds notNil and: [ seconds > 0 ]) ifTrue: [
		self lastModifiedTime: seconds
	].

	"Extended behavior - if the label was left empty, 
	 or was left at its initial default, then discard the thumbtack;
	 this gives a  way to 'cancel' note-creation 
	 by dismissing the popup editor that appears."

	(tmp isEmpty or: [tmp = (self class defaultContents)]) ifTrue: [
		(self contents = self class defaultContents) 
			ifTrue: [ self destroySoon ]
			ifFalse: [ self contents: tmp ]
	] ifFalse: [
		self contents: tmp.
		self hiliteBriefly.
	].! !

!QStickyNote methodsFor: 'accessing' stamp: 'bgf 9/2/2008 01:59'!
topicFirstLineMax: nChars

	"The first m < n characters of the topic, truncated with ellipsis, with CR's changed to space."
	^ QStyles firstLinesFrom: label contents max: nChars! !

!QStickyNote methodsFor: 'events' stamp: 'bgf 8/1/2009 15:55'!
warpThresholdAdjustment
	"Notes are small, and we want to drag/arrange them from afar."
	^ 10! !

!QStickyNote methodsFor: 'printing' stamp: 'bgf 10/4/2008 02:02'!
writeTranscriptOn: stream
	"For assembling text reports of note-collections."
	stream nextPutAll: (self titleString).
	self lastModifiedTime ifNotNil: [
		stream nextPut: $ .
		stream nextPutAll: (QStyles dateStringFromUTC: self lastModifiedTime).
	].
	stream cr.
	stream nextPutAll: (self contents).
	stream cr.
	! !

QAnnotation subclass: #QThumbtack
	instanceVariableNames: 'label2D image2D textColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QThumbtack commentStamp: '<historical>' prior: 0!
A thumbtack is a 2-D (constant pixel size) annotation,
comprising a thumbtack graphic and a label string floating above it.

Relies on QTextEditJob support in the harness for editing support.
!

!QThumbtack class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QThumbtack class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 23:24'!
defaultBackgroundColor
	^ Color r:0.99 g: 0.94 b:0.5 alpha:0.7! !

!QThumbtack class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 23:30'!
defaultContents
	^ '<Thumbtack Label>'! !

!QThumbtack class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QThumbtack class methodsFor: 'defaults' stamp: 'bgf 2/11/2007 09:36'!
defaultTextColor
	"The default color for thumbtack labels."

	^ Color r: 0.04 g:0.08 b:0.12! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 8/23/2009 17:32'!
attachmentOffset
	^0.1! !

!QThumbtack methodsFor: 'persist' stamp: 'bgf 2/10/2007 23:29'!
constructFrom:pStream
	super constructFrom:pStream.
	self contents: (pStream readString: 'contents' default: self contents).
	! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 2/10/2007 23:29'!
contents
	^label2D contents! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 2/10/2007 23:30'!
contents: aString
	label2D contents: aString! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 7/19/2008 12:27'!
creator: anAvatar
	"Stub, for workalike with sticky notes."! !

!QThumbtack methodsFor: 'initialize' stamp: 'bgf 7/12/2006 23:44'!
defaultForm
	| form |
	form := self island at: #ThumbTackForm.
	form ifNil:[
		form := Form fromFileNamed: (FileDirectory pathFromURI:'icons/Qwaq/thumbtack32.png').
		form flipVertically.
		form := self island at: #ThumbTackForm put: form.
	].
	^form
! !

!QThumbtack methodsFor: 'updating' stamp: 'bgf 3/1/2009 12:52'!
editByUser: anAvatar

	"Open an editor on the label's text, for the given avatar."

	QTextEditJob
			edit: (label2D contents)
			title: ('Thumbtack')
			forUser: anAvatar
			recipient: self
			selector: #labelFromEditor:.

	self hiliteBriefly.
! !

!QThumbtack methodsFor: 'updating' stamp: 'bgf 2/9/2008 22:36'!
hilite
	self hilite: (self class defaultBackgroundColor darker).! !

!QThumbtack methodsFor: 'updating' stamp: 'bgf 2/11/2007 09:24'!
hilite:aColor
	label2D textColor: (Color red darker).
	label2D backgroundColor: aColor asColor! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 1/10/2007 01:22'!
image

	"The Form to display as the 'thumbtack' portion of the annotation;
	 the default is a small thumbtack glyph."

	^ image2D form! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 2/10/2007 22:15'!
image: aForm

	"Supply an  image form to use in place of the default thumbtack.
       This image appears on-screen at its natural pixel size;
	 it is not persisted, however, so this is only useful for subclasses."

	image2D form: aForm! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 1/10/2007 01:21'!
imageOffset

	"The raster offset from the tack origin to the label origin, in screen raster space.
	 The default is x 0,0 - the image form bottom is at the tack location."

	^ image2D xyOffset! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 1/10/2007 01:10'!
imageOffset: aPoint

	"The raster offset from the tack origin to the label origin, in screen raster space.
	 The default is x 0,0 - the image form bottom is at the tack location."

	image2D xyOffset:aPoint.! !

!QThumbtack methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:38'!
initialize
	super initialize.
	self visible: true. "!!!!!! otherwise no picking"

	adapter overColor: (self class defaultBackgroundColor darker alpha: 0.3).
	adapter downColor: (self class defaultBackgroundColor alpha:0.5).

	label2D := QLabel2D new.
	label2D xyOffset: (0@(self defaultForm height)) negated.
	label2D align: #bottomCenter.
	label2D backgroundColor: (self class defaultBackgroundColor).
	self addChild: label2D.

	self textColor: (self class defaultTextColor).
	self contents: (self class defaultContents).

	image2D := QImage2D new.
	image2D form: self defaultForm.
	image2D align: #bottomCenter.
	self addChild: image2D.

! !

!QThumbtack methodsFor: 'updating' stamp: 'bgf 2/10/2007 23:30'!
labelFromEditor:  aString

	"Labels are touchy about leading and trailing linefeeds."
	| tmp |

	tmp := aString.
	[tmp isEmpty not and: [tmp last = (Character cr)]] whileTrue: [
		tmp := tmp allButLast.
	].
	[tmp isEmpty not and: [tmp first = (Character cr)]] whileTrue: [
		tmp := tmp allButFirst.
	].

	"Extended behavior - if the label was left empty, 
	 or was left at its initial default, then discard the thumbtack;
	 this is an easy way to 'cancel' thumbtack creation by dismissing the initial editor."
	(tmp isEmpty or: [tmp = (self class defaultContents)]) ifTrue: [
		self destroySoon
	] ifFalse: [
		self contents: tmp
	].! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 1/10/2007 01:21'!
labelOffset

	"The raster offset from the tack origin to the label origin, in screen raster space.
	 The default is x 0, y half the height of the image form.  Negative y is up."

	^ label2D xyOffset! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 1/10/2007 01:07'!
labelOffset: aPoint

	"The raster offset from the tack origin to the label origin, in screen raster space.
	 The default is x 0, y half the height of the image form.  Negative y is up."

	label2D xyOffset:aPoint.! !

!QThumbtack methodsFor: 'persist' stamp: 'bgf 2/10/2007 23:31'!
persistOn:pStream
	super persistOn:pStream.
	pStream storeString: 'contents' value: self contents default: self class defaultContents.
	! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:35'!
textColor
	^ textColor! !

!QThumbtack methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:35'!
textColor: aColor
	textColor := aColor.
	label2D textColor: aColor! !

!QThumbtack methodsFor: 'updating' stamp: 'bgf 2/11/2007 00:24'!
translateIntoContactWith: aPoint alongNormal: aNormal

	"As a 2-D billboard element, I don't have much to do here;
	 just place my origin at the given point, and apply my fixed offset."
	| |
	self translation: (aPoint + (self attachmentOffset * aNormal)).! !

!QThumbtack methodsFor: 'updating' stamp: 'bgf 2/11/2007 10:09'!
unhilite
	label2D ifNotNil:[
		label2D textColor: (self textColor).
		label2D backgroundColor: (self class defaultBackgroundColor).
	].! !

QAnnotation subclass: #QTopicCard
	instanceVariableNames: 'title body creatorColor chatLog footer extent sizer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Annotations'!

!QTopicCard commentStamp: '<historical>' prior: 0!
A note-card is a structured annotation resembling a 3x5 index card,
for use in "brainstorming" and generally for more discursive annotations
than basic sticky notes.

Note cards carry some visual structure - a color coded title at the top,
some summary information at the bottom.
!

!QTopicCard class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTopicCard class methodsFor: 'inspector' stamp: 'sn 3/9/2013 14:51'!
addPropertyRows: propertyDialog forFrame: tcref

	|  cRow aRow |

	"No Super - for cards we override the property sheet."
	"Does the label have a non-nil creatorLogin? "

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [:frame | frame get: #extent ];
			acceptBlock: [ : ext :frame |  frame future extent: ext ];
			balloonHelp: 'Dimensions of the rectangle'translateMe;
		yourself).

	propertyDialog addRow: (
		cRow := (QChoicesPropertyRow new)
			titleString: 'Title Color'translateMe;
			setChoices: (self colorChoices);
			fetchBlock: [ :frame |(frame get: #creatorColor) ];
			acceptBlock: [ : c :frame | frame future creatorColor: c ];
			balloonHelp: 'The color of the title strip.\(Automatic if the author is recorded.)'translateMe withCRs;
			comparisonBlock: [ :v1 : v2 | QColors sameColor: v1 as: v2 ];
			inputEnabled: (propertyDialog targets allSatisfy: [ : ref |  (ref get: #creatorLogin) isNil ]);
		yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Last Modified'translateMe;
			fetchBlock: [:frame | (frame get: #lastModifiedTime) 
							ifNil: [''] 
							ifNotNilDo: [ :seconds| QStyles dateStringFromUTC:  seconds ] ];
			balloonHelp: 'When last edited, if known.'translateMe;
			inputEnabled: false;
		yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Author'translateMe;
			fetchBlock: [ :frame |(frame get: #creatorLogin) ifNil: [''] ];
			balloonHelp: 'The original author, if known.'translateMe;
			inputEnabled: false;
		yourself).

	propertyDialog addRow: (
		aRow := (QBooleanPropertyRow new)	
			titleString: 'Attributed to Original Author'translateMe;
			fetchBlock: [ :frame | (frame get: #creatorLogin) notNil ];
			acceptBlock: [ : v :frame | v ifFalse: [ frame future creator: nil ]];
			balloonHelp: 'When checked, the author is recorded, and only that person can edit.\Cannot be restored if unchecked.'translateMe withCRs;
			inputEnabled: (propertyDialog targets anySatisfy: [ : ref | (ref get: #creatorLogin) notNil ]);
		yourself).

	propertyDialog addRow: (
		(QBooleanPropertyRow new)
			titleString: 'Lock Placement'translateMe;
			fetchBlock: [:frame | frame get: #isPlacementLocked ];
			acceptBlock: [ : b :frame | frame future isPlacementLocked: b ];
			balloonHelp: 'Can the card be moved?'translateMe;
		yourself).

	"Keep consistency about what is editable..."		
	propertyDialog  startScript: [
			| notes | 
			notes := propertyDialog targets.
			notes ifNotNil: [ 
				cRow inputEnabled: (notes allSatisfy: [ : ref |  (ref get: #creatorLogin) isNil ]).
				aRow inputEnabled: (notes anySatisfy: [ : ref | (ref get: #creatorLogin) notNil ]).
			]]
		when: { propertyDialog . #accepted }.
		! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/27/2008 15:36'!
colorChoices
	
	"A dictionary of value names->choices for the paper color.
	 The names are the avatar-color names, plus 'Default' and 'Paper';
	 the values are VectorColors ... the paper-pastel variants of avatar colors."

	^ QStickyNote colorChoices
! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:11'!
defaultBackingColor
	"The (post-processed, i.e. lightened) backing color to use
	  when no creatorColor is set.
	 Post-processed color so that persistOn: can detect equality."
	^ QColors topicCardColor! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 11/24/2009 19:57'!
defaultCanTakeAnnotations
	^ false! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/31/2008 20:30'!
defaultContents
	^'<Describe the topic here>'! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/31/2008 20:38'!
defaultExtent
	"Default size for an index card.  Huge in world-scale, more of a sheet-o-paper.."
	^ 1.8@1.3		"Not 3x5, which looks oddly oblong, but bigger than a sticky"! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/27/2008 15:50'!
defaultMargin
	^ 8! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/27/2008 22:45'!
defaultMaterial
	"The material for the card body (which underlies both title and body)"
	^ (QMaterials flatWhiteMaterial)
		ambientColor: (self defaultBackingColor)/2.0;
		diffuseColor: (self defaultBackingColor);
		cullFace:false;
		fullBright:true;
		yourself.! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/27/2008 16:35'!
defaultStyleName
	^ #CardStyle! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/27/2008 22:47'!
defaultTextColor
	^ Color r: 0.04 g: 0.05 b: 0.09		"Renders slightly better than black."! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:12'!
defaultTitleColor
	"The (post-processed, i.e. lightened) backing color to use
	  when no creatorColor is set.
	 Post-processed color so that persistOn: can detect equality."
	^ QColors topicCardTitleColor! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/27/2008 15:51'!
defaultWrap
	"Pixel length of a line before it autowraps - relates to default style"
	^ 280! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 8/31/2008 18:27'!
editorTitle
	^ 'Topic and Discussion'! !

!QTopicCard class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 10:24'!
hiliteColor
	^VectorColor r: 0.5 g: 0.5 b: 0.55 a:0.3! !

!QTopicCard methodsFor: 'accessing' stamp: 'das 10/18/2009 15:44'!
addMessage: txt fromUser: cuser atTime: utcSeconds

	| secs |
	secs := utcSeconds ifNil:[ DateAndTime now asUTCSeconds ].
	chatLog addMessage: txt fromUser: cuser atTime: secs.! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/28/2008 00:25'!
asPlace
	^ QPlace
		category: 'Notes'
		label: self titleString
		frame: self
		sticky: true! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/15/2009 22:35'!
asTopicRecord
	"The QTopicRecord describing me for topic review.
	  This must be side-effect free. 
	  QTopicRecord is a pass-by-copy simple data record."

	^ (QTopicRecord new)
			cardRef: self;
			author: (self creatorLogin);
			color: (self creatorColor);
			creationTime: ((self creationTime) ifNil: [0]);
			lastModifiedTime: ((self  lastModifiedTime) ifNil: [ 0 ]);
			lastCommentTime: ((self lastCommentTime) ifNil: [ 0 ]);
			topic: ((self topicFirstLineMax: 34) ifNil: ['']);
			nComments: (self commentCount);	
		yourself! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/23/2009 17:33'!
attachmentOffset
	"For z-buffer relief"
	^ 0.025! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 15:34'!
backingScale
	"How much bigger"
	^ 1.1! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/23/2009 19:31'!
ballotColor
	"Note colors are light paper colors... darken back for ballot/poll use."
	^ (self creatorColor 
		ifNotNilDo: [ : cc |
			cc asColor adjustSaturation: 0.2 brightness: -0.25 
		] 
		ifNil: [ Color gray ]) asVectorColor! !

!QTopicCard methodsFor: 'testing' stamp: 'bgf 10/25/2009 14:53'!
canEditInBox
	"The edit-box is inappropriate for these.  Distorts the surface-attachment."
	^ false
! !

!QTopicCard methodsFor: 'testing' stamp: 'bgf 2/19/2009 16:09'!
canEditMaterial
	"The material editor wants to apply the material across the subtree; just say no."
	^ false
! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 20:30'!
chatLog
	"My discussion record."
	^ chatLog! !

!QTopicCard methodsFor: 'render' stamp: 'bgf 9/3/2008 23:38'!
child: c setDistance: ogl

	"Treat children as equidistant so we can control subdistance.
	 The resize grip needs a negative subdistance so it is rendered as if closer than the bottom strip."
	frameParent ifNotNil: [
		^ (frameParent child: self setDistance: ogl) - Float epsilon.
	].
	^ super child: c setDistance: ogl! !

!QTopicCard methodsFor: 'render' stamp: 'bgf 9/3/2008 22:01'!
childSubdistance: c
	"The resize grip needs a negative subdistance so it is rendered as if closer than the bottom strip."
	^ c == (sizer grip) ifTrue: [ -0.1 ] ifFalse: [ 0 ]! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 9/1/2008 18:26'!
commentCount
	"How many user-issued comments do I contain?"
	^ chatLog realMessageCount! !

!QTopicCard methodsFor: 'persist' stamp: 'bgf 1/20/2009 17:14'!
constructFrom:pStream
	super constructFrom:pStream.
	self extent: (pStream readPoint: 'extent' default: self extent).
	self contents: (pStream readString: 'contents' default: self class defaultContents).
	self creatorColor: (pStream readColor: 'creatorColor' default: (self creatorColor)).

	chatLog constructLogFrom: pStream.

	"Update footer if we have chat messages."
	chatLog messages ifNotEmpty: [
		self onCommentAdded.
	] ifEmpty: [
		self onCommentsCleared.
	].! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 17:17'!
contents
	^ body contents ! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 1/6/2009 17:25'!
contents: aString

	"The string is the note."
	| txt |
	(txt := aString asText) = (body contents) ifFalse: [
		body wrap: (self labelWrapForContents: aString).
		body contents: txt.
		self signal: #contentsChanged.
	].! !

!QTopicCard methodsFor: 'testing' stamp: 'bgf 2/9/2010 13:49'!
contextMenuMaskForAvatar: avatar
	"Remove the system 'Attach Note' command from the menu."
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	mask remove: #attachNote ifAbsent: [].
	mask add: #createPoll.
	^ mask! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 9/1/2008 15:43'!
creationTime
	"Time I was made - the timestamp of my 'created' log entry, if known"
	^ chatLog messages 
		ifEmpty: [ nil ]
		ifNotEmpty: [ chatLog messages first timestamp ]! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 22:37'!
creator: anAvatar

	"Have the note remember who made it.
	 Also set the backing color to the current paperColor for that avatar...
	 (And note that so long as the creatorLogin is remembered, 
	  subsequent avatar-preference color updates will be reflected in my color.)"

	| color |
	anAvatar ifNotNil: [
		color := anAvatar avatarColor.
		color ifNotNil: [
			self creatorColor: (QColors paperColorForAvatarColor: color).
		] ifNil: [
			self creatorColor: nil.
		].
		self creatorLogin: (anAvatar userLogin).
	] ifNil: [
		self creatorLogin ifNotNil: [
			self creatorLogin: nil.		"Decouple further updates."
			self creatorColor: nil.		"And forget the old association."
		].
	].! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 22:34'!
creatorColor
	"The 'paper color' of the note.
 	 Note that if the creatorLogin is set,
	 then the color will be updated to match the current appearance of that user,
	 if they are present in a forum session,
	 and the persisted color will be the last such color.
	 If creatorLogin is cleared to nil, that updating is disabled."

	^ creatorColor! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 15:49'!
creatorColor: aColor
	"The backing color for the note...
	 often, the personal color of the person introducing the note.
	 Default is a yellow post-it style note.

	 To respect compact persistence for the default,
	 we record <nil> as-is, applying the default to the material in that case."

	aColor ifNotNil: [
		creatorColor := aColor asVectorColor.
	] ifNil: [
		creatorColor := nil.
	].

	title backgroundColor: (creatorColor ifNil: [ self class defaultTitleColor ]).
	footer backgroundColor: (title backgroundColor asColor muchLighter asVectorColor).! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 17:59'!
creatorLogin: aName
	super creatorLogin: aName.
	title contents: (self titleString).! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 8/30/2009 10:42'!
deletionLabel
	^ 'Topic Card'! !

!QTopicCard methodsFor: 'transform' stamp: 'bgf 7/26/2009 20:18'!
destinationExtent
	^ (self extent * 5) min: 12@12 max: 4@2! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 7/27/2009 00:03'!
destinationStanceFor: anAvatar
	^ QGotoButton 
			stanceFor: anAvatar
			toView: self
			inDestination: self		! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 3/8/2010 11:15'!
editByUser: anAvatar

	"Actually 'view or edit by user...'.
	 We open the dialog in either case,  but only allow editing/updating for 
 	 those allowed to edit (usually the owner.)"

	self hiliteBriefly.

	"Placeholder - this will use a the note-taking tool."
	QTextEditJob
		edit: (body contents asString)
		title: (self editorTitleFor: anAvatar)
		forUser: anAvatar
		recipient: self
		selector: ((self editableByUser:anAvatar)
					ifTrue: [ #textFromEditor:utcSeconds:avatar: ]
					ifFalse: [ nil ] ).

	! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 21:29'!
editToolClass
	"This is what the text-edit job uses to edit me."
	^ QTopicTool! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 8/27/2008 15:34'!
editableByUser: anAvatar

	"A note is editable by its creator only.
 	 If the note has no remembered creator, it is editable by anyone."

	^ (creatorLogin isNil) 
			or: [ 
				(anAvatar notNil 
				 and: [ anAvatar userLogin = creatorLogin ]) ].
! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 17:21'!
editorTitleFor: anAvatar

	"How to caption the editor?  Say who owns it, if we know."
	(anAvatar userLogin = creatorLogin) ifTrue: [
		^ (self class editorTitle), ' by You'
	].

	creatorLogin ifNotNil: [
		^ 'Topic by ', (self creatorName)
	].
	^ self class editorTitle. ! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 22:09'!
extent
	^ extent
! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 22:10'!
extent: xy
	extent := xy clone.
	self shapeChanged.! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 2/15/2010 17:58'!
floatingMessageFor: anAvatar

	"Adjust the floating message according to whether I can edit or not,
	 whether there's a persisted creator and/or last-mod date, etc."

	|  cnt |

	^ String streamContents: [ : s |
		s nextPutAll: (self topicFirstLineMax: 32).
		s nextPut: $\.

		self creatorLogin ifNotNilDo: [  : cl |
			cl = (anAvatar userLogin) ifTrue: [ s nextPutAll: 'Topic by You' ] ifFalse: [ s nextPutAll: 'By ', cl ].
		] ifNil: [
			s nextPutAll: 'Open Topic'
		].

		self lastModifiedTime ifNotNil: [
			s nextPutAll: ' -  ';
			   nextPutAll: (QStyles dateStringFromUTC: self lastModifiedTime);
			   nextPut: $\..
		].
		cnt := self commentCount.
		cnt > 0 ifTrue: [
			s nextPutAll: ( cnt > 1 ifTrue: [ cnt printString, ' Comments' ] ifFalse: [ 'One Comment' ] );
			  nextPutAll: ' - ', (QStyles dateStringFromUTC: ((self lastCommentTime) ifNil: [0]));
			  nextPut: $\.
		].
		(self editableByUser: anAvatar)
			ifTrue: [ s nextPutAll: 'Double-click to edit.' ]
			ifFalse: [ s nextPutAll: 'Double-click to comment.' ].
	].	
! !

!QTopicCard methodsFor: 'persist' stamp: 'bgf 9/1/2008 18:27'!
footerString
	| sz |
	^ (sz := (self commentCount)) caseOf: {
		[1] -> [ 'One comment' ].
		[0] -> [ 'No comments' ].
	} otherwise: [
		'', sz, ' Comments' 
	].
	! !

!QTopicCard methodsFor: 'accessing' stamp: 'das 8/5/2009 12:46'!
getChatLog
	"My discussion record."
	^ chatLog getLog.! !

!QTopicCard methodsFor: 'accessing' stamp: 'das 10/18/2009 15:57'!
getColor

	^ self creatorColor.! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 8/27/2008 15:34'!
hilite
	self hilite: self class hiliteColor! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 22:16'!
hilite: aColor
	"Confine hiliting to the title... just looks cleaner"
	title ifNotNil: [ title hilite: aColor ].
	footer ifNotNil: [ footer hilite: aColor ].! !

!QTopicCard methodsFor: 'initialize' stamp: 'bgf 11/8/2009 11:00'!
initialize
	super initialize.
	self objectName: 'Topic Card'.
	self extent: self class defaultExtent.

	adapter overColor: (QColors whiteOverColor).	"We're like file-cards for hiliting."
	chatLog := QChatLog new.
	title := (QLabel new)
		material: (self class defaultMaterial);
		styleName:  self class defaultStyleName;
		textColor: self class defaultTextColor;
		margin: 1;
		cullFace: false;
		yourself.
	self addChild: title.

	footer := (QLabel new)
		material: (self class defaultMaterial);
		styleName:  self class defaultStyleName;
		textColor: self class defaultTextColor;
		margin: 1;
		contents: self footerString;
		cullFace: false;
		yourself.
	self addChild: footer.

	body := (QLabel new)
		styleName:  self class defaultStyleName;
		backgroundColor: (self class defaultBackingColor);
		textColor: self class defaultTextColor;
		margin: self class defaultMargin;
		cullFace: false;
		yourself.
	self addChild: body.

	self solid: false.
	self contents: (self class defaultContents).
	self shapeChanged.

	self startScript: #onCommentAdded when: { chatLog . #messageAdded }.
	self startScript: #onCommentsCleared when: { chatLog . #resetChat }.

	sizer := QCorner2DResizer forLuggage: self.! !

!QTopicCard methodsFor: 'testing' stamp: 'bgf 9/1/2008 19:42'!
isListedTopic
	^ true
	"Yes, this is a managed item visible in the topics tool"! !

!QTopicCard methodsFor: 'testing' stamp: 'bgf 8/23/2009 17:55'!
isPollBallot
	"Include me when making a poll from a group of attachments?"
	^ true! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 17:21'!
labelForBrowser
	^ self titleString! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 15:34'!
labelWrapForContents: aString

	"Returns desired pixel width of label compositing box.
	 Semismart - if there aren't more than a couple of CR's and the string is small
	 use a default small width.  If the string is long, use a wider limit.
	 If there are already lots of CRs in the source string, don't autowrap."

	^ (aString occurrencesOf: (Character cr)) > 3  "The user seems to want control"
		ifTrue: [ nil ]
		ifFalse: [ 
			(aString size < 200)
			ifTrue: [ (aString size > 120)
						ifTrue: [ self class defaultWrap ]
						ifFalse: [ self class defaultWrap // 1.25 ]	"less width for shorter notes" ]
			ifFalse: [ (self class defaultWrap * 1.5) rounded ]  "extra width for longer notes" ].

! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 9/1/2008 16:52'!
lastCommentTime
	"Timestamp of the last message added."
	chatLog realMessageCount > 0 ifTrue: [
		^ chatLog messages last timestamp
	].
	^ nil! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 16:57'!
lastModifiedTime: aTime

	super lastModifiedTime: aTime.
	title contents: (self titleString).! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 22:09'!
material
	"Nominal, so the mat-editor doesn't get confused."
	^ self class defaultMaterial.! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/30/2009 10:41'!
menuTitle
	^ (QStyles shorten: (self contents) toLabelOfLength: 12), ' ...'! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 9/3/2008 21:23'!
minExtent
	"Corner-resize protocol."
	^ 0.5@0.3! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 9/3/2008 23:52'!
normal
	"So the resizer can use me as if I were a rectangle."
	^ Vector3 x: 0 y: 0 z: -1! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 8/27/2008 15:34'!
onAvatarColorChanged: anAvatar

	"Try always to appear in the color of my avatar."
	(anAvatar notNil 
	 and: [anAvatar userLogin = (self creatorLogin)])
	 ifTrue: [
		"Since we know the creator matches, this just refreshes the color"
		self creator: anAvatar
	].! !

!QTopicCard methodsFor: 'events' stamp: 'bgf 9/2/2008 14:16'!
onCommentAdded
	"The footer counts comments."
	footer contents: self footerString.
	chatLog messages ifNotEmpty: [
		"Paranoia"
		self signal: #commentAdded with: chatLog messages last.
	].! !

!QTopicCard methodsFor: 'events' stamp: 'bgf 1/20/2009 16:59'!
onCommentsCleared
	"The footer counts comments."
	footer contents: self footerString.
	! !

!QTopicCard methodsFor: 'persist' stamp: 'bgf 8/31/2008 00:55'!
persistOn:pStream

	super persistOn:pStream.
	pStream storePoint: 'extent' value: self extent default: self class defaultExtent.
	pStream storeString: 'contents' value: (self contents) default: (self class defaultContents).
	pStream storeColor: 'creatorColor' value: (self creatorColor) default: nil.
	pStream storeObject: 'cardMaterial' value: (self material) default: (self class defaultMaterial).

	chatLog persistLogOn:pStream.! !

!QTopicCard methodsFor: 'transform' stamp: 'bgf 1/10/2010 12:19'!
portraitExtent
	^ self extent! !

!QTopicCard methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:59'!
privateScale: aFactor

	"Keep my layout consistent."
	nomScale := aFactor * nomScale.
	self extent: (self extent * aFactor).! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 15:34'!
root: aSpace

	"Extended to watch for avatar color changes."
	self stopScript: #onAvatarColorChanged:.
	super root: aSpace.
	aSpace ifNotNil: [
		self startScript: #onAvatarColorChanged: when: { aSpace . #avatarColorChanged }.
	].! !

!QTopicCard methodsFor: 'accessing' stamp: 'das 10/18/2009 15:58'!
setColor: col

	self creatorColor: col.! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 9/23/2008 10:26'!
setPlacementLocked: aBool

	super setPlacementLocked: aBool.
	sizer ifNotNil: [
		sizer isEnabled: (adapter isEnabled).
	].! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 3/15/2010 17:17'!
textFromEditor: aString utcSeconds: seconds avatar: anAvatar

	"This is the callback for updating the text from a client-side edit job.
	 Labels are touchy about leading and trailing linefeeds,
	 so we do some surgery to remove leading and trailing feeds."

	| tmp |

	self root ifNil: [ ^ self ].	"I've been removed"

	tmp := aString.
	[tmp isEmpty not and: [tmp last = (Character cr)]] whileTrue: [
		tmp := tmp allButLast.
	].
	[tmp isEmpty not and: [tmp first = (Character cr)]] whileTrue: [
		tmp := tmp allButFirst.
	].

	"Extended behavior - if the content was left empty, 
	 or was left at its initial default, then discard the card;
	 this gives a  way to 'cancel' note-creation 
	 by dismissing the popup editor that appears."
	(tmp = (self class defaultContents)) ifTrue: [
		(self contents = self class defaultContents) ifTrue: [ 
			self destroySoon.
			^ self.
		]].

	(tmp = self contents) ifTrue: [
		^ self
	].
	"Nil indicates a non-change, so don't update the modified time."
	(seconds notNil and: [ seconds > 0 ]) ifTrue: [
		lastModifiedTime ifNil: [	
			"If my stored value was nil, I'm just being made."
			chatLog addMessage: 'Created' fromAvatar: nil atTime: seconds
			].
		self lastModifiedTime: seconds
	].
	self contents: tmp.
	self hiliteBriefly.! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/27/2008 17:56'!
title
	"The label across my top."
	^ title! !

!QTopicCard methodsFor: 'updating' stamp: 'bgf 8/31/2008 18:28'!
titleString
	"The composed string for my title line."

	^ self creatorName ifNil: [
		'Topic (Open)'
	] ifNotNilDo: [ :nm|
		'Topic (By ', nm, ')'
	].! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 9/2/2008 01:59'!
topicFirstLineMax: nChars
	"The first m < n characters of the topic, truncated with CR's changed to space."
	^ QStyles firstLinesFrom: body contents max: nChars
! !

!QTopicCard methodsFor: 'accessing' stamp: 'bgf 8/31/2008 22:16'!
unhilite
	title ifNotNil: [ title unhilite ].
	footer ifNotNil: [ footer unhilite ].! !

!QTopicCard methodsFor: 'initialize' stamp: 'bgf 9/3/2008 21:56'!
updateShape
	"Fiddly, so the there is overlap to prevent gaps...
	 the title takes the top 0.15; the footer is the bottom 0.15, and the body is the rest."
	title
		extent: (extent x @ (extent y * 0.15 ))*1.001;
		translation: (0@ (extent y * 0.426 ) @ 0).
	body
		extent: (extent x)@  (extent y * 0.705);
		translation: (0 @ 0 @ 0).
	footer
		extent: (extent x @ (extent y * 0.15 ))*1.001;
		translation: (0@ (extent y * -0.426 ) @ 0).
	sizer updateShape.
		! !

!QTopicCard methodsFor: 'events' stamp: 'bgf 8/1/2009 15:55'!
warpThresholdAdjustment
	"Cards are small, and we want to drag/arrange them from afar."
	^ 6! !

!QTopicCard methodsFor: 'printing' stamp: 'bgf 10/4/2008 02:01'!
writeTranscriptOn: stream
	"For assembling text reports of note-collections."
	stream nextPutAll: (self titleString).
	stream cr.
	stream nextPutAll: (self contents).
	stream cr.
	self chatLog messages size > 0 ifTrue: [
		self chatLog writeTranscriptOn: stream.
		stream nextPutAll: '- end of topic -'.
		stream cr.
	].

	! !

TGroup subclass: #QBarChart
	instanceVariableNames: 'data extent base baseHeight motionGrip midplane clicker title lastModifiedTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QBarChart commentStamp: '<historical>' prior: 0!
A chart depicts a small array of scalar values for comparison;
each value has a color and a short label.

The base implementation is a simple bar chart thows each value as a vertical slab
arising from the base of the chart, and extending up to its value.
Labels array along the base, scaled to fit.
!

!QBarChart class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QBarChart class methodsFor: 'inspector' stamp: 'sn 3/9/2013 14:19'!
addPropertyRows: propertyDialog forFrame: chartref

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [:f | f get: #extent ];
			acceptBlock: [:ext :f| f future extent: ext ];
			balloonHelp: 'Dimensions of the entire chart.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Plinth Height'translateMe;
			fetchBlock: [ :f | f get: #baseHeight ];
			acceptBlock: [ :ext :f  | f future baseHeight: ext ];
			balloonHelp: 'Height of the label slab.'translateMe;
		yourself).

	super addPropertyRows: propertyDialog forFrame: chartref.! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 18:08'!
baseMaterial
	 ^ (QMaterials flatGrayMaterial)! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/14/2009 18:02'!
defaultBaseHeight
	^ 0.3! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 02:50'!
defaultExtent
	"Default dimensions of a new empty chart object"
	^ Vector3 x: 4 y: 2 z: 0.2! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/14/2009 15:59'!
defaultTextColor
	^ Color white! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 17:48'!
defaultTitleBackground
	^ QColors qwaqGray slightlyDarker! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 14:44'!
defaultTitleString
	^ 'Bar Chart'! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 19:12'!
labelMax
	"Maximum letters in an item label"
	^ 18! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 14:35'!
midplaneMaterial
	"We put an invisible-but-pickable sheet through the body
	 so that picking for drag-placement always works."
	^ QMaterials clearMaterial! !

!QBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 18:07'!
stripeMaterial
	 ^ (QMaterials flatGrayMaterial)! !

!QBarChart methodsFor: 'hierarchy' stamp: 'bgf 8/14/2009 15:12'!
addChild: child atIndex: index
	"Private - make sure children aren't individually editable."
	super addChild: child atIndex: index.
	child isPlacementLocked: true; 	
		  doContextMenu: false.! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/23/2009 16:14'!
asPlace
	^ (QPlace
		category: 'Notes'
		label: self titleString
		frame: self
		sticky: true)
		cameraSelector: #cameraStanceFor:;
		yourself
! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 2/3/2010 10:41'!
asTopicRecord
	"Mostly for testing - if self-contained bar-charts participate in note-collecting,
	 they are testable as tabbed wall attachments."

	^ (QTopicRecord new)
			cardRef: self;
			author: '';
			color: Color blue duller;
			creationTime: 0;
			lastModifiedTime: (self lastModifiedTime ifNil: [0]);
			topic: (self titleString);
			nComments: (self stickers) size;
		yourself! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 14:44'!
assureBarsAndLabels: nKids

	| count bars fronts backs |
	bars := self barCubes.
	fronts := self frontLabels.
	backs := self backLabels.

	count := nKids asInteger max: 0.

	bars size + 1 to: count do: [ : i |
		self addChild: (self makeBar: i).
	].
	fronts size + 1 to: count do: [ : i |
		base addChild: (self makeLabel: i name: 'frontLabel').
	].
	backs size + 1 to: count do: [ : i |
		base addChild: (self makeLabel: i name: 'backLabel').
	].

	count+1 to: bars size do: [ : i |
		(bars at: i ) obliterateLater.
	].
	count+1 to: fronts size do: [ : i |
		(fronts at: i) obliterateLater.
	].
	count+1 to: backs size do: [ : i |
		(backs at: i) obliterateLater.
	].
! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 14:45'!
assureStripes: nKids 
	| count stripes |
	stripes := self stripeCubes.

	count := nKids asInteger max: 0.
	stripes size + 1 to: count do: [ : i |
		self addChild: (self makeStripe: i).
	].
	count+1 to: stripes size do: [ : i |
		(stripes at: i) obliterateLater.
	].
	
	^ self stripeCubes! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/23/2009 21:35'!
attachmentOffset
	"Above most annotations."
	^ 0.2! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/14/2009 18:39'!
backLabels
	^ base children select: [ : k | k class = QLabel and: [ k objectName = 'backLabel' ] ].
! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/18/2009 20:57'!
backgroundClick: event
	"The user has clicked outside the individual samples."
	data zeroValue = 0 ifFalse: [
		event avatar signal:#alertMessage 
					with: ( (self objectName), 
						 '\Bottom Value: ', (data zeroValue), 
						 '\Largest Sample	: ', (data maxValue) ).
	] ifTrue: [ 
		event avatar signal:#alertMessage 
					with: ( (self objectName), 
						 '\Largest Sample	: ', (data maxValue) ).
	].
! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/14/2009 16:51'!
barCubes
	^ self children select: [ : k | k class = TCube and: [ k objectName = 'bar' ] ].
! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/14/2009 18:02'!
baseHeight
	^ baseHeight! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/14/2009 18:05'!
baseHeight: yDimension
	"Base height - height of the black plinth the labels are arranged over."
	baseHeight = yDimension ifFalse: [
		baseHeight := yDimension.
		self shapeChanged.
	].! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/23/2009 16:13'!
cameraStanceFor: anAvatar
	"Because I use grounded stances."
	^ QGotoButton frame: self cameraStanceFor: anAvatar! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/15/2009 08:57'!
colorForBar: n

	^ data colorAt: n! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/23/2009 19:33'!
colorForLabel: i
	"Darkened label background, as the text letters are white."
	^((self colorForBar: i) * 0.8) alpha: 1! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:26'!
colors
	^ data colors! !

!QBarChart methodsFor: 'persist' stamp: 'bgf 8/26/2009 18:23'!
constructFrom: pStream
	super constructFrom: pStream.
	
	self extent: (pStream readVector3: 'extent' default: self extent).
	self baseHeight: (pStream readNumber: 'baseHeight' default: (self baseHeight)).
	self data: (pStream readObject: 'data' default: data).
	self lastModifiedTime: (pStream readNumber: 'lastModifiedTime' default: (self lastModifiedTime)).! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:53'!
data
	^ data! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 9/8/2009 13:40'!
data: barChartData
	barChartData = data ifFalse: [
		self stopScript: #onDataShapeChanged.
		self stopScript: #onDataChanged.

		data := barChartData.
		self runScript: #onDataShapeChanged when: { data . #shapeChanged }.
		self runScript: #onDataChanged when: { data . #dataChanged }.
		self updateDisplay.
	].
	! !

!QBarChart methodsFor: 'initialize' stamp: 'bgf 8/14/2009 15:23'!
defaultIsAtomic 
	"I'm a singleton object from the inspector's POV'"
	^ true! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/30/2009 01:18'!
destinationExtent
	"Relaxed back to see the chrome and context."
	^ (extent x @ extent y)  * 1.667! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/23/2009 15:57'!
destinationStanceFor: anAvatar

	"The stance to apply to the avatar to bring it to me."
	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self.
! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 15:46'!
doubleClick: event
	
	event selection frame = title ifTrue: [
		self titleDoubleClick: event
	].! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:46'!
dragAdapter
	^ motionGrip! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/24/2009 19:00'!
editByUser: aUser
	"Used by the topics tool... so give a response. "
	aUser goToPlace: self asPlace
! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 15:40'!
editToolClass
	"These text edit jobs (renaming tabs) always want one-liners."
	^ QTextLinePane! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/14/2009 15:20'!
extent
	"3D extent of the chart.   XY are width and height"
	^ extent! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/23/2009 20:31'!
extent: xyz
	xyz isPoint ifTrue: [
		extent := (xyz x @ xyz y @ extent z).
	] ifFalse: [
		extent := xyz.
	].
	self shapeChanged.! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/14/2009 18:39'!
frontLabels
	^ base children select: [ : k | k class = QLabel and: [ k objectName = 'frontLabel' ] ].
! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/14/2009 18:48'!
hasAlpha 
	^ true! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/18/2009 16:10'!
height
	"Not including the base"
	^ extent y! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 19:32'!
index: idx pointerClick: event
	"The user has clicked on one of my samples."
	| lbl |
	event avatar signal: #alertMessage with: nil.
	data size >= idx ifTrue: [
		lbl := QStyles formatForAlert: (data descriptionAt: idx).
		event avatar signal: #alertMessage with: ( lbl, '\\', (data valueAt: idx) ).
	].! !

!QBarChart methodsFor: 'initialize' stamp: 'bgf 8/30/2009 01:12'!
initialize
	super initialize.
	self objectName: (self class defaultTitleString).
	self extent: self class defaultExtent.	
	self baseHeight: self class defaultBaseHeight.
	data := QChartData new.

	base := TCube new.
	base material: (self class baseMaterial).
	self addChild: base.

	midplane := TSimpleRectangle new.
	midplane material: self class midplaneMaterial.
	self addChild: midplane.

	title := QLabel new.
	title objectName: 'chart-title'.
	title textColor: (self class defaultTextColor).
	title backgroundColor: (self class defaultTitleBackground).
	title contents: (self objectName).
	title margin: 4.
	self addChild: title.

	clicker := QClickHandler forFrame: self.
	self runScript: #pointerClick: when: { self . #pointerClick }.
	self runScript: #doubleClick: when: { title . #doubleClick }.
	self runScript: #doubleClick: when: { self . #doubleClick }.
	self isPlacementLocked: (self isPlacementLocked).! !

!QBarChart methodsFor: 'testing' stamp: 'bgf 8/14/2009 14:54'!
isEditableGroup
	"No reparenting in or out; use the prototype and count instead."
	^ false! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/14/2009 18:31'!
isGroundBound
	^ false! !

!QBarChart methodsFor: 'testing' stamp: 'bgf 8/18/2009 13:20'!
isListedTopic
	"Mostly for testing - if self-contained bar-charts participate in note-collecting,
	 they are testable as tabbed wall attachments."
	^ true! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 9/30/2009 10:42'!
isPlacementLocked: bLocked
	super isPlacementLocked: bLocked.
	base ifNotNil: [
		bLocked ifFalse: [
			motionGrip 
				ifNotNil: [ motionGrip enable ]
				ifNil: [ self makeMotionGrip ].
		] ifTrue: [
			motionGrip ifNotNil: [
				motionGrip luggage: nil; grip: nil.
				motionGrip := nil.
			].
		]].! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/23/2009 19:01'!
labelStringForBar: barNum
	^ QStyles shorten: (data labelAt: barNum) 
			  toLabelOfLength: (self class labelMax)! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/26/2009 18:22'!
lastModifiedTime
	"UTC seconds of my creation or most recent edit, or nil if unknown"
	^ lastModifiedTime! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/26/2009 18:22'!
lastModifiedTime: utcSeconds
	"UTC seconds of my creation or most recent edit, or nil if unknown"
	lastModifiedTime = utcSeconds ifFalse: [ 	
		lastModifiedTime := utcSeconds.
		self signal: #lastModifiedTimeChanged
	].! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/14/2009 16:04'!
makeBar: barNum
	| label |
	label := (TCube new) 
			objectName: 'bar';
			material: (QMaterials fullWhiteAlphaMaterial);
			checkAlpha;
			yourself.

	^ label! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/30/2009 01:08'!
makeLabel: barNum name: objName
	| label |
	label := (QLabel new) 
			objectName: objName;
			contents: (self labelStringForBar: barNum);
			styleName: #CardStyle;
			textAnchorPoint: #center;
			margin: 2;
			textColor: self class defaultTextColor;
			backgroundColor: (self colorForLabel:  barNum);
			checkAlpha;
			yourself.

	^ label! !

!QBarChart methodsFor: 'private' stamp: 'bgf 9/8/2009 13:42'!
makeMotionGrip
	"Make the motion grip that allows sliding me around by my base."
	motionGrip ifNil: [
		motionGrip := QMotionAdapter new.
		motionGrip
			luggage: self ;
			isGroundBound: false;
			isAttractedToSurfaces: true;
			attractionRange: 6.0;	" Factor based on original distance "
			openSpacePolicy: #viewer;
			surfacePolicy: #upright;
			contactOffset: 0.15;		"Above nearby notes, etc."
			overColor: nil.
	].
	self runScript: #pointerClick: when: { title . #pointerClick }.
	motionGrip grip: title.
! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 18:22'!
makeStripe: stripeNum
	| label |
	label := (TCube new) 
			objectName: 'stripe-', stripeNum;
			material: (self class stripeMaterial);
			yourself.

	^ label! !

!QBarChart methodsFor: 'menu' stamp: 'bgf 8/29/2009 02:26'!
menuTitle
	^ QStyles shorten: (self titleString) toLabelOfLength: 32! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/28/2009 15:46'!
numBars
	^ data count! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/23/2009 14:58'!
objectName: aString
	super objectName: aString.
	title ifNotNil: [
		title contents: aString.
	].! !

!QBarChart methodsFor: 'private' stamp: 'bgf 9/8/2009 13:46'!
onDataChanged
	TObjectID for: 'Bar Chart Data ', (data ifNotNil: [data size]).	"Sync debugging"
	self shapeChanged! !

!QBarChart methodsFor: 'private' stamp: 'bgf 9/8/2009 13:47'!
onDataShapeChanged
	TObjectID for: 'BarChart Data Shape ', (data ifNotNil: [data size]).		"Sync debugging."
	self updateDisplay! !

!QBarChart methodsFor: 'persist' stamp: 'bgf 8/26/2009 18:23'!
persistOn: pStream
	super persistOn: pStream.

	pStream storeVector3: 'extent' value: extent default: (self class defaultExtent).
	pStream storeNumber: 'baseHeight' value: baseHeight default: (self class defaultBaseHeight).
	pStream storeObject: 'data' value: data default: nil.
	pStream storeNumber: 'lastModifiedTime' value: (self lastModifiedTime) default: nil.
! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/23/2009 19:29'!
photographerClass
	^ QDirectPhotoJob! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 15:33'!
pointerClick: event
	
	"Translate pointer-clicks into data-index clicks."

	(title = event selection frame) ifTrue: [
		^ self titleClick: event
	].

	(self barCubes indexOf: (event selection frame) ifAbsent: [nil]) ifNotNilDo: [ : idx |
		data size >= idx ifTrue: [ self index: idx pointerClick: event ].
		^ self].

	(self frontLabels indexOf: (event selection frame) ifAbsent: [nil]) ifNotNilDo: [ : idx |
		data size >= idx ifTrue: [ self index: idx pointerClick: event ].
		^ self].

	(self backLabels indexOf: (event selection frame) ifAbsent: [nil]) ifNotNilDo: [ : idx |
		data size >= idx ifTrue: [ self index: idx pointerClick: event ].
		^ self].

	self backgroundClick: event.
	
! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 1/8/2010 17:06'!
portraitTransform
	"My front-on transform for photos taken with the QDirectPhotoJob."
	| camera ext |
	camera := (self globalTransform clone).
	ext := (self destinationExtent).
	camera translation: (
		camera translation
		+ (camera orientation transformVector: (0@ (self extent y * 0.5) @ (ext maxDimension * 1.1)))).

	^ camera! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:55'!
privateScale: s
	"I assume I manage all my children, so no generic child resize is called."
	nomScale := nomScale * s.
	extent := extent * s.
	baseHeight := baseHeight * s.
	self shapeChanged.! !

!QBarChart methodsFor: 'initialize' stamp: 'bgf 8/15/2009 08:26'!
setColors: arrayOfVectorColors
	data setColors: arrayOfVectorColors! !

!QBarChart methodsFor: 'initialize' stamp: 'bgf 8/15/2009 08:26'!
setLabels: arrayOfShortStrings
	data setLabels: arrayOfShortStrings! !

!QBarChart methodsFor: 'initialize' stamp: 'bgf 8/15/2009 08:26'!
setLabels: labelArray colors: colorArray
	data setLabels: labelArray.
	data setColors: colorArray.! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/15/2009 08:32'!
setValue: value atIndex: i
	data setValue: value atIndex: i! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 14:40'!
stripeCubes
	^ self children select: [ : k | k class = TCube and: [ k objectName beginsWith: 'stripe' ] ].
! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 15:32'!
titleClick: event
	"For subclass intervention."
	self backgroundClick: event.! !

!QBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 17:59'!
titleDoubleClick: event
	"For subclass intervention; default is to rename the chart."

	QTextEditJob 
		edit: (self titleString)
		title: 'Enter Title for Chart' 
		forUser: event avatar 
		recipient: self 
		selector: #titleString:
	
! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/23/2009 14:57'!
titleString
	^ self objectName.! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/23/2009 14:58'!
titleString: aString
	self objectName: aString.
	title contents: aString.! !

!QBarChart methodsFor: 'hierarchy' stamp: 'bgf 8/15/2009 08:56'!
updateDisplay
	"Rebuild the entire current display."

	| nbars |
	nbars := data size.
	self assureBarsAndLabels: nbars.
	self shapeChanged.! !

!QBarChart methodsFor: 'hierarchy' stamp: 'bgf 11/20/2009 21:22'!
updateShape
	| heightScale bars fronts backs zeroValue topValue range le |
	super updateShape.
	zeroValue := data zeroValue.
	topValue := data topValue.
	range := (topValue - zeroValue) max: ( data minRange ifNil: [ 0 ]).
	(range abs < Float epsilon) ifTrue: [
		heightScale := self extent y
	] ifFalse: [
		heightScale := (self extent y) /range.
	].

	"Lay out all the bars."
	bars := self barCubes.
	fronts := self frontLabels.
	backs := self backLabels.
	base extentX: self extent x y: baseHeight  z: self extent z * 0.98;
		 translationX: 0 y: (baseHeight * -0.5) z: 0.
	bars doWithIndex: [ : bar : i |
		bar extentX: (extent x / (bars size)) * 0.85
		      		y: (self visibleValueAt:i) * (heightScale)
		      		z: (self extent z * 0.8);
			translationX: ((extent x / bars size) * (i - 0.5)) - (extent x * 0.5)
				y: (bar extentY *0.5)
				z: 0.
		self setBar: bar colorForIndex: i.
		bar visible: (bar extentY > Float epsilon).
	].

	"The labels are children of the base."
	le := (extent x / bars size) * 0.97 @ (base extentY * 0.9).
	fronts doWithIndex: [ : lbl : i |
		lbl extent: le;
		     translationX: (bars at: i) translation x
			y: (0)
			z: (base extentZ * 0.506).
		lbl backgroundColor: (self colorForLabel: i);
			contents: (self labelStringForBar: i);
			checkAlpha.
	].
	backs doWithIndex: [ : lbl : i |
		lbl rotationAroundY: 180;
			extent:  le;
			translationX: (bars at: i) translation x
				y: (0)
				z: (base extentZ * -0.506).
		lbl backgroundColor: (self colorForLabel: i);
			contents: (self labelStringForBar: i);
			checkAlpha.
	].

	midplane extent: (self extent x @ self extent y);
			 translationX: 0 y: (self extent y * 0.5) z: 0.
	title extent: (self extent x * 1.005)@(1.5 * base extentY);
		translationX: (0) y: (self extent y) + (base extentY * 0.75) z: 0.0.

	self updateStripesForZero: zeroValue 
			top: (topValue max: (range + zeroValue)) 
			heightScale: heightScale.
	self updateBounds.! !

!QBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 15:40'!
updateStripesForZero: zeroValue top: topValue heightScale: heightScale
	| stripes thick height numStripes range stripY |

	"Decide the horizontal scale as a multiple of a power of ten."
	range := topValue - zeroValue.
	stripY := (10 raisedTo: (((range max:1) log floor))) ceiling.
	numStripes := (range / stripY) ceiling.
	numStripes < 3 ifTrue: [
		"But subdivide if we have (e.g.) a basic range that gives too few.."
		stripY := stripY * 0.1.
		numStripes := (range / stripY) ceiling.
	].

	 ((((numStripes *stripY) ceiling) - zeroValue) > (topValue + Float epsilon)) ifTrue: [
		numStripes := numStripes - 1.		
	].
	
	numStripes > 0 ifTrue: [
		| y |
		height := self  extent y.
		stripes := self assureStripes: numStripes + 2.
		thick := (height) * 0.04 / ((numStripes max: 5) asFloat).
		numStripes > 0 ifTrue: [
			1 to: numStripes do: [ : i | | stripe |
				y :=  (((i*stripY) ceiling)- zeroValue) * heightScale.
				stripe := stripes at: i.
				stripe extentX: base extentX  y: thick  z: thick;
					   translationX: 0 y: ( y ) z: 0.
			]].

			"Sides up to the top stripe"
			(stripes at: numStripes+1)
				extentX: thick y: y*1.01 z: thick;
				translationX: (self extent x * 0.5) y: y*0.5 z: 0.
		 	(stripes at: numStripes+2)
				extentX: thick y:  y*1.01 z: thick;
				translationX: (self extent x * -0.5) y: y*0.5 z: 0.
	] ifFalse: [
		self assureStripes: 0.
	].! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 11/20/2009 22:04'!
visibleValueAt: barIndex
	"When plotting in space, the value to draw for the given column"
	^ data valueAt: barIndex! !

!QBarChart methodsFor: 'transform' stamp: 'bgf 8/28/2009 17:56'!
warpThresholdAdjustment
	"Make it easier to click-interact."
	^ 5! !

!QBarChart methodsFor: 'accessing' stamp: 'bgf 8/18/2009 16:10'!
width
	^ extent x! !

QBarChart subclass: #QPollBarChart
	instanceVariableNames: 'votes isSingleChoice isOpenForVoting isSecretUntilClosed resizer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QPollBarChart commentStamp: '<historical>' prior: 0!
A QPollBarChart is a bar chart for voting among a small set of choices.

As choices are made, the bars adjust to depict the results.
(So the balloting is not particularly secret, although once a vote is made,
only the voter can use the interface to see how they voted.)

Voting is by click ing the choices; confimration menus are offered.

The Poll can be single choice or multiple choice.
In a multiple choice poll, each user can vote for as many different options
as they desire.  There is still  a limit of one count-contribution from any user to a choice;
you cannot vote many times for the same choice.

Votes are revocable by clicking again on what you voted for.

The choices are defined by the #data - a QChartData;
setting those up is not managed here.!

!QPollBarChart class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPollBarChart class methodsFor: 'inspector' stamp: 'sn 3/9/2013 14:32'!
addPropertyRows: dialog forFrame: pollref
	| row |

	row := self namePropertyRow: pollref.
	row titleString: 'Title'translateMe.
	row extraLong.
	dialog addRow: row.

	row := QBoolean3PropertyRow 
							newForFlag: 'Open for Votes'translateMe 
							flag: 'Secret til Closed'translateMe
							flag: 'Single Choice'translateMe.

	row
		fetchBlock: [ : f | { (f get: #isOpenForVoting) . (f get: #isSecretUntilClosed) . (f get: #isSingleChoice) .  } ];
		acceptBlock: [ :b2 : f| 
			f future isOpenForVoting: (b2 first). 
			f future isSecretUntilClosed: (b2 second).
			f future isSingleChoice: (b2 third).

		];
		balloonHelp: ('Can users (still) vote?\Hide votes until poll closed?\Vote for a single option?'translateMe withCRs).		
	dialog addRow: row.

	super addPropertyRows: dialog forFrame: pollref.

	(dialog rows detect: [ : r | r titleString = 'Name' ] ifNone: [ nil ]) ifNotNilDo: [ : r |
		dialog removeRow: r titleString
	].! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'bgf 8/30/2009 01:24'!
askForVote: pollRef suggesting: seeds  fromUser: pubName forClient: owner
	"Given a poll, pop a dialog asking this user for a vote. 
	 This is the per-user handler when someone uses 'Call for Votes' features;
	 the user in this client-session is being asked to vote in a poll.
	 If the request came from some other user, their publicName is pubName. "

	| existing |
	pollRef isValidFarRef ifFalse: [ ^ self ].
	pollRef wait.
	existing := owner elements asArray
				detect: [ : player | (player propertyValueAt: #pollReference) = (pollRef value)]
				ifNone: [ nil ].

	existing ifNil: [
		"Indirect script since the dialog uses a nested run...
	 	 and we don't want the main UI signal to queue subsequent requests until the run is ended."
		^ self startScript: #runBallotDialog:suggesting:fromUser:forClient: 
			   withArguments: { pollRef . seeds . pubName . owner }
	].! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'bgf 8/29/2009 02:33'!
ballotButtons: choices multipleChoice: multi currentVotes: votes 
	"Private - build the ballot buttons or checkbox entries for a poll ballot dialog."
	| radios |

	radios := CPlayer new.
	radios layout: CTableLayout new.
	radios hResizing: #spaceFill; vResizing: #spaceFill.
	radios layout layoutInset: 4@3; cellInset: 2@2; cellPositioning: #topLeft.
	radios borderWidth: 1; borderColor: (QColors qwaqOrange alpha: 0.3); borderStyle: #simple.
	radios fillColor: (Color white alpha: 0.8).
	choices doWithIndex: [ : c : i |
		radios add: ( (multi ifTrue: [CCheckBox new] ifFalse: [ CRadioButton new ])
				label: c;
				textAnchorPoint: #leftCenter;
				textEmphasis: ((votes includes: i) ifTrue: [1] ifFalse: [0]);
				selected: false;
				height: 18;
				vResizing: #rigid;
				yourself).
		 ].
	
	^ radios! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 18:23'!
baseMaterial
	^ QMaterials qwaqGrayMaterial! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/30/2009 10:35'!
defaultBaseHeight
	^ 0.3! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 17:48'!
defaultClosedTitleBackground
	^ QColors qwaqGray slightlyDarker ! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/30/2009 10:34'!
defaultExtent
	^ 2.2 @ 1.8 @ 0.05! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 20:27'!
defaultIsOpenForVoting
	^ true! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 11/20/2009 21:24'!
defaultIsSecretUntilClosed
	^ false! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 20:18'!
defaultIsSingleChoice
	^ true! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 16:54'!
defaultOpenTitleBackground
	^ QColors qwaqOrange muchDarker! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 11/21/2009 10:20'!
defaultSecretTitleBackground
	"So they look different from other polls."
	^ Color fromHtmlColorRef: '4022C0'! !

!QPollBarChart class methodsFor: 'utilities' stamp: 'bgf 2/9/2010 14:21'!
isBallotCandidate: aTFrame
	"True if the given frame is one of the ones usable as a ballot description when generating a poll."
	^ (aTFrame isKindOf: QAnnotation) and: [ aTFrame isPollBallot ]! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'bgf 8/29/2009 01:14'!
isBallotChange: radios from: votes
	"Compare the button array to the votes (indeces) array"
	^ (votes hasEqualElements: (
				(radios select: [ : r | r selected ] thenCollect: [ : r | radios indexOf: r ])))
		not! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 19:32'!
labelMax
	"Maximum letters in a poll-choice label"
	^ 25! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/23/2009 17:01'!
midplaneMaterial
	^ QMaterials paperMaterial! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'sn 12/27/2012 21:08'!
prepareQuestionForClient: owner
	"Ask the user to input a question text,
	 create a poll, and send it to my avatar;
	 the in-space operation will add the poll-chart to the forum,
	 and use call-for-votes to pose it as a question to everyone."
	
	| dlog instructions  secret |

	instructions := ('Enter your question here.\'  ,
			'Each person will be asked to choose Yes or No. '  ) withCRs.

	dlog := CUserDialog new.
	dlog dialogLabel value: 'Tally a Question' translateMe .
	dlog messageText 
		contents: (instructions);
		extent: 280@40;
		hResizing: #rigid;
		vResizing: #rigid.
	dlog costume.

	secret := CCheckBox new.
	secret label: 'Hide Results Until Voting is Closed.'translateMe .
	secret value: false; 
		   textAnchorPoint: #leftCenter;
		   textOffset: (3@0);
		   hResizing: #shrinkWrap.
	
	dlog group add: secret after: (dlog messageText).

	dlog buttonRow removeAll.
	(dlog addButton: 'Ask Everyone' translateMe  value: true) 	width: 95; textEmphasis: 1.
	(dlog addButton: 'Cancel' translateMe value: false)			width: 95.
	[
		(dlog run: nil centerAt: owner dialogCenter) = true ifTrue: [
			| question |
			owner readyForEvents ifFalse: [ ^ self ].
			question := (dlog messageText contents asString).
			(question size > 0 and: [ question ~= instructions] ) ifTrue: [ 
				owner avatar replica future  
							questionForTally: question 
							secretUntilClosed: (secret value)
			].
		].
	] ensure: [
		owner remove: dlog.
		dlog destroy.
	].
	! !

!QPollBarChart class methodsFor: 'undo-redo' stamp: 'bgf 6/15/2010 13:02'!
resetPoll: pollRef  votes: votesArray wasOpen: bOpen

	"Undoable action to clear all the votes out of the poll.
	 The votesArray is an array of pairs:   (userLogin -> arrayOfIndexes)  "

	^ QBlockAction 
		undoBlock: [ : avatar | pollRef future restoreVotes: votesArray wasOpen: bOpen ]
		redoBlock: [ : avatar | pollRef future clear.  pollRef future isOpenForVoting: true  ]
		description: 'Reset Poll'
		requiredFrames: { pollRef }.
! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'hrs 5/11/2010 14:32'!
runBallotDialog: pollRef suggesting: seeds  fromUser: pubName forClient: owner
	"Given a poll, pop a dialog asking for a vote. "
	
	|  choices multi |

	owner readyForEvents ifFalse: [ ^ self ].
	(pollRef waitTimeoutSeconds: 5) ifFalse: [ 
		QLogger logInfo: 'Poll check timed out.'.
		^ self ].

	multi := (pollRef get: #isSingleChoice) = false.
	choices := pollRef get: #labels.
	choices size > 1 ifFalse: [ ^ self ].
	
	owner mobileProxy ifNotNilDo: [:proxy |
		^proxy offerPoll: pollRef suggesting: seeds fromUser: pubName.
	].

	"Catch short yes/no agree/disagree style questions... make 'em simple dialogs."
	(multi not and: [ choices size = 2]) ifTrue: [
		(choices anySatisfy: [ :  choiceLabel | choiceLabel size > 11 ]) 
		 ifFalse: [ 
			^ self runSimpleBallotDialog: pollRef fromUser: pubName forClient: owner
		]].

	^ self runFatBallotDialog: pollRef suggesting: seeds fromUser: pubName forClient: owner
! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'bgf 9/13/2009 02:54'!
runFatBallotDialog: pollRef suggesting: seeds  fromUser: pubName forClient: owner
	"Given a poll, pop a dialog asking for a vote. "
	
	| dlog  question choices radios voteButton votes multi group |

	owner readyForEvents ifFalse: [ ^ self ].

	multi := (pollRef get: #isSingleChoice) = false.
	choices := pollRef get: #labels.
	choices size > 1 ifFalse: [ ^ self ].

	question := pollRef get: #titleString.
	votes := pollRef get: #loginVotes: with: (owner userLogin).
	votes := votes ifNil: [#()].

	dlog := CUserDialog new.
	dlog propertyValueAt: #pollReference put: pollRef value.
	(pubName isNil or: [  pubName = owner avatar publicName ]) ifTrue: [ 
		dlog dialogLabel value: 'Ballot'.
	] ifFalse: [
		dlog dialogLabel value: 'Ballot From ', pubName 
	].

	dlog messageText 
			contents: question;
			allowEditing: false;
			hResizing: #shrinkWrap; 
			vResizing: #shrinkWrap.
	dlog costume.

	dlog buttonRow removeAll.
	voteButton := (dlog addButton: 'Vote' value: #vote) 
					width: 80; yourself.
	(dlog addButton: ((votes size > 0) ifTrue: [ 'Cancel'] ifFalse: [ 'Abstain'])
		value: #abstain) 
		width: 80.

	group := self ballotButtons: choices multipleChoice: multi currentVotes: votes.
	radios := group elements asArray.

	voteButton disable.
	radios doWithIndex: [ : rb :idx | 
		dlog startScript: [ voteButton enabled: (self isBallotChange: radios from: votes) ] 
			 when: { rb . #selectedChanged }.
		((seeds ifNil: [ votes ]) includes: idx) ifTrue: [ rb select ]].

	dlog group add: group after: dlog messageText.
	dlog group layout cellInset: 0@3.
	dlog group add: (CPlayer new height: 0; yourself) after: dlog messageText.
	votes isEmptyOrNil ifFalse: [
		dlog group add: 
			((CLabel on: 'You have already voted.\Previous choices appear in bold.' withCRs)
				hResizing: #shrinkWrap;	
				yourself)
			after: group.
		dlog group add: (CPlayer new height: 0; yourself) after: group.
	].
	dlog group add: (CPlayer new height: 4; yourself) before: dlog buttonRow.

	dlog startScript: #destroy when: { pollRef . #deleted }.
	dlog startScript: #destroy when: { pollRef . #destroyed }.

	"We're careful to take away the dialog when done, 
	 so the modal run-loop doesn't conflict with re-use via poseQuestion: "
	[
		(pubName = (owner avatar publicName)) ifFalse: [
			owner playSoundEffect: #chime.
		].
		(dlog run: nil centerAt: owner dialogCenter) = #vote ifTrue: [
			owner readyForEvents ifFalse: [ ^ self ].
			pollRef island = owner avatar replica island ifTrue: [
				pollRef future
					user: (owner avatar replica) 
					setVotes: (radios select: [ : r | r selected ] thenCollect: [ : r | radios indexOf: r ])
					atTime: (Time utcSeconds).
			] ifFalse: [
				owner inform: 
					'That poll is in a different Telespace.\You must return there to vote.' withCRs.
			].
		].
	] ensure: [
		owner remove: dlog.
		dlog destroy.
	].
	! !

!QPollBarChart class methodsFor: 'dialogs' stamp: 'bgf 9/13/2009 02:53'!
runSimpleBallotDialog: pollRef fromUser: pubName forClient: owner
	"For a ballot with two brief choices,  run a normal 'Yes' 'No' 'Cancel' style dialog."
	| question choices dlog ans |
	question := pollRef get: #titleString.
	choices := pollRef get: #labels.
	choices size > 1 ifFalse: [ ^ self ].

	dlog := CUserDialog new.
	dlog propertyValueAt: #pollReference put: pollRef value.
	(pubName isNil or: [  pubName = owner avatar publicName ]) ifTrue: [ 
		dlog dialogLabel value: 'Question'.
	] ifFalse: [
		dlog dialogLabel value: 'Question From ', pubName 
	].
	dlog messageText 
			contents: question;
			allowEditing: false;
			hResizing: #shrinkWrap; 
			vResizing: #shrinkWrap.
	dlog buttonRow removeAll.
	
	(dlog addButton:  (choices first)  value: 1)  width: 80.
	(dlog addButton: (choices second) value: 2) width: 80.
	(dlog addButton: 'Cancel' value: 0)  width: 80.

	dlog startScript: #destroy when: { pollRef . #deleted }.
	dlog startScript: #destroy when: { pollRef . #destroyed }.

	"We're careful to take away the dialog when done, 
	 so the modal run-loop doesn't conflict with re-use via poseQuestion: "
	[
		(pubName = (owner avatar publicName)) ifFalse: [
			owner playSoundEffect: #chime.
		].
		ans := (dlog run: nil centerAt: owner dialogCenter).
		((ans = 1 or: [ ans = 2]) and: [ owner readyForEvents ]) ifTrue: [
			pollRef island = owner avatar replica island ifTrue: [
				pollRef future
					user: (owner avatar replica) 
					setVotes: ( {ans} )
					atTime: (Time utcSeconds).
			]  ifFalse: [
				owner inform: 'That poll is in a different Telespace.\You must return there to vote.' withCRs.
			].
		].
	] ensure: [
		owner remove: dlog.
		dlog destroy.
	].
	! !

!QPollBarChart class methodsFor: 'utilities' stamp: 'bgf 5/25/2010 19:01'!
sortBallots: listOfBallots

	^ (listOfBallots allSatisfy: [ : b | b contents isAllDigits ])
		ifTrue: [ listOfBallots asArray sortBy: [ : b1 : b2 | b1 contents asInteger < b2 contents asInteger ] ]
		ifFalse: [ listOfBallots asArray sortBy: [ : b1 : b2 | b1 contents caseInsensitiveLessOrEqual: b2 contents ] ]! !

!QPollBarChart class methodsFor: 'defaults' stamp: 'bgf 8/18/2009 18:22'!
stripeMaterial
	^ QMaterials qwaqOrangeMaterial! !

!QPollBarChart class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:02'!
undoForVotes: pollRef oldVotes: oldVotes newVotes: newVotes
	^ QBlockAction 
		undoBlock: [ : avatar | pollRef future user: avatar resetVotes: oldVotes  ]
		redoBlock: [ : avatar | pollRef future user: avatar resetVotes: newVotes ]
		description: 'Vote'
		requiredFrames: { pollRef }.
! !

!QPollBarChart class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:02'!
undoableClose: pollRef
	 ^ QBlockAction
		undoBlock: [ pollRef future isOpenForVoting: true  ]
		redoBlock: [ pollRef future isOpenForVoting: false ]
		description: 'Close Voting'
		requiredFrames: { pollRef }.
! !

!QPollBarChart methodsFor: 'menu' stamp: 'sn 12/27/2012 16:19'!
addMenuActions: tmenu forEvent: event
	"Custom context menu."
	| editProtect who |
	who := event avatar.

	editProtect := (event avatar canModifyForum not) 
				 or: [ self fromTemplate and: [ event avatar isEditingTemplate not ] ].
	who allowedToMove ifTrue: [
		tmenu addItemSpec: { 'Approach' translateMe . #bringAvatarHere: } target: self ].

	self isOpenForVoting ifTrue: [
		tmenu add: 'Vote...' translateMe
			    target: self 
			    action: #provokeVoteDialog:from: 
			    arguments: {who . event avatar publicName}.
		tmenu addSeparator.
		(who isMeetingLeader or: [ who meetingLeader isNil ]) ifTrue: [
			(tmenu addItemSpec: { 'Call for Votes' translateMe . #pushPoll: } target: self)
				enabled: self allHaveVoted not.
			tmenu addItemSpec: { 'Close Voting' translateMe . #closePoll: } target: self.
		]].
	tmenu addSeparator.

	editProtect ifFalse: [
		self isPlacementLocked ifFalse: [
			(tmenu add:  'Resize...' translateMe target: who action: #editFrame: argument: self)
				enabled: (self isInEditBox not).
		].
		tmenu addItemSpec: { 'Properties...' translateMe. #editProperties: } target: self ].

	(event avatar canUseInventory and: [ self canBeCopiedBy: event avatar ])  ifTrue: [
		tmenu addSeparator.
		tmenu addItemSpec: { 'Copy to Inventory' translateMe. #copyToInventory: } target: self ].

	(event avatar canDeleteFrame: self) ifTrue: [
		tmenu addSeparator.	
		self hasVotes ifTrue: [
			tmenu addItemSpec: { 'Reset (Clear Votes)' translateMe. #clearPoll: } target: self .
		].
		tmenu addItemSpec: { 'Delete' translateMe. #confirmDelete: } target: self
	]. ! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 8/18/2009 19:15'!
addUserVote: avatar forIndex: idx
	"The user has voted for the given choice (idx) "
	| va | 
	va := votes at: (avatar userLogin) ifAbsent: [ #() ].
	(va includes: idx) ifFalse: [
		votes at: (avatar userLogin) put: (va copyWith: idx).
		data setValue: (((data valueAt: idx) ifNil: [0]) + 1) 
			 atIndex: idx.
	].

	^ idx! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 8/29/2009 02:37'!
allHaveVoted
	"Have all users (avatars) currently here in the forum, already voted?"

	^ (self root allUsers anySatisfy: [ : u | (votes includesKey: (u userLogin)) not ]) not.
	! !

!QPollBarChart methodsFor: 'transform' stamp: 'bgf 11/21/2009 08:37'!
asPlace
	| str |

	"Since these appear in Places, need some distinguishing name."
	str := self titleString.
	(str includesSubstring: 'poll' caseSensitive: false)  ifFalse: [ 
		str := 'Poll: ', str.
	].

	^ (QPlace
		category: 'Panels'
		label: str
		frame: self
		sticky: true)
		cameraSelector: #cameraStanceFor:;
		yourself
! !

!QPollBarChart methodsFor: 'accessing' stamp: 'bgf 8/26/2009 19:06'!
asTopicRecord
	"Mostly for testing - if self-contained bar-charts participate in note-collecting,
	 they are testable as tabbed wall attachments."

	^ (QTopicRecord new)
			cardRef: self;
			author: 'Poll';
			color: Color blue duller;
			creationTime: 0;
			lastModifiedTime: (self lastModifiedTime ifNil: [0]);
			topic: (self titleString);
			nComments: (votes size);
		yourself! !

!QPollBarChart methodsFor: 'events' stamp: 'bgf 11/21/2009 10:25'!
backgroundClick: event
	"The user has clicked outside the samples... give a summary."
	| mine myChoices |
	myChoices := self userVotes: (event avatar).
	myChoices isEmptyOrNil ifFalse: [
		mine := String streamContents: [ : str |
					str nextPutAll: '\You chose:'.
					myChoices do: [ : vidx |
						str nextPut:$\;  space:3;  nextPutAll: (self labelStringForBar: vidx)
					]].
	] ifTrue: [
		mine := (isOpenForVoting ifTrue: ['\You have not voted.' ] ifFalse: [ '\You did not vote.' ]).
	].
	event avatar signal:#alertMessage with: ( 
			(self objectName), 
			(self isSecretUntilClosed ifTrue: [ '\Secret until voting is closed.'] ifFalse: ['']),
			(self isSingleChoice ifFalse: [ '\Multiple-Choice'] ifTrue: ['']),
				'\Number of Voters: ', (votes size) ,
			mine).
! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 6/15/2010 11:57'!
clear
	"Low-level operation to reset the poll to be empty."
	votes := Dictionary new.
	1 to: (data count) do: [ : idx | 
		data setValue: 0 atIndex: idx.
	].		
! !

!QPollBarChart methodsFor: 'menu' stamp: 'bgf 6/15/2010 13:01'!
clearPoll: eventOrAvatar

	"For 'Reset Poll' menu action"
	| old |
	old := Array streamContents: [  : stream |
			votes keysAndValuesDo: [ : userLogin : votes |   
				stream nextPut: { userLogin . votes asArray}.
			]].

	eventOrAvatar avatar
		recordAction: #resetPoll:votes:wasOpen: 
		class: QPollBarChart
		arguments: {self . old . self isOpenForVoting }.
	
	self clear.
	self isOpenForVoting: true.! !

!QPollBarChart methodsFor: 'menu' stamp: 'bgf 3/13/2010 08:53'!
closePoll: eventOrAvatar
	"Triggers take care of showing results, etc..."

	self isOpenForVoting ifTrue: [
		eventOrAvatar avatar
			recordAction: #undoableClose: 
			class: QPollBarChart
			arguments: { self }
	].
	self isOpenForVoting: false.! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 8/23/2009 19:26'!
colorForBar: n
	"We cannot use translucents since they wink behind the backing material."
	^ (data colorAt: n) alpha: 1! !

!QPollBarChart methodsFor: 'persist' stamp: 'bgf 11/20/2009 21:26'!
constructFrom: pStream
	| tmp |
	super constructFrom: pStream.
	self isSingleChoice: (pStream readBool: 'isSingleChoice' default: isSingleChoice).
	self isOpenForVoting: (pStream readBool: 'isOpenForVoting' default: isOpenForVoting).
	self isSecretUntilClosed: (pStream readBool: 'isSecretUntilClosed' default: isSecretUntilClosed).

	"See persistOn:"
	tmp := pStream readStringMap: 'votes' default: nil.
	[ 	tmp ifNotNil: [
			votes := tmp collect: [ : votesString | self stringToVotes: votesString ].
		].
	] on: Error do: [ : ex |
		QLogger logWarning: 'Survey votes could not be recovered: ', ex.
	].
	self reconcile.! !

!QPollBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 21:41'!
drag: cornerPt corner:sym  delta: deltaExt
	"Corner 2D resize handler.
	 We need a custom handler since my origin y is the base, rather than center."
	| newSize dt old |
	old := extent asPoint.
	newSize := (old + (deltaExt x @ (deltaExt y negated))) max: ( baseHeight @ baseHeight ).
	self extent: newSize.
	dt := (newSize - old).
	dt := (dt x * 0.5 @ (dt y negated)).
	self
	        translation: ((self translation) 
					+ (self orientation transformPoint: (dt @ 0.001))).
					"The z fudge factor fights numeric creep away from the pick plane."
! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 6/15/2010 12:40'!
hasVotes
	"True if any votes at all have been recorded."
	^ data values anySatisfy: [ : v | v > 0 ]! !

!QPollBarChart methodsFor: 'events' stamp: 'bgf 8/29/2009 03:22'!
index: idx pointerClick: event
	"A specific data item (i.e. vote choice) has been clicked."
	| who va |
	who := event avatar userLogin.
	who ifNil: [ ^ super index: idx pointerClick: event ].
	self isOpenForVoting ifFalse: [
		"No voting, so just summarize."
		^ super index: idx pointerClick: event
	].
	va := self userVotes: event avatar.
	(va includes: idx) ifTrue: [
		self isSingleChoice ifTrue: [
			^ event avatar signal: #alertMessage 
					with: 'You already voted\for ', (self labelStringForBar:  idx).
		]].

	self isSingleChoice ifTrue: [ 
		self provokeVoteDialog: event avatar suggesting: {idx} from: (event avatar publicName).
	] ifFalse: [
		|suggest|
		suggest := self userVotes: event avatar.
		(suggest includes: idx) ifTrue: [ 
			suggest := suggest copyWithout: idx
		] ifFalse: [
			suggest := suggest copyWith: idx
		].
		self provokeVoteDialog: event avatar suggesting: suggest from: (event avatar publicName).
	].! !

!QPollBarChart methodsFor: 'initialize' stamp: 'bgf 11/20/2009 21:25'!
initialize
	super initialize.
	self objectName: 'Poll'.
	self isSingleChoice: self class defaultIsSingleChoice.
	self isSecretUntilClosed: self class defaultIsSecretUntilClosed.
	self isOpenForVoting: self class defaultIsOpenForVoting.
	votes := Dictionary new.
	data minRange: 5.

	resizer := QCorner2DResizer forLuggage: self.
! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 11/20/2009 22:08'!
isListedLandmark 
	"By popular demand."
	^ (self visible)! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 8/18/2009 20:29'!
isOpenForVoting
	"Is this poll still available to voters?"
	^ isOpenForVoting! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 11/21/2009 10:12'!
isOpenForVoting: bOpen
	"Is this poll still available to voters?"
	isOpenForVoting := bOpen.

	self shapeChanged.		"In case votes are hidden/revealed."! !

!QPollBarChart methodsFor: 'accessing' stamp: 'bgf 8/23/2009 19:47'!
isPlacementLocked: bLocked
	super isPlacementLocked: bLocked.
	resizer ifNotNil: [
		resizer isEnabled: (bLocked not)
	].! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 11/20/2009 21:17'!
isSecretUntilClosed
	"Are results shown during voting ?"
	^ isSecretUntilClosed! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 11/20/2009 22:05'!
isSecretUntilClosed: bSecret
	"Are results shown during voting ?"
	isSecretUntilClosed := bSecret.
	self shapeChanged.! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 8/18/2009 20:19'!
isSingleChoice
	"Does each user get a single vote, or can they make multiple choices?"
	^ isSingleChoice! !

!QPollBarChart methodsFor: 'testing' stamp: 'bgf 8/18/2009 20:19'!
isSingleChoice: bSingle
	isSingleChoice = bSingle ifFalse: [
		isSingleChoice := bSingle.
		self reconcile.
	].! !

!QPollBarChart methodsFor: 'accessing' stamp: 'bgf 8/28/2009 15:46'!
labels
	"Array of strings, one per vote choice."
	^ Array streamContents: [ : strm |
		1 to: self numBars do: [ : i |
			strm nextPut: (self labelStringForBar: i).
		]].
	! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 8/29/2009 02:37'!
loginVotes: userLogin
	"For the given user, return the array of integers represent his vote;
	 each element in the resulting array is an index into my chart-data.
	 For single-choice polls, the result will either be empty or an array of one integer."
	
	^ votes at: userLogin ifAbsent: [ #() ].
! !

!QPollBarChart methodsFor: 'persist' stamp: 'bgf 11/20/2009 21:26'!
persistOn: pStream
	| tmp |
	super persistOn: pStream.
	pStream storeBool: 'isSingleChoice' 
			value: isSingleChoice 
			default: (self class defaultIsSingleChoice).
	pStream storeBool: 'isOpenForVoting' 
			value: isOpenForVoting 
			default: (self class defaultIsOpenForVoting).
	pStream storeBool: 'isSecretUntilClosed' 
			value: isSecretUntilClosed 
			default: (self class defaultIsSecretUntilClosed).

	"TXML has no easy to persist a dictionary<String,<Array-of-Integers>>,
	 so we write the list of votes for specific user as an inline comma-separated string."
	tmp := votes collect: [ : voteArray | self votesToString: voteArray ].
	pStream storeStringMap: 'votes'  value: tmp default: (Dictionary new).
! !

!QPollBarChart methodsFor: 'menu' stamp: 'bgf 8/29/2009 03:23'!
provokeVoteDialog: user from: pname
	"Signal the user's harness to open a voting dialog."
	self provokeVoteDialog: user suggesting: (self userVotes: user) from: pname! !

!QPollBarChart methodsFor: 'menu' stamp: 'bgf 8/29/2009 03:15'!
provokeVoteDialog: user suggesting: seeds from: requesterName
	"Signal the user's harness to open a voting dialog."
	user signal: #callForVote with: self with: seeds with: requesterName! !

!QPollBarChart methodsFor: 'menu' stamp: 'bgf 8/30/2009 00:46'!
pushPoll: eventOrAvatar
	"Ask for votes from anyone present who hasn't voted; 	
	 signal their avatars to trigger the client to raise a dialog."
	|avatar asked|
	avatar := eventOrAvatar avatar.
	asked := 0.
	self root allUsers 
		do: [ : u |
			(votes includesKey: (u userLogin)) ifFalse: [
				asked := asked + 1.
				self provokeVoteDialog: u from: (avatar publicName) ]].
	asked = 0 ifTrue: [ 
		 avatar signal: #alertMessage 	
				with: 'Everyone here has voted already.'.
	] ifFalse: [
		 avatar signal: #alertMessage 
				with: 'Asked ', asked,   ((asked > 1) ifTrue: [ ' people'] ifFalse: [' person']), ' to vote.'.
		QActivityCapture action: #CallForVotes 
					details: (asked asString) 
					forSession: ( avatar sessionHex).
	].! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 11/20/2009 21:28'!
reconcile
	"Make the votes, and only the votes, count."
	votes ifNil: [ ^ self ].

	votes clone keysAndValuesDo: [ : login : voteArray |
		| va |
		(voteArray isArray and: [ voteArray isEmpty not]) ifTrue: [
			va := self isSingleChoice 
						ifTrue: [ voteArray first: 1 ]
						ifFalse: [ voteArray select: [ : v | v > 0 and: [ v <= data size ]]]].
		va isEmptyOrNil 
			ifTrue: [ votes removeKey: login ] 
			ifFalse: [ votes at: login put: va ].
		].
	1 to: data size do: [ :i |
		data setValue: 0 atIndex: i
	].
	votes do: [ : voteIndexArray |
		voteIndexArray do: [ : idx | data setValue: (data valueAt: idx) + 1 atIndex: idx ]
	].! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 6/15/2010 13:05'!
restoreVotes: votesArray wasOpen: bOpen

	"Part of undoing a reset"

	votes := Dictionary new.
	votesArray do: [ : pair |
		votes at: pair first put: pair second.
		].
	self reconcile.
	self isOpenForVoting: bOpen.! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 8/18/2009 19:22'!
retractUserVote: avatar 
	"Retract all the ballots of this user."
	avatar userLogin ifNil: [ ^ self ].
	(self userVotes: (avatar)) do: [ : oldIdx |
		data setValue: (((data valueAt: oldIdx) - 1) max: 0) atIndex: oldIdx.
	].
	votes removeKey: (avatar userLogin) ifAbsent: [].
! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 8/30/2009 00:49'!
setBar: bar colorForIndex: i
	| clr |
	clr := self colorForBar: i.
	bar material
			diffuseColor: clr; 
			ambientColor: (clr * 0.75).
	bar checkAlpha.
! !

!QPollBarChart methodsFor: 'menu' stamp: 'bgf 8/29/2009 02:36'!
showContextMenu: event
	"Custom context menu; see addMenuActions..."
	self showMenu: event! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 19:32'!
stringToVotes: comSepString

	^ ((comSepString findTokens: ',') collect: [ : istr | istr asInteger ]) asArray! !

!QPollBarChart methodsFor: 'events' stamp: 'bgf 8/23/2009 18:00'!
titleDoubleClick: event
	QTextEditJob 
		edit: (self titleString)
		title: 'Enter Title for Poll' 
		forUser: event avatar 
		recipient: self 
		selector: #titleString:
	! !

!QPollBarChart methodsFor: 'transform' stamp: 'bgf 11/21/2009 10:17'!
updateShape
	| corner |
	super updateShape.
	title backgroundColor: (
			isOpenForVoting 
				ifTrue: [ 
					isSecretUntilClosed 
						ifTrue: [self class defaultSecretTitleBackground ]
						ifFalse: [self class defaultOpenTitleBackground ] ]
				ifFalse: [ self class defaultClosedTitleBackground ]
			).

	"Fit the resize-corner."
	corner := base extentY * 0.5.
	resizer grip 
		extent: corner@corner;
		translationX: (base extentX - corner *0.5) 
				   y:   (corner *0.5) - base extentY
				   z: (base extentZ * 0.6)! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 3/8/2010 14:44'!
user: avatar resetVotes: votes  
	"Votes is an array of choice indexes; this is support for undoing a vote-action"

	self isOpenForVoting ifFalse: [
		"We can no longer undo votes if the poll is closed by someone else."
		^ avatar signal: #alertMessage with: 'The poll has been closed.'.
	].

	self retractUserVote: avatar.
	votes do: [ :idx | 
		self addUserVote: avatar forIndex: idx.
	].

! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 3/13/2010 08:53'!
user: avatar setVotes: votes  atTime: t
	"Votes is an array of choice indexes"

	| oldVotes |

	self isOpenForVoting ifFalse: [
		^ avatar signal: #alertMessage with: 'The poll has been closed.'.
	].

	oldVotes := (self userVotes: avatar) clone.
	self retractUserVote: avatar.
	self lastModifiedTime: t.
	votes do: [ :idx | 
		self addUserVote: avatar forIndex: idx.
	].

	avatar recordAction: #undoForVotes:oldVotes:newVotes:
		    class: QPollBarChart
		    arguments: { self . oldVotes . (self userVotes: avatar) clone }! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 8/18/2009 21:04'!
userVotes: avatar
	"For the given user, return the array of integers represent his vote;
	 each element in the resulting array is an index into my chart-data.
	 For single-choice polls, the result will either be empty or an array of one integer."
	avatar ifNil: [ ^ nil ].
	^ votes at: (avatar userLogin) ifAbsent: [ #() ].
! !

!QPollBarChart methodsFor: 'accessing' stamp: 'bgf 11/20/2009 22:18'!
visibleValueAt: barIndex

	"Suppress drawing the bars if we're secret."
	(self isOpenForVoting and: [ self isSecretUntilClosed ]) 
	 ifTrue: [
		^ 0
	].
	^ data valueAt: barIndex! !

!QPollBarChart methodsFor: 'private' stamp: 'bgf 8/18/2009 20:00'!
votesToString: iarray
	"Return the integer-array as a comma-separated string,
	 suitable for writing as the string-value when persisting the vote table into PersistStream."

	"The most common case by far is a single-choice poll..."
	iarray ifEmpty: [^ ''].
	iarray size = 1 ifTrue: [ ^ iarray first asString ].

	^ String streamContents: [ : strm |
		| first | 	
		first := true.
		iarray do: [ : i |  
			first ifFalse: [ strm nextPut: $,].
			first := false.
			strm nextPutAll: (i asString) ]].! !

!QPollBarChart methodsFor: 'voting' stamp: 'bgf 8/26/2009 19:03'!
writeTranscriptOn: stream
	"For assembling text reports of note-collections."
	stream nextPutAll: 'Poll -- ', (self titleString), ' -- '.
	self lastModifiedTime ifNotNil: [
		stream nextPutAll: ' ('.
		stream nextPutAll: (QStyles dateStringFromUTC: self lastModifiedTime).
		stream nextPutAll: ')'.
	].
	stream cr.
	1 to: data size do: [ : i | 
		stream nextPutAll: ('  Choice ',  (i asString),  ':	', (self labelStringForBar: i), ':	', (data valueAt: i)).
		stream cr.
	].
	stream nextPutAll: 'End  --'.
	stream cr.
	! !

TGroup subclass: #QCampfire
	instanceVariableNames: 'particles'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QCampfire class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCampfire methodsFor: 'initialize' stamp: 'jcg 12/20/2007 02:57'!
initialize
	| ptxtr |
	super initialize.
	ptxtr := QMaterials flameMaterial asTFrame.
	particles := TParticleTxtr new initializeWithSize: 100.
	particles 
		rectangle: ptxtr;
		setLifetimeRange: (250 to: 600);
		setPositionRangeMin:(-1.0@0.0@-1.0) max: (1.0@1.5@1.0);
		setVelocityRangeMin:(-1.2@4.0@-1.2) max:(1.2@8.0@1.2);
		setAccelerationRangeMin:(0@-4@0) max:(0@0@0);
		startColor: #(1.0 1.0 1.0 0.99);
		endColor: #(1.0 1.0 1.0 0.0).
	self addChild: particles.! !

TGroup subclass: #QDisplayNiche
	instanceVariableNames: 'largeWall smallWall panel noteboard width depth height fraction thickness'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QDisplayNiche commentStamp: '<historical>' prior: 0!
A specialized furniture/construction element for project forums,
comprising two surface area walls at an oblique angle,
with a fixed display panel as part of the larger wall,
and a fixed noteboard as part of the smaller wall.

The niche walls themselves are not drop-target walls;
the fixed display panel is meant to be the operating display area
for the niche, not one of many.!

!QDisplayNiche class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDisplayNiche class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 11:56'!
addPropertyRows: propertyDialog forFrame: nicheref

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ :f | f get:#extent ];
			acceptBlock: [ :v3 :f | f future extent: v3 ];
			balloonHelp: 'Dimensions of the niche.';
	 	yourself	).

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Fraction for Panel';
			fetchBlock: [ :f | f  get: #fraction ];
			acceptBlock: [ :s :f  | f future fraction: s ];
			balloonHelp: 'Between 0.1 and 0.9.  The portion of width given to the left wall';
		yourself).

	super addPropertyRows: propertyDialog forFrame: nicheref.! !

!QDisplayNiche class methodsFor: 'defaults' stamp: 'bgf 7/24/2008 14:16'!
defaultExtent
	^ (15@10@6).! !

!QDisplayNiche class methodsFor: 'defaults' stamp: 'bgf 7/24/2008 17:51'!
defaultFraction
	"The ratio between large panel wall and smaller noteboard wall."
	^ 0.7! !

!QDisplayNiche class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:11'!
defaultIsFromTemplate
	"Reduces xml volume since these are often premade."
	^true ! !

!QDisplayNiche class methodsFor: 'defaults' stamp: 'bgf 6/21/2009 15:04'!
defaultIsPlacementLocked
	"Eases interactions if these are unlikely to be unlocked."
	^true! !

!QDisplayNiche class methodsFor: 'defaults' stamp: 'bgf 8/29/2009 13:55'!
defaultMaterial
	"For the wall slabs"
	^ (TMaterial new)
		texture: (QTextures find:'Content/Qwaq/TexturePalette/woodGradient.jpg');
		ambientColor: #(0.67 0.8 0.77) asVectorColor;
		diffuseColor: #(0.85 1.0 0.94 1) asVectorColor;
		specularColor: #(0.2 0.18 0.14 1) asVectorColor;
		yourself.	
! !

!QDisplayNiche methodsFor: 'transform' stamp: 'bgf 6/29/2009 19:58'!
cameraStanceFor: anAvatar
	^ QGotoButton frame: self cameraStanceFor: anAvatar! !

!QDisplayNiche methodsFor: 'testing' stamp: 'bgf 2/18/2009 21:45'!
canEditMaterial 
	^ true! !

!QDisplayNiche methodsFor: 'persist' stamp: 'bgf 7/24/2008 23:28'!
constructFrom: pStream

	| child |
	super constructFrom:pStream.
	self extent: (pStream readVector3: 'extent' default: (self extent)).
	self fraction: (pStream readNumber: 'fraction' default: (self fraction)).
	self material: (pStream readObject: 'material' default: (self material)).

	"We have two fixed children whose details are entirely persisted..."
	child := pStream readObject: 'noteboard' default: noteboard.
	child = noteboard ifFalse: [
		noteboard destroy.
		noteboard := child.
		smallWall addChild: noteboard.
	].

	child := pStream readObject: 'panel' default: panel.
	child = panel ifFalse: [
		panel destroy.
		panel := child.
		largeWall addChild: panel.
	].

	"For future variation once member-persisting is handled...
		pStream constructObject: 'noteboard' initialObject: noteboard.
		pStream constructObject: 'panel' initialObject: panel.
	"! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 16:50'!
defaultMaterial
	"Deliberately shared between large and small walls."
	^ largeWall material! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:17'!
depth
	^ depth! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 16:15'!
depth: aScalar

	"The distance from front to the inside corner; this is the z-depth of the niche."
	depth := aScalar.
	self shapeChanged.! !

!QDisplayNiche methodsFor: 'transform' stamp: 'bgf 7/26/2009 23:35'!
destinationStanceFor: anAvatar
	"Floor-standing front-gathering destination (we also have a camera stance)"
	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:16'!
extent
	^ (width@height@depth)! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:16'!
extent: aV3

	self width: aV3 x.
	self height: aV3 y.
	self depth: aV3 z.! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 15:10'!
fraction
	^ fraction! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 15:10'!
fraction: aScalar
	fraction := aScalar min: 0.9 max: 0.1.
	self shapeChanged.! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:17'!
height
	^ height! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 13:16'!
height: aScalar
	height := aScalar.
	self shapeChanged.! !

!QDisplayNiche methodsFor: 'initialize' stamp: 'bgf 7/24/2008 15:11'!
initialize
	super initialize.

	fraction := self class defaultFraction.
	thickness := 0.2.		"Thickness of the wall panels."

	largeWall := TCube new.
	largeWall material: self class defaultMaterial.
	self addChild: largeWall.

	smallWall := TCube new.
	smallWall material: (largeWall material).
	self addChild: smallWall.

	panel := QDisplayPanel withSimpleFrame.
	largeWall addChild: panel.

	noteboard := QNoteBoard new.
	smallWall addChild: noteboard.

	self extent: self class defaultExtent.! !

!QDisplayNiche methodsFor: 'testing' stamp: 'bgf 7/27/2008 19:55'!
isGroundBound
	^true! !

!QDisplayNiche methodsFor: 'transform' stamp: 'bgf 2/12/2010 14:31'!
localCenter
	| lc |
	lc := super localCenter.
	"Place an approaching avatar on the ground, almost."
	^ (lc x) @ (5.5*(self scale)) @ ((lc z) + (depth * 0.25)).! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 16:16'!
material
	"The same material is used for both faux-wall cubes."
	^ largeWall material.! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 16:16'!
material: aMaterial

	"The same material is used for both faux-wall cubes."
	largeWall material: aMaterial.
	smallWall material: aMaterial.! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 8/29/2009 13:58'!
noteboard
	"The QNoteboard on my right side."
	^ noteboard! !

!QDisplayNiche methodsFor: 'persist' stamp: 'bgf 7/28/2009 23:34'!
persistOn: pStream

	super persistOn: pStream.
	pStream storeVector3: 'extent' value: (self extent) default: (self class defaultExtent).
	pStream storeNumber: 'fraction' value: (self fraction) default: (self class defaultFraction).
	pStream storeObject: 'material' value: (self material) default: (self class defaultMaterial).

	"Persist the noteboard and panel because we want their content,
	 customization, etc persisted.   
	 But we do not want to persist their containing 'wall' elements which are private.
	 (It is inefficient that we have to destroy the original
	 if we reconstruct from persistence, but the persist model doesn't really have
 	 a clean way to 'recover from persistence' without this.)"

	[ 
	  self addChild: noteboard. "Must be parented, or some content may not persist properly."
	  pStream storeObject: 'noteboard' value: noteboard
	] ensure: [
	  smallWall addChild: noteboard.
	].

	[ 
	  self addChild: panel.	"Must be parented, or some content may not persist properly."
	  pStream storeObject: 'panel' value: panel
	] ensure: [
	  largeWall addChild: panel.
	].! !

!QDisplayNiche methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:55'!
privateScale: aFactor
	"Let my shape manager take care of it all"
	self extent: (self extent * aFactor).
	super privateScale: aFactor.! !

!QDisplayNiche methodsFor: 'menu' stamp: 'bgf 2/16/2010 10:50'!
reallyDelete: event
	"Since we build in a noteboard, we want to clean up all the attachments to all the interior parts."
	self root ifNotNilDo: [ : space |
		self allFramesDo: [ : f |
			(space stickersOn: f) do: [ : item |  item reallyDelete: event ]
		]].
	super reallyDelete: event.
! !

!QDisplayNiche methodsFor: 'hierarchy' stamp: 'bgf 2/12/2010 14:31'!
updateShape
	| adj1 adj2 theta1 theta2 longLen shortLen pe py |
	
	width := width max: 0.1.
	height := height max: 0.1.
	depth := depth max: 0.

	adj1 := width * fraction.
	theta1 := (depth / adj1 ) arcTan.
	adj2 := width * (1.0-fraction).
	theta2 := ( depth / adj2) arcTan.

	longLen := adj1 / (theta1 cos).
	largeWall localTransform: Matrix4x4 identity.
	largeWall extent: (longLen @ height @ thickness).
	largeWall translation: ( (adj2 *-0.5) @ (height * 0.5) @ 0);
			  addRotationAroundY: theta1 radiansToDegrees.

	shortLen := adj2  / (theta2 cos).
	smallWall localTransform: Matrix4x4 identity.
	smallWall extent: (shortLen @ height @ thickness).
	smallWall translation: ( (adj1*0.5) @ (height * 0.5) @ 0);
			 addRotationAroundY: theta2 negated  radiansToDegrees.

	"Adjust panel height so avatars stay on ground plane.
	 (The panel is a child of the wall,
	 so it's position is relative to mid-wall.   wall ht  / -2.0 is the ground height.
	 5.5 is working eye-height for average avatar."
	pe := (largeWall extent x*0.8) @ (height * 0.6).
	py := (((largeWall extent y)*-0.5) + (5.5 * (self scaleMemo))) min: (height * 0.6) max: (height * -0.2).
	panel extent: pe;
	         translationX: (largeWall extent x * 0.01) y: py z:  thickness * 1.03.

	panel isPlacementLocked: true.

	noteboard extent: (smallWall extent x - (thickness*0.5) - (2*noteboard borderWidth)) 
						@ (height - (2.1*noteboard borderWidth));
			   translationX: (thickness*0.25) y:  0 z:  thickness * 0.53.

! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 16:16'!
width
	"The distance across the front of the niche."
	^ width! !

!QDisplayNiche methodsFor: 'accessing' stamp: 'bgf 7/24/2008 16:16'!
width: aScalar
	
	"The distance across the front of the niche."
	width := aScalar.
	self shapeChanged.! !

TGroup subclass: #QDisplayPanel
	instanceVariableNames: 'sheet gotoButton handler menuButton startStopButton rolodex isFadeEnabled nextFade menuBtnHiliter startStopHiliter cardExtent ioIndicator dropbox'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QDisplayPanel commentStamp: '<historical>' prior: 0!
Manages a QDisplaySheet within a QPanelHandler,
to provide a framed, movable, sizable display tool
for 2-dimensional content.

This is the general-purpose resizeable content panel
for 2D content in the treefort and conferencing demos.

It works with QFileCard and the QDragAdapter protocol
so that rolodexes, display panels, and file cards
provide a rudimentary content organization and viewing scheme.

Although QDisplaySheet handles most drag/drop activity,
the panel also participates in drag/drop as a target, so that QFileCards
can be dropped on a completely closed panel
(which springs open to display their contents.)
!

!QDisplayPanel class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDisplayPanel class methodsFor: 'inspector' stamp: 'sn 3/8/2013 19:14'!
addPropertyRows: propertyDialog forFrame: panelref

	self addNameRowIfNeeded: propertyDialog forFrame: panelref.

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :f | f get:#extent ];
			acceptBlock: [:v :f  | f future extent: v ];
			validateBlock: [ : pt | 
				((pt isNil) or:[ (pt x <= 0) | (pt y <= 0)])  ifTrue: [
					self inform: 'Extent must be positive.'translateMe.
					false
				] ifFalse: [
					true
				]];
			balloonHelp: 'Dimensions of the display area.'translateMe;
	 	yourself	).
			
	propertyDialog addRow: (
		(QBooleanPropertyRow newForFlag: 'Fade Frame'translateMe)
			fetchBlock: [ :f | f get: #isFadeEnabled ];
			acceptBlock: [ :b : f  | f future setFadeEnabledFromInspector: b ];
			balloonHelp: 'Fading hides all border pieces.'translateMe withCRs;
		yourself).

	super addPropertyRows: propertyDialog forFrame: panelref.
! !

!QDisplayPanel class methodsFor: 'defaults' stamp: 'bgf 10/30/2008 17:30'!
cardExtent
	"Target extent of cards in the rolodex."
	^ (3.8@2.2) * (QPanelFrame defaultThickness)! !

!QDisplayPanel class methodsFor: 'undo-redo' stamp: 'bgf 4/2/2010 21:52'!
clearDisplay: dref oldCard: pickle
	"Record the undoable clear of a fixed-display."
	^ QBlockAction
		undoBlock: [ : avatar | dref future restoreCard: pickle forAvatar: avatar ]
		redoBlock: [ : avatar | dref future clearPanel: avatar ]
		description: 'Clear'
		requiredFrames: { dref }.! !

!QDisplayPanel class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QDisplayPanel class methodsFor: 'defaults' stamp: 'bgf 8/6/2009 23:29'!
defaultExtent
	^ 7.6 @ 5.7! !

!QDisplayPanel class methodsFor: 'defaults' stamp: 'bgf 6/2/2009 18:54'!
defaultIsTemplateExclusion
	"It seems common to keep display panels browsable"
	^ true! !

!QDisplayPanel class methodsFor: 'defaults' stamp: 'bgf 1/18/2008 23:17'!
minRecentDocCapacity

	"Even when narrow, I won't restrict my shelf to fewer than this man document cards."
	^ 10! !

!QDisplayPanel class methodsFor: 'undo-redo' stamp: 'bgf 4/3/2010 09:26'!
panelData: panelPair convertedFrom: prevPair inParent: parentRef
	"Undo support for 'Convert to Panel' operations in poster and fixed-display.
	 The 'pair' arguments are each a 2-array:
		first element objectID, second capturedFrameList.
	 prevPar captures the original poster or fixedisplay;
	 panelPair captures  the panel it was converted to."

	^ QBlockAction	
		undoBlock: [ : avatar | avatar island future 
						in: QDisplayPanel
						perform: #restore:removing:forAvatar:
						withArguments: { prevPair . panelPair . avatar }  ]
		redoBlock: [ : avatar | avatar island future 
						in: QDisplayPanel
						perform: #restore:removing:forAvatar:
						withArguments: { panelPair . prevPair . avatar }  ]
		description: 'Convert to Panel'
		requiredFrames: { parentRef }.! !

!QDisplayPanel class methodsFor: 'undo-redo' stamp: 'bgf 3/19/2010 11:29'!
projection: panelRef newSlaves: newSlaves  oldSlaves: oldSlaves

	"Generic support for undoable  displayPanel -> projectionScreen  setups.
	  The slaves arrays are arrays of pairs,   {   {displayPanel . projectionSlave} }
	  where nil on either side of the pair is an indication to clear the binding of the non-nil side."

	^ QBlockAction 
		undoBlock: [ panelRef island future 
			in: QDisplayPanel perform: #setProjections: withArguments: {oldSlaves} ]
		redoBlock: [ panelRef island future 
			in: QDisplayPanel perform: #setProjections: withArguments: {newSlaves} ]
		description: 'Projection'
		requiredFrames: (((newSlaves, oldSlaves) gather: [  : pair | pair ]) copyWithout: nil)! !

!QDisplayPanel class methodsFor: 'undo-redo' stamp: 'bgf 4/4/2010 10:57'!
restore: picklePair removing: curPair forAvatar: avatar
	| cur |
	" Undo support for 'Convert to Panel' operations in poster and fixed-display. "

	cur := avatar island lookup: curPair first.
	cur ifNil: [  self error: 'Cannot undo: replacement is absent.'].
	cur root ifNil: [  self error: 'Cannot undo: replacement has been removed.'].
	cur obliterateLater.

	picklePair second restoreForAvatar: avatar.
	
! !

!QDisplayPanel class methodsFor: 'undo-redo' stamp: 'bgf 3/19/2010 11:29'!
setProjections: masterSlavePairs
	"Undo support - set the master-slave pairings"
	masterSlavePairs do: [ : pair | 
		(pair second) 
			ifNotNil: [ : slave | slave master: (pair first ifNotNil: [ : display | display sheet] ) ]
			ifNil: [ pair first stopProjectorDisplay ] ].! !

!QDisplayPanel class methodsFor: 'undo-redo' stamp: 'bgf 4/16/2010 12:24'!
undoableClearHistory: capList inPanel: panelref

	^ QBlockAction 
		undoBlock: [ : avatar | avatar future undoDeleteFrames: capList ]
		redoBlock: [ : avatar | avatar future redoDeleteFrames: capList ]
		description: 'Clear History'
		requiredFrames: { panelref }! !

!QDisplayPanel class methodsFor: 'instance creation' stamp: 'bgf 1/3/2007 15:16'!
withFrame: aPanelFrame

	"Create a display panel with the given frame."
	^ self new withFrame: aPanelFrame! !

!QDisplayPanel class methodsFor: 'instance creation' stamp: 'bgf 1/3/2007 15:15'!
withSimpleFrame

	"Create a display panel with a simple rectilinear resizing frame."
	^ self new withSimpleFrame.
! !

!QDisplayPanel methodsFor: 'private' stamp: 'das 11/10/2008 08:45'!
abortPython

	"The stop-button has been pressed."
	sheet hasEmbeddedApp ifTrue: [
		sheet onAbort.
		"Take enough time to let the save control-s find its way upstream."
		(sheet childContent future: 1500) stopRunning.
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 11/12/2009 14:42'!
actionsList
	"This is the QInspector action list... advanced or rare operations."

	(handler thickness =  (QFlatFrame defaultThickness)) ifFalse: [
		^ { {#resetProportions . 'Reset Panel Frame'} }
	].
	^ #()
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 5/20/2008 20:59'!
activityDetails

	"ActivityCapture details string to summarize this display panel.
	 Use the file card activity details if possible;
	 lets us see correlation of go-to's with other document-related actions."
	self contentCard ifNotNilDo: [ :card|
		^ (self class name asString), ': ', (card activityDetails)
	].
	^ (self class name asString), ': ', self title.! !

!QDisplayPanel methodsFor: 'menu' stamp: 'sn 12/28/2012 20:13'!
addMenuActions: aMenu forEvent: event

	"Build the panel's menu (hung from the blue title-bar button, fired by the panel-handler."
	| sz |

	"The top is determined by whatever content we have open."
	sz := aMenu menuItems size.
	sheet contentCard content ifNotNil: [
		sheet contentCard addMenuActions: aMenu forEvent: event.
		"The file card inserts Properties... for itself.  We don't want that entry in the panel's menu."
		(aMenu menuItems detect: [ : item | item label = 'Properties...' ] ifNone: [nil])
		 ifNotNilDo: [ : propsItem |
			aMenu removeItem: propsItem.
		].
	].
	aMenu menuItems size > sz ifTrue: [ aMenu addSeparator ].

	"Windowing..."
	sheet hasEmbeddedApp ifTrue: [
		aMenu add: 'Floating Window...' translateMe target: self action: #showFloatingAppPane:.
	].	
	sheet contentCard ifNotNil: [
		aMenu add: 'Zoom to Window'  translateMe target: self  action: #zoomToWindow100:
	].
	aMenu menuItems size > sz ifTrue: [ aMenu addSeparator ].

	"Slave projection display"
	self addProjectorOptionsToMenu: aMenu.

	"Panel things."
	"aMenu addSeparator.
	(aMenu add: 'Share as Video Feed' target: self action: #toggleShareAsVideoFeedViaEvent:)
		checked: sheet camera notNil;
		enabled: true;
		separator: true."

	(aMenu add: 'Become Dropbox' translateMe target: self action: #beDropbox:)
		enabled: dropbox isNil.

	(aMenu addItemSpec: {'Clear Display' translateMe . #confirmClear: } target: self)
		enabled: (self canClear).

	(aMenu add: 'Clear Recent Documents' translateMe target: self action: #confirmClearHistory:)
		disabled: (rolodex cards size < (self hasContent ifTrue: [2] ifFalse: [1])).		

	self isPlacementLocked ifFalse: [
		aMenu addSeparator.
		aMenu addItemSpec: {'Delete Panel' translateMe . #confirmDelete:} target: self.
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 3/12/2010 08:04'!
addProjectorOptionsToMenu: aMenu

	"The menu for projecting to slave display screens.
	 Content depends on what's in the space
	 (slave screens are space resources)
	 and whether or not I am projecting already."

	|  currentSlave projectors |

	self root ifNil: [ ^ self ].
	sheet contents ifNil: [ ^ self ].
	projectors := (self availableSlaveScreens ifEmpty: [ ^ self ]).
	currentSlave := self slaveProjector.

	"Keep it simple. Offer stop if being projected, or else offer a choice of screens to start."
	currentSlave ifNotNil: [
		aMenu add: 'Stop Projector' target: self action: #stopProjectorDisplay: .
	] ifNil: [
		self hasContent ifTrue: [
			projectors do: [ : slave |
				aMenu add: 'Project on ', (slave menuTitle) 
				    target: self 
				    action: #displayToProjector:forEvent:
				    arguments: { slave }.
			]].
	].
	aMenu menuItems last separator: true.

	! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 2/17/2008 14:27'!
arrow
	^ gotoButton arrow! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 7/22/2009 13:51'!
asPlace
	"Return the QPlace description to bring the avatar into focus on this panel."
	^ (gotoButton asPlace)
		hasApp: (sheet notNil and: [ sheet hasEmbeddedApp ] );
		cameraSelector: #cameraStanceFor:;
		yourself! !

!QDisplayPanel methodsFor: 'annotations' stamp: 'sn 12/28/2012 20:17'!
asTopicRecord	
	"The record presenting me as a line-item in a topics-tool."
	self document ifNotNilDo: [ : doc |
		^ (QTopicRecord new)
			cardRef: self;
			author: (self document ctime who);
			color: (Color black);
			creationTime: (doc ctime time);
			lastModifiedTime: (doc mtime time);
			topic: ('Doc: ', doc shortName);
			nComments: (self root stickersAttachedTo: self) size;
		yourself
	].

	^ (QTopicRecord new)
			cardRef: self;
			author: '';
			color: Color black;
			creationTime: 0;
			lastModifiedTime: 0;
			topic: 'Empty Panel' translateMe;
			nComments: (self root stickersAttachedTo: self) size;
		yourself! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 7/16/2009 22:26'!
availableSlaveScreens

	"All slave displays currently registered in my space."
	^ ((self island resourcesAt: (QSlaveProjectionScreen resourceType)) 
		select: [ : sd | sd root = self root ]) 
		asArray 
		sort: [: a : b | a objectName <= b objectName ]
! !

!QDisplayPanel methodsFor: 'menu' stamp: 'hrs 3/11/2009 21:58'!
beDropbox: event

	"Make this the designated Dropbox for the event's avatar."
	(event notNil and: [event avatar notNil]) ifTrue: [
		self beDropboxFor: event avatar.
	].
! !

!QDisplayPanel methodsFor: 'events' stamp: 'ar 4/16/2011 17:11'!
beDropboxFor: anAvatarOrNil
	"Make this the designated Dropbox for the avatar, or for no one if anAvatarOrNil is nil.
	#findDropboxFor:among: defines what is recognized as a Dropbox,
	and #dropCard:at:for: is what uses them.
	
	IFF we are rooted:
	* anAvatarOrNil must be an avatar if not nil, and is used to undesignate any other Dropboxes 
		in the space.  This implementation ensures no more than one Dropbox per space.
	* An instruction card is added.  (If we're not rooted, we assume we're constructing from
		persistence and we'll either have a card or not in the usual way."

	| card |
	dropbox ifNotNil: [
		self removeChild: dropbox.
		dropbox destroy.
	].
	dropbox := anAvatarOrNil ifNotNil: [
		self root ifNotNilDo: [:root |
			"There can only be one. Kill the others."
			(root findDropboxFor: anAvatarOrNil among: anAvatarOrNil currentLandmarks)
				ifNotNilDo: [:old | old beDropboxFor: nil].
	
			"Give some instructions."
			card := QFileCard forText: 'The Dropbox is the default destination for cards that are not dropped to a specific landmark by the user that created them.

For example, this can be used when importing a document to a Forum that you are not in at the time.'.
			card content shortName: 'Dropbox-Instructions.txt'.
			card content ctime: Time utcSeconds who: 'support@openqwaq.com' what: 'Creation'.
			self displayCard: card.
			self embeddedApp ifNotNil: [ : app | app isDirty: true ].
					"So the next save will create the corresponding file."
		].
		
		QPlaceToken consistingOf: self named: 'Dropbox'.
	].
	self onTitleChanged.
	! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 12/21/2008 09:00'!
broadcastColor: col
	"Relays the avatar color of an active user back to the embedded-app,
	 so that floating displays can reflect that activity."
	sheet ifNotNil:[
		sheet broadcastColor: col.
	].! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 4/26/2009 13:49'!
cameraStanceFor: anAvatar
	^ QGotoButton frame: self cameraStanceFor: anAvatar! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 3/3/2009 20:56'!
canClear
	"Should the (panel-handler) Clear menu item be enabled?"
	^ self contentCard content notNil
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/17/2008 13:39'!
canMove
	"Is interactive moving via the title bar turned on?"
	^ handler canMove! !

!QDisplayPanel methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 14:17'!
canOpenCard: aFileCard

	"Is the card allowable as an argument to 'openCard:fromUser:
	 See QFileCard #canBeOpened"
	^ QDisplaySheet canAcceptDrop: aFileCard.! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/17/2008 13:39'!
canResize
	"Is interactive resizing (boundary-dragging) turned on?"
	^ handler canResize! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 3/20/2007 15:31'!
cancelFade
	"suppress any pending fade schedule"
	nextFade := nextFade + 1.
	nextFade > 10000 ifTrue: [
		nextFade = 0.	"Just in case, you know..."
	].! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/17/2008 12:29'!
cardExtent
	"The desired basicExtent of my recentDocuments shelf"
	^ cardExtent! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/17/2008 12:29'!
cardExtent: aPoint
	cardExtent := aPoint.
	self shapeChanged.! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 7/28/2009 23:40'!
checkProjectionForRootChange
	"When reparenting, give my display screen notice that they might lose the master..."
	self availableSlaveScreens do: [ : slave |
		slave master = sheet ifTrue: [
			slave future checkRoot
		]].
	! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 9/4/2007 12:08'!
childrenForBrowser
	^ rolodex childrenForBrowser! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 4/16/2010 12:28'!
clearHistory: evtOrAvatar
	"Toss the recent-documents history, save for the currently open card."
	| cards |
	cards := (rolodex cards copyWithout: (self contentCard)).
	evtOrAvatar avatar
		recordAction: #undoableClearHistory:inPanel:
		class: QDisplayPanel
		arguments: {  (evtOrAvatar avatar captureFrames: cards) .  self }.
	cards do: [ : cd | cd destroy ].
! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 4/2/2010 21:54'!
clearPanel: evOrAvatar
	"Make the display empty."
	| writer |
	sheet isEmpty ifTrue: [ ^ self ].

	(evOrAvatar notNil and: [evOrAvatar avatar notNil])
	 ifTrue: [
		writer := TSceneWriter new.
		writer preserveObjectIDs: true.
		evOrAvatar avatar recordAction: #clearDisplay:oldCard:
				class: QDisplayPanel
				arguments: { self . (writer storeSceneAsZip: (sheet contentCard)) }.
		self stopRunning: evOrAvatar action: #StopAppClearPanel.
	].
	sheet future clear.
	self unhilite.
! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 6/13/2009 10:29'!
closeCard: aFileCard forEvent: anEvent
	"Notice from a file card that it is being removed while I'm showing it."
	(self hasCardOpen: aFileCard) ifTrue: [
	 	self clearPanel: anEvent.
	].! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 10/12/2006 12:35'!
closePortal

	"Close the TPortal of the display sheet."

	sheet closePortal.

! !

!QDisplayPanel methodsFor: 'menu' stamp: 'sn 12/28/2012 20:19'!
confirmClear:event
	"Ask the user to confirm clear of the panel.  
	 First stage action for the clearPanel command."

	event avatar ifNil: [ ^ self ].
	self hilite: (QColors confirmDeleteColor) for: 600.

	event avatar confirm: 'Clear Display (Close Contents) ?' translateMe
		selector: #clearPanel:
		target: self
	
! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 4/16/2010 12:17'!
confirmClearHistory: event
	"Ask the user to confirm clearing the history rolodex."
	| prompt |
	event avatar ifNil: [ ^ self ].

	prompt := (self hasContent 
					ifTrue: ['Remove all but the open document?']
					ifFalse: [ 'Clear the recent documents tray?']).

	event avatar confirm: prompt
				selector: #clearHistory:
				target: self.
! !

!QDisplayPanel methodsFor: 'persist' stamp: 'bgf 3/15/2010 16:37'!
constructFrom: pStream
	| tmp cards frame locked |
	super constructFrom: pStream.

	frame := pStream readObject: 'frame' default: nil.
	frame ifNotNil: [ self withFrame: frame ] ifNil: [ self withSimpleFrame ].

	"Unpack the inline persistence of the current content card..."
	tmp := (QFileCard new).
	tmp constructContentFrom: pStream.
	tmp isRetiredDemoContent ifTrue: [
		tmp clear
	].

	self displayCard: tmp from: (pStream sessionID).
	self cardExtent: (pStream readPoint: 'cardExtent' default: (self class cardExtent)) .
	"read maxExtent first because it constraints extent."
	self maxExtent: (pStream readPoint: 'maxExtent' default: self maxExtent).
	self minExtent: (pStream readPoint: 'minExtent' default: self minExtent) .
	self extent: (pStream readPoint: 'extent' default: (self extent)).

	cards := pStream readSequence: 'cards' default: #().
	cards do: [ :card|
		(card content = tmp content) ifTrue: [
			card shallowCopyFrom: tmp
		].
		card isRetiredDemoContent ifFalse: [
			rolodex insertCard: card.
		] ifTrue: [
			card destroy
		]].

	pStream preserveObjectIDs ifTrue: [
		(pStream readString: 'sheetOID' default: nil) ifNotNil: [ : hex | 
			sheet objectID: (TObjectID readHexFrom: hex)
		]].

	self rolodexVisible: (pStream readBool: 'rolodexVisible' default: (self rolodexVisible)).
	self title: (pStream readString: 'title' default: self defaultTitle).
	self isFadeEnabled: (pStream readBool: 'isFadeEnabled' default: false).

	"isPlacementLocked was isEmbedded until 2.0.18; allow it to control if present."
	locked := (pStream readBool: 'isEmbedded' default: (self isPlacementLocked)).
	self isPlacementLocked: locked.

	(pStream readBool: 'dropbox' default: false) ifTrue: [
		self beDropboxFor: 'unpickle'
	].
! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 12/10/2006 17:20'!
consumeDrop: delivery avatar: avatar selection: aTXSel

	"Drops pass through to the display sheet"

	^ sheet consumeDrop: delivery avatar: avatar selection: aTXSel! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/4/2009 00:04'!
contentCard
	"The filecard (from my rolodex) representing the current display contents."
	^ (rolodex detect: sheet contentCard) 
		ifNil: [ sheet contentCard ]! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:09'!
defaultIsAtomic
	^ false! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'hrs 3/11/2009 20:24'!
defaultTitle
	"The title text we display automatically based on our current card content;
	 the actual title text can be overridden in software, although it will
	 be updated when the content changes."

	(sheet isNil or: [sheet contentCard content isNil]) ifTrue: [
		^ self titleBase
	].
	^ sheet title! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 4/26/2009 15:04'!
destinationExtent
	"The x-y extent of the panel as a goto destination... 
	 adds some slop to keep controls away from the edge."

	| adjust |
	adjust := 2.4.
	^ (handler ifNotNil: [ handler extent  + (0@(self cardExtent y*adjust) ) ]
			    ifNil: [ sheet extent "Safety fallback" ])
		! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 7/26/2009 23:36'!
destinationStanceFor: anAvatar

	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self
! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 9/9/2008 16:29'!
destroy

	"Make sure portals and embedded apps are cleared right away."

	self stopProjectorDisplay.
	sheet destroy.
	self stopAllScripts.
	rolodex clear.
	super destroy.! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 10/4/2007 09:48'!
displayCard: aCard
	"Take the content of the given QFileCard as my display."
	sheet displayCard: aCard.
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'das 9/2/2008 10:59'!
displayCard: aCard from: userID
	"Take the content of the given QFileCard as my display, 
	 on behalf of the given user.   The avatar parameter identifies the principal,
 	 and this form also indicates that the action originated from a live user request."
	sheet displayCard: aCard from: userID
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'das 2/1/2009 18:37'!
displayCard: aCard fromUser: anAvatar
	"Take the content of the given QFileCard as my display, 
	 on behalf of the given user.   The avatar parameter identifies the principal,
 	 and this form also indicates that the action originated from a live user request."
	QLogger panelLog.
	sheet displayCard: aCard fromUser: anAvatar
! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 3/13/2010 09:33'!
displayToProjector: slaveScreen forEvent: anEvent
	| oldMasterSlaves |
	slaveScreen ifNil: [ ^ self ].
	slaveScreen root ifNil: [ ^ self ].
	oldMasterSlaves := self stopProjectorDisplay.
	QActivityCapture action: #ProjectorDisplayOn details: (self activityDetails).

	"Add the previous master of the given slave, so that undo works properly."
	oldMasterSlaves := oldMasterSlaves  
						copyWith: {(slaveScreen master ifNotNil: [ : dispsheet | dispsheet parent]) . slaveScreen }.

	slaveScreen master: (self sheet).
	anEvent avatar
		recordAction: #projection:newSlaves:oldSlaves:
		class: QDisplayPanel	
		arguments: { self .   
					{ { self . slaveScreen } } . 
					oldMasterSlaves }! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 7/14/2007 21:12'!
doFade: fadeRequest
	"If fading is still desired, vanish the frame parts now."
	(isFadeEnabled & (fadeRequest = nextFade)) ifTrue: [
		self doHideFrame.
		nextFade := nextFade +1.
	].! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 1/18/2008 02:22'!
doHideFrame

	"make the frame pieces invisible."
	handler frame visible: false.
	menuButton visible: false.
	rolodex visible: false.
	startStopButton visible: false.
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 8/1/2009 16:21'!
document
	"The QAbstractDocument I currently present, if any, or nil if I'm not showing one."
	^ sheet document ! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 8/19/2009 12:38'!
drag: cornerPt corner: compassPt delta: delta
	"Protocol relay from my resize-corner handler to my panel handler's 
	 edge-drag resize handling.  
	 The FlatFrame corner resizer must have a TFrame as its luggage (that is, me)
	 and it will resize by calling drag:corner:delta: since I support it.
 
	 But we want the panel handler to deal with resize, so pass the drags along.

	 (That allows, for example, DisplayStands to keep their
 	 feet on the ground when sizing-corner is dragged... the panel handler 
	 repairs the stand height to match vertical changes in my bottom edge.)" 

	handler drag: cornerPt corner: compassPt delta: delta! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 8/16/2009 01:51'!
dragAdapter
	^ handler motionAdapter! !

!QDisplayPanel methodsFor: 'annotations' stamp: 'bgf 8/1/2009 16:00'!
editByUser: replica
	"Protocol infill for treating panels as attachments to topic collectors;
	 this is the response to the Topic Tool 'Open' (and topic-list double-click) command."
	sheet hasEmbeddedApp ifTrue: [
		 replica signal: #floatForEmbeddedApp with: (sheet childContent)
	] ifFalse: [
		replica goToPlace: (self asPlace)
	].! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 12/20/2008 11:16'!
embeddedApp
	"If an embedded app is showing, return it; otherwise nil."
	^ sheet embeddedApp! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 1/2/2007 17:34'!
extent
	"The x-y extent of the panel."

	^ handler ifNotNil: [ handler extent ] ifNil: [ nil ].! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 1/2/2007 17:11'!
extent: newExtent

	"Set the current extent to the given rectangle, 
	 after enforcing the maximum extent bounds."

	handler ifNil: [ ^ self ].
	self extent ~= newExtent ifTrue: [
		handler extent: newExtent.
	].
	self shapeChanged.! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 9/21/2006 12:02'!
frame
	"The QFlatFrame around the display area."
	^ handler resizeFrame! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 5/28/2007 09:08'!
getKeyboardControl: avatar

	self signal: #getKeyboardControl with: avatar.
! !

!QDisplayPanel methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 12:49'!
hasCardOpen: aFileCard
	"Is this filecard my current open content?"
	^ self contentCard 
		ifNotNilDo: [ : cd | cd kind = aFileCard kind and: [ cd content = aFileCard content ]]
		ifNil: [ false ]! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/19/2008 18:08'!
hasContent
	"Is there content open in the panel?"
	^ self contentCard ifNotNilDo: [ : cd | cd content notNil ] ifNil: [ false ]! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 10/28/2006 16:56'!
height

	"The current total height of the display panel, including edges."
	^ self extent y + (handler resizeFrame thickness * 2)! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 9/4/2007 16:08'!
historyRolodex
	^ rolodex! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 4/30/2009 09:37'!
holdsDocument: aSharedDocument

	"Locate-support for the inspector... return true if I currently display the given document or if my shelf does."
	^ (self contentCard content = aSharedDocument) 
	    or: [rolodex ifNotNil: [rolodex holdsDocument: aSharedDocument]]! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'sn 1/2/2013 14:45'!
initialize
	"Create the basic display panel - without handler or frame.
	 Use 'withSimpleFrame' to complete initialization."

	super initialize.

	isFadeEnabled := false.

	nextFade := 0. 	"Serial counter to match fade requests."
	cardExtent := self class cardExtent.

	handler := QPanelHandler forAssembly: self withFrame: nil.
	handler extent:6@4.5;
		minExtent: 2@1;
		maxExtent: 500@500;		"Big as a campus... nominal numeric guard for resizing."
		menuLabel: 'Panel' translateMe;			
		noRotation.

	gotoButton :=  QGotoButton new.
	self addChild: gotoButton.
	gotoButton label: 'Display Panel' translateMe;
			     category: 'Panels';
			     suggestedPOV: 'First Person'.
	gotoButton disable.

	menuButton := TRectangle new.
	menuButton material: (QMaterials menuButtonRectangleMaterial).
	self addChild: menuButton.
	menuButton doesClickSelect: false.
	self addChild: (startStopButton := QStartStopButton new).
	startStopButton doesClickSelect: false.
	startStopButton disable.

	"MouseOverHiliters use only by scripting weak refs... to keep them alive,
	 we have to keep explicit references. 
	 That's the  purpose for my hiliters variables."
	menuBtnHiliter := QMouseoverHiliter forFrame: menuButton.
	menuBtnHiliter floaterString: 'Panel Menu' translateMe.
	menuBtnHiliter overColor: QColors controlOverColor.
	startStopHiliter :=  QMouseoverHiliter forFrame: startStopButton.
	startStopHiliter overColor: (Color orange asVectorColor alpha: 0.4).

	rolodex := QDocumentShelf new.
	rolodex axis: #horizontal.
	rolodex numVisibleSlots: nil.	"Until laid out."
	rolodex numSlots: nil.		"Until laid out."
	rolodex infoMessage: 'Recent Documents' translateMe .
	rolodex borderMaterial: (QFlatFrame defaultMaterial).
	rolodex baseMaterial: (QMaterials greyMaterial).
	self addChild: rolodex.

	sheet := QDisplaySheet new.
	self addChild: sheet.

	sheet isDropEnabled: false.	"Route drops through me (the container.)"
	gotoButton destination: self.

	ioIndicator := QAppTrafficIndicator new.

	self startScript: #onContentChanged when: {sheet . #contentChanged}.
	self startScript: #onTitleChanged when: {sheet . #titleChanged}.
	self startScript: #onDirtyChanged when: {sheet . #dirtyChanged}.
	self runScript: #getKeyboardControl: when: {sheet. #getKeyboardControl}.
	self runScript: #loseKeyboardControl when: {sheet. #loseKeyboardControl}.
	self startScript: #noFade when: {sheet . #pointerDown}.
	self startScript: #onRolodexChanged when: {rolodex . #contentChanged }.
	self onContentChanged.
	self shapeChanged.

	"Reassert with all the pieces in place."
	self class = QDisplayPanel ifTrue: [
		self isPlacementLocked: (self class defaultIsPlacementLocked).
	].
	sheet setEventMask.
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 7/31/2009 04:42'!
isComponent
	^ self shouldPersist! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 9/27/2006 19:55'!
isDropTarget: payload

	"We can accept anything the display sheet can present."
	^ QDisplaySheet canAcceptDrop:payload! !

!QDisplayPanel methodsFor: 'testing' stamp: 'bgf 12/21/2008 20:23'!
isEmpty
	"Is the panel without any content?"
	^ sheet isEmpty! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/20/2007 11:55'!
isFadeEnabled
	"Does the frame vanish when the display panel has no pointers on it?"
	^ isFadeEnabled! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 10/2/2007 15:32'!
isFadeEnabled: bool
	"Does the frame vanish when the display panel has no pointers on it?"
	isFadeEnabled := bool.
	bool ifTrue: [
		self scheduleFade.
		self startScript: #noFade when: { sheet . #pointerDown }.
		self startScript: #scheduleFade when: { self . #pointerLeave }.

		"Don't permit inter-frame motion to prematurely fade."
		self startScript: #cancelFade when: {self . #pointerOver }.
		self startScript: #cancelFade when: {sheet . #pointerOver }.
		self startScript: #cancelFade when: {handler frame . #pointerOver}.
	] ifFalse: [
		self stopScript: #noFade.
		self stopScript: #scheduleFade.
		self stopScript: #cancelFade.
		self noFade.
	].
	sheet setEventMask.! !

!QDisplayPanel methodsFor: 'testing' stamp: 'bgf 12/20/2008 16:50'!
isListedLandmark

	"Should this panel be included in automatic places-menu generation?"
	^ self contentCard content notNil and: [ self visible ]
! !

!QDisplayPanel methodsFor: 'annotations' stamp: 'bgf 8/1/2009 16:00'!
isListedTopic
	"When attached to a note-collector (e.g. a brainstorming wall), I participate in annotation stacking
	 and can appear in the topics list (shown by the document I display.)"
	^true
	! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 6/1/2009 12:42'!
isPlacementLocked: aBool
	"Whether the menu offers deletion and size controls.
	 If true, the panel is made immovable and resize is suppressed,
	 and those options are taken out of the normal panel menu.

	 For use by aggregating components, which must set this up on construction.
	 Can also be used by scene-creators to discourage rearrangement of panel setups,
	 using the property sheet."

	super isPlacementLocked: aBool.
	aBool ifTrue: [
		self panelHandler disableMove; disableResize.
	] ifFalse: [
		self panelHandler enableMove; enableResize.
	].

	"Ensure the come-hither event masks are intact."
	self resetClickHandler.! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 5/31/2009 19:57'!
isReparentable
	"Usually reparentable freely, unless marked as embedded in a fixed component."
	^ (super isReparentable) and: [ self isPlacementLocked not ]! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'hrs 3/11/2009 20:26'!
labelForBrowser
	(self title = self titleBase) ifTrue: [
		^ self titleBase
	] ifFalse: [
		^ self titleBase, ' - ', self title
	].
! !

!QDisplayPanel methodsFor: 'annotations' stamp: 'bgf 8/2/2009 19:42'!
lastModifiedTime
	"The topic tool sorts summaries using this."
	^ self document ifNotNil: [ : doc | doc mtime time ] ifNil: [ 0 ]! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 2/11/2010 10:58'!
localCenter

	"Fine-tunes the goto button destination to keep the title bar in view.
	 (The sheet is centered on the origin.)"

	^  self translation - (0@(self cardExtent y/4 * (self scaleInMatrix))@0)! !

!QDisplayPanel methodsFor: 'events' stamp: 'das 4/30/2007 14:48'!
loseKeyboardControl

	self signal: #loseKeyboardControl.
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 10/28/2006 16:52'!
maxExtent
	^ handler maxExtent! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 10/28/2006 16:59'!
maxExtent:  xY
	
	handler maxExtent: xY.
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 9/20/2006 22:42'!
maximize
	self frame maximize.! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 9/14/2007 21:25'!
menuStyle
	^ QStyles menuStyle! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 3/17/2009 02:43'!
menuTitle
	"Don't mention the content, as this title is for menus
	 that operate on the panel as furniture."
	^ 'Panel'
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 1/5/2007 18:43'!
minExtent
	^ handler minExtent! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 8/16/2007 21:40'!
minExtent:  xY
	
	handler minExtent: (xY max: (1@(2*self frame thickness))).
! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 1/18/2008 02:22'!
noFade
	"Clear the faded state, and suppress any forthcoming fade."
	
	self cancelFade.
	handler frame visible: true.
	menuButton visible: true.
	startStopButton visible: true.
	rolodex visible: true.
	
	sheet setEventMask.! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 9/21/2006 12:01'!
noMotion
	"Disallow interactive moves for this panel.
	 This is an initialization choice that discards the motion handler entirely."
	handler noMotion! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 3/5/2010 18:28'!
objectID: oid
	super objectID: oid.
	gotoButton ifNotNil: [ gotoButton destination: self ].! !

!QDisplayPanel methodsFor: 'events' stamp: 'sn 12/28/2012 20:24'!
onContentChanged
	"The display sheet content has changed."
	| currentCC |
	currentCC := sheet contentCard.

	"Content card becomes my deck's first card, unless there's no content."
	(sheet contentCard isNil or: [ sheet contentCard content isNil ]) ifTrue: [
		handler menuLabel: 'Panel: Empty' translateMe.
	] ifFalse: [
		rolodex firstCard: 
			((rolodex detect: (currentCC)) 
				 ifNil: [QFileCard new shallowCopyFrom: (currentCC)]).
		handler menuLabel: (currentCC kindSummary).
	].	

	self onTitleChanged.
	self updateStartStop.
	self stopScript: #startRunning:.
	self stopScript: #stopRunning:.
	self stopScript: #updateStartStop.

	self embeddedApp ifNotNilDo: [ :app|
		app isRemote ifTrue: [ 	
			self addChild: ioIndicator. 
			ioIndicator embeddedApp: app
		] ifFalse: [
			self removeChild: ioIndicator.
			ioIndicator embeddedApp: nil
		].
		"Connections between the content running state and the start/stop button:"
		self startScript: #stopRunning: when: { startStopButton . #stop }.
		self startScript: #startRunning: when: { startStopButton . #start }.
		self startScript: #updateStartStop when: { sheet childContent . #isRunningChanged }.
	] ifNil: [
		self removeChild: ioIndicator.
		ioIndicator embeddedApp: nil.
	].

	"We don't have good emphasizers, full-brighting the content card at least gives a hint."
	rolodex cards do: [ : card | card bright: (card = sheet contentCard) ].

	"Provoke a refresh in any tools watching the collectors I'm attached to..."
	self signal: #contentsChanged.
	! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 3/25/2010 10:46'!
onDirtyChanged
	handler ifNotNil: [
		handler frame titleColor: (
			sheet isDirty 
				ifTrue: [  Color r: 0.88 g: 0.88 b: 0.7 ] 
				ifFalse: [ Color white ] )
	]! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 5/26/2008 12:15'!
onRolodexChanged

	"When the card set in the rolodex changes"
	rolodex visible: true.
! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 11/21/2009 01:14'!
onTitleChanged
	"The display sheet content has changed."
	sheet contentCard content ifNotNilDo: [:ct| 
		self title: sheet title.
		gotoButton category: (sheet contentCard panelPlaceCategory).
		handler menuLabel: (sheet contentCard kindSummary).
	] ifNil: [
		self title: self titleBase.
		gotoButton category: 'Panels'.
		handler menuLabel: 'Panel: Empty'.
	].
	
! !

!QDisplayPanel methodsFor: 'recent-documents' stamp: 'das 2/1/2009 18:37'!
openCard: aFileCard fromUser: anAvatar

	QLogger panelLog.
	"Generic protocol for managing a rolodex of recent content."
	^ sheet  displayCard: aFileCard fromUser: anAvatar
! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 8/5/2006 18:52'!
openPortal

	"Open the TPortal of the display sheet."

	sheet openPortal.

! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 12/20/2007 13:22'!
panelHandler
	"My QPanelHandler"
	^ handler! !

!QDisplayPanel methodsFor: 'persist' stamp: 'bgf 3/15/2010 16:42'!
persistOn: pStream

	super persistOn: pStream.

	pStream storePoint: 'cardExtent' value: self cardExtent default: (self class cardExtent).
	pStream storePoint: 'extent' value: self extent default: 6@4.5.
	pStream storePoint: 'maxExtent' value: self maxExtent default: (8@6).
	pStream storePoint: 'minExtent' value: self minExtent default: (0@0).
	pStream storeString: 'title' value: self title default: self defaultTitle.
	pStream storeBool: 'isFadeEnabled' value: self isFadeEnabled default: false.
	self isPlacementLocked ifTrue: [ 
		pStream storeBool: 'resizable' value: false.	"Pre 2.0.18"
		pStream storeBool: 'movable' value: false.
		pStream storeBool: 'isEmbedded' value: true.
	].
	
	sheet contentCard persistContentOn: pStream.
	pStream preserveObjectIDs ifTrue: [  
		"Undo-redo of drops is much easier if sheet ID's survive panel undo/redo"
		pStream storeString: 'sheetOID' value: (sheet objectID hex).
	].

	pStream storeObject: 'frame' value: handler frame default: nil.

	pStream storeSequence: 'cards' value: (rolodex cards) default: #().
	pStream storeBool: 'rolodexVisible' value: (self rolodexVisible) default: true.
	pStream storeBool: 'dropbox' value: dropbox notNil default: false.! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 8/1/2009 17:52'!
photographerClass
	^ QDirectPhotoJob! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 1/6/2007 03:16'!
pinEdges: listOfCompassPtSymbols

	"Pin (disable interactive drag/resizing) of the named edges,
	 the names are #north #south #east #west."

	handler pinEdges: listOfCompassPtSymbols! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 1/1/2009 13:04'!
placeMenuActions: anAvatar
	"Label-symbol pairs for the Places context menu."
	| list |
	list := OrderedCollection new.
	sheet hasEmbeddedApp ifTrue: [
		list add: { 'Floating Window...' . #showFloatingAppPane: } .
	].
	sheet contentCard content ifNotNil: [ 
		list add: { 'Zoom to Window' . #zoomToWindow100: } .
	].
	^ list asArray.
! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 3/25/2010 10:42'!
pointerEnter: event

	event avatar ifNil: [ ^ self ].

	self parent = self root ifTrue: [ gotoButton enable ].

	"If the frame isn't set to fade, display an info message describing the content.
	 We omit this when the frame is faded to reduce busy-ness;
	 people use faded frames for decorative content."

	handler frame visible ifTrue: [
		sheet contentCard content ifNotNil: [ 
			| msg | 
			msg := sheet contentCard shortSummary.
			sheet isDirty ifTrue: [
				msg := msg, ' * Modified'
			 ].
			 event avatar signal: #statusMessage with: msg
		].
	].! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 8/5/2006 18:54'!
pointerLeave: event

	"Clear out all pointer interaction state."

	gotoButton disable.! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 8/5/2006 18:57'!
portal

	^ sheet portal.! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/19/2009 15:39'!
positioningBox
	"For materialization, we want to see the display even if we're at exact wall-distance 
	 for it... to be sure, use only the inner bounds to figure distances so we err too close.
	 Helps prevent panel-behind-wall problems."
	| box |
	box := sheet boundingBox.	
	box min: (box min x - 0.2) @ (box min y -0.1) @ (-0.1).
	box max: (box max x + 0.2) @ (box max y) + (0.1)  @ 0.2.
	^ box
! !

!QDisplayPanel methodsFor: 'printing' stamp: 'jcg 3/7/2007 22:42'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ':'''.
	aStream nextPutAll: self defaultTitle.
	aStream nextPutAll: ''''.! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:55'!
privateScale: aFactor

	"Unlike x-y resizing, scaling the panel scales its decorations.
	 This lets you use the edit box to make very small panels."
	nomScale := nomScale * aFactor.
	self cardExtent: (self cardExtent * aFactor).
	gotoButton scale: aFactor.
	handler scale: aFactor.
	self shapeChanged.! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 3/12/2010 11:08'!
projectOnSlaves: arrayOfSlaves
	"Undo-redo support"
	arrayOfSlaves do: [ : slave | slave master: (self sheet) ]! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/4/2009 00:07'!
pyAppName: aString
	[
		"NB these are probably redundant.  The file-cards are distinct,
	 	 but both share reference to the underlying document."
		sheet contentCard content pyAppName: aString.
		self contentCard content pyAppName: aString.

	] on: Error do: [ : ex |
		"I fear timing risks since these calls come back in from python,
	 	 but a clickety user may have dropped something non-python 
		 into the display by the time we get back."
		QLogger logWarning: 'pyAppName failed. ', (ex description).
	].
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 11/21/2009 01:14'!
pyDocName: aString
	[
		"NB these are probably redundant.  The file-cards are distinct,
	 	 but both share reference to the underlying document."
		sheet contentCard content docName: aString.
		self contentCard content docName: aString.
	] on: Error do: [ : ex |

		"I fear timing risks since these calls come back in from python,
	 	 but a clickety user may have dropped something non-python 
		 into the display by the time we get back."
		QLogger logWarning: 'pyDocName failed. ', (ex description).
	].
	handler menuLabel: (sheet contentCard kindSummary).
! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/4/2009 00:07'!
pyStartupOwner: aUserLogin
	"set the startup owner for a Python app - if it isn't this guy, you can't run the app"
	[
		"NB these are probably redundant.  The file-cards are distinct,
	 	 but both share reference to the underlying document."
		sheet contentCard content startupOwner: aUserLogin.
		self contentCard content startupOwner: aUserLogin.

	] on: Error do: [ : ex |
		"I fear timing risks since these calls come back in from python,
	 	 but a clickety user may have dropped something non-python 
		 into the display by the time we get back."
		QLogger logWarning: 'pyStartupOwner failed. ', (ex description).
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 8/1/2009 15:58'!
reallyDelete: event
	"Implementation of the menu Delete Panel command, which destroys me."

	"Forget attachment links ASAP (analogous to un-rooting early)"
	self root ifNotNil: [ self root detach: self ].
	(event notNil and: [event avatar notNil]) ifTrue: [

		QActivityCapture action: #Delete 
						 details: (self activityDetails)
						 forSession: (event avatar sessionHex).

		self stopRunning: event action: #StopAppDeletePanel.
		sheet future clear.
	].
	super reallyDelete: event.! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 4/3/2010 09:28'!
recordConversion: prePair forAvatar: avatar
	"Support for undoable convert-to-panel operations.
	 This is a common worker routine called by convert to panel methods.
	 This method is future-called, 
	 so that by the time we capture the panel's state here, it is wearing its content."

	| postPair |

	postPair := { self objectID . ( avatar captureFrames: { self }) }.

	avatar recordAction: #panelData:convertedFrom:inParent:
				class: QDisplayPanel
				arguments: { postPair . prePair . (self parent) }.! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 1/19/2010 16:06'!
resetClickHandler

	"Reinforce the title bar click event handler (needed after some disables to keep the masking in place)."
	self stopScript: #titleClick:.
	handler frame ifNotNil: [
		self startScript: #titleClick: when: { handler frame northBar . #pointerClick }.
		handler frame northBar doesClickSelect: false.
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 2/17/2008 13:19'!
resetProportions
	"Restore the rolodex and panel-handler thickness to default."

	handler thickness: (QFlatFrame defaultThickness).
	self cardExtent: (self class cardExtent).
	self shapeChanged.! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 4/2/2010 21:37'!
restoreCard: pickle forAvatar: aUser
	sheet replayDrop: pickle forAvatar: aUser! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/27/2007 12:00'!
rolodexVisible
	^ rolodex parent = self! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 2/27/2007 12:00'!
rolodexVisible: aBool
	aBool ifTrue: [self addChild: rolodex] ifFalse: [self removeChild: rolodex].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'hrs 9/30/2009 16:11'!
root: aSpaceOrNil

	"Clear any projector connections when I'm uprooted, just in case."
	| |
	self root = aSpaceOrNil ifTrue: [ ^ self ].
	self checkProjectionForRootChange.
	super root: aSpaceOrNil.
	
	aSpaceOrNil ifNil: [^self].
	self root name: self uniquelyAmongClass: QDisplayPanel.! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 3/20/2007 15:33'!
scheduleFade
	"Trigger a fade after the delay."
	|  |
	self cancelFade. "Any previous requests no longer obtain."
	(self future:2000) doFade:nextFade.

	"If nextFade stays unchanged, we'll fade in 2sec."! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 2/1/2009 14:33'!
setFadeEnabledFromInspector: aBool

	self isFadeEnabled: aBool.
	self isFadeEnabled ifTrue: [
		"Make it immediate, so the user knows it worked."
		self doFade:nextFade.
	].! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 7/31/2009 13:13'!
setupAttachment
	"My attachments are always surface-attachments out from my back."
	((self root = self parent) and: [ (self root surfaceAttachmentOf: self) isNil ]) 
	 ifTrue: [
		self root pickSurfaceFor: self alongNormal: (0@0@-1) withinDistance: 0.2
	].! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 8/5/2006 19:18'!
sheet
	"My QDisplaySheet"
	^ sheet! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 11/21/2009 01:15'!
showContextMenu: event
	(event frame = menuButton) ifTrue: [ 
		"Make right-clicking the menu button itself do the same as left-clcking the MB;
	  	 users can right-click the panel body or title for the 'pure' context menu."
		handler showMenu: event
	] ifFalse: [
		super showContextMenu: event
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 1/1/2009 12:54'!
showFloatingAppPane: evOrAvatar

	"Raise the content into a floating window in the client."
	sheet showFloatingAppPane: evOrAvatar! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 8/20/2009 15:49'!
showOrHide: bShow forPrincipal: evOrAv
	"Frame notification that a collection manager has changed or will change my visible state.
	 We stop 'expensive' apps when hidden,
	 and make sure that all but expensive apps are started when revealed. "
	bShow ifFalse: [
		sheet embeddedApp ifNotNilDo: [ : app | 
			app shouldRunUponLoad ifFalse: [
				self stopRunning: evOrAv action: #Hide
			]]
	] ifTrue: [
		sheet embeddedApp ifNotNilDo: [ : app | 
			app shouldRunUponLoad ifTrue: [
				self startRunning: evOrAv
			]]].
	! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 7/25/2008 01:20'!
slaveProjector

	"If I am being projected onto a QSlaveProjectionScreen, return it.
	 (Only the slave projectors hold state about the relationship, so we ask them.)"

	^ self root ifNotNilDo: [ : space |
			self availableSlaveScreens detect: [ : sd | sd master = sheet ] ifNone: [nil]
			].! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 8/20/2009 15:32'!
startRunning: eventOrAv
	"Sometimes we start running in response to a gesture, sometimes during load/startup.
	If during load-startup, there's no triggering event, so we don't know the responsible user;
	the eventOrAv will be nil in those cases."
	sheet hasEmbeddedApp ifTrue: [
		"Dignostics magic: click-start via the panel (the green button) 
	       with <Shift> down to start fine-grained tracing."
		(eventOrAv notNil 
		  and: [eventOrAv class = CroquetEvent ]  
		  and: [eventOrAv shiftPressed ]) ifTrue: [
			sheet childContent 
				     traceName: ('App-', (sheet childContent hashedName)).
		] ifFalse: [
			sheet childContent traceName: nil.
		].
		sheet childContent lastEvent: (
				(eventOrAv class = CroquetEvent) ifTrue: [ eventOrAv ]ifFalse: [nil]).
		sheet childContent startRunning: (eventOrAv ifNotNil: [eventOrAv userID]).
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 3/12/2010 11:32'!
stopProjectorDisplay
	"Stop any slave displays from projecting my content."
	| slaves |
	slaves := OrderedCollection new.
	self availableSlaveScreens do: [ : slave |
		slave master = sheet ifTrue: [
			slave master: nil.
			slaves add: slave.
		]].

	"Return the previous master-slave data"
	^ (slaves collect: [ : slave | { self . slave } ])  asArray! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 3/13/2010 08:52'!
stopProjectorDisplay: event 
	"Stop any slave displays from projecting my content.  Undoable"
	| oldSlaves |
	oldSlaves := self stopProjectorDisplay.
	event avatar ifNotNil: [ : avatar |
		avatar recordAction: #projection:newSlaves:oldSlaves:
			class: QDisplayPanel
			arguments: { self .  {{self . nil }}  . oldSlaves }
	]! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 8/20/2009 15:33'!
stopRunning: eventOrAv

	self stopRunning: eventOrAv action: #StopAppButton.
! !

!QDisplayPanel methodsFor: 'private' stamp: 'hrs 9/23/2009 14:32'!
stopRunning: eventOrAv action: aSymbol

	"The stop-button has been pressed."
	sheet hasEmbeddedApp ifTrue: [
		sheet onSaveForUser: eventOrAv userID.
		sheet onQuitForUser: eventOrAv userID action: aSymbol.

		"Take enough time to let the save control-s find its way upstream."
		(sheet childContent future: 1500) stopRunning: (eventOrAv userID).

		self sheet recordAction: 'paused' avatar: (eventOrAv ifNotNil: [eventOrAv avatar]).
	].! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 3/4/2007 15:12'!
title
	"The string displayed in the title widget"
	^ self frame title! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'hrs 3/11/2009 20:28'!
title: aString
	self frame ifNotNilDo: [:frame| frame title: aString].
	gotoButton label: (aString ifNil: [ self titleBase ]).! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'hrs 3/11/2009 20:23'!
titleBase
	"The basic label that we add on to as content changes."
	^dropbox ifNil: ['Display Panel'] ifNotNil: ['Dropbox'].
	! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 1/5/2010 14:48'!
titleClick: event
	"A click on my title (or background) - bring the avatar hither"
	event avatar ifNotNilDo: [ : av |	
		av doWarpMoves ifTrue: [
			gotoButton bringToDestination: event
		]]! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 5/5/2007 15:28'!
toggleFadingForUser: event

	event avatar ifNil: [ ^ self ].

	self isFadeEnabled: self isFadeEnabled not.
	self isFadeEnabled ifTrue: [
		"Make it immediate, so the user knows it worked."
		self doFade:nextFade.
		event avatar signal: #alertMessage 
				with: 'Click on the display content to show the frame.'.
		"This is an alert so it isn't overridden by a normal over-feedback info message."
	].! !

!QDisplayPanel methodsFor: 'menu' stamp: 'crl 2/19/2010 18:01'!
toggleShareAsVideoFeedViaEvent: event
	"Share my sheet's content using a video feed."

	self sheet toggleShareAsVideoFeedViaEvent: event! !

!QDisplayPanel methodsFor: 'private' stamp: 'ar 3/9/2010 11:46'!
updateShape
	"Update positions and extents
	 of components to the current extent of the panel."
	| slots ex |
	self frame ifNotNilDo: [ :f |
	 	f doUpdateShape.
		f layoutGoto: gotoButton menu: menuButton startStop: startStopButton.
	].
	ex := startStopButton extent.
	ioIndicator
		extent: ((ex x * 0.8))@(ex y * 0.8);
		translation: (startStopButton translation - ((ex x)@0@0)).

	"Keep the sheet far enough from the bar that the backdrop under-bar shadow appears.
	 The sheet is a bit in front of the backdrop ...
	 it has to be in front of the front of the resize bars in the flat frame."
	sheet
		translationX: 0  y: (self frame thickness* -0.38)  z: (self frame thickness*0.035);
		extent: self extent - (self frame thickness*0.5 min:0.5 max:0.01);
		inset: (self frame thickness*0.2) @ (self frame thickness * 0.6).

	"Size the card shelf to fit exactly 
	 (use the wanted card-size versus my width to derive the visible-slot count,
	 and then adjust the card-size so that (slotcount* cardwidth) is my width.)
	 The doc-shelf layout does the rest, including squeezing overflow cards to fit."

	self canResize ifTrue: [
		"Allow for lower-right resize doober.  (See QFlatFrame updateShape.)"
		slots := (self extent x-(self frame thickness *2.8) / (self cardExtent x)) truncated.
		rolodex basicExtent: (self extent x - (self frame thickness *2.8) / slots)@ (self cardExtent y);
				numVisibleSlots: slots;
				numSlots: ((slots max: rolodex size) max: (self class minRecentDocCapacity));
				translationX: (self frame thickness * -1.39)
						y: ((self extent y + rolodex extent y * -0.5) - (self frame thickness/2))
						z: (self frame thickness * 0.25).
	] ifFalse:[
		slots := (self extent x-(self frame thickness / 2.0) / (self cardExtent x)) truncated.
		rolodex basicExtent: (self extent x - (self frame thickness / 2.0) / slots)@ (self cardExtent y);
			numVisibleSlots: slots;
			numSlots: ((slots max: rolodex size) max: (self class minRecentDocCapacity)) ;
			translationX: 0
					   y: ((self extent y + rolodex extent y * -0.5) - (self frame thickness/2))
					   z: (self frame thickness * 0.25).
	].
	rolodex rotationAroundX: -12.	"(So the cards don't fall out.)"

	self updateBounds.! !

!QDisplayPanel methodsFor: 'events' stamp: 'bgf 12/20/2007 14:24'!
updateStartStop
	"The embedded app has signalled that its running state has changed."

	sheet hasEmbeddedApp ifTrue: [
		sheet childContent isRunning ifTrue: [
			startStopButton enableAsStarted.
		] ifFalse: [
			startStopButton enableAsStopped.
		].
		startStopHiliter floaterString: (startStopButton actionLabel, ' ', sheet shortName).
	] ifFalse: [
		startStopButton disable.
		startStopHiliter floaterString: nil.
	].! !

!QDisplayPanel methodsFor: 'accessing' stamp: 'bgf 1/27/2010 17:39'!
visibleBoundingBox
	"Avoid oddities when the fade-frame stuff is going on."
	menuButton visible ifFalse: [ 
		^ sheet boundingBox transformedBy: (sheet localTransform)
	] ifTrue: [
		^ super visibleBoundingBox
	]! !

!QDisplayPanel methodsFor: 'private' stamp: 'bgf 8/4/2009 23:53'!
warpThresholdAdjustment
	"These are adjustable."
	^ (self isPlacementLocked ifTrue: [ 1 ] ifFalse: [1.5])! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 5/31/2009 20:58'!
withFrame: aPanelFrame
	"Initialize the panel to be surrounded by the given panel frame."
	handler withFrame: aPanelFrame.
	handler 
		motionGrip: (handler frame northBar) onDemand: false;
		noRotation;
		resizeFrameOnDemand: false;
		menuAccess: menuButton;
		contactOffset: 0.01.
	"My title bar is busy enough (full of subcontrols) so having it hilite on mouse-over is too much;
	 Suppress that aspect of the motion adapter setup."
	handler motionAdapter overColor: nil.
	self title: self defaultTitle.
	self shapeChanged.
	self isPlacementLocked: self isPlacementLocked.
	self resetClickHandler.! !

!QDisplayPanel methodsFor: 'initialize' stamp: 'bgf 8/6/2009 23:33'!
withSimpleFrame
	"Initialize the display panel with a default simple flat frame."

	| frame |
	frame := (QFlatFrame new)
		disable.
	self withFrame: frame.
! !

!QDisplayPanel methodsFor: 'menu' stamp: 'bgf 1/1/2009 11:05'!
zoomToWindow100: evOrAvatar
	"Ask the UI to zoom my display-sheet to be shown as the entire content of the main window."
	sheet zoomToWindow100: evOrAvatar! !

QDisplayPanel subclass: #QDisplayStand
	instanceVariableNames: 'leg base wings leftWing rightWing rgrip height stand shadow'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QDisplayStand commentStamp: '<historical>' prior: 0!
A Display panel on a stand, for use in mid-room.
!

!QDisplayStand class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDisplayStand class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:10'!
addPropertyRows: propertyDialog forFrame: standref

	self addNameRowIfNeeded: propertyDialog forFrame: standref.

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Height';
			fetchBlock: [ :f | f get:#standHeight ];
			acceptBlock: [ : ht :f | f future standHeightFromInspector: ht ];
			validateBlock: [ : ht :f | 
				(ht isNil or: [ ht < 0 ]) ifTrue: [
					self inform: 'Height must be positive.'.
					false
				] ifFalse: [
					true
				]];
			balloonHelp: 'Height from foot to bottom of display.';
	 	yourself	).

	super addPropertyRows: propertyDialog forFrame: standref.! !

!QDisplayStand methodsFor: 'private' stamp: 'bgf 3/15/2010 13:47'!
applyResizerState: stateVec
	super applyResizerState: stateVec.
	self standHeight: stateVec fourth.! !

!QDisplayStand methodsFor: 'as yet unclassified' stamp: 'bgf 3/31/2010 12:08'!
childrenForPositioningBox
	"Include the wings so the we drag properly across floors, 
		even though these are not click-selectable."
	^ super childrenForPositioningBox, { wings }! !

!QDisplayStand methodsFor: 'persist' stamp: 'bgf 5/31/2009 20:08'!
constructFrom: pStream
	super constructFrom: pStream.
	
	self standHeight: (pStream readNumber: 'height' default: 0).
	! !

!QDisplayStand methodsFor: 'private' stamp: 'bgf 1/27/2010 17:40'!
doHideFrame
	"Make the frame invisible"
	super doHideFrame.
	{ wings . leg . base . shadow } do: [ :f | f visible: false ].! !

!QDisplayStand methodsFor: 'initialize' stamp: 'bgf 9/1/2010 17:19'!
initialize

	| baseUnit baseTh |

	super initialize.

	baseUnit := 0.8.
	baseTh := 0.08.
	height := 0.

	"The leg and central part of the base comprise the motion grip."
	stand := TGroup new.
	self addChild: stand.

	leg := TCube new.
	leg material: (QFlatFrame defaultMaterial).
	leg extent: (baseTh@1.0@(baseTh*0.5)).
	stand addChild: leg.

	base := TCube new.
	base material:  (QFlatFrame defaultMaterial);
		  extent: (baseUnit*3@baseTh@ (baseUnit*0.3)).
	stand addChild: base.

	"Two extensions flank the central base,
	 sized to appear contiguous.  These are the rotation grips,
	 so we group them and pass the group to the pivot adapter."
	wings := TGroup new.
	self addChild: wings.
	
	leftWing := TCube new.
	leftWing material:  (QFlatFrame defaultMaterial);
			 extent: (baseUnit@baseTh@(baseUnit*0.3));
			 translation: (baseUnit*2)@0@0.
	wings addChild: leftWing.

	rightWing := TCube new.
	rightWing material:  (QFlatFrame defaultMaterial);
		 extent:  (baseUnit@baseTh@(baseUnit*0.3));
		 translation: ((baseUnit *2) negated)@0@0.
	wings addChild: rightWing.
	rgrip := (QPlanePivot forLuggage: self grip: wings).
	rgrip overColor: QColors controlOverColor.

	shadow := QMaterials newRectangularShadow.
	shadow addRotationAroundX:-90;
			extent: ((baseUnit * 5)@(baseUnit*0.3)) *0.99;
			borderWidth: (baseUnit *0.125).

	self addChild: shadow.
	shadow doesClickSelect: false.
	stand doesClickSelect: false.
	wings doesClickSelect: false.

	self class = QDisplayStand ifTrue: [
		self isPlacementLocked: (self class defaultIsPlacementLocked).
	].
! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 3/19/2009 12:54'!
isGroundBound
	^ true! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 6/1/2009 12:33'!
isPlacementLocked: bool
	"Suppress my leg-wing rotation when embedded
	 (i.e. when placement-locked), just for completeness."
	"Turn on/off the foot-wing rotation behavior."
	rgrip isEnabled: bool not.
	super isPlacementLocked: bool.

! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 1/27/2010 17:39'!
noFade
	"Clear the faded state, and suppress any forthcoming fade."
	
	super noFade.
	{ wings . leg . base . shadow } do: [ :f | f visible: true ].! !

!QDisplayStand methodsFor: 'persist' stamp: 'bgf 5/31/2009 20:12'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeNumber: 'height' value: height default: nil.
	self isPlacementLocked ifTrue: [
		pStream storeBool: 'turnable' value: false.	"Pre 2.0.18"
	].! !

!QDisplayStand methodsFor: 'private' stamp: 'bgf 1/10/2010 12:47'!
portraitTransform
	"My front-on transform for photos taken with the QOnsitePhotoJob,
	 which does not take the frame out to an empty studio space."
	| camera ext |
	camera := (self globalTransform clone).
	ext := (self portraitExtent *1.3) + (0@(self standHeight)).
	camera translation: (
		camera translation
		+ (camera orientation transformVector: (0@ (self standHeight *-0.5) @ (ext maxDimension * 1.05)))).

	^ camera! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 3/31/2010 12:02'!
positioningBox
	"Underride the DisplayPanel override"
	^ (self positioningBox: Matrix4x4 new) scaled: self scaleInMatrix.
! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:55'!
privateScale: aFactor
	super privateScale: aFactor.
	stand scale: aFactor.
	wings scale: aFactor.
	shadow scale: aFactor.
	height := height * aFactor.
! !

!QDisplayStand methodsFor: 'private' stamp: 'bgf 2/1/2009 14:30'!
resetClickHandler

	"Reinforce the title bar click event handler 
	 (needed after some disables to keep the masking in place).
	 Since the stand does not attach a motion grip to the title,
	 we use the flat-frame itself as the pointer-click negotiator."
	self stopScript: #titleClick:.
	handler canMove ifTrue: [
		handler disableMove; enableMove.
	].
	handler frame ifNotNil: [
		self startScript: #titleClick: when: { handler frame  . #pointerClick }.
	].
! !

!QDisplayStand methodsFor: 'private' stamp: 'bgf 3/15/2010 13:48'!
resizerState
	^ (super resizerState),  { self standHeight }! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 7/31/2009 13:14'!
setupAttachment
	"My attachments are always surface-attachments out from my bottom."
	((self root = self parent) and: [ (self root surfaceAttachmentOf: self) isNil ]) 
	 ifTrue: [
		self root pickSurfaceFor: self alongNormal: (0@-1@0) withinDistance: 0.25
	].! !

!QDisplayStand methodsFor: 'initialize' stamp: 'bgf 3/19/2009 11:57'!
setupForAvatar: avatar atFloorPoint: dropPt

	"Place and size myself as a newly dropped-in stand for the given user,
	 in contact with the given global point on a floor surface.
	
	 The magic numbers here are the display size,
	  and the 5.5. translation to the display center which keeps viewer feet on the ground.
	  The 2.95 in the height part derives from those and the stand chrome details,
	  to keep the stand foot on the ground without a further down-pick."

	self extent: 7.6@5.7;
		  rotationAroundY: (avatar yaw detentBy: 15 atMultiplesOf: 15 snap: true);
 		  translation: (dropPt x @ (dropPt y + 5.5) @ dropPt z);
		  standHeight: (self translation y - (dropPt y + 2.95)).! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 7/14/2007 18:51'!
standHeight
	^ height! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 7/14/2007 23:35'!
standHeight: h
	height := h.
	self shapeChanged.! !

!QDisplayStand methodsFor: 'accessing' stamp: 'bgf 2/20/2008 22:24'!
standHeightFromInspector: aScalar

	"When changing the height from an inspector, 
	 update the vertical position to match, 
	 so the base preserves absolute height."

	self translation: (self translation + (0@(aScalar - (self standHeight))@0));
		standHeight: aScalar.
! !

!QDisplayStand methodsFor: 'accessing' stamp: 'hrs 3/11/2009 20:24'!
titleBase
	"The basic label that we add on to as content changes."
	^dropbox ifNil: ['Display Stand'] ifNotNil: ['Dropbox'].
	! !

!QDisplayStand methodsFor: 'private' stamp: 'bgf 8/25/2009 13:33'!
updateShape
	| ground |
	super updateShape.

	"My origin is always at center of the display rectangle...
	ground is the y offset from there to the bottom of the leg."

	ground := (height + (super extent y/2.0)) negated.

	leg extent: (leg extent x @ (height + super extent y) @ leg extent z).
	leg translation: (0@ (ground + (leg extent y/2.0))@-0.07).

	wings translation: (0@ground@0).
	base translation: (0@ground@0).

	shadow translationX: 0 y: (ground - (base extent y/2.0)) + 0.01 z:0.
	"Often either the local extent nor translation of the base parts have changed,
	 only their group, but that needs updating."
	stand updateBounds.! !

!QDisplayStand methodsFor: 'initialize' stamp: 'bgf 11/8/2009 09:55'!
withFrame: aFrame
	"Configure the frame and panel-handler for my interactions."
	super withFrame: aFrame.

	handler motionAdapter isAttractedToSurfaces: false.
	handler motionGrip: stand onDemand: false.
	(handler motionAdapter) 
				overColor: (QColors controlOverColor);
				isGroundBound: true;
				isAttractedToSurfaces: true;
				attractionRange: 1.5;
				openSpacePolicy: #viewer;
				surfacePolicy: #uprightOnly.
	handler frame unpinEdge: #north.	"Since motion is at the base, the top can resize."
	handler enableMove; enableResize.

	self resetClickHandler.! !

TGroup subclass: #QDisplaySheet
	instanceVariableNames: 'pictureRect portal card childContent inset extent isDropEnabled glass camera'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QDisplaySheet commentStamp: '<historical>' prior: 0!
A rectangular display surface
that can present textures (images), 
as well as morphic textures (applications) and portals.    

Accepts drops of QFileCards,
and displays the payload.
Textures, portals, and postcards (also handled as portals)
can be accepted and displayed.

Textures are displayed to fill the sheet, but respecting aspect ratios.
Portals fill the sheet.   In both cases, a small inset margin is maintained.

Content can be dragged from the display sheet
into other drop sites.  (The drag/drop vehicle is a QFileCard
to which the drag activity is delegated.)

Emits two event signals:
	#contentChanged (with QFileCard) whenever the visible content changes.
	#contentDropped (with QFileCard) whenever an interactive drop is accepted.

(So #contentChanged will be emitted for every #contentDropped,
but not vice-versa; #contentDropped is not emitted for programmatic changes.)

Mainly used as the interior of a QDisplayScreen.
!

!QDisplaySheet class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDisplaySheet class methodsFor: 'inspection' stamp: 'bgf 1/26/2010 12:01'!
addPropertyRows: propertyDialog forFrame: sheetref
	"Display sheets add extent, and an option for emptying.
	 (It is rare to encounter a user-editable display sheet in the wild,
	 but we use them sometimes for putting fixed-content displays into templates.)"

	self addNameRowIfNeeded: propertyDialog forFrame: sheetref.
	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ : f | f get:#extent ] ;
			acceptBlock: [ :pt :f | f future extent: pt ] ;
			balloonHelp: 'Dimensions of the display surface.';
		yourself
	). 
	super addPropertyRows: propertyDialog forFrame: sheetref.

	propertyDialog addRow: (
		(QBooleanPropertyRow new)
			titleString: 'Is Drop Enabled';
			fetchBlock: [ :f | (f get: #isDropEnabled) = true ];
			acceptBlock: [ : b :f | f future isDropEnabled: b ];
			balloonHelp: 'Can new content be dropped into the display?';
		yourself).

	propertyDialog addRow: (
		(QBooleanPropertyRow new)
			titleString: 'Is Empty';
			fetchBlock: [ :f | (f get: #contents) isNil ];
			acceptBlock: [ : b :f | b ifTrue: [ f future clear ]];
			balloonHelp: 'Check to remove any old content.';
		yourself).
! !

!QDisplaySheet class methodsFor: 'drag & drop' stamp: 'bgf 12/26/2008 14:25'!
canAcceptDrop: payload

	"For preflight checks - can the given drag-drop payload be used
       as a content drop onto a display sheet?"

	^ ((payload class includesBehavior: QFileCard) and: [payload canEmbedInDisplay]).! !

!QDisplaySheet class methodsFor: 'defaults' stamp: 'bgf 9/6/2008 00:45'!
defaultInset
	^ 0.12@0.12! !

!QDisplaySheet class methodsFor: 'drag & drop' stamp: 'bgf 3/19/2010 11:30'!
dropIntoSheet: sheet oldCard: oldC3Z newCard: newC3Z verb: aVerb

	"Return the undoable record of a card drop into the given display sheet."
	"An undoable-action that undoes creation of the given list of frames"
	^ QBlockAction
		undoBlock: [ : avatar | sheet future replayDrop: oldC3Z forAvatar: avatar ]
		redoBlock: [  : avatar | sheet future replayDrop: newC3Z forAvatar: avatar ]
		description: aVerb
		requiredFrames: { sheet }! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 3/14/2010 12:39'!
announceDropForUndo: avatar oldCard: oldCard newCard: newCard description: aVerb
	| writer |
	writer := TSceneWriter new.
	writer preserveObjectIDs: true.
	avatar recordAction: #dropIntoSheet:oldCard:newCard:verb:
		   class: QDisplaySheet 
		   arguments: { 
					self . 
					writer storeSceneAsZip: oldCard .
					writer storeSceneAsZip: newCard .
					aVerb
				}.! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 12/21/2008 09:00'!
broadcastColor: col
	"Relays the avatar color of an active user back to the embedded-app,
	 so that floating displays can reflect that activity."
	childContent signal: #broadcastColor with: col.! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'crl 2/19/2010 17:59'!
camera
	^camera! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 11/2/2007 10:19'!
childContent
	"In various cases where the sheet is showing a custom content item,
	 such as an embedded application, return that item.
	May be nil if the sheet is showing a simple texture or is blank."

	^ childContent! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 12/21/2008 20:09'!
clear
	"Empty the display."

	card clear.
	self clearOldContent.
	self privateDisplayCard: card.
	self textureContent: nil.
	self signal:#contentChanged with: card.
	self shapeChanged.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 3/25/2010 10:19'!
clearOldContent

	QLogger panelLog.
	self closePortal.
	self stopScript: #onTextChanged.
	self stopScript: #onDocNameChanged.
	self stopScript: #onAppExtentChanged.
	self stopScript: #onAppDirtyChanged.
	self stopScript: #updateContentURL: .

	childContent ifNotNil: [
		childContent stopScript: #getKeyboardControl.
		childContent stopScript: #loseKeyboardControl.
		childContent destroy.
		childContent := nil.
	].
	self signal: #dirtyChanged with: (self isDirty).! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 12/12/2006 11:11'!
closePortal

	"Hide and disable my reusable TPortal."

	glass visible:false.
	portal visible: false.
	portal postcardLink: nil.
! !

!QDisplaySheet methodsFor: 'persist' stamp: 'bgf 2/17/2008 14:39'!
constructFrom: pStream

	| tmp |
	super constructFrom: pStream.

	self extent: (pStream readPoint: 'extent' default: (self extent)).
	self inset: (pStream readPoint: 'inset' default: (self inset)).	
	tmp := (QFileCard new).
	tmp constructContentFrom: pStream.
	self displayCard: tmp.

	self isDropEnabled: (pStream readBool: 'isDropEnabled' default: (self isDropEnabled)).

	self setEventMask.! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 3/14/2010 12:40'!
consumeDrop: payload avatar: avatar selection: aTXSel
	QLogger panelLog.

	"Can I display this content?"
	(payload isKindOf: QFileCard) ifFalse: [ ^ false ].
	(self displayCard: payload fromUser: avatar withUndoName: 'Drop')
	  ifTrue: [
		QSoundEffects play: #acceptDrop for: self.
		payload kind = #postcardLink ifTrue: [
			avatar signal: #activateLinkForFrameAndGo 
					with: payload contentLink 
					with: self 
					with: false.
		].
		self signal: #contentDropped with: card.
	].
	"I always copy, never assume ownership of the card itself."
	^ false! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 2/14/2007 16:00'!
contentCard
	"The display sheet maintains a QFileCard representing its current content;
	 it is unparented by the sheet and can be displayed in a containing frame."

	^ card! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 7/25/2008 01:13'!
contents

	"The current rectangle I'm showing, or nil if I'm empty.
	 TRectangle of some sort; may be  a Portal or video-rectangle."

	childContent ifNotNil:[ ^childContent ].		"Embedded apps, etc..."
	pictureRect ifNotNil:[ ^pictureRect ].			"Images, for which I custom-construct a TRectangle"
	portal ifNotNil:[ ^portal ].					"The portal itself is the rectangle"
	^ nil ! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'das 2/18/2009 20:00'!
contentsExtent

	"The extent of whatever I currently display, or my own extent if I am empty."
	^ self contents ifNotNilDo: [ : r | r extent ] ifNil: [ self extent ].
! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 7/26/2009 23:38'!
destinationStanceFor: anAvatar
	"Provided so that Locate works properly for documents in unpanelled display sheets."
	"If we're normally packaged inside a panel, use it so arrival is consistent."
	(self parent isKindOf: QDisplayPanel) ifTrue: [
		^ self parent destinationStanceFor: anAvatar 
	].
	"Unrooted display sheets hold welcome-apps...
	  don't go to a nonsense transform."
	self root ifNil: [ ^ nil ].
	^ QGotoButton stanceFor: anAvatar toView: self inDestination: self! !

!QDisplaySheet methodsFor: 'initialize' stamp: 'bgf 9/9/2008 16:31'!
destroy
	self stopScript: #onCardDestroyed.
	self clearOldContent.
	card destroy.
	super destroy.! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 2/5/2010 12:30'!
displayCard: aFileCard
	"Make me display the contents of the given file card, if I can;
	 returns boolean indicating whether display was possible.
	 If the display failed, the previous display is removed."

	^ self displayCard: aFileCard from: nil! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 2/5/2010 15:23'!
displayCard: aFileCard from: userID
	"Make me display the contents of the given file card, if I can;
	 returns boolean indicating whether display was possible.
	 This action comes from a specific user's session, but not from an interactive gesture."
	| tmp |

	"To make persist/restore safer, check validity here,
	  so that reloading a forum will work even if some rules change.
	  (E.g. if we withdraw certain app-definitions, we don't want to try to open such documents.)."
	aFileCard canEmbedInDisplay ifFalse: [ ^ false ].

	"Get the existing document saved in its current state,
	 since I'm closing that app and opening a new one."
	self onSaveForUser: userID.
	self onQuitForUser: userID action: #StopAppReplace.

	tmp := QFileCard new shallowCopyFrom: aFileCard.
	(self privateDisplayCard: tmp) ifTrue:[
		self shapeChanged.
		(self hasEmbeddedApp and: [ childContent shouldRunUponLoad ] ) 
		 ifTrue: [
			childContent startRunning: userID.
		].
		self signal: #contentChanged with: card.
		tmp destroy.
		^true
	].
	tmp destroy.
	^ false! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 3/14/2010 12:37'!
displayCard: aFileCard fromUser: anAvatar
	
	"Perform a display change without adding an undo record."
	^ self displayCard: aFileCard fromUser: anAvatar withUndoName: nil! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 3/24/2010 12:43'!
displayCard: aFileCard fromUser: anAvatar withUndoName: aVerb
	"Make me display the contents of the given file card, if I can;
	 returns boolean indicating whether display was possible."
	| tmp  |

	"Get the existing document saved in its current state,
	 since I'm closing that app and opening a new one."
	QLogger panelLog.
	self onSaveForUser: anAvatar userID.
	self onQuitForUser: anAvatar userID action: #StopAppReplace.

	"If this isn't a real copying of ownership, don't duplicate the document.
	 This catches the case of dropping the card from the panel shelf onto the same panel. "
	( aFileCard parent notNil 
	  and: [ (aFileCard parent component) = (self component)]) ifTrue: [
		tmp := QFileCard new shallowCopyFrom: aFileCard.
		tmp isTransient: true.
	] ifFalse: [
		tmp := QFileCard new copyFrom: aFileCard.
		(tmp transferToCurrentForum: anAvatar) ifTrue: [
			QLogger panelLog:'Transferring ', aFileCard, ' to forum'.
			self startScript: #finishDisplayCard:fromUser:withUndoName:
				withArguments: {tmp . anAvatar . aVerb } 
				when: {tmp . #transferComplete }.
			^ true.
	]].

	"If unrooted, we may be still in the construction process."
	self root ifNil: [
		self future finishDisplayCard: tmp fromUser: anAvatar withUndoName: aVerb.
		^ true
	].	
	^ self finishDisplayCard: tmp fromUser: anAvatar withUndoName: aVerb.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 2/2/2008 09:38'!
docContent: aDocument
	"Show an embedded application for the document."

	self installEmbeddedApp: (aDocument newEmbeddedAppForExtent: (self extent)).

	"Mainly used by text documents."
	self runScript: #onTextChanged when: {aDocument . #textChanged }.
	self runScript: #onDocNameChanged when: {aDocument . #shortNameChanged}.

	^ true.! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 8/1/2009 16:20'!
document
	"The QAbstractDocument I currently present, if any, or nil if I'm not showing one."
	^  card document 
! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 12/20/2008 11:16'!
embeddedApp
	"If an embedded app is running, return it; otherwise nil."
	^ self hasEmbeddedApp ifTrue: [
		self childContent
	] ifFalse: [
		nil
	].! !

!QDisplaySheet methodsFor: 'private' stamp: 'ar 2/18/2009 19:12'!
embeddedContent: anEmbeddedApp
	"Make my body a copy of the given embedded application
	 (create a new instance of the application, using the same extent and data.)

	 This is for a card whose content is explicitly a TEmbeddedApp.
	 There are other xxxContent: methods for when the content is a document,
	 movie, or URL, that also create embedded applications."
	| rect |
	rect := anEmbeddedApp class 
				name: (anEmbeddedApp appName)  
				extent: (anEmbeddedApp appExtent) 
				data: (anEmbeddedApp appData copy).
	self installEmbeddedApp: rect.
	^ true.! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 10/28/2006 23:49'!
extent 

	"The x-y extent of the display sheet rectangle"
	^ extent! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 10/28/2006 13:44'!
extent: newExtent

	"Set the current extent to the given rectangle, 
	 after enforcing the maximum extent bounds."

	extent := newExtent.
	self shapeChanged.! !

!QDisplaySheet methodsFor: 'hierarchy' stamp: 'bgf 1/31/2009 19:06'!
fileTextures
	"So resource-finding mechanisms can find dependencies in my content..
	 might be a document with items that are offscreen right now."
	^ (card fileTextures, super fileTextures) asSet asArray! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 3/14/2010 12:38'!
finishDisplayCard: tmpCard fromUser: anAvatar withUndoName: aVerb

	"A one-time handler for the transferComplete callback...
	 opens a document after it has been transferred from inventory to forum storage.
	 Note that this destroys the card, since the callers generate the card."

	| result |
	result := false.

	"Make sure the avatar is still participating, since this is a callback after doc transfer."
	(anAvatar notNil and: [ anAvatar root notNil ]) ifTrue: [
		aVerb ifNotNil:[
			self announceDropForUndo: anAvatar  oldCard: card  newCard: tmpCard description: aVerb
		].

		(self privateDisplayCard: tmpCard) ifTrue:[
			QLogger panelLog:'privateDisplayCard true'.
			self shapeChanged.
			(self hasEmbeddedApp) ifTrue: [
				QLogger panelLog:'hasEmbeddedApp true'.
				"When given a card by user gesture, always launch the app at creation."
				childContent startRunning: anAvatar userID. "triggered directly by a UI event"
			].
			self signal: #contentChanged with: card.
			result := true.
		].
	].

	tmpCard destroy.
	self stopScript: #finishDisplayCard:fromUser:.
	^ result ! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 5/28/2007 09:11'!
getKeyboardControl: avatar

	self signal: #getKeyboardControl with: avatar.
! !

!QDisplaySheet methodsFor: 'testing' stamp: 'bgf 11/2/2007 10:14'!
hasEmbeddedApp
	"True if the sheet currently contains an embedded application."
	^ childContent class includesBehavior: TEmbeddedApp! !

!QDisplaySheet methodsFor: 'testing' stamp: 'bgf 5/20/2008 09:25'!
hasRunningApp
	^ self hasEmbeddedApp and: [ childContent isRunning ].! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 8/5/2006 18:22'!
height

	^ extent y! !

!QDisplaySheet methodsFor: 'testing' stamp: 'bgf 2/5/2008 16:57'!
holdsDocument: aSharedDocument

	"Locate-support for the inspector... return true if I currently display the given document."
	^ (self contentCard content = aSharedDocument)! !

!QDisplaySheet methodsFor: 'initialize' stamp: 'bgf 5/26/2009 09:24'!
initialize

	super initialize.
	self setEventMask.
						"EPD so laser-tracking works; it is dependent on events"
	inset := self class defaultInset.
	extent := (4@3).	
	isDropEnabled := true.

	"The glass ensures pickability no matter what the content rectangle."
	glass := (TRectangle new)
		material: QMaterials portalGlassMaterial;
		divisions: 2@2;
		visible: false;
		yourself.
	self addChild:glass.

	pictureRect := TRectangle new.
	pictureRect material: (QMaterials fullWhiteAlphaMaterial);
		divisions: 3.
	pictureRect material emissiveColor: (Color white /2).
	self addChild: pictureRect.

	portal := QPortal new.
	portal visible: false.
	portal blocked: true.		"No walking through display portals."
	self addChild:portal.
	self startScript: #onGlassClick: when: { glass . #pointerUp }.
	self makeMyCard.
	self textureContent: nil.! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 2/17/2008 14:36'!
inset
	^ inset! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 2/17/2008 14:37'!
inset: aPoint
	inset := aPoint.
	self shapeChanged.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 9/1/2009 23:48'!
installChild: newChild

	"Display the given object by parenting it into the shade.
	 If newChild is nil, simply clears any previous child content.
	 The display sheet assumes ownership for newChild
	 and will destroy when content is later changed."

	QLogger panelLog.
	self clearOldContent.
	newChild ifNil: [
		^self.
	].

	pictureRect visible: false.
	childContent := newChild.
	childContent translation: 0@0@(childContent extent maxDimension * 0.001). 
	self addChild: childContent.

	self runScript: #signalPointerDown: when: {childContent . #pointerDown}.
	self runScript: #signalPointerDown: when: {childContent . #mouseDown}.
	self runScript: #getKeyboardControl: when: {childContent. #getKeyboardControl}.
	self runScript: #loseKeyboardControl when: {childContent. #loseKeyboardControl}.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 3/25/2010 10:18'!
installEmbeddedApp: eAppRect

	"Adopt the given embedded app as my content;
	 this extends '#installChild:' to cover the common extra steps for applications."

	QLogger panelLog.
	(frameParent isKindOf: QDisplayPanel) ifTrue: [eAppRect displayPanel: frameParent].
	eAppRect hashedName: (card hashedName).			"For preview identification."
	eAppRect material emissiveColor: (Color gray:0.5).	"For some immunity to space lighting."
	self installChild: eAppRect.
	self startScript: #onAppExtentChanged when: { eAppRect . #appExtentChanged }.

	"This one is a signal resignaler, so can be runScript for immediacy"
	self runScript: #onAppDirtyChanged when: { eAppRect . #dirtyChanged }.! !

!QDisplaySheet methodsFor: 'testing' stamp: 'bgf 3/25/2010 10:15'!
isDirty
	^ self hasRunningApp and: [ childContent isDirty ]! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 10/28/2006 00:36'!
isDropEnabled
	"True if the screen accepts drops right now."

	^ isDropEnabled! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 10/28/2006 00:36'!
isDropEnabled: aBool
	"True if the screen accepts drops right now."

	isDropEnabled := aBool! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 1/11/2009 18:22'!
isDropTarget: payload

	(isDropEnabled and: [self class canAcceptDrop:payload]) ifFalse: [ ^ false ].

	"Now we know it is a file card that we can show.
	 Check to protect against python apps which assume a containing DisplayPanel
	 (e.g. see #getDataZip, reads the displayPanel extent rather than the embeddedApp's own.)"
	(payload kind = #panelApp) ifTrue: [
		^ self parent isKindOf: QDisplayPanel
	].
	^ true! !

!QDisplaySheet methodsFor: 'testing' stamp: 'bgf 12/21/2008 20:23'!
isEmpty
	"Am I empty of content (and therefore transparent)?"
	^ card content isNil! !

!QDisplaySheet methodsFor: 'events' stamp: 'das 4/30/2007 14:43'!
loseKeyboardControl

	self signal: #loseKeyboardControl.
! !

!QDisplaySheet methodsFor: 'initialize' stamp: 'bgf 8/16/2009 00:52'!
makeMyCard

	"We keep a QFileCard whose content is the stuff we're displaying;
	 it gives us access to the QFileCard classification methods."

	card := QFileCard new.
	card isTransient:false. "(No rolodex, but doesn't evaporate on drop.)"
	card dragAdapter isDeletable: false.  "Don't offer to delete this, it makes little sense here."
	self privateDisplayCard: card.

	"In case the user destroys my content via the card's delete menu."
	self stopScript: #onCardDestroyed.
	self runScript: #onCardDestroyed when: { card . #destroyed }.
	self stopScript: #onSaveForUser.
	self runScript: #onSaveForUser: when: { card . #saveForUser }.

	self signal:#contentChanged with: card.
! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 12/16/2007 19:26'!
mpegFileContent: mpegDocument

	"Make my body an embedded text editor;
	 the MPEG must appear in the prepackaged content!! "
	|app|
	app := QEmbeddedMpegPlayer new.
	app document: mpegDocument.
	self installEmbeddedApp: app.
	^ true

! !

!QDisplaySheet methodsFor: 'private' stamp: 'das 11/10/2008 08:45'!
onAbort

	"Relay quit requests to the embedded app if I have one."

	childContent signal: #quitForUser with: nil.	
! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 3/25/2010 10:48'!
onAppDirtyChanged
	"Relay dirty-changes in apps I happen to hold, so my container can just observe my dirty state."
	self signal: #dirtyChanged with: (self isDirty)! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 8/18/2009 01:09'!
onAppExtentChanged
	"When the app extent changes, its aspect ratio does also.
	 So we re-lay-out the sheet to get the interior properly proportioned."
	self updateShape! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 2/19/2007 17:25'!
onCardDestroyed

	self makeMyCard.! !

!QDisplaySheet methodsFor: 'private' stamp: 'das 1/5/2009 10:45'!
onDocNameChanged
	"The underlying document has changed its short-name"

	QLogger panelLog.
	self signal: #titleChanged with: card.
	self hasEmbeddedApp ifTrue: [
		childContent hashedName: (card hashedName).			"For preview identification."
		card faceAppPreview.
	].
! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 5/26/2009 09:22'!
onGlassClick: event

	"Clicking on an open, unresolved portal triggers login."
	((glass isOrContains: (event selection frame)) and: [portal visible]) ifTrue: [
		^ event avatar signal: #activateLinkForFrameAndGo 
				with: (portal postcardLink)
				with: self 
				with: false.	"No traversal"
	].! !

!QDisplaySheet methodsFor: 'private' stamp: 'das 8/5/2009 14:31'!
onQuitForUser: aUserID action: aSymbol
	"Relay save+quit requests to the embedded app if I have one."
	self hasRunningApp ifTrue: [
		QActivityCapture action: aSymbol
						 details: (childContent shortName)
						 forSession: aUserID.
		childContent isDirty ifTrue: [
			card docChange: 'Close' byID: aUserID.
		].
	]. 

	childContent signal: #quitForUser with: aUserID.	
! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 3/12/2009 08:04'!
onSaveForUser: aUserID

	"Relay save request to the embedded app, if any.
 	 There are two steps: foreground in-island protocol,
	 and then a signal so arbitrary observers (e.g. the off-island resources) can act.
	 The foreground protocol is only for embedded app content.
	 The saveForUser signal is issued regardless of what kind of content I have."

	QLogger panelLog:'userID = ', (aUserID asString).

	(childContent isKindOf: TEmbeddedApp) ifTrue: [
		childContent acceptForUser: aUserID.
		childContent signal: #saveForUser with: aUserID.
	].
! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 2/19/2007 15:34'!
onTextChanged

	"We're displaying the editor for a text document,
	 and the underlying text has changed.  Signal the embedded application,
	 so the off-island text morph can be updated."

	childContent signal: #textChanged with: card content text.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 12/16/2007 19:28'!
openPortal

	"Enable my reusable TPortal."

	self installChild: nil.
	pictureRect visible: false.
	glass visible: true.

	"To avoid transient mirrors in the (quite old) teleconference-desk demo, 
	 hide the portal until content is installed; 
	 see for instance #portalContent:"

	portal visible: false.		

	"Keep pointer-tracking happy."
	self setEventMask.

! !

!QDisplaySheet methodsFor: 'private' stamp: 'das 1/5/2009 10:45'!
panelAppContent: panelDoc
	"A Panel app. Specifies the application by way of module name like:
		Qwaq.SimplePaint
	etc. which must exist in scripts/ directory."
	|app|
	QLogger panelLog.
	app := PyPanelApp new.
	app document: panelDoc.
	self installEmbeddedApp: app.
	self startScript: #onDocNameChanged when: { panelDoc . #docNameChanged }.
	^ true

! !

!QDisplaySheet methodsFor: 'persist' stamp: 'bgf 2/17/2008 14:38'!
persistOn: pStream

	super persistOn: pStream.

	pStream storePoint: 'extent' value: extent default: (4@3).
	card persistContentOn: pStream.
	pStream storeBool: 'isDropEnabled' value: isDropEnabled default: true.
	pStream storePoint: 'inset' value: self inset default: (self class defaultInset).! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 12/12/2007 23:45'!
pointerEnter: event

	self pointerOver: event.! !

!QDisplaySheet methodsFor: 'events' stamp: 'das 4/16/2007 14:08'!
pointerOver: event

	QArrowCursor setArrowCursor: event frame: self.
! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 8/5/2006 18:16'!
portal

	^ portal.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 12/15/2007 23:13'!
portalContent: newPortal

	"Display the portal as the content of the shade,
	 link it to newPortal,
      and send this shade's postcard to newPortal for 2-way display."

	self openPortal.

	"Set up for two-way viewing."

	(newPortal = portal) ifTrue: [
		portal postcardLink: nil.
		portal visible: false.
	] ifFalse: [
		portal visible: true.
		portal postcardLink: newPortal postcard. 
		newPortal postcardLink: portal postcard.
	].

	^ true.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 12/15/2007 23:15'!
postcardContent: aPostcard

	"Display a portal onto the given postcard url."

	self openPortal.
	portal visible: true.
	portal postcardLink: aPostcard.

	^ true! !

!QDisplaySheet methodsFor: 'private' stamp: 'crl 12/16/2009 17:13'!
privateDisplayCard: aCard

	"Set the content as given; cont must be a texture, portal, postcard, or TFrame.
	 Returns true if the content is applied, false if rejected."

	"Cancel any old content listeners."
	| prevCard result |
	QLogger panelLog: aCard kind.
	self stopScript: #onTextChanged.
	self stopScript: #onDocNameChanged.
	self stopScript: #updateContentURL:.

	"It is convenient to have 'card' refer to what we're trying to display
	 while setting up the display..."
	prevCard := QFileCard new shallowCopyFrom: card.
	card shallowCopyFrom: aCard.
	result := (card kind caseOf: {
		[#document] -> [ self docContent: card content].
		[#url] -> [self urlContent: card content].
		[#ssh] -> [self sshContent: card content].
		[#application] -> [ self embeddedContent: card content].
		[#image] -> [ self textureContent: card fileTexture].
		[#postcardLink] -> [self postcardContent: card content].
		[#object] -> [ false ].
		[#mpeg] -> [ self mpegFileContent: card content].
		[#qms] -> [self qmsExistingAppContent: card content].
		[#videoLink] -> [ self postcardContent: card content ].	
		[#streamingVideoV1] -> [ self streamingVideoContentV1: card content ].
		[#panelApp] -> [ self panelAppContent: card content ].
		[#webcast] -> [self webcastContent: card content].
		[nil] -> [ self textureContent: nil ].
	} otherwise: [
		QLogger logError: 'Unknown FILE CARD KIND: ', (card kind).
		false 
	]).

	(result == true) ifFalse: [
		card shallowCopyFrom: prevCard.
	].
	prevCard destroy.
	^ result.
! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:55'!
privateScale: aFactor
	
	nomScale := nomScale * aFactor.
	self extent: (self extent * aFactor).! !

!QDisplaySheet methodsFor: 'private' stamp: 'ar 2/18/2009 18:32'!
qmsExistingAppContent: existingAppDescription
	| app |
	app := QRemoteServerApp 
			name: #QExistingRemoteApp
			extent: existingAppDescription extent
			data: { '' . '' . existingAppDescription persistString }.   "App-server, media stream, url "
	self installEmbeddedApp: app.
	^ true.! !

!QDisplaySheet methodsFor: 'events' stamp: 'hrs 9/23/2009 14:32'!
recordAction: aString avatar: avatarReplicaOrNil
	"Record the stated action in the text chat window.
	Events here have an avatar, but those lower down (in TEmbeddedApp) have only 
	a user/sessionID. Hence it is the sender's responsibility to get the avatar."
	self root ifNotNilDo: [:space | 
			space textMessage: 
					((avatarReplicaOrNil ifNotNilDo: [:a | a publicName])
						ifNotNilDo: [:name | name, ' ', aString]
						ifNil: [aString])
					fromAvatar: nil  "host message"
					atTime: DateAndTime now asUTCSeconds
					place: self contentCard identifier.
			].! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 7/19/2009 14:11'!
removeUser: user selection: selection

	selection userInfo ifNotNil:["remove the frame that is in this slot"
		selection userInfo delete. 
		selection userInfo: nil.
	].! !

!QDisplaySheet methodsFor: 'undo-redo' stamp: 'bgf 3/3/2010 14:41'!
replayDrop: cardC3Z forAvatar: anAvatar
	| cd find | 
	cd := (TSceneReader new 
			preserveDocumentNames: true;
			readSceneFromZip: cardC3Z).
	
 	find := (self component find: [ : cd1 | cd1 = cd ]).
	find isEmptyOrNil ifFalse: [
		cd destroy.
		cd := find first.
	].

	(self privateDisplayCard: cd) ifTrue:[
		self shapeChanged.
		(self hasEmbeddedApp) ifTrue: [
			"When given a card by user gesture, always launch the app at creation."
			childContent startRunning: anAvatar userID. "triggered directly by a UI event"
		].
		self signal: #contentChanged with: card.
	].
! !

!QDisplaySheet methodsFor: 'initialize' stamp: 'das 10/6/2007 15:12'!
setEventMask

	"Something in the trigger setup of display panels,
	 or in event handling itself, clears my event mask.
 	 This restores it."
"	self eventMask: EventPointerOver + EventPointerDown."! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 12/20/2007 00:01'!
shortName
	self hasEmbeddedApp ifTrue: [
		^ childContent shortName
	].
	^ card kindName! !

!QDisplaySheet methodsFor: 'editing' stamp: 'bgf 1/19/2010 17:38'!
shouldSelectFromClick
	"Protocol test - should a simple click on this frame select the frame?
	 If there is anything inside the sheet right now, then don't do edit-selecting on it; allow interaction."
	self isEmpty ifFalse: [ ^ false ].
	^ super shouldSelectFromClick! !

!QDisplaySheet methodsFor: 'menu' stamp: 'bgf 1/4/2009 20:44'!
showFloatingAppPane: anEvent
	"Show my content in a floating window in the client."
	self hasEmbeddedApp ifTrue: [
		anEvent avatar signal: #floatForEmbeddedApp with: (self childContent)
	].! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 5/28/2007 17:46'!
signalPointerDown: event
	"For fade management in the display panel.
	Note this will sometimes raise the signal without a CroquetEvent
	(we observe mouseDown from the embedded application)."
	self signal: #pointerDown with: event
! !

!QDisplaySheet methodsFor: 'private' stamp: 'ar 2/18/2009 18:32'!
sshContent: appDescription
	"Make my body a copy of the given embedded application
	 (create a new instance of the application, using the same extent and data.)"
	| app |
	app := QRemoteServerApp 
			name: #QRemoteSSHApp
			extent: appDescription extent
			data: { '' . '' . appDescription persistString }.   	"App-server, media stream, appData "

	self installEmbeddedApp: app.
	^ true.! !

!QDisplaySheet methodsFor: 'private' stamp: 'jcg 1/25/2008 12:59'!
streamingVideoContentV1: desc "a QStreamingVideoDescription"
	| rect |

	"Create a new rectangle with a QVideoTexture."
	rect := QVideoRectangle new initializeWithDescription: desc.

	"Add the video rectangle to this sheet."
	self installChild: rect.
	^true! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 2/14/2007 15:02'!
textEditorSize

	"Size in pixels for a text editor in my current display."

	(extent x > extent y)  ifTrue: [
		^ 512@(512 * (extent y / extent x)) rounded
	] ifFalse: [
		^ (512 * extent x/extent y) rounded @ 512
	].! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 12/21/2008 20:00'!
textureContent: txtrOrNil

	"Display the given texture as the screen content." 

	self installChild: nil.
	pictureRect visible: true.
	self closePortal.

	txtrOrNil ifNil: [
		pictureRect material: QMaterials clearMaterial.
	] ifNotNil: [
		pictureRect material: (QMaterials fullWhiteAlphaMaterial).
		pictureRect material texture: txtrOrNil.
	].

	self setEventMask.
	^ true
! !

!QDisplaySheet methodsFor: 'accessing' stamp: 'bgf 12/21/2008 18:22'!
title
	"The title to use in panels and other containers for this display."
	^ self contentCard shortSummary! !

!QDisplaySheet methodsFor: 'events' stamp: 'crl 2/19/2010 18:02'!
toggleShareAsVideoFeedViaEvent: event
	"Share my contents using a video feed."

	QLogger logInfo: camera class printString.
	camera
		ifNil: [
			camera := QVideoCamera new.
			camera sessionID: event avatar sessionID.
			camera sheet: self.
			camera visible: false.
			self addChild: camera]
		ifNotNil: [camera := nil]! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 7/8/2009 10:29'!
updateContentURL: aURL
	"Notice that the URL of a firefox display has changed.
	 Update the content card url for persistence, and update my title."

	"Make sure we're still displaying a url
	 (protection from harness-client races in url updating)"
	card kind = #url ifFalse: [ ^ self   "Stale signal from a replaced app." ].

	"If I live in a panel that manages my state, then it needs the the update."
	(self parent respondsTo: #contentCard) ifTrue: [
		| ownerCard |
		ownerCard := self parent contentCard.
		(ownerCard notNil and: [ ownerCard kind = #url ]) ifTrue: [ 		
			"The kind check is just a safeguard in case web-traversal overlaps an app-change."
			ownerCard updateContent: aURL
		]].
	"And keep my own information up to date."
	card updateContent: aURL.
	self onDocNameChanged.! !

!QDisplaySheet methodsFor: 'events' stamp: 'bgf 3/15/2007 01:50'!
updatePostcard: newPostcard originalUrl: orig

	"Callback after postcard resolution - update my postcard link in place."
	(self contentCard kind = #postcardLink) ifTrue: [
		self contentCard content urlString: newPostcard urlString.
		portal postcardLink: newPostcard.
		self signal: #contentChanged with: self contentCard
	].! !

!QDisplaySheet methodsFor: 'initialize' stamp: 'bgf 9/6/2009 12:36'!
updateShape
	| isTexture cwidth cheight |
	"Update the content geometry to the current extent of the sheet."

	cwidth := (extent x - inset x) max: 0.
	cheight := (extent y - inset y) max: 0.

	"We size all three content variants even though only one is visible at a time.
	 (Doesn't take very long, and we don't do this often.)"
	isTexture := (card fileTexture isKindOf: TTexture). 
	isTexture ifTrue: [
		pictureRect extent: cwidth@cheight  aspect: card fileTexture originalAspect.
	] ifFalse: [
		pictureRect extent: cwidth@cheight.
	].
	pictureRect translation:0@0@0.
	
	glass extent: cwidth@cheight;
			translation: 0@0@0.02.
	portal extent: cwidth@cheight;
			translation: 0@0@0.01. 

	childContent ifNotNil: [
		childContent extent: cwidth@cheight aspect: childContent originalAspect;
				translation: 0@0@((cwidth max: cheight) * 0.001). 
	].

	self updateBounds.! !

!QDisplaySheet methodsFor: 'private' stamp: 'bgf 5/24/2009 14:42'!
urlContent: aURL
	"Make my body a copy of the given embedded application
	 (create a new instance of the application, using the same extent and data.)"
	| app |
	app := QRemoteServerApp
			name: #QRemoteFirefox
			extent: QRemoteFirefox defaultExtent
			data: { '' . '' . aURL }.   	"App-server, media stream, url "
	self installEmbeddedApp: app.
	self runScript: #updateContentURL: when: {app. #urlChanged}.
	
	^ true.! !

!QDisplaySheet methodsFor: 'private' stamp: 'crl 12/17/2009 11:36'!
webcastContent: record "a QWebcastRecord"
	| rect |

	QLogger logInfo: 'setting webcast content of display sheet'.

	"Create a new rectangle with a QVideoTexture."
	rect := QVideoRectangle new initializeWithWebcastRecord: record.

	"Add the video rectangle to this sheet."
	self installChild: rect.
	^true! !

!QDisplaySheet methodsFor: 'menu' stamp: 'bgf 7/18/2009 15:38'!
zoomToWindow100: evOrAvatar
	"Ask the UI to zoom my display-sheet to be shown as the entire content of the main window."
	evOrAvatar ifNil: [ ^ self ].
	evOrAvatar avatar ifNil: [ ^ self ].
	evOrAvatar avatar signal: #zoomToWindow100 
					    with: self asPlace
					    with: self globalTransform
					    with: (self contentsExtent + 0.01).
! !

TGroup subclass: #QDocumentShelf
	instanceVariableNames: 'basicExtent base cards openCard direction infoMessage border numSlots axis numVisibleSlots'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QDocumentShelf commentStamp: '<historical>' prior: 0!
A linear shelf to organize a set of QFileCards.  
Provides mouse-over "enlarge the focal card" behavior for browsing.  
Accepts drops (copies the dropped card into a new card to retain a reference to its content.)

Visually, a framed rectangle with cards laid across.  
There are two orientations: #vertical or #horizontal.
Default is horizontal.

Sizing is automatic, based on 
	basicExtent - extent of a single shelved card
	numSlots    - number of card positions
	
If numSlots is nil, the shelf is potentially infinite, shrink-to-fit around its current card set.!

!QDocumentShelf class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDocumentShelf class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:16'!
addPropertyRows: propertyDialog forFrame: shelfRef

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Card Positions';
			fetchBlock: [ :f | (f get:#numVisibleSlots) rounded];
			acceptBlock: [ :pns :f | f future numVisibleSlots: pns rounded ];
			validateBlock: [ : float : f | 
				(float isNil or: [ float < 1 ])  ifTrue: [
					self inform: 'Counts must be positive.'.
					false
				] ifFalse: [
					true
				]];
			balloonHelp: 'Visible card positions.\Together with Card Extent, sets the size of the shelf.'withCRs;
	 	yourself	).

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Maximum Cards';
			fetchBlock: [: f |  ((f get:#numSlots)  ifNil: [0]) rounded] ;
			acceptBlock: [ :slots : sr | sr future numSlots: ((slots rounded = 0)ifTrue: [nil] ifFalse: [slots rounded])] ;
			validateBlock: [ : float : sr| 
				(float < 0) ifTrue: [
					self inform: 'Counts must be positive.'.
					false
				] ifFalse: [
					true
				]];
			balloonHelp: 'Total capacity of the shelf. 0 means unlimited.';
	 	yourself	).

	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: ('Card Extent');
			fetchBlock: [ :shelf | shelf get: #basicExtent ] ;
			acceptBlock: [ :pt : shelf| shelf future basicExtent: pt ] ;
			balloonHelp: 'Size of a single card in the shelf.';
		yourself).

	super addPropertyRows: propertyDialog forFrame: shelfRef.! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 1/18/2008 15:25'!
defaultAxis
	^ #horizontal! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 2/7/2008 18:16'!
defaultBaseMaterial
	^ (QMaterials mediumWoodMaterial)
			cullFace: false;
			yourself.! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 1/17/2008 23:31'!
defaultBasicExtent
	"Default collapsed card size - the basis for layout."
	^ (1.0@0.667)! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 2/7/2008 18:16'!
defaultBorderMaterial
	^ (QMaterials darkWoodMaterial)
			cullFace: false;
			yourself.! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 1/17/2008 23:31'!
defaultDirection
	^ #right! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:20'!
defaultEventMask
	^ EventPointerOver! !

!QDocumentShelf class methodsFor: 'undo-redo' stamp: 'bgf 3/19/2010 11:33'!
shelf: shelfRef dropCard: cardRef capture: cardCapture
	"Undoable action when dropping a (newlycreated transient) card into the shelf"
	^ QBlockAction
		undoBlock: [ : avatar | shelfRef future undoDrop: cardRef forUser: avatar ]
		redoBlock: [ : avatar | shelfRef future redoDrop: cardCapture forUser: avatar ]
		description: 'Drop'
		requiredFrames: { shelfRef }! !

!QDocumentShelf class methodsFor: 'undo-redo' stamp: 'bgf 3/19/2010 11:33'!
shelf: shelfRef moveCard: cardRef oldParent: oldRolodex
	"Undoable action when moving a card into the shelf from some other container."
	^ QBlockAction
		undoBlock: [ : avatar | oldRolodex future insertCard: cardRef ]
		redoBlock: [ : avatar | shelfRef future insertCard: cardRef ]
		description: 'Move Card'
		requiredFrames: { shelfRef . cardRef . oldRolodex }! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 1/17/2008 22:10'!
verticalSlope
	"The forward lean of cards in a vertical rolodex"

	^ -16 "degrees"! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 1/17/2008 22:10'!
verticalSlopeSin
	"The sin of vertical slope, which is used a lot in visual animation updates."

	^   0.275637355816999! !

!QDocumentShelf class methodsFor: 'defaults' stamp: 'bgf 6/13/2009 11:23'!
zoomFactor

	"The maximum enlargement created by opening a card."

	^ 1.333! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/8/2010 13:41'!
addChild: aChild atIndex: anIndex
	"We manage card-children carefully... divert to that process 
	 for arbitary child additions."
	(aChild isKindOf: QFileCard) ifTrue: [
		^ self insertCard: aChild allowDuplicates: true sort: true
	].
	super addChild: aChild atIndex: anIndex.
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/19/2008 13:56'!
addContent: cardContent kind: cardKind

	"Ensure that I contain the given content.
	 If I'm not already holding a card of the given kind+content, add one.
	 Return the card found or created."

	(cards detect: [ : c | c content = cardContent and: [ c kind = cardKind ]] ifNone: [nil])  
		ifNotNilDo: [ : found | 
			^ found 
		] ifNil: [
			^ self insertCard: (QFileCard new kind: cardKind content: cardContent; yourself)
		].


! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/4/2010 14:01'!
addPastedChild: aChild
	"Protocol for <paste> when selected.  Not used, as this is not an editable group."
	(aChild isKindOf: QFileCard) ifTrue: [
		^ self insertCard: aChild
	].
	^ super addPastedChild: aChild! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 3/22/2010 09:27'!
asTopicRecord
	"Just a summary, for when I appear in a grouping context (like on a tabbed wall)"
	| n msg |
	n := self size.
	msg := n caseOf: {
		[ 0 ] -> [ 'Empty Shelf' ] .
		[ 1 ] -> [ 'Shelf with 1 Card' ] .
	} otherwise: [
		'Shelf with ', n, ' Cards'.
	].

	^ (QTopicRecord new)
		cardRef: self;
		color: (Color gray:0.2);
		creationTime: 0;
		lastModifiedTime: 0;
		topic: msg;
		nComments: n;
		yourself! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 15:21'!
axis

	"#horizontal or #vertical depending on how cards are aligned."
	^ axis
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 15:20'!
axis: anAxis
	"One of #vertical, #horizontal (default horizontal)."

	"Make sure the symbol we get is one we know."
	axis := (anAxis = #vertical) ifTrue: [#vertical] ifFalse: [#horizontal].
	self shapeChanged.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/7/2008 18:44'!
baseMaterial
	"Material of the backing wood - the shelf proper.  Default is medium wood."
	^ base material! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/7/2008 18:44'!
baseMaterial: aMaterial
	"Material of the backing wood - the shelf proper.  Default is medium wood."
	base material: aMaterial.
	base cullFace: false.
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:11'!
basicExtent

	"The cardsize used for cards inside."

	^ basicExtent! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 02:37'!
basicExtent: rect

	"Set the cardsize used for cards inside... not the shelf!! "

	basicExtent := rect copy.
	self shapeChanged.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/7/2008 18:44'!
borderMaterial
	"Material of the twindow-like surround framing."
	^ border material! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/7/2008 18:44'!
borderMaterial: aMaterial
	"Material of the twindow-like surround framing."
	border material: aMaterial.
	border cullFace: false.
	! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 3/30/2008 00:14'!
borderThickness
	
	"Nominal thickness of the surround (window frame)"

	^  ((basicExtent x min: (basicExtent y)) / 8) min: 0.075 max: 0.0125.! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 3/14/2010 10:29'!
cardComplete: aFileCard fromUser: anAvatar

	"A file card that needed transfer to/from inventory is now ready for insertion."
	| newCard |
	newCard := self insertCard: aFileCard.
	openCard ifNotNil: [
		self fileCard: openCard.
	].
	
	self isInAvatar ifFalse: [
		anAvatar recordAction: #shelf:dropCard:capture:
				 class: QDocumentShelf
				 arguments: { self . aFileCard . (anAvatar captureFrames: { newCard } ) }.
	].

	self future stopScript: #cardComplete:fromUser:

	"Hopefully these space-harness relays are quicker than the users' drag-drop reflexes;
	 we don't want the trigger left after the operation,
	 but we don't have a nice way to have multiple one-time triggers set up."! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/19/2008 09:36'!
cardForBasePosition: v3

	"The card to focus on when the pointer is at the given y relative to the base.
	 This is how scanning is managed for the simple flat shelf...
	 picking is always done relative to the base 
	 so that pointer-stealing by neighboring animated cards
	 doesn't interrupt the zoom flow."
	|  index |

	(v3 x abs > (base extent x * 0.55)) ifTrue: [ ^ nil ].
	(v3 y abs > (base extent y * 0.55)) ifTrue: [ ^ nil ].

	self isVertical ifTrue: [
		index := (v3 y + (base extent y/2.0)) / (self effectiveBasicExtent y).
	] ifFalse: [
		index := (v3 x + (base extent x/2.0)) / (self effectiveBasicExtent x).	
	].
	 index := (index truncated) min: (self numVisibleSlots max: (cards size))-1 max: 0.

	"Sleight of hand...
	 When we lay out the cards, we reverse the indexes so the
	 rolodex grows in the human-expected order...."
	(direction == #left) ifTrue: [
		index := cards size - index.
	] ifFalse: [
		index := index + 1.
	].
	^ cards at: index ifAbsent: [nil].! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 1/19/2008 09:24'!
cardForEvent: ev

	"The card position involved in the given (pointer) event."
	| ray |

	"Use the position of the pointer on the base sheet;
	 this makes cards scannable by sweeping the mouse,
	 without having the open card hog all the picks and occlude neighbors."
	ray := ev makePointer.
	ray pointerPickTree: base.
	ray selectedPoint ifNotNilDo: [ :pt |
		^ self cardForBasePosition:  pt.
	].

	 ^nil 
		
! !

!QDocumentShelf methodsFor: 'render' stamp: 'bgf 11/28/2009 12:09'!
cardSubdistance: cd
	"For the shelf, cards do not allow other cards to occlude,
	 when scanning at a sharp angle."
	cd == openCard ifTrue: [ ^ -0.1 ].
	^ -0.01! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 23:21'!
cards	
	"Ordered collection of file cards."
	^ cards! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 3/18/2010 09:39'!
checkForShelfTransfer: aFileCard byUser: anAvatar
	"Check for card-drops that are really moves inside a storage container (shelf stack)"

	| oldDex |

	oldDex := aFileCard rolodex.
	oldDex ifNil: [ ^ false ].
	oldDex = self ifTrue: [ ^ true ].

	(self parent ~= self root and: [ oldDex parent = self parent]) ifTrue: [
		self insertCard: aFileCard.
		anAvatar recordAction: #shelf:moveCard:oldParent:
				 class: QDocumentShelf
				 arguments: { self . aFileCard .  oldDex }.
		^ true
		].

	^ false! !

!QDocumentShelf methodsFor: 'render' stamp: 'bgf 11/28/2009 12:08'!
childSubdistance: child
	"See rolodex.   We have to adjust the card distances carefully
	 for the vertical-rolodex slot-angle display to show the cards overlap properly."
	child class = QFileCard ifTrue: [ ^ self cardSubdistance: child ].
	^ 0! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
childrenForBrowser
	^ self cards asArray
	! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 9/7/2008 17:47'!
clear
	"Destroy all cards, leaving me empty."
	cards copy do: [ : card |
		card destroy.
	].! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 1/17/2008 22:10'!
closeAll

	"If a card is open (enlarged and hilite by browsing), close it"
	openCard ifNotNil: [ openCard possibleTransition:#close].! !

!QDocumentShelf methodsFor: 'persist' stamp: 'bgf 9/25/2009 22:16'!
constructFrom: pStream

	super constructFrom:pStream.

	self direction: (pStream readString: 'direction' default: (self direction)).
	self axis: (pStream readString: 'axis' default: (self axis)).
	self infoMessage: (pStream readString: 'infoMessage' default: nil).
	self basicExtent: (pStream readPoint: 'basicExtent' default: self basicExtent).

	(pStream readSequence:'cards' default:#()) 
		do: [ :cd| cd isRetiredDemoContent
					 ifFalse: [ self insertCard: cd ]
					 ifTrue: [ 
						QLogger logInfo: 'Withdrawn: ', (cd shortSummary).
						cd destroy ] ].

	self numVisibleSlots: (pStream readNumber: 'numVisibleSlots' default: nil).
	self numSlots: (pStream readNumber: 'numSlots' default: nil).

	self baseMaterial: (pStream readObject: 'baseMaterial' default: (self class defaultBaseMaterial)).
	self borderMaterial: (pStream readObject: 'borderMaterial' default: (self class defaultBorderMaterial)).

	self shapeChanged.! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 3/14/2010 09:59'!
consumeDrop: delivery avatar: avatar selection: aTXSel
	| card |

	"Standard drop-receipt protocol - can I absorb this content into a card?
	 Note that we always return 'false' here because a doc-shelf always
 	 leaves the original alone, it just copies in the material."

	"Disallowed if I'm someone's inventory, and the dropper isn't my owner."
	(self notMyAvatar: avatar) ifTrue: [ ^ false ].
	base hilite.

	"If the drop is a card, or a document that we can easily put in a card, simply insert it..."
	card := nil.
	(delivery class includesBehavior: QFileCard) ifTrue: [
		card := delivery
	] ifFalse: [
		(delivery class includesBehavior: QSharedDocument) ifTrue: [
			card := QFileCard forDocument: delivery.
		]].
	card ifNotNil: [
		(self hasContent: card content) ifTrue: [ 
			avatar signal: #infoMessage with: 'Already includes that ', (card kindSummary).
		] ifFalse: [
			self insertDroppedCard: card fromUser: avatar.	"Handles content-copying to inventory, etc."
		].
		base unhilite.
		^ false
	].

	"Fallthrough - cannot do anything with this drop."
	base unhilite.
	^ false! !

!QDocumentShelf methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:09'!
defaultIsAtomic
	^ false! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 1/18/2008 00:19'!
deferredCardComplete: aFileCard

	"A file card that needed transfer to/from inventory is now ready for insertion."
	self insertCard: aFileCard.

	openCard ifNotNil: [
		self fileCard: openCard.
	].

	self stopScript: #deferredCardComplete:
	"Hopefully these space-harness relays are quicker than the users' drag-drop reflexes;
	 we don't want the trigger left after the operation,
	 but we don't have a nice way to have multiple one-time triggers set up."! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 3/22/2010 09:26'!
destinationExtent
	"Approach, but not fully, since I'm furniture rather than a viewing display."
	^ (base extent) * (QGotoButton defaultSlopFactor*2)! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 10:05'!
detect: aFileCard

	"Like collection detect but based on the content of the card,
	 rather than true equality."

	^ cards detect: [ : cd | cd kind = aFileCard kind and: [ cd content = aFileCard content ] ] ifNone: [nil]! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 23:28'!
direction
	"#left or #right - from which edge the cards begin.  Use #left for top in a vertical shelf."
	^ direction! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 15:22'!
direction: aDirection
	"One of #left, #right (default left)."

	"Retrofit old rolodex terminology."
	(aDirection = #vertical or: [aDirection = #horizontal]) ifTrue: [
		^ self axis: aDirection.
	].

	"Make sure the symbol we get is one we know."
	direction := (aDirection = #left) ifTrue: [#left] ifFalse: [#right].
	self shapeChanged.! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 3/22/2010 09:05'!
editByUser: replica
	"Protocol infill for treating panels as attachments to topic collectors;
	 this is the response to the Topic Tool 'Open' (and topic-list double-click) command."
	
	replica goToPlace: (self asPlace)
	! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/18/2008 23:37'!
effectiveBasicExtent

	"My size is computed from basicExtent and numVisibleSlots.
	 But when I do layout, I have to show all my cards.
	 effectiveBasicExtent is the modified basicExtent (size of a card)
	 that is actually used to size the cards.
	 It will be = basic Extent when I'm not overfull.
	 The cards are made narrow during overflow to allow them all to fit."

	(cards size > self numVisibleSlots) ifTrue: [
		self isVertical ifTrue: [
			^ basicExtent x @ (basicExtent y * self numVisibleSlots / cards size)
		] ifFalse: [
			^ (basicExtent x * self numVisibleSlots / cards size) @ basicExtent y
		].
	].
	^ basicExtent! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 21:01'!
extent

	| slots |
	slots := self numVisibleSlots.
	self isVertical ifTrue: [
		^ basicExtent x@(slots * basicExtent y)
	] ifFalse: [
		^ (basicExtent x * slots)@basicExtent y
	].! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
fileCard: cd 

	"Put the card away (restore it to normal position amidst its peers)."
	| index |
	index := (self indexOfCard: cd).
	(index >= 0) ifTrue: [
		self fileCard: cd at: index.
	].
	(openCard = cd) ifTrue: [
		openCard := nil.
	].! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 3/9/2009 10:42'!
fileCard: cd at: index

	"Collapse the card back to its small size.
	 If the card is the open card, this is done gradually;
	 otherwise, it merely resets the card position and visuals."

	| increment |
	cd isTranslucent: true.

	increment := (cd extent x / basicExtent x) - 1.0.
	(cd = openCard) ifTrue: [
		openCard := nil.
		(self future: 20) placeCard: cd at: index scale: increment * 0.67 + 1.
		(self future: 50) placeCard: cd at: nil scale: increment * 0.33 + 1.
		(self future: 90) placeCard: cd at: nil scale: 1.
		(cd future: 95) endTransition:#close.

	] ifFalse: [

		self placeCard: cd at: nil scale: 1.
		cd endTransition: #close.
	].

	"If the timing changes, see also QFileCard possibleTransition:,
	 as well as QDocumentShelf openCard:at:"
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 10/25/2008 23:06'!
firstCard: card
	"Take the given card into the rolodex; make it the first card.  No duplicates;
	 if the given content is already in the rolodex, that occurrence will be replaced."
	| cd |

	"Use the given card to replace any existing duplicates,
	 since those may be in various states of drag-transition."
	(self detect: card)
		ifNotNilDo: [  : existingcard | 
			(existingcard == card)
				ifFalse: [ existingcard destroy ]		"Redundant non-first entry there."
				ifTrue: [ cd := existingcard ]].		"Already there"
	cd ifNil: [
		cd := self insertCard: card allowDuplicates: true sort: false.
	].

	(cards size > 1 and: [cards first ~~ cd]) ifTrue: [
		cards remove: cd ifAbsent: [nil].
		cards addFirst: cd.
		self shapeChanged.
	].! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
getCard: index

	^cards at: index.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'das 2/23/2009 14:08'!
getFileCards
	
	"get list of child names for Python scripts"
	| array |
	array := cards collect:[:frm | frm objectID asUIDString].
	^ array asArray.
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
hasContent: ct
	"Does any card have the given content already?"
	^ cards anySatisfy: [:cd|cd content = ct]! !

!QDocumentShelf methodsFor: 'testing' stamp: 'bgf 2/5/2008 17:01'!
holdsDocument: aSharedDocument

	"Locate-support for the inspector... return true if I hold the given document."

	self isInAvatar ifTrue: [ ^ false ].   "Since rolodex contents are private,
									   XXX should perhaps work through the details 
									   of showing docs that are in my own inventory."

	^ (cards anySatisfy: [ :cd | cd holdsDocument: aSharedDocument ])! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/28/2008 11:54'!
horizontalCard: cd at: cardIndex scale: scale

	"Place the given card in position, for a #horizontal shelf.
	 The cards are aligned along the center of the rolodex.
	 Assume scale=1 for collapsed and > 1 for expanded."

	| x y z index be |

	"Sleight of hand - we want card 1 at the left, rather than bottom..."
	direction == #left ifTrue: [
		index := cards size + 1 - cardIndex.
	] ifFalse: [
		index := cardIndex.
	].
	be := self effectiveBasicExtent.
	x := base translation x - (base extent x /2) + (be x *0.5 *1.05).
	x := x + ((index-1)* (be x)).
	y := (be y)*(scale-1.0)/6.0.		"Enlarge upwards."
	z  := 0.03 + (scale-1 * (be y)).	"Step forward"

	cd translationX: x y: y z: z; 
		extent: ((2-scale * be) + ((scale-1 *0.95 * basicExtent x)@(scale-1 * 1.15 * basicExtent y))) * scale * 0.95.
				"When squeezed to overflow, this animates the card towards the target size.
				 We also have a slight bias to growth in y"
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 10:06'!
includes: aFileCard

	"Like collection includes but based on the content of the card,
	 rather than true equality."

	^ (self detect: aFileCard) notNil! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/17/2008 22:10'!
indexOfCard: cd

	"Array index of the given card in my rolodex, or -1 if it's not present."
	^ cards indexOf:cd ifAbsent: [ -1 ].! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 23:22'!
infoMessage
	"The information message displayed when the shelf is moused-over"
	^ infoMessage! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 23:23'!
infoMessage:aMessage
	"The information message displayed when the  shelf is moused-over"
	infoMessage := aMessage! !

!QDocumentShelf methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:31'!
initialize

	super initialize.

	axis := self class defaultAxis.
	direction := self class defaultDirection.

	openCard := nil.
	basicExtent := (self class defaultBasicExtent).

	cards := OrderedCollection new.
	base := TRectangle new extent:(basicExtent x)@(basicExtent y).
	base material: (self class defaultBaseMaterial).
	self addChild: base.

	border := TWindowFrame new.
	border material: (self class defaultBorderMaterial).
	self addChild: border.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 10/25/2008 22:53'!
insertCard: card

	"Normal path to add cards to a shelf...
	 don't accept duplicates.
	 Return the argument, or the original if we discover that card is a duplicate."

	^ self insertCard: card allowDuplicates: false sort: true.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 10/25/2008 22:54'!
insertCard: card allowDuplicates: bAllow2

	"Take the given card into the rolodex,
  	 unless what it describes is already present,
	 or bAllow2 is true."

	^ self insertCard: card allowDuplicates: bAllow2 sort: true! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/14/2010 21:55'!
insertCard: card allowDuplicates: bAllow2 sort: bSort
	"Take the given card into the rolodex,
  	 unless what it describes is already present, or bAllow2 is true.
	 If my 'isSorted' property is true, and bSort is true, then the card is inserted
	 alphabetically according to its summary string."

	bAllow2 ifFalse: [
		(self detect: card) ifNotNilDo: [ :cd | ^ cd ].
	].
	"If we're taking it, reset pose since we'll take over geometry."
	card localTransform: Matrix4x4 identity.

	"If limited, make room."
	numSlots ifNotNil: [
		numSlots < (cards size+1) ifTrue: [
			QLogger logInfo: 'Shelf full, discarding: ', (cards last shortSummary).
			cards last destroy.
		].
	].

	"We assume control of pointer over handling for cards inside."
	card eventMask: (card eventMask bitClear: EventPointerOver).
	card rolodex: self.
	cards add: card.
	"Inefficient, but the faster ways keep block-valued instance variables..."
	(bSort and: [self isSorted]) ifTrue: [
		cards := cards sortBy: [:c1 :c2 | c1 summary <= c2 summary ].
	].

	"The -super- here is important: we intercept #addChild:atIndex:
	 and route it into this method,  so a self call here would just recur-loop."
	super addChild: card atIndex: (self children size + 1).

	self signal: #contentChanged.
	self shapeChanged.

	^ card! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 3/18/2010 09:27'!
insertDroppedCard: aFileCard fromUser: anAvatar
	| newCard deferUpdate |

	(self checkForShelfTransfer: aFileCard byUser: anAvatar) ifTrue: [
		^ aFileCard
	].

	newCard := QFileCard new copyFrom: aFileCard.
	newCard isTransient: true.
	self isInAvatar ifTrue: [
		deferUpdate := newCard transferToInventory: (self myAvatar).
	] ifFalse: [
		deferUpdate := newCard transferToCurrentForum: (anAvatar).
	].

	"If we're changing the card in the background,
	 wait until that's done before inserting.
	 (After insert, for example, the avatar inventory is auto-saved,
	 and we want to save the new rather than old information.)"

	deferUpdate ifTrue: [
		self startScript: #cardComplete:fromUser:
			withArguments: {newCard . anAvatar} 
			when: { newCard . #transferComplete }.
	] ifFalse: [
		self cardComplete: newCard fromUser: anAvatar
	].
	^ newCard! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
isDropTarget:payload

	"Assume we can try to absorb anything;
	 file cards attempt to capture a streamed snapshot if the content isn't 
	 one of the supported 2-D content types."
	^ true! !

!QDocumentShelf methodsFor: 'testing' stamp: 'bgf 2/14/2010 15:32'!
isEditableGroup
	^ (self isAtomic not)! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
isGroundBound
	^ true.! !

!QDocumentShelf methodsFor: 'testing' stamp: 'bgf 3/22/2010 08:56'!
isListedTopic
	"I play in the note-management schemes if I'm independently draggable."
	^ true! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 09:24'!
isSorted
	^ false! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 15:23'!
isVertical
	^ axis = #vertical! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
length
	"The extent of the rolodex along its length (or height, if vertical)."
	^ base extent y.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 10:05'!
matchingCard: aFileCard

	"Like collection includes but based on the content of the card,
	 rather than true equality."

	^ cards detect: [ : cd | cd kind = aFileCard kind and: [ cd content = aFileCard content ] ] ifNone: [nil]! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/17/2008 22:10'!
myAvatar

	"If this rolodex is owned (part of an avatar) return the owning avatar.
	 Otherwise, nil."

	| p |
	p := frameParent.
	[p notNil] whileTrue: [
		p isAvatarProxy ifTrue: [ ^p avatar].
		p isAvatar ifTrue: [ ^p ].
		p := p parent.
	].
	^ nil
! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/17/2008 22:10'!
notMyAvatar: anAvatar

	"Psuedo-security check. 
	 Returns true if this rolodex is owned by an avatar other than the given one;
	 returns false in other cases."

	^ self myAvatar ifNotNilDo: [ :av | av ~= anAvatar ] ifNil: [ ^ false ]! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/20/2008 11:50'!
numSlots
	"The number of cards I will allow myself to grow to during normal operation."
	^ numSlots ifNil: [0]! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 21:00'!
numSlots: nCardsOrNil
	"Max number of cards maintained by the rolodex in response to user activity...
	 if cards are dropped so as to exceed this number, then older ones are discarded.
	 (This limit is -not- enforced when internalizing from persistence.)
	 Value is allowed to be nil, or [1....n] but not less than 1.
 	 Nil means no fixed size... shelf will expand as needed."

	numSlots := nCardsOrNil ifNotNil: [ nCardsOrNil max: 1].
	numSlots ifNotNil:[
		((cards size - numSlots) max: 0) timesRepeat: [
			self removeChild: (cards last).
		].
	].
	self shapeChanged.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 20:58'!
numVisibleSlots
	"If the deck has a fixed number of visible slots, 
	 return that number; 
	 else return the number of cards,
	 but always at least one."

	^ numVisibleSlots ifNil: [ cards size max: 1 ]! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/18/2008 20:57'!
numVisibleSlots: nCardsOrNil
	"Value is allowed to be nil, or [1....n] but not less than 1.
 	 Nil means no fixed size... shelf will expand as needed."

	numVisibleSlots := nCardsOrNil ifNotNil: [ nCardsOrNil max: 1].
	self shapeChanged.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
openCard: cd 

	"Make the given card the center of attention;
	 the pointer is over it.  (In a vertical dex, cards pop out to front; in a horizontal tray, they expand.)"

	openCard ifNotNil: [ 
		openCard = cd 
			ifTrue: [cd endTransition:#open. ^self] 
			ifFalse:[openCard possibleTransition:#close]
		].

	openCard := cd.
	self openCard: cd at: (self indexOfCard: cd).
! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/28/2008 11:37'!
openCard: cd at: index

	self openCard: cd at: index zoom: self class zoomFactor.
! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 3/9/2009 10:43'!
openCard: cd at: index zoom: aFactor

	| increment |

	"Enlarge the card and make it opaque.  
	 This is called via #openCard, which deals with card transition schedules."

	cd isTranslucent: false.

	increment := aFactor - 1.

	self placeCard: cd at: index scale: increment*0.2 + 1.0.
	(self future: 30) placeCard: cd at: nil scale: increment*0.4 +1.0.
	(self future: 60) placeCard: cd at: nil scale: increment*0.6 +1.0.
	(self future: 90) placeCard: cd at: nil scale: increment*0.8 +1.0.
	(self future: 120) placeCard: cd at: nil scale: increment  +1.0.
	(cd future: 125) endTransition:#open.

	"If the timing changes, see also QFileCard possibleTransition"
! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
outerExtent

	^ self extent + (self borderThickness*2)! !

!QDocumentShelf methodsFor: 'persist' stamp: 'bgf 2/7/2008 18:17'!
persistOn: pStream
	super persistOn: pStream.

	pStream storePoint: 'basicExtent' value: basicExtent default: (self class defaultBasicExtent).
	pStream storeNumber: 'numVisibleSlots' value: numVisibleSlots default: nil.	"nil means size-to-fit"
	pStream storeNumber: 'numSlots' value: numSlots default: nil.	"nil means infinite"

	pStream storeSequence: 'cards' value: cards default: #().
	pStream storeString: 'direction' value: direction default: (self class defaultDirection).
	pStream storeString: 'axis' value: axis default: (self class defaultAxis).
	pStream storeString: 'infoMessage' value: infoMessage default: nil.

	pStream storeObject: 'baseMaterial' value: (self baseMaterial) default: (self class defaultBaseMaterial).
	pStream storeObject: 'borderMaterial' value: (self borderMaterial) default: (self class defaultBorderMaterial).! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 3/9/2009 10:45'!
placeCard: cd at: indexOrNil scale: scale

	"Do layout placement of the given individual card.
	 Scale is normally one, for a card in place at rest.
	 If scale > 1, then the card will be brought forward and enlarged.
	 Scale is not (should not be) less than one in our usage.
	 (Scale here is a transient multiplier during mouseover enlargement.)"
	| i |
	i := indexOrNil ifNil: [ cards indexOf: cd ].
	i = 0 ifTrue: [ ^ self ]. 
	axis caseOf: {
		[#vertical]->[self verticalCard: cd at: i scale: scale].
		[#horizontal]->[self horizontalCard: cd at: i scale: scale].
	}.
! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 1/28/2008 21:37'!
pointerEnter: ev

	self pointerOver: ev! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 11/13/2008 13:11'!
pointerLeave: ev
	self closeAll.
	"Clear the most recent card floater."
	ev avatar ifNotNil: [
		ev avatar clearFloatMessageForEvent: ev
	]! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 11/13/2008 13:02'!
pointerOver: ev

	ev avatar ifNil: [ ^ self ].
	(self notMyAvatar: ev avatar) ifTrue: [ ^ self ].

	(self cardForEvent: ev) ifNotNilDo: [ : cd | 
		cd possibleTransition: #open.
		cd showFloatFor: ev.
	] ifNil: [
		self closeAll.
		infoMessage ifNotNil:[
			ev avatar showFloatMessage: infoMessage forEvent: ev.
		].
	].! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:56'!
privateScale: aFactor

	"Intercept so we can do geometry management."
	nomScale := aFactor * nomScale.
	self basicExtent: (self basicExtent * aFactor).! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 3/15/2010 14:19'!
redoDrop: capturedCard forUser: anAvatar
	| cd |
	cd := (capturedCard restoreForAvatar: anAvatar) first.
	self insertCard: (cd).
	anAvatar selectFrames: { cd }.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
removeCard: card

	self removeChild: card.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 3/18/2010 09:35'!
removeChild: card

	cards remove: card ifAbsent:[ ].
	super removeChild: card.

	card = openCard ifTrue: [ openCard := nil ].
	self signal: #contentChanged.
	self shapeChanged.! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
removeContent: content
	"Remove the card bearing the given content, if any."

	cards do: [ :c|
		c content = content ifTrue: [
			^ self removeCard: c.
		]
	].
! !

!QDocumentShelf methodsFor: 'render' stamp: 'bgf 11/28/2009 12:09'!
renderChildren: ogl
	"This is a cheap performance hack - simply ignore cards through portals.
	 The problem is that cards and their textures really pile up."
	| cull count |

	frameChildren ifNil: [ ^ 0 ].

	cull := ogl inPortal2D | ogl inPortal3D.
	count := 0.
	1 to: frameChildren size do: [ :i |
		(cull and: [((frameChildren at: i) isKindOf: QFileCard)]) ifFalse: [
			count := count + ((frameChildren at: i) renderFrame: ogl).
		].
	].
	^ count
	! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 22:10'!
size 
	"The number of cards in the rolodex."
	^ cards size! !

!QDocumentShelf methodsFor: 'events' stamp: 'bgf 1/17/2008 22:10'!
spatiallyTransparentToDrag 

	"Special drop-site option - do not pull dragged items 'in front' of the 
	 rolodex when dragged over them; the rolodex hiliting suffices.
	 This makes it easier to drag huge things into the avatar's wallet rolodex."

	^ true! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 4/4/2010 10:58'!
undoDrop: card forUser: anAvatar
	"Remove/destroy a card that was added in a drop gesture by the given user."
	(cards detect: [ : c | card kind = c kind and: [ c content = card content ]] ifNone: [nil])  
	ifNotNil: [ : cd |
		self removeCard: cd.
		cd obliterateLater.
	] ifNil: [
		QLogger logWarning: 'No match to undo drop: ', card.
	].! !

!QDocumentShelf methodsFor: 'initialize' stamp: 'bgf 9/6/2009 12:36'!
updateShape

	|  th ext |

	"Nominal thickness of the surround border."
	th := self borderThickness.

	ext := self extent.
	base extent: ext  + th.
	base translation: 0@0@0.

	cards doWithIndex:[ :card :index |
		(card = openCard) ifFalse:[
			self fileCard: card at: index.
		].
	].

	border extentX: base extent x - 0.01
		     		  y: base extent y - 0.01.
	border radius: th.
	border translation: (base translation - (0@0@(th/4))).

	self updateBounds.! !

!QDocumentShelf methodsFor: 'private' stamp: 'bgf 1/28/2008 11:46'!
verticalCard: cd at: cardIndex scale: scale

	"Place the given card in position, for a #vertical rolodex.
	 The cards are aligned along the center of the rolodex.
	 Assume scale=1 for collapsed and > 1 for expanded."

	| x y z index be |

	"Sleight of hand - we want card 1 at the left, rather than bottom..."
	direction == #left ifTrue: [
		index := cards size + 1 - cardIndex.
	] ifFalse: [
		index := cardIndex.
	].
	be := self effectiveBasicExtent.
	x := 0.
	y := base translation y - (base extent y /2) + (be y * 0.5 * 1.05).
	y := y + ((index-1)* (be y)).
	z  := 0.03 + (scale-1 * 0.1).

	cd translationX: x y: y z: z; 
		extent: ((2-scale * be)+ (scale-1 * basicExtent)) * scale * 0.95.
				"When squeezed to overflow, this animates the card towards the target size."! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 3/22/2010 09:19'!
warpThresholdAdjustment
	"We want to drag/arrange them from afar."
	^ 5! !

!QDocumentShelf methodsFor: 'accessing' stamp: 'bgf 1/17/2008 23:23'!
width
	"The x extent of the shelf proper within the border."
	^ base extent x.! !

QDocumentShelf subclass: #QRolodex
	instanceVariableNames: 'spacing toRight cardRotation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QRolodex commentStamp: '<historical>' prior: 0!
Organizes a set of QFileCards.  
Provides mouse-over "enlarge the focal card" behavior for browsing.  
Accepts drops (copies the dropped card into a new card to retain a reference to its content.)

Visually, a rolodex is a flat rectangle with its cards displayed as spaced
elements, with a slight angle.  There are two orientations: #vertical acts
a bit like one of those time-card racks you see in some workplaces;
#horizontal is a flat platform with the cards standing upright on it.
Default is horizontal.

Rolodexes can be set for use on the right-hand or left-hand side of the viewer,
using 'toRight: true' or false.   This adjusts the card angle so that a viewer
in the favored position has a better view of the card set.

Rolodexes adjust isTranslucent of cards to false when the cards 
are small and inactive, and to true when a card is focused.
This allows easier examination of a set of cards by scanning the cursor 
across them... one by one, the cards become opaque and spring up larger.
!

!QRolodex class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRolodex class methodsFor: 'defaults' stamp: 'bgf 1/18/2008 15:26'!
defaultAxis
	^ #horizontal! !

!QRolodex class methodsFor: 'defaults' stamp: 'bgf 4/18/2007 23:23'!
defaultBasicExtent
	"Default collapsed card size - the basis for rolodex layout."
	^ (1.0@0.667)! !

!QRolodex class methodsFor: 'defaults' stamp: 'bgf 1/18/2008 13:36'!
defaultDirection
	^ #right! !

!QRolodex class methodsFor: 'defaults' stamp: 'bgf 1/12/2008 11:33'!
verticalSlope
	"The forward lean of cards in a vertical rolodex"

	^ -16 "degrees"! !

!QRolodex class methodsFor: 'defaults' stamp: 'bgf 1/12/2008 11:33'!
verticalSlopeSin
	"The sin of vertical slope, which is used a lot in visual animation updates."

	^   0.275637355816999! !

!QRolodex methodsFor: 'accessing' stamp: 'bgf 1/18/2008 15:23'!
autoRotateForAvatar: avatar
	"Orient and spin into position to be easier to traverse by the given avatar."

	| apos rpos bearing yaw delta rotation |

	(avatar notNil and: [avatar root = self root]) ifFalse: [
		^ self
	].

	apos := avatar globalTransform translation.
	rpos := self globalTransform translation.

	"bearingToPoint uses a different basis than TFrame yaw;
	 bring them both into a 0-360 basis, counterclockwise from above."
	bearing := 360 - ((rpos x@rpos z) bearingToPoint: (apos x)@(apos z)).
	yaw := 180 + self globalTransform pitchYawRoll y.
	delta := yaw - bearing.
	
	"Make sure the card edge alignment is OK."

	self toRight: (delta > 0).

	axis = #horizontal ifTrue: [
		"Stay near a 45 degree relative rotation to the avatar viewing angle,
		 so it's easy to scan through the cards."
		rotation :=  (delta sign * 45).
	] ifFalse: [
		rotation := (delta sign * 10).
	].

	self rotateToYaw: (bearing - 180) + rotation.
! !

!QRolodex methodsFor: 'events' stamp: 'bgf 1/6/2008 23:23'!
cardForBasePosition: v3

	"The card to focus on when the pointer is at the given y relative to the base.
	 This is how scanning is managed in #vertical mode."
	|  index |

	(v3 x abs > (base extent x * 0.55)) ifTrue: [ ^ nil ].
	(v3 y abs > (base extent y * 0.55)) ifTrue: [ ^ nil ].

	index := ((v3 y + (base extent y/2.0) - (basicExtent y-(spacing*0.65))) / spacing).
	index := (index truncated) min: (cards size-1) max: 0.
 
	"Sleight of hand...
	 When we lay out the cards, we reverse the indexes so the
	 rolodex grows in the human-expected order...."
	index := cards size - index.

	^ cards at: index ifAbsent: [nil].! !

!QRolodex methodsFor: 'events' stamp: 'bgf 1/19/2008 09:25'!
cardForEvent: ev

	"The card involved in the given (pointer) event."
	| frame |

	axis caseOf: {
		[#horizontal] -> [
			"This is a standing-up coverflow-ish presentation;
			  the card itself, not the effective pick on the backdrop, decides."
			frame := ev selection frame.
			[frame isNil or:[ frame isKindOf: QFileCard]] whileFalse: [
				frame = self ifTrue: [
					 frame := nil 
				] ifFalse: [
					frame := frame parent.
				]
			].
			^frame
		].
		[#vertical] -> [
			^ super cardForEvent: ev
		].
	}.
	^ nil.
! !

!QRolodex methodsFor: 'accessing' stamp: 'bgf 10/28/2006 13:42'!
cardRotation: r

	"Degrees by which the cards are skewed relative to the base.  Default is 15 degrees.
	 The rotation need not be signed differently for left-handed alignment; the placement code does that conversion."
	cardRotation := r.
	self shapeChanged.! !

!QRolodex methodsFor: 'render' stamp: 'bgf 1/24/2008 03:24'!
cardSubdistance: card

	"Support for rendering my file cards.
	 We modulate the ogl sorting distance so that the bottomMost card
	 in the deck sorts closest. Because the cards lean forward, this ensures
	 the card-over-card drawing order matches visual expectations.  	
	 (The default ray distance from camera to center of bottom card
	  is generally greater than the distance to the next one up,
	  but to appear properly in front of that card, we sort it as if it were closer."

	| step |
	step := self isVertical ifTrue: [-10] ifFalse: [10].
	^ (card == openCard) & (self isVertical)
		ifTrue: [ (cards size+1) * (-20 * Float epsilon) ]
		ifFalse: [ (self indexOfCard: card) * (step*Float epsilon) ]! !

!QRolodex methodsFor: 'render' stamp: 'bgf 11/28/2009 11:56'!
child: child setDistance: ogl
	
	"In a rolodex, we use the child/subsorting render tricks
	 to give the rolodex control over the alpha-sort order of cards,
	 so it decides which appears in front.
	 To engage that, we give all cards the identical basic distance (the same as the rolodex distance)."
	^ (frameParent notNil)
		ifTrue: [ (frameParent child: self setDistance: ogl) - Float epsilon]
		ifFalse: [ super child: child setDistance: ogl ]
! !

!QRolodex methodsFor: 'persist' stamp: 'bgf 1/18/2008 13:14'!
constructFrom: pStream

	super constructFrom:pStream.
! !

!QRolodex methodsFor: 'accessing' stamp: 'bgf 10/28/2006 14:32'!
extent

	^basicExtent@(cards size*spacing).! !

!QRolodex methodsFor: 'private' stamp: 'bgf 3/17/2008 01:10'!
horizontalCard: cd at: cardIndex scale: scale

	| edgePt hyp xoff zoff side z by ext index |

	"The cards are aligned along one side of the rolodex,
      so that all can be picked even when big ones are in front of smaller ones.
	 Since the cards are rotated, and have different extents (we respect aspect ratio)
	 there is some trig to be done.  We identify the edge coordinate,
	 and then reckon the center from there."

	"Card 1 goes at  the front, rather than the back."
	index := cards size + 1 - cardIndex.

	ext := basicExtent * scale.
	toRight ifTrue: [side := -1] ifFalse: [side :=1.0].
	by := base translation y.
	z := spacing.
	1 to: index-1 do: [ :i|
		z := z + spacing.
	].

	edgePt := Vector3 x: (basicExtent x/2.0 * side) y: by z: z.

	hyp := ext x/2.0.
	xoff := (cardRotation*side) degreeCos * (hyp * side).
	zoff := (cardRotation*side) degreeSin * (hyp * side negated).

	cd 	extent: ext;
		translationX: (edgePt x - xoff) y: (by+(ext y/2))  z:((edgePt z - zoff) - (self length/2));
		rotationAroundY: (toRight ifTrue: [ cardRotation negated ] ifFalse: [ cardRotation ]);
		allUpdateShape.! !

!QRolodex methodsFor: 'private' stamp: 'bgf 1/12/2008 11:39'!
indexOfCard: cd

	"Array index of the given card in my rolodex, or -1 if it's not present."
	^ cards indexOf:cd ifAbsent: [ -1 ].! !

!QRolodex methodsFor: 'initialize' stamp: 'bgf 1/18/2008 13:56'!
initialize

	| |
	super initialize.
	toRight := false.
	cardRotation := 15.! !

!QRolodex methodsFor: 'persist' stamp: 'bgf 1/18/2008 13:20'!
persistOn: pStream
	super persistOn: pStream.
! !

!QRolodex methodsFor: 'events' stamp: 'bgf 1/18/2008 13:25'!
pointerOver: event

	event avatar ifNil: [ ^ self ].
	(self notMyAvatar: event avatar) ifTrue: [ ^ self ].

	self autoRotateForAvatar: event avatar.
	super pointerOver: event.! !

!QRolodex methodsFor: 'accessing' stamp: 'bgf 8/17/2006 17:09'!
rotateToYaw: yaw

	"Rotate visibly to the given ABSOLUTE yaw."

	| curYaw delta offset |
	curYaw := self globalTransform pitchYawRoll y.
	offset := (self yaw) - curYaw.
	delta := yaw - curYaw.

	(self future:50) rotationAroundY: (offset + curYaw + (delta/4)).
	(self future: 100) rotationAroundY: (offset + curYaw + (delta/2)).
	(self future: 150) rotationAroundY: (offset + curYaw + (3*delta/4)).
	(self future: 200) rotationAroundY: (offset + yaw).! !

!QRolodex methodsFor: 'accessing' stamp: 'bgf 1/6/2008 21:33'!
spacing
	"The current inter card spacing, which is a function of basicExtent and card count"
	^  spacing! !

!QRolodex methodsFor: 'events' stamp: 'bgf 1/13/2007 16:35'!
spatiallyTransparentToDrag 

	"Special drop-site option - do not pull dragged items 'in front' of the 
	 rolodex when dragged over them; the rolodex hiliting suffices.
	 This makes it easier to drag huge things into the avatar's wallet rolodex."

	^ true! !

!QRolodex methodsFor: 'accessing' stamp: 'bgf 10/28/2006 13:42'!
toRight: bool

	"Specify the edge-alignment of the cards in the rolodex.	If toRight is true, then cards are placed with their right-hand edge sitting at the right-hand edge of the rolodex edge; otherwise the left-hand edges align.

	The aligned edge is the one that should be closer to the viewer for ease of use, since card scanning works best from that vantage.

	Note that this property is dynamic.. it is adjusted automatically by openCard:event:"

	(toRight ~= bool) ifTrue: [
		toRight := bool.
		self shapeChanged.
	].! !

!QRolodex methodsFor: 'initialize' stamp: 'bgf 1/18/2008 23:32'!
updateShape

	|  len th |

	"Nominal thickness of the surround border."
	th := self borderThickness.

	"Pack cards in a bit closer as more cards are added."
	spacing := (basicExtent y/2.5) * (18/(cards size max: 18)).

	self isVertical ifTrue: [
		len := basicExtent y + ((cards size-1)*spacing).
	] ifFalse: [
		len := 2*spacing.
		cards do: [ :cd |
			len := len + spacing.
		].
		len := len max: (basicExtent y).
		border rotationAroundX:-90.
		base rotationAroundX:-90.
	].

	base extent: (basicExtent x+th)@(len + th+ th).
	cards doWithIndex:[ :card :index |
		(card = openCard) ifFalse:[
			self fileCard: card at: index.
		].
	].

	border extentX: base extent x - 0.01
		     		  y: base extent y - 0.01.
	border radius: th.
	border translation: (base translation -
						(self isVertical ifTrue:[0@0@(th/3)] ifFalse:[0@(th/3)@0])).

! !

!QRolodex methodsFor: 'private' stamp: 'bgf 1/24/2008 02:59'!
verticalCard: cd at: cardIndex scale: scale

	| x y z index |

	"Place the given card in position, for a #vertical rolodex.
	 The cards are aligned along the center of the rolodex.
	 Assume scale=1 for collapsed and > 1 for expanded."

	"Sleight of hand - we want card 1 at the top, rather than bottom..."
	index := cards size + 1 - cardIndex.
	x := 0.
	z :=  0.02 + (1.1 * (basicExtent y/2.0) * (self class verticalSlopeSin)).

	z := z + ((((scale-1)*(basicExtent y*4))) * (self class verticalSlopeSin)).
				"Pulls the open card forward."
	
	y := base translation y - (base extent y /2) + (cd extent y *0.6).
	y:= y + ((index-1)*spacing).

	cd translationX: x y:y z:z; 
		extent: basicExtent;
		rotationAroundX: (self class verticalSlope* (-3 + (2*scale)))
		"Instead of making the card larger, 
		 scale (which runs from 1 for filed to 1.4 for open)
	       pulls the card forward, and makes it a bit more vertical."
! !

QRolodex subclass: #QStandRolodex
	instanceVariableNames: 'leg mgrip shadow'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QStandRolodex commentStamp: '<historical>' prior: 0!
A Rolodex with a leg supplied by default.

A StandRolodex has direction #vertical by default,
since that orientation is easier to use for a freestanding rolodex.!

!QStandRolodex class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStandRolodex class methodsFor: 'defaults' stamp: 'bgf 1/18/2008 15:26'!
defaultAxis
	^ #vertical! !

!QStandRolodex class methodsFor: 'defaults' stamp: 'bgf 1/10/2007 22:05'!
defaultBasicExtent
	"Default collapsed card size - the basis for rolodex layout."
	^ (1.0@0.667)! !

!QStandRolodex class methodsFor: 'defaults' stamp: 'bgf 1/18/2008 13:36'!
defaultDirection
	^ #right! !

!QStandRolodex class methodsFor: 'defaults' stamp: 'bgf 1/13/2007 00:39'!
defaultSpacing
	^ 0.2! !

!QStandRolodex methodsFor: 'persist' stamp: 'bgf 10/28/2006 13:40'!
constructFrom: pStream
	super constructFrom: pStream.
	self height: (pStream readNumber: 'height' default: self height).
	self shapeChanged.! !

!QStandRolodex methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:49'!
dragAdapter
	^ mgrip! !

!QStandRolodex methodsFor: 'accessing' stamp: 'bgf 10/19/2006 17:39'!
height

	^ leg extent y! !

!QStandRolodex methodsFor: 'accessing' stamp: 'bgf 8/11/2006 16:19'!
height: y

	leg extent: (leg extent x)@y@(leg extent z).
	leg translation: 0@(y negated)@0.! !

!QStandRolodex methodsFor: 'initialize' stamp: 'bgf 8/29/2009 11:36'!
initialize

	super initialize.

	leg := TCylinder new.
	leg height: 3;
		topRadius: 0.08;
		baseRadius: 0.08;
		material: (QMaterials mediumDarkWoodMaterial).
	self addChild: leg.

	(mgrip := QMotionAdapter forLuggage: self grip: leg)
		isGroundBound: true.

	shadow := QMaterials newRadialShadow.
	shadow addRotationAroundX:-90.
	self addChild: shadow.

	self axis: #vertical.
	self shapeChanged.

	! !

!QStandRolodex methodsFor: 'accessing' stamp: 'bgf 8/11/2006 16:24'!
isGroundBound
	^ true.! !

!QStandRolodex methodsFor: 'accessing' stamp: 'bgf 6/2/2009 09:36'!
isPlacementLocked: bool
	"No leg-dragging if locked in place."
	super isPlacementLocked: bool.
	mgrip isEnabled: (bool not).! !

!QStandRolodex methodsFor: 'persist' stamp: 'bgf 10/19/2006 17:45'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeNumber: 'height' value: self height default: 3.! !

!QStandRolodex methodsFor: 'initialize' stamp: 'bgf 2/11/2010 15:58'!
privateScale: aFactor
	self height: (self height * aFactor).
	super privateScale: aFactor.! !

!QStandRolodex methodsFor: 'initialize' stamp: 'bgf 6/27/2007 18:35'!
updateShape

	leg ifNotNil: [
		"When vertical, shift so the leg is behind the base."
		leg translationX: 0 y: -0.05 z: (self isVertical ifTrue:[-0.1] ifFalse:[-0.05]).
		base translationX: 0 y: (leg extent y) z: 0.
	].

	shadow translationX: 0 y: 0.01 z: 0.
	shadow extent: base extent.
	super updateShape.
	! !

TGroup subclass: #QEditBox
	instanceVariableNames: 'contents extent outerExtent dragFaces rotateEdges scaleCorners dragIcon ringIcon scaleIcon borderSize border selectedFrame lastSpin rotPoint boxPoint rotateDirection selectedPoint selectedNormal box scaleFrame justAClick baseScale doingInteraction savePoint shadow shadowRay grid initialState latestState undoVerb'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QEditBox commentStamp: '<historical>' prior: 0!
QEditBox allows the user to arbitrarily move, rotate, and scale the contents of the object.

Keys:

	u - move up the hierarchy (space not included).
	d - move to the first child (if any)
	l - move to my peer to the left
	r - move to my peer to the right
	x - cut the contents
	c - copy the contents
	v - paste the contents
	backspace - remove this edit box
	delete - remove edit box and contents!

!QEditBox class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEditBox class methodsFor: 'undo-redo' stamp: 'bgf 3/18/2010 09:58'!
applyEdit: tframes state: ebState forUser: anAvatar
	| box |
	tframes size > 1 ifTrue: [
		box := tframes first parent editBoxAroundMe.
		box ifNotNil: [
			(tframes allSatisfy: [ : f | f parent = tframes first parent ]) ifFalse: [ box := nil ].
		].
	] ifFalse: [
		box := tframes first editBoxAroundMe.
	].

	"Undo support for an edit box transition."
	box ifNotNil: [
		box applyStateVector: ebState
	] ifNil: [
		tframes with: ebState fourth do: [ : f : frameState |
			f nominalScale: (ebState first * (frameState first));
			  localTransform: (ebState second * frameState second);
			  translation: (frameState third)
		].
		anAvatar selectFrames: tframes
	].! !

!QEditBox class methodsFor: 'defaults' stamp: 'bgf 1/21/2010 14:46'!
defaultDoesClickSelect
	^ false
! !

!QEditBox class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:20'!
defaultEventMask
	^ EventPointerOver + EventPointerDown + EventKeyboard! !

!QEditBox class methodsFor: 'undo-redo' stamp: 'bgf 3/18/2010 09:08'!
editContents: tframes oldState: oState newState: nState description: descrip
	"Make the undo-redo record for one change to the edit box."

	^ QBlockAction
		undoBlock: [ : avatar |
			avatar island future 
			  in: QEditBox perform: #applyEdit:state:forUser:  withArguments: { tframes . oState . avatar} ]
		redoBlock: [ : avatar |
			avatar island future 
			  in: QEditBox perform: #applyEdit:state:forUser: withArguments: { tframes . nState . avatar } ]
		description: descrip
		requiredFrames:  tframes ! !

!QEditBox class methodsFor: 'materials' stamp: 'bgf 11/27/2009 17:59'!
faceMaterial
	"The  box faces."
	|c|
	c := QColors editboxFaceColor.
	^ (TMaterial new)
		ambientColor: c/3;
		diffuseColor: c;
		emissiveColor: (c/2);
		flipFace: true;
		yourself! !

!QEditBox class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:00'!
indicatorMaterial
	"Used by each active feedback indicator (rotation ring, scale boxes, etc.)"
	| c |
	c := QColors editboxIndicatorColor.

	^ (TCube defaultTransparentMaterial)
		ambientColor: (c * 0.25);
		diffuseColor: (c);
		emissiveColor: (c * 0.75);
		cullFace: true;
		fullBright: true;
		yourself! !

!QEditBox class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:04'!
menuStyle
	"TMenuStyle for editing tool menus -
	 use this style to tie things together visually."
	| c |
	c := QColors editboxCornerColor asColor.

	^ (QStyles menuStyle)
		borderColor: (c duller alpha:0.6);
		separatorColor: (c lighter duller alpha:0.5);
		disabledTextColor: (c muchDarker);
		selectedItemColor: (c lighter);
		yourself.! !

!QEditBox class methodsFor: 'materials' stamp: 'bgf 11/27/2009 20:33'!
shadowMaterial
	"The conic drag-shadow color."
	| shade |
	shade := (QColors editboxShadowColor).
	^ (TMaterial new)
		texture: (QTextures find: 'LinearShadow.png');
		ambientColor: shade/2;
		diffuseColor: shade;
		specularColor: shade/5;
		emissiveColor: shade;
		cullFace: false;
		yourself! !

!QEditBox class methodsFor: 'utilities' stamp: 'bgf 4/9/2010 13:46'!
snapVectorToIntegers: vector ifWithin: epsilon
	
	"Destructively alter the given vector by replacing any components in with integers
	 when those components are within the given epsilon of an integer."

	vector doWithIndex:  [ : comp : idx | 
		(comp - (comp rounded)) abs <= epsilon ifTrue: [
			vector at: idx put: (comp rounded)
		]].! !

!QEditBox class methodsFor: 'defaults' stamp: 'bgf 4/9/2010 13:53'!
unitDetente
	"The limit for doing snap-to detentes when dragging near a unit coordinate."
	^ 0.045! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 2/26/2010 18:03'!
actionsList

	"This is the QInspector action list... useful in emergencies."
	^{
		{#dismiss . 'Dismiss Editor'}.
		}! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 5/20/2008 22:14'!
activityDetails
	"For details: in QActivityCapture items."
	^ 'EditBox: ', (self contents ifNotNilDo: [ :f | f activityDetails ] ifNil: [ 'empty' ])! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 8/24/2010 10:23'!
addMenuActions: aTMenu forEvent: event
	"Populate the custom context menu."
	| mask |
	mask := contents isFrame 
				ifTrue: [ contents contextMenuMaskForAvatar: (event avatar) ] 
				ifFalse: [ #() ].

	(mask includes: #material) ifTrue: [
		aTMenu addItemSpec: {'Materials...' . #editMaterial: } target: self.
	].
	(mask includes: #properties) ifTrue: [
		aTMenu addItemSpec: { 'Properties...' . #editProperties: } target: self.
	].
	(contents respondsTo: #addContainedMenuActions:forEvent:) ifTrue: [
		aTMenu addSeparator.
		contents addContainedMenuActions: aTMenu forEvent: event.
	].
	aTMenu addSeparator.

	(aTMenu addItemSpec: {'Reset Rotation' . #resetRotation} target: self)
		disabled: (self orientation = Matrix4x4 new).

	aTMenu addSeparator.
	(self parent isSpace or: [ self parent isPlacementLocked ]) ifFalse: [
		aTMenu addItemSpec: { 'Select Parent' . #selectOutward: } target:self.
	].
	contents isEditableGroup ifTrue: [
		contents frameChildren ifNotNil: [
			aTMenu addItemSpec: { 'Select First Child' . #selectFirstChild: } target: self.
		].
	].
	self parent ~= self root ifTrue: [
		self nextVisiblePeer ~= self ifTrue: [
			aTMenu addItemSpec: {'Select Next' . #selectNextPeer: } target: self.
		].
		self previousVisiblePeer ~= self ifTrue: [
			aTMenu addItemSpec: {'Select Previous' . #selectPreviousPeer: } target: self.
		].
	].
	aTMenu addSeparator.
	(mask includes: #copyToInventory) ifTrue: [
		aTMenu addItemSpec: {'Copy to Inventory' . #copyToInventory: } target: self.
		aTMenu addSeparator.
	].

	aTMenu addItemSpec: {'Delete' . #confirmDelete:} target: self.
	aTMenu addSeparator.

	initialState ifNotNil: [
		aTMenu addList: {
			{ 'Cancel (Revert)' . #undoAllChanges: }.
			#- .
			{'Done'. #dismiss:} .
		} target: self. 
	] ifNil: [
		aTMenu addList: {
			{'Done'. #dismiss:} .
		} target: self. 
	].! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/10/2010 09:33'!
alignIndicator: dragIndicator toNormal:  aNormal
	| gvector abs orient |
	"Align the supplied drag-indicator to the plane of motion for dragging the selectedFace;
	 this is the nearest ortho plane (in parent frame) to the face-plane."

	self root ifNil: [ ^ nil ].		"Destroy / interaction race guards."
	contents ifNil: [ ^ nil ].
	gvector := (self orientation transformVector: aNormal).
	abs := gvector abs.
	orient := self parent globalOrientation.
	orient := (abs max caseOf: {
		"The x=z checks resolve the ambiguity at the 45-degree points in y."
		 [abs x] -> [
				((abs x ~= abs z) or: [ gvector z ~= gvector x ])
					ifTrue: [ orient * (Matrix4x4 new rotationAroundY: 90 ) ]
					ifFalse: [ orient ] 
				    ].
		 [abs y] -> [ orient * (Matrix4x4 new rotationAroundX: 90 ) ]. 
		 [abs z] -> [ orient  ].
		} otherwise: [
			^ self
		]).
	dragIndicator orientation: (dragIndicator parent globalOrientation inverse) * orient.! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/17/2010 14:06'!
announceForUndo: event
	"If the most recent undo is more than a no-op,
	 then announce it to the undo-mechanism."
	| state |
	latestState ifNil: [ ^ self ].
	event avatar ifNil: [ ^ self ].

	state := self stateVector.
	(state hasEqualElements: latestState)  ifTrue: [ ^ self ].

	"Register the undoable step"
	(contents notNil and: [ self root notNil ]) ifTrue: [
		event avatar 
			recordAction: #editContents:oldState:newState:description:
			class: QEditBox
			arguments: { 
				(contents isKindOf: QTemporaryGroup) ifTrue: [ contents children ] ifFalse: [ { contents } ].
				 latestState .  
				state .
				(undoVerb ifNil: [ 'Adjust' ])  }.
	].! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/17/2010 11:57'!
applyStateVector: stateVec
	"For undo.  The state of the edit box itself is a triplet: 
		(scale,  localTransform, and bounding box extent).
	 This resets the edit box to a previously-captured state."

	extent := (stateVec third).
	scaleFrame nominalScale: (stateVec first / (contents nominalScale)).
	self extent: (stateVec third).
	self localTransform: (stateVec second).
	self updateShadow.
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/26/2010 18:05'!
cancelInteraction
	"Interrupt, undo, and stop the current scale/move/rotate interaction;
	 this is the <escape> key handler."

	doingInteraction ifTrue: [
		latestState := nil.
		doingInteraction := false.
	].! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 11/7/2009 15:44'!
childrenForBrowser
	"Scene browsers see the edit box and its contents as a single item."
	^ contents
		ifNil: [ #() ]
		ifNotNil: [ contents childrenForBrowser ].! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 3/27/2009 10:07'!
component
	"Break the upward recursion - the thing inside me has to be treated as the editable part."
	^ contents ifNil: [ self ]! !

!QEditBox methodsFor: 'persist' stamp: 'bgf 1/20/2010 09:21'!
constructFrom: pStream
	"Persist the receiver"
	| trans scl |
	super constructFrom: pStream.
	trans := self localTransform.
	self edit: (pStream readObject: 'contents' default: self contents) forAvatar: nil.
	"fix up for new, improved scale management"
	scl := (pStream readNumber: 'scale' default: self scale).
	scaleFrame nominalScale: scaleFrame nominalScale*scl.
	contents ifNotNil: [
		contents nominalScale: 1.0.
	].
	self extent: (pStream readVector3: 'extent' default: self extent).
	self localTransform: trans.! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/17/2010 14:44'!
contentPropertiesChanged
	"Followup from prop-editor apply... redo my shrinkwrap."
	contents ifNotNil: [ self updateBox ].! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 1/21/2007 10:46'!
contents
	"The frame being edited."
	 ^ contents! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 4/5/2007 10:25'!
contents: aFrame
	"The frame being edited."
	contents := aFrame.
	contents ifNotNil: [
		scaleFrame addChild: contents.
	].! !

!QEditBox methodsFor: 'initialize' stamp: 'das 5/27/2007 13:54'!
contentsDeleted

	"Clear the now deleted contents and then kill the box.
	 and destroy the edit box itself."
	
	contents:= nil.
	(self future:500) destroy.	"After event deliveries."! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 11/7/2009 15:44'!
contextMenuMaskForAvatar: avatar
	"Remove resize from the menu. (Just in case.)"
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	mask remove: #resize ifAbsent: [].
	^ mask! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 1/20/2010 09:26'!
copyToInventory: event

	"Copy the selected contents to the avatar's wallet rolodex."
	| frame |
	event avatar ifNil: [ ^ self ].

	"Temporarily unbox the contents, so that the size we record in the inventory
	 is the current size.  (Otherwise, this is lost, since the scaleframe doesn't modify 
	 the content until it is unboxed.)"

	frame := contents.
	frame ifNil: [ ^ self ].
	self outOfBox.
	frame copyToInventory: event.		
			"Dispatch to the target, so posters and cards are copied intelligently;
			 this makes edit-box copy-to-inventory consistent with context-menu CTI.
			Do -not- pass the avatar so we don't redirect to a different group-selection."
	self edit: frame forAvatar: nil.

! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 10/20/2009 14:17'!
defaultDestinationTransform
	"For 3d-model locate, make this give the same 'Locate...' result 
	 whether or not the model is in a box."
	contents ifNotNil: [
		^ contents defaultDestinationTransform
	].
	^ super defaultDestinationTransform! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:10'!
defaultIsAtomic
	^ false! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 8/1/2009 22:13'!
deletionLabel
	^  contents ifNotNil: [ contents deletionLabel ]
				ifNil: [ 'Edit Box' ]! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 6/23/2009 12:45'!
destroy
	"The is not my child; it is usually unrooted.
	 Destroy it to clean up, and to prevent orphans
	 when  overlapping events of multiple users delete me in mid-drag."
	grid ifNotNil: [
		grid destroy.
		grid := nil.
	].
	super destroy.! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/10/2010 11:53'!
detenteAngle: rawTheta forEvent: event
	"Constrain to whole-number degrees. 
	 (Until we sort out how to detente the one angle we're moving, 
	 as it resolves out after applying to a rotated matrix, this simply wheel-detente is best we can do.)"
	|theta|
	theta := rawTheta roundTo: 1.0.
	^ theta! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 12/5/2009 13:03'!
dismiss

	"Extract contents back into their context (if any)
	 and destroy the edit box itself."

	self outOfBox.
	self obliterateLater.! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 1/29/2010 15:38'!
dismiss: event
	"Extract contents back into their context (if any) and destroy the edit box itself."
	| |
	event avatar ifNotNil: [
		QActivityCapture 
				action: #DismissEditBox
				details: self activityDetails
				forSession: (event avatar sessionHex).
	].
	self dismiss.
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 4/12/2010 11:35'!
dragBox: event

	| delta  pointer n |

	pointer := event makePointer.

	n := self showGridForNormal: selectedNormal at: (self translationForContent).
	n ifNil: [ ^ false ].
	n := self orientation inverse transformVector: n.		"Use ortho plane."

	(pointer frame: self pickPlane: selectedPoint normal: n) 
	 ifTrue:[
		| trans |
		event avatar laserShow: (self globalTransform transformPoint: (pointer selectedPoint)).
		"Arbitrarily cap motion to prevent union sphere space death."
		delta := (selectedPoint - pointer selectedPoint).
		delta abs max > 1000 ifTrue: [
			delta := (selectedPoint - pointer selectedPoint) * 1000/(delta abs max).
		].
	
		trans := (self translation - (self orientation transformPoint: delta)).

		trans := self translationBoxToContent: trans.
		event shiftPressed ifFalse: [
		 	self class snapVectorToIntegers: trans ifWithin: (self class unitDetente)
		].

		event avatar 
			cornerInfo: '', (trans x roundTo: 0.1), ' @ ', (trans y roundTo: 0.1), ' @ ', (trans z roundTo: 0.1).

		trans := self translationContentToBox: trans.
		self translation: trans.
		self updateShadow.
		^ true

	] ifFalse: [
		self noGrid.
		^ false.
	].! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 5/24/2010 09:56'!
edit: conts forAvatar: anAvatarOrNil

	(conts isKindOf: QEditBox) ifTrue: [ ^ self ].

	contents = conts ifTrue:[^ self ].
	contents ifNotNil:[ 
		self stopScript: #contentsDeleted.  
		self stopScript: #contentPropertiesChanged.
		self outOfBox  ].
	contents := conts.
	self runScript: #contentsDeleted			when: { contents. #destroyed.}.
	self startScript: #contentPropertiesChanged 	when: { contents . #propertiesChanged }.
	contents ifNotNil: [	
		self intoBox.
	].

	anAvatarOrNil ifNotNil: [ 
		anAvatarOrNil selectFrame: contents
	].

	TObjectID for: 'E-Box activated.'.! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 11/7/2009 19:18'!
editBoxAround: aFrame
	"Private TFrame protocol... 
	 Shortcut the generic search here, since I'm probably the box we're looking for."
	^ contents = aFrame 
		ifTrue: [ self ] 
		ifFalse: [ super editBoxAround: aFrame ]! !

!QEditBox methodsFor: 'editing' stamp: 'bgf 1/25/2010 11:54'!
editClass
	^ nil
	"Do nothing... base-case protection against 3-D editing an edit-box."! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 1/30/2010 14:07'!
editGroup: frames inParent: parent forAvatar: anAvatarOrNil
	"Worker routine for putting an edit box around a group of several separate frames;
	 we enclose them in a temporary group for the duration of the edit-boxing session."

	| group idx  |
	group := QTemporaryGroup new.
	idx := parent indexOfChild: frames first.
	frames do: [ :f |
		group addChild: f.
	].
	parent addChild: group atIndex: idx.

	^ self edit: group forAvatar: anAvatarOrNil
! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 7/26/2010 10:49'!
editMaterial: event
	"Edit the material of the contents."

	event avatar signal: #editFrameMaterial with: self contents.! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 1/20/2010 11:34'!
editProperties: event
	"Open property sheet of contents."
	event avatar ifNil: [ ^ self ].
	contents ifNotNil: [
		contents editProperties: event
	].
	! !

!QEditBox methodsFor: 'accessing' stamp: 'hrs 10/2/2007 14:23'!
editableObject
	self contents ifNil: [^self].
	^self contents editableObject.! !

!QEditBox methodsFor: 'accessing' stamp: 'das 12/14/2006 16:54'!
editor: editorClass

	"do not apply editors to editors"! !

!QEditBox methodsFor: 'accessing' stamp: 'ar 1/26/2007 15:35'!
extent
	^extent! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 4/1/2010 12:39'!
extent: ext
	| mid |

	extent := ext max:(0.3 @ 0.3 @ 0.3).  " Big enough to get hold of."

	"Scale the border to the sqrt of the content extent; this allows
 	 the eye to read scale changes properly, rather than as receding/approaching,
	 while keeping the border thick enough to be usable for large objects.
	 We use the middle dimension, so the extent suits (for example) posts and sheets,
	 as well as more cubical forms."

	mid := extent middleDimension.

	border := (borderSize * (mid sqrt)).
	outerExtent := extent + (border*0.5).

	self updateParts.! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 6/14/2009 03:41'!
grid
	"My QGrid, which may or may not be visible."
	^ grid ! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 11/1/2009 14:52'!
hilite: aColor

	"Make it clear that the box is hilit,
	 but don't obscure the contents too much."
	contents ifNotNil: [
		contents hilite: aColor
	].

! !

!QEditBox methodsFor: 'events' stamp: 'bgf 1/20/2010 11:26'!
inform: event

	contents ifNotNil: [
		event avatar signal: #infoMessage with: 'Selected: ', (contents menuTitle).
	]! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 2/26/2010 18:03'!
initialize
	super initialize.

	borderSize := 0.125.	"Basis for widths of the unseen pickable edges and corners."
	self makeParts.
	self extent: 2@2@2.
	scaleFrame := QEditBoxOutline new.
	scaleFrame lineWidth: 3.1; 
				lineColor: (QColors editboxWireColor);
				hiddenLineWidth: 2.0;
				hiddenLineColor: (QColors editboxHiddenWireColor);
				matte: (self class faceMaterial);
				pickable: false.	"This prevents interaction with the contents while boxed."
	
	self addChild: scaleFrame.
	doingInteraction := false.
! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 3/24/2010 11:20'!
intoBox

	"Absorb the contents into the edit box,
	 splicing the edit box into the scene graph in its place."

	box:= contents editorBoundingBox copy.

	self localTransform: (contents localTransform).
	self translation: self translation + (self orientation transformPoint:box center).
	contents translation: (box center negated); 
			 pitchYawRoll: (Vector3 zero).
	contents parent ifNotNilDo: [ : p | | idx| 
		idx := p indexOfChild: contents.
		p removeChild: contents.
		p addChild: self atIndex: idx.
		].

	scaleFrame nominalScale: 1.0.
	scaleFrame addChild: contents.
	
	self extent: box extent.
	self updateShadow.

	scaleCorners do: [ : corner | corner pickable: (contents canScale) ].
	self root ifNotNil: [ : spc |
		spc future signal: #intoEditBox with: self.
		initialState := (self stateVector).
	].! !

!QEditBox methodsFor: 'testing' stamp: 'bgf 11/7/2009 15:40'!
isEditBoxFor: aFrame
	"TFrame protocol.  This makes the scene-graph aware that I'm in control of the contents."
	^ aFrame = contents! !

!QEditBox methodsFor: 'events' stamp: 'bgf 2/26/2010 18:04'!
keyStroke: event

	| c |
	"Escape interrupts/resets during a pointer-interaction;
	 so if you scale into unrecoverable limbo, you can bail out"
	doingInteraction ifTrue: [
		event keyValue = 27 "escape" ifTrue: [
			^ self cancelInteraction.
		].
	].

"
	127 - delete
	8 - backspace
	13 - enter
	32 - space.
"
	c := event keyCharacter.
	event buttons = 80 ifFalse:["not control keys"
		event keyValue = 8 ifTrue:[
			contents ifNotNil:[
				^ self dismiss: event.
			].
		].
		c = $u ifTrue:[ "move to parent"
			^ self selectOutward: event.
		].
		c = $d ifTrue:[ "move to child"
			^ self selectFirstChild: event.
		].
		c = $[ ifTrue:[ 
			^ self selectPreviousPeer: event.
		].
		c = $] ifTrue:[ 
			^ self selectNextPeer: event.
		].
		c = $x ifTrue:[self addRotationAroundX:1. ^ self. ].
		c = $X ifTrue:[self addRotationAroundX:-1. ^ self. ].
		c = $y ifTrue:[self addRotationAroundY:1. ^ self. ].
		c = $Y ifTrue:[self addRotationAroundY:-1. ^ self. ].
		c = $z ifTrue:[self addRotationAroundZ:1. ^ self. ].
		c = $Z ifTrue:[self addRotationAroundZ:-1. ^ self. ].

		c = $i ifTrue:[ ^ self resetRotation ].
	].! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 6/10/2009 14:19'!
labelForBrowser
	^ contents ifNil: [ '3D Editor' ] ifNotNil: [ contents labelForBrowser, ' [In 3D Editor]']! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 7/1/2010 12:13'!
makeParts
	|  edge corner cm |
	cm := QMaterials clearMaterial.

	"These faces, edges, and corners are completely transparent;
	 they exist to catch the various picks that lead to different size/move/rotate interactions.
	 (The appearance of the edit-box is supplied by the scaleFrame,  an OutlineGroup.)"

	dragFaces := Array new: 6.
	1 to: 6 do: [  :i | |f| 
		f := TRectangle new.
		self addChild: f.
		dragFaces at: i put: f.
		f objectName: #dragBox:.
		f material: cm  ].
	
	(dragFaces at: 2) rotationAroundY: 90.		
	(dragFaces at: 3) rotationAroundY: 180.		
	(dragFaces at: 4) rotationAroundY: -90.		
	(dragFaces at: 5) rotationAroundX: 90.		
	(dragFaces at: 6) rotationAroundX: -90.		

	rotateEdges := Array new:12.
	1 to:4 do:[:index |
		edge := TCube new. 
		self addChild: edge. 
		edge objectName:#rotateX:. 
		rotateEdges at:index put:edge.
		edge material: cm.
	].
	1 to:4 do:[:index |
		edge := TCube new. 
		self addChild: edge. 
		edge objectName:#rotateY:. 
		rotateEdges at:index+4 put:edge.
		edge material: cm.
	].
	1 to:4 do:[:index |
		edge := TCube new. 
		self addChild: edge. 
		edge objectName:#rotateZ:. 
		rotateEdges at:index+8 put:edge.
		edge material: cm.
	].

	scaleCorners := Array new:8.
	1 to:8 do:[:index |
		corner := TCube new. 
		self addChild: corner. 
		corner objectName:#scaleBox:. 
		scaleCorners at:index put:corner.
		corner material: cm.
	].
	frameChildren do:[:fc | fc solid: false.].

	ringIcon := QEditRingIcon new.
	ringIcon objectName:#ring.

	dragIcon := QEditDragIcon new.
	dragIcon objectName: #drag.

	scaleIcon := QEditScaleIcon new.
	scaleIcon objectName: #scale.

	shadow := TConic new.
	shadow baseRadius:0.0.
	shadow capped: false.
	self addChild: shadow.
	shadowRay := TRay new.
	shadowRay isDownRay: true.
	self addChild: shadowRay.
	
	shadow objectName:#moveBox: ;
		solid:false;
		material: self class shadowMaterial;
		uvScale: 1@0.75. "apply uvScale after material"

	grid := (QGrid new).
	grid extent: 100@100.
	grid shouldPersist: false.
	grid material 
			diffuseColor: (QColors editboxGridColor).! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 9/14/2007 21:28'!
menuStyle 
	"TMenuStyle for popup."

	^ self class menuStyle! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 9/14/2007 21:01'!
menuTitle

	^ contents ifNil: [ super menuTitle ] 
			ifNotNil:[ contents menuTitle ]! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 5/5/2010 10:32'!
moveBox: event
	"Called via perform for pointer-move events on the move-shadow."
	| delta  pointer trans1 trans norm |

	pointer := event makePointer.

	trans := self translationForContent.
	trans := (trans x) @ (self translation y) @ (trans z).
	norm := self showGridForNormal: (self orientation transformVector: (0@1@0)) at: trans.
	norm ifNil: [ ^ false ].

	(pointer frame: self 
			pickPlane: selectedPoint 
			normal: ((self orientation inverseByTranspose transformPoint:(0@1@0)))) 
	 ifTrue:[
		event avatar laserShow: (self globalTransform transformPoint: (pointer selectedPoint)).
		delta := selectedPoint - pointer selectedPoint.
		trans := (self translation - (self orientation transformPoint: delta)).

		"Confine the x and z to integers if nearby..."
		trans := self translationBoxToContent: trans.
		event shiftPressed ifFalse: [
			| epsilon | 
			epsilon := (self class unitDetente).
			(trans x - (trans x rounded)) abs <= epsilon ifTrue: [ 	trans x: (trans x rounded) ].
			(trans z - (trans z rounded)) abs <= epsilon ifTrue: [ 	trans z: (trans z rounded) ].
		].

		trans1 := self translationContentToBox: trans.
		self xzGridAt: (trans x) @ (trans1 y - (shadow height*0.99)) @ (trans z).
		event avatar cornerInfo: '', (trans x roundTo: 0.1), '  @  ', (trans z roundTo: 0.1).

		self translation: trans1.
		self updateShadow.

		^ true.].

	^ false.! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/2/2010 17:48'!
noGrid
	"Retract the grid."
	grid ifNotNil:  [
		grid parent ifNotNil: [ grid parent removeChild: grid ].
		grid crosshairOrigin: nil.
	]	! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 1/29/2010 15:40'!
outOfBox

	"Put the content back into its original place in the scene,
	 leaving the edit box unrooted."

	contents ifNil:["bullet-proof for nil contents"
		frameParent ifNotNil:[self parent removeChild: self].
		^self].

	self root miramarUI ifTrue:[
		| wrapper mat |
		contents nominalScale: contents nominalScale * scaleFrame nominalScale.
		mat := contents globalTransform scaleBy: (1.0 / contents cumulativeScale).
		(contents isKindOf: TGeometry) ifTrue:[
			contents localTransform: mat.
			self parent addChild: contents.
			contents freeRange: true.
		] ifFalse:[
			wrapper := MViewWrapper new.
			wrapper addForumsObject: contents.
			wrapper localTransform: mat.
			self parent addChild: wrapper.
		].
		contents updateBounds.
	] ifFalse:[
		self transferPose.
		self parent addChild: contents atIndex: (self parent indexOfChild: self).
		self parent removeChild: self.
	].
	contents restoreFromEditing.

	scaleFrame nominalScale: 1.0.
	selectedFrame ifNotNil:[selectedFrame unhilite.].
	selectedFrame := nil.
	contents := nil.! !

!QEditBox methodsFor: 'menus' stamp: 'bgf 2/4/2010 12:14'!
performDelete: event
	"Safe User-delete of the edit box and its contents."
	event avatar ifNotNil: [
		QActivityCapture 
					action: #DeleteEditBox
					details: self activityDetails
					forSession: (event avatar sessionHex).
	].

	contents 
		ifNil: [ ^ self reallyDelete: event  ]
		ifNotNil: [ 
			self visible: false.
			self transferPose.			"So undoing the delete gets the final pose."

			"This is partial out-of-box, omitting the child notification,
	 		  so that the temporary group receives the performDelete: without having disposed itself."
			self parent addChild: contents atIndex: (self parent indexOfChild: self).
			self parent removeChild: self.
			contents allUpdateShape.
			contents performDelete: event.
		]

! !

!QEditBox methodsFor: 'persist' stamp: 'ar 1/26/2007 15:37'!
persistOn: pStream
	"Persist the receiver"
	super persistOn: pStream.
	pStream storeNumber: 'scale' value: self scale default: (1@1@1).
	pStream storeVector3: 'extent' value: self extent default: nil.
	pStream storeObject: 'contents' value: self contents default: nil.
! !

!QEditBox methodsFor: 'events' stamp: 'bgf 4/15/2010 09:51'!
pointerDown: event

	justAClick := true.			"until proven otherwise by a mouse move"

	self pointerOver: event. 	"sets selection and icons for THIS user"
	selectedFrame ifNil: [
		^self 
	].

	undoVerb := selectedFrame objectName caseOf: {
		[ #dragBox:]->['Move'] .
		[ #rotateX: ] ->['Rotate'] .
		[ #rotateY: ] ->['Rotate'] .
		[ #rotateZ: ] ->['Rotate'] .
		[ #scaleBox: ] -> [ 'Scale' ] .
		} otherwise: [ 'Move' ].

	baseScale := scaleFrame nominalScale.
	self pushLatestState..
	savePoint := event selection point.
	selectedFrame = shadow 
				ifTrue: [ shadow hilite: QColors editboxShadowDownColor ].
	selectedNormal := selectedFrame orientation transformPoint:  (
							event selection normal ifNil: [ selectedFrame lookAt ]).
	selectedPoint := selectedFrame localTransform transformPoint: event selection point.

	boxPoint := (selectedFrame localTransform transformPoint: event selection point).
	rotPoint := nil.
	doingInteraction := true.
! !

!QEditBox methodsFor: 'events' stamp: 'das 12/10/2006 16:37'!
pointerEnter: event

	self pointerOver: event.! !

!QEditBox methodsFor: 'events' stamp: 'bgf 11/15/2009 14:47'!
pointerLeave: event

	| lastSelection lastIcon id |
	self inform: event.

	self noGrid.
	currentUser ifNotNil:[
		id := event userID.
		id ifNil:[ id := currentUser userOver. ].
		id ifNil:[ ^self ].

		lastSelection := currentUser at: id.
		lastSelection ifNotNil:[
			lastSelection frame ~= event selection frame ifTrue:[
				lastSelection frame ifNotNilDo: [ :f | f unhilite. ]
			].
			lastIcon := lastSelection userInfo.
		].
	].

	selectedFrame := event selection frame.
	(selectedFrame isNil or: [selectedFrame objectName = #shadow]) ifTrue:[
		selectedFrame :=nil.
		lastIcon ifNotNil:[lastIcon delete.].
		^ self.
	].

! !

!QEditBox methodsFor: 'events' stamp: 'bgf 11/27/2009 18:00'!
pointerMove: event

	doingInteraction ifFalse:[
		^ self.
	].
	selectedFrame~= event selection frame ifTrue:[.
		selectedFrame ifNotNil:[selectedFrame unhilite.].
	].
	selectedFrame := event selection frame.
	selectedFrame = shadow 
				ifTrue: [ shadow hilite: QColors editboxShadowDownColor ].

	selectedFrame objectName ifNotNil:[ 
		(self respondsTo: (selectedFrame objectName)) ifTrue: [
			self perform: selectedFrame objectName with: event.
		].
	].
	event selection userInfo: (currentUser userInfoAt: event userID).
	event selection point: savePoint.

	justAClick := false.! !

!QEditBox methodsFor: 'events' stamp: 'bgf 4/15/2010 09:52'!
pointerOver: event

	| ext lastSelection lastIcon icon dist |

	self inform: event.
	self noGrid.

	lastSelection := currentUser at: event userID.
	lastSelection ifNotNil:[
		lastSelection frame ~= event selection frame ifTrue:[
			lastSelection frame ifNotNilDo: [ :f | f unhilite. ]
		].
		lastIcon := lastSelection userInfo.
	].

	selectedFrame := event selection frame.
	(selectedFrame isNil or: [selectedFrame objectName = #shadow]) ifTrue:[
		selectedFrame :=nil.
		lastIcon ifNotNil:[
			lastIcon delete.	
			"See above... we got lastIcon out of lastSelection."	
			lastSelection userInfo: nil ].
		^ self.
	].

	selectedPoint := event selection point.
	selectedFrame = shadow  ifTrue: [ shadow hilite: QColors editboxShadowOverColor ].
	selectedNormal := selectedFrame orientation transformPoint:  (
							event selection normal ifNil: [ selectedFrame lookAt ]).

	"Distance, adjusted for the viewarea, so the indicators are vaguely fixed-size."
	dist := (event selection distance min: 200 max: 0.8)  * 300.0
			/ (event avatar viewExtent minDimension max: 300).

	selectedFrame objectName caseOf:{
		[#dragBox:]->[ 
			lastIcon ifNotNil:[ 
				lastIcon objectName = #drag ifTrue:[icon := lastIcon.]
				ifFalse:[lastIcon delete.].
			].
			icon := icon ifNil:[dragIcon duplicate.].
			event selection userInfo: icon.
			selectedFrame addChild: icon.
			icon translation: selectedPoint + (0@0@0.01).
			self alignIndicator: icon toNormal: selectedNormal.
			ext := dist sqrt * 0.32.
			icon extent: ext@ext.
		].
		[#rotateX:]->[ event selection userInfo: (self ringX: selectedPoint x-border icon: lastIcon).].
		[#rotateY:]->[ event selection userInfo: (self ringY: selectedPoint y-border icon: lastIcon).].
		[#rotateZ:]->[event selection userInfo: (self ringZ: selectedPoint z-border icon: lastIcon). ].
		[#scaleBox:]->[ 
			lastIcon ifNotNil:[ 
				lastIcon objectName = #scale ifTrue:[icon := lastIcon.]
				ifFalse:[lastIcon delete.].
			].
			icon := icon ifNil:[scaleIcon duplicate.].
			self addChild: icon.
			event selection userInfo: icon.
			ext := dist sqrt * 0.1.
			icon translation: selectedFrame translation 
				 extent: (ext @ ext @ ext).
		].
		[#moveBox:]->[ 
			lastIcon ifNotNil:[ 
				lastIcon objectName = #drag ifTrue:[icon := lastIcon.]
				ifFalse:[lastIcon delete.].
			].
			icon := icon ifNil:[dragIcon duplicate.].
			event selection userInfo: icon.
			selectedFrame addChild: icon.
			icon translation: selectedPoint + (0@0@0.01);
				 rotationAroundX: 90.
			ext := dist sqrt * 0.3.
			icon extent: ext@ext.
		].

	} otherwise: [			
		lastIcon ifNotNil:[ lastIcon delete.].
	].

! !

!QEditBox methodsFor: 'events' stamp: 'bgf 4/1/2010 12:54'!
pointerUp: event
	| pyr |
	self noGrid.
	((justAClick isNil not) and:[ justAClick ]) ifTrue: [
		selectedFrame ifNil: [ ^ self ].

		"If we click on the edges, then set orientation to nearest 45 degrees. 
		Otherwise, throw up a menu"
		selectedFrame objectName caseOf:{
			[#rotateX:]->[ 
				pyr := self pitchYawRoll. 
				pyr x: (pyr x detentBy: 45.0 atMultiplesOf: 45.0 snap: true).
				self pitchYawRoll: pyr.
				self updateShadow.
			].
			[#rotateY:]->[ 				
				pyr := self pitchYawRoll. 
				pyr y: (pyr y detentBy: 45.0 atMultiplesOf: 45.0 snap: true).
				self pitchYawRoll: pyr.
				self updateShadow.
			].
			[#rotateZ:]->[ 				
				pyr := self pitchYawRoll. 
				pyr z: (pyr z detentBy: 45.0 atMultiplesOf: 45.0 snap: true).
				self pitchYawRoll: pyr.
				self updateShadow.
			].
		} otherwise:[			
			"Remove the undo message pushed by the click that got the menu."
			latestState := nil.
			"DoubleClick means edit-properties."
			event nrClicks = 2 ifTrue: [ contents editProperties: event ].
		].
		self announceForUndo: event.

	] ifFalse: [
		doingInteraction ifTrue: [
			self announceForUndo: event
		].
	].
	self updateBox.	"Refresh the shrink-wrapping in case we've grown back after scaling-down"! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/11/2010 17:06'!
privateNominalScale: newScale

	self extent: extent * (newScale / scaleFrame nominalScale).
	scaleFrame nominalScale: newScale.
! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 2/11/2010 17:16'!
privateScale: scl
	"Unused, as e-box uses matrix scaling."
	nomScale := nomScale * scl.
	self extent: extent*scl.
	scaleFrame nominalScale: scl.! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/1/2010 12:34'!
pushLatestState
	"Push the complete current state.
	 This is in preparation for an action, so we have a record of the 'before' for undo."
	latestState := (self stateVector).! !

!QEditBox methodsFor: 'events' stamp: 'bgf 7/19/2009 13:43'!
removeUser: user selection: selection

	selection userInfo ifNotNil:["remove the frame that is in this slot"
		selection userInfo delete. 
		selection userInfo: nil.
	].
	selection frame ifNotNilDo: [ : sf | sf unhilite ].! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/1/2010 12:28'!
resetRotation

	"Set the rotation to identity.
 	 Check first so we don't push no-ops on the undo state."
	self orientation = Matrix4x4 new ifFalse: [
		self pushLatestState.
		self orientation: Matrix4x4 new. 
		self updateShadow.
	].! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 3/23/2010 11:35'!
ringX: offset icon: lastIcon

	| icon |
	lastIcon ifNotNil:[ 
		lastIcon objectName = #ring ifTrue:[icon := lastIcon.]
			ifFalse:[lastIcon delete.].
		].
	icon := icon ifNil:[ringIcon duplicate.].
	self addChild: icon.
	icon height: border.
	icon radius:((outerExtent y*outerExtent y/4.0)+(outerExtent z*outerExtent z/4.0))sqrt.
	icon localTransform:(Matrix4x4 new rotationAroundZ:-90).
	icon translation:offset@0@0.
	^ icon.! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 3/23/2010 11:56'!
ringY: offset icon: lastIcon

	| icon |
	lastIcon ifNotNil:[ 
		lastIcon objectName = #ring ifTrue:[icon := lastIcon.]
			ifFalse:[lastIcon delete.].
		].
	icon := icon ifNil:[ringIcon duplicate ].
	self addChild: icon.
	icon height: border.
	icon radius:((outerExtent x*outerExtent x/4.0)+(outerExtent z*outerExtent z/4.0))sqrt.
	icon localTransform:(Matrix4x4 new).
	icon translation:0@offset@0.
	^icon.! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 3/23/2010 11:56'!
ringZ: offset icon: lastIcon

	| icon |
	lastIcon ifNotNil:[ 
		lastIcon objectName = #ring ifTrue:[icon := lastIcon.]
			ifFalse:[lastIcon delete.].
		].
	icon := icon ifNil:[ringIcon duplicate ].
	self addChild: icon.
	
	icon height: border.
	icon radius:((outerExtent x*outerExtent x/4.0)+(outerExtent y*outerExtent y/4.0))sqrt.
	icon localTransform:(Matrix4x4 new rotationAroundX:90).
	icon translation:0@0@offset.

	^ icon.! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 3/24/2010 11:32'!
root: rt
	super root: rt.
	rt ifNotNil:[
		self future updateShadow.
		self future saveInitialState.
	].! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 4/10/2010 12:19'!
rotateX: event
	| len delta deltaLoc spto spfrom theta trans pointer radius icon |
	" Rotate around the x-axis "

	pointer := event makePointer.
	icon := currentUser userInfoAt: event userID.
	radius := icon radius.
	len := 100.
	(pointer frame: self 
			pickRightCylinderFrom: ( Vector3 x: len  y: 0 z: 0)
			to:( Vector3 x: len negated y: 0 z: 0)
			radius: radius) 
		ifFalse: [  "not on the cylinder anymore...."
			delta := 2*radius.
			((pointer rayPointer row2 dot: (pointer rayPosition))  > 0) ifFalse: [ delta := delta negated ].
			deltaLoc := pointer rayPointer row2 * delta.
			(pointer frame: self
					pickRightCylinderFrom: (Vector3 x: len  y: deltaLoc y z: deltaLoc z)
					to: (Vector3 x: len negated y: deltaLoc y z: deltaLoc z)
					radius: radius)
				ifFalse:[
					lastSpin := nil. 
					^ false
				].
			spto := pointer selectedPoint copy- deltaLoc.
			spto x: 0  y: spto y negated  z: spto z negated.
		] ifTrue: [
			spto := pointer selectedPoint copy
		].
	event avatar laserShow: (self globalTransform transformPoint: (pointer selectedPoint)).
	rotPoint ifNil:[
		rotPoint := spto. 
		(rotPoint - boxPoint) length >0.01
			ifTrue: [ rotateDirection := -1.0 ]
			ifFalse: [ rotateDirection := 1.0 ].
	].
	spfrom := rotPoint copy.
	spfrom x: 0.0.
	spfrom normalize.
	spto x: 0.0.
	spto normalize.
	theta := ((spfrom y * spto z) - (spto y * spfrom z)).
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.
	theta  := self detenteAngle: theta forEvent: event.

	lastSpin := Matrix4x4 new rotationAroundX: theta.
	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	localTransform =* lastSpin.
	self translation: trans.
	self updateShadow.
	self showGridForAxis: (1@0@0).
	event avatar cornerInfo: '', (localTransform rotation x roundTo: 1), ' Degrees'.
	^ true.
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 4/10/2010 12:18'!
rotateY: event
	| trans spfrom spto deltaLoc theta len delta pointer radius icon |
	" Rotate around the y-axis "

	pointer := event makePointer.
	icon := currentUser userInfoAt: event userID.
	radius := icon radius.
	len := 100.
	(pointer frame: self 
			pickRightCylinderFrom: ( Vector3 x: 0.0 y: len z: 0.0)
			to: (Vector3 x: 0.0 y: len negated z: 0.0) 
			radius: radius)
		ifFalse: [  "not on the cylinder anymore...."
			delta := 2*radius.
			((pointer rayPointer row1 dot: (pointer rayPosition)) > 0)  ifFalse:[ delta := delta negated ].
			deltaLoc := pointer rayPointer row1 * delta.
			(pointer frame: self 
					pickRightCylinderFrom: (Vector3 x: deltaLoc x y: len z: deltaLoc z )
					to: (Vector3 x: deltaLoc x y: len negated z: deltaLoc z) 
					radius: radius)
				ifFalse:[
					lastSpin := nil.
					^ false
				].
			spto := pointer selectedPoint copy - deltaLoc.
			spto x: spto x negated  y: 0  z: spto z negated.
		] ifTrue: [
			spto := pointer selectedPoint copy
		].
	event avatar laserShow: (self globalTransform transformPoint: (pointer selectedPoint)).
	rotPoint ifNil:[
		rotPoint := spto. 
		(rotPoint - boxPoint) length >0.01
			ifTrue: [ rotateDirection := -1.0 ]
			ifFalse: [ rotateDirection := 1.0 ].
	].
	spfrom := rotPoint copy.
	spfrom y: 0.0.
	spfrom normalize.
	spto y: 0.0.
	spto normalize.
	theta := ((spfrom z * spto x) - (spto z * spfrom x)).
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.
	theta  := self detenteAngle: theta forEvent: event.

	lastSpin := Matrix4x4 new rotationAroundY: theta.
	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	localTransform =* lastSpin.
	self translation: trans.
	self updateShadow.
	self showGridForAxis: (0@1@0).
	event avatar cornerInfo: '', (localTransform rotation y roundTo: 1), ' Degrees'.
	^ true.
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 4/10/2010 12:18'!
rotateZ: event
	| delta trans spfrom spto deltaLoc theta len pointer radius icon |
	" Rotate around the z-axis "

	pointer := event makePointer.
	icon := currentUser userInfoAt: event userID.
	radius := icon radius.
	len := 100.
	(pointer frame: self 
			pickRightCylinderFrom: (Vector3 x: 0.0 y: 0.0 z: len)
			to: (Vector3 x: 0.0 y: 0.0 z: len negated) 
			radius: radius)
		ifFalse: [  "not on the cylinder anymore...."
			delta := 2*radius.
			((pointer rayPointer row2 dot: (pointer rayPosition)) > 0) ifFalse:[  delta := delta negated. ].
			deltaLoc := pointer rayPointer row2 * delta.
			(pointer frame: self 
					pickRightCylinderFrom: (Vector3 x: deltaLoc x y: deltaLoc y z: len)
					to: (Vector3 x: deltaLoc x y: deltaLoc y z: len negated) 
					radius: radius)
				ifFalse:[
					lastSpin := nil. 
					^ false
				].
			spto := pointer selectedPoint copy - deltaLoc.
			spto x: spto x negated y: spto y negated   z: 0.
		] ifTrue: [
			spto := pointer selectedPoint copy
		].
	event avatar laserShow: (self globalTransform transformPoint: (pointer selectedPoint)).
	rotPoint ifNil:[
		rotPoint := spto. 
		(rotPoint - boxPoint) length >0.01
			ifTrue: [ rotateDirection := -1.0 ]
			ifFalse: [ rotateDirection := 1.0 ].
	].
	spfrom := rotPoint copy.
	spfrom z: 0.0.
	spfrom normalize.
	spto z: 0.0.
	spto normalize.
	theta := ((spfrom x * spto y) - (spto x * spfrom y)).
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.
	theta  := self detenteAngle: theta forEvent: event.

	lastSpin := Matrix4x4 new rotationAroundZ: theta.
	trans := self translation.
	self translationX: 0.0 y:0.0 z:0.0.
	localTransform =* lastSpin.
	self translation: trans.
	self updateShadow.
	self showGridForAxis: (0@0@1).
	event avatar cornerInfo: '', (localTransform rotation z roundTo: 1), ' Degrees'.

	^ true.
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/24/2010 11:32'!
saveInitialState
	"Finishing construct-from initialization of a saved edit box."

	(initialState isNil and:  [ contents notNil ] and: [ self root notNil ] ) 
	ifTrue: [
		initialState := (self stateVector).
	].! !

!QEditBox methodsFor: 'accessing' stamp: 'das 10/26/2008 16:56'!
scale
	^scaleFrame nominalScale! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 4/12/2010 11:35'!
scaleBox: event
	"Called via perform for pointer-move events on the scale-corners."
	| newScale pointer len1 len2 center notNormal ns |

	pointer := event makePointer.
	(pointer frame: self pickPlane: selectedPoint normal: selectedNormal) ifTrue:[
		event avatar laserShow: (self globalTransform transformPoint: (pointer selectedPoint)).
		center := selectedPoint *selectedNormal.
		len1 := (selectedPoint - center)length.
		len1 = 0 ifTrue:[^ self].
		len2 := (pointer selectedPoint - center)length.
		newScale := len2/len1 +1.0/2.0.
		newScale < 0.1 ifTrue:[newScale := 0.1].
		ns := newScale * scaleFrame nominalScale.

		"Guards against numeric freefall.."
		ns < (baseScale *0.25) ifTrue:[^self].
		ns > (baseScale *4.0) ifTrue:[^self].

		"Detent at unit scale factors (especially useful for 1.0) "
		event shiftPressed ifFalse: [
			(ns - (ns rounded)) abs < 0.0125 ]  ifTrue: [
				| detent |
				detent := (ns rounded) / (scaleFrame nominalScale).
				detent > (100*Float epsilon) ifTrue: [	"Don't even flirt with zero."
					newScale := detent.
				]].

		event avatar 
			cornerInfo: 'Scale: ',  ((newScale * scaleFrame nominalScale) roundTo: 0.001).
		selectedPoint := selectedPoint * newScale.
		notNormal := (selectedPoint - center)sign.
		notNormal := notNormal-(notNormal * selectedNormal abs).
		self translation: self translation + (self orientation transformPoint:(extent-(extent*newScale)*(selectedNormal-notNormal/2.0))).
		self extent: extent * newScale.
		scaleFrame nominalScale: (scaleFrame nominalScale*newScale).
		currentUser ifNotNil:[
			currentUser users valuesDo:[:sel | 
				sel userInfo ifNotNil:[
					sel userInfo objectName = #scale ifTrue:[
						sel userInfo
							translation: sel frame translation.
					].
				].
			].
		].
		self showGridForNormal: (selectedNormal) at: (self translationForContent).
	].! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/7/2010 18:54'!
selectFirstChild: event
	contents childrenForBrowser isEmptyOrNil ifFalse: [
		| c |
		c := contents childrenForBrowser first.
		self edit: c forAvatar: (event avatar).
	].! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/7/2010 18:52'!
selectNextPeer: event
	| p |
	p:=self nextVisiblePeer. 
	p~= self ifTrue:[
		self edit: p forAvatar: (event avatar).
	 ]. ! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/7/2010 18:51'!
selectOutward: event
	| p |
	p := self parent.
	(p isSpace or: [ p isPlacementLocked ]) ifFalse: [
		self edit: p  forAvatar: (event avatar).
	].
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 2/7/2010 18:51'!
selectPreviousPeer: event
	| p |
	p:=self previousVisiblePeer. 
	p~= self ifTrue:[
		self edit: p forAvatar: (event avatar).
	 ]. ! !

!QEditBox methodsFor: 'menu' stamp: 'bgf 10/25/2009 16:14'!
showContextMenu: event
	"The edit box has its own custom context menu,
	 which currently is also available in single-click."
	self showMenu: event! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 11/16/2009 00:15'!
showGridForAxis: aNormal 
	| abs origin |
	"For grids during rotation, we detect the case where we're turning around y,
	 and put the grid on the base plane (shadow bottom)
	 since otherwise it tends to be horizontal near the horizon.
	 For the vertical grids, place them through the object normally."

	self root ifNil: [ ^ nil ].		"Destroy / interaction race guards."
	contents ifNil: [ ^ nil ].
	grid ifNil: [ ^ nil ].

	origin := self translation.

	abs := (self globalTransform orientation transformVector: aNormal) abs.
	abs max caseOf: {
		[abs x] -> [self yzGridAt: origin ].
		[abs y] -> [shadow visible ifTrue: [
						self xzGridAt: (origin - (0@(shadow height*0.98)@0))
					] ifFalse: [
						self xzGridAt: origin.
					]].
		[abs z] -> [self xyGridAt: origin ].
	} otherwise: [
		self noGrid
	].


! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/10/2010 11:34'!
showGridForNormal: aNormal at: origin
	| abs gvector |
	"Show the grid for motion.  The normal is the normal to the plane of motion.
	 We draw the grid in the nearest orthonormal plane, through the center of the object."

	self root ifNil: [ ^ nil ].		"Destroy / interaction race guards."
	contents ifNil: [ ^ nil ].
	grid ifNil: [ ^ nil ].
	gvector := (self orientation transformVector: aNormal).
	abs := gvector abs.
	abs max caseOf: {
		"The x=z checks resolve the ambiguity at the 45-degree points in y."
		[abs x] -> [   
				((abs x ~= abs z) or: [ gvector z ~= gvector x ])
					ifTrue: [ self yzGridAt: origin .  ^ Vector3 x: 1 y: 0 z: 0 ]
					ifFalse: [ self xyGridAt: origin .  ^ Vector3 x: 0 y: 0 z: 1 ] 
				   ].
		[abs y] -> [self xzGridAt: origin .  ^ Vector3 x: 0 y: 1 z: 0 ].
		[abs z] -> [self xyGridAt: origin .  ^ Vector3 x: 0 y: 0 z: 1 ].
	} otherwise: [
		self noGrid.
		^ nil
	].


! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/18/2010 10:18'!
stateVector
	"The complete current pose-edit state.
	 This is a that can be used to restore the edit-box after further adjustemts;
	 see applyStateVector,   and applyEdit:state:. "

	| boxState |

	boxState := { 
	     (scaleFrame nominalScale * contents nominalScale) . 
	     self localTransform copy . 
	     self extent copy 
	  }.

	"Fifth is an array of inner { scale . localTransfrom . translation } 
	 items for each content item inside the box.  We need this because when group-editing,
	 if the edit-box is dismissed, its QTemporaryGroup is lost, 
	 and we want to undo-redo the adjustments on the things inside."

	(contents isKindOf: QTemporaryGroup) ifTrue: [
		^ boxState copyWith: 
			(contents children collect: [ : f | 
				{
					(f nominalScale).
					(f localTransform clone) .		"Relative to contents"
					(self parent relativeTransform: f) translation
				}])
	] ifFalse: [
		"For a singleton, the content position is used directly"
		^ boxState  
			copyWith: {{ 1.0 . (Matrix4x4 identity) .  (self parent relativeTransform: contents) translation }}
	].		
	! !

!QEditBox methodsFor: 'hierarchy' stamp: 'bgf 1/21/2010 16:21'!
structureChanged
	"De nada - don't notify tools about changes within the tools."! !

!QEditBox methodsFor: 'hierarchy' stamp: 'bgf 1/29/2010 15:40'!
transferPose

	"Part of releasing the editor... move the current posture into the child."

	box := contents editorBoundingBox scaled: scaleFrame nominalScale.
	contents nominalScale: contents nominalScale * scaleFrame nominalScale.
	contents localTransform: self localTransform.
	contents translation: contents translation - (contents orientation transformPoint: box center).! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 4/12/2010 11:29'!
translationBoxToContent: trans
	"Edit-box translation to content-translation."

	^ trans - (self orientation transformPoint: box center)! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 4/12/2010 11:30'!
translationContentToBox: trans
	"Content translation to edit-box-translation."

	^ trans +  (self orientation transformPoint: box center)! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 4/12/2010 11:35'!
translationForContent

	"Property sheet support. 
	 The property sheet shows coordinates for the content,
	 which have to be adjusted from our recentered box coordinates."

	^ self translationBoxToContent: (self translation)! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 6/2/2010 13:30'!
translationFromContent: trans
	"Property sheet support. 
	 The property sheet sends us coordinates for the content,
	 which have to be adjusted to our recentered box coordinates."

	self translation: (self translationContentToBox: trans).
	self updateShadow.! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 3/17/2010 11:14'!
undoAllChanges: event
	"Rewind all changes,
	 extract contents back into their context (if any)
	 and destroy the edit box itself."
	event avatar ifNotNil: [
		QActivityCapture 
					action: #UndoAllChanges
					details: self activityDetails
					forSession: (event avatar sessionHex).
	].
	initialState ifNotNil: [ 
		self applyStateVector: initialState.
	].
	self dismiss.! !

!QEditBox methodsFor: 'accessing' stamp: 'bgf 11/1/2009 14:52'!
unhilite
	"undo hilite:..."
	contents ifNotNil: [
		contents unhilite
	].
! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 4/1/2010 12:38'!
updateBox
	"Shrinkwrap my outlines around the outside of the contents at its current size."
	box := contents editorBoundingBox copy.
	contents translation: (box center negated); 
			 pitchYawRoll: (Vector3 zero).
	self extent: (box extent * (scaleFrame nominalScale)).
	self updateShadow.
	scaleCorners do: [ : corner | corner pickable: (contents canScale) ].
! !

!QEditBox methodsFor: 'initialize' stamp: 'bgf 4/1/2010 13:09'!
updateParts
	"Position  the pickable faces, edges, and corners, and the shadow."
	|  trans transN bw faceOffsets cext bext |

	bw := (outerExtent average - extent average)  * 0.25.	

	faceOffsets := (extent + bw) * 0.5.
	(dragFaces at: 1)
		extent: extent x @ extent y;  translationX: 0 y: 0 z: (faceOffsets z) .
	(dragFaces at: 2)
		extent: extent z @ extent y;  translationX: (faceOffsets x) y: 0 z: 0 .
	(dragFaces at: 3)
		extent: extent x @ extent y;  translationX:  0 y: 0 z: (faceOffsets z negated) .
	(dragFaces at: 4)
		extent: extent z @ extent y;  translationX: (faceOffsets x negated) y: 0 z: 0 .
	(dragFaces at: 5)
		extent: extent x @ extent z;  translationX: 0 y: (faceOffsets y negated) z: 0.
	(dragFaces at: 6)
		extent: extent x @ extent z;  translationX:  0 y: (faceOffsets y) z: 0.

	trans := (extent * 0.5) +  (0.67*bw).
	transN := trans negated.
	bext := extent x @ border @ border.
	(rotateEdges at: 1)
		extent: bext;   translationX: 0 y: trans y z: trans z.
	(rotateEdges at: 2)
		extent: bext;   translationX: 0 y: trans y z: transN z.
	(rotateEdges at: 3)
		extent: bext;  translationX: 0 y: transN y z: trans z.
	(rotateEdges at: 4)
		extent: bext;  translationX: 0 y: transN y z: transN z.

	bext := border@extent y@border.
	(rotateEdges at: 5)
		extent: bext;  translationX: trans x y: 0 z: trans z.
	(rotateEdges at: 6)
		extent: bext;   translationX: transN x y: 0 z: trans z.
	(rotateEdges at: 7)
		extent: bext;  translationX: trans x y: 0 z: transN z.
	(rotateEdges at: 8)
		extent: bext; translationX: transN x y: 0  z: transN z.

	bext := border@border@extent z. 
	(rotateEdges at: 9)
		extent: bext;   translationX: trans x y: trans y z:  0.
	(rotateEdges at:10)
		extent: bext;  translationX: transN x y: trans y  z:  0.
	(rotateEdges at:11)
		extent: bext;  translationX: trans x y: transN y  z:  0.
	(rotateEdges at:12)
		extent: bext;  translationX: transN x y: transN y  z:  0.

	trans := (extent * 0.5) +  (1.3*bw).
	transN := trans negated.
	cext := (border@border@border) + (5*bw).	"Make the corners a little easier to pick."

	(scaleCorners at:1)
		extent: cext;  translation:trans.
	(scaleCorners at:2)
		extent: cext;  translationX: trans x y: trans y z: transN z.
	(scaleCorners at:3)
		extent: cext;  translationX: trans x y: transN y z: trans z.
	(scaleCorners at:4)
		extent: cext;  translationX: transN x y: trans y z: trans z.
	(scaleCorners at:5)
		extent: cext;  translationX: trans x y: transN y z: transN z.
	(scaleCorners at:6)
		extent: cext;  translationX: transN x y: transN y z: trans z.
	(scaleCorners at:7)
		extent: cext;  translationX: transN x y: trans y z: transN z.
	(scaleCorners at:8)
		extent: cext;  translation:transN.

	shadow topRadius: ((extent x+ extent y + extent z)-extent max)/(2.0*1.414213562373095).
	self updateShadow.
! !

!QEditBox methodsFor: 'actions' stamp: 'bgf 11/6/2009 20:45'!
updateShadow

	shadow localTransform:(
			self globalTransform orientation inverseByTranspose addRotationAroundX:180 ).
	frameRoot ifNotNil:[
		shadowRay resetTotal.
		frameRoot rayTest: shadowRay ignore: {self}.
		(shadowRay selectedDistance = Float infinity or:[shadowRay selectedDistance = 0])ifTrue:[
			shadow visible: false.
		] ifFalse:[
			shadow height:shadowRay selectedDistance.
			shadow visible: true.
		].
	].! !

!QEditBox methodsFor: 'scale' stamp: 'bgf 2/11/2010 00:10'!
useMatrixScaling 
	^ true! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 2/21/2010 17:06'!
xyGrid
	"Activate the grid (add to root) and align to my containment xz plane."
	grid localTransform: (self parent globalOrientation).
	self parent = self root ifTrue: [
		grid extent: 100@100.
		grid nominalScale: 1.0.
	] ifFalse: [
		| bb chunk |
		chunk := 10.0 * (self parent nominalScale).
		bb := ((self parent boundingBox clone) scale: (1.0 / self parent cumulativeScale)).
		grid extent: ( (((bb extentX @ bb extentY)* 1.1) 
						min: (20@20)*chunk  
						max: (chunk@chunk)) roundTo: chunk );
			nominalScale: (self parent cumulativeScale).
	].
	self root ifNotNil: [
		self root addChild: grid.
	].
	^ grid! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/9/2010 13:15'!
xyGridAt: origin
	"Show the grid in the xy plane, centered on the origin (rounded to space-coordinates.)
	 The origin is in my parent's coordinates (e.g. my translation, meaning through my center.)"
	| trans pt griddedOrigin |

	griddedOrigin :=	Vector3 x: (origin x roundTo: 5)  y: (origin y roundTo: 5) z: (origin z roundTo: 0.02).

	trans := self parent globalTransform transformPoint: griddedOrigin.
	pt := origin - griddedOrigin.

	self xyGrid translation: trans.
	grid crosshairOrigin: pt x @ pt y.
! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 2/21/2010 15:51'!
xzGrid
	"Activate the grid (add to root) and align to my containment xz plane."
	grid localTransform: (self parent globalOrientation *  (Matrix4x4 new rotationAroundX: 90)).
	self parent = self root ifTrue: [
		grid extent: 100@100.
		grid nominalScale: 1.0.
	] ifFalse: [
		| bb chunk |
		chunk := 10.0 * (self parent nominalScale).
		bb := ((self parent boundingBox clone) scale: (1.0 / self parent cumulativeScale)).

		grid extent: ( (((bb extentX @ bb extentZ)* 1.1) 
						min: (20@20)*chunk  
						max: (chunk@chunk)) roundTo: chunk );
			nominalScale: (self parent cumulativeScale).
	].
	self root ifNotNil: [
		self root addChild: grid.
	].
	^ grid! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/9/2010 13:15'!
xzGridAt: origin
	"Show the grid in the xz plane, centered on the origin (rounded to space-coordinates.)
	 The origin is in my parent's coordinates (e.g. my translation, meaning through my center.)"	
	| trans pt griddedOrigin |

	griddedOrigin := Vector3 x: (origin x roundTo: 5) y: (origin y roundTo: 0.02) z: (origin z roundTo: 5).

	trans := self parent globalTransform transformPoint: griddedOrigin.
	pt := origin - griddedOrigin.

	self xzGrid translation: trans.
	grid crosshairOrigin: pt x @ pt z.
	! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 2/21/2010 15:51'!
yzGrid
	"Activate the grid (add to root) and align to my containment xz plane."
	grid localTransform: (self parent globalOrientation *  (Matrix4x4 new rotationAroundY: 90)).
	self parent = self root ifTrue: [
		grid extent: 100@100.
		grid nominalScale: 1.0.
	] ifFalse: [
		| bb chunk |
		chunk := 10.0 * (self parent nominalScale).
		bb := ((self parent boundingBox clone) scale: (1.0 / self parent cumulativeScale)).

		grid extent: ( (((bb extentZ @ bb extentY)* 1.1) 
						min: (20@20)*chunk  
						max: (chunk@chunk)) roundTo: chunk );
			nominalScale: (self parent cumulativeScale).
	].
	self root ifNotNil: [
		self root addChild: grid.
	].
	^ grid
! !

!QEditBox methodsFor: 'grid' stamp: 'bgf 4/9/2010 13:15'!
yzGridAt: origin
	"Make the current grid the yz grid centered at the given point.
	 The origin is in my parent's coordinates (e.g. my translation, meaning through my center.)"
	| trans pt griddedOrigin |

	griddedOrigin := Vector3 x: (origin x roundTo: 0.02)  y: (origin y roundTo: 5) z: (origin z roundTo: 5).

	trans := self parent globalTransform transformPoint: griddedOrigin.
	pt := origin - griddedOrigin.

	self yzGrid translation: trans.
	grid crosshairOrigin: ((pt z negated) @ pt y).
! !

TGroup subclass: #QGate
	instanceVariableNames: 'top left right portal knob height width thickness foot gotoButton glass iris rolodex label isOpen returnFrame selfCard door menuButton buttonHandler mgrip leftFoot rightFoot rgrip style'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Portals'!

!QGate commentStamp: '<historical>' prior: 0!
A basic portal gateway with simple framing, a goto button,
a menu button, and a history rolodex.

A gate is like a window in being a container for a portal,
with a goto button.  It is unlike a window in that it doesn't 
accept other kinds of content.

When closed, a gate has a plain wooden inside its frame.

When open, the content is a QPortal covered with edge-frosted glass
(the glass is a transparent texture that prevents any picking through the gate,
and gives a drop-target surface.)

The content QPortal is given a texture that is either the preview of the destination,
or a swirling clound if no preview is available.   
This preview is visible when the gate is "open" (in the replicated sense)
but when its destination space is not yet logged in for the current harness
(so the destination space cannot be rendered.)   See QPortal
for how this per-harness rendering decision is made.

Clicking the handle on a gate opens or closes the portal.
Clicking on the portal surface is a gesture that is bounced
to the harness, and means 'log in to the far space, if I'm not already synced to it.'

QFileCards of type #postcardLink can be dropped onto the gate
to connect them.      Gates also accept drops that from the UI's Forums...
and History.. tools; these are converted to the right kind of cards
by the drop process managed from those tools.!

!QGate class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QGate class methodsFor: 'inspector' stamp: 'sn 3/8/2013 17:18'!
addPropertyRows: propertyDialog forFrame: gateref

	propertyDialog addRow: (
		(QStringPropertyRow new )
			titleString: 'Label'translateMe;
			fetchBlock: [ :f |  (f get: #label) ifNil: [''] ];
			acceptBlock: [ :v : f | f future label: ((v = '') ifTrue: [nil] ifFalse: [v]) ];
			balloonHelp: 'Label on the gate.'translateMe;
		yourself).
	propertyDialog addRow: (
		(QPointPropertyRow new)
			titleString: 'Extent'translateMe;
			fetchBlock: [ :f | f  get:#extent ];
			acceptBlock: [ : pt :f | f future extent: pt ];
			balloonHelp: 'Dimensions of the portal.'translateMe;
	 	yourself	).
	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Thickness'translateMe;
			fetchBlock: [ : f | f  get:#thickness ];
			acceptBlock: [ :r : f | f future thickness: (r atLeast: 0.05 atMost: 0.3 ) ];
			balloonHelp: 'Thickness of doorframe.'translateMe;
		yourself).

	propertyDialog addRow: (
		(QChoicesPropertyRow new)
			titleString: 'Style'translateMe;
			setChoices: (self styleChoices);
			fetchBlock: [  :f | f get: #style ];
			acceptBlock: [  : s : f| f future style: s ];
			balloonHelp: [ 'Color of the gate'translateMe ];
		yourself).

	super addPropertyRows: propertyDialog forFrame: gateref.! !

!QGate class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QGate class methodsFor: 'defaults' stamp: 'bgf 7/23/2007 10:42'!
defaultExtent

	"Default basic size of a QGate."

	^ 5.3@7.8! !

!QGate class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:12'!
defaultIsFromTemplate
	"Reduces xml volume since these are often premade."
	^true ! !

!QGate class methodsFor: 'defaults' stamp: 'bgf 10/3/2007 16:59'!
defaultStyle

	"Default style for a gate."

	^ #black! !

!QGate class methodsFor: 'defaults' stamp: 'bgf 5/31/2009 13:51'!
defaultThickness
	"Basis for the thickness of all the framework elements;
	 eg. this is used as the radius of the top lozenge."
	^ 0.135! !

!QGate class methodsFor: 'accessing' stamp: 'bgf 11/8/2009 10:23'!
glassHiliteColor
	"Glass tint used for mouse-over - makes the edges glow a bit"
	^ VectorColor r:0.45 g:0.125 b:0 a:0.35! !

!QGate class methodsFor: 'inspector' stamp: 'bgf 5/29/2009 15:24'!
styleChoices
	^ Dictionary newFromPairs: {
			'Black' . #black . 'Wood' . #wood . 'Metal' . #metal
		}! !

!QGate class methodsFor: 'accessing' stamp: 'bgf 7/22/2007 02:15'!
wallStandoff
	"The default standoff from a gate to a wall surface;
	 enough to keep the portal surface in front,
	 by enough to avoid most z-buffer rendering mistakes."

	^ 0.3! !

!QGate methodsFor: 'menus' stamp: 'sn 12/27/2012 23:20'!
addMenuActions: menu forEvent: event
	|block|
	block := event avatar canModifyForum not.
	event avatar canUseInventory ifTrue: [
		menu addItemSpec: {'Copy Address to Inventory' translateMe  . #copyToInventory: } target: selfCard.
		(menu addItemSpec: {'Copy Destination to Inventory' translateMe . #copyDestToInventory: } target: self)
			 disabled: (self postcardLink isNil).
	].

	menu addSeparator. 
	(menu addItemSpec: {'Disconnect Gate' translateMe . #confirmClear: } target: self)
		disabled:  (block or: [ self postcardLink isNil] ).

	(event avatar isEditingTemplate or: [ self fromTemplate not  ]) ifTrue: [
		menu addSeparator.
		(menu addItemSpec: {'Delete Gate' translateMe  . #confirmDelete:} target: self)
			disabled: (block or: [ self isPlacementLocked ] ).
	].
! !

!QGate methodsFor: 'private' stamp: 'bgf 3/14/2010 17:50'!
applyStateVector: stateVec
	"See #stateVector - this reapplies a captured state."	self postcardLink: (stateVec first).
	self label: (stateVec second).
	self isOpen: (stateVec third).! !

!QGate methodsFor: 'private' stamp: 'bgf 7/28/2010 13:26'!
applyStyle

	| frameStuff doorStuff doorUV |

	doorUV := 1@1.
	style caseOf: {
		[#metal] -> [ frameStuff := QMaterials aluminumMaterial. 
					  doorStuff := QMaterials aluminumMaterial.
					  doorUV := self extent * 2. ].
		[#black] -> [ frameStuff := QMaterials darkGreyMaterial. 
					  doorStuff := QMaterials mediumDarkGreyMaterial].
	} otherwise: [
		frameStuff := QMaterials mediumDarkWoodMaterial. 
		doorStuff := QMaterials mediumWoodMaterial
	].

	{top. left. right} do: [:part|
		part material: frameStuff copy.
	].
	door material: doorStuff copy.
	door material cullFace: false.
	door uvScale: doorUV.

	rolodex borderMaterial: frameStuff copy.
	rolodex baseMaterial: doorStuff copy.

	label ifNotNil: [
		label textColor: self labelColor
	].
! !

!QGate methodsFor: 'menus' stamp: 'bgf 3/15/2008 00:32'!
asPlace
	"The QPlace descriptor for this destination."
	^ gotoButton asPlace! !

!QGate methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 13:02'!
canOpenCard: aFileCard

	"Is the card allowable as an argument to 'openCard:fromUser:"
	^ aFileCard kind = #postcardLink! !

!QGate methodsFor: 'accessing' stamp: 'bgf 9/4/2007 12:13'!
childrenForBrowser
	^ rolodex childrenForBrowser! !

!QGate methodsFor: 'recent-documents' stamp: 'bgf 3/12/2010 18:07'!
closeCard: aFileCard forEvent: anEvent
	"The card is being deleted from my recent-rolodex."
	(self hasCardOpen: aFileCard) ifTrue: [
		self disconnect
	].
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 7/12/2008 19:13'!
closeContents

	"Close the portal.  The underlying link remains in place,
	 but the portal is removed from the visible space.
	This prevents outbound travel, but not inbound travel."

	isOpen ifTrue:[
		self removeChild: portal.
		isOpen := false.
		self signal: #closed.
		self updateKnob.
	].
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/31/2006 11:52'!
closeContentsVisibly

	"Close the portal, animating its disappearance."

	|  |

	iris ~= 1 ifTrue: [
		^ self  "no competing transitions!!"
	].
	iris := 2.

	QSoundEffects play: #iris for: self.
	self stepClosed.	
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/16/2006 15:49'!
closeInboundLink

	"Suspend passage into this portal; 
	other portals connected to me will not be traversible while I am closed to inbound linking."

	self removeChild: returnFrame.! !

!QGate methodsFor: 'menus' stamp: 'sn 12/27/2012 22:45'!
confirmClear: event
	"Ask the user to confirm clearing the portal link. "
	event avatar ifNil: [ ^ self ].
	self root ifNil: [ ^ self ].
	event avatar confirm: 'Disconnect Portal Gate?' translateMe   
				selector:  #reallyDisconnect:
				target: self.
! !

!QGate methodsFor: 'persist' stamp: 'bgf 6/1/2009 22:29'!
constructFrom: pStream
	| cards defaultLinkName rval oldCanMove |
	super constructFrom: pStream.
	self extent: (pStream readPoint:'extent' default: self extent).
	self isOpen: (pStream readBool: 'isOpen' default: true).
	
	(pStream readBool: 'isInboundLinkOpen' default: true) ifFalse: [
		self closeInboundLink.
	].

	"The following information is for establishing the linkage between portals;
	 if a return link name was persisted, make it the only global name of the return frame."
	defaultLinkName := returnFrame globalName.  "Registered implicitly at construction."
	(pStream readString: 'linkTargetName' default: nil) 
		ifNotNilDo:[ : linkName | 
			returnFrame globalName: linkName.
			Processor activeIsland removeGlobal: defaultLinkName.
			defaultLinkName := nil.
	].

	"Compatibility:
	Old persistence snapshots may link by objectID, with no global name.
	If we encounter one of them,
	use the objectID as the global name for the target frame 
      (link ID and globalName  were mutually exclusive in the past.)"
	defaultLinkName ifNotNil: [
		(pStream readString: 'linkTargetID' default: nil) 
			ifNotNilDo:[:linkID| 
				returnFrame globalName: linkID.
				Processor activeIsland removeGlobal: defaultLinkName].
	].

	"Initialize filled the self card using the now-forgotten default id of my return frame; update it."
	selfCard kind:#postcardLink content: (self postcard).

	self postcardLink: (pStream readObject: 'postcardLink' default: nil).
	self label: (pStream readString: 'label' default: nil).
	rval := pStream readNumber: 'thickness' default: nil.
	rval ifNil: [
		"Old gates used a default radius of 0.175 but didn't properly omit the default from xml"
		rval := pStream readNumber: 'radius' default: (self thickness).
		rval = 0.175 ifTrue: [ rval := self class defaultThickness ].
	].
	self thickness: rval.

	"Filtering duplicates here avoids duplicating the card for the current link (set above)
	 and cleans up older rolodexes on reload."
	cards := pStream readSequence: 'cards' default: #().
	cards do: [ :card|
		(self rolodex hasContent: card content) ifFalse: [
			card content ifNotNil: [
				self rolodex insertCard:card.
			].
		].
	].
	"Is Placement Locked used to be 'canMove' until 2.0.18"
	oldCanMove := (pStream readBool: 'canMove' default: (self isPlacementLocked not)).
	oldCanMove ~= self isPlacementLocked not ifTrue: [
		self isPlacementLocked: oldCanMove not.
	].
	self style: (pStream readString: 'style' default: style) asSymbol.
! !

!QGate methodsFor: 'events' stamp: 'bgf 3/13/2010 08:52'!
consumeDrop: delivery avatar: avatar selection: aTXSel

	| old new |
	(delivery respondsTo: #kind) ifFalse: [ ^ false ].
	(delivery kind = #postcardLink) ifFalse: [ ^ false ].

	old := self stateVector.	

	self	rolodex consumeDrop: delivery avatar: avatar selection: aTXSel.
	self establishLink: delivery contentLink forUser: avatar.
	
	new := self stateVector.

	(old hasEqualElements: new) ifFalse: [	
		avatar recordAction: #frame:oldState:newState:description:
			class: QStateVectorAction
			arguments: {  self . old. new .  'Drop Link' }.
	].
	^ false  	"Card is not consumed, even if drop is accepted"! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/16/2006 15:44'!
contents
	"Convenience accessor for the return-linking frame, 
	 so that a gate can stand in code as a replacement for TWindow."
	^ returnFrame! !

!QGate methodsFor: 'menus' stamp: 'bgf 1/2/2008 23:51'!
copyDestToInventory: event

	"Copy the destination URL this gate opens towards, to the user's inventory."
	(QFileCard forServiceURL: (self postcardLink urlString)) copyToInventory: event.! !

!QGate methodsFor: 'initialize' stamp: 'bgf 4/17/2009 11:52'!
defaultIsAtomic
	"So the link history cards show in the inspector."
	^ false! !

!QGate methodsFor: 'accessing' stamp: 'bgf 4/25/2009 12:40'!
destinationStanceFor: anAvatar

	| stance |

	stance := QGotoButton stanceFor: anAvatar toView: glass inDestination: self.

	"Gates are for walking through, so align the avatar bottom with the gate bottom
	 rather than centering on the gate... this works best for gates on level ground."

	stance yHere: (self globalPositionReal y) + (anAvatar footToEyes). 
	stance yThere: (stance yHere).

	^ stance! !

!QGate methodsFor: 'menus' stamp: 'bgf 3/12/2010 18:07'!
disconnect

	self postcardLink: nil.
	self label: nil.
	! !

!QGate methodsFor: 'private' stamp: 'bgf 3/12/2010 17:27'!
displayCard: fileCard fromUser: anAvatar
	"Workalike with display classes, for consistent undo handling of filecard drops."
	self establishLink: fileCard contentLink forUser: anAvatar.
	! !

!QGate methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:47'!
dragAdapter
	^ mgrip! !

!QGate methodsFor: 'events' stamp: 'bgf 11/23/2007 15:49'!
establishLink: contentLink forUser: avatar

	"The given user has supplied the given content link for opening...
	 link to it and request resolution of the linked space (via the avatar and its harness,
	 so the effect is only on the requesting avatar's session."

	self postcardLink: contentLink.
	self openContentsVisibly.
	avatar signal: #activateLinkForFrameAndGo 
			with: contentLink 
			with: self 
			with: false.	"Let the user see the result in-place, rather than jumping-through."
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/20/2006 00:33'!
extent
	"The x,y rectangular extent of the gateway.
	 The frame pieces are arranged around the outside of this rectangle."
	^ width@height
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 5/29/2009 15:07'!
extent: ext

	"Set the extent (width and height) of the gate."

	(ext x = width & (ext y = height)) ifTrue: [ ^ self ].	"Cut circular recursions with my grip."

	width := ext x.
	height := ext y.
	self shapeChanged.
! !

!QGate methodsFor: 'private' stamp: 'sn 3/4/2013 18:26'!
floatingMessage

	"The floating-message about me, to use for routine mousevers."

	self isOpen ifTrue: [
		"Note that this message goes only to the local user, 
		  so we can use 'resolved' to distinguish the strings."
		portal resolved ifFalse: [
			^ 'Click to login to:\'translateMe, self postcardLink shortDisplayString
		] ifTrue: [
			 ^ 'Click to jump to:\'translateMe, self postcardLink shortDisplayString.
		].
	] ifFalse: [
		^'Gate to:\'translateMe, self postcardLink shortDisplayString, ' (Closed)'.
	].! !

!QGate methodsFor: 'accessing' stamp: 'bgf 7/17/2007 21:39'!
gotoButton 
	"The goto arrow button positioned centered above the gate"
	^ gotoButton! !

!QGate methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 13:05'!
hasCardOpen: aFileCard
	"Is this filecard my current open content?"
	^ aFileCard contentLink = self postcardLink! !

!QGate methodsFor: 'initialize' stamp: 'sn 12/27/2012 22:57'!
initialize
	| sideFeet |
	super initialize.

	iris := 1.
	thickness := self class defaultThickness.
	top := QLozenge new.
	left := TCube new.
	right := TCube new.
	foot := TCube new.
	leftFoot := TCube new.
	rightFoot := TCube new.
	glass := TSimpleRectangle new.

	self addChild: top.
	self addChild: left.
	self addChild: right.
	self addChild: foot.
	self addChild: glass.
	sideFeet := TGroup new.
	sideFeet addChild: leftFoot.
	sideFeet addChild: rightFoot.
	self addChild: sideFeet.

	{foot . leftFoot . rightFoot } do: [:part|
		part material: QMaterials gateFootMaterial.
		part solid: false.		"No twitchy jumps when approaching."
	].
	top uvAngle: 90.0.
	glass material: QMaterials portalGlassMaterial;
		   objectName: #glass;
		   solid: false.

	(mgrip := QMotionAdapter forLuggage: self grip: foot)
		isGroundBound: true.
	rgrip := QPlanePivot forLuggage: self grip: sideFeet.

	knob := QLozenge new.
	knob material: QMaterials controlOnMaterial;
		  objectName: #knob.
	self addChild: knob.

	portal := QPortal new.
	self addChild: portal.
	portal objectName: #gatePortal.
	isOpen := true.

	returnFrame := TGroup new.
	self addChild: returnFrame.
	returnFrame rotationAroundY:180.
	returnFrame visible: false.

	gotoButton :=  QGotoButton new.
	gotoButton destination: self.
	gotoButton 
		category:  'Portal Gates' ;
	 	label: (self label ifNil: ['Portal Gate']);
	 	suggestedPOV: nil.	"Not sticky - interferes with tabbing."

	self addChild: gotoButton.
	gotoButton disable.

	rolodex := QDocumentShelf new.
	rolodex basicExtent: 0.6@0.4;
			direction: #vertical;
			infoMessage: 'Recent Destination Links' translateMe .
	self addChild: rolodex.

	door := (TRectangle new)
				objectName: #door;
				divisions: 4;
				cullFace:false;
				normal: nil;		"So it is an obstacle from the rear also"
				yourself.
	self addChild: door.

	"The file-card representing a link to this gate (used for copy-to-inventory)"
	selfCard := QFileCard forPostcard: (self postcard).
	selfCard isTransient: false.
	selfCard dragAdapter isDeletable: false.	

	menuButton := (TRectangle new)
			material: (QMaterials menuButtonRectangleMaterial);
			objectName: #mbutton;
			yourself.
	self addChild: menuButton.
	buttonHandler := QButtonHandler
		body: menuButton
		target: self
		action: #showMenu:.
	buttonHandler label: 'Gate Menu.' translateMe .

	self style: self class defaultStyle.			
	self extent: (self class defaultExtent).
	self postcardLink: nil.

	self startScript: #titleClick: when: { top . #pointerDown }.
	self startScript: #onRolodexChanged when: {rolodex . #contentChanged }.	
	self onRolodexChanged.! !

!QGate methodsFor: 'events' stamp: 'bgf 2/20/2007 19:23'!
isDropTarget: payload

	^(payload respondsTo:#kind) and: [payload kind = #postcardLink]! !

!QGate methodsFor: 'accessing' stamp: 'bgf 1/28/2007 18:46'!
isGroundBound
	^ true! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/16/2006 15:51'!
isInboundLinkOpen

	"When true, I can receive inbound traffic from other portals."
	^ self hasChild:returnFrame! !

!QGate methodsFor: 'testing' stamp: 'bgf 3/15/2008 17:58'!
isListedLandmark
	"Should I appear in a places list?"
	^ self postcardLink notNil! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/20/2006 00:16'!
isOpen
	^ isOpen! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/20/2006 00:15'!
isOpen: bFlag
	bFlag ifTrue: [
		self openContents
	] ifFalse: [
		self closeContents
	].! !

!QGate methodsFor: 'accessing' stamp: 'bgf 6/1/2009 12:43'!
isPlacementLocked: bLocked
	"We present as one boolean property 
	 the control over whether the gates resize menu option, 
	 and the gates drag-by-foot behaviors are avaliable."
	super isPlacementLocked: bLocked.
	mgrip isEnabled: bLocked not.
	rgrip isEnabled: bLocked not.
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/20/2006 00:20'!
label
	"The string shown in the sign over the gate, if any."

	label ifNotNil: [
		^ label contents asString
	].
	^ nil! !

!QGate methodsFor: 'accessing' stamp: 'bgf 3/15/2008 00:30'!
label: aString
	"Add or update the label marquee floating above the portal."

	label ifNotNil: [
		self removeChild: label.
	].

	(aString notNil and: [ aString size > 0 ]) ifTrue: [
		label := (QLabel new) 
			material: (QMaterials flatWhiteMaterial);
			contents: aString;
			styleName: #SignStyle;
			textColor: self labelColor;
			backgroundColor: Color transparent;
			checkAlpha;
			yourself.
		self addChild: label.
		self startScript: #titleClick: when: {label . #pointerDown}.
		self updateAccessories.
	] ifFalse: [
		label := nil
	].

	gotoButton label: (aString ifNil: [ 'Portal Gate' ]).! !

!QGate methodsFor: 'private' stamp: 'bgf 6/2/2009 18:10'!
labelColor
	"Color for the top-bar label text... depends on the style in use."
	^ style caseOf: {
		[ #metal ] -> [ Color r: 0.1 g: 0.05 b: 0.03 ].
		[ #black ] -> [ Color gray:0.9 ].
	} otherwise: [
		Color paleBuff darker
	].
	! !

!QGate methodsFor: 'accessing' stamp: 'bgf 9/4/2007 12:49'!
labelForBrowser
	self postcardLink ifNotNil: [
		^ 'Portal Gate - ', (self postcardLink shortDisplayString)
	] ifNil: [
		^ 'Portal Gate (QGate)'
	].! !

!QGate methodsFor: 'accessing' stamp: 'bgf 2/12/2010 14:19'!
localCenter
	"Cheat - substitute the plain avatar height above ground;
	 this is a workaround so that Forums gates leave the avatar on the ground."
	^ self translation + (0@(5*(self scaleMemo))@0)! !

!QGate methodsFor: 'menus' stamp: 'bgf 9/14/2007 21:26'!
menuStyle 
	^ QStyles menuStyle! !

!QGate methodsFor: 'menus' stamp: 'bgf 9/14/2007 19:59'!
menuTitle
	^ 'Portal Gate'! !

!QGate methodsFor: 'initialize' stamp: 'bgf 3/5/2010 18:28'!
objectID: oid
	super objectID: oid.
	gotoButton ifNotNil: [ gotoButton destination: self ].! !

!QGate methodsFor: 'events' stamp: 'bgf 1/19/2008 14:12'!
onRolodexChanged

	"Only show the recent-links when there are links other than the installed one;
	 keeps gates looking simple for default use."

	self rolodex visible:(
		self rolodex cards anySatisfy: [:cd | cd contentLink ~= self postcardLink] ).! !

!QGate methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 14:10'!
openCard: aFileCard fromUser: anAvatar

	"Generic protocol for managing a deck of recent content."
	^self establishLink: aFileCard contentLink forUser: anAvatar! !

!QGate methodsFor: 'accessing' stamp: 'bgf 7/12/2008 19:14'!
openContents

	"Open the portal."
	isOpen ifFalse:[
		self addChild: portal.
		isOpen := true.
		self signal: #opened.
		self updateKnob.
	].
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 9/3/2007 05:28'!
openContentsVisibly

	"Display the portal, animating its appearance."

	|  |

	iris ~= 1 ifTrue: [
		^ self  "no competing transitions!!"
	].
	iris := 32.

	isOpen ifFalse:[
		self addChild: portal.
		isOpen := true.
		self signal: #opened.
		self updateKnob.
	].
	QSoundEffects play: #iris for: self.
	self stepOpen.


! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/16/2006 15:50'!
openInboundLink

	"Resume passage into this portal - portals linked to my returnFrame
	 can be traversed when my inbound link is open.  This is the default state."

	self addChild: returnFrame.! !

!QGate methodsFor: 'persist' stamp: 'bgf 6/1/2009 12:37'!
persistOn: pStream
	super persistOn: pStream.
	pStream storePoint: 'extent' value: self extent default: self class defaultExtent.  
	pStream storeBool: 'isOpen' value: self isOpen default: true.
	pStream storeBool: 'isInboundLinkOpen' value: self isInboundLinkOpen default: true.

	"The following information is for establishing the linkage between portals"
	pStream storeString: 'linkTargetName' value: returnFrame globalName default: nil.

	pStream storeObject: 'postcardLink' value: self postcardLink default: nil.
	pStream storeString: 'label' value: self label default: nil.
	"Thickness was radius prior to 2.0.18 ... for two-way load, duplicate it."
	pStream storeNumber: 'thickness' value: self thickness default: (self class defaultThickness).	
	pStream storeNumber: 'radius' value: self thickness default: (self class defaultThickness).
	pStream storeSequence: 'cards' value: (self rolodex cards) default: #().

	"For pre-2.0.18 back-forth compatibility."
	self isPlacementLocked ifTrue: [
		pStream storeBool: 'canMove' value: false.
		pStream storeBool: 'canRotate' value: false.
	].

	pStream storeString: 'style' value: style default: self class defaultStyle.! !

!QGate methodsFor: 'menus' stamp: 'bgf 8/1/2009 17:52'!
photographerClass
	^ QDirectPhotoJob! !

!QGate methodsFor: 'events' stamp: 'bgf 5/22/2007 01:13'!
pointerDown: event
	"Prep for pointerUp:, which does the work."

	event avatar ifNil: [ ^ self ].	"Just in case we're logging in or out."

	"If the portal is open and unresolved, this makes it brighten a little
	 in response to the click.  Otherwise, no effect."
	portal hilite.
	! !

!QGate methodsFor: 'events' stamp: 'bgf 2/8/2008 03:53'!
pointerEnter: event

	self pointerOver: event.
! !

!QGate methodsFor: 'events' stamp: 'bgf 11/23/2007 14:18'!
pointerLeave: event

	gotoButton disable.
	knob unhilite.
	glass unhilite.! !

!QGate methodsFor: 'events' stamp: 'ar 4/16/2011 17:12'!
pointerOver: event

	event selection frame ifNil: [ ^ self ].
	event avatar ifNil: [ ^ self].

	(self rolodex isOrContains: (event selection frame)) ifTrue: [
		glass unhilite.
		^ self
	].
	(gotoButton isOrContains: (event selection frame)) ifTrue: [
		glass unhilite.
		^ self 
	].
	portal postcardLink ifNil: [
		glass unhilite.
		^ event avatar 
			signal: #statusMessage with: 'To connect the gate, drag a Forum link onto it.'  
	].

	(event selection frame = knob) ifTrue: [
		glass unhilite.
		^ self pointerOverKnob: event.
	].

	event selection frame = glass ifTrue: [
		glass hilite: self class glassHiliteColor.
	].

	event avatar doWarpMoves ifTrue: [
		self floatingMessage ifNotNilDo: [ :msg |
			event avatar showFloatMessage: msg forEvent: event.
		]].! !

!QGate methodsFor: 'private' stamp: 'bgf 11/13/2008 12:48'!
pointerOverKnob: event
	self postcardLink ifNotNil: [
		knob hilite.
		self isOpen ifTrue: [
			event avatar showFloatMessage: 'Click knob to close gate' forEvent: event.
		] ifFalse: [
			event avatar showFloatMessage:  'Click knob to open gate' forEvent: event.
		].
	].
! !

!QGate methodsFor: 'events' stamp: 'bgf 11/28/2009 09:21'!
pointerUp: event
	"Toggle visibility and (if opening) request resolution of the underlying portal link"

	| steps |
	event avatar ifNil: [ ^ self ].	"Just in case we're logging in or out."

	(portal future:300) unhilite.	"See pointer-down"

	(event frame = knob) ifTrue: [		"Knob toggles portal iris."
		self postcardLink ifNotNil: [
			^ self toggleVisiblyFor: event avatar.
		]
	].

	(self root miramarUI) ifTrue:[ ^self ].

	"Clicking on an open, unresolved portal triggers login;
	 if already resolved, the action is to teleport through."

	event avatar doWarpMoves ifTrue: [
		(event frame = glass and: [self isOpen & (self postcardLink notNil)]) ifTrue:[

			event avatar clearFloatMessageForEvent: event.
			steps := (gotoButton bringToDestination: event).
			event avatar clearFloatMessageForEvent: event.
			self future signalActivation: event avatar after: (steps*40).   
					"About right for 25fps; in slower environments, jump a bit early"
		]].

	"Clicking on a closed portal toggles it open."
	(event frame = glass and: [self isOpen not & (self postcardLink notNil)]) ifTrue: [
		^ self toggleVisiblyFor: event avatar.
	].! !

!QGate methodsFor: 'accessing' stamp: 'bgf 3/25/2006 08:35'!
portal

	^ portal! !

!QGate methodsFor: 'accessing' stamp: 'bgf 1/8/2010 14:19'!
portraitTransform
	"My front-on transform for photos taken with the QOnsitePhotoJob,
	 which does not take the frame out to an empty studio space."
	| camera ext |
	camera := (self globalTransform clone).
	ext := (self destinationExtent).
	camera translation: (
		camera translation
		+ (camera orientation transformVector: (0@ (self extent y * 0.525) @ (ext maxDimension * 1.1)))).

	^ camera! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/16/2006 16:18'!
postcard
	"the postcard to use when linking to this gateway."
	^ returnFrame postcard! !

!QGate methodsFor: 'accessing' stamp: 'bgf 2/10/2008 11:30'!
postcardLink
	"The link to which I am connected."
	^ portal postcardLink! !

!QGate methodsFor: 'accessing' stamp: 'bgf 2/10/2008 11:30'!
postcardLink: postcard

	"Connect me to the given link."
	postcard ifNotNil: [
		rolodex firstCard:(
			rolodex addContent: postcard kind: #postcardLink ).
	].
	portal postcardLink: postcard.
	postcard ifNotNil: [
		self label: postcard shortDisplayString
	] ifNil: [
		self closeContents
	].
	self updateKnob.
	self onRolodexChanged.! !

!QGate methodsFor: 'accessing' stamp: 'bgf 3/28/2007 23:15'!
postcardUrl

	"The URL String of my current postcard link, which may be nil"
	^ portal postcardLink ifNotNilDo: [ :pc| pc urlString ].! !

!QGate methodsFor: 'accessing' stamp: 'bgf 2/11/2010 17:19'!
privateScale: aFactor

	width := width * aFactor.
	height := height * aFactor.
	thickness := thickness * aFactor.
	super privateScale: aFactor.

! !

!QGate methodsFor: 'menus' stamp: 'bgf 3/13/2010 08:52'!
reallyDisconnect: event

	| old new |
	old := self stateVector.	
	self disconnect.
	new := self stateVector.

	event avatar ifNotNil: [ : avatar |
		(old hasEqualElements: new) ifFalse: [	
			event avatar 
				recordAction: #frame:oldState:newState:description:
				class: QStateVectorAction
				arguments: {  self . old. new .  'Disconnect' }.
		]].
! !

!QGate methodsFor: 'persist' stamp: 'bgf 10/4/2007 03:17'!
registerGlobal: aSymbol
	"Register the portal for postcard linking...
	 we intercept this and delegate to the return frame.
	 This must be done ONLY ONCE, and ONLY during world-building!! "

	"Link consistency on save/restore requires a single global name."
	returnFrame globalName ifNotNilDo: [ :gn | 
		Processor activeIsland removeGlobal: gn
	].
	returnFrame registerGlobal:aSymbol.

	"The default self-link is stale."
	selfCard kind: #postcardLink content: (returnFrame postcard).! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/16/2006 15:47'!
returnFrame
	"This is the TFrame that should be set up as the global postcard link point
	 for portal connections -to- this portal.   Also available  as #contents"
	^ returnFrame! !

!QGate methodsFor: 'accessing' stamp: 'bgf 7/17/2007 21:40'!
rolodex
	"The history rolodex."
	^ rolodex! !

!QGate methodsFor: 'menus' stamp: 'bgf 10/25/2009 16:13'!
showContextMenu: event
	(event frame = menuButton) ifTrue: [ 
		"Make right-clicking the menu button itself do the same as left-clcking the MB;
	  	 users can right-click the panel body or title for the 'pure' context menu."
		self showMenu: event
	] ifFalse: [
		super showContextMenu: event
	].! !

!QGate methodsFor: 'private' stamp: 'bgf 3/26/2008 22:00'!
signalActivation: avatar after: msecs
	"Slightly-delayed indication to the harness that I've been clicked;
 	 the delay allows the approach-goto to run before we leap.
	 Note that we cannot time this differently for resolved versus unresolved,
	 as that would be a identical-replication violation."

	(avatar future: msecs)
			signal: #activateLinkForFrameAndGo 
			with: self postcardLink
			with: self 
			with: true.	"Step through."! !

!QGate methodsFor: 'private' stamp: 'bgf 3/12/2010 07:14'!
stateVector
	"Capture the state of link-setup, for undo-redo of link drops."
	^ { self postcardLink .
	     self label .
	     self isOpen }! !

!QGate methodsFor: 'private' stamp: 'bgf 7/12/2008 19:14'!
stepClosed

	"Make the portal smaller in steps until it disappears,
	 then remove it from view.
	 Worker routine for closeContentsVisibly"

	iris = 1 ifTrue: [
		^ self.	"Try to catch competing loops."
	].

	iris < 32 ifTrue: [
		portal extent: (width/iris)@(height/iris).
		iris := iris*1.5.
		(self future:100) stepClosed.
	] ifFalse: [
		self closeContents.
		portal extent: width@height.
		iris := 1.
	].
! !

!QGate methodsFor: 'private' stamp: 'bgf 7/12/2008 19:14'!
stepOpen

	"Make the portal larger in steps until it fills the gate.
	 Worker routine for openContentsVisibly."

	iris > 1 ifTrue: [
		portal extent: (width/iris)@(height/iris).
		iris := iris/1.5.
		(self future:100) stepOpen.
	] ifFalse: [
		portal extent: width@height.
		iris := 1.
	].
! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/3/2007 16:40'!
style
	"A named style - #wood, #metal, #black."
	^ style! !

!QGate methodsFor: 'accessing' stamp: 'bgf 10/3/2007 18:46'!
style: aSymbol

	"Choose the visual style for the gate framing.  
      Currently just a coloring choice: #wood, #black, #metal." 

	style = aSymbol ifFalse: [
		style := aSymbol.
		self applyStyle.
	].! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/27/2007 19:17'!
teleportTransform

	"Return the newTransform to assign an avatar after teleporting through the gate."

	| front |

	front := self globalTransform copy.

	"A bit in front, and set to the default avatar height, and facing away.
	  Falling will repair the height after we enter for different avatar heights;
	  we don't use the avatar in this protocol since at the time it's called there isn't one
  	  in the space with me."
	front translation: front translation  + (front transformVector: (0 @ 5 @ 5)).
	front addRotationAroundY: 180.
	^ front! !

!QGate methodsFor: 'accessing' stamp: 'bgf 5/31/2009 13:55'!
thickness
	"The basis for the thickness of the gate framing pieces."
	^ thickness! !

!QGate methodsFor: 'accessing' stamp: 'bgf 5/31/2009 13:55'!
thickness: r

	"The dimension for the thickness of posts and crown bars;
	 specifically, it is the lozenge radius of the crown bar
 	 (and the basis for layout of the rest of the frame)"

	(thickness ~= r) ifTrue: [
		thickness := r.
		knob radius: r/2.0.
		self shapeChanged.
	].! !

!QGate methodsFor: 'events' stamp: 'bgf 1/5/2010 14:48'!
titleClick: event
	"A click on my title (or background) - bring the avatar hither"
	event avatar ifNotNilDo: [ : av |	
		av doWarpMoves ifTrue: [
			gotoButton bringToDestination: event
		]]! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/23/2007 15:51'!
toggleVisiblyFor: avatar

	"open or close the gate, with slow visible feedback"

	isOpen ifTrue: [ 
		self closeContentsVisibly
	] ifFalse: [ 
		avatar signal: #activateLinkForFrameAndGo 
					with: self postcardLink
					with: self 
					with: false.	"Certainly don't jump just because we a closed gate."
		self openContentsVisibly
	].! !

!QGate methodsFor: 'initialize' stamp: 'bgf 7/14/2009 20:45'!
updateAccessories

	"updateShape worker: update the buttons and accessories."

	| upperY |
	upperY := top translation y.

	"Goto button centered above."
	gotoButton translation: 0@(height+(thickness*8)+(gotoButton extent y/2))@0.

	"recent-links deck to left of left jamb."
	self rolodex translation: ((width + rolodex basicExtent x + (1.5*thickness)) * -0.5)
							@ (height/2)
							@ (1.2*thickness).

	"Knob in front of right jamb."
	knob translation: (width/2)@(height/2)@(0.75*thickness);
		  extent: (thickness @ (1.5*thickness));			
		  radius: (thickness/2.0).

	"Menu button at upper right"
	menuButton 
		extent: (((thickness*3.4) min: (width*0.16))@(thickness*1.25))+ (thickness *0.55);
		translation: ((width-(4*thickness))/-2)@upperY@(thickness*1.02).
			

	"label on top bar between menu and tearaway."
	label ifNotNil: [
		label extent: (width * 0.72)@(thickness * 2.9);
			  translationX: (thickness*1.5) y: upperY+(thickness/12.0) z: (1.1*thickness).
	].
! !

!QGate methodsFor: 'initialize' stamp: 'bgf 5/31/2009 14:18'!
updateFrame
	"Update shape worker - fix the basic size of the parts."

	top translation: 0@(height+(thickness*1.5))@0;
		 extent: (width+(thickness))@(2*thickness); 
		 radius: (thickness).

	left translation: (width/2)@(height/2)@0;
		extent: (thickness*0.95)@height@(thickness*1.6).

	right translation: (width/-2)@(height/2)@0;
		 extent: (left extent).

	foot extent: (width-(6*thickness))@(thickness/2.2)@(thickness*3);
		 translation: 0@(thickness/3.5)@0.

	leftFoot extent: (4*thickness)@(foot extent y)@(foot extent z);
			translation: (width-(2*thickness))/-2.0@(foot translation y)@0.
	rightFoot extent: (leftFoot extent);
			translation: (width-(2*thickness))/2.0@(foot translation y)@0.

	portal extent: width@height;
		translation: 0@height/2@(thickness * 0.35).	"Behind glass (glass intercepts clicks)."

	"The door is solid so that it can be an obstacle to walking through closed or unresolved portals,
	 and it is 2-sided so it is an obstacle for walking through freestanding portals from behind.  
	 But... it has to be far enough behind the TPortal so that 
	 people teleporting INTO me are not obstructed.
	 Teleport-arrivals arrive a tiny bit 'behind' the TPortal they link into
	 as part of the staging that gives visual continuity to the transition."
	door translation: 0@(height/2)@(thickness*-0.72);
		 extent: width@height.

	returnFrame translation: (0 @ (height/2) @ (thickness*3 max: 0.501)). 
			"Out ahead in Z, so none of the frame is visible
	 		 from the far side, and so teleports don't get trapped behind the return frame."

	"The glass doesn't quite meet the threshold.
	 We leave a tiny gap.  This prevents the glass from intercepting
	 picks on the foot (so it's easier to move the gate around.)"
	glass translation: 0@(height/2 +0.01)@(thickness*0.37);
		  extent: (width-(thickness/2.0))@(height-0.02).


! !

!QGate methodsFor: 'private' stamp: 'bgf 9/3/2007 06:06'!
updateKnob
	knob material: (
		self postcardLink ifNil: [
			QMaterials controlDisabledMaterial.
		] ifNotNil: [
			self isOpen ifTrue: [
				QMaterials controlOnMaterial.
			] ifFalse: [
				QMaterials controlMaterial.
			]]).
		
! !

!QGate methodsFor: 'events' stamp: 'bgf 7/17/2007 21:41'!
updatePostcard: newPostcard originalUrl: orig

	"Callback after postcard resolution - update my postcard link in place."

	"Throw away the stale postcard from my history."
	self	rolodex cards 
			select: [ : cd | cd kind = #postcardLink and: [ cd content urlString = orig]]
			thenDo: [:cd| self rolodex removeCard:cd].
	self postcardLink ifNotNilDo: [:pc |
		"for side effects, in case the postcard is shared with any far-away rolodexes."
		pc urlString = orig ifTrue: [ pc urlString: newPostcard urlString].
	].
	self postcardLink: newPostcard.
! !

!QGate methodsFor: 'initialize' stamp: 'bgf 7/17/2007 22:20'!
updateShape

	self updateFrame.
	self updateAccessories.! !

!QGate methodsFor: 'accessing' stamp: 'bgf 11/17/2009 16:54'!
visible: aBool
	"Portals are rendered without regard to the visibility of their containers."
	portal ifNotNil: [ portal visible: aBool ].
	super visible: aBool! !

QGate subclass: #QToriGate
	instanceVariableNames: 'leftWing rightWing wings wingGrip'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Portals'!

!QToriGate commentStamp: '<historical>' prior: 0!
A fixed size, fixed location holder for 2D portals,
framed within a stylized gateway reminiscent of a Tori.

A gate is like a window in being a container for a portal,
with a goto button.  It is unlike a window in that it doesn't 
accept other kinds of content.

When closed, a gate has only smoked glass as content 
(so it is nearly transparent.)

When open, a gate retains the smoked glass,
so there is no picking through a gate.

Clicking the handle on a gate opens or closes the portal.
Clicking on the portal surface is a gesture that is bounced
to the harness, and means 'log in to the far space, if I'm not already synced to it.'
!

!QToriGate class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QToriGate methodsFor: 'persist' stamp: 'bgf 7/17/2007 23:35'!
constructFrom: pStream
	| |
	super constructFrom: pStream.

	wingGrip isEnabled: (pStream readBool: 'canRotate' default: true).
! !

!QToriGate methodsFor: 'initialize' stamp: 'bgf 7/17/2007 23:37'!
initialize

	super initialize.

	leftWing := TCube new.
	rightWing := TCube new.

	wings := TGroup new.
	wings addChild: leftWing.		"The decorative wings also serve as pivot grips."
	wings addChild: rightWing.
	self addChild: wings.

	wingGrip := QPlanePivot forLuggage: self grip: rightWing.
	{leftWing . rightWing} do: [:part|
		part material: left material clone.
	].
! !

!QToriGate methodsFor: 'accessing' stamp: 'bgf 5/29/2009 14:49'!
isPlacementLocked: aBool
	"We present as one boolean property 'lockPlacement'
	 the control over whether the gates resize menu option, 
	 and the gates drag-by-foot behaviors are avaliable."
	super isPlacementLocked: aBool.
	wingGrip isEnabled: aBool not.! !

!QToriGate methodsFor: 'persist' stamp: 'bgf 7/17/2007 23:35'!
persistOn: pStream
	super persistOn: pStream.
	! !

!QToriGate methodsFor: 'initialize' stamp: 'bgf 11/14/2009 10:28'!
updateShape

	super updateShape.

	top translation: 0@(height)@0.
	top extent: (width+(4*thickness))@thickness@(thickness*5).

	left translation: (width/2)@(height/2)@0.
	left extent: thickness@height@(thickness*4).

	right translation: (width/-2)@(height/2)@0.
	right extent: thickness@height@(thickness*4).

	portal extent: width@height.
	portal translation: 0@height/2@(thickness).
	returnFrame translation: (0@(height/2)@(thickness*4)).	"In plane of toe of feet"

	leftFoot extent: (thickness*1.5)@thickness@(thickness*8).
	leftFoot translation: (width/2)@(thickness/2)@0.

	rightFoot extent: (thickness*1.5)@thickness@(thickness*8).
	rightFoot translation: (width/-2)@(thickness/2)@0.

	leftWing extent: (thickness*8)@(thickness)@(thickness*5).
	leftWing translation: (width/2)@(height+(top extent y/2)+(thickness*0.54))@0.

	rightWing extent: (thickness*8)@(thickness)@(thickness*5).
	rightWing translation: (width/-2)@(height+(top extent y/2)+(thickness*0.54))@0.

	gotoButton translation: 0@(height+(thickness*3)+(gotoButton extent y/2))@0.

	rolodex translation: ((width+(3.5*thickness)+(rolodex width))/-2)@(height/2)@0.1.
	knob translation: (width+(3.6*thickness)/2)@(height/2)@(thickness/2).

	foot extent: width@(thickness/2)@thickness.
	foot translation: 0@(thickness/4)@0.

	glass translation: 0@(height/2)@(thickness*1.5).
	glass extent: width@height@(thickness*1.5).

	door translation: 0@(height/2)@(thickness*-1.65);
		  extent: width@height.

	menuButton 
		translation: (rightWing translation + ((thickness*1.9)@(thickness*-0.55)@(thickness*2.55)));
		extent: ((thickness*3)@(thickness*1.4)) + 0.22.

	label ifNotNil: [
		label translationX: 0 y: (height+(3*thickness)) z: 0.
	].
! !

TGroup subclass: #QLectern
	instanceVariableNames: 'base front shelf destination shadow'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QLectern commentStamp: '<historical>' prior: 0!
A Lectern object, used to define a listed landmark
that places one in a speaker position.
Has an angled shelf suitable for terminal-style scaled down display panels.

!

!QLectern class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLectern class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:25'!
addPropertyRows: dialog forFrame: lecternref

	dialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ :f | f get:#extent ];
			acceptBlock: [ :pt :f | f future extent: pt ];
			balloonHelp: 'Overall dimensions of the lectern.';
		yourself
	). 
	dialog addRow: (
		(QChoicesPropertyRow new)
			titleString: 'Places Category';
			setChoices: (QDestination categoryChoices);
			fetchBlock: [ : f | f get: #category ];
			acceptBlock: [ : t :f  | f future category: t ];
			balloonHelp: 'Where the entry appears\in Places' withCRs;
		yourself).

	super addPropertyRows: dialog forFrame: lecternref.

	"We fix some properties."
	dialog removeRow: 'Scale Factor'.! !

!QLectern class methodsFor: 'defaults' stamp: 'bgf 8/6/2009 17:34'!
defaultBodyMaterial
	^ QMaterials lightWoodMaterial! !

!QLectern class methodsFor: 'defaults' stamp: 'bgf 8/6/2009 17:27'!
defaultCategory
	^'Destinations'! !

!QLectern class methodsFor: 'defaults' stamp: 'bgf 8/6/2009 17:27'!
defaultDestinationMode
	^ #viewpoint! !

!QLectern class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:21'!
defaultEventMask
	^ EventPointerDown! !

!QLectern class methodsFor: 'defaults' stamp: 'bgf 8/6/2009 19:31'!
defaultExtent
	"Width, height, shelf-height"
	^ 3@3.9@1.2! !

!QLectern methodsFor: 'places' stamp: 'bgf 8/6/2009 21:15'!
asPlace
	^ self isOccupied ifFalse: [
		destination asPlace
	] ifTrue: [ 
		nil
	]! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 19:53'!
category
	"In Places list"
	^ destination category! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 19:52'!
category: aString
	"In Places list"
	destination category: aString! !

!QLectern methodsFor: 'hierarchy' stamp: 'bgf 8/6/2009 17:29'!
childrenForBrowser
	^ self childrenForPersistence! !

!QLectern methodsFor: 'hierarchy' stamp: 'bgf 11/14/2009 10:23'!
childrenForPersistence
	^ super children
			copyWithoutAll: { front  . base . shelf  . destination . shadow }.! !

!QLectern methodsFor: 'persist' stamp: 'bgf 8/6/2009 19:57'!
constructFrom: pStream
	super constructFrom: pStream.
	self class == QLectern ifTrue:[
		(pStream readSequence:'children' default: #()) do: [ :child | self addChild: child ].
	].
	self extent: ( pStream readVector3: 'extent' default: (self extent) ).
	self category: (pStream readString: 'category' default: (self category)).
! !

!QLectern methodsFor: 'initialize' stamp: 'bgf 8/6/2009 22:46'!
defaultIsAtomic
	"Parts like panels can be added right in to the lectern;	 
	 these are recognized places that the speaker might be at.   (See pointerDown: )"
	^ false! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 17:47'!
depth
	^ shelf extent y! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 22:42'!
depth: aScalar
	"Surface y of the lectern shelf."
	shelf extent: (shelf extent x) @ aScalar.
	self shapeChanged.! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 18:14'!
destination
	^ self! !

!QLectern methodsFor: 'places' stamp: 'bgf 8/6/2009 22:45'!
destinationStanceFor: anAvatar
	"So I always have an Approach context menu."
	^ destination destinationStanceFor: anAvatar! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 17:58'!
extent
	^ Vector3 x: self width y: self height z: self depth! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 17:57'!
extent: vXYZ
	self width: (vXYZ x).
	self height: (vXYZ y).
	self depth: (vXYZ z).! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 17:37'!
height
	^ front extentY! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 22:42'!
height: y
	"The height is from the ground to the intersection of post and lectern shelf"
	front extentY: y.
	self shapeChanged.! !

!QLectern methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:33'!
initialize
	super initialize.
	self objectName: 'Lectern'.

	destination := (QDestination new)
						destinationMode: #viewpoint;
						suggestedPOV: 'First Person';
						buttonLabel: (self objectName)
						yourself.
	self addChild: destination.

	base := QLozenge new.
	base material: (self class defaultBodyMaterial).
	base addRotationAroundX: -90.
	self addChild: base.

	front := TCube new.
	front material: (QMaterials darkGreyMaterial).
	self addChild: front.

	shelf := QLozenge new.
	shelf material: (self class defaultBodyMaterial).
	shelf addRotationAroundX: -40.
	self addChild: shelf.

	shadow := QMaterials newRectangularShadow.
	shadow addRotationAroundX: -90.
	self addChild: shadow.

	{ destination . base . front . shelf  . shadow } do: [ : f |
		f shouldPersist: false
	].
	self extent: (self class defaultExtent).! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 17:31'!
isComponent
	^ self shouldPersist! !

!QLectern methodsFor: 'testing' stamp: 'bgf 8/6/2009 17:30'!
isEditableGroup
	"C.F. TGroup; plain destination instances are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QLectern) and: [ self isAtomic not ]! !

!QLectern methodsFor: 'testing' stamp: 'bgf 8/6/2009 17:30'!
isGroundBound
	^ true! !

!QLectern methodsFor: 'testing' stamp: 'bgf 8/6/2009 21:16'!
isOccupied
	"Anyone here?"
	self root ifNotNilDo: [ : root |
		^ (root allUsers anySatisfy: [ : av | 
			av currentPlace ifNotNilDo: [ : aplace | aplace destinationID = objectID ] ifNil: [false]])
	].
	^ false! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 9/3/2009 15:16'!
objectName: name
	destination ifNotNil: [
		destination buttonLabel: name
	].
	super objectName: name! !

!QLectern methodsFor: 'persist' stamp: 'bgf 8/6/2009 19:57'!
persistOn: pStream
	super persistOn: pStream.
	self class == QLectern ifTrue:[
		pStream storeSequence: 'children' value: (self childrenForPersistence) default: #().
	].
	pStream storeVector3: 'extent' value: (self extent) default: (self class defaultExtent).
	pStream storeString: 'category' value: (self category) default: (self class defaultCategory).
! !

!QLectern methodsFor: 'events' stamp: 'bgf 8/6/2009 22:44'!
pointerDown: event
	" A couple of usability accelerators mapped to direct-click on the lectern parts..."

	event avatar ifNil: [ ^ self ].
	( event selection notNil  and: [ event selection frame notNil ] ) ifFalse: [ ^ self ].

	"If the lectern is free, clicking the black-body takes you to it."
	(self isOccupied not and: [ event selection frame = front ]) 
	 ifTrue: [
		^ event avatar goToPlace: (self asPlace).
	].

	"If the clicker is at the lectern or any panel it contains,
	 then clicking on the lectern shelf returns to first-person out-facing view.
	 So there is a two-click sequence to go from panel to audience to panel,
	 or vice-versa"
	event avatar nominalPlace ifNil: [ ^ self ].
	event avatar nominalPlace destinationFrame ifNil: [ ^ self ].
	(self isOrContains: (event avatar nominalPlace destinationFrame))
	ifTrue: [
		event avatar goToPlace: (self asPlace).
	].! !

!QLectern methodsFor: 'transform' stamp: 'bgf 2/11/2010 17:49'!
privateScale: aFactor
	aFactor = 1 ifFalse: [
		self height: (self height * aFactor).
		self width: (self width * aFactor).
		self depth: (self depth * aFactor).
	].
	nomScale := nomScale * aFactor.

	"Scale non-builtin children."
	self childrenForBrowser do: [ : bc | bc scale: aFactor ].
	self shapeChanged.! !

!QLectern methodsFor: 'hierarchy' stamp: 'bgf 2/11/2010 17:54'!
updateShape
	| thickness backoff |
	super updateShape.

	"We pull the defining dimensions out of the parts we keep them in,
	 and use those to compute everything else."
	thickness := (self extent max * 0.02 * self scale).
	backoff := (-2.3 * self scale).
	front
		translation: 0 @ (self height * 0.5) @  backoff;
		extentX: (self width * 0.333);
		extentZ: thickness.
	base	
		extent: (self width *0.5 + (3*thickness) @ (5*thickness)) ;
		radius: thickness*0.5;
		translation: (0@(thickness*0.5)@backoff).
	shelf 
		radius: thickness;
		translation: (0@(self height + (thickness*0.5))@(backoff)).
	destination 
		buttonPosition: (0@8.0@0) * (self scale).

	shadow extent: (base extent);
		      borderWidth: (thickness*2);
		      translation: (0@(0.1*thickness)@backoff).
	self unionSphereChanged.! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 18:41'!
width
	^shelf extent x! !

!QLectern methodsFor: 'accessing' stamp: 'bgf 8/6/2009 22:41'!
width: x
	"The width is across the main lectern shelf"
	shelf extent: (x @ shelf extent y).
	self shapeChanged.! !

TGroup subclass: #QMagicMap
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QMagicMap commentStamp: 'das 2/21/2006 09:02' prior: 0!
TMagicMap

I am used to drag the avatar around in a larger world. This feels like you are dragging the world around, though.

Important!! The Avatar must do all of the work. The TMagicMap cannot hold dynamic state itself because it needs to be available to multiple users.!

!QMagicMap class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMagicMap class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:22'!
defaultEventMask
	^ EventPointerDown! !

!QMagicMap methodsFor: 'initialize' stamp: 'jdm 8/21/2007 08:16'!
content: cont

	contents := cont.
	self localTransform: cont localTransform.
	cont localTransform: Matrix4x4 new.
	cont insertFrame: self.
! !

!QMagicMap methodsFor: 'events' stamp: 'bgf 6/21/2007 21:51'!
doPointerDown: event

	event isAvatarValid ifFalse: [ ^ self ].
	event selection frame ifNil:[
		event selection frame: (self getOwnedObject: event selection frameID).
	].
	self pointerDown: event.
! !

!QMagicMap methodsFor: 'events' stamp: 'bgf 6/21/2007 21:52'!
doPointerMove: event

	event isAvatarValid ifFalse: [ ^ self ].
	event selection frame ifNil:[
		event selection frame: (self getOwnedObject: event selection frameID).
	].
	self pointerMove: event.
! !

!QMagicMap methodsFor: 'events' stamp: 'bgf 6/21/2007 21:53'!
doPointerUp: event

	event isAvatarValid ifFalse: [ self ].
	event selection frame ifNil:[
		event selection frame: (self getOwnedObject: event selection frameID).
	].
	self pointerUp: event.
! !

!QMagicMap methodsFor: 'events' stamp: 'das 3/7/2007 16:41'!
pointerDown: event 

	| avt |
	avt := event avatar.
	"Is the avatar an extended Meredith avatar in this space?"
	(avt isKindOf: QAvatarReplica) ifTrue: [
		avt selectedFrame: event selection frame.
		avt selectedPoint: event selection point.
		].! !

!QMagicMap methodsFor: 'events' stamp: 'jdm 8/16/2007 16:54'!
pointerMove: event

	| pointer avt  delta |
	avt := event avatar.

	"Is the avatar an extended Meredith avatar in this space?"
	(avt isKindOf: QAvatarReplica) ifTrue: [
		pointer := event makePointer.
		pointer frame: avt selectedFrame pickPlane: avt selectedPoint normal: 0@1@0.
		delta := avt selectedPoint - pointer selectedPoint.
		delta := avt selectedFrame globalTransform orientation transformPoint: delta.
		avt translation: avt translation + delta.
	event selection cameraTransform translation: event selection cameraTransform translation+delta.
	].! !

!QMagicMap methodsFor: 'events' stamp: 'das 3/7/2007 16:41'!
pointerUp: event


	| avt |
	avt := event avatar.
	"Is the avatar an extended Meredith avatar in this space?"
	(avt isKindOf: QAvatarReplica) ifTrue: [
		avt selectedPoint: nil.
		avt selectedFrame: nil.
	].! !

TGroup subclass: #QMultiBoardEditButtons
	instanceVariableNames: 'board extent plus minus up down clear'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QMultiBoardEditButtons commentStamp: '<historical>' prior: 0!
Part of the QTabbed Wall.   
Keeps a  row of square labeled buttons
for the basic editing actions on a tabbed-wall.
(Add/remove, reorder up and down, and clear content).
!

!QMultiBoardEditButtons class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMultiBoardEditButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 14:43'!
defaultButtonBackground
	^ Color r: 0 g: 0 b: 0 alpha: 0.3! !

!QMultiBoardEditButtons class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QMultiBoardEditButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 01:09'!
defaultExtent
	^ 1.0@2.0! !

!QMultiBoardEditButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 01:09'!
defaultTextColor
	^ Color white! !

!QMultiBoardEditButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 10:27'!
disabledTextColor
	^ (Color gray: 0.5) alpha: 0.6! !

!QMultiBoardEditButtons methodsFor: 'actions' stamp: 'bgf 3/8/2010 11:44'!
actionAddTab: event
	"I'll add the tab in the editor callback  #newTabTitle: "
	QTextEditJob 
			edit: '<Untitled>'
			title: 'New Tab' 
			forUser: event avatar 
			recipient: self 
			selector: #newTabTitle:utcSeconds:avatar:.! !

!QMultiBoardEditButtons methodsFor: 'actions' stamp: 'bgf 3/12/2010 15:04'!
actionClearTab: event

	event avatar ifNil: [ ^ self ].
	board checkStaleNotes.
	board clearCurrentTab: event
! !

!QMultiBoardEditButtons methodsFor: 'actions' stamp: 'bgf 3/13/2010 08:52'!
actionDownTab: event
	| crsr |
	crsr := board cursor.
	board tradeTabForward: (board cursor) forEvent: event.
	event avatar recordAction: #board:tab:direction:
				class: QMultiBoard
				arguments:  { board . crsr . 1 }.
! !

!QMultiBoardEditButtons methodsFor: 'actions' stamp: 'bgf 8/5/2009 02:39'!
actionRemoveTab: event
	board checkRemoveTab: event! !

!QMultiBoardEditButtons methodsFor: 'actions' stamp: 'bgf 3/13/2010 08:53'!
actionUpTab: event

	| crsr |
	crsr := board cursor.
	board tradeTabBackward: (board cursor) forEvent: event.
	event avatar recordAction: #board:tab:direction:
				class: QMultiBoard
				arguments:  { board . crsr . -1 }.
! !

!QMultiBoardEditButtons methodsFor: 'initialize' stamp: 'bgf 8/5/2009 01:09'!
board: aSheet

	board := aSheet.
	self updateButtons.
	
	board ifNotNil: [
		self startScript: #updateButtons when: { board . #notesChanged }
	].! !

!QMultiBoardEditButtons methodsFor: 'events' stamp: 'bgf 8/5/2009 02:25'!
button: aButton enabled: bEnabled

	bEnabled ifTrue: [
		aButton textColor: (self class defaultTextColor); pickable: true
	] ifFalse: [
		aButton textColor: (self class disabledTextColor); pickable: false
	].! !

!QMultiBoardEditButtons methodsFor: 'accessing' stamp: 'bgf 8/5/2009 01:09'!
buttons
	^ frameChildren ifNil: [ #() ]! !

!QMultiBoardEditButtons methodsFor: 'actions' stamp: 'bgf 2/9/2010 23:47'!
clearCurrentTab: event
	board clearCurrentTab: event! !

!QMultiBoardEditButtons methodsFor: 'initialize' stamp: 'bgf 8/5/2009 04:54'!
editToolClass
	"These text edit jobs (renaming tabs) always want one-liners."
	^ QTextLinePane! !

!QMultiBoardEditButtons methodsFor: 'accessing' stamp: 'bgf 8/5/2009 01:09'!
extent: xy

	extent = xy ifFalse: [
		extent := xy.
		self shapeChanged
	].! !

!QMultiBoardEditButtons methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:33'!
initialize
	super initialize.
	extent := self class defaultExtent.

	plus := self makeButton: #actionAddTab: label: '+'.
	minus := self makeButton: #actionRemoveTab: label: '-'.
	up := self makeButton: #actionUpTab: label: '^'.
	down := self makeButton: #actionDownTab: label: '^'.
	down addRotationAroundZ: 180.

	clear := self makeButton: #actionClearTab: label: ' Clear \Tab' withCRs.! !

!QMultiBoardEditButtons methodsFor: 'initialize' stamp: 'bgf 8/20/2009 22:11'!
makeButton: action label: string
	"My button is just a QLabelButton."
	| label |
	label := (QLabelButton new) 
			target: self action: action;
			contents: (string);
			textColor: (self class defaultTextColor);
			backgroundColor: (self class defaultButtonBackground);
			yourself.

	self addChild: label.
	^ label! !

!QMultiBoardEditButtons methodsFor: 'accessing' stamp: 'bgf 3/13/2010 08:53'!
newTabTitle: string utcSeconds: secs avatar: myAvatar 
	"Callback from the 'edit new label' edit job we fire automatically on tab creation.
	 If the user has entered a tab label, create a tab and label it."

	(string = '<Untitled>')  ifFalse:[
		| oldCursor |
		oldCursor := board cursor.
		board addTab: (board cursor) withLabel: (string withBlanksTrimmed).
		board cursor: (board cursor + 1).
		myAvatar
			recordAction: #board:newTab:name:oldCursor:
			class: QMultiBoard
			arguments: { board . oldCursor .  (string withBlanksTrimmed) . oldCursor }.
	].! !

!QMultiBoardEditButtons methodsFor: 'events' stamp: 'bgf 8/5/2009 01:09'!
pointerLeave: event
	self buttons do: [ : b | b unhilite ].
	! !

!QMultiBoardEditButtons methodsFor: 'events' stamp: 'bgf 8/20/2009 20:42'!
pointerOver: event

	self buttons do: [ : b | b unhilite ].
	! !

!QMultiBoardEditButtons methodsFor: 'hierarchy' stamp: 'bgf 8/5/2009 04:55'!
updateButtons
	"Button availability... depends on the board cursor position and how many tabs there are."
	self button: plus enabled: true.
	self button: minus enabled: (board numTabs > 1).
	self button: up enabled: (board cursor > 1).
	self button: down enabled: (board cursor < board numTabs).
	self button: clear enabled: (board notes size > 0).

	self updateShape.! !

!QMultiBoardEditButtons methodsFor: 'hierarchy' stamp: 'bgf 6/10/2010 12:46'!
updateShape
	| bs firstX dim |
	super updateShape.

	self buttons isEmptyOrNil ifTrue: [ ^ nil ].	"Being dismantled."

	"Fit the single row of five buttons exactly how we want them."
	dim := extent y min: (extent x / (self buttons size)).
	bs := (dim @ dim).
	firstX := (extent x * 0.925 - bs x) * -0.5.
	self buttons doWithIndex: [ : button : i |
		button extent: bs * 0.8;
			     translation: (firstX + ((i-1)*(dim*0.9))) @ (dim * 0.05) @ 0.
	].

	self unionSphereChanged.! !

TGroup subclass: #QMultiBoardTabButtons
	instanceVariableNames: 'board extent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Project'!

!QMultiBoardTabButtons commentStamp: '<historical>' prior: 0!
Part of the QTabbed Wall.   
Keeps a column of stylized labels as a radio-button group
from which one can select the current page to show in the multiboard.!

!QMultiBoardTabButtons class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMultiBoardTabButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 14:47'!
currentButtonBackground
	"The active tab"
	^ Color r: 0.8 g: 0.35 b: 0.005 alpha: 0.8! !

!QMultiBoardTabButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 14:43'!
defaultButtonBackground
	^ Color r: 0 g: 0 b: 0 alpha: 0.3! !

!QMultiBoardTabButtons class methodsFor: 'defaults' stamp: 'bgf 8/5/2009 14:48'!
defaultButtonSize
	"Used as basis for proportions when fitting buttons into the outer extent."
	^ 1.4@0.225! !

!QMultiBoardTabButtons class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QMultiBoardTabButtons class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 22:28'!
defaultExtent
	^ 1.0@2.0! !

!QMultiBoardTabButtons class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 22:44'!
defaultTextColor
	^ Color white! !

!QMultiBoardTabButtons methodsFor: 'initialize' stamp: 'bgf 8/4/2009 21:19'!
board: aSheet

	board := aSheet.
	self updateButtons.
	
	board ifNotNil: [
		self startScript: #updateButtons when: { board . #notesChanged }
	].! !

!QMultiBoardTabButtons methodsFor: 'accessing' stamp: 'bgf 8/4/2009 20:22'!
buttons
	^ frameChildren ifNil: [ #() ]! !

!QMultiBoardTabButtons methodsFor: 'events' stamp: 'bgf 3/29/2010 10:38'!
doubleClick: event clientData: index
	board numTabs >= index ifTrue: [ 	
		QTextEditJob 
			edit: (board tabNameAt: index) 
			title: 'Rename Tab' 
			forUser: event avatar 
			recipient: self 
			selector: #setTabTitle:utcSeconds:avatar:index:
			extraData: index.
	].
! !

!QMultiBoardTabButtons methodsFor: 'accessing' stamp: 'bgf 8/5/2009 04:53'!
editToolClass
	"These text edit jobs (renaming tabs) always want one-liners."
	^ QTextLinePane! !

!QMultiBoardTabButtons methodsFor: 'accessing' stamp: 'bgf 8/4/2009 20:48'!
extent: xy

	extent = xy ifFalse: [
		extent := xy.
		self shapeChanged
	].! !

!QMultiBoardTabButtons methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:33'!
initialize
	super initialize.
	extent := self class defaultExtent.

	self startScript: #doubleClick:clientData: when: { self . #buttonDoubleClick }.

	"Most of the setup happens when my board: is set."! !

!QMultiBoardTabButtons methodsFor: 'accessing' stamp: 'bgf 8/5/2009 14:49'!
labelForTab: i
	"The spaces give us a bit of horizontal-only margin in the labels."
	^' ', (board tabNameAt: i), ' '
		! !

!QMultiBoardTabButtons methodsFor: 'initialize' stamp: 'bgf 8/20/2009 21:33'!
makeButton: pageNum
	| label |
	label := (QLabelButton new) 
			contents: (self labelForTab: pageNum);
			textAnchorPoint: #centerLeft;
			backgroundColor: (Color black alpha: 0.2);
			target: self action: #singleClick:clientData:;
			yourself.

	^ label! !

!QMultiBoardTabButtons methodsFor: 'events' stamp: 'bgf 8/4/2009 21:08'!
pointerLeave: event
	self buttons do: [ : b | b unhilite ].
	! !

!QMultiBoardTabButtons methodsFor: 'events' stamp: 'bgf 11/8/2009 11:05'!
pointerOver: event

	self buttons do: [ : b | b unhilite ].
	event selection ifNil: [ ^ self ].
	event selection frame ifNotNilDo: [  : sel |
		(self buttons includes: sel) ifTrue: [
			sel hilite: (VectorColor r: 8 g: 0.9 b: 0.2 alpha: 0.8).		
			event avatar signal: #statusMessage with: 'Double-click to edit the title of the tab.'
		]].

	! !

!QMultiBoardTabButtons methodsFor: 'accessing' stamp: 'bgf 3/29/2010 10:43'!
setTabTitle: string utcSeconds: secs avatar: avatar index: index

	board numTabs >= index ifTrue: [
		board tabNameAt: index put: string withBlanksTrimmed.
	] ifFalse: [
		avatar signal: #alertMessage with:  'Cannot set tab title; tab is absent.'
	].! !

!QMultiBoardTabButtons methodsFor: 'events' stamp: 'bgf 8/20/2009 22:01'!
singleClick: event clientData: index
	"Callback from any tab button."
	board cursor: index! !

!QMultiBoardTabButtons methodsFor: 'hierarchy' stamp: 'bgf 8/20/2009 22:15'!
updateButtons
	"Bring the tab-buttons into agreement with the current set of tabs in the multiboard."
	[self buttons size < board numTabs] whileTrue: [
		self addChild: (self makeButton: (self buttons size + 1)).
	].

	[self buttons size > board numTabs] whileTrue: [
		| b |
		b := self buttons last.
		b obliterateLater.
	].

	1 to: board numTabs do: [ : pi |
		(self buttons at: pi) 
			clientData: pi;
			contents: (self labelForTab: pi );
			backgroundColor: (
					board cursor = pi 
						ifTrue: [ self class currentButtonBackground ] 
						ifFalse: [ self class defaultButtonBackground ]  ).
	].
	self shapeChanged.! !

!QMultiBoardTabButtons methodsFor: 'hierarchy' stamp: 'bgf 8/10/2009 10:35'!
updateShape
	| bs topLine hot |
	super updateShape.

	bs := (self class defaultButtonSize) * (extent x / (self class defaultButtonSize x)).
	bs := (extent x) @ (bs y min: (extent y / (self buttons size max: 1))).
	topLine := (extent y * 0.5 - (bs y )).

	self buttons doWithIndex: [ : button : i |
		hot := (i = board cursor).
		button extent: (hot ifTrue: [ bs x * 0.96 ] ifFalse: [ bs x * 0.9 ])
						@ (bs y * 0.89);
			     translation: (hot ifTrue: [ bs x * 0.03] ifFalse: [ bs x * 0.002 ]) 
						@ (topLine - ((i-1) * (bs y))) @ 0.
	].

	self unionSphereChanged.! !

TGroup subclass: #QPanelFrame
	instanceVariableNames: 'extent thickness movableBars resizeCase selectedPoint wasDragged assembly isEnabled aspect originalState'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QPanelFrame commentStamp: '<historical>' prior: 0!
Abstract class for a resizable 2D rectangular picture frame,
used by display panels and other resizeable flat objects.

Panel frames are generally resizable (by dragging their edges)
and usually affixed to a flat surface.

A concrete subclass must implement accessors for the TFrames
that represent the edges of the frame...
	#bars, #eastBar, #westBar, #northBar, #southBar
(where 'bars' is usually a collection of the 4 individual objects.)
It should also implement 'updateShape', and in that method,
resize the frame to surround 'extent'.

The resizing behavior is partly a responsibility of the PanelFrame's #assembly.
The assembly of a QPanelFrame must implement:
	#dragEdge:delta:
	#clickEdge:

(See QDisplayPanel for examples, and #pointerMove:, #pointerUp: in this class.)

Edges can be pinned to suppress resize through them; see #pinEdge:, #pinEdges.
(Edges are identified by compass cardinal names #north, #south, #east, #west).

Vertical alignment (where the vertical origin of the assembly falls)
is specified through vAlign:; the default is #center.
A wall uses a handler with the south edge pinned, and vAlign #bottom.

!

QPanelFrame subclass: #QBoxFrame
	instanceVariableNames: 'northBar southBar eastBar westBar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QBoxFrame commentStamp: '<historical>' prior: 0!
A simple, symmettric 4-sided panel frame,
for use as the resizing control of a flat object.!

!QBoxFrame class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QBoxFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:20'!
bars
	^ { northBar . westBar . eastBar . southBar }.! !

!QBoxFrame methodsFor: 'persist' stamp: 'bgf 4/17/2007 15:26'!
constructFrom: pStream
	super constructFrom:pStream.
	self frameMaterial: 
			(pStream readObject: 'frameMaterial' default: (QMaterials darkGreyMaterial)).! !

!QBoxFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:20'!
eastBar
	^ eastBar! !

!QBoxFrame methodsFor: 'events' stamp: 'bgf 4/17/2007 15:27'!
enable

	"Handle interactions."
	super enable.

	"Workaround loss of event routing - 
		force updates to bar objectOwner and event mask.
	 These are lost in QDisplayPanel, 
		which re-uses some bars for grips for rotate and move.
	 Installing and removing event scripts for objects which sometimes are owned,
	 then become their own handlers, and then return to being owned, seems
	 to confuse something in the on: notify:  and/or objectOwner code."

	{northBar . southBar . eastBar . westBar } do: [ :bar|
		self removeChild: bar.
		self addChild: bar.
	].! !

!QBoxFrame methodsFor: 'initialize' stamp: 'bgf 4/17/2007 16:24'!
initialize

	| frameStuff |

	super initialize.

	frameStuff := QMaterials darkGreyMaterial.

	northBar := TCube new.
	northBar material: frameStuff copy.
	self addChild: northBar.

	westBar := TCube new.
	westBar material: frameStuff copy.
	self addChild: westBar.

	eastBar := TCube new.
	eastBar material: frameStuff copy.
	self addChild: eastBar.

	southBar := TCube new.
	southBar material: frameStuff copy.
	self addChild: southBar.

	movableBars := {northBar. westBar . eastBar . southBar }.
	self shapeChanged.! !

!QBoxFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:20'!
northBar
	^ northBar! !

!QBoxFrame methodsFor: 'persist' stamp: 'bgf 4/17/2007 15:27'!
persistOn: pStream
	super persistOn:pStream.
	pStream storeObject: 'frameMaterial' 
				value: (northBar material) default: (QMaterials darkGreyMaterial).! !

!QBoxFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:20'!
southBar
	^ southBar! !

!QBoxFrame methodsFor: 'initialize' stamp: 'bgf 7/23/2007 10:32'!
updateShape

	| wid |

	"Size the bars."

	wid := 2.5 * thickness.

	northBar extent:(extent x +wid + thickness)@wid@thickness;
			translation: 0@(extent  y + thickness /2)@0.

	southBar extent: (extent x+wid + thickness)@wid@thickness;
			translation: 0@(extent y + thickness /-2)@0.

	westBar extent: wid@(extent y + wid + thickness)@(thickness);
			translation: (extent x + thickness / -2)@0@0.

	eastBar extent: wid@(extent y +wid + thickness)@(thickness);
			translation: (extent x+ thickness /2)@0@0.
! !

!QBoxFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:20'!
westBar
	^ westBar! !

QBoxFrame subclass: #QFadingBoxFrame
	instanceVariableNames: 'fader'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QFadingBoxFrame commentStamp: '<historical>' prior: 0!
A 2-D resizing frame with auto-fade-out behavior,
so it is only visible when requested.

The resizable assembly calls enable to make the frame visible,
and disable to start its fadeout.   (This is normally done
by #pointerOver: and #pointerLeave: handlers on the assembly.)!

!QFadingBoxFrame class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFadingBoxFrame methodsFor: 'action' stamp: 'bgf 8/25/2006 20:51'!
disable
	fader disable.! !

!QFadingBoxFrame methodsFor: 'action' stamp: 'bgf 8/25/2006 20:51'!
enable
	fader enable.! !

!QFadingBoxFrame methodsFor: 'initialize' stamp: 'bgf 8/25/2006 20:49'!
initialize
	super initialize.
	
	fader := QControlFader for: self.! !

!QFadingBoxFrame methodsFor: 'accessing' stamp: 'bgf 1/21/2007 13:22'!
overColor
	^ (QColors controlOverColor alpha: (fader baseAlpha))! !

QPanelFrame subclass: #QFlatFrame
	instanceVariableNames: 'backdrop northBar label avatarFocus avatarColors corner resizer backface'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QFlatFrame commentStamp: '<historical>' prior: 0!
The resizing and title frame for a QDisplayPanel.
(Should be called QDisplayFrame, but older versions
used this name and are named in persistence files.)

The top bar is extra-wide, usually dark gray, and bears a title string.
Insted of edge bars, the rest of the resizing is done with a 'corner'
which is placed to align exactly with the display panel's old-document rolodex.

Supports an event conspiracy with QPanelHandler
to show pointer-over feedback.!

!QFlatFrame class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFlatFrame class methodsFor: 'defaults' stamp: 'bgf 9/5/2008 18:36'!
cornerMaterial
	^ (QMaterials resizeCornerBlackMaterial)	
! !

!QFlatFrame class methodsFor: 'defaults' stamp: 'bgf 9/2/2010 12:14'!
defaultMaterial
	^ QMaterials darkGreyMaterial! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/7/2008 03:16'!
assembly: panelHandler
	super assembly: panelHandler.
	resizer luggage: (panelHandler ifNotNil: [ panelHandler assembly ]).! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/7/2008 03:17'!
backMaterial: m
	"Set the backdrop material (both sides)."
	backdrop material: m.
! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/5/2008 17:06'!
bars
	^ { northBar }.! !

!QFlatFrame methodsFor: 'private' stamp: 'bgf 4/17/2007 15:20'!
clearAssemblyEvents
	"If we have registered for events with the assembly, remove them."
	assembly ifNil: [^self].
	assembly stopScript: #highlightForAvatarFocus:.
	assembly stopScript: #removeHighlightForAvatarFocus.
	! !

!QFlatFrame methodsFor: 'private' stamp: 'das 10/22/2007 16:54'!
colorizeBackdrop: aColor

	"Apply the given avatar color as the material color behind the backdrop;
	 if the color is nil, reset the backdrop."

	aColor ifNil: [
		backdrop material colorize: (QColors backdropColor).
		backdrop material emissiveColor: Color black.
		assembly ifNotNil:[
			assembly broadcastColor: nil.
		].
	] ifNotNil: [
		backdrop material colorize: (QColors backdropColor / 4.0).
		backdrop material emissiveColor: aColor.
		assembly ifNotNil:[
			assembly broadcastColor: aColor.
		].
	].! !

!QFlatFrame methodsFor: 'persist' stamp: 'bgf 9/1/2010 17:19'!
constructFrom: pStream
	super constructFrom:pStream.
	self privateFrameMaterial: 
			(pStream readObject: 'frameMaterial' 
			 default: (QFlatFrame defaultMaterial))! !

!QFlatFrame methodsFor: 'events' stamp: 'bgf 9/23/2008 15:13'!
disable
	"Suppress resize interaction."
	super disable.

	"Work around loss of event routing - force updates to bar objectOwner and event mask.
	 These are lost in QDisplayPanel, which re-uses some bars for grips for rotate and move.
	 Installing and removing event scripts for objects which sometimes are owned,
	 then become their own handlers, and then return to being owned, seems
	 to confuse something in the on: notify:  and/or objectOwner code."
	resizer enableClickOnly.
	{northBar} do: [ :bar|
		self removeChild: bar.
		self addChild: bar.
	].! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/5/2008 17:06'!
eastBar
	^ nil! !

!QFlatFrame methodsFor: 'events' stamp: 'bgf 9/7/2008 03:18'!
enable
	"Allow interactive resizing and moving."
	super enable.
	resizer enable.
	
	"Workaround loss of event routing - force updates to bar objectOwner and event mask.
	 These are lost in QDisplayPanel, which re-uses some bars for grips for rotate and move.
	 Installing and removing event scripts for objects which sometimes are owned,
	 then become their own handlers, and then return to being owned, seems
	 to confuse something in the on: notify:  and/or objectOwner code."
	{northBar . corner } do: [ :bar|
		self removeChild: bar.
		self addChild: bar.
	].! !

!QFlatFrame methodsFor: 'private' stamp: 'bgf 12/21/2008 20:31'!
highlightForAvatarFocus: avatar
	"Show which avatar has focus by coloring the backdrop 	
	 the same as the avatar's laser color."
	| col |
	avatar ifNil: [ ^ self ].	"Don't try to claim focus across spaces."
	assembly isEmpty ifTrue: [ ^ self ].	"Don't colorize unused panels."

	col := avatar avatarColor.
	col ifNil: [ ^ self ].		"Not yet configured."

	col := (col*3+(1@1@1@1))/4.0.

	self colorizeBackdrop: col.
	avatarFocus := true.
! !

!QFlatFrame methodsFor: 'private' stamp: 'das 10/3/2007 20:36'!
highlightForAvatarLeave: userID
	"User left panel; update highlight"
	avatarColors := avatarColors reject:[:a| a key = userID].
	self updateAvatarColor.! !

!QFlatFrame methodsFor: 'private' stamp: 'bgf 12/21/2008 20:31'!
highlightForAvatarOver: event
	"Show which avatar has weak focus by coloring the backdrop 	
	 a faded version of the avatar's laser color."
	| avatar col |
	avatarFocus ifTrue:[^self].
	assembly isEmpty ifTrue: [ ^ self ].	"Don't colorize unused panels."
	avatar := event avatar.
	avatar ifNil: [ ^ self ].	"Don't try to claim focus across spaces."

	col := avatar avatarColor.
	col ifNil: [ ^ self ].		"Not yet configured."

	col := (col+(1@1@1@1))/2.0.

	"Not sure if we can get multiple enters here so make sure we only use one"
	(avatarColors anySatisfy:[:a| a key = avatar userID])
		ifFalse:[avatarColors add: avatar userID -> col].
	self updateAvatarColor.! !

!QFlatFrame methodsFor: 'private' stamp: 'bgf 9/7/2008 03:13'!
hiliteBar: bar
	"My only bar is the title bar. It has a lot of decorations, so we don't hilite it."! !

!QFlatFrame methodsFor: 'initialize' stamp: 'bgf 9/1/2010 17:11'!
initialize

	| |

	super initialize.
	northBar := QLozenge new.
	northBar material: (self class defaultMaterial).
	self addChild: northBar.

	corner := TSimpleRectangle new.
	corner material: (QMaterials resizeCornerDPMaterial).
	corner material cullFace: false.
	self addChild: corner.

	resizer := (QCorner2DResizer new)
				grip: corner;
				yourself.

	backdrop := TRectangle new.
	backdrop colorize: QColors backdropColor;
		divisions: 4.
	backdrop material texture: (QTextures blankPanel).
	backdrop uvScale: 0.998@0.99999.	"So the dark edge of the texture works without recycling."
	self addChild: backdrop.

	backface := TRectangle new.
	backface colorize: (QColors paneDarkerColor darker duller);
		divisions: 4.
	backface addRotationAroundY: 180.
	self addChild: backface.

	"By default the lower bars are movable, and invisible,
	 so the appearance is spare.  You can resize by dragging near edges."
	movableBars := { }.
	label := QLabel new.
	label 
		material: (QMaterials flatWhiteMaterial);
		textColor: Color white; 
		backgroundColor: Color transparent;
		margin: 0;
		contents: ''.
	northBar addChild: label.
	self shapeChanged.
	avatarFocus := false.
	avatarColors := OrderedCollection new.! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:42'!
label
	^ label! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:20'!
northBar
	^ northBar! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/5/2008 17:11'!
outerExtent
	"The 2-d extent of the frame plane (to the outsides of the bars.)"
	^ extent + ((thickness/2)@(thickness*2)).! !

!QFlatFrame methodsFor: 'persist' stamp: 'bgf 9/1/2010 17:19'!
persistOn: pStream
	super persistOn:pStream.
	pStream storeObject: 'frameMaterial' 
				value: (northBar material) default: (QFlatFrame defaultMaterial).! !

!QFlatFrame methodsFor: 'private' stamp: 'bgf 9/1/2010 18:48'!
privateFrameMaterial: aMaterial

	"Update the material to match the coloring of the given material,
	 withouth changing opacity... used to load custom titles from persistence."
	| o |
	self bars do: [:b|
		o := b material opacity.
		b material: aMaterial.
		b material opacity: o.
		b checkAlpha.
	].
! !

!QFlatFrame methodsFor: 'private' stamp: 'das 10/3/2007 20:36'!
registerForAssemblyEvents 
	"Register for events produced by our assembly."
	assembly ifNil: [^self].

	self runScript: #highlightForAvatarFocus: when: { assembly. #getKeyboardControl }.
	self runScript: #removeHighlightForAvatarFocus when: { assembly. #loseKeyboardControl }.
	self runScript: #highlightForAvatarOver: when: { assembly. #pointerEnter }.
	self runScript: #highlightForAvatarLeave: when: { assembly. #removeUser }.
! !

!QFlatFrame methodsFor: 'private' stamp: 'ar 5/17/2007 14:16'!
removeHighlightForAvatarFocus
	"No avatar has focus; change backdrop to default color."
	avatarFocus := false.
	self updateAvatarColor.! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/5/2008 17:11'!
southBar
	^ nil! !

!QFlatFrame methodsFor: 'private' stamp: 'bgf 12/21/2008 20:31'!
updateAvatarColor
	"Choose a color from the list of avatars currently over this panel.
	Currently we use the *last* color as this allows us to aquire focus
	by moving the cursor out and back into a panel. Alternatively, you 
	might choose the first color but this has a bit of a lock-out feel to it.
	If the panel is empty, clear the highlighting - we don't colorize empty panels,
	as the visuals are too strong and distracting."

	| color |
	avatarFocus ifTrue:[^self].

	assembly isEmpty 
		ifTrue: [ color := nil ]
		ifFalse: [ 
			color := avatarColors isEmpty 
				ifTrue:[nil]
				ifFalse:[avatarColors last value]].

	self colorizeBackdrop: color.
! !

!QFlatFrame methodsFor: 'initialize' stamp: 'bgf 9/6/2009 12:37'!
updateShape
	"Size all the parts to reflect the current extent."
	backdrop
		extent: ((extent +((thickness*0.25) @ (thickness*0.3))) max:(0.4@0.4));
		translationX: 0  y: (thickness*-0.075)  z: (thickness*-0.25).
	backface
		extent: ((extent +((thickness*0.25) @ (thickness*0.3))) max:(0.4@0.4));
		translationX: 0  y: (thickness*-0.075)  z: (thickness*-0.251).

	northBar
		extentX:(extent x+(thickness/2)) y: (2*thickness); 
		radius: (thickness*0.3);
		translationX: 0  y: ((extent y/2 + (thickness*0.98)))  z: 0.

	"Corner has to align best we can with the doc shelf, which has a 3-d frame surround;
	 these numbers tie in with the doc-shelf setup in Display Panel updateShape."
	corner
		extent: (3 * thickness)@ (3.02 *thickness) ;
		translationX: (extent x * 0.5 - (thickness*1.18)) 
				   y: (extent y * -0.5  - (1.63*thickness)) 
				   z: (thickness*0.22);
		rotationAroundX: -12.

	"Title is a child of the north bar."
	label
		translationX: 0 y: (thickness*-0.08) z: (thickness*0.4);
		extent: (extent x * 0.6) @ (thickness*2).

	self updateBounds.! !

!QFlatFrame methodsFor: 'accessing' stamp: 'bgf 9/5/2008 17:11'!
westBar
	^ nil! !

!QPanelFrame class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPanelFrame class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QPanelFrame class methodsFor: 'defaults' stamp: 'bgf 8/21/2009 01:06'!
defaultThickness

	"Basis for thickness of frame bars. 
	 (This is an overloaded magic number, 
	 since the details of display panel title bar and bottom-strip layout depend on it.)"
	^ 0.13! !

!QPanelFrame methodsFor: 'private' stamp: 'bgf 3/15/2010 11:04'!
applyStateVector: vec
	"Undo/redo support"
	assembly applyStateVector: vec
! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/6/2007 01:55'!
aspect
	"If not nil, a fixed aspect ratio that constrains interactive resize."
	^ aspect! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/6/2007 01:55'!
aspect: yOverX
	"If not nil, a fixed aspect ratio that constrains interactive resize."
	aspect := yOverX! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:21'!
assembly
	"The assembly that I resize; may be a QPanel, or an ancestor TFrame of mine.
	 Must respond to the interactive resizing methods."

	^ assembly! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 8/19/2009 12:37'!
assembly: anAncestor

	"The luggage is the outer TFrame that represents the assembly
	  of this frame and its content.  
	  Moving the FlatFrame moves the whole assembly,
	  and the assembly handles resize operations."

	"Must respond to the interactive resizing methods."

	assembly == anAncestor ifTrue: [^self].

	self clearAssemblyEvents.

	anAncestor ifNotNil: [
		(anAncestor respondsTo: #drag:corner:delta:) 
		ifFalse: [
			self error: 'Frame assembly must support edge protocol.'.
		].
	].

	assembly := anAncestor.
	self registerForAssemblyEvents.
! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:11'!
bars
	self subclassResponsibility! !

!QPanelFrame methodsFor: 'events' stamp: 'jdm 9/26/2007 16:11'!
classifyResize: event

	"Depending on where the event selection point lies,
	 choose one of the 8 compass points of resize.
	 The event must be a click on one of my drag bars.
	 Returns a symbol such as #north, #northeast, etc."

	| p north south east west |

	p := (self relativeTransform: (event selection frame)) transformPoint:(event selection point).

	north := false.
	south := false.
	east := false.
	west := false.

	"Which edges are under the pick?"
	(p x > ((extent x/2)-(thickness*1.5))) ifTrue: [
		east := true.
	] ifFalse: [ 
		(p x <((extent x/-2) + (thickness*1.5))) ifTrue: [ west := true ].
	].
	(p y > ((extent y/2)-(thickness*1.5))) ifTrue: [
		north := true.
	] ifFalse: [ 
		(p y < ((extent y/-2)+(thickness*1.5))) ifTrue: [ south := true ].
	].

	"Fold them together into a single result."
	(north and: [self movableBars includes: self northBar]) ifTrue: [
		(east and: [self movableBars includes: self eastBar]) ifTrue: [ ^ #northeast ].
		(west and: [self movableBars includes: self westBar]) ifTrue: [ ^ #northwest ].
		^ #north
	].
	(south and: [self movableBars includes: self southBar]) ifTrue:[
		(east and: [self movableBars includes: self eastBar]) ifTrue: [ ^ #southeast ].
		(west and: [self movableBars includes: self westBar]) ifTrue: [ ^ #southwest ].
		^ #south
	].
	(east and: [self movableBars includes: self eastBar]) ifTrue: [ ^ #east ].
	(west and: [self movableBars includes: self westBar]) ifTrue: [ ^ #west ].
	^ nil.! !

!QPanelFrame methodsFor: 'private' stamp: 'jcg 2/9/2007 18:47'!
clearAssemblyEvents
	"If we have registered for events with the assembly, remove them."
! !

!QPanelFrame methodsFor: 'persist' stamp: 'bgf 1/6/2007 01:56'!
constructFrom: pStream

	super constructFrom: pStream.

	self extent: (pStream readPoint: 'extent' default: (self extent)).
	self thickness: (pStream readNumber: 'thickness'  default: (self thickness)).
	self aspect: (pStream readNumber: 'aspect'  default: (self aspect)).
	isEnabled := (pStream readBool: 'isEnabled' default: isEnabled).
	
	"TODO - handle the menu, motion, and rotation suppression options;
	 for the moment, these are not persisted, and so are available only 
	 to objects which use the display panel as internal structure."! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 1/2/2007 16:07'!
disable
	"Stop handling interactions."
	isEnabled := false! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 3/15/2010 10:45'!
dispatchResize: selPoint forEvent: event
	"Send a dragCorner message to the assembly
	 to carry out the resizing step.   
	 We have to enforce the fixed aspect ratio of the mesh."
	| d sign |
	wasDragged := true.	"No click at the end. (Belt&braces)"
	d := (selPoint - selectedPoint).
	"Convert the delta into a delta along a diagonal.
	 We have to decide which dimension of the delta should be passed up."
	aspect ifNotNil: [
		(#(north south) includes: resizeCase) ifTrue: [
			d x: (d y / aspect).		
		] ifFalse: [
			(#(east west) includes: resizeCase) ifTrue: [
				d y: (d x * aspect).
			] ifFalse: [
				sign := 1.
				resizeCase caseOf: {
					[#northwest] -> [sign := -1].
					[#southeast] -> [sign := -1].
					[#northeast ] -> [ sign := 1].
					[#southwest ] -> [ sign := 1].
				}.

				d x abs > (d y abs) ifTrue: [
					d y:  d x * aspect * sign.
				] ifFalse: [
					d x: d y / aspect * sign.
				].
			].
		].
	] ifNil: [	"No aspect ratio constraint. "
		"We use 'dragCorner' and substitute a corner name when passing
		 this to the assembly.  Suppress the corresponding coordinate."
		(#(north south) includes: resizeCase) ifTrue:[ d x: 0].
		(#(east west) includes: resizeCase) ifTrue: [ d y: 0].
	].

	resizeCase caseOf: {
		[ #north ] -> [ self assembly drag: selPoint corner: #northeast delta: d].
		[ #south ] -> [ self assembly drag: selPoint corner: #southwest delta: d].
		[ #east ] -> [ self assembly drag: selPoint corner: #northeast delta: d].
		[ #west ] -> [ self assembly drag: selPoint corner: #southwest delta: d].
	} otherwise: [
		self assembly drag: selPoint corner: resizeCase delta:d.
	].
! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:11'!
eastBar
	self subclassResponsibility! !

!QPanelFrame methodsFor: 'private' stamp: 'bgf 1/2/2007 16:17'!
edgeBarAt: compassPt
	"Private: return the edge bar corresponding to the given compass point name"

	| |
	^ compassPt caseOf: {
		[#north] -> [self northBar].
		[#south] -> [self southBar].
		[#east] -> [self eastBar].
		[#west] -> [self westBar].
	} otherwise: [
		nil.
	].! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 1/2/2007 16:14'!
enable

	"Handle interactions."
	isEnabled := true.
! !

!QPanelFrame methodsFor: 'private' stamp: 'bgf 3/15/2010 11:32'!
endMove: event
	"Make an undoable version of whatever was done."
	| new |
	assembly doUpdateShape.
	new := assembly stateVector.
	(originalState hasEqualElements: new) ifFalse: [	
		event avatar
			recordAction: #frame:oldState:newState:description:
			class: QStateVectorAction
			arguments: {  self . originalState . new .  'Resize' }.
	].
		

! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:56'!
extent

	"The 2-d extent of the frame plane (to the insides of the bars.)"
	^ extent! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
extent: e
	extent := e.
	self shapeChanged.! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 12/13/2007 16:14'!
frameMaterial: m

	"Change the material of the frame bars, but not the backdrop." 
	| |
	self bars do: [:b|
		b material: m copy.
		b checkAlpha.
	].
! !

!QPanelFrame methodsFor: 'private' stamp: 'bgf 12/20/2007 02:38'!
hiliteBar: aBar
	"Hilite the given edge bar for resize feedback"
	aBar hilite: (QColors controlHiliteColor).! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 9/5/2008 17:13'!
hiliteEdgeByCase: aSymbol

	"Hilite the edge bar(s) denoted by the given resize case symbol."
	(aSymbol caseOf: {
			[#north] -> [ {self northBar} ].
			[#south] -> [ {self southBar} ].
			[#east] -> [ {self eastBar} ].
			[#west] -> [ {self westBar} ].
			[#northeast ] -> [ {self northBar . self eastBar} ].
			[#northwest ] -> [ {self northBar . self westBar} ].
			[#southeast ] -> [ {self southBar . self eastBar} ].
			[#southwest ] -> [ {self southBar . self westBar} ].
		} otherwise: [ {} ]) 
	do: [:bar|
		bar ifNotNil: [ self hiliteBar: bar ]
	].
! !

!QPanelFrame methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:34'!
initialize
	super initialize.

	isEnabled := true.
	thickness := self class defaultThickness.   "Thickness of frame bars."
	extent := 4@3.
	wasDragged := false.

	movableBars := Set new.
! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/23/2007 09:42'!
isEnabled
	^ isEnabled! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/23/2007 09:42'!
isEnabled: aBool
	isEnabled := aBool! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:37'!
label
	"If present, the QLabel that bears my title.  Absent by default."
	^ nil! !

!QPanelFrame methodsFor: 'initialize' stamp: 'bgf 8/21/2009 01:12'!
layoutGoto: gotoButton menu: menuButton startStop: ssButton 
	"Position the given display-frame controls in my layout.
	 The controls are a QGotoButton and two TCubes;
	 this default implementation suits the layout of QFlatFrame."

	gotoButton arrow extent: (
				QGotoButton defaultExtent 
					* ((self extent maxDimension / (QDisplayPanel defaultExtent maxDimension)) 
						atLeast: 0.1 atMost: 2)).
	gotoButton translationX:0 
				 y: (self outerExtent y/2)+(3*thickness)+(gotoButton extent y * 0.66) 
				 z:0.
	menuButton translationX: extent x*-0.5 + (thickness*2.2) 
				   	y: extent y*0.5 + (thickness*0.9)
					z: thickness*0.315;
		extent: (thickness*3.9)@(thickness*1.5).
	ssButton ifNotNil: [
		ssButton translationX: (extent x*0.5 - (thickness*2.1))
					y: (extent y *0.5 + (thickness*0.9)) 
					z: (thickness*0.315);
				extent: (thickness*3.5)@(thickness*1.8)@(thickness*0.55).
	].! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
movableBars
	^ movableBars! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:10'!
northBar
	self subclassResponsibility! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 2/27/2007 12:09'!
outerExtent

	"The 2-d extent of the frame plane (to the outsides of the bars.)"
	^ extent + (2*thickness).! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/21/2007 13:21'!
overColor
	"The hilite color of the frame during normal mouseover"

	^ QColors controlOverColor.! !

!QPanelFrame methodsFor: 'persist' stamp: 'bgf 1/3/2008 11:45'!
persistOn: pStream

	super persistOn: pStream.

	pStream storePoint: 'extent' value: self extent default: 6@4.5.
	pStream storeNumber: 'thickness' value: self thickness default: (self class defaultThickness).
	pStream storeNumber: 'aspect' value: self aspect default: nil.

	pStream storeBool: 'isEnabled' value: isEnabled default: true.
	
	"TODO - handle the menu, motion, and rotation suppression options;
	 for the moment, these are not persisted, and so are available only 
	 to objects which use the display panel as internal structure."! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
pinEdge: compassPt
	"Compass pt is one of #north, #south, #east, #west.
	 Fix the corresponding edge bar so that it cannot be dragged."

	| exclude |
	exclude := self edgeBarAt: compassPt.
	exclude ifNotNil: [
		movableBars := movableBars copyWithout: exclude.	
	].! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
pinEdges: collectionOfSymbols
	"Each edge is one of #north, #south, #east, #west.
	 Fix the corresponding edge bar so that it cannot be dragged."

	collectionOfSymbols do: [ :each|
		self pinEdge:each.
	].! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 10/8/2007 14:07'!
pointerClick: event
	"A non-dragging click in the frame."
	self signal: #pointerClick with: event.! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 3/15/2010 11:02'!
pointerDown: event

	"If the pointer is down in an edge bar, prepare for resize dragging.
	 We note what direction we're working in (by setting resizeCase)
	 and remember where exactly the resize action starts."
	wasDragged := false.
	isEnabled ifFalse: [
		^self
	].

	originalState := assembly stateVector.
	resizeCase := self classifyResize:event.
	resizeCase ifNotNil: [
		selectedPoint := (self relativeTransform: event selection frame)
							transformPoint: (event selection point).
		self hiliteEdgeByCase: resizeCase.
	].
! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 1/5/2007 21:35'!
pointerEnter: event

	self pointerOver: event! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 1/2/2007 16:07'!
pointerLeave: event

	isEnabled ifFalse: [
		^self
	].

	"Clear out all pointer interaction state."

	wasDragged := false.
	resizeCase := nil.
	self bars do: [ :b | b unhilite. ].
	self unhilite.! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 3/15/2010 10:45'!
pointerMove: event
	|  pointer |
	wasDragged := true.	"I.e. not a click."
	isEnabled ifFalse: [
		^self
	].

	"Resizing interaction... drag the edge(s) we started with."
	resizeCase ifNil: [
		self unhilite.
		^ self
	].

	pointer := event makePointer.
	(pointer frame: self pickPlane: selectedPoint normal: (Vector3 zAxis)) ifTrue:[
		"Do it."
		self dispatchResize: pointer selectedPoint forEvent: event.
	
		"Once the resize has been done, update the remembered selected point,
	 	 since our transforms may have changed."
		(pointer frame: self pickPlane: selectedPoint normal: (Vector3 zAxis)) ifTrue:[
			selectedPoint := pointer selectedPoint copy.
		].

	] ifFalse: [
		QLogger logWarning: 'No pick in panel frame pointer move.'
	].

! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 9/23/2008 15:30'!
pointerOver: event

	wasDragged := false.
	isEnabled ifFalse: [
		^self
	].
	QArrowCursor setArrowCursor: event frame: self.
	self bars do: [ :b|
		b unhilite	"Clear any leaked state."
	].
	event selection frame ifNotNil: [
		self hiliteEdgeByCase: (self classifyResize: event).
	].
	! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 3/15/2010 10:46'!
pointerUp: event
	event avatar ifNil: [ ^ self ].
	isEnabled ifTrue: [
		wasDragged ifTrue: [
			self pointerMove: event.
			self endMove: event.
		].
		self bars do: [:b|b unhilite].
	].
	"Clicks are for goto-approach, so they always go through"
	wasDragged ifFalse: [
		self pointerClick: event.
	].
	wasDragged := false.
! !

!QPanelFrame methodsFor: 'private' stamp: 'jcg 2/10/2007 20:02'!
registerForAssemblyEvents
	"Register for events produced by our assembly."
! !

!QPanelFrame methodsFor: 'events' stamp: 'bgf 7/19/2009 14:10'!
removeUser: user selection: selection

	selection userInfo ifNotNil:["remove the frame that is in this slot"
		selection userInfo delete. 
		selection userInfo: nil.
	].! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 4/7/2007 19:32'!
sideBars
	^ { self eastBar . self westBar }! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:10'!
southBar
	^ self subclassResponsibility! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
thickness
	^ thickness! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
thickness: t
	thickness := t.
	self shapeChanged.! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:39'!
title
	"The contents of the title label, or nil"
	^ self label ifNotNil: [self label contents]! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 4/17/2007 15:37'!
title: aString
	"The string in the title label"
	self label ifNotNil: [self label contents: aString]! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 3/25/2010 10:31'!
titleColor: aColor
	self label ifNotNil: [ : label |  label textColor: aColor ]! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 3/15/2010 10:44'!
unpinEdge: compassPt
	"Compass pt is one of #north, #south, #east, #west.
	 Un-pin the corresponding edge bar so that it can be dragged."
	| include |
	include := self edgeBarAt: compassPt.
	include ifNotNil: [
		(movableBars includes: include) ifFalse: [
			movableBars := movableBars copyWith: include.
		].
	].! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:07'!
unpinEdges: collectionOfSymbols

	"Each edge is one of #north, #south, #east, #west.
	 Unfix the corresponding edge bar so that it can be dragged."

	collectionOfSymbols do: [ :each|
		self unpinEdge: each.
	].! !

!QPanelFrame methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:09'!
westBar
	self subclassResponsibility! !

TGroup subclass: #QRepeatingGroup
	instanceVariableNames: 'prototype count'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QRepeatingGroup commentStamp: '<historical>' prior: 0!
A an abstract class that arranges to have as children,
a specified number of instances of its 'prototype' .. which can be any frame.

Initial use case is to create ranges of chairs.

This class neither provides a layout policy for its children,
nor persists their positions.   
Subclasses are expected to provide one or the other
(usually, an updateShape method that positions the children.)

(This base class sets placement-locked on its dynamic children
so the user has no easy access to positioning-tools for them,
since that is not expected to be persistent.)
!

QRepeatingGroup subclass: #QCircleGroup
	instanceVariableNames: 'radius'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QCircleGroup commentStamp: '<historical>' prior: 0!
A cloning manager for placing chairs around a circle;
the children are spaced equally around the circle of given radius,
each facing towards the center.!

!QCircleGroup class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCircleGroup class methodsFor: 'inspector' stamp: 'bgf 2/13/2010 10:59'!
addPropertyRows: propertyDialog forFrame: raref
	"The base version of the QPropertyDialog row-population protocol;
	 add property rows for the user-editable aspects of this frame."
	self addNameRowIfNeeded: propertyDialog forFrame: raref.

	propertyDialog addRow: (
		(QScalarPropertyRow new )
			allowNil: false;
			titleString: 'Radius';
			fetchBlock: [ :frame | (frame get: #radius) ];
			acceptBlock: [ :v :frame | frame future radius: v ] ;
			balloonHelp: 'Distance from the center to the items.';
		yourself).

	super addPropertyRows: propertyDialog forFrame: raref.! !

!QCircleGroup class methodsFor: 'defaults' stamp: 'bgf 2/13/2010 11:00'!
defaultRadius
	^ 8.0! !

!QCircleGroup methodsFor: 'persist' stamp: 'bgf 2/13/2010 10:59'!
constructFrom: pStream
	super constructFrom: pStream.

	"We don't persist children; the QRepeatingGroup takes care of that."
	self radius: (pStream readNumber: 'radius' default: (self class defaultRadius)).
! !

!QCircleGroup methodsFor: 'initialize' stamp: 'bgf 2/13/2010 11:00'!
initialize
	super initialize.
	self radius: (self class defaultRadius).! !

!QCircleGroup methodsFor: 'persist' stamp: 'bgf 2/13/2010 11:00'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeNumber: 'radius' value: radius default: (self class defaultRadius).
! !

!QCircleGroup methodsFor: 'transform' stamp: 'bgf 2/13/2010 10:57'!
privateScale: aFactor
	super privateScale: aFactor.
	aFactor ~= 1 ifTrue: [
		self radius ifNotNil: [
			self radius: (self radius * aFactor).
		]].! !

!QCircleGroup methodsFor: 'accessing' stamp: 'bgf 2/13/2010 10:57'!
radius
	^ radius
! !

!QCircleGroup methodsFor: 'accessing' stamp: 'bgf 2/13/2010 11:00'!
radius: aFloat
	radius := aFloat atLeast: 0.1 atMost: 1000.0.
	self shapeChanged.
! !

!QCircleGroup methodsFor: 'hierarchy' stamp: 'bgf 2/13/2010 11:05'!
updateShape
	"My layout method - just lay out an inward-facing circle. "
	| a kids step |
	kids := (self children ifEmpty: [ ^ self ]).
	step := (360.0 / kids size max: 1).
	a := 0.

	"A little trigonometry. " 
	kids doWithIndex: [ : kid : i |
		| x z t |
		x := a degreeSin * radius.
		z := (a degreeCos * radius).
		t := kid translation.
		kid translation: (Vector3 x: x y: (t y) z: z).
		kid rotationAroundY: a .

		a := (a + step) asSmallPositiveDegrees.
		].

		! !

QRepeatingGroup subclass: #QRectangularGroup
	instanceVariableNames: 'width length'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QRectangularGroup commentStamp: '<historical>' prior: 0!
A cloning layout manager that is meant to place chairs around a rectangular table.
Given width, length, and count of chairs, arrays the chairs best-fit along the outline of the rectangle,
facing inwards (turned toward the middle a bit, but mostly facing the table edge.)
!

!QRectangularGroup class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRectangularGroup class methodsFor: 'inspector' stamp: 'bgf 2/12/2010 16:10'!
addPropertyRows: propertyDialog forFrame: raref
	"The base version of the QPropertyDialog row-population protocol;
	 add property rows for the user-editable aspects of this frame."
	self addNameRowIfNeeded: propertyDialog forFrame: raref.

	propertyDialog addRow: (
		(QScalarPropertyRow new )
			allowNil: false;
			titleString: 'Length';
			fetchBlock: [ :frame | (frame get: #length) ];
			acceptBlock: [ :v :frame | frame future length: v ] ;
			balloonHelp: 'Long axis of the rectangle of items.';
		yourself).

	propertyDialog addRow: (
		(QScalarPropertyRow new )
			titleString: 'Width';
			fetchBlock: [ :frame | (frame get: #width) ] ;
			acceptBlock: [ :v :frame | frame future width: v ] ;
			balloonHelp: 'Short axis of the rectangle of items.';
		yourself).

	super addPropertyRows: propertyDialog forFrame: raref.! !

!QRectangularGroup class methodsFor: 'defaults' stamp: 'bgf 2/12/2010 15:58'!
defaultAngle
	^ 15.0! !

!QRectangularGroup class methodsFor: 'defaults' stamp: 'bgf 2/12/2010 16:01'!
defaultLength
	^ 10! !

!QRectangularGroup class methodsFor: 'defaults' stamp: 'bgf 2/12/2010 15:58'!
defaultRadius
	^ 10.0! !

!QRectangularGroup class methodsFor: 'defaults' stamp: 'bgf 2/12/2010 16:01'!
defaultWidth
	^ 5! !

!QRectangularGroup methodsFor: 'persist' stamp: 'bgf 2/12/2010 16:01'!
constructFrom: pStream
	super constructFrom: pStream.

	"We don't persist children; the QRepeatingGroup takes care of that."

	self width: (pStream readNumber: 'width' default: (self class defaultWidth)).
	self length: (pStream readNumber: 'length' default: (self class defaultLength)).
! !

!QRectangularGroup methodsFor: 'accessing' stamp: 'bgf 2/12/2010 16:14'!
count: nKids
	"Constrain the number of children to be even"
	super count: (nKids roundTo: 2)! !

!QRectangularGroup methodsFor: 'initialize' stamp: 'bgf 2/13/2010 11:02'!
initialize
	super initialize.
	self width: self class defaultWidth.
	self length: self class defaultLength.! !

!QRectangularGroup methodsFor: 'accessing' stamp: 'bgf 2/12/2010 16:00'!
length
	^ length! !

!QRectangularGroup methodsFor: 'accessing' stamp: 'bgf 2/12/2010 16:11'!
length: l 
	length := l max: 1.0.
	self shapeChanged.! !

!QRectangularGroup methodsFor: 'persist' stamp: 'bgf 2/12/2010 16:08'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeNumber: 'width' value: width default: (self class defaultWidth).
	pStream storeNumber: 'length' value: length default: (self class defaultLength).
! !

!QRectangularGroup methodsFor: 'transform' stamp: 'bgf 2/12/2010 16:03'!
privateScale: aFactor
	super privateScale: aFactor.
	aFactor ~= 1 ifTrue: [
		self width ifNotNil: [ self width: (self width * aFactor) ].
		self length ifNotNil: [ self length: (self length * aFactor) ].
		].! !

!QRectangularGroup methodsFor: 'hierarchy' stamp: 'bgf 8/18/2010 10:20'!
updateShape
	"My layout method... spread the children around the sides of the table. "
	| nKids ratio kids smaller larger k xStep zStep x z |
	self children ifEmpty: [ ^ self ].
	kids := self children.
	nKids := (self children size) roundTo: 2.
	ratio := width / ((width+length * 1.1) asFloat).
	smaller :=  (nKids * 0.5 * ratio) rounded.
	larger := (nKids // 2) - smaller.

	"Below, using the x or z as the rotation-around-y is just a dumb pet-trick,
 	 to get a slight camber inwards for chairs near the corners.
	 (Using feet-from-midline as degrees-of-camber-inward is the cheap pet trick...
	 works fine for table-sized objects.)"

	larger > 0 ifTrue: [
		xStep := (length asFloat) / (larger +1 max: 1).
		x  := (larger even ifTrue: [ 0.5 ] ifFalse: [ 0 ]) * xStep.
		1 to: (larger min: (kids size)) do:  [ : i |
			k := kids at: i.
			k rotationAroundY: (x).
			k translationX: x y: 0 z: (width -(x abs * 0.1))*0.5.
			x := x negated.
			x negative ifFalse: [ x := x + xStep ].
		].
		x  := (larger even ifTrue: [ 0.5 ] ifFalse: [ 0.0 ]) * xStep.
		(larger +1) to: (larger+larger min: (kids size)) do: [ : i |
			k := kids at: i.
			k rotationAroundY: -180 - x.
			k translationX:  x y: 0  z: ((width -(x abs * 0.1)) * -0.5).
			x := x negated.
			x negative ifFalse: [ x := x + xStep ].
		]].

	smaller > 0 ifTrue: [
		| zero |
		zero := larger + larger.
		zStep := (width asFloat) /  (smaller +1 max: 1).
		z := (smaller even ifTrue: [ 0.5 ] ifFalse: [ 0]) * zStep.
		zero+1 to: ((zero+ smaller)  min: (kids size))  do: [ : i |
			k := kids at: i.
			k rotationAroundY: 90 - (z).
			k translationX: (length - (z abs * 0.1))*0.5 y:  0 z: z.
			z := z negated.
			z negative ifFalse: [ z := z + zStep ]. 
		].
		z := (smaller even ifTrue: [ 0.5 ] ifFalse: [ 0 ]) * zStep.
		(zero+smaller +1) to: (zero+smaller+smaller min: (kids size)) do: [ : i |	
			k := kids at: i.
			k rotationAroundY: -90 + z.
			k translationX: (length - (z abs * 0.1)) * -0.5  y:  0 z: z.
			z := z negated.
			z negative ifFalse: [ z := z + zStep ]. 
		]].
	
		! !

!QRectangularGroup methodsFor: 'accessing' stamp: 'bgf 2/12/2010 16:00'!
width
	^ width! !

!QRectangularGroup methodsFor: 'accessing' stamp: 'bgf 2/12/2010 16:12'!
width: w 
	width := w max: 1.0.
	self shapeChanged.! !

QRepeatingGroup subclass: #QRepeatingArc
	instanceVariableNames: 'radius angle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QRepeatingArc commentStamp: '<historical>' prior: 0!
A layout-managing group of repeating clones of a prototype child.

Arranges its chlidren's positions to be evenly spaced along an arc of the given radius,
each at the given angle from the next, reckoning from my origin as the arc-center.

The arc is in the X-Z plane, fanning outward such
that the first (innermost) child is placed at x=0, positive z = radius 
(from the group origin).   

Children's rotationAroundY is set to face them inward towards the origin. !

!QRepeatingArc class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRepeatingArc class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:33'!
addPropertyRows: propertyDialog forFrame: raref
	"The base version of the QPropertyDialog row-population protocol;
	 add property rows for the user-editable aspects of this frame."
	self addNameRowIfNeeded: propertyDialog forFrame: raref.

	propertyDialog addRow: (
		(QScalarPropertyRow new )
			allowNil: false;
			titleString: 'Radius';
			fetchBlock: [ :frame | (frame get: #radius) ];
			acceptBlock: [ :v :frame | frame future radius: v ] ;
			balloonHelp: 'Distance from the arc origin to the items.';
		yourself).

	propertyDialog addRow: (
		(QScalarPropertyRow new )
			titleString: 'Angle';
			fetchBlock: [ :frame | (frame get: #angle) ] ;
			acceptBlock: [ :v :frame | frame future angle: v ] ;
			balloonHelp: 'Degrees between centers of items on the arc.';
		yourself).

	super addPropertyRows: propertyDialog forFrame: raref.! !

!QRepeatingArc class methodsFor: 'defaults' stamp: 'bgf 7/21/2009 10:27'!
defaultAngle
	^ 15.0! !

!QRepeatingArc class methodsFor: 'defaults' stamp: 'bgf 7/21/2009 10:27'!
defaultRadius
	^ 10.0! !

!QRepeatingArc methodsFor: 'accessing' stamp: 'bgf 7/21/2009 10:27'!
angle
	^ angle
! !

!QRepeatingArc methodsFor: 'accessing' stamp: 'bgf 7/21/2009 10:57'!
angle: aFloat
	angle := (aFloat atLeast: 0.01 atMost: 180).
	self shapeChanged.
! !

!QRepeatingArc methodsFor: 'seating' stamp: 'bgf 7/26/2009 14:40'!
bestSeatWithin: tbox nearest: avatarPosition

	"Linear search - finds the seat nearest the middle."
	frameChildren ifNotNil: [
		frameChildren do: [ : kid | 
			(kid bestSeatWithin: tbox nearest: avatarPosition) ifNotNilDo: [ : seat | ^ seat  ].
		]].

	^ nil! !

!QRepeatingArc methodsFor: 'persist' stamp: 'bgf 7/21/2009 14:42'!
constructFrom: pStream
	super constructFrom: pStream.

	"We don't persist children; the QRepeatingGroup takes care of that."

	self angle: (pStream readNumber: 'angle' default: (self class defaultAngle)).
	self radius: (pStream readNumber: 'radius' default: (self class defaultRadius)).
! !

!QRepeatingArc methodsFor: 'initialize' stamp: 'bgf 2/13/2010 11:01'!
initialize
	super initialize.
	self radius: (self class defaultRadius).
	self angle: (self class defaultAngle).! !

!QRepeatingArc methodsFor: 'persist' stamp: 'bgf 7/21/2009 14:42'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeNumber: 'angle' value: angle default: (self class defaultAngle).
	pStream storeNumber: 'radius' value: radius default: (self class defaultRadius).
! !

!QRepeatingArc methodsFor: 'transform' stamp: 'bgf 2/11/2010 15:58'!
privateScale: aFactor
	super privateScale: aFactor.
	aFactor ~= 1 ifTrue: [
		self radius ifNotNil: [
			self radius: (self radius * aFactor).
		]].! !

!QRepeatingArc methodsFor: 'accessing' stamp: 'bgf 7/21/2009 10:26'!
radius
	^ radius
! !

!QRepeatingArc methodsFor: 'accessing' stamp: 'bgf 7/21/2009 10:26'!
radius: aFloat
	radius := aFloat atLeast: 0.1 atMost: 2000.0.
	self shapeChanged.
! !

!QRepeatingArc methodsFor: 'hierarchy' stamp: 'bgf 7/26/2009 14:42'!
updateShape
	"My layout method - starting from the middle (i.e. at x 0,  facing z negative),
	 lay out the children in an arc symetrical fanning out from the z axis. "
	| even a |
	self children ifEmpty: [ ^ self ].
	a := 0.0.
	even := self children size even.
	even ifTrue: [
		a := angle * 0.5.
	].

	"A little trigonometry.   Note that the first seats land innermost in the arc;
	 this matters when using me to hold a bank of QSeats for a theatre audience." 
	self children doWithIndex: [ : kid : i |
		| x z t |
		x := a degreeSin * radius.
		z := (a degreeCos * radius).
		t := kid translation.
		kid translation: (Vector3 x: x y: (t y) z: z).
		kid rotationAroundY: a .
		(even = i even) ifTrue: [ 
			a := a negated + angle.
		] ifFalse: [
			a := a negated.
		].
		a := a asSmallAngleDegrees ].

		! !

!QRepeatingGroup class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRepeatingGroup class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:33'!
addPropertyRows: propertyDialog forFrame: rgref
	"The base version of the QPropertyDialog row-population protocol;
	 add property rows for the user-editable aspects of this frame."
	self addNameRowIfNeeded: propertyDialog forFrame: rgref.

	propertyDialog addRow: (
		(QScalarPropertyRow new )
			allowNil: false;
			titleString: 'Count';
			fetchBlock: [ :frame | (frame get: #count) ];
			acceptBlock: [ :v : frame | frame future count: (v asInteger)] ;
			balloonHelp: 'Number of times to repeat the prototype.';
		yourself).

	propertyDialog addRow: (
		(QTFramePropertyRow new )
			titleString: 'Prototype';
			fetchBlock: [ :frame| (frame get: #prototype) ] ;
			acceptBlock: [ :v :frame | frame future setPrototype: v ] ;
			balloonHelp: 'The frame object to use as the prototype.';
		yourself).

	super addPropertyRows: propertyDialog forFrame: rgref.! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 1/15/2010 16:54'!
addChild: child atIndex: index
	super addChild: child atIndex: index.
	child isPlacementLocked: true.
	child isFromTemplate: (self isFromTemplate).! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 2/13/2010 11:30'!
bestSeatWithin: tbox nearest: avatarPosition
	"Repeating groups are generally used to supply seats."
	^ self genericBestSeatWithin: tbox nearest: avatarPosition
	! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 2/14/2010 12:55'!
childrenForBrowser
	^ #()! !

!QRepeatingGroup methodsFor: 'persist' stamp: 'bgf 7/21/2009 14:43'!
constructFrom: pStream
	super constructFrom: pStream.
	self setPrototype: (pStream readObject: 'prototype' default: prototype).
	self count: (pStream readNumber: 'count' default: (self count)).
! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 7/21/2009 14:09'!
count
	"The number of instances of the prototype to be managed as children."
	^ count
! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 7/21/2009 14:30'!
count: nKids
	count := nKids asInteger max: 0.
	prototype ifNil: [
		self destroyAll.
		^ self 
	].	
	[count > self children size] whileTrue: [
		self addChild: (prototype duplicate).
	].
	[count < self children size]  whileTrue: [
		self removeChild: (self children last).
	].


! !

!QRepeatingGroup methodsFor: 'initialize' stamp: 'bgf 7/21/2009 14:54'!
defaultIsAtomic 
	^ false! !

!QRepeatingGroup methodsFor: 'initialize' stamp: 'bgf 7/21/2009 14:22'!
destroyAll
	| kids |
	kids := self children.
	kids isEmptyOrNil ifFalse: [
		kids clone do: [ : k | k destroy ].
	].! !

!QRepeatingGroup methodsFor: 'initialize' stamp: 'bgf 7/21/2009 14:50'!
initialize
	super initialize.
	count := 0.

	"My subclasses are expected to provide an updateShape implementation with a layout policy,
 	 as there is no persistence of child positions provided in the base."

	self startScript: #shapeChanged when: { self . #structureChanged }.! !

!QRepeatingGroup methodsFor: 'testing' stamp: 'bgf 7/21/2009 14:55'!
isEditableGroup
	"No reparenting in or out; use the prototype and count instead."
	^ false! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 1/18/2010 12:11'!
isFromTemplate: bFrom
	"Since my children are managed automatically, they inherit this."
	super isFromTemplate: bFrom.
	self children do: [ : c | c isFromTemplate: bFrom ].! !

!QRepeatingGroup methodsFor: 'persist' stamp: 'bgf 7/21/2009 14:40'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeNumber: 'count' value: count default: 0.
	pStream storeObject: 'prototype' value: prototype default: nil.! !

!QRepeatingGroup methodsFor: 'initialize' stamp: 'bgf 2/11/2010 17:15'!
privateScale: aFactor
	super privateScale: aFactor.
	"And keep the prototype consistent with the instances."
	prototype ifNotNil: [ prototype scale: aFactor ].! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 7/21/2009 14:09'!
prototype
	^ prototype! !

!QRepeatingGroup methodsFor: 'accessing' stamp: 'bgf 7/21/2009 17:01'!
setPrototype: aTFrame
	"Set the prototype... always an unparented copy of the incoming frame."
	| parent i |
	prototype = aTFrame ifTrue: [ ^ self ].
	self destroyAll.
	aTFrame ifNil: [
		prototype ifNotNil: [ 
			prototype destroy.
			prototype := nil.
		].
	] ifNotNil:[
		[
			aTFrame parent ifNotNilDo: [ : p |
				parent := p.
				i := parent indexOfChild: aTFrame.
				parent removeChild: aTFrame.
				].  	
			[ prototype := aTFrame duplicate.
			  prototype localTransform: Matrix4x4 identity.
			] on: Error do: [ :ex|
				QLogger logWarning: 'Cannot use ', aTFrame, ' as prototype: ', (ex description).
				prototype := nil.
			].
		] ensure: [
			parent ifNotNil: [ parent addChild: aTFrame atIndex: i ]
		].
	].
	self count: count.! !

TGroup subclass: #QSequenceGroup
	instanceVariableNames: 'delta playing index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QSequenceGroup class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSequenceGroup methodsFor: 'persist' stamp: 'das 1/1/2008 12:07'!
constructFrom: pStream
	"Construct the receiver"
	super constructFrom: pStream.
	self delta: (pStream readNumber: 'delta' default: self delta).
	self children: (pStream readSequence:'children' default: self children).
! !

!QSequenceGroup methodsFor: 'accessing' stamp: 'das 12/27/2007 14:19'!
delta
	
	^delta.! !

!QSequenceGroup methodsFor: 'accessing' stamp: 'das 12/27/2007 14:19'!
delta: dt
	
	delta := dt.! !

!QSequenceGroup methodsFor: 'initialize' stamp: 'das 12/27/2007 14:07'!
initialize

	super initialize.
	playing := false.
	index := 1.
	delta := 100. "start with a 100 msec delay between 'frames'"! !

!QSequenceGroup methodsFor: 'persist' stamp: 'das 1/1/2008 12:07'!
persistOn: pStream

	super persistOn: pStream.
	pStream storeNumber: 'delta' value: self delta default: 100.
	pStream storeSequence: 'children' value: self children default: #().
	! !

!QSequenceGroup methodsFor: 'initialize' stamp: 'bgf 3/30/2009 21:42'!
renderChildren: ogl

	((frameChildren notNil and:[frameChildren size>0]) and:[ index<=frameChildren size ])ifTrue:[
		^(frameChildren at: index) renderFrame: ogl.
	].
	^ 0.! !

!QSequenceGroup methodsFor: 'stepping' stamp: 'das 12/27/2007 14:06'!
start

	playing := true.
	(self future:100) step.! !

!QSequenceGroup methodsFor: 'stepping' stamp: 'das 12/27/2007 14:09'!
step

	playing ifTrue:[
		(frameChildren notNil and:[frameChildren size>0]) ifTrue:[
			index := index + 1.
			index > frameChildren size ifTrue:[index := 1].
		].
		(self future: delta) step.
	].
! !

!QSequenceGroup methodsFor: 'stepping' stamp: 'das 12/27/2007 14:06'!
stop

	playing := false.
! !

TGroup subclass: #QSimpleTable
	instanceVariableNames: 'legs extent cover thickness label mgrip rgrip contents bars top shadow'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QSimpleTable commentStamp: '<historical>' prior: 0!
A table geometry made of primitive cubes.
The edges of the table are rotation grips.
The legs are motion grips.
Shift-click on the table-top gives it a mirror-finish. 
An optional QLabel carries a single-line label
that appears hanging under the front lip.!

!QSimpleTable class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSimpleTable class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:35'!
addPropertyRows: propertyDialog forFrame: stref

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Label';
			fetchBlock: [ :frame | (frame get: #label) ifNil: [''] ];
			acceptBlock: [ :v :frame | frame future label: ((v = '') ifTrue: [nil] ifFalse: [v]) ];
			balloonHelp: 'Label on the front of the table.';
		yourself).

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Extent';
			fetchBlock: [ :frame | frame get:#extent ];
			acceptBlock: [ :v3 :frame | frame future extent: v3 ];
			balloonHelp: 'Dimensions of the table.';
	 	yourself	).

	super addPropertyRows: propertyDialog forFrame: stref.! !

!QSimpleTable class methodsFor: 'defaults' stamp: 'bgf 4/13/2008 09:14'!
defaultThickness
	"Basis for leg and trim geometry"
	^ 0.2! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
addContents: aTFrame

	"Put something on the table.  Doesn't handle positioning, just ensures that
	 the frame moves when the table moves, and is  saved with it persistently."

	(contents includes: aTFrame) ifFalse: [
		contents add: aTFrame.
	].
	self addChild: aTFrame.! !

!QSimpleTable methodsFor: 'private' stamp: 'bgf 4/13/2008 09:14'!
barHorizontal
	"The thickness of the edge bar horizontally."
	^ thickness * 0.67! !

!QSimpleTable methodsFor: 'private' stamp: 'bgf 4/13/2008 09:14'!
barVertical
	"The thickness of the edge bar vertically."
	^ thickness! !

!QSimpleTable methodsFor: 'persist' stamp: 'bgf 4/13/2008 09:14'!
constructFrom: pStream
	| kids |
	super constructFrom: pStream.

	thickness := (pStream readNumber: 'thickness' default: thickness).
	self extent: (pStream readVector3: 'extent' default: (self extent)).
	self isReflecting: (pStream readBool: 'isReflecting' default: (self isReflecting)).
	self label:(pStream readString: 'label' default:nil).

	kids := (pStream readSequence: 'contents' default: #()).
	kids  do: [ :kid | self addContents: kid ].! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
contents
	"The user-supplied TFrames being carried by this table."
	^ contents! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:48'!
dragAdapter
	^ mgrip! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
extent

	"The cube occupied by the table.
	 Internally, the table's elements are arranged to grow upward from y=0,
	 so the top surface is the plane through extent y."

	^ extent! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
extent: anExtent

	"Change the extent of the table;
       the value is a cube corresponding to the volume of the table."

	extent ~= anExtent ifTrue: [
		extent := anExtent.
		self shapeChanged.
	].! !

!QSimpleTable methodsFor: 'initialize' stamp: 'bgf 6/2/2009 19:30'!
forwardKeystroke:  event

	"Indirect forwarding of keystrokes (see #initialize)
	 so that a container can observe my #keyStroke event.
	 It is important that keyStroke event handling NOT be enabled in my event mask. 

	 This is so that <Escape> can be delivered to the leg and edge motion-rotation adapters."
	self signal: #keyStroke with: event.! !

!QSimpleTable methodsFor: 'initialize' stamp: 'bgf 6/28/2010 11:11'!
initialize
	| leg |
	super initialize.
	self objectName: 'Simple Table'.
	contents := OrderedCollection new.
	extent := 2@2@2.
	thickness := 0.2.
	self isReflecting: false.
	cover := TSimpleRectangle new
			material: QMaterials mediumWoodMaterial;
			rotationAroundX:-90;
			yourself.
	cover material 
			opacity: 0.5;   
			cullFace: false.
	cover checkAlpha.
	self addChild:cover.

	shadow := TSimpleRectangle new.
	shadow addRotationAroundX:-90.
	shadow material texture: (QTextures find: 'TableShadow.png').
	shadow material color: (VectorColor r:1 g: 1 b: 1 a:0.9).
	shadow checkAlpha.
	shadow pickable: false.
	self addChild: shadow.

	bars := TGroup new.
	1 to: 4 do: [ :i|
		bars addChild:(
			(TCube new)
				material: (QMaterials mediumDarkWoodMaterial);
				yourself).
	].
	self addChild:bars.

	rgrip := QPlanePivot forLuggage: self grip: bars.
	legs := TGroup new.
	self addChild: legs.
	1 to: 4 do: [ :i|
		leg :=(TCylinder new) material: (QMaterials darkWoodMaterial); 
			yourself.
		legs addChild: leg.	].
	(mgrip := QMotionAdapter forLuggage: self grip: legs)
		 isGroundBound: true;
		contactOffset: 0.05.

	self runScript: #forwardKeystroke: when: { legs . #keyStroke }.

	self shapeChanged.! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
isGroundBound
	^ true! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 6/1/2009 23:27'!
isPlacementLocked: bLock
	bLock = (self isPlacementLocked) ifFalse: [
		super isPlacementLocked: bLock.
		mgrip isEnabled: bLock not.
		rgrip isEnabled: bLock not.
	].! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
isReflecting
	^ top isKindOf:TPortal! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 8/27/2009 17:02'!
isReflecting: bool

	"Set the top, reflective or plain."

	(top notNil and: [self isReflecting = bool])
		ifTrue: [ ^ self ].

	top ifNotNil: [ top destroy ].
	bool ifFalse: [
		top := (TSimpleRectangle new)
			material: QMaterials mediumWoodMaterial;
			rotationAroundX:-90;
			cullFace:false;
			yourself.
		self addChild: top.
	] ifTrue: [
		top := TPortal new.
		top rotationAroundX: -89.95.
		self addChild: top.
	].

	self shapeChanged.! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
label
	"The string appearing as a label for this table."
	label ifNotNil: [
		^  label contents asString.
	].
	^nil! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:17'!
label: aString
	"Hang a label under the front face of the table.
	 Note a bit of historical confusion... the label instance variable is a QLabel,
	 but my label/label: accessors refer to its content string."

	label ifNotNil: [
		self removeChild: label.
	].

	(aString notNil and: [ aString size > 0 ]) ifTrue: [
		label := (QLabel new) 
			extent: (extent x /1.5) @ (self barVertical * 1.5);
			contents: aString;
			textColor: QColors signTextForeground;
			backgroundColor: QColors signTextBackground;
			yourself.
		self addChild: label.
	] ifFalse: [
		label := nil.
	].
	self shapeChanged.
! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:37'!
labelTag
	"The widget holding my label string"
	^ label! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 2/12/2010 14:24'!
localCenter
	"Cheat - substitute the plain avatar height above ground;
	 this is a workaround so that jumps leave the avatar on the ground."
	^ self translation + (0@(5*(self scaleMemo))@0)! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
luggage: anAncestor

	"Make the motion handling of the table's elements work 
	 to manipulate the given ancestor frame;
	 this can be used when the table is the foundation of a larger assembly."

	mgrip luggage: anAncestor.
	rgrip luggage: anAncestor.! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
menuStyle 
	^ QStyles menuStyle! !

!QSimpleTable methodsFor: 'persist' stamp: 'bgf 4/13/2008 09:14'!
persistOn: pStream

	super persistOn: pStream.
	pStream storeNumber: 'thickness' value: (self thickness) default: (self class defaultThickness).
	pStream storeVector3: 'extent' value: (self extent).
	pStream storeString: 'label' value: self label default: nil.
	pStream storeSequence: 'contents' value: self contents default:#().
	pStream storeBool: 'isReflecting' value: self isReflecting default: false.! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:58'!
privateScale: aFactor

	thickness := thickness * aFactor.
	label ifNotNil: [
		label extent: (label extent * aFactor).
	].
	self extent: (self extent * aFactor).
	super privateScale: aFactor.
! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 13:27'!
removeChild: aTFrame
	
	"If this was in my contents, it should be forgotten now..."
	contents remove: aTFrame ifAbsent: [].
	super removeChild: aTFrame.! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 13:21'!
removeContents: aTFrame
	"Does the work both to forget contents and drop the child:"
	self removeChild: aTFrame.! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
thickness
	"Base dimension for trim and accessories (the unit is the vertical height of the edge trim"
	^ thickness! !

!QSimpleTable methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:14'!
thickness: aScalar
	"Base dimension for trim and accessories (the unit is the vertical height of the edge trim"
	thickness := aScalar.
	self updateShape.! !

!QSimpleTable methodsFor: 'initialize' stamp: 'bgf 4/13/2008 09:39'!
updateShape

	|  inset |

	"The origin is at the feet,  so everything translates up from the floor."
	1 to: 2 do: [ :i|
		(bars frameChildren at: i) 
			extent: (extent x +(self barHorizontal) @ thickness@(self barHorizontal));
			translation: (0@(extent y-(thickness*0.4))@(extent z*0.5 * ((i-1.5) sign)))
	].
	3 to: 4 do: [ :i|
		(bars frameChildren at: i) 
			extent: ((self barHorizontal) @ thickness@(extent z+(self barHorizontal)));
			translation: (extent x * 0.5 * ((i-3.5)sign) @(extent y-(thickness*0.4))@0)
	].

	top extent:(extent x@extent z) - 0.05.
	top translation:(0@(extent y)@0).
	cover extent:(extent x@extent z).
	cover translation: (0@(extent y+0.01)@0).

	inset := thickness * 1.5 min: ((extent x/8) min: (extent z/8)).

	legs frameChildren doWithIndex: [ :leg :i |
		leg height: extent y - thickness.
		leg radius: (thickness/2).
		leg translation:
			(i caseOf: {
				[1] -> [ (extent x/2-inset)@(thickness/2)@(extent z/2-inset) ].
				[2] -> [ (extent x/-2+inset)@(thickness/2)@(extent z/2-inset) ].
				[3] -> [ (extent x/2-inset)@(thickness/2)@(extent z/-2+inset) ].
				[4] -> [ (extent x/-2+inset)@(thickness/2)@(extent z/-2+inset) ].
			}).
	].

	shadow extent: (top extent)*(1.08@1.05);
			translationX: 0 y: ((thickness/2.0)+0.01) z:0.
	label ifNotNil: [
		"The label hangs just under the front edge trim"
		label translationX: 0 
			  y: (extent y - (1.55*self barVertical)) 
			  z: (extent z/2) + (self barHorizontal/2.0) - 0.01
	].! !

TGroup subclass: #QSlaveProjectionObject
	instanceVariableNames: 'displayObject master displayMaterial scaleUV'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QSlaveProjectionObject class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSlaveProjectionObject class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QSlaveProjectionObject class methodsFor: 'defaults' stamp: 'das 6/29/2009 14:24'!
resourceType
	^ #displaySlave! !

!QSlaveProjectionObject class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:03'!
slave: slaveRef displayObject: objRef oldState: stateVec
	"Undo for dropping a display object onto the screen object in the scene inspector."

	^ QBlockAction
		undoBlock: [ : avatar | slaveRef future undoDrop: objRef savedState: stateVec forUser: avatar ]
		redoBlock: [ : avatar | slaveRef future doDrop: objRef savedState: stateVec forUser: avatar ]
		description: 'Drop Display'
		requiredFrames: { slaveRef . objRef }.

		! !

!QSlaveProjectionObject methodsFor: 'initialize' stamp: 'bgf 3/14/2010 09:20'!
acceptDroppedFrame: aTFrame forUser: anAvatar

	"A point for subclass interception...
	 scene-editing tools use this 
	 to reparent an existing tframe into me.
	 The default implementation preserves my global stance."

	"run the gauntlet of things you can't transfer"
	| cap oldXForm oldScale oldKidXForm oldKidScale oldParent vec |
	aTFrame isFrameThumb ifTrue:[^ self].
	(aTFrame class includesBehavior: QEditBox) ifTrue:[^self].
	(aTFrame class includesBehavior: QSoundFrame) ifTrue:[^self]. 
	(aTFrame class includesBehavior: QDisplayPanel) ifTrue:[^self]. 
	(aTFrame class includesBehavior: QSlaveProjectionScreen) ifTrue:[^self]. 
	(aTFrame class includesBehavior: QSlaveProjectionObject) ifTrue:[^self]. 

	"Save state for undo"
	displayObject ifNotNil: [
		cap := anAvatar captureFrames: { displayObject }
	].
	oldXForm := self localTransform copy.
	oldScale := self nominalScale.
	oldKidXForm := aTFrame localTransform copy.
	oldKidScale := aTFrame nominalScale.
	oldParent := aTFrame parent.
	vec :=  { cap . oldXForm . oldScale . oldKidXForm . oldKidScale . oldParent }.

	self doDrop: aTFrame savedState: vec forUser: anAvatar.

	anAvatar recordAction: #slave:displayObject:oldState:
			 class: QSlaveProjectionObject
			 arguments: {  self .  aTFrame .  vec }! !

!QSlaveProjectionObject methodsFor: 'transform' stamp: 'das 7/29/2009 15:12'!
addChild: child atIndex: indexOrNil

	"I only hold one object at a time"
	self removeAll.
	displayObject := child.
	super addChild: child atIndex: nil.
	self future sheetContentChanged.
! !

!QSlaveProjectionObject methodsFor: 'transform' stamp: 'das 7/1/2009 19:29'!
applyMaterial

	| muv |
	displayMaterial texture: master contents material texture.
	muv := master contents material uvScale.
	displayMaterial uvScale: muv*scaleUV.
	displayMaterial uvOffset:	muv-(muv*scaleUV)/2.0.		
! !

!QSlaveProjectionObject methodsFor: 'transform' stamp: 'das 7/1/2009 19:19'!
applyScale

	| muv |
	muv := master contents material uvScale.
	displayMaterial uvScale: muv*scaleUV.
	displayMaterial uvOffset:	muv-(muv*scaleUV)/2.0.		
! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/30/2009 15:04'!
canEditMaterial

	^ true.! !

!QSlaveProjectionObject methodsFor: 'transform' stamp: 'das 7/28/2009 22:55'!
checkContents

	frameChildren ifNil:[
		self createDefaultObject.
		self sheetContentChanged.
	].! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 7/29/2009 16:43'!
checkRoot

	"a future send from root:"
	frameRoot ifNil: [ self master: nil.].
	master ifNotNil: [ 
		master root ifNil: [ self master: nil ].
	].! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'bgf 3/14/2010 09:06'!
clear

	"Erase the display-sharing (but leave the master-relationship active)."
	displayMaterial ifNotNil: [
		displayMaterial texture:nil.
	].
	self stopScript: #onAppExtentChanged:.
! !

!QSlaveProjectionObject methodsFor: 'persist' stamp: 'bgf 3/14/2010 09:30'!
constructFrom: pStream

	| panel |
	super constructFrom: pStream.
	"this is not a TGroup, so we need to explicitly load the children"
	self children: (pStream readSequence:'children' default: self children).
	self scaleUV: (pStream readPoint: 'scaleUV' default: 1@1).
	panel := (pStream readObject: 'sourcePanel' default: nil).
	displayMaterial := (pStream readObject: 'displayMaterial' default: displayMaterial ).
	panel ifNotNil: [
		"Future, so that rooting and construction has had a chance to finish."
		self future master: (panel sheet).
	].
! !

!QSlaveProjectionObject methodsFor: 'initialize' stamp: 'bgf 3/14/2010 09:26'!
createDefaultObject

	displayMaterial ifNil:[
		displayMaterial := QMaterials fullWhiteAlphaMaterial.
	].

	displayObject ifNil:[
		displayObject := TCube new.
		displayObject extent: 2@2@2.
		displayObject material: displayMaterial.
		self addChild: displayObject.
		scaleUV := 1@1.
	].! !

!QSlaveProjectionObject methodsFor: 'transform' stamp: 'das 7/28/2009 22:34'!
destinationStanceFor: anAvatar

	"Standard 2-D view gathering."
	^ QGotoButton stanceFor: anAvatar toView: self inDestination: self.! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/29/2009 14:24'!
destroy

	self master: nil.
	super destroy.
! !

!QSlaveProjectionObject methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 09:33'!
doDrop: objRef savedState: stateVec forUser: anAvatar

	objRef transferTo: self! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'bgf 3/12/2010 08:11'!
fillContextMenu:aMenu forEvent: event

	master ifNotNil: [
		(event avatar canModifyForum) ifTrue: [
			(aMenu add: 'Stop Projecting' target: self action: #stopProjecting:) 
				separator: true.
			].
		event avatar allowedToMove ifTrue: [
			aMenu add: 'Locate Source' target: self action: #locateSource: 
		].
	] ifNil: [
		(aMenu add: 'Not Projecting' target: self action: #noSuchAction:) 
			disabled: true
	].
	event avatar canModifyForum ifTrue: [
		aMenu add: 'Release Projection Object' target: self action: #releaseObject:.
		aMenu addSeparator.
		aMenu add: 'Flip Image Horizontal' target: self action: #flipImageHorizontal:.
		aMenu add: 'Flip Image Vertical' target: self action: #flipImageVertical:.
		aMenu add: 'Flip Visible Side' target: self action: #flipVisibleSide:.
	].! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'das 7/1/2009 19:29'!
flipImageHorizontal: event

	scaleUV := scaleUV x negated @ scaleUV y.
	self applyMaterial.		
! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'das 7/1/2009 19:30'!
flipImageVertical: event

	scaleUV := scaleUV x @ scaleUV y negated.
	self applyMaterial.		
! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'das 7/1/2009 18:26'!
flipVisibleSide: event

	displayMaterial flipFace: displayMaterial flipFace not.! !

!QSlaveProjectionObject methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:36'!
initialize
	super initialize.	
	self createDefaultObject.! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 7/1/2009 15:27'!
isAtomic

	^ false.! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/29/2009 17:04'!
isEditableGroup

	^ true.! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'das 6/29/2009 14:24'!
locateSource: anEvent

	"Menu action - take the avatar to the master display panel that is driving me."

	anEvent avatar ifNotNil:[
		master ifNotNil: [
			anEvent avatar takeMeToFrame: master.
		] ifNil: [
			anEvent avatar signal: #infoMessage with: 'The projector is not in use.'.
		].
	].! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/29/2009 14:24'!
master
	"QDisplaySheet - the guy whose content I'm mirroring right now."
	^ master! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/29/2009 17:11'!
master: aDisplaySheet

	"Become slaved to the given display sheet's content."
	master := aDisplaySheet.
	self stopScript: #sheetContentChanged.

	master ifNotNil: [
		self startScript: #sheetContentChanged when: { master . #contentChanged }.
		self sheetContentChanged.
	] ifNil: [
		self clear.
	].
! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/30/2009 15:04'!
material

	^ displayMaterial.
! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/30/2009 15:05'!
material: mat

	| txtr |
	txtr := displayMaterial texture.
	displayMaterial := mat.
	displayMaterial texture: txtr.
! !

!QSlaveProjectionObject methodsFor: 'events' stamp: 'das 7/1/2009 19:29'!
onAppExtentChanged: appExtent

	"On the fly change in my master's extent, triggered from app resizing;
	 we just follow suit."

	(displayObject isNil or: [ master isNil ] or: [ master contents isNil]) ifTrue: [
		^ self
	].
	self applyMaterial.
! !

!QSlaveProjectionObject methodsFor: 'persist' stamp: 'bgf 3/14/2010 09:07'!
persistOn: pStream

	| txtr |
	super persistOn: pStream.
	"this is not a TGroup, so we need to explicitly save the children"
	pStream storeSequence: 'children' value: self children default: #().
	"this is the material, with the texture removed"
	pStream storePoint: 'scaleUV' value: scaleUV default: 1@1.
	displayMaterial ifNotNil: [
		txtr := displayMaterial texture.
		displayMaterial texture: nil.
		pStream storeObject: 'displayMaterial' value: displayMaterial default: nil.
		displayMaterial texture: txtr.
	].

	"We can only persist the relationship if we're hooked into a valid DisplayPanel;
	 that's what constructFrom: expects."

	(master notNil 
	 and: [master parent isKindOf: QDisplayPanel]
	 and: [master root = self root] ) 
	 ifTrue: [
		pStream storeObject: 'sourcePanel' value: (master parent) default: nil
	].
! !

!QSlaveProjectionObject methodsFor: 'events' stamp: 'bgf 10/25/2009 15:25'!
pointerEnter: anEvent
	"A simple mouseover, to remind  everyone that this is a projection,
	 whose contents cannot be interacted with directly.
	 We could bridge event-interaction eventually 
	 but the focus-details,  geometry-conversions, etc will be painful."
	self showFloatFor: anEvent.! !

!QSlaveProjectionObject methodsFor: 'events' stamp: 'bgf 10/25/2009 15:25'!
pointerOver: event
	"Show the floating details if over the off-screen background
	 or if the projector is off, but don't keep showing them atop an active projection,
	 so you can point without the floating-message distraction." 
	(master isNil or: [ event selection frame = self "I.e. not the contentRect" ]) ifTrue: [
		self showFloatFor: event.
	].

! !

!QSlaveProjectionObject methodsFor: 'initialize' stamp: 'das 6/29/2009 14:24'!
registerResource
	"Register this slave-display as a resource.
	 We keep these as resources mostly for quick access at the space level."
	self island registerResource: self type: self resourceType.
! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'das 7/29/2009 16:47'!
releaseObject: event

	displayObject ifNotNil:[
		displayObject transferTo: self root. 
		displayObject material: TMaterial new.
		displayObject := nil.
	].
	self createDefaultObject.
! !

!QSlaveProjectionObject methodsFor: 'transform' stamp: 'das 7/29/2009 14:19'!
removeChild: child

	super removeChild: child.
	displayObject := nil.
	self future checkContents.

	! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/29/2009 14:24'!
resourceType
	^ self class resourceType! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 6/29/2009 14:24'!
root: newRoot

	frameRoot 
		ifNotNil: [ self unregisterResource ].

	super root: newRoot.

	frameRoot 
		ifNotNil: [ self registerResource ].

	self future checkRoot. "check if we have a nil root. If so, remove master"
! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 7/1/2009 18:40'!
scaleUV: suv
	
	scaleUV := suv! !

!QSlaveProjectionObject methodsFor: 'accessing' stamp: 'das 7/1/2009 19:30'!
sheetContentChanged

	"Any change in the content of my master sheet...
	 update the slave display by attaching to the material all over again.
	 This syncs us with the texture, the aspect details, etc."

	| src |
	displayObject ifNil:[^self].
	self stopScript: #onAppExtentChanged:.
	master ifNotNil: [
		(master root = self root) ifFalse: [
			"Drop the hot potato, somebody's probably been destroyed."
			^ self master: nil.
		].
		src := master contents.
		src ifNotNil: [
			self startScript: #onAppExtentChanged: when: { src. #appExtentChanged }.
			self applyMaterial.
			displayObject material: displayMaterial.
		] ifNil: [
			self clear.
		].
	] ifNil: [
		self clear.
	].
! !

!QSlaveProjectionObject methodsFor: 'events' stamp: 'das 6/29/2009 14:24'!
showFloatFor: event

	"A simple mouseover message, 
	 mostly as a reminder that this is a projector and not an interactive panel."	
	event avatar ifNil: [ ^ self ].
	master ifNil: [
		event avatar showFloatMessage: (self objectName), '\Display Panels can be projected here.'
					 forEvent: event.
	] ifNotNil: [
		event avatar showFloatMessage: 'Projecting: ', master parent title
				       forEvent: event.
	].! !

!QSlaveProjectionObject methodsFor: 'menu' stamp: 'bgf 3/13/2010 08:53'!
stopProjecting: event
	"Menu action - turn me off.  Needs to be made undoable."
	| oldmaster |
	oldmaster := self master ifNotNil: [ : m | m parent ].
	event avatar
		recordAction: #projection:newSlaves:oldSlaves:
		class: QDisplayPanel	
		arguments: { self .   
					{{ nil . self } } . 
					{{ oldmaster . self }}  }.
	self master: nil.! !

!QSlaveProjectionObject methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 09:50'!
undoDrop: objRef savedState: stateVec forUser: anAvatar

	| capture oldXForm oldScale  oldKidXForm oldKidScale oldParent |

	capture := stateVec first.
	oldXForm := stateVec second.
	oldScale := stateVec third.
	oldKidXForm := stateVec fourth.
	oldKidScale := stateVec fifth.
	oldParent := stateVec sixth.

	oldParent ifNotNil: [
		| kid |
		kid := displayObject.
		kid transferTo: oldParent;
			nominalScale: oldKidScale;  
			localTransform: oldKidXForm.
		].

	self nominalScale: oldScale;  localTransform: oldXForm.

	capture ifNotNil: [
		(capture restoreForAvatar: anAvatar) ifNotNil: [ : array |
			| do |
			do := array first.
			do parent = self ifFalse: [
				self addChild: array first
			]]].
! !

!QSlaveProjectionObject methodsFor: 'initialize' stamp: 'das 6/29/2009 14:24'!
unregisterResource
	"Unregister this slave display as a resource"
	self island unregisterResource: self type: self resourceType.! !

TGroup subclass: #QSlider
	instanceVariableNames: 'rule slide hair value length dragObject selectedPoint target action label'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QSlider commentStamp: 'das 12/22/2006 17:52' prior: 0!
QSlider

I am used to set a value between 0 and 1. I am made up of two main parts, a rule and a slide. Click anywhere on the rule, and the slide will jump to that location. Click on the slide and you can drag it where you like. You can't drag it off of the ends of the rule. The dragging sets the "value" variable.

rule - the base object that the slide moves up and down on. 
slide - the object moving across the rule to show the current value.
hair - a hair on the slide showing the precise value of the slider.
value - the current value of the slider - must be between 0 and 1.
dragObject - the current drag object, usually the slider.
length - the current length of the slider.
selectedPoint - where on the slider we just clicked.
target - who do we inform that we are changing the value.
action - how do we inform the target that we are changing the value.!

!QSlider class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSlider class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QSlider methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:37'!
initialize
	| tframe |
	super initialize.

	rule := TGroup new.
	self addChild: rule.

	11 timesRepeat: [
		tframe := (TCube new).
		tframe material: (QMaterials flatWhiteMaterial).
		tframe material color: (VectorColor r:0.72 g:0.71  b:0.7).
		tframe checkAlpha.
		rule addChild: tframe.
	].
	rule frameChildren first material color: (VectorColor r:0.33 g:0.32 b:0.3).
	rule frameChildren first checkAlpha.

	slide := TCylinder new.
	slide material: (QMaterials translucentEditorMaterial).
	slide objectName: #slide.
	hair := TCylinder new.
	hair material: (QMaterials blackMaterial).
	hair checkAlpha.
	hair objectName: #hair.
	rule addChild: hair.  "Don't want the hair hilit when the slide is, so manage it separately."
	rule addChild: slide.

	dragObject := nil.
	value := 0.0.
	label := 'QSlider'.! !

!QSlider methodsFor: 'accessing' stamp: 'bgf 1/23/2007 13:24'!
label
	^ label ifNil: ['Slider'].! !

!QSlider methodsFor: 'accessing' stamp: 'das 12/23/2006 15:35'!
label: lbl

	label := lbl.! !

!QSlider methodsFor: 'accessing' stamp: 'bgf 1/23/2007 16:10'!
length: len

	| ln |
	length := len.
	ln := len/10.0.

	"Black inner pipe."
	(rule frameChildren first)
		extent: (ln*0.9)@(length*1.02)@(ln*0.9);
		translationX:0 y: (len/2) z:0.

	"White collars, creating a series of etched grooves."
	2 to: 11 do:[:index | 
		(rule frameChildren at: index)
			extent: (ln*0.91)@(ln*0.85)@(ln*0.91);
			translation:0@(ln*(index-2) + (ln*0.525))@0.
	].

	rule translation: 0@(len/-2)@0.
	slide height: ln.
	slide radius: ln * 0.9.
	hair radius: ln.
	hair height: ln/8.0.
	self value: value.! !

!QSlider methodsFor: 'accessing' stamp: 'bgf 1/23/2007 15:40'!
paintAlphaGradient: fullColor

	"Paint the rule to indicate alpha variation on the given color."

	| box |

	(rule frameChildren at:1) material transparency:0.5.
	1 to: 10 do: [ :i|
		box := (rule frameChildren at: (i+1)).
		box	material color: (fullColor copy alpha: ((i/10.0)+0.05)).
		box checkAlpha.
	].! !

!QSlider methodsFor: 'accessing' stamp: 'bgf 1/23/2007 16:09'!
paintValueGradient: fullColor

	"Paint the rule to indicate value indication on the given color."

	| alpha box hue sat |
	alpha := 0.99.
	hue := fullColor asColor hue.
	sat  := fullColor asColor saturation.

	1 to: 10 do: [ :i|
		box := (rule frameChildren at: (i+1)).
		box	material color: (Color h: hue s: sat v: (i/10.0) alpha:alpha) asVectorColor.
		box checkAlpha.
	].! !

!QSlider methodsFor: 'events' stamp: 'das 12/22/2006 17:44'!
pointerDown: event

	| val |

	(event selection frame objectName = #slide or:[event selection frame objectName = #hair])ifTrue:[
		dragObject := slide.
		selectedPoint := event selection point.
		event selection frame objectName = #hair ifTrue:[
			selectedPoint := selectedPoint + hair translation.].
		selectedPoint := selectedPoint + slide translation.
		selectedPoint y:0.0.
	]ifFalse:[
		val := event selection frame translation y + event selection point y.
		self setValue: val/length.
	].! !

!QSlider methodsFor: 'events' stamp: 'bgf 11/13/2008 12:59'!
pointerEnter: event

	event avatar ifNil: [ ^ self ].

	slide hilite: (QColors generalDragColor).
	event avatar showFloatMessage: (self label), ' ', (self value roundTo: 0.01)
				 forEvent: event.
! !

!QSlider methodsFor: 'events' stamp: 'bgf 1/23/2007 11:51'!
pointerLeave: event

	slide unhilite.! !

!QSlider methodsFor: 'events' stamp: 'bgf 11/13/2008 13:00'!
pointerMove: event

	| ptr |
	event avatar ifNil: [ ^ self ].

	ptr := event makePointer.
	dragObject ifNotNil:[
		(ptr frame: rule pickPlane: selectedPoint normal: 
			(rule inverseGlobalOrientation transformPoint: 
			event selection cameraTransform lookAt negated)) ifTrue:[
				self setValue: ptr selectedPoint y/length.
			event avatar showFloatMessage: (self label), ' ', (self value roundTo: 0.01)
				   forEvent: event.
		^ true.].
	^ false.	].! !

!QSlider methodsFor: 'events' stamp: 'das 12/22/2006 17:35'!
pointerUp: event

	self pointerMove: event.
	dragObject := nil.
! !

!QSlider methodsFor: 'accessing' stamp: 'das 12/23/2006 14:21'!
setValue: val

	self value: val.
	self signal:#setValue with: value.
	target ifNil:[ ^self ].
	action ifNil:[ ^self ].
	target perform: action with: value.! !

!QSlider methodsFor: 'accessing' stamp: 'das 12/22/2006 17:23'!
target: trgt action: act

	target := trgt.
	action := act.! !

!QSlider methodsFor: 'accessing' stamp: 'bgf 1/23/2007 13:26'!
value
	^ value! !

!QSlider methodsFor: 'accessing' stamp: 'bgf 1/23/2007 13:02'!
value: val

	value := val.
	value < 0.0 ifTrue:[value := 0.0].
	value > 1.0 ifTrue:[value := 1.0].
	length ifNotNil:[
		slide translation: 0@(value*length-(slide height/2.0))@0.
		hair  translation: 0@(value*length-(hair height/2.0))@0.
	].
! !

TGroup subclass: #QStartStopButton
	instanceVariableNames: 'extent hotControl startIndicator stopIndicator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QStartStopButton commentStamp: '<historical>' prior: 0!
A button whose shape is a right-pointing triangle or a red square.
X-Y extent is of the 'face' of the button; Z extent is its thickness.

Supports three states with appearances:
	Disabled - Neither shape is visible.
	Started  - Square stopIndicator is visible.
	Stopped - Start indicator is visible.

Signals self #start when the start indicator is pressed, and goes to the started state.
Signals self #stop when the stop indicator is pressed, and goes to the stopped state.
The default initial state is stopped, with the start indicator visible.

!

!QStartStopButton class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStartStopButton class methodsFor: 'materials' stamp: 'hrs 8/8/2008 21:29'!
offMaterial
	"Material for the button shown when I not running, and my action is to start."
	^ self shade: (TMaterial new
					texture: (QTextures find: 'StartButton.png');
					opacity: 0.99;
					yourself)
		as: QColors controlRunningColor.! !

!QStartStopButton class methodsFor: 'materials' stamp: 'hrs 8/8/2008 21:30'!
runningMaterial

	"Material for the button shown when I am running, and my action is to stop."
	^ self shade: (TMaterial new
					texture: (QTextures find: 'PauseButton.png');
					opacity: 0.99;
					yourself)
		as: QColors controlRunningColor.
! !

!QStartStopButton class methodsFor: 'materials' stamp: 'hrs 8/8/2008 21:27'!
shade: aMaterial as: aColor 

	"colorize aMaterial by aColor in our style, answering aMaterial"
	^ aMaterial
		ambientColor: (aColor/1.1);
		diffuseColor: (aColor/1.1);
		specularColor: (aColor/4);
		emissiveColor: (aColor/1.1);
		yourself! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 12/19/2007 23:47'!
actionLabel

	^ hotControl caseOf: {	
		[startIndicator] -> ['Start'].
		[stopIndicator] -> ['Stop'].
	} otherwise: [
		'Start or Stop' 
	]! !

!QStartStopButton methodsFor: 'accessing' stamp: 'hrs 9/1/2008 11:52'!
click: event
	"Fire whichever action is hot."
	QSoundEffects play: #switch for: self.
	hotControl = startIndicator ifTrue: [
		self signal: #start with: event.
		self future enableAsStarted.
	] ifFalse: [
		self signal: #stop with: event.
		self future enableAsStopped.
	].! !

!QStartStopButton methodsFor: 'persist' stamp: 'bgf 12/20/2007 02:29'!
constructFrom: pStream
	"Note - state is not persisted!!"
	super constructFrom: pStream.
	self extent: (pStream readVector3: 'extent' default: (self extent)).! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 12/19/2007 22:29'!
disable
	hotControl := nil.
	self removeAll.! !

!QStartStopButton methodsFor: 'private' stamp: 'hrs 9/1/2008 11:57'!
disableIndicator: disabled enabled: enabled
	"If indicators are to be visible when disabled, remove them."

	self singleIndicator ifTrue: [
		enabled ifNotNil: [self addChild: enabled].
		self removeChild: disabled.
	] ifFalse: [
		"The control colors here are intentionally 'backwards.'
		The running color doesn't highlight well."
		enabled ifNotNil: [
			(self hasChild: enabled) ifFalse: [self addChild: enabled].
			self class shade: enabled material as: QColors controlDisabledColor.
			].
		(self hasChild: disabled) ifFalse: [self addChild: disabled].
		self class shade: disabled material as: QColors controlRunningColor.
		].! !

!QStartStopButton methodsFor: 'accessing' stamp: 'hrs 8/8/2008 21:14'!
enableAsStarted

	"Set state when the underlying activity is already running."
	self disableIndicator: startIndicator enabled: (hotControl := stopIndicator).
	self future unhilite.! !

!QStartStopButton methodsFor: 'accessing' stamp: 'hrs 8/8/2008 21:14'!
enableAsStopped

	"Set state when the underlying activity is already running."
	self disableIndicator: stopIndicator enabled:  (hotControl := startIndicator).
	self future unhilite.! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 12/19/2007 22:10'!
extent
	^ extent! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 12/19/2007 22:10'!
extent: v3
	
	extent := v3.
	self shapeChanged.! !

!QStartStopButton methodsFor: 'initialize' stamp: 'bgf 8/27/2009 17:02'!
initialize
	super initialize.

	self addChild: (startIndicator := TSimpleRectangle new).
	startIndicator material: self class offMaterial.

	self addChild: (stopIndicator := TSimpleRectangle new).
	stopIndicator material: self class runningMaterial.

	self singleIndicator ifTrue: [
		self startScript: #start: when: { startIndicator . #pointerDown }.
		self startScript: #stop: when: { stopIndicator. #pointerDown }.
	] ifFalse: [
		self startScript: #click: when: {self . #pointerDown }.
	].

	self enableAsStopped.
	self extent: (self singleIndicator ifTrue: [1] ifFalse: [2])@1@1.! !

!QStartStopButton methodsFor: 'persist' stamp: 'bgf 12/20/2007 02:29'!
persistOn: pStream
	"Note - state is not persisted!!"
	super persistOn: pStream.
	pStream storeVector3: 'extent' value: extent default: nil.! !

!QStartStopButton methodsFor: 'pick' stamp: 'bgf 6/9/2008 09:06'!
pick: pointer

	self state = #disabled ifTrue: [ 
		^ false
	].
	^ super pick: pointer.! !

!QStartStopButton methodsFor: 'transform' stamp: 'bgf 2/11/2010 17:19'!
privateScale: aFactor
	
	extent := extent * aFactor.
	super privateScale: aFactor.
! !

!QStartStopButton methodsFor: 'accessing' stamp: 'hrs 8/10/2008 20:36'!
singleIndicator
	"Should we swap between play an pause indicators (vs showing both side-by-side)?"
	^false! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 11/8/2009 08:48'!
start: event

	"Fire the go action."
	QSoundEffects play: #switch for: self.
	self hilite: (Color darkGray alpha: 0.6).
	self signal: #start with: event.
	self future enableAsStarted.! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 12/20/2007 02:27'!
state

	"Mostly for self-documentation... the state is held implicitly
	 by which indicator is the current hot (visible and clickable) control."

	^ hotControl caseOf: {	
		[startIndicator] -> [#stopped].
		[stopIndicator] -> [#started].
	} otherwise: [ #disabled ]! !

!QStartStopButton methodsFor: 'accessing' stamp: 'bgf 11/8/2009 08:48'!
stop: event

	"Fire the stop action."
	QSoundEffects play: #switch for: self.
	self hilite: (Color darkGray alpha: 0.6).
	self signal: #stop with: event.
	self future enableAsStopped.
! !

!QStartStopButton methodsFor: 'hierarchy' stamp: 'bgf 9/6/2009 12:33'!
updateShape
	| indicatorWidth offset |
	indicatorWidth := (self singleIndicator ifTrue: [self extent x] ifFalse: [self extent x / 2]).
	offset := ((self extent x - indicatorWidth) / 2).

	stopIndicator 
		extent: (indicatorWidth@ extent y);
		translation: offset negated @ 0 @ (extent z / 2.0).
	
	startIndicator 
		extent: (indicatorWidth@ extent y);
		translation: offset @ 0 @ (extent z / 2.0).

	self updateBounds.! !

TGroup subclass: #QTemporaryGroup
	instanceVariableNames: 'proxy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Editing'!

!QTemporaryGroup commentStamp: '<historical>' prior: 0!
The transient group created when edit-boxing an ad-hoc collection of children.

One instance exists for each QEditBox that has been given a collection of frames to edit rather than a single frame.!

!QTemporaryGroup class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTemporaryGroup class methodsFor: 'properties' stamp: 'bgf 1/27/2010 13:40'!
addPropertyRows: dialog forFrame: group
	| commonBase kids |
	commonBase := group get: #commonBaseClass.
	commonBase isSymbol ifFalse: [ ^ nil ].

	kids := group get: #children.
	dialog targets: kids.
	(Smalltalk at: commonBase ifAbsent: [ ^ nil ]) 
		addPropertyRows: dialog forFrame: (kids first).! !

!QTemporaryGroup class methodsFor: 'properties' stamp: 'bgf 1/28/2010 11:37'!
genericName
	^ 'Group'! !

!QTemporaryGroup methodsFor: 'editing' stamp: 'bgf 1/30/2010 15:20'!
canScale
	"The space-scaled nominalScale stuff means we can't rely on relative transform operations,
	 so we cannot scale ad-hoc groups."
	^ false! !

!QTemporaryGroup methodsFor: 'editing' stamp: 'bgf 1/30/2010 15:20'!
commonBaseClass
	"The Name of the most-specific class including all of my children."
	| klasses |
	klasses := self children collect: [ : k | k class ].
	^ (QPropertyDialog commonBaseClass: klasses) name
	! !

!QTemporaryGroup methodsFor: 'persist' stamp: 'bgf 1/20/2010 11:32'!
constructFrom: pStream
	"Persist the receiver"
	super constructFrom: pStream.
	self children: (pStream readSequence:'children' default: self children).
	! !

!QTemporaryGroup methodsFor: 'editing' stamp: 'bgf 1/30/2010 15:22'!
contextMenuMaskForAvatar: anAvatar
	| mask |
	"Wrong!!  Should create a group-menu merging the child menu masks.
	 But this is only used indirectly by the edit-box, so this suffices."
	mask := super contextMenuMaskForAvatar: anAvatar.
	mask remove: #copyToInventory ifAbsent: [].
	mask remove: #material ifAbsent: [].
	^ mask! !

!QTemporaryGroup methodsFor: 'initialize' stamp: 'bgf 2/7/2010 00:26'!
defaultIsAtomic
	"So the deletion rule-checks don't trip over this parent of the selected-frames
	 when editing a group in a box."
	^ false! !

!QTemporaryGroup methodsFor: 'initialize' stamp: 'bgf 1/28/2010 12:42'!
initialize
	super initialize.
! !

!QTemporaryGroup methodsFor: 'testing' stamp: 'bgf 2/6/2010 21:35'!
isEditableGroup
	^ true! !

!QTemporaryGroup methodsFor: 'menu' stamp: 'bgf 1/25/2010 16:49'!
labelForBrowser
	^ 'Group of ', (self children size), ' Items'! !

!QTemporaryGroup methodsFor: 'menu' stamp: 'bgf 1/20/2010 11:30'!
menuTitle
	^ '', (self children size), ' Items'! !

!QTemporaryGroup methodsFor: 'menu' stamp: 'bgf 2/4/2010 12:23'!
performDelete: event
	"See reallyDelete:  .  I never put myself on the recorded-path of the avatar,
	 undoing this delete brings back my chlidren, not myself"
	self reallyDelete: event! !

!QTemporaryGroup methodsFor: 'persist' stamp: 'bgf 1/20/2010 11:31'!
persistOn: pStream
	"Persist the receiver"
	super persistOn: pStream.
	pStream storeSequence: 'children' value: self children default: #().
	! !

!QTemporaryGroup methodsFor: 'editing' stamp: 'bgf 1/20/2010 14:07'!
photographerClass
	"No implicit out-of-box, since that entails my self-destruction."
	^ QDirectPhotoJob! !

!QTemporaryGroup methodsFor: 'menu' stamp: 'bgf 2/3/2010 17:16'!
reallyDelete: evtOrAvatar
	"A temp-group deletes its children formally, and simply self-destructs;
	 this is so delete+pasteback brings back the original kids, 
	 not the temporary editing group."
	| kids |
	self signal: #deleted with: self.
	kids := self children shallowCopy.
	self restoreFromEditing.
	evtOrAvatar avatar deleteFrames: kids event: evtOrAvatar.
	self obliterateLater.! !

!QTemporaryGroup methodsFor: 'initialize' stamp: 'bgf 1/20/2010 15:31'!
restoreFromEditing
	"Put my children back into my parent.   
	 The corrections here don't work under scaling effects."
	| p idx kids outerXForm outerScale |
	p := frameParent.
	p ifNil: [ 
		QLogger logWarning: 'Temporary group - was unparented before disposal.'.
		^ nil ].
	self allUpdateShape.

	outerXForm := self localTransform.		"Descaled"
	outerScale := self nominalScale.

	idx := p indexOfChild: self.
	p removeChild: self.
	kids := self children shallowCopy.
	kids do: [ : k |
		| innerXForm innerScale | 
		innerXForm := (k localTransform).
		innerScale := (k nominalScale).
		self removeChild: k.	
		k setParentScale: 1.0.
		k nominalScale: (innerScale * outerScale).
		k localTransform: (outerXForm * innerXForm).
		p addChild: k atIndex: idx.
		idx := idx + 1.
		].

	self obliterate.
	kids do: [ : k |
		k updateTransform.
		k setupAttachment.
		].! !

CTextEditor subclass: #QTextAppEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-TextApp'!

!QTextAppEditor commentStamp: '<historical>' prior: 0!
A non-replicated editor for a replicated QEmbeddedTextApp. Uses QTextAppParagraphEditor and QTextAppEditorPane to forward editing and scroll requests to the replicated app. Receives updates from the replicated app via events and applies them separately.!

QTextAppEditor defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="embeddedApp" toSet="embeddedApp:" changeEvent="embeddedAppChanged">embeddedApp</field>
	<field toGet="croquetSession" toSet="croquetSession:" changeEvent="croquetSessionChanged">croquetSession</field>
	<field toGet="doBlink" toSet="doBlink:" changeEvent="doBlinkChanged">doBlink</field>
	<field toGet="doBroadcast" toSet="doBroadcast:" changeEvent="doBroadcastChanged">doBroadcast</field>
</fields>'!

!QTextAppEditor class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 12/30/2007 11:47'!
checkUserOver
	| isOver |
	isOver := embeddedApp get: #isUserOver: with: croquetSession userID.
	isOver ifTrue:[
		doBlink := true.
	] ifFalse:[
		doBlink := false.
		editor showCaret: false. "ensure caret hidden after blink"
	].! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 12/30/2007 09:10'!
croquetSession
	"Answer the croquetSession of the receiver"
	<bewareOf: #croquetSessionChanged>
	^self propertyValueAt: #croquetSession! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 12/30/2007 09:10'!
croquetSession: aValue
	"Modify the receiver's croquetSession"
	^self propertyValueAt: #croquetSession put: aValue with: #croquetSessionChanged! !

!QTextAppEditor methodsFor: 'initialize' stamp: 'ar 4/19/2007 18:16'!
defaultTarget
	"Use a QTextAppPane to catch scroll requests"
	^QTextAppPane new! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 12/30/2007 09:29'!
doBlink
	"Answer the doBlink of the receiver"
	<bewareOf: #doBlinkChanged>
	^self propertyValueAt: #doBlink! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 12/30/2007 09:29'!
doBlink: aValue
	"Modify the receiver's doBlink"
	^self propertyValueAt: #doBlink put: aValue with: #doBlinkChanged! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 6/2/2008 20:14'!
doBroadcast
	"Answer the doBroadcast of the receiver"
	<bewareOf: #doBroadcastChanged>
	^self propertyValueAt: #doBroadcast! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 6/2/2008 20:14'!
doBroadcast: aValue
	"Modify the receiver's doBroadcast"
	^self propertyValueAt: #doBroadcast put: aValue with: #doBroadcastChanged! !

!QTextAppEditor methodsFor: 'operations' stamp: 'bgf 3/20/2009 10:09'!
editMenu
	| menu item |
	menu := CMenu new.
	menu label: nil.
	item := menu add: 'Find' target: self action: #findText.
	item := menu add: 'Find Next' target: self action: #findNext.
	item := menu add: 'Find and Replace' target: self action: #replace.
	menu addSeparator.
	item := menu add: 'Undo' target: self action: #undo.
	item := menu add: 'Redo' target: self action: #undo.
	menu addSeparator.
	item := menu add: 'Copy' target: self action: #copySelection.
	item := menu add: 'Cut' target: self action: #cutSelection.
	item := menu add: 'Paste' target: self action: #pasteSelection.

	QStyles installMenuFontInMenu: menu.
	^menu! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 4/19/2007 14:59'!
embeddedApp
	"Answer the embeddedApp of the receiver"
	<bewareOf: #embeddedAppChanged>
	^self propertyValueAt: #embeddedApp! !

!QTextAppEditor methodsFor: 'accessing' stamp: 'ar 4/19/2007 14:59'!
embeddedApp: aValue
	"Modify the receiver's embeddedApp"
	^self propertyValueAt: #embeddedApp put: aValue with: #embeddedAppChanged! !

!QTextAppEditor methodsFor: 'initialize' stamp: 'ar 6/2/2008 20:15'!
initialize
	super initialize.
	doBlink := false.
	doBroadcast := true.! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 12/30/2007 09:31'!
onBlinkCursor
	"Overridden to only blink if local user owns keyboard control of the app. 
	Avoids the problem that tons of text editors are wildly blinking when nobody
	is using them"
	doBlink ifTrue:[^super onBlinkCursor].! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 6/2/2008 20:50'!
onEmbeddedAppChanged
	"Update the editor after we aquired an embedded app"
	| ed |
	<on: embeddedAppChanged>
	ed := QTextAppParagraphEditor new.
	target textHolder editor: ed.
	"Set the following before installing the embedded app to avoid extra round-trips"
	ed replaceSelectionWith: (embeddedApp get: #string).
	ed selectFrom: (embeddedApp get: #markIndex)
		to: (embeddedApp get: #pointIndex).
	target scrollOffset: (embeddedApp get: #scrollOffset).
	"Then install embedded app"
	ed embeddedProxy: self.
	target embeddedProxy: self.
! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 6/30/2008 12:34'!
onKeyDown: evt
	"Delay delivery of the event in the case that we're currently suppressing broadcasts.
	We do this (suppressing broadcasts) to avoid the issue of run-away scrolling in text
	editors due to delayed events being handled later. However, that means that there
	can be keyboard events coming in while broadcast is suppressed so we have to delay
	the event delivery until such a time that broadcast is turned on again."
	[doBroadcast] whileFalse:[self waitTick].
	^super onKeyDown: evt! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 6/30/2008 12:34'!
onKeyStroke: evt
	"Delay delivery of the event in the case that we're currently suppressing broadcasts.
	We do this (suppressing broadcasts) to avoid the issue of run-away scrolling in text
	editors due to delayed events being handled later. However, that means that there
	can be keyboard events coming in while broadcast is suppressed so we have to delay
	the event delivery until such a time that broadcast is turned on again."
	[doBroadcast] whileFalse:[self waitTick].
	^super onKeyStroke: evt! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 12/30/2007 10:39'!
onPointerEnter
	"Limit blinking to situations where someone is actually interacting with the panel"
	<on: pointerEnter in: embeddedApp>
	self checkUserOver.! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 12/30/2007 10:39'!
onPointerLeave
	"Limit blinking to situations where someone is actually interacting with the panel"
	<on: pointerLeave in: embeddedApp>
	self checkUserOver.! !

!QTextAppEditor methodsFor: 'events' stamp: 'ar 6/2/2008 20:51'!
onUpdateText: updateRecord with: event
	"Update events that occurred in the replicated space"
	| record |
	<on: updateText in: embeddedApp>
	doBroadcast := false.
	[
		event withDroppedEventsDo:[:evt|
			record := evt arguments first.
			record first caseOf: {
				[#updateSelection] -> [editor selectFrom: record second to: record third].
				[#replaceSelection] -> [editor replaceSelectionWith: record second].
				[#scrollOffset] ->[target scrollOffset: record second].
			}.
		].
		"Delay re-enabling the broadcast script until the point that all the changes
		caused by the above operations trickled through the system. Mostly relevant
		for scroll offsets."
		self syncScript.
	] ensure:[
		doBroadcast := true.
	].! !

!QTextAppEditor methodsFor: 'protocol' stamp: 'bgf 3/11/2009 23:47'!
saveSharedContent
	"This is called from the general saveSharedContent regime;
	 we just dump the text files up to the server in their current form."

	| docRef participant location |
	embeddedApp isValidFarRef ifFalse: [ ^ self ].
	(docRef := (embeddedApp get: #document)) isValidFarRef ifFalse: [ ^ self ].
	(participant := (croquetSession owner)) ifNil: [ ^ self ].
	(location := (docRef get: #asServiceManagerPath)) ifNil: [ ^ self ].

	participant exitTripped ifTrue: [ ^ self ].

	"And share it on the server..."
	[ 
		participant serviceMgr 
				put: (contents asString squeakToUtf8)
				name: location 
				backup: true
				informing: nil.

		(((embeddedApp get: #isDirty) == true) and: [docRef isValidFarRef]) ifTrue: [
			docRef future mtime: (Time utcSeconds) who: (participant userLogin) what: 'Save'.
		].
	  ] on: Error do: [ :ex |
		QLogger logError: 'Cannot save ', location details: ex signalerContext shortStack from: self.
	  ]! !

!QTextAppEditor methodsFor: 'proxy' stamp: 'ar 9/17/2008 18:12'!
sendBackspace
	"Send a replicated backspace command.
	Answers true if the command has been broadcast, false if it should be applied locally."
	doBroadcast ifFalse:[^false]. "not broadcast"
	embeddedApp ifNotNil:[
		embeddedApp future backspace.
		^true].
	^false! !

!QTextAppEditor methodsFor: 'proxy' stamp: 'ar 6/2/2008 20:56'!
sendReplaceSelectionWith: aText
	"Send a replicated replace selection command.
	Answers true if the command has been broadcast, false if it should be applied locally."
	doBroadcast ifFalse:[^false]. "not broadcast"
	embeddedApp ifNotNil:[
		embeddedApp future replaceSelectionWith: aText.
		^true].
	^false! !

!QTextAppEditor methodsFor: 'proxy' stamp: 'ar 6/2/2008 20:56'!
sendScrollOffset: newOffset
	"Send a replicated scrollOffset change command.
	Answers true if the command has been broadcast, false if it should be applied locally."
	doBroadcast ifFalse:[^false]. "not broadcast"
	embeddedApp ifNotNil:[
		embeddedApp future scrollOffset: newOffset.
		^true].
	^false! !

!QTextAppEditor methodsFor: 'proxy' stamp: 'ar 6/2/2008 20:56'!
sendSelectFrom: markIndex to: pointIndex
	"Send a replicated selection update command.
	Answers true if the command has been broadcast, false if it should be applied locally."
	doBroadcast ifFalse:[^false]. "not broadcast"
	embeddedApp ifNotNil:[
		embeddedApp future selectFrom: markIndex to: pointIndex.
		^true].
	^false! !

!QTextAppEditor methodsFor: 'initialize' stamp: 'bgf 12/3/2009 10:13'!
setupCostume
	"Set up my costume"
	super setupCostume.
	self color: Color white.
	self hScrollable: false.
	self vScrollable: true.
	! !

TSimpleRectangle subclass: #QAppTrafficIndicator
	instanceVariableNames: 'app'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QAppTrafficIndicator commentStamp: '<historical>' prior: 0!
Displays an IO/indicator glyph that changes color according to the traffic level
of an associated remote embedded application.
!

!QAppTrafficIndicator class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAppTrafficIndicator class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:19'!
defaultEventMask
	^ EventPointerOver ! !

!QAppTrafficIndicator methodsFor: 'accessing' stamp: 'bgf 2/17/2008 20:04'!
embeddedApp: anEmbeddedApp

	"The app whose traffic level we'll use as the basis for the indicator color."
	app := anEmbeddedApp
	! !

!QAppTrafficIndicator methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:30'!
initialize
	| mat |
	super initialize.
	mat := QMaterials grayIndicatorLightMaterial.
	self material: mat.
	self solid: false.
	mat texture: (QTextures find: 'Content/Qwaq/Textures/appIOindicator.png').
	self checkAlpha.
! !

!QAppTrafficIndicator methodsFor: 'events' stamp: 'bgf 2/17/2008 18:33'!
pointerEnter: event
	self pointerOver: event! !

!QAppTrafficIndicator methodsFor: 'events' stamp: 'bgf 2/17/2008 20:05'!
pointerLeave: event
	"Clear the rollover feedback"
	self material emissiveColor: (Color gray:0.4) asVectorColor.! !

!QAppTrafficIndicator methodsFor: 'events' stamp: 'bgf 11/13/2008 13:08'!
pointerOver: event
	"Rollover feedback - a little floating help message."
	event avatar ifNil: [ ^ self ].
	self material emissiveColor: (Color gray: 0.6) asVectorColor.
	event avatar showFloatMessage: 'Application network activity' forEvent: event! !

!QAppTrafficIndicator methodsFor: 'render' stamp: 'bgf 2/17/2008 20:07'!
renderAlpha: ogl

	"Intercepted to substitute (temporarily) 
 	 the emissive color specified by the app's traffic level;
	 this is how we use harness-side data to paint the level-indication."

	| color save |
	app ifNotNil: [
		color := ogl view session trafficColorForApp: app appID.
		"Transcript cr; show: color."
		color ifNotNil: [
			[
				save := self material emissiveColor.
				self material emissiveColor: color.
				^ super render: ogl
			] ensure:[
				"Back as we were, before we leave the rendering cycle"
				self material emissiveColor: save
			].
		].
	].
	"Fallthrough - render as plain gray if we have no traffic data"
	^ super render: ogl

! !

TSimpleRectangle subclass: #QGrid
	instanceVariableNames: 'unitScale crosshairOrigin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QGrid commentStamp: '<historical>' prior: 0!
The grid renders a  regular series of unit-spaced orthogonal lines.
The grid has stronger lines for every fifth step.

It does not render the normal rectangle texture surface;
only its lines are drawn.
!

!QGrid class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QGrid class methodsFor: 'defaults' stamp: 'bgf 6/14/2009 02:31'!
defaultIsPickable
	^ false! !

!QGrid class methodsFor: 'defaults' stamp: 'bgf 11/13/2009 18:00'!
defaultMaterial
	^ (TMaterial new)
		diffuseColor: (Color r:0.5 g: 0.9 b: 0.65 alpha: 0.65);
		" Predates line-render:
			 texture: (QTextures find: 'Content/Qwaq/Textures/GridWhiteOnBlack.png') ; 
			premultiplied: true;"
		cullFace: false;
		yourself
! !

!QGrid class methodsFor: 'defaults' stamp: 'bgf 11/16/2009 17:03'!
defaultUnitScale
	^ 1.0 ! !

!QGrid methodsFor: 'accessing' stamp: 'bgf 4/2/2010 18:15'!
crosshairOrigin: aPointOrNil
	"If non-nil, the x-y origin of an extent-spanning pair of ortho lines, 
	 (serving as crosshairs for whatever movement the grid is informing.)"
	crosshairOrigin := aPointOrNil! !

!QGrid methodsFor: 'render' stamp: 'bgf 4/2/2010 18:05'!
doCrosshair: origin extent: extent ogl: ogl
	"Plot crossed lines meeting at the given origin."

	ogl glColor4fv: QColors editboxCrosshairColor.
	ogl glBegin: GLLines.
	ogl 
		glVertex3f:  (origin x - (extent x*0.5)) with: (origin y )  with: 0;
		glVertex3f:  (origin x + (extent x*0.5)) with: (origin y)  with: 0;

		glVertex3f:  (origin x ) with: (origin y - (extent y*0.5))  with: 0;
		glVertex3f:  (origin x ) with: (origin y + (extent y*0.5))  with: 0.

	ogl glEnd.! !

!QGrid methodsFor: 'render' stamp: 'bgf 11/16/2009 00:01'!
doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: dx 
	"Spit out a series of parallel lines, all running from lowy to highy;
	 the first line is at lowx, and the last at highx, spaced by dx."
	|  curx |
	curx := lowx.
	ogl glBegin: GLLines.
	[ curx <= (highx + Float epsilon) ] whileTrue: [
		ogl 
			glVertex3f:  curx with: lowy  with: 0;
			glVertex3f:  curx with: highy with: 0.
		curx := curx + dx.
	].
	ogl glEnd.! !

!QGrid methodsFor: 'render' stamp: 'bgf 11/16/2009 00:09'!
doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: dy 
	"Spit out a series of parallel lines, all running from lowx to highx;
	 the first line is at lowy, and the last at highy, spaced by dy."
	|  cury  |
	cury := lowy.
	ogl glBegin: GLLines.
	[ cury <= (highy+Float epsilon) ] whileTrue: [
		ogl 
			glVertex3f:  lowx with: cury  with: 0;
			glVertex3f:  highx with: cury with: 0.
		cury := cury + dy.
	].
	ogl glEnd.! !

!QGrid methodsFor: 'accessing' stamp: 'bgf 6/13/2009 19:29'!
extent: xy
	super extent: xy.
	self uvScale: (xy / 10.0).! !

!QGrid methodsFor: 'initialize' stamp: 'bgf 11/16/2009 17:03'!
initialize
	super initialize.
	self checkAlpha.
	unitScale := (self class defaultUnitScale).! !

!QGrid methodsFor: 'render' stamp: 'bgf 4/2/2010 18:16'!
renderPrimitive: ogl alpha: alphaPass
	| highx lowy highy  lowx color mcolor |

	highx := (extent x * 0.5).
	lowx := highx negated.
	highy := (extent y * 0.5).
	lowy := highy negated.

	color := self material diffuseColor.
	mcolor := (color clone - 0.2) alpha: ((color alpha * 1.33) min: 1.0).	"Major lines are more opaque."

	ogl	glLineWidth: 1.0;
		glDisable: GLLighting; 
		glColor4fv: color.

	"Minor lines, width 1, at spacing of 1 unit."
	self doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: unitScale.
	self doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: unitScale.

	"Minor lines at spacing of 5 units."
	ogl glLineWidth: 2.0;
		glColor4fv: mcolor.
	self doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: 5*unitScale.
	self doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: 5*unitScale.   

	crosshairOrigin ifNotNil: [ 
		ogl glLineWidth: 1.0.
		self doCrosshair: crosshairOrigin extent: extent ogl: ogl.
	].

	ogl glLineWidth: 1.0;
		glEnable: GLLighting.
! !

!QGrid methodsFor: 'render' stamp: 'bgf 11/13/2009 22:07'!
renderPrimitiveStippled: ogl alpha: alphaPass
	| highx lowy highy  lowx color rcolor |

	highx := (extent x * 0.5).
	lowx := highx negated.
	highy := (extent y * 0.5).
	lowy := highy negated.

	color := self material diffuseColor.
	rcolor := ((VectorColor r: 1 g: 1 b: 1) - color) alpha: (color alpha).

	ogl	glLineWidth: 1.0;
		glDisable: GLLighting; 
		glEnable: GLLineStipple;
		glColor4fv: color.

	"Minor lines, width 1, at spacing of 1 unit."
	ogl glLineStipple: 1 with: 4r30303030;
	       glColor4fv: color.
	self doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: 1.0.
	self doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: 1.0.
	ogl glLineStipple: 1 with: 4r03030303;
	       glColor4fv: rcolor.
	self doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: 1.0.
	self doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: 1.0.

	color := color clone alpha: ((color alpha * 1.25) min: 1.0).	"More opaque."
	rcolor := ((VectorColor r: 1 g: 1 b: 1) - color)  alpha: (color alpha).

	"Major lines at spacing of 5 units."
	ogl glLineWidth: 2.0;
		glColor4fv: color;
		glLineStipple: 1 with: 4r30303030.
	self doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: 5.0.
	self doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: 5.0.
	ogl glLineStipple: 1 with: 4r03030303;
	       glColor4fv: rcolor.
	self doXLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dx: 5.0.
	self doYLines: ogl lowX: lowx highX: highx lowY: lowy highY: highy dy: 5.0.

	ogl glLineWidth: 1.0;
		glDisable: GLLineStipple;
		glEnable: GLLighting.
! !

!QGrid methodsFor: 'accessing' stamp: 'bgf 11/16/2009 16:57'!
unitScale
	"The space between ticks."
	^ unitScale! !

!QGrid methodsFor: 'accessing' stamp: 'bgf 11/16/2009 16:57'!
unitScale: aFactor
	"The space between ticks."
	unitScale := aFactor! !

TRectangle subclass: #QCullingRectangle
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QCullingRectangle commentStamp: '<historical>' prior: 0!
(c) 2007 Qwaq Inc.  Culls rendering of the rectangle if the back is facing the camera.  

This is useful with QVideoTexture/QClientVideoTexture, where the video stream is paused when the texture is not being used... with a regular TRectangle, the texture is reported as being used even if OpenGL culls the rectangle face.

Note: Brad voiced his distaste for deciding to pause the stream based on texture use, preferring to use a heuristic that more directly solves the problem at hand.  I proceeded anyway because 1) the heuristic would have been basically the same as QCullingRectangle (only differing in the mechanism for notifying the stream to stop) and 2) the texture-use approach is general (you can slap a video texture onto any kind of object, and it will still pause when it isn't being rendered).!

!QCullingRectangle class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCullingRectangle methodsFor: 'render' stamp: 'bgf 10/13/2008 00:22'!
renderPrimitive: ogl alpha: alphaPass
	| toCamera myLookAt camLookAt flippedDot |

	toCamera := ogl camera globalTransform translation - self globalTransform translation.
	myLookAt := self globalTransform lookAt.
	camLookAt := ogl camera globalTransform lookAt.

	"Check whether the camera is pointing toward this rectangle."
	(toCamera dot: camLookAt) <= 0 ifTrue: [
		"Camera is pointing away from recangle.  This isn't a perfect test... for example, the field of view is probably less than 180 degrees, so this won't cull some stuff that it should."
		^self
	].

	"Check whether this rectangle is facing the camera."
	flippedDot := toCamera dot: myLookAt.
	self flip ifTrue: [flippedDot := flippedDot negated]. 
	flippedDot < 0 ifTrue: [^self "culled!!"].

	super renderPrimitive: ogl alpha: alphaPass
	
! !

TRectangle subclass: #QFileCard
	instanceVariableNames: 'content adapter rolodex thumbnail kind isTransient nextTransition transitionInProgress description customFaceTexture'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QFileCard commentStamp: '<historical>' prior: 0!
A small, draggable token holding  a 'document' of some sort 
(picture, video, text, url,  embedded application, portal, or 3D model).
Cards work in close conspiracy with 
QDisplayPanel/Sheet, QDocumentShelf, and QSharedDocument

Draggable; delivers its content into a DisplayScreen 
(or any other QDraggable drop target) if dropped on one.

The 'kind' attribute identifies the nature of the content in the card:

		#application		TEmbeddedApp
		#document		QSharedDocument or QTextDocument.
		#image			TTexture (usually a file texture)
		#mpeg			QSharedDocument identifying an MPEG-2 movie file.
		#object			QSharedDocument whose underlying content is the 
						XML persistence snapshot of a TFrame; can drop into space.
		#postcardLink	TPortal or TPostcard (these cards can be dropped onto QGates)
		#url			URL - dropping onto a sheet creates a remote browser session
		#videoLink		TPortal - the 'peer' fake video camera in the QPeeringDesk demo stuff.

#kind is used to allow many different types of content to be used in file cards
without needing a sea of subclasses
and without imposing presentation-protocols across them all.
See 'contentSummary' for example....
QFileCard is a junction-box in the system, where we get down to cases.

A card of kind=#object can be dropped into empty space,
to materialize a new instance of the assembly whose xml snapshot it contains.
Other kinds can be dropped onto QDisplaySheets to present the content
in a suitable viewer or editor.

The 'face' of the card can be specified as any texture;
most of the construction paths pick a default face for the situation.
The face is distinct from the backdrop, which is one of 
a set of standard iconic backgrounds - see #backdropForKind:
!

!QFileCard class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFileCard class methodsFor: 'inspector' stamp: 'bgf 1/28/2010 12:38'!
addPropertyRows: dialog forFrame: cardref
	| docs |

	"File-cards are just icons for their content,
	 so we offer very little  in the property dialog.
	 Mainly just the document attributes, if any."

	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Kind';
			fetchBlock: [  : f | f   get: #kindSummary ];
			acceptBlock: [ : v |];
			balloonHelp: 'Description of the card type';
		      inputEnabled: false;
			extraLong;
			yourself).

	docs := (dialog targets collect: [ : cd | cd get: #document ]).
	(docs allSatisfy: [ : d | d isValidFarRef ]) ifTrue: [
		(QPropertyDialog commonRefClass: docs) ifNotNilDo: [ : klass |
			(klass includesBehavior: QAbstractDocument) ifTrue: [
				klass addPropertyRows: dialog forFrame: (docs first).
				dialog rows allButFirst do: [ : row | row targets: docs ].
			] ifFalse: [
				QLogger logInfo: 'Common document class base problem: ', klass
			].
		
		]].
! !

!QFileCard class methodsFor: 'defaults' stamp: 'bgf 1/4/2010 16:12'!
backdropForKind: aKind
	"Return the background texture for cards of the given 'kind' value. "
	
	^ QTextures find: 'Content/Qwaq/Textures/Cards2/', (self backingNameForKind: aKind).! !

!QFileCard class methodsFor: 'defaults' stamp: 'bgf 6/16/2010 10:22'!
backingNameForKind: aKind

	"Return the background icon-file name for cards of the given 'kind' value. 
	  (We carry two flavors of each, one in client-side ./icons/Qwaq/Cards/. for the UI,
	  and one in bundled content (and the bundled forms table) for the texture inside cards."

	^ aKind caseOf: {
		[#document] -> ['DocumentCard.png'].
		[#panelApp]->['DocumentCard.png'].
		[#application] -> ['AppCard.png'].
		[#image] -> ['ImageCard.png'].
		[#object] -> ['ObjCard.png'].
		[#postcardLink] -> ['PortalCard.png'].
		[#url] -> ['PageCard.png'].
		[#mpeg] -> ['MovieCard.png'].
		[#qms] -> ['AppCard.png'].
		[#videoLink] -> ['CamCard.png'].
		[#streamingVideoV1] -> ['CamCard.png'].
		[#webcast] -> ['SecureVideoCard.png'].
		[#ssh] -> ['AppCard.png'].
	}  otherwise: [ 'ImageCard.png' ].

	! !

!QFileCard class methodsFor: 'defaults' stamp: 'bgf 11/28/2009 10:53'!
defaultExtent
	^ Point x: 1.8 y: 1.2! !

!QFileCard class methodsFor: 'defaults' stamp: 'bgf 11/28/2009 10:48'!
defaultMaterial
	^ QMaterials flatWhiteAlphaMaterial! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 1/7/2010 20:40'!
doStoreObject: aTFrame recipient: card forUser: anAvatar
	"Continuation for object-storage."
	QStoreObjectJob storeObject: aTFrame recipient: card forUser: anAvatar.
! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 2/15/2007 14:23'!
forApplication: anEmbeddedApp

	"Generally used when the application is set up with its own appIcon."
	^ (self new)
			kind: #application content: anEmbeddedApp;
			yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 4/29/2009 13:52'!
forDocument: document 
	"Give the user a shared-document file card representing the imported content"

	"Handy aliases for unifying harness-side doc/resource management..."
	document class = TFileTexture ifTrue: [
		^ self forTexture: document
	].
	(document isKindOf: QImageDocument) ifTrue: [
		^ (self new) kind: #image content: document; yourself.
	].

	document class = PyPanelDocCS ifTrue:[
		^ self forPythonCode: document.
	].

	^ (self new)
		kind: #document content: document;
		yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 1/2/2008 22:39'!
forFileNamed: aShortFileName text: aString

	"A file card representing the content of the given named file;
	 the file name is the short name, a reminder for later, not the full path file name."
	| doc |
	doc := (QTextDocument new)
			shortName: aShortFileName;
			text: aString;
			yourself.

	^ (self new)
			kind: #document content: doc;
			yourself.
	! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 9/7/2008 17:25'!
forMPEG: aSharedDoc

	^ (self new)
			kind: #mpeg content: aSharedDoc;
			yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 12/3/2008 14:09'!
forMPEGFile: aStringOrDoc sha: aShaString

	"Retrofit for demo-style prepackaged mpeg files (Alien, Qwaq-Intro)"
	| doc |
	doc := aStringOrDoc.
	doc isString ifTrue: [
		doc := (QSharedDocument new)
					sharedName: aStringOrDoc;
					suffix: 'mpeg';
					sha: aShaString;
					yourself.
	].
	^ self forMPEG: doc! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 1/2/2008 23:48'!
forPostcard: aPostcard 

	"Generally used when the application is set up with its own appIcon."
	^ (self new)
			kind: #postcardLink content: aPostcard;
			yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'das 9/2/2008 12:51'!
forPythonApp: appName

	"A file card representing a launchable Python app."
	| content |
	content := PyPanelDocCS new.
	content appName: appName.
	^ (self new)
			kind: #panelApp content: content;
			yourself.
	! !

!QFileCard class methodsFor: 'instance creation' stamp: 'das 12/22/2008 11:04'!
forPythonCode: codeDoc

	^(self new) 
			kind: #panelApp content: codeDoc;
			yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 1/2/2008 23:51'!
forServiceURL: aURL 

	| postcard |
	postcard := TPostcard urlString: aURL.
	^ (self new)
			kind: #postcardLink content: postcard;
			yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 1/8/2010 21:42'!
forTFrame: aTFrame user: anAvatar container: aRolodex
	"Handler for creating an object-template card from a live TFrame.
	 The destination rolodex is used to determine where the backing document is stored
	 (it will go in the user's private storage, if the rolodex is an inventory,
	 or into the current forum shared storage otherwise.)

	 The result is a file card whose 
		type will be#object,  
		content a QSharedDocument,
		face a TFileTexture (from a photo job.)"

	| desc card |
	desc := (aTFrame objectName ifNil: [aTFrame class name]) copy.
	card := (self new) 
		kind: #object content: nil;
		description: desc;
		yourself.

	"The harness of this user has to generate the icon-picture, 
	 and generate the stored scene archive;
	 we'll be called back via #facePhoto: and #objectStored: when done."
	aTFrame photographerClass
		photograph: aTFrame forUser: anAvatar recipient: card selector: #facePhoto:.

	"Defer this, so the handling of the picture can complete before we create the catalog entry."
	(self future: 250) 
		doStoreObject: aTFrame recipient: card forUser: anAvatar.

	^ card
! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 2/15/2009 02:05'!
forText: aString

	| doc |
	doc := (QTextDocument new)
			text: aString;
			shortName: 'New-Text-Document.txt';
			yourself.

	^ (self new)
			kind: #document content: doc;
			yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 4/29/2009 18:59'!
forTexture: aTexture
	^ (self new)
		kind: #image content: (QImageDocument forTexture: aTexture);
		yourself.! !

!QFileCard class methodsFor: 'instance creation' stamp: 'bgf 1/30/2009 01:32'!
forURL: aURL
	"A file card representing a web-browser session."
	^ (self new) 
		kind: #url content: aURL;
		yourself.


! !

!QFileCard class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 17:57'!
space: ref welcomeDoc: newDocCard oldWelcomeDoc: oldDocCard

	"Undoable change to the forum welcome"
	^ QBlockAction 
		undoBlock: [ : avatar | ref future welcome: oldDocCard fromUser: avatar ]
		redoBlock: [ : avatar | ref future welcome: newDocCard fromUser: avatar ]
		description: 'Set Welcome'
		requiredFrames: { ref }! !

!QFileCard methodsFor: 'comparing' stamp: 'bgf 9/9/2008 17:47'!
= other
	^ self class == other class
		and: [self kind = other kind ]
		and: [self content = other content]! !

!QFileCard methodsFor: 'editing' stamp: 'ar 4/16/2011 17:11'!
actionsList
	self canUseAsForumWelcome ifTrue: [
	 	^   {{#useAsForumWelcome: . 'Use as Forum Welcome' } }.
	].
	^ #()
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 4/29/2009 13:52'!
activityDetails

	"The details: argument to use for activity reports involving this file card.
	 For the important payloads, we use hash-anonymized names but include the document type.
	 For other stuff, just the kind of content."
	^ kind caseOf: {
		[#document] -> [ content ifNotNil: [ QActivityCapture fileDetails: (content shortName) ] ifNil: [ 'empty']].
		[#image] -> [ content ifNotNil: [ QActivityCapture fileDetails: (content shortName) ] ifNil: [ 'empty' ]].
		[#url ] -> [ 'url-', (content ifNotNil: [ content hash] ifNil: ['empty']) ].
		[#object ] -> [ content ifNotNil: [ QActivityCapture fileDetails: (content shortName) ] ifNil: [ 'empty' ]].
	  } otherwise: [
		self kindSummary.
	].! !

!QFileCard methodsFor: 'menus' stamp: 'sn 5/22/2013 13:39'!
addMenuActions: aMenu forEvent: event

	"Add my content-specific menu actions to the given menu."
	| item |

	(self canBeOpenedPrivately and: [self isInInventory not]) ifTrue: [
		item := aMenu addItemSpec: {'Compose Mail...' translateMe  . #mailOut:} target: self.
	].

	(self canBeOpened and: [ self isOpenInOwner not ]) ifTrue: [
		"Operation to open the content in the display panel, since I'm in its shelf."
		(item := aMenu addItemSpec: { 'Open' translateMe . #openInOwner: } target: self) 
	].
	(self canBeSavedLocally and: [self isInInventory not]) ifTrue: [
		item := aMenu addItemSpec: {'Open Document Page' translateMe . #openWebPage:} target: self.
	].

	self canBeExecuted ifTrue:[
		item := aMenu addItemSpec:{'Install Python Application...' translateMe . #savePython:} target: self.
	].

	self canBeSavedLocally ifTrue: [
		item := aMenu addItemSpec: {'Save Local Copy...' translateMe  . #saveAs:} target: self.
	].
	self canBeOpenedPrivately ifTrue: [
		item := aMenu addItemSpec: {'View Privately...' translateMe  . #openLocally:} target: self.
	].

	item ifNotNil: [ item separator: true ].

	(event avatar canUseInventory and: [ self canBeCopiedBy: event avatar ])  ifTrue: [
		item := aMenu addItemSpec: {'Copy to Inventory' translateMe . #copyToInventory: } target: self.
	].

	(content notNil and: [self root notNil]) ifTrue: [
		(self isInAvatar ) 
			ifTrue: [item := aMenu addItemSpec: {'Drop' translateMe . #release: } target: self]
			ifFalse: [item := aMenu addItemSpec: { 'Pick Up' translateMe . #pickUp: } target: self].
	].

	(self space notNil
	   and: [self canUseAsForumWelcome]
	   and: [self isInInventory not]
	   and: [event avatar canModifyForum ]
         and: [self ~= (self space welcomeDocument)])
	  ifTrue:[
		item := aMenu addItemSpec: {'Use as Forum Welcome' . #useAsForumWelcome: } target: self ].

	self holdsDocument ifTrue: [
		item ifNotNil: [ item separator: true ].
		item :=(aMenu addItemSpec: { 'Properties' translateMe  . #editProperties: } target: self).
	].

	(event avatar canModifyForum and: [ self canBeDeleted ])  ifTrue: [
		item ifNotNil: [ item separator: true ].
		(aMenu addItemSpec: {'Delete' translateMe . #confirmDelete:} target: self)
	].

	"Don't let it be empty below the title"
	aMenu menuItems size = 1 ifTrue: [
		(aMenu addItemSpec: { 'Cancel' translateMe . #noSuchAction } target: self) enabled: false
	].

! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/4/2010 14:07'!
backdropForKind: aKind
	"Return the background texture for the given 'kind' value;
	 the backdrop texture is the physical appearance of the card as a whole.
	 (It's not the face icon texture, which is placed atop the backdrop.)"

	"Allow shared documents of all kinds to provide their own index-card appearance."
	(content respondsTo: #fileCardTexture)
	  ifTrue: [
		content fileCardTexture ifNotNilDo: [ :custom | ^ custom ].
 	  ].

	^ self class backdropForKind: aKind! !

!QFileCard methodsFor: 'events' stamp: 'das 6/20/2008 23:35'!
beginDrag: event

	"Reject this operation if I'm owned and this isn't my owner."
	rolodex ifNotNil: [
		(rolodex notMyAvatar: event avatar) ifTrue: [ ^ false ].
		rolodex cards first = self ifTrue:[
			self saveForEvent: event.
		].
	].
	self isTranslucent: false.
	rolodex ifNotNilDo: [ :r | r openCard: self ].
	^ true! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 11/28/2009 10:46'!
bright: aBool
	"Some shelves use bright: true to emphasize a card.
	 (And are expected to apply bright: (true or false) to all their cards,
	 since the non-bright color is not the plain card color.)
	 This state is not persisted."
	aBool ifTrue: [
		self material
			ambientColor: (QColors pureWhite);
			diffuseColor: (QColors whiteAlphaColor);
			emissiveColor: (QColors gray3).
	] ifFalse: [
		self material
			ambientColor: (QColors gray5); 
			diffuseColor: (QColors gray6Alpha);
			emissiveColor: (QColors gray2).
	].! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 4/1/2010 08:48'!
canBeCopiedBy: anAvatar
	^ (super canBeCopiedBy: anAvatar)
	    and: [ self canBeCopiedToInventory ]		"Document-specific rules"! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 4/7/2009 12:42'!
canBeCopiedToInventory
	"Can I be copied to any user's inventory ?
	 (Controls whether menu actions are offered.)"

	self isInInventory ifTrue: [ ^ false ]. 	"Already in inventory."

	"Transient cards are fresh-dropped.  They are used to prevent
	 an extra-renaming document-copy between drop and arrival in a panel
	 (so when you drop foo.txt it lands in the panel as foo.txt).
	 But that copy-default also makes them unsuitable for copy-to-inventory,
	 since those paths always want to rename."
	self isTransient ifTrue: [ ^ false ].

	"Some simple kinds can always be copied to inventory."
	(kind = #url) ifTrue: [ ^ true ].
	(kind = #postcardLink) ifTrue: [ ^ true ].

	(content respondsTo: #canBeCopiedToInventory ) ifTrue: [
		^ content canBeCopiedToInventory
	].
	
	^ false! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 6/13/2009 09:34'!
canBeDeleted
	"Can I have a Delete entry in my menu?"
	^ adapter isDeletable ! !

!QFileCard methodsFor: 'menus' stamp: 'das 9/8/2008 18:42'!
canBeExecuted
	"Can I be saved locally by any user?"

	(content respondsTo: #canBeExecuted) ifTrue: [
		^ content canBeExecuted.
	].
	^ false! !

!QFileCard methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 14:16'!
canBeOpened
	"Can I be opened directly in my owner?
	 Part of recent-documents support.
	 An <owner> is an object that holds a rolodex child,
	 and implements
				canOpenCard: aCard
				hasCardOpen: aCard
				openCard: aCard fromUser: anAvatar

	 The card menu setup uses these methods to offer appropriate actions
	 (Open) and restrict others (Delete - disallowed if the card is active.)
	 when the card appears in a rolodex whose container implements the protocol."

	self owner ifNotNilDo: [ :frame|
		^ (frame respondsTo: #canOpenCard:) and: [ frame canOpenCard: self ].
	].
	^ false.
! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 12/30/2008 14:39'!
canBeOpenedPrivately
	"Can I be opened in a local private application?"

	kind == #object ifTrue: [ ^ false ].
	content ifNil: [ ^ false ].

	(content respondsTo: #canBeOpenedPrivately) ifTrue: [
		^ content canBeOpenedPrivately
	].

	"Web URL's - content is just a string."
	^ (kind = #url) and: [  (content beginsWith: 'local:') not]
! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 12/8/2008 06:50'!
canBeSavedLocally
	"Can I be saved locally by any user?"

	(content respondsTo: #canBeSavedLocally) ifTrue: [
		^ content canBeSavedLocally
	].
	^ false! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 10/25/2009 15:02'!
canEditInBox
	^ false! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 2/18/2009 22:00'!
canEditMaterial
	^ false! !

!QFileCard methodsFor: 'accessing' stamp: 'jcg 2/5/2009 10:37'!
canEmbedInDisplay
	"We cannot embed stored-objects (its sorta meaningless).
	 We can embed documents only if they have an application registered.
	 Other types (e.g. images) are OK. "
	^kind caseOf: {
		[#object]->[false].
		[#document]->[content notNil and: [ content embeddedAppClassName notNil]].
		[#panelApp]->[content notNil and: [ content embeddedAppClassName notNil]]
		} otherwise: [true].
! !

!QFileCard methodsFor: 'accessing' stamp: 'jcg 4/10/2009 11:27'!
canUseAsForumWelcome
	"Any embedded-app can be used as the Forum Welcome, except for streaming-video.  Audio is problematic because the DisplaySheet is not in the space, so we can't spatialize the sound.  Video also doesn't work for QRecordedSessionDocument (it's not set as an active-player in the QNetVidClient; not sure why." 
	self isApplicationContent ifFalse: [^false].
	kind == #document ifFalse: [
		"Want to disallow some documents, but this isn't a document, so it's OK."
		^true  
	].
	(content respondsTo: #canUseAsForumWelcome) ifFalse: [
		"We always want to be explicit about this, but in case we missed
		 implementing the method somewhere, we log a warning."
		QLogger logWarning: 'content does not implement #canUseAsForumWelcome'.
		^true
	].
	^content canUseAsForumWelcome! !

!QFileCard methodsFor: 'events' stamp: 'bgf 1/13/2007 16:06'!
cancelDrag: ev

	rolodex ifNotNil: [rolodex fileCard: self].! !

!QFileCard methodsFor: 'drag/drop' stamp: 'bgf 2/14/2010 15:24'!
carryWhenPasted
	"Currently the only thing for which do this..."
	^ true! !

!QFileCard methodsFor: 'render' stamp: 'bgf 11/28/2009 12:02'!
child: child setDistance: ogl
	
	"In a rolodex, we use the child/subsorting render tricks
	 to give the rolodex control over the alpha-sort order of cards,
	 so it decides which appears in front.
	 To engage that, we give all cards and their thumbs
	 the identical basic distance (the same as the rolodex distance)."
	^ rolodex 
		ifNotNil: [ (rolodex child: self setDistance: ogl) ]
		ifNil: [ super child: child setDistance: ogl ]! !

!QFileCard methodsFor: 'render' stamp: 'bgf 11/28/2009 12:01'!
childSubdistance: child

	"Give the rolodex control over the alpha-sort order of elements,
	 so it decides which appears in front."

	^ rolodex 
		ifNil: [ 2* Float epsilon negated ]
		ifNotNil: [ (rolodex cardSubdistance: self) + (2* Float epsilon negated ) ]! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 3/3/2009 22:47'!
clear
	"Make the card empty."
	self kind: nil content: nil! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/3/2008 01:03'!
clearThumbnail
	thumbnail ifNotNil: [
		self removeChild: thumbnail.
		thumbnail destroy.
		thumbnail := nil.
		customFaceTexture := nil.
	].
! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 1/9/2007 14:01'!
close
	rolodex fileCard:self.! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 6/2/2009 21:47'!
component
	"There are too many behaviors wired in here to allow another answer"
	^ self! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 2/1/2010 12:25'!
confirmDelete: event
	"Ask the user to confirm deletion.
	 Extended to warn of closing the content, if I am open-in-owner."
	| question |

	event avatar ifNil: [ ^ self ].

	question := (self isOpenInOwner) ifTrue: [
					('Delete ', (self deletionLabel), ' ?' , '\(Will clear display)') withCRs.
				] ifFalse: [
					('Delete ', (self deletionLabel), ' ?' )
				].

	event avatar confirm: question
			       selector: #performDelete: 
			       target: self 
				! !

!QFileCard methodsFor: 'persist' stamp: 'bgf 4/30/2009 12:37'!
constructContentFrom: pStream
	| k fc ct |
	"See persistContentOn:"

	k := (pStream readString: 'kind' default: 'image') asSymbol.
	description := (pStream readString: 'description' default: nil).
	fc := (pStream readObject: 'face' default: nil).

	"The interpretation of 'content' depends on the kind..."
	ct := (k caseOf: {
			[#object] -> [[
					(pStream readObject: 'content' default: nil).
				] on: Error do: [ :ex|
					"backward compatible (pre 1.0.33) read of encoded inline scene-xml" 
					| txt |
					txt := (pStream readString: 'content' default: nil) 
							ifNotNilDo: [:data|data base64Decoded].

					"Ideally we'd generate a shared document and upload the text;
	 		             to make the inventory xml smaller.  But that's awkward 
					 and not worth the fuss for this little-used case."
					QTextDocument forString: txt.
				]].

			[#url] 	-> [pStream readString: 'content' default: nil].
			[#mpeg] -> [pStream readObject: 'content' default: nil].

			"xsetroot before 1.1.34 - file card had a custom kind"
			[#xsetroot] -> [pStream readString: 'content' default: 'lightblue' ].
			
			"mpegFile is for pre-1.0.16 saves; no longer generated."
			[#mpegFile] -> [pStream readString: 'content' default: nil].
		} otherwise: [
			"Clumsy backward compatible case for text stored as a plain string; 
                    wrap the old saved text in a text document."
			[ 
				pStream readObject: 'content' default: nil
			] on: Error do: [ :ex| 
				k = #text ifTrue: [
					| txt |
					txt := pStream readString: 'content' default: nil.
					QTextDocument forString: txt.	
				] ifFalse: [
					self error: ex description.
				].
			].
		]).

	"Forward conversion... as of 1.1.34, the #text kind is retired,
	 although we still write it, for reverse-compatibility with recent pre-1.1.34 versions."
	(k == #text) ifTrue: [ k := #document ].

	"Old kind #xsetroot (pre 1.1.34) is just an application-card"
	(k == #xsetroot) ifTrue: [
		k := #application.
		ct := QRemoteServerApp name: #QRemoteXSetRoot
							  extent: QRemoteXSetRoot defaultExtent
					              data: { '' . '' . ct } 
	].

	"mpegFile (pre 1.0.16) predates QSharedDocument."
	(k == #mpegFile) ifTrue: [
		k := #mpeg.
		ct := (QSharedDocument new) sharedName: ct; suffix: 'mpeg'; yourself
	].

	"Retrofit the sha for this into its document, to enable unbundling it from the package
	 without needing a resource-bundling sha mapper for movies."
	k == #mpeg ifTrue: [
		( ct sharedName = 'Content/Qwaq/Movies/QwaqCroquetSmall.mpg') 
			ifTrue: [ ct sha: '5SY28ZQ1WCUQNKTX27QYCXSAB6KVGHI'].
	].

	"Provision for modernizing document types based on the suffix."
	((k == #document) and: [ ct class == QSharedDocument ]) ifTrue: [
		ct := self updateToSpecificDocument: ct.	
	].

	"For post 2.0.13, when QImageDocument was introduced.
	 We either have the document written beside the raw texture 'content',
	 or we simply invent one... this is how forums auto-upgrade to QImageDocument."
	((k == #image) and: [ ct isKindOf: TFileTexture ]) ifTrue: [
		ct := pStream readObject: 'imgdoc' default: (QImageDocument forTexture: ct)
	].

	"Initialize kind+content first, and then apply the custom face, since kind+content sets a face."
	self kind: k content: ct.

	"Pre 1.1.34, we saved the firefox file texture with -every- url file card;
	 now that association is builtin, so we ignore such custom faces."
	(fc isKindOf: TFileTexture) ifTrue: [
		(k == #url and: [ fc fileName = 'Content/Qwaq/Textures/firefox.jpg' ]) ifTrue: [
			fc := nil.
		].
	].
	"Pre 1.1.34 versions were sometimes haphazard abut what got saved as a custom face;
	 only TFileTextures internalize properly."
	(fc isKindOf: TFileTexture) ifTrue: [
		self customFaceTexture: fc
	].
! !

!QFileCard methodsFor: 'persist' stamp: 'bgf 1/13/2007 09:12'!
constructFrom: pStream
	| |
	super constructFrom: pStream.
	self extent: (pStream readPoint: 'extent'default: self extent).
	self constructContentFrom: pStream.
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 4/29/2009 13:53'!
content
	"The underlying data in the card,
	 whose type varies according to #kind."
	^ content! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 2/21/2007 00:55'!
contentLink

	"Return the TPostcard that this card is a link to,
	 if the card is a link; nil otherwise."
	(kind = #postcardLink) ifFalse: [
		^nil
	].
	content ifNil: [ ^ nil].
	^ content isPostcard ifTrue: [content] ifFalse: [content postcard].
! !

!QFileCard methodsFor: 'accessing' stamp: 'crl 1/12/2010 16:44'!
contentSummary

	"Return a string naming the content of the card, or nil if there is none."

	| appClass |

	description ifNotNil: [ ^ description ].
	content ifNil: [ ^ nil].

	^ (kind caseOf: {
		[#document] -> [content ifNil: ['Document (Empty)'] 
							ifNotNil: [content contentSummary]].
		[#panelApp] -> [content ifNotNil:[content contentSummary]].
		[#application] -> [
			appClass := QEmbeddedApp appNamed: content appName.
			appClass 
					ifNil: [content appName] 
					ifNotNil:[appClass summaryFor: content appData].
		].
		[#image] -> [
			thumbnail ifNil: [
				nil
			] ifNotNil: [
				| txtr sz |
				txtr := thumbnail material texture ifNil: [ ^ nil ].
				sz := (txtr memUsage/ 1024.0) truncated.
				sz := ' ', txtr extent, ' (', sz, ' kb)'.
				(txtr respondsTo: #fileName) ifTrue: [
					(self shortenPath:(txtr fileName ifNil: ['(Imported)'])), sz.
				] ifFalse: [
					(self shortenPath:(txtr url ifNil: ['(Imported)'])), sz.
				].
			].
		].
		[#mpeg] -> [content sharedName].	"Legacy - replaced by #document"
		[#ssh] -> [content shortName].
		[#object] -> [
			(description ifNil: ['Model'] ifNotNil: [description])
		].
		[#postcardLink] -> [(self qurlSummary)].
		[#qms] -> [content shortName].
		[#url] -> [(self content ifNil:['']) unescapePercents].
		[#videoLink] -> ['Video'].	"Peculiar to the peering-desk demo, not real video."
		[#streamingVideoV1] -> [ 'From ', (self publicNameForLogin: (self content userName))].
		[#webcast] -> ['Feed from room ''', content service, ''''].
	} otherwise: [nil]).
! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/3/2008 02:22'!
contentVisible: aBoolean

	"Determine whether the content of the card is visible."
	thumbnail ifNotNil: [
		thumbnail visible: aBoolean
	].! !

!QFileCard methodsFor: 'initialize' stamp: 'bgf 4/30/2009 19:53'!
copyFrom: aCard
	"Adopt the content of the given card as my own."
	| ct |
	"When documents have backing storage 
	 (such as for textures and shared-documents)
	 we create a new instance of the document object, 
	 so that when inventory-to-forum data transfer is used,
	 the updated file name (written into the doc object by the copy job) 
       doesn't clobber the original document and/or texture."

	((aCard kind = #image and: [ aCard content notNil ]) 		"Always defeat TTextures auto-reuse."
	 or: [ aCard hasDocumentStorage and: [aCard isTransient not]] ) 
	ifTrue: [
		ct := TSceneReader new duplicateScene: aCard content
	] ifFalse: [
		ct := aCard content
	].

	self kind: aCard kind content: ct.
	aCard customFaceTexture ifNotNil: [ 
		self customFaceTexture: aCard customFaceTexture 
	].! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 3/8/2008 22:26'!
copyToInventory: event

	"Copy my contents to the avatar's inventory rolodex."
	event avatar ifNotNilDo: [ :avatar| 
		"In case there's an app open on my content... get it saved."
		self saveForEvent: event.

		"Take the transfer forward to give the above save signal a chance.
	  	 For remote apps, event has to be noticed by the harness,
	 	 and the request sent via QMS, before the document save can complete."
		(avatar future: 1000) addCardToInventory: self.
	].! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 4/29/2009 15:01'!
creationTime
	"Creation utc timestamp of my content, if known, else 0 "
	(((kind = #document) or: [kind = #image]) and: [ content notNil ]) ifTrue: [
		^ content creationTime
	].
	^ 0
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/3/2008 01:03'!
customFaceTexture

	"The file texture used for the icon on this file card.
	 This is non-nil only when a per-instance texture has been supplied from outside;
	 the url firefox icon and other builtins are provided internally and not held here.
	 This texture, if non-nil, is persisted, and copied when the card is copied."

	^ customFaceTexture ! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/24/2008 03:15'!
customFaceTexture: aFileTexture

	"Use the given texture as the 2D thumbnail on the card,
	 and persist the texture with the card.
	 This facility should be used sparingly, as it bloats persistence.
	 Ordinarly the face textures are determined by the content + kind."
	self faceTexture: aFileTexture.
	customFaceTexture := aFileTexture.! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 9/14/2007 22:36'!
deletionLabel
	^ self shortSummary! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/8/2010 20:44'!
description
	"The string to use in describeTo: summaries.
	 Usually nil, since the summary is computed from the content.
	 If set, this is also returned as the result of  #contentSummary "
	^ description! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/8/2010 11:17'!
description: aString
	"The string to use in describeTo: summaries.
	 Usually nil, since the summary is computed from the content.
	 If set, this becomes the result of  #contentSummary "
	description := aString! !

!QFileCard methodsFor: 'initialize' stamp: 'bgf 3/3/2009 23:43'!
destroy
	"Lifecycle -release any document registration we've arranged."
	self releaseContent.
	super destroy.
! !

!QFileCard methodsFor: 'drag/drop' stamp: 'bgf 11/28/2009 10:45'!
disable
	adapter disable! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 11/28/2009 12:38'!
divisions: divisions
	^ 1@1		"Ignore - we optimize render"! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 8/21/2008 12:07'!
doNextTransition
	
	"If the nextTransition is set, perform it now;
	 this is sent as a timed future message to create some tolerance for rapid pointer motion
	 when animating opens and closes in the rolodex."
	nextTransition ifNotNil: [
		(self startTransition: nextTransition) ifTrue: [
			self perform: nextTransition.
			nextTransition := nil.
		] ifFalse: [
			QLogger logWarning:  'Overlapping card transitions (', nextTransition, ' ' , transitionInProgress, ')'.
		].
	].! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/12/2009 20:12'!
docChange: descrip by: avatar
	"Mark the document as having been changed by this person."

	self docChange: descrip by: avatar time: (avatar lastAliveUTC).! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 3/3/2009 22:29'!
docChange: descrip by: avatar time: utc
	"Mark the document as having been changed by this person."
	self holdsDocument ifTrue: [
		content mtime who: avatar userLogin what: descrip utc: utc.
	].! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/12/2009 18:32'!
docChange: descrip byID: userSessionID
	"If the avatar and its current time can be identified,
	 mark the document as having been changed by that person (now)"
	userSessionID ifNil: [^ self ].
	self space ifNotNilDo: [ : spc | 
		(spc avatarForSession: userSessionID) ifNotNilDo: [ :avatar|
			self docChange: descrip by: avatar
		]]. ! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/19/2009 18:54'!
docCreate: descrip by: avatar
	"Mark the document as having been created by this person."

	self docCreate: descrip by: avatar time: (avatar lastAliveUTC).! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 3/3/2009 22:30'!
docCreate: descrip by: avatar time: utc
	"Mark the document as having been created by this person."
	self holdsDocument ifTrue: [
		content ctime who: avatar userLogin what: descrip utc: utc.
	].! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/19/2009 18:55'!
docCreate: descrip byID: userSessionID
	"If the avatar and its current time can be identified,
	 mark the document as having been created by that person (now)"
	userSessionID ifNil: [^ self ].
	self space ifNotNilDo: [ : spc | 
		(spc avatarForSession: userSessionID) ifNotNilDo: [ :avatar|
			self docCreate: descrip by: avatar
		]]. ! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 8/1/2009 15:57'!
document
	"The QAbstractDocument I hold, if I hold one."
	^ (content class includesBehavior: QAbstractDocument) ifTrue: [ content ] ifFalse: [nil]! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 2/7/2010 00:24'!
documents
	"Protocol - the QAbstractDocuments I hold."
	^ self document ifNotNil: [ : d | { d } ]  
		   ifNil: [ #() ]
! !

!QFileCard methodsFor: 'drag/drop' stamp: 'bgf 11/28/2009 10:45'!
dragAdapter
	^ adapter! !

!QFileCard methodsFor: 'events' stamp: 'bgf 2/27/2007 11:10'!
dragInFrontOf: aTXSel graspDistance: aDistance

	"If the selection identifies an frame that is between
	 the selection's pick origin and the card, move the card
	 back so it appears in front of that frame.
	 This is protocol for external drag and drop, in which the harness determines
	 the selection before placing a card into the space.
	 Returns true if an obstacle was identified and the card repositioned."

	^ adapter dragInFrontOf: aTXSel graspDistance: aDistance
	! !

!QFileCard methodsFor: 'events' stamp: 'bgf 1/29/2009 23:45'!
dropInSpace: event pointer: pointer
	"Override for QDragAdapter #dropInSpace:pointer:
	 The card has been dropped into space, with no drop target in line.
       Only a few kinds materialize into something in this case.
	 Return true if something was created, false otherwise."

	event avatar ifNil: [ ^ false ].
	self isInInventory ifTrue: [
		QActivityCapture action: #DropInSpaceFromInventory 
						 details: (self activityDetails) 
						 forSession: (event avatar sessionHex)  ].

	"Stored objects are persisted TFrames that we unpickle into the space when dropped."
	self isStoredObject ifTrue: [
		| finisher |
	 	"The stored-object-xml  is a QSharedDocument;
		  We send this to a harness-side QCreateStoredObject job 
		  to fetch the document-file and internalize it.
	        Since we'll need the event and pointer to pose the resulting object,
	 	  we save them in a QCreateObjectCallback
	        which does the final posing after the harness builds the object."

		event avatar showFloatMessage: ('Creating ', self summary, '...') forEvent: event.
		finisher := QCreateObjectCallback new event: event avatar: event avatar pointer: pointer.
		QCreateStoredObjectJob instantiate: content recipient: finisher forUser: event avatar.
		self isTransient ifTrue: [ self destroy ].
		^ true
	].

	"Sound files are documents that we materialize in-space as sound frames."
	self isSampledSound ifTrue: [
		self dropSoundFrame: event pointer: pointer.
		self isTransient ifTrue: [ self destroy ].
		^ true
	].

	"Fallthrough for undroppable content... space-drop does nothing at all."
	^ false! !

!QFileCard methodsFor: 'events' stamp: 'bgf 1/29/2009 16:01'!
dropOnFrame: aFrame avatar: avatar selection: aTXSel 
	"Support for direct-drop, in which the harness determines
	 the drop target as part of handling a Squeak-level file drop event.
	 Returns true if a drop target was identified and the card sent to it."
	| result |
	result := false.
	adapter identifyDropTarget: (aTXSel frame).
	adapter currDropTarget ifNotNil: [
		adapter dropOn: adapter currDropTarget avatar: avatar selection: aTXSel.
		result := true.
	].
	adapter clearDragState.
	^ result! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/29/2009 01:46'!
dropSoundFrame: event pointer: pointer
	"Handler for drop-in-space when my content is a soundfile document...
	 make a sound-frame (a simplistic in-space player) holding the sound."
	| sf |
	sf := (QSoundFrame new)
				objectName: '', (content shortName);
				displayCard: self fromUser: event avatar;
				yourself.
	QDragAdapter drop: sf
				inSpace: (event avatar root) 
				forEvent: event 
				pointer: pointer.		


! !

!QFileCard methodsFor: 'editing' stamp: 'bgf 1/25/2010 11:55'!
editClass
	"File cards and containers manage the size and position."
	^ nil! !

!QFileCard methodsFor: 'drag/drop' stamp: 'bgf 11/28/2009 10:44'!
enable
	adapter enable! !

!QFileCard methodsFor: 'events' stamp: 'bgf 1/9/2007 13:51'!
endDrag: ev

	"Called at the end of a normal drag or no-target drop."

	rolodex 
		ifNotNil: [ rolodex fileCard: self ]
		ifNil:[ adapter isMovable ifFalse:[self snapBack]].

	"(We have to do the snap-back manually since we've taken delegation of  endDrag:)"! !

!QFileCard methodsFor: 'events' stamp: 'bgf 12/10/2006 16:49'!
endDrop: ev

	"Called at the end of a successful drop."

	isTransient ifTrue: [
		self delete.
	] ifFalse: [
		rolodex ifNotNil: [ rolodex fileCard: self ].
	].! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 1/9/2007 15:27'!
endTransition: aSymbol

	"Note that the card transition has ended; 
	 this allows a subsequent transition to take place."
	transitionInProgress := nil.
	! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 11/28/2009 12:12'!
extent: ext
	"Optimization... we are never aspect-Fixed."
	(extent ~= ext) ifTrue: [
		extent := ext asPoint.
		self shapeChanged.
	].

! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/30/2009 23:06'!
faceAppPreview
	"Make my face the standard QEmbeddedApp preview,
	 which is a semi-dynamic texture produced when apps generate snapshots.
	 The off-island image is stored and cached in a stylized way,
	 using a SHA generated from my <fullSummary>."

	self faceTexture:(QEmbeddedApp makeIconTexture: (self hashedName)).
	self shapeChanged.

! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/24/2008 13:41'!
faceDocPreview

	"We rely on the application preview-snapshot mechanism."
	self faceAppPreview.

! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/3/2008 01:00'!
facePhoto: aTTexture

	"Private; change the icon-image depicted on the card, leaving the content the same;
	 this is the callback by which we receive portraits of stored objects."

	self customFaceTexture: aTTexture.
	self shapeChanged.

	"Miserable hack to convince the auto-persisting inventory rolodex to save itself."
	(content notNil & rolodex notNil) ifTrue: [
		rolodex signal: #contentChanged.
	].
! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/3/2008 02:24'!
facePortal: aPostcard

	"As face, use the preview we have available for the given postcard.
	 That involves render-time lookup in the session preview cache,
	 and for that we use a specialized object."
	self clearThumbnail.
	thumbnail := QPortalBillboard new.
	thumbnail fullColor: true.
	thumbnail postcard: aPostcard.
	self addChild: thumbnail.
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/7/2010 22:51'!
faceTexture
	"If the card has an image nested within it, that is based on a normal thumbnail texture,
	 return that TFileTexture.   Otherwise, nil.
	 This will return, for example:
	 	The preview texture, if any, for application documents
		The image itself, for image documents
		 The mini-photo, for stored objects."

	^ self customFaceTexture 
		ifNil: [ 
			(thumbnail class = TSimpleRectangle) ifTrue: [
				thumbnail material texture
			] ifFalse: [
				self fileTexture
			]].! !

!QFileCard methodsFor: 'private' stamp: 'bgf 8/27/2009 18:16'!
faceTexture: aTexture

	"Use the given texture as the 2D thumbnail on the card."
	self clearThumbnail.
	aTexture ifNotNil: [
		thumbnail := TSimpleRectangle new.
		thumbnail material: ((QMaterials flatWhiteAlphaMaterial)
								texture: aTexture;	
								yourself).
		self addChild: thumbnail.
	].
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 4/30/2009 12:31'!
fileTexture
	"The content of an #image file card is a QImageDocument,
	 which always holds a TFileTexture; this returns that TFileTexture.
	 Returns nil for any other kind of content."

	(kind = #image and: [ content isKindOf: QImageDocument ]) ifTrue: [
		^ content fileTexture	
	].
	^ nil! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 3/3/2009 22:30'!
fileTextures
	"Include any textures carried inside my document..."
	^ self holdsDocument  ifTrue: [
			(content fileTextures, (super fileTextures)) asSet asArray
		] ifFalse: [
			super fileTextures	
		]! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/13/2010 09:15'!
forPythonApp: appName
	"A file card representing a launchable Python app."
	| doc |
	doc := PyPanelDocCS new.
	doc appName: appName.
	self kind: #panelApp content: doc.
	! !

!QFileCard methodsFor: 'render' stamp: 'bgf 11/28/2009 11:28'!
hasAlpha
	^ true! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/15/2009 02:22'!
hasDocumentStorage

	"Does the given content have a forum or user-specific server-side document-file
	 (if it does, then it usually needs to be copied into or out of user inventory, for example.)"

	^ (content respondsTo: #isDocumentInForumStorage) 
		or: [ content respondsTo: #isDocumentInInventory ]! !

!QFileCard methodsFor: 'render' stamp: 'bgf 11/28/2009 11:28'!
hasOpaque
	^ false! !

!QFileCard methodsFor: 'comparing' stamp: 'bgf 9/9/2008 16:47'!
hash
	^ (kind hash bitXor: (content hash)).
! !

!QFileCard methodsFor: 'accessing' stamp: 'ar 9/2/2010 11:39'!
hashedName
	"A name for preview images, etc, for the content of this file card.
	 This is -the- origin for the hashedName used in TEmbeddedApps to name preview-textures "

	^ (SecureHashAlgorithm new 
		hashStream: (ReadStream on: (self summary squeakToUtf8))) radix:36
! !

!QFileCard methodsFor: 'private' stamp: 'bgf 2/19/2009 18:38'!
holdsDocument
	"Used internally to determine some behavior choices."
	^ content isKindOf: QAbstractDocument! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/5/2008 16:57'!
holdsDocument: aSharedDocument

	"Locate-support for the inspector... return true if I hold the given document."
	^ (self content = aSharedDocument)! !

!QFileCard methodsFor: 'accessing' stamp: 'hrs 4/4/2009 20:29'!
identifier
	"A mostly human-readable string that distinguishes this file card from others in this forum 
	that are detectably different content. 
	Identifiers are made visible to the users through QRLs.
	An identifier should:
	- be stable under other changes to the forum (but could change under action on the card itself);
	- not include additional helpful labeling info that a user would not use in referring to the card;
	- be recognizablely matchable (modulo identity distinctions):
		-- when moved via inventory from one forum to another;
		-- when compared with in-world and inspector labels."

	kind caseOf: {
		[#object] -> [
			^ content ifNotNil: [ content sharedName ] ifNil: [ 'stored-object' ]  
						"Take the shared-document path (the c3z path) itself"
		].
		} otherwise: [].

	(content respondsTo: #identifier) ifTrue: [^content identifier].

		"Cards left over from older crashes during testing, etc,
 		 may simply produce nil as the content summary.
	  	 Give them an identifier that we can see reflects this."
		^self contentSummary ifNil: ['Empty'].! !

!QFileCard methodsFor: 'initialize' stamp: 'bgf 11/28/2009 11:05'!
initialize
	super initialize.

	adapter := QDragAdapter forLuggage: self grip: self.
	adapter overColor: (QColors whiteOverColor).
	isTransient := true.			"Until we're put into a rolodex."
	transitionInProgress := nil.

	self
		cullFace: false;
		normal: nil.

	self kind: nil content: nil.

! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 4/29/2009 14:21'!
isApplicationContent
	"Return true if this card is of a type that leads to an embedded application when displayed."

	^ self canEmbedInDisplay
		and: [ kind ~~ #image ]
	      and: [ #(document application url mpeg ssh) includes: kind ]

	"Should include panelApp, but right now py-apps presume that they are contained
	 in QDisplayPanels.   This method is used to validate content for setting up welcome-apps,
	 which are not empanelled."! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 6/2/2009 21:04'!
isFromTemplate
	^ false! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 6/2/2009 21:04'!
isFromTemplate: bool
	"No-op for file cards."! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 12/31/2008 01:42'!
isInInventory
	"Am I a card in someone's inventory?
	 (If so, then my contents are generally not available freely as resources in the space,
	  since they are in my owner's private storage.)"
	^ rolodex notNil and: [ rolodex isInAvatar ]! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 1/19/2008 14:18'!
isOpenInOwner
	"Am I already open (in the owner - the container of my rolodex)?
	 Meaningful in gates, display panels, 
	 and other devices using a rolodex of recent/available documents."

	self owner ifNotNilDo: [ :frame|
		^ (frame respondsTo: #hasCardOpen:) and: [ frame hasCardOpen: self ]
	].
	^ false.
! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 3/3/2009 22:30'!
isPDF
	"Am I a PDF shared-document (opened via QMS, through special-cases on the QRemoteFirefox path)"
	^ self holdsDocument and: [ content docType isPDF].! !

!QFileCard methodsFor: 'persist' stamp: 'bgf 10/21/2009 14:18'!
isRetiredDemoContent
	"See display-panel and doc-shelf. 
	 This rule is used to ignore/discard certain items
	 on recovery from persistence. 
	 Should only return true for content we pre-populated
	 into forums, that we no longer want to have available."

	"The ancient 'QwaqCroquetMovie' littered about old style conference rooms."	 
	(kind == #mpeg 
	 and: [ content notNil ] and: [ content sha = '5SY28ZQ1WCUQNKTX27QYCXSAB6KVGHI' ])
	ifTrue: [ ^ true ].

	"Help on an in-space browser in old conf rooms, which predates having user guide on web."
	(kind == #url
	 and: [ content isString ] and: [ content beginsWith: 'local:docs/wwhelp/wwhimpl/js' ] )
	ifTrue: [ ^ true ].
 
	^ false.! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 3/17/2009 20:39'!
isSampledSound
	"For sound-frame support... is my document a sampled sound?"
	^ self holdsDocument 
		and: [content class == QSharedDocument]	"E.g. not personal share"
		and: [content docType isSampledSound]! !

!QFileCard methodsFor: 'testing' stamp: 'bgf 1/29/2009 23:43'!
isStoredObject
	"Do I contain a pickled object... a document holding a .c3z of a frame-tree?"
	^ kind = #object and: [ content notNil ]! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/9/2007 15:21'!
isTransient

	"A transient file card deletes itself after a successful drop into a drop target.
	 Cards are transient by default.  Specifying a rolodex: sets transient to false."

	^isTransient! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 8/17/2006 23:14'!
isTransient: aBool

	"A transient file card deletes itself after a successful drop into a drop target.
	 Cards are transient by default.  Specifying a rolodex: sets transient to false."

	isTransient := aBool! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 11/28/2009 10:49'!
isTranslucent
	"Cards display either as slightly translucent (normal)
       or fully opaque (for focus display.)"

	self transparency < 0.98! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 11/28/2009 11:40'!
isTranslucent: aBoolean

	"Cards display either as slightly translucent (normal)
       or fully opaque (for focus display.)"
	
	thumbnail ifNotNil:[
		self transparency: (aBoolean ifTrue: [0.88] ifFalse:[0.96]).	
		thumbnail transparency: (aBoolean ifTrue: [0.93] ifFalse: [0.99]).
		thumbnail checkAlpha.
		false ifTrue: [
			"If in-card previews are too expensive, this would be one workaround...
	 	 	 turn them off when the card is meant to be 'away' in a shelf."
			thumbnail visible: (aBoolean not).
		].
	] ifNil: [
		self transparency: (aBoolean ifTrue: [0.92] ifFalse:[0.99]).	
	].
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 12/3/2008 23:36'!
kind

	"The classification of 'content', the thing represented and shown by this card;
	 one of #application, #image, #object, #postcardLink, #url, #videoLink, and others.

	 kind is used to allow many different types of content to be used in file cards
	 without needing a sea of file-card subclasses
	 and without imposing a great many presentation-protocols across them;
	 see 'shortSummary' for example....
	 QFileCard is a junction-box in the system, where we get down to cases."

	^ kind! !

!QFileCard methodsFor: 'initialize' stamp: 'bgf 11/28/2009 10:50'!
kind: aKind content: cont
	"Configure the card to represent the given content.
	 If the content is nil, revert to #image as kind, 
	 since the code paths handle nil properly for #image."

	cont == content ifFalse: [ self releaseContent ].
	kind := aKind ifNotNil: [aKind asSymbol].
	content := cont.
	self root ifNotNil: [ self registerContent ].
	"The kind determines the backdrop (the texture of the card as a whole)"
	self material texture: (self backdropForKind: kind).
	self clearThumbnail.

	"Various cases determine the face texture, which is the center icon on the card."
	[
		(kind == #object) ifTrue: [
			^ self clearThumbnail.	"A custom texture may be passed in later, via #facePhoto:"
		].
		(kind == #url) ifTrue: [
			"Shorthand for a web-application document (i.e. a url opened in firefox)."
			^ self faceAppPreview
		].
		(kind == #document) ifTrue: [ 	
			^self faceDocPreview
		].
		(kind == #panelApp)  ifTrue: [
			content ifNil:[^self].
			content isCodeDoc ifTrue:[
			] ifFalse: [
				^ self faceAppPreview
			].
		].

		(cont isKindOf: TObject) ifFalse: [
			"Early return before the remain isXXX cases; otherwise they might give MNU."
			^ self clearThumbnail.
		].
		"Now we know content is a TObject of some sort (and is non-nil)."
		kind == #image ifTrue: [
			^ self faceTexture: (cont fileTexture).
		].
		cont isPostcard ifTrue: [
			^ self facePortal: cont.
		].
		(cont isKindOf: TEmbeddedApp) ifTrue: [
			cont hashedName: self hashedName.
			^ self faceAppPreview.
		].

	] ensure: [
		self shapeChanged.
	].
! !

!QFileCard methodsFor: 'accessing' stamp: 'crl 1/12/2010 16:45'!
kindName

	"User-friendly string for the kind of card this is."
	content ifNil: [ ^ 'Empty' ].
	^ self kind caseOf: {
		[#document] ->['Document'].
		[#application] -> ['Application'].
		[#image] -> ['Image'].
		[#object] -> ['Stored Object'].	
		[#panelApp] -> ['Python'].
		[#postcardLink] -> ['Link'].
		[#qms] -> ['Multi-share'].
		[#url] -> ['Web URL'].
		[#mpeg] -> ['Movie'].
		[#ssh] -> ['SSH'].
		[#videoLink] -> ['Video Link'].
		[#streamingVideoV1] -> ['Webcam Video'].
		[#webcast] -> ['Secure Video'].
	} otherwise: [ 'File Card' ].
! !

!QFileCard methodsFor: 'accessing' stamp: 'crl 1/12/2010 15:42'!
kindSummary

	"Return a word summarizing the kind of the card, or nil if there is none."

	^ (kind caseOf: {
		[#document] -> [(content ifNil: [^ 'Document'])  kindSummary]. 
		[#application] -> [ 'Application'].
		[#image] -> ['Image'].
		[#mpeg] -> ['Movie'].
		[#ssh] -> ['SSH'].
		[#object] -> ['Stored Object'].
		[#panelApp] -> [content ifNotNil:[content isCodeDoc ifTrue:['[Python App]'] ifFalse:['Python Doc']] ifNil: [ 'Python']] .
		[#postcardLink] -> ['Portal Link'].
		[#qms] -> ['Multi-Share'].
		[#url] -> [ ((self contentSummary ifNil: ['']) beginsWith: 'local:')
					ifTrue: ['Shared'] 
					ifFalse: ['URL']].
		[#videoLink] -> ['Chat'].	"Peculiar to the peering-desk demo, not real video."
		[#streamingVideoV1] -> ['Webcam Video'].
		[#webcast] -> ['Secure Video'].
	} otherwise: ['Empty']).
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 12/26/2008 13:31'!
labelForBrowser
	"We want to equate the card with the content in the users mind."
	^ self shortSummary! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 3/1/2010 12:20'!
mailOut: event

	"Mail Out menu action:
	 open an external mail application on the content of this card.
	 Offered only if (self canBeOpenedPrivately)."

	event avatar ifNil: [ ^ self ].

	"Best-effort to get current content saved before going on to the download..."
	self saveForEvent: event.

	"Run the appropriate job to open local mail client on (maybe a link to) the content."
	(content respondsTo: #canBeOpenedPrivately) ifTrue: [
		content mailFromAvatar: event avatar.
		^ self
	].
	(kind = #url)  ifTrue: [
		| url |
		content isString ifFalse: [ ^ self ].
		url := content.
		"Firefox tolerates unheaded URL's and so must we."
		(url beginsWith: 'http' caseSensitive: false) 
			ifFalse: [ url := 'http://', content ].
		QMailDocumentJob mail: url fromUser: event avatar
	].
! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 6/2/2009 21:02'!
markAsFromTemplate
	"This is a no-op for file cards."
	^ self! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 3/3/2009 23:25'!
markAsInventory: bInventory
	"If my content requires, mark it is being in or not-in user-inventory."
	(content respondsTo: #isInInventory:) ifTrue: [
		content isInInventory: bInventory.
		self root ifNotNil: [ self registerContent ].
	].! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 6/2/2009 21:02'!
markAsNotFromTemplate
	"This is a no-op for file cards."
	^ self! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 1/2/2009 13:54'!
menuTitle
	"When clicking the card itself, label the menu with the most user-relevant thing we have..."
	^ (self shortContentSummary)! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 4/29/2009 13:56'!
modificationTime
	"Most recent utc timestamp of my content, if known, else 0 "
	(((kind = #document) or: [kind = #image]) and: [ content notNil ]) ifTrue: [
		^ content modificationTime
	].

	^ 0
	! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/3/2009 23:35'!
objectStored: aSharedDocument

	"Callback after generated data for a stored-object
	 has been uploaded to the server as a shared document
	 and the resulting shared document object is returned into the space;
	 if the given document is nil, the original copy-to-inventory operation was bust."
	content == aSharedDocument ifFalse: [ self releaseContent ].
	content := aSharedDocument.
	self markAsInventory: (self isInInventory).	 "So the content is flagged."
	self root ifNotNil: [self registerContent].
	rolodex ifNotNilDo: [ :dex|
		content ifNil: [ self destroy ].
		dex signal: #contentChanged
	].! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 1/9/2007 14:00'!
open
	rolodex openCard: self.! !

!QFileCard methodsFor: 'recent-documents' stamp: 'bgf 1/19/2008 14:18'!
openInOwner: event

	"Menu callback - open in the display panel or gate or whatever owns my rolodex.
	 See also #canBeOpened."

	event avatar ifNil: [ ^ self ].
	
	(self owner respondsTo: #openCard:fromUser:)  ifTrue: [
		self owner openCard: self fromUser: event avatar
	].! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 3/1/2010 12:18'!
openLocally: event

	"Open Private Copy menu action:
	 open an external desktop application on the content of this card.
	 Offered only if (self canBeOpenedPrivately)."

	event avatar ifNil: [ ^ self ].

	"Best-effort to get current content saved before going on to the download..."
	self saveForEvent: event.

	"Run the appropriate job to open a private copy in a local application."
	(content respondsTo: #doOpenForAvatar:) ifTrue: [
		"The delay is to create room for the save to complete."
		(content future: 1000) doOpenForAvatar: event avatar.
		^ self
	].

	(kind = #url)  ifTrue: [
		| url |
		content isString ifFalse: [ ^ self ].
		url := content.
		"Firefox tolerates unheaded URL's,  and so must we.
	          (In years of use, the only problem arising from  this crude test  was  https:;
	 	  which is why the beginsWith test doesn't include the :// part.)"
		(url beginsWith: 'http' caseSensitive: false) 
			ifFalse: [ url := 'http://', content ].
		QOpenDocExternalJob open: url forUser: event avatar
	].! !

!QFileCard methodsFor: 'menus' stamp: 'hrs 4/6/2009 21:50'!
openWebPage: event

	"Open Web Page action:
	 open an external web browser on the url for this card.
	 Offered only if (self canBeOpenedPrivately)."

	event avatar ifNil: [ ^ self ].

	"Best-effort to get current content saved before going on to the download..."
	self saveForEvent: event.

	"Run the appropriate job to open local mail client on (maybe a link to) the content."
	(content respondsTo: #canBeOpenedPrivately) ifTrue: [
		content openWebPageForAvatar: event avatar.
		^ self
	].
! !

!QFileCard methodsFor: 'private' stamp: 'jdm 8/21/2007 08:16'!
orientContent: frame

	"Orient the frame so that the least dimension of its bounding box is in my Z direction,
	 and the largest is along my X axis.  No longer used,
	 but kept because it was a headache to work out properly."

	| box x y z yAxis zAxis swap rotation |
	frame localTransform: Matrix4x4 new.

	box := frame boundingBox.

	x := box extentX.
	y := box extentY. yAxis := #y.
	z := box extentZ.  zAxis := #z.

	"The following three clauses are a minimal bubble-sort."
	z > y ifTrue: [	
		swap := y. y := z. z := swap.
		yAxis := #z.
		zAxis := #y.
	].
	y > x ifTrue: [	
		swap := y.  y := x. x := swap.   
		yAxis := #x.
	].
	z > y ifTrue: [	"Y must have changed... so X took its place in check two."
		yAxis := zAxis.
		zAxis := #x.
	].
	
	rotation := Vector3 new. 
	zAxis caseOf: {
		[#x] -> [ rotation y: -90 ].
		[#y] -> [ rotation x: -90 ].
		[#z] -> [ ].
	}.
	yAxis = #x ifTrue: [
		"Depends on whether we've rotated z ."
		zAxis caseOf: {
			[#x] -> [ rotation x: 90 ] .
			[#y] -> [ rotation y: -90 ] .
			[#z] -> [ rotation z: -90 ] 
		}
	].
	frame localTransform: (frame localTransform rotation: rotation).

	^ x.! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 1/19/2008 13:12'!
owner
	"Not my rolodex, but rather the container that holds my rolodex.
	 For instance, the display panel, or gate, if I'm in their recent-documents list."
	^ rolodex ifNotNil: [ rolodex parent ]! !

!QFileCard methodsFor: 'accessing' stamp: 'crl 1/12/2010 16:45'!
panelPlaceCategory

	"For panel support - where does a panel appear in the Places groupings?"
	^ kind caseOf: {
		[ #streamingVideoV1 ] -> [ #Video ] .
		[#webcast] -> [#Video].
	} otherwise: [
		'Panels'		"Default - just batch it in the documents area"
	].! !

!QFileCard methodsFor: 'private' stamp: 'bgf 6/28/2010 11:07'!
parent: aFrame
	"Reset fullbright state when reparented, in case a card moves into or out of an avatar carry.
	 UpdateShape will fix it for avatars."
	super parent: aFrame.
	self allMaterialsDo: [ : m | m fullBright: false ].
! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 11/28/2009 10:44'!
payload
	"Implementing subclasses may override this to return the payload
	 that a drop should make available to the #acceptDrop: recipient.
	 By default, the draggable presents itself as the dropped content."

	^ self
! !

!QFileCard methodsFor: 'persist' stamp: 'bgf 4/30/2009 12:37'!
persistContentOn: pStream
	"Persistence helper - persist the portions of the card state that describe its content,
	 leaving out the inherited TFrame geometry stuff, which we recompute anyway."

	pStream storeString: 'kind' value: kind default: nil.
	pStream storeString: 'description' value: description default: nil.
	"Below we defend against misuse of customFaceTexture, which must be a persistable file texture..."
	pStream storeObject: 'face' 
			 value: ((customFaceTexture isKindOf: TFileTexture) ifTrue: [ customFaceTexture ] ifFalse: [nil]) 
			 default: nil.

	kind caseOf: {
		[#url] -> [pStream storeString: 'content' value: content].
		[#image] -> [ 
			"We write the texture as 'content', and the document beside it;
	 		 this gives backward-forward compatibility, since the pre-2.0.14 reader
	             finds the texture 'content' it expects."
			pStream storeObject: 'content' value: self fileTexture default: nil.
			pStream storeObject: 'imgdoc' value: content default: nil
		 ].
	} otherwise: [
		content ifNotNil:[ 
			pStream storeObject: 'content' value: content default: nil.
		].
	].
! !

!QFileCard methodsFor: 'persist' stamp: 'bgf 3/14/2010 10:41'!
persistOn: pStream
	"We only persist the card's TFrame information if not in a rolodex;
	 if we are in a rolodex, we only persist the content.  This is a way of
	 forcing all the other information (transforms, etc) to be treated as default;
	 since the rolodex imposes all that anyway."
	rolodex ifNil:[
		super persistOn: pStream.
		pStream storePoint: 'extent' value: self extent default: (self class defaultExtent).
	] ifNotNil: [
		self persistIdentityOn: pStream
	].
	self persistContentOn: pStream.! !

!QFileCard methodsFor: 'render' stamp: 'bgf 11/16/2009 13:42'!
photographerClass
	^ QDirectPhotoJob! !

!QFileCard methodsFor: 'drag/drop' stamp: 'bgf 11/28/2009 12:07'!
pick: pointer
	"Drag adapter protocols."
	adapter isBeingDragged ifTrue: [
		^false
	].
	^super pick: pointer! !

!QFileCard methodsFor: 'menus' stamp: 'das 1/8/2009 21:29'!
pickUp: event

	| grabbed xform |

	self saveForEvent: event.
	"Pick up the card for carrying elsewhere."
	self isTransient ifTrue: [
		grabbed := self
	] ifFalse: [
		grabbed := QFileCard new copyFrom: self.
	].
	grabbed extent: (0.75@0.5).
	xform := event avatar globalTransform clone.
	xform translateBy: (event avatar lookAt negated * 2.5).
	
	event avatar root addChild: grabbed.
	grabbed localTransform: xform; updateTransform.
	event avatar startCarrying: grabbed.! !

!QFileCard methodsFor: 'events' stamp: 'bgf 10/25/2009 12:36'!
pointerClick: event

	"Disallowed if I'm held in a rolodex that is owned by an avatar."
	rolodex ifNotNil: [
		(rolodex notMyAvatar: event avatar) ifTrue: [ ^ self ].
		"Try to keep the card from staying big.."
		rolodex future closeAll.
	].
	! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 1/24/2008 03:28'!
possibleTransition: aSymbol

	"aSymbol is one of #open or #close, the two animated transitions
       that we do inside a rolodex.   If we're not already in the process
	 of the transition, schedule it ASAP. "

	transitionInProgress = aSymbol ifTrue: [
		"We are transitioning to the desired state already."
		nextTransition := nil.
	] ifFalse: [
		nextTransition ifNil: [
			"If we're doing the opposite, let if finish before we start."
			"The 130 suits the rolodex transitions
			 (longer than the QDocumentShelf openCard:at: duration)"
			(self future: (transitionInProgress ifNil: [0] ifNotNil:[130]))
				perform: #doNextTransition.	
		].
		nextTransition := aSymbol.
	].! !

!QFileCard methodsFor: 'private' stamp: 'bgf 6/23/2009 09:40'!
publicNameForLogin: aLoginString
	"Find the best public name we can for the given user login.
	 If I'm rooted, and the user in the space, we'll find their screen name."
	^ self space ifNotNilDo: [ :spc |
			spc publicNameForLogin: aLoginString
		] ifNil: [
			QStyles simplerLoginName: aLoginString
		].! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 4/27/2007 17:59'!
qurlSummary

	"Readable form of a postcard link url ('qurl')."
	| url |
	url := self contentLink ifNotNilDo: [ : cl | cl urlString ].
	url ifNil: [ ^ '' ].
	((url size - (url lastIndexOf:$/)) = 36) ifTrue: [
		url := (url copyFrom: 1 to: (url size-36)), '...' 
	].

	^  url! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 6/13/2009 09:43'!
reallyDelete: event

	"User-delete of the file card."
	event avatar ifNotNil: [
		QActivityCapture 
			action: (self isInInventory ifTrue: [ #DeleteCardInventory ] ifFalse: [ #DeleteCard ])
			details: self activityDetails
			forSession: (event avatar sessionHex).
	].

	self isOpenInOwner ifTrue: [
		self owner closeCard: self forEvent: event
	].

	super reallyDelete: event.! !

!QFileCard methodsFor: 'private' stamp: 'bgf 4/30/2009 19:16'!
registerContent 
	"Do whatever is appropriate before relinquishing access to 'content'.
	 The owner of a document is the rooted card pointing to it.."
	self holdsDocument ifTrue: [
		content registerResource.
	].! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 11/25/2009 09:02'!
release: event
	"Stop carrying me. Just leave me where I am."

	| copy |

	self root ifNil: [ ^ self ].	"Not carried, been deleted, just go away."

	self root miramarUI ifTrue:[
		rolodex ifNil:[
			copy := event avatar releaseCard: self.
		] ifNotNil:[
			copy := QFileCard new copyFrom: self.
		].
		^event avatar dropCard: copy
	].

	rolodex ifNil: [event avatar carryingRelease: self]
		ifNotNil: ["Drop a copy"
			copy := QFileCard new copyFrom: self.
			self root addChild: copy.
			copy localTransform: self globalTransform.
			].! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/3/2009 23:34'!
releaseContent
	"Unregister the content (if it is a document resource that we have registered.)"
	self root ifNotNil: [	
		self unregisterContent
	].
	content := nil.! !

!QFileCard methodsFor: 'render' stamp: 'bgf 11/28/2009 13:40'!
renderFrame: ogl
	" OPTIMIZED - we know we have at most one child, 
	 and that our child bounds are always inside our own bound sphere. "
	| count ac hilite |

	(bitflags anyMask: FlagVisible) ifFalse:[^ 0 ].	"Inlined self visible"
	count := 0.
	ac := ogl camera.

	"Is this frame itself inside the viewing pyramid?"
	(ac testBounds: boundSphere  scale: cumScale) ifTrue:[
		hilite := self pushHilite:ogl.
		self setupAlpha: ogl.
		count := 1.
		thumbnail ifNotNil: [
			count := count + (thumbnail renderFrame: ogl).
		].
		hilite ifTrue: [self popHilite:ogl] 
	].
	^ count.! !

!QFileCard methodsFor: 'render' stamp: 'bgf 6/25/2010 11:22'!
renderPrimitive: ogl alpha: alphaPass
	| rMesh |
	"Optimized - we know we are a 1-strip procedural mesh with fixed material."
	rMesh := ogl getMesh: self changed: (lastMeshChange).
	rMesh ifNil: [ ^ self ].

	self defaultMaterial enable: ogl.
	ogl beginMesh: rMesh.
	ogl drawStrip: (rMesh stripData at: 1) mesh: rMesh.
	self defaultMaterial disable: ogl.
	ogl endMesh.

! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 1/9/2007 15:22'!
rolodex
	"The rolodex containing the card, if any.
	 The card notifies the rolodex
	 of mouse pointer activity; rolodexes adjust the card size 
	 during mouse-overs and drag operations."

	^ rolodex! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 12/21/2008 12:52'!
rolodex: r
	"The rolodex containing the card.  The card notifies the rolodex
	 of mouse pointer activity; rolodexes adjust the card size 
	 during mouse-overs and drag operations."

	isTransient := false.
	rolodex := r.
	self extent: r basicExtent.

	self markAsInventory: (self isInInventory).! !

!QFileCard methodsFor: 'hierarchy' stamp: 'bgf 3/3/2009 23:23'!
root: aSpace
	"Maintain registration of documents as resources
	 so long as the card is rooted in space (reachable in the scene.)"
	(aSpace ~= self root) ifTrue: [ 
		aSpace ifNil: [ self unregisterContent ]
				ifNotNil: [ self registerContent ] ].
	super root: aSpace.
 ! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 2/8/2008 19:21'!
saveAs: event

	"Assume save-copy means save the current stuff...
	 i.e. induce a saveSharedContent on the container app if applicable."

	self saveForEvent: event.

	"And then (after a pause for station identification - to let the save-command reach the app)
	 steer this through the replica where the common code is."
	event avatar ifNotNilDo: [ :av | (av future: 1000) doLocalSave: content ]! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 2/8/2008 19:54'!
saveForEvent: event

	"Trigger a save of my content where appropriate,
	 using the most direct path we can find.
	 This attempts to induce saves in the underlying embedded applications
	 that may have the content open."

	(self isOpenInOwner and: [self owner isKindOf: QDisplayPanel]) ifTrue: [			
		self owner sheet onSaveForUser: (event userID)
	] ifFalse: [
		"This gets to the same place, but takes an extra signal/cycle."
		self signal: #saveForUser with: (event userID). 
	].
! !

!QFileCard methodsFor: 'menus' stamp: 'das 1/9/2009 16:31'!
savePython: event

	| app |
	event avatar ifNil: [ ^ self ].

	"Run the appropriate job to open a private copy in a local application."
	(content respondsTo: #doInstallPython:) ifTrue: [
		"The delay is to create room for the save to complete."
		(self isOpenInOwner and: [self owner isKindOf: QDisplayPanel]) ifTrue: [			
			app := self owner sheet contents. "probably a PyPanelApp"
			(app isKindOf: PyPanelApp) ifTrue:[
				app pyInstalledUserID: event avatar userID.
				app runScript: #onPyAppInstalled when:{content. #pyAppInstalled}.
			].
		].
		content doInstallPython: event avatar.
		^ self
	].

! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/5/2009 10:14'!
serviceURL 
	"The forums URL I live in."
	^ (self island at: #ServiceURLString) ifNil: [
			QLogger logWarning: 'Early serviceURL (', (self summary), ')'.
			nil
	 	]! !

!QFileCard methodsFor: 'private' stamp: 'bgf 8/21/2008 20:17'!
serviceURLString
	"Private, for debugging."
	^ (self island at: #ServiceURLString) ifNil: [
		QLogger logWarning: 'Premature service URL access in file card.'.
		'*Harness*'
		].! !

!QFileCard methodsFor: 'initialize' stamp: 'das 9/2/2008 09:47'!
setCardText: aString
	| doc |

	"This allows Python to create a text document"
	doc := (QTextDocument new)
			text: aString;
			yourself.

	self  kind: #document content: doc.! !

!QFileCard methodsFor: 'initialize' stamp: 'bgf 5/5/2009 16:44'!
setCardTexture: aTexture

	self  kind: #image content: (QImageDocument forTexture: aTexture).! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 3/3/2009 23:33'!
shallowCopyFrom: aFileCard
	"Shallow in the sense that we are sharing the document reference;
	 contrast with copyFrom: where we pickle-unpickle to duplicate the document.
	 This is for limited use... allowing temporary cards,  without excess duplication."

	self kind: aFileCard kind content: aFileCard content! !

!QFileCard methodsFor: 'private' stamp: 'bgf 1/8/2010 11:29'!
shortContentSummary
	"Make path-like content summaries shorter by eliding portions;
	 we do this by cases because (for instance) 
	 the host site of a URL is the useful summary,
	 while the service name (second-last part) of a service URL is better,
	 and the very last part - the filename - is best for file paths."

	description ifNotNil: [
		^ self shortenPath: self contentSummary
	].

	content ifNil: [ ^ 'Empty'].

	kind = #image ifTrue: [
		^ self shortenPath: (
				thumbnail material texture
					ifNotNilDo: [ :txtr | txtr fileName ifNil: [ txtr url ]] 
					ifNil: [nil])
	].
	kind = #url ifTrue: [
		^ self shortenWebPath: self contentSummary
	].
	kind = #postcardLink ifTrue: [
		^ self contentLink ifNotNilDo: [:pc| pc shortDisplayString] 
				ifNil: [self contentSummary].
	].

	^ self shortenPath: self contentSummary! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 3/3/2009 22:37'!
shortSummary
	"Return a string tersely summarizing the content of the card."
	| cts |
	content ifNil: [ ^ 'Blank'].
	cts := self shortContentSummary.
	cts ifNil: [ ^ self kindSummary].
	^ cts,' - ', (self kindSummary).
! !

!QFileCard methodsFor: 'private' stamp: 'das 1/11/2009 12:11'!
shortenPath: aURLString

	"Casual shortening of file path strings - strip to the rightmost component - the file name."
	| slash str start end |
	aURLString ifNil: [^ ''].

	"We can encounter either pathing scheme on any platform - depends
	 where it was imported from, not where this copy of the island is running."
	slash := aURLString lastIndexOf: $/.
	slash = 0 ifTrue: [slash := aURLString lastIndexOf: $\]. 	

	str := aURLString copyFrom: slash+1 to: aURLString size.
	kind = #panelApp ifFalse:[^ str].

	"remove anything between [ and ]"
	start := str lastIndexOf: $[.
	end := str lastIndexOf: $].
	^(start < end)ifTrue:[(str copyFrom:1 to: start-1),(str copyFrom: end+1 to: str size)] ifFalse:[str].
	! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/4/2007 04:05'!
shortenServiceURL: aURLString

	"Casual shortening of Qwaq Service URL's - strip to the Service name."
	| slash result |
	aURLString ifNil: [^ ''].

	result := aURLString.

	"Take off the last component - the tframe reference inside the service...
	 this will need to be made more general as more interesting postcards come into use."
	slash := aURLString lastIndexOf: $/.
	slash > 0 ifTrue: [
		result := aURLString copyFrom: 1 to: slash-1.
	].

	"Now treat it like a file path - peel off everything but the rightmost component,
	 so host and realm don't appear in the short summary - just the service name."
	^ self shortenPath: result.
! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/14/2007 13:06'!
shortenWebPath: aURLString

	"Casual shortening of path-like strings,for summaries and titles."
	| slash slashslash |
	aURLString ifNil: [^ ''].
	(aURLString beginsWith: 'local:') ifTrue: [
		^ (aURLString copyFrom: ((aURLString lastIndexOf:$/) + 1) to:  aURLString size)
	].

	slashslash := aURLString indexOfSubCollection:'//'.
	slash := aURLString indexOf:$/ startingAt: slashslash+2.
	slashslash = 0 ifTrue: [slashslash := -1].

	slash = 0 ifTrue: [slash := (aURLString indexOf:$?)].
	slash = 0 ifTrue: [slash := (aURLString size)+1].

	^ (aURLString copyFrom:slashslash+2 to: slash-1) 

! !

!QFileCard methodsFor: 'menus' stamp: 'bgf 10/25/2009 20:23'!
showContextMenu: event
	"Cards are really document icons, and provide a complete context menu.
	 (Most system-supplied operations for object-handling don't apply,
	 and the document has many specific actions that need to be here."
	self showMenu: event.
	rolodex ifNotNil: [  rolodex future closeAll ].! !

!QFileCard methodsFor: 'rolodex' stamp: 'das 3/31/2009 17:47'!
showFloatFor: event
	"Show floating info message with a description of the card."
	| msg |

	event avatar ifNil: [ ^ self ].
	"Don't say anything if this card is empty."
	content ifNotNil: [
		kind = #panelApp ifTrue:[ 
			msg := self kindSummary, ': ', (self shortenPath:self contentSummary)
		]ifFalse:[
			msg := self summary.
		].
		msg ifNotNil: [
			rolodex ifNotNil: [
				msg := '(', (rolodex cards indexOf: self ifAbsent: ['?']), ') ', msg.
			].
			event avatar showFloatMessage: msg forEvent: event
		].
	].! !

!QFileCard methodsFor: 'private' stamp: 'bgf 5/28/2007 22:51'!
signalOnce: aSymbol
	"A one-time start-script used to await #transferComplete and re-signal it on myself."
	self signal: aSymbol.
	self stopScript: #signalOnce:! !

!QFileCard methodsFor: 'drag/drop' stamp: 'bgf 11/28/2009 10:42'!
snapBack
	"If in the midst of a drag operation, return the luggage to the initial position."
	adapter snapBack! !

!QFileCard methodsFor: 'private' stamp: 'bgf 12/26/2008 15:22'!
space
	"Since file-cards are often unrooted members of other objects,
	 it is convenient to have access to the (presumed) containing space
	 for operations like 'Use as Forum Welcome' "
	^ self root ifNil: [
		((self island resourcesAt: (TSpace resourceType)) ifEmpty: [^nil])  first
	].! !

!QFileCard methodsFor: 'rolodex' stamp: 'bgf 10/21/2006 18:18'!
startTransition: aSymbol

	"If no transition is in progress, note this one, and return true.
	 If a transition is already in progress, return false."

	transitionInProgress ifNotNil: [^ false].
	transitionInProgress := aSymbol.
	^ true.! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 2/18/2009 16:10'!
summary
	"Return a string summarizing the content of the card."
	| cts |
	content ifNil: [ ^ 'Blank'].
	cts := self contentSummary.
	cts ifNil: [ ^ self kindSummary].

	"Keep in sync with QAbstractDocument summary.
	 (File card covers other kinds of content also, but should share this format with QAD)"
	^ self kindSummary, ': ', self contentSummary
! !

!QFileCard methodsFor: 'accessing' stamp: 'hrs 2/17/2009 09:26'!
thumbnailSHA
	"The sha of the preview picture, for use by harness-side tools."
	^thumbnail ifNotNil: [thumbnail defaultMaterial texture sha]! !

!QFileCard methodsFor: 'document store' stamp: 'das 12/20/2009 14:17'!
transferToCurrentForum: anAvatar
	"If this card is a document that lives in a user's inventory storage,
	 or in the storage of a different forum,
	 then copy that document into this forum's shared storage.
	 Sent when dropping a card onto a display sheet, 
	 before launching its application.
	 Returns true if a copy job is initiated --
	 we will have to wait for the transferComplete signal
	 before launching the application in that case."
	
	"Set the holder-status flag in any case."
	self markAsInventory: false.

	(((content respondsTo: #isDocumentInInventory) 
		and: [ content isDocumentInInventory ])
	 or: [
	  (content respondsTo: #isDocumentInForum:)  
		and: [ (content isDocumentInForum: (self serviceURL)) not ]])
	 ifTrue: [
		self startScript: #signalOnce: 
			withArguments: {#transferComplete} 
			when: {content . #transferComplete}.
		QCopyDocumentJob copyToForum: content forUser: anAvatar.
		self docCreate: 'Copy In' by: anAvatar.
		^true 
	].

	^ false! !

!QFileCard methodsFor: 'document store' stamp: 'bgf 2/19/2009 18:55'!
transferToInventory: anAvatar

	"The holder status flag is set in any case."
	self markAsInventory: true.

	"If this contains a URL, then make sure it is absolute,
	 so it refers back to the origin when used in a different forum."
 	((kind = #postcardLink) and: [self contentLink isAbsolute not]) ifTrue: [
		self kind: #postcardLink content: 
			(TPostcard urlString: 
				(TPostcard makeUrl: (self contentLink urlString) 
						    relativeTo: anAvatar islandURL)).
			^ false 
	].

	"If this card is a document that lives in a forum shared storage,
	 then copy that document into the user's private inventory storage.
	 Sent when dropping a card into inventory.
	 Returns true if a copy job is initiated.. false indicates that
	 the document is either already in normal forum store, 
	 or that this card has some other kind of content altogether.."

	((content respondsTo: #isDocumentInForumStorage) 
	 and: [ content isDocumentInForumStorage ])
	 ifTrue: [
		self startScript: #signalOnce: 
			withArguments: {#transferComplete} 
			when: {content . #transferComplete}.
		QCopyDocumentJob copyToInventory: content forUser: anAvatar.
		self docCreate: 'Into Inventory' by: anAvatar.
		(content respondsTo: #isInInventory:) ifTrue: [
			content isInInventory: true
		].
		^true 
	].
	^ false
! !

!QFileCard methodsFor: 'private' stamp: 'bgf 4/30/2009 19:15'!
unregisterContent 
	"Unregister the document resource (if I have one)."
	self holdsDocument ifTrue: [
		content unregisterResource.
		].! !

!QFileCard methodsFor: 'accessing' stamp: 'bgf 3/3/2009 23:26'!
updateContent: cont
	"Change the content, assuming that classification remains the same."
	content = cont ifTrue: [^self "no change"].
	content == cont ifFalse: [ self releaseContent ].
	content := cont.
	self root ifNotNil: [ self registerContent ].
	self updateShape.! !

!QFileCard methodsFor: 'initialize' stamp: 'bgf 6/28/2010 11:07'!
updateShape
	"Update geometry.  Note that my extent is only stored in my rectangles.
	 My 'thumbnail' is a rectangle depicting my face texture.
	 It may be the content itself, or a face supplied to depict more complex content.."
	| inner |

	inner := self extent * 0.667.
	thumbnail ifNotNil: [
		"Fit any aspect-aware rectangle inside the card.
		 (Otherwise, the rectangle makes itself larger than we ask for in extent,
		 which messes up the appearance of the card.)"
		(thumbnail class includesBehavior: TSimpleRectangle) 
		ifTrue: [
			thumbnail extent: (inner x)@(inner y) aspect: thumbnail originalAspect.
		] ifFalse: [
			thumbnail extent: (inner x)@(inner y).
		].
		"XXX Without this, the thumbnail occasionally rejects the extent
		 and displays at a size proportional to the image; I don't know why."
		thumbnail doUpdateShape.
		isTransient ifTrue: [
			thumbnail visible: true.
		].
		thumbnail translationX: (self extent x * 0.025) y: (self extent y * 0.05) z: 0.003.
	].

	"Avatar-carried cards (including inventory) are lighting-immune."
	self isInAvatar ifTrue: [ self allMaterialsDo: [ : m | m fullBright: true ] ].
	super updateShape.! !

!QFileCard methodsFor: 'persist' stamp: 'bgf 4/6/2009 18:06'!
updateToSpecificDocument: sharedDoc
	"The argument is (exactly) a QSharedDocument.
	  If required, convert it to a more specific subclass suitable to the type of its content.
	  Return sharedDoc, or its replacement."

	| klass result convertible |

	"The target classes we want to update to."
	convertible := {QStreamingVideoDocument . QRecordedSessionDocument}.

	"What is the doc-type prescribed class for the content of sharedDoc?"
	klass := ((QDocumentType forSuffix: (sharedDoc suffix)) ifNil: [ ^ sharedDoc ])
			   docClass.

	klass == sharedDoc class ifTrue: [ ^ sharedDoc ].		"No change."
	(convertible includes: klass) ifFalse: [ ^ sharedDoc ].		"Not one of our targets."

	"OK, convert to the specified class... 
	  reread the same persistence, but into an instance of the substitute class."
	result := klass new.
	(TSceneReader new) 
		read: result from: (sharedDoc persistString).

	^result! !

!QFileCard methodsFor: 'private' stamp: 'bgf 3/3/2009 21:35'!
uploadComplete: aSharedDocument

	"Callback after generated data has been uploaded to the server as a shared document
	 and the resulting shared document object is returned into the space."
	content == aSharedDocument ifFalse: [
		self releaseContent.
		self root ifNotNil: [ aSharedDocument ifNotNil:[ aSharedDocument registerResource ]].
	].
	content := aSharedDocument.
	self markAsInventory: (self isInInventory).
	rolodex ifNotNil: [ rolodex signal: #contentChanged ].! !

!QFileCard methodsFor: 'editing' stamp: 'ar 4/16/2011 17:12'!
useAsForumWelcome: evOrAvatar
	"Install this card as the forum welcome-document."
	self space ifNotNilDo: [ : space | 
		| prev | 
		prev := space welcomeDocument.
		space welcomeDocument: (self duplicate).
		space signal: #infoMessage with: ('This ', (self kindSummary), ' is now the Welcome for this Forum.').

		evOrAvatar avatar ifNotNil: [ : avatar | 
			avatar recordAction: #space:welcomeDoc:oldWelcomeDoc: 
				    class: QFileCard
				    arguments: { space . (self persistString) .  (prev ifNotNil: [ prev persistString ]) }
			].
	].! !

!QFileCard methodsFor: 'events' stamp: 'bgf 5/20/2008 20:57'!
willBeDroppedOn: aDropTarget byUser: avatar
	"Optional protocol - called before a drop involving this card as payload."

	| action |
	action := (self isInInventory ifTrue: [
				#DropFromInventory
			] ifFalse: [
				(aDropTarget isFrame and: [aDropTarget isInAvatar]) ifTrue: [
					#DropToInventory
				] ifFalse: [
					#DropFileCard
				]]).

	QActivityCapture action: action details: (self activityDetails) forSession: avatar sessionHex! !

TRectangle subclass: #QPortalBillboard
	instanceVariableNames: 'portal postcard fullColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Portals'!

!QPortalBillboard class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPortalBillboard methodsFor: 'render' stamp: 'bgf 1/3/2008 02:10'!
cachedPreview: ogl

	fullColor ifTrue: [
		^ ogl view session colorPreviewForUrl: (self postcardUrl).
	] ifFalse: [
		^ ogl view session previewForUrl: (self postcardUrl).
	].! !

!QPortalBillboard methodsFor: 'accessing' stamp: 'bgf 1/3/2008 02:12'!
fullColor: aBool
	"If true, use full-color previews rather than washed-out grayscale.  
	 Default is false, to use the grayscale versions."
	fullColor := aBool! !

!QPortalBillboard methodsFor: 'initialize' stamp: 'bgf 11/17/2009 17:21'!
initialize

	super initialize.
	self material: QMaterials fullWhiteMaterial.
	self solid: false.
	self material texture: (QTextures find: 'portal-void.gif').
	self fullColor: false.! !

!QPortalBillboard methodsFor: 'accessing' stamp: 'bgf 1/3/2008 02:34'!
portal: prtl

	"The portal for which I am a billboard.
	Exclusive with postcard; the most recent supplied is that used.
	One or the other must be supplied before the billboard is rendered."
	portal := prtl.
	postcard := nil.
	! !

!QPortalBillboard methodsFor: 'accessing' stamp: 'bgf 1/3/2008 02:35'!
postcard: aPostcard

	"The postcard whose preview I represent.
	 Exclusive with portal; the most recent supplied is that used.
	One or the other must be supplied before the billboard is rendered."

	postcard := aPostcard.
	portal := nil.! !

!QPortalBillboard methodsFor: 'accessing' stamp: 'bgf 1/3/2008 02:05'!
postcardUrl

	"The url string of the destination of the portal or postcard I represent "
	^ portal ifNotNil: [ portal postcardUrl ] ifNil: [ postcard urlString ].! !

!QPortalBillboard methodsFor: 'render' stamp: 'bgf 11/17/2009 17:22'!
render: ogl

	"Normally, the portal renders itself as an invisible (material opacity 0) rectangle;
	 this happens in the normal course of TFrame renderFrame: recursion,
	 after the portal-rendering phase of TSpace renderPortals.

	However, to provide a billboard for unresolved portals (rather than a mirror)  
	we enhance this if the portal has a billboard (a rectangle child).
	We locally (private to render) make the billboard visible for the duration of render.
	Since all of rendering takes place inside a lockout from the time cycle,
	no replication difficulties should arise.  (The billboard exists in all replicants,
	so its object ID is reachable and consistent in all; all we're doing is 
	local transient change to an instance variable of the billboard.)"

	| txtr |

	(portal isNil or: [portal resolved not and: [ portal beMirrorWhenUnresolved not] ]) ifTrue: [	
		"Resolved is a private flag set in render:depth:"
		txtr := self material texture.
		[
			(self cachedPreview: ogl) ifNotNilDo: [ :svcTexture |
				self material texture: svcTexture
			].
			super render: ogl.
		] ensure: [
			self material texture: txtr
		].
	].
! !

TRectangle subclass: #QSlaveProjectionScreen
	instanceVariableNames: 'displayRect master'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Desktop'!

!QSlaveProjectionScreen commentStamp: '<historical>' prior: 0!
Re-displays the content of a display panel,
simply by mapping the master display's content 
into an aspect-ratio adjusted rectangular child.
(The material is simply shared,
after ensuring appropriate aspect ratio.)

A display panel's menu is used to initiate and end display;
there is also a click-menu on the slave display
that can stop a running projection.

It is a simple rectangle without much geometry management of its own.
Click-menu, and mouse-over floatMessage, are the only interactive hooks.

All slave displays are registered as resources;
see QSpace allSlaveDisplays, which provides support
for the display panel projection menu.
(It is assumed that very few such screens would be used in a  single forum;
otherwise the panel menu will grow ungainly.)
!

!QSlaveProjectionScreen class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSlaveProjectionScreen class methodsFor: 'defaults' stamp: 'bgf 11/24/2009 19:53'!
defaultCanTakeAnnotations
	^ false! !

!QSlaveProjectionScreen class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QSlaveProjectionScreen class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:36'!
defaultMaterial
	^ QMaterials paperMaterial
! !

!QSlaveProjectionScreen class methodsFor: 'defaults' stamp: 'bgf 7/24/2008 18:43'!
resourceType
	^ #displaySlave! !

!QSlaveProjectionScreen methodsFor: 'space' stamp: 'bgf 7/21/2009 16:59'!
asPlace
	| place |
	"This lets 'Approach...' work reasonably for a plain un-managed projection screen."
	place :=	
		QPlace 
			category: 'Projections'
			label: (self objectName)
			frame: self
			selector: #destinationStanceFor: 
			sticky: true.
	place cameraSelector: #cameraStanceFor:.
	^ place
! !

!QSlaveProjectionScreen methodsFor: 'transform' stamp: 'bgf 6/29/2009 22:34'!
cameraStanceFor: anAvatar
	"I offer ground-aware full-on viewing"
	^ QGotoButton frame: self cameraStanceFor: anAvatar! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/28/2009 23:41'!
checkRoot
	"A future send from root:, or from a display panel being unrooted."
	frameRoot ifNil: [ self master: nil.].
	master ifNotNil: [ 
		master root ifNil: [ self master: nil ].
	].
! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/27/2008 23:38'!
clear

	"Erase the display-sharing (but leave the master-relationship active)."
	self removeChild: displayRect.
	displayRect material: (TMaterial new).
	self stopScript: #onAppExtentChanged:.

! !

!QSlaveProjectionScreen methodsFor: 'persist' stamp: 'bgf 7/25/2008 01:36'!
constructFrom: pStream

	| panel |
	super constructFrom: pStream.

	panel := (pStream readObject: 'sourcePanel' default: nil).
	panel ifNotNil: [
		"Future, so that rooting and construction has had a chance to finish."
		self future master: (panel sheet).
	].
! !

!QSlaveProjectionScreen methodsFor: 'menu' stamp: 'bgf 10/25/2009 15:15'!
contextMenuMaskForAvatar: avatar
	"Remove the system 'Attach Note' command from the menu."
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	mask remove: #attachNote ifAbsent: [].
	mask remove: #copyToInventory ifAbsent: [].	"Avoids confusion - people think it means
												 the content being projected."
	^ mask! !

!QSlaveProjectionScreen methodsFor: 'transform' stamp: 'bgf 7/26/2009 23:37'!
destinationStanceFor: anAvatar
	"I offer ground-aware full-on viewing"
	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/24/2008 19:10'!
destroy

	self master: nil.
	super destroy.
! !

!QSlaveProjectionScreen methodsFor: 'menu' stamp: 'bgf 3/12/2010 08:10'!
fillContextMenu: aMenu forEvent: event

	master ifNotNil: [
		(event avatar canModifyForum) ifTrue: [
			(aMenu add: 'Stop Projecting' target: self action: #stopProjecting:) 
				separator: true.
			].
		event avatar allowedToMove ifTrue: [
			aMenu add: 'Locate Source' target: self action: #locateSource: 
		].
	] ifNil: [
		(aMenu add: 'Not Projecting' target: self action: #noSuchAction:)
			disabled: true.
	].! !

!QSlaveProjectionScreen methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:37'!
initialize
	super initialize.	

	displayRect := TSimpleRectangle new.
	displayRect material: (TMaterial new).! !

!QSlaveProjectionScreen methodsFor: 'menu' stamp: 'bgf 10/25/2009 15:15'!
locateSource: anEvent
	"Menu action - take the avatar to the master display panel that is driving me."

	anEvent avatar ifNotNil:[
		master ifNotNil: [
			anEvent avatar takeMeToFrame: master.
		] ifNil: [
			anEvent avatar signal: #infoMessage with: 'The projector is not in use.'.
		].
	].! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/25/2008 01:33'!
master
	"QDisplaySheet - the guy whose content I'm mirroring right now."
	^ master! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/24/2008 23:57'!
master: aDisplaySheet

	"Become slaved to the given display sheet's content."

	master := aDisplaySheet.
	self stopScript: #sheetContentChanged.

	master ifNotNil: [
		self startScript: #sheetContentChanged when: { master . #contentChanged }.
		self sheetContentChanged.
	] ifNil: [
		self clear.
	].
! !

!QSlaveProjectionScreen methodsFor: 'events' stamp: 'bgf 7/27/2008 23:49'!
onAppExtentChanged: appExtent

	"On the fly change in my master's extent, triggered from app resizing;
	 we just follow suit."

	(displayRect isNil or: [ master isNil ] or: [ master contents isNil]) ifTrue: [
		^ self
	].
	displayRect
		extent: self extent aspect: (master contents originalAspect);
		uvScale: (master contents uvScale).! !

!QSlaveProjectionScreen methodsFor: 'persist' stamp: 'bgf 7/25/2008 01:35'!
persistOn: pStream

	super persistOn: pStream.

	"We can only persist the relationship if we're hooked into a valid DisplayPanel;
	 that's what constructFrom: expects."

	(master notNil 
	 and: [master parent isKindOf: QDisplayPanel]
	 and: [master root = self root] ) 
	 ifTrue: [
		pStream storeObject: 'sourcePanel' value: (master parent) default: nil
	].! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 10/19/2009 20:24'!
photographerClass
	"Prevents being reordered in a Destination parent by copy-to-inventory"
	^ QDirectPhotoJob! !

!QSlaveProjectionScreen methodsFor: 'events' stamp: 'bgf 10/25/2009 15:13'!
pointerEnter: anEvent
	"A simple mouseover, to remind  everyone that this is a projection,
	 whose contents cannot be interacted with directly.
	 We could bridge event-interaction eventually 
	 but the focus-details,  geometry-conversions, etc will be painful."

	self showFloatFor: anEvent.! !

!QSlaveProjectionScreen methodsFor: 'events' stamp: 'bgf 10/25/2009 15:13'!
pointerOver: event
	"Show the floating details if over the off-screen background
	 or if the projector is off, but don't keep showing them atop an active projection,
	 so you can point without the floating-message distraction." 
	(master isNil or: [ event selection frame = self "I.e. not the contentRect" ]) ifTrue: [
		self showFloatFor: event.
	].
! !

!QSlaveProjectionScreen methodsFor: 'initialize' stamp: 'bgf 7/24/2008 23:02'!
registerResource
	"Register this slave-display as a resource.
	 We keep these as resources mostly for quick access at the space level."
	self island registerResource: self type: self resourceType.
! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/24/2008 18:46'!
resourceType
	^ self class resourceType! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'das 7/28/2009 21:27'!
root: newRoot

	frameRoot 
		ifNotNil: [ self unregisterResource ].

	super root: newRoot.

	frameRoot 
		ifNotNil: [ self registerResource ].

	self future checkRoot.
	! !

!QSlaveProjectionScreen methodsFor: 'accessing' stamp: 'bgf 7/27/2008 23:43'!
sheetContentChanged

	"Any change in the content of my master sheet...
	 update the slave display by attaching to the material all over again.
	 This syncs us with the texture, the aspect details, etc."

	| src |

	self stopScript: #onAppExtentChanged:.
	master ifNotNil: [
		(master root = self root) ifFalse: [
			"Drop the hot potato, somebody's probably been destroyed."
			^ self master: nil.
		].
		src := master contents.
		src ifNotNil: [
			self startScript: #onAppExtentChanged: when: { src. #appExtentChanged }.
			displayRect 
				aspectFixed: true;
				 material: src material;
				 extent: self extent aspect: (src originalAspect);
				 localTransform: (Matrix4x4 identity);
				 translation: (0@0@( self extent length * 0.005));
				 uvScale: (src uvScale).
			self addChild: displayRect.
		] ifNil: [
			self clear.
		].
	] ifNil: [
		self clear.
	].
! !

!QSlaveProjectionScreen methodsFor: 'events' stamp: 'bgf 6/26/2009 22:55'!
showFloatFor: event

	"A simple mouseover message, 
	 mostly as a reminder that this is a projector and not an interactive panel."	
	event avatar ifNil: [ ^ self ].
	master ifNil: [
		event avatar showFloatMessage: (self objectName), '\Display Panels can be projected here.'
					 forEvent: event.
	] ifNotNil: [
		"No float if there's content... it gets in da way."
		event avatar signal: #infoMessage with: ( 'Projecting: ', master parent title )		
	].! !

!QSlaveProjectionScreen methodsFor: 'menu' stamp: 'bgf 3/13/2010 08:53'!
stopProjecting: event
	"Menu action - turn me off.  Needs to be made undoable."
	| oldmaster |
	oldmaster := self master ifNotNil: [ : m | m parent ].
	event avatar
		recordAction: #projection:newSlaves:oldSlaves:
		class: QDisplayPanel	
		arguments: { self .   
					{{ nil . self } } . 
					{{ oldmaster . self }}  }.
	self master: nil.! !

!QSlaveProjectionScreen methodsFor: 'initialize' stamp: 'bgf 7/24/2008 18:45'!
unregisterResource
	"Unregister this slave display as a resource"
	self island unregisterResource: self type: self resourceType.! !

TButton subclass: #QGotoButton
	instanceVariableNames: 'place suppressed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-Basic'!

!QGotoButton commentStamp: '<historical>' prior: 0!
A 'come-hither' arrow button, extracted from the TWindow control set.

The QGotoButton takes a 'destination:' frame (usually, the larger object
of which it is a part). Its firing behavior is to bring the avatar
into viewing position for that destination frame.

The button provides 'enable' and 'disable' operations which
make the button visible (and invisible.)  These meanted to be triggered
by mouse enter (mouse leave) on the destination, so the goto button
appears only when the destination frame has in the user's attention.
The disable behavior fades the button slowly, so there is time to
notice and use it with casual mouse steering.


!

!QGotoButton class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 3/15/2008 00:21'!
defaultCategory
	^ 'Gathering Places'! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 1/19/2010 16:04'!
defaultDoesClickSelect
	^ false
! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 2/17/2008 14:20'!
defaultExtent

	"The down arrow is a simple triangle..."

	^ (1@1@0.2)
			! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 04:08'!
defaultIsSolid
	"Since these are transient accessories, they need never be solid."
	^ false! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 3/4/2009 12:01'!
defaultIsSticky	
	"Default is to let suggestedPOV govern stickiness."
	^ nil! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 3/15/2008 00:22'!
defaultLabel
	^ 'Approach'! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 4/24/2009 17:01'!
defaultSlopFactor
	"The multiplier we use to create some slop 
	 when we compute a destination extent using the x-y of the bounding box.
	 We need some slop to avoid harsh cutoffs
	 due to depth/parallax and due to general numeric crappitiness.
	 This does not apply when the destination defines its own #destinationExtent,
	 only when we use the bounding box to derive it."
	^ 1.03 ! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 4/14/2009 14:12'!
defaultSuppressed
	^ false! !

!QGotoButton class methodsFor: 'utilities' stamp: 'bgf 4/24/2009 16:25'!
distanceFor: anAvatar toSee: aRectangleExtent
	"Basic trig utility.  
	 The avatar gives access to viewing window extent (aspect ratio)
	 and rendering camera angle for that user's client. 
	 Derive the distance from which a rendering looking straight at a rectangle 
	 of the given extent will fit the rectangle into the viewing window"
	| aspect opposite distance |
	aspect := (aRectangleExtent x max: 1.0) / (aRectangleExtent y max: 1.0).
	opposite := (aRectangleExtent y) / 2.0.
	distance := opposite / ((anAvatar viewAngle degreesToRadians / 2.0) tan) abs.
	"If the target is wider than the user's view, adjust to get x to fit."
	aspect > (anAvatar viewAspect) ifTrue: [
		distance := distance * (aspect / anAvatar viewAspect).
	].

	^ distance * 1.003	"(Much smaller bump than slop-factor; this is just to fight pixel-truncation)"! !

!QGotoButton class methodsFor: 'utilities' stamp: 'bgf 7/26/2009 17:58'!
distanceToSee: aRectangleExtent
	"Basic trig utility.  
	 Assuming default viewing (i.e not specific to avatar conditions)
	 Derive the distance from which a rendering looking straight at a rectangle 
	 of the given extent will fit the rectangle into the viewing window.

	 Used to place avatars for objects that also have a camera transform;
 	 we want all the avatars at a consistent distance, so they don't get in front of one another."

	| aspect opposite distance |
	aspect := (aRectangleExtent x max: 1.0) / (aRectangleExtent y max: 1.0).
	opposite := (aRectangleExtent y) / 2.0.
	distance := opposite / 0.41421 abs.		"(tan of (45/2) degrees)"
	"If the target is wider than the user's view, adjust to get x to fit."
	aspect > (1.3) ifTrue: [
		distance := distance * (aspect / 1.3).
	].

	^ distance * 1.1! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 2/17/2008 14:20'!
downArrow

	"The down arrow is a simple triangle..."

	^ (QArrowHead new)
			extent:  self defaultExtent;
			material: QMaterials placemarkMaterial;
			rotationAroundX:180;
			solid: false;
			yourself.
! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 08:46'!
downHilite
	^ (Color green /3) alpha: 0.5! !

!QGotoButton class methodsFor: 'utilities' stamp: 'bgf 4/26/2009 14:36'!
frame: frame cameraStanceFor: avatar
	"Trigonometric utility:
	 The stance to bring an avatar (camera) face-to-face with this viewable frame.
	 This is a shortcut to the inner implementation behind normal destination-finding,
	 avoiding all the gathering-circle offset-angle stuff."
	| distance center spot orientation |

	orientation := frame globalTransformReal orientation.
	center := frame localCenter.
	(frame parent = frame root) ifFalse: [
		center := frame parent globalTransformReal transformPoint: center
	].
	distance := QGotoButton distanceFor: avatar 
							 toSee: (frame destinationExtent).
	spot := center + (orientation transformPoint: (Vector3 x: 0 y: 0 z: distance)).
	
	^ Stance here: spot there: center! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 7/23/2009 10:34'!
gatheringSlotsForDistance: radius

	"Generate a circle of inward-facing gathering position (angles), for the given radius.
	 This is a circle of places, sorted by desireability, for forming a circular gathering.
	 The result list starts with the compass points,  and then has
	 slots that step down into more 'off-cardinal' points.
	 The 45 degree slots come after the 90, and then the 22.5, and so on."

	| result divisions angles personalRadius |

	result :=OrderedCollection new.
	angles := OrderedCollection new.

	"Subdivide the circle finer and finer until the next division 
	 would make the slot separation less than 4.  (16 = 4 squared)"

	divisions := 1.		"Number of divisions per quadrant"
	angles add: 0.0.
	personalRadius := 2.7.		"Min avatar-avatar center distance..."

	"The following is cosine law... squared length of the chord between
	 two points on the circle, given an angle subtended.
	 I.e. we're subdividing so long as the distance between avatar slots is > personalRadius."
	[  ((radius*radius*2) * (1.0 - ((90.0/divisions) degreeCos)))  > (personalRadius*personalRadius) ] whileTrue: [
		| da |
		divisions := divisions * 2.
		da := 90.0 / divisions.
		angles := angles, (angles collect: [ :a | a + da ]).	"Add the next fine-grained set of staggers"
	].

	1 to: divisions do: [  : i | 
		result addAll: (self quadrantSlotsForDistance: radius angleOffset: (angles at: i)).
	].
	^ result! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 7/26/2009 22:01'!
groundAdjust: stance forAvatar: avatar
	"Modify the given stance to keep the avatar on the ground standing.
	 This is used for viewing destinations that also supply a focal camera transform;
	 we can modulate the avatar position for such destinations,
	 as the focal transform preserves the correct first-person view."
	| y |
	y := (avatar root groundHeightAt: stance here) + avatar footToEyes.
	(y - (stance here y)) abs < (avatar avatarHeight * 3) ifTrue: [
		stance yHere: y.
		stance yThere: (stance yHere).
	].
	
	^ stance! !

!QGotoButton class methodsFor: 'utilities' stamp: 'bgf 7/27/2009 00:07'!
groundedStanceFor: anAvatar toView: frame inDestination: destFrame
	"The transform to apply to the avatar to bring it face-to-face with my destination,
	 but standing on the ground if the ground is fairly close to the vertical result.
	 For destinations whose  Place  supplies a cameraStanceFor: selector. 
	 Provides fanout viewpoints for a gathering in front of a flat-display."

	"Assigns the position assuming that the destination has a cameraStanceFor: definition,
	 so we always place avatars at the 'nominal' position for a 45 degree viewangle.
	 Camera adjustment, and render-time tricks to hide occluding avatars, help with the rest."

	^ self groundAdjust: (
			self stanceFor: anAvatar
				toView:  frame
				inDestination: destFrame)  
		   forAvatar: anAvatar
! !

!QGotoButton class methodsFor: 'defaults' stamp: 'bgf 11/8/2009 09:09'!
overHilite
	^ (Color yellow /2) alpha:0.2! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 7/28/2009 12:06'!
quadrantSlotsForDistance: distance angleOffset: degrees

	"Generate 4 compass-point inward-looking viewpoints on a circle of the given radius,
	 twisted from 0-90-180-270 by the given offset degrees.   
	 This is used to subdivide the circle when we make gathering circles.

	 There is no parallax adjustment for gatherings, so each pair
 	 uses the same angle for the slot itself and the avatar yaw to use for that slot.

	 The angles are normalized to be comparable to the results of Integer asSmallAngleDegrees
	 (we assume <degrees> is a positive offset) 
	 so they have to fall in the interval: (-180, 180]  "

	degrees = 0 ifTrue: [ ^ { 0 . -180 . 90 . -90 }  ].

	^ { degrees .
		-180 + degrees . "Opposite first... then the cross-views."
		90 + degrees . 
		-90 + degrees . 
		}.

! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 8/30/2009 15:12'!
slotForIndex: occupantIndex viewSlots: slots 
	"The slot (viewing angle) for the occupant in the given occupant-index of a viewing destination."
	| tier slot idx stagger fanout tierCompaction |
	tier := occupantIndex-1 // slots size.
	idx := (occupantIndex-1 \\ slots size) + 1.
	slot := slots at: idx.

	"A bit of adjustment to pull the outer rings (which have greater distance)
	 in closer to the centerline, to keep their perspectives a little nicer."
	tier = 0 ifTrue: [ ^ slot ].
	stagger := tier even ifTrue: [ -2 ] ifFalse: [ 2 ].	"Sideways offset switches row by row."
	idx = 1 ifTrue: [
		^ slot  + stagger
	] ifFalse: [ 
		| flanking |
		fanout := slots max.
		fanout > 45 ifTrue: [	" Wide arc, don't compress outer spreads."
			^ slot + stagger.
		] ifFalse: [
			"Viewing arc... reduce the spread angle a bit as we move back,
	 		 to reduce the width-growth of the outer radials."
			tierCompaction := (fanout *0.175 / (slots size+1.0)) * (tier).
		].
		flanking := (idx-1 / 2) ceiling. 
		slot < 0 ifTrue: [
			slot := slot + stagger + (tierCompaction*flanking).	
		] ifFalse: [
			slot > 0 ifTrue: [
				slot := slot + stagger -  (tierCompaction*flanking).
			] ifFalse: [
				slot := slot  + stagger
			].
		]
	].
	^ slot
! !

!QGotoButton class methodsFor: 'utilities' stamp: 'bgf 7/26/2009 23:50'!
stanceFor: anAvatar toView: frame inDestination: destFrame
	"The transform to apply to the avatar to bring it face-to-face with my destination,
	 or into an arc of viewing positions around that ideal position if there are more viewers.
	 Provides the proper fanout viewpoints for a flat-display destination."
	^ self stanceFor: anAvatar
		   toView: frame 
		   inDestination: destFrame
		   inAngle: 30.0      "I.e. arc around the front of the 2D x-y extent"
		   atDistance: (self distanceToSee: frame destinationExtent).
! !

!QGotoButton class methodsFor: 'utilities' stamp: 'bgf 2/16/2010 02:55'!
stanceFor: avatar toView: frame inDestination: destFrame  inAngle: angle atDistance: distance
	"Find an empty vantage point where
	 this avatar can go to view the given frame (a target within destFrame).
	 (Vantage slots come in a series of rings facing the frame's localCenter.)
	 Result is global Stance - the destination for the avatar to go to."
	| angles center occupants pos legalBox  avpositions xform orient slop |

	"Just a safety guard... in case we asynchronously go-to something that's been unrooted."
	frame root ifNil: [ ^ avatar globalTransformReal transformStance: (Stance new) ].
	center := frame localCenter.
	frame root avatarBox ifNotNil: [ : b |
		"If the destination is inside avatar-motion limits, then constrain the search in there also."
		(b pointInBox: (frame globalTransformReal transformPoint: center)) ifTrue: [
			legalBox := b.
		]].

	"Who's here viewing this already?  "
	occupants := (frame root usersAtDestination: destFrame).
	angles := self viewingSlotsForDistance: distance withinAngle: angle.

	" Look for a vacant slot.
	  (The slots may be sparsely occupied, after folk come and go)
	  Use x-z  to test occupancy since avatar positions are often y-adjusted."
	xform := frame parent globalTransformReal.
	orient := frame orientation.
	avpositions := occupants collect: [ : occ |  |v3|
							v3 := occ destinationPosition. 
							Point x: v3 x y: v3 z ].

	slop := (distance*0.08) squared min: 3.2.   "Avatar spacing tolerance, squared" 

	"Scan upwards slot by slot, using the number of occupants to cap the scan
  	 (if there is no free slot in the first n, then the n occupants must be in them,
	 or some part of the destination may be occluded by avatarBox limits."
	1 to: ((occupants size + 1 * angles size)) do: [ : i |
		| slot spot  d |
		slot := self slotForIndex: i viewSlots: angles.
		d := distance + ((i -1 // angles size) * 1.5).
		spot := 
			xform transformPoint: (
				center +
				(orient transformPoint:
					(Vector3 x: (slot degreeSin * d) y: 0 z: (slot degreeCos * d)))).
		(legalBox isNil or: [ legalBox pointInBox: spot ]) ifTrue: [
			(avpositions anySatisfy: [ : gpt |   
				(gpt squaredDistanceTo: (spot x @ spot z)) < slop ]
			)  ifFalse: [
				"No one here!!  Use this spot as the target stance."
				^ Stance here: (spot) there: (xform transformPoint: center)
			]].
	].

	"Cannot find a place within the box."
	pos := occupants indexOf: avatar ifAbsent: [ occupants size + 1 ].
	^ self stanceForSlot: (self slotForIndex: pos viewSlots: angles)
		   frame: frame
		   center: center
	   	   distance: distance + ((pos-1 // angles size)*1.5)

	

		 

! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 2/16/2010 02:35'!
stanceForSlot: slotAngle frame: frame center: centerV3 distance: dist
	"Worker routine for viewpoint finding.
		slot:  view-angle defines the view slot relative to my center.
		frame :  the destination frame to view
		center: its #localCenter
	 	distance: from avatar to frame's center.
	Return the transform that puts the avatar into the defined viewing position. "
	| spot |
	spot := centerV3 +
			(frame orientation transformPoint:
				(Vector3 x: (slotAngle degreeSin * dist) y: 0 z: (slotAngle degreeCos *dist))).

	^ Stance here: (frame parent globalTransformReal transformPoint: spot)
	 		there: (frame parent globalTransformReal transformPoint: centerV3).
! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 7/28/2009 12:17'!
viewingSlotsForDistance: distance withinAngle: anAngle

	"Main entry to generate a list of viewpoints 
	 (each result is an angle (0 meaning straightOn down the z axis)
	 around a destination.  
	 If the given limit angle is small,  then a packed arc is generated... 	
		appropriate for viewing a panel.

	 If the limit angle is 180,  a nicely ordered round gathering is generated,
	 which supplies compass points in order of desireablity
	 N, S, E, W, and then NE, SW, NW, SE, and so into the finer grain."	

	(anAngle >= (180-(Float epsilon))) ifTrue: [
		^ self gatheringSlotsForDistance: distance
	] ifFalse: [
		^ self viewingSlotsForDistance: distance withinArc: anAngle
	]! !

!QGotoButton class methodsFor: 'private' stamp: 'bgf 7/23/2009 10:32'!
viewingSlotsForDistance: distance withinArc: anAcuteAngle

	"Generate an arc of viewing angles for the given viewing distance, 
	 within the given arc, looking at my localCenter.
	 Heuristically attempts to fit a reasonable spread of avatars on the arc
	 fitted on front+center... the greater the distance, the more slots.
	 The slots are ordered by desirability - front-center is first.

	 The results are the { angle , adjusted-angle } pairs expected by findViewpoint,
	 where the adjusted angles do a bit of parallax correction."

	| result fanout da |
	result := OrderedCollection new.

	"Direct facing view."
	result add: 0.
	fanout := true caseOf: {
				[ distance < 6.0 ] -> [1].
				[ distance < 10.0 ] -> [2].
				[ distance < 15.0 ] -> [3].
			} otherwise: [ (distance / 5) rounded ].
	da := (anAcuteAngle / (fanout max: 1)) min: 11.0.	"11 degree cap - minimizes disadvantage"

	1 to: fanout do: [ :i |
		result add: da*i .
		result add: da negated * i.
	].

	^ result 
! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/17/2008 14:23'!
arrow
	"The down-arrow itself"
	^ frame! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 09:48'!
asPlace
	"Mostly for use as a delegation point for components that contain buttons."
	^place
! !

!QGotoButton methodsFor: 'events' stamp: 'bgf 3/10/2010 08:03'!
bringToDestination: event 

	"Callback: bring the user to see my destination,
	 and bring them to the suggested POV for the destination.
	 Returns the number of drive-steps used in the resulting goto cycle,
	 which can be used to invent an estimate of how long it will take;
	 the real timing is determined by the avatar-user's harness cycles.

	 Honors leader-mode, when allowedToMove is false."

	| atar dest |
	atar := event avatar ifNil:[^0].
	atar allowedToMove ifFalse: [ ^ 0 ].
	(self hasAncestor: atar) ifTrue:[^ 0]. "don't go there"
	(self root = atar root) ifFalse:[^0]. "don't go there either"
	(dest := (self destination)) ifNil:[ ^ 0 ].

	self root miramarUI ifTrue:[
		(self alternateBringToDestination:event) ifNil:[ ^0 ].
	].
	QActivityCapture action: #GoToButton
					 details: (dest activityDetails)
					 forSession: (atar sessionHex).
	(atar isAtPlace: place) ifFalse: [
		^ atar goToPlace: self asPlace
	] ifTrue: [		
		"If already here, just reset to restore from look-mode."
		atar suggestPOV: (place suggestedPOV) forPlace: place;
			 updateFocalTransform
	].
	^ 0
! !

!QGotoButton methodsFor: 'testing' stamp: 'bgf 2/18/2009 21:58'!
canEditMaterial
	^ false! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 03:32'!
category
	"String - the Places-menu category when I am presented as a Place"
	^ place category! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 03:33'!
category: aPlaceCategory
	"String - the Places-menu category when I am presented as a Place"
	place category: aPlaceCategory! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 9/30/2007 14:19'!
color: aColor
	"Ignore ... tbutton uses this for setup,
	 but we already control the material color of the arrow head."! !

!QGotoButton methodsFor: 'persist' stamp: 'bgf 4/22/2009 06:48'!
constructFrom: pStream
	super constructFrom: pStream.

	"TButton properties"
	self spin: (pStream readBool: 'spin' default: spin).
	self persist: (pStream readString: 'persist'  default: persist).

	"My own properties"
	self destination: (pStream readObject: 'destination' default: self destination).
	self label: (pStream readString: 'label' default: (self label)).
	self category: (pStream readString: 'category' default: self category).
	self suggestedPOV: (pStream readString: 'suggestedPOV'  default: self suggestedPOV).
	self isSticky: (pStream readBool: 'sticky' default: (self class defaultIsSticky)).
	self suppressed: (pStream readBool: 'suppressed' default: (self class defaultSuppressed)).

	self arrow extent: (pStream readVector3: 'arrowExtent' default: (self arrow extent)).
	self posture: (pStream readString: 'posture' default: self posture).
	self seatingRange: (pStream readNumber: 'seatingRange' default: (self seatingRange)).! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 3/1/2009 14:32'!
destination
	[
		^ place destinationFrame
	] on: Error do: [ :ex |
		self island class == TIsland ifTrue: [
			ex pass	
		]
		"Otherwise, not an error... we expect destination to be nil off-island."
	].! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 3/5/2010 11:29'!
destination: aTframe
	"Set the frame this button brings you to."

	place destinationID: (aTframe ifNotNil: [ aTframe objectID ]).

	"If we're being attached to something with camera-focus, take its cam selector.
	 Don't do this for QDestinations since they borrow my button and modify it directly.
	 This is for panels, fixed displays, etc."
	(aTframe notNil and: [(aTframe isKindOf: QDestination) not])
	 ifTrue: [
		aTframe asPlace ifNotNilDo: [ : plc | 
			place cameraSelector: plc cameraSelector
			]].

	self target: self action: #bringToDestination:.
	self target2: self action2: #bringToDestination:.

	self resetTriggers.! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 7/26/2009 23:37'!
destinationStanceFor: anAvatar
	"The transform to apply to the avatar to bring it face-to-face with my destination.
	 Delegate to the destination frame if possible, otherwise calculate a nice default position."
	| result df |
	df := place destinationFrame.
	df ifNotNil: [
		result := place destinationStanceFor: anAvatar
	].

	"The default frame implementation returns nil if no container provides a destination...
 	 so the final fallback is to directly calculate a destination based on the frame extent."
	result ifNil: [
		| dest | 
		dest := (self destination ifNil: [ self ]).
		place hasCameraStance ifFalse: [
			result := self class stanceFor: anAvatar toView: dest inDestination: dest.
		] ifTrue: [
			result := self class groundedStanceFor: anAvatar toView: dest inDestination: dest.
		].
	].
	^ result! !

!QGotoButton methodsFor: 'action' stamp: 'bgf 4/14/2009 14:16'!
enable
	self suppressed ifFalse: [ super enable ].! !

!QGotoButton methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:32'!
initialize

	super initialize.
	doSpin := 0.
	place := QPlace new.
	suppressed := self class defaultSuppressed.
	self suggestedPOV: 'First Person'.
	self initializeWithFrame: (self class downArrow).
	self label: self class defaultLabel.
	self category: self class defaultCategory.! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 3/4/2009 12:37'!
isSticky
	"The value of the sticky-but-no-POV override flag."
	^ place isSticky! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 3/4/2009 12:37'!
isSticky: aBool
	"The value of the sticky-but-no-POV override flag."
	place isSticky: aBool! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 09:12'!
label	
	^ place label! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 09:11'!
label: aLabel
	place label: aLabel.	
	super label: aLabel.! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/8/2008 03:50'!
persist: yesNo
	super persist: yesNo.
	self resetTriggers.! !

!QGotoButton methodsFor: 'persist' stamp: 'bgf 4/22/2009 12:40'!
persistOn: pStream
	super persistOn: pStream.

	"TButton properties"
	pStream storeBool: 'spin' value: spin default: false.
	pStream storeBool: 'persist' value: persist default: false.

	"My own properties"
	pStream storeObject: 'destination' value: (self destination) default: nil.
	pStream storeString: 'label' value: (self label) default: 'Approach'.
	pStream storeString: 'posture' value: (self posture) default: nil.
	pStream storeString: 'category' value: (self category) default: nil.
	pStream storeString: 'suggestedPOV' value: (self suggestedPOV ifNil: ['none']) default: 'First Person'.
	pStream storeBool: 'sticky' value: (self isSticky) default: (self class defaultIsSticky).
	pStream storeBool: 'suppressed' value: (self suppressed) default: (self class defaultSuppressed).
	pStream storeVector3: 'arrowExtent' value: (self arrow extent) default: (self class defaultExtent).
	pStream storeNumber: 'seatingRange' value: (self seatingRange) default: 0.
! !

!QGotoButton methodsFor: 'pick' stamp: 'bgf 4/12/2008 18:24'!
pickDownTree: ray
	"Goto-buttons are solid w/respect to normal picking,
	 but not to down-ray (avatar falling) picking."
	^ false
! !

!QGotoButton methodsFor: 'events' stamp: 'bgf 2/24/2009 03:38'!
pointerOver: event
	"Provoke the warp-cursor."
	self destination ifNotNil: [
		event avatar signal: #overWarpPlace with: (self asPlace).
	].! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 03:33'!
posture
	"The suggested avatar posture to apply once w've arrived here."
	^ place posture! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 03:34'!
posture: pSym
	"Specify a default posture for this destination."
	place posture: (pSym ifNotNil: [pSym asSymbol]).! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:56'!
privateScale: aFactor
	"Goto buttons are fixed-size UI elements 
	 that ought not be resized too extremely 
	 when the destination construct is casually resized.
	 So we override scale to reduce the scale factor proportionally."

	super privateScale: (aFactor sqrt).		"Even forced, don't be proportional, to preserve UI feel."
! !

!QGotoButton methodsFor: 'initialize' stamp: 'bgf 2/24/2009 03:39'!
resetTriggers

	"When the persist flag and/or destination change, update the auto-over triggers."

	self stopScript: #enable.
	self stopScript: #disable.

	(persist not and: [ self destination notNil ]) ifTrue: [

		self startScript: #enable when: { self destination . #pointerOver }.
		self startScript: #enable when: { self destination . #pointerEnter }.
		self startScript: #disable when: { self destination . #pointerLeave }.
	].! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 4/22/2009 07:31'!
seatingRange
	"If zero, no auto-seating within the destination.
	 If > zero, then this is a distance in standard units added to the bounding box,
	 to define the total volume in which seats are detected."
	^ place seatingRange! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 4/22/2009 16:26'!
seatingRange: radius
	"If zero, no auto-seating within the destination.
	 If > zero, then this is a distance in standard units added to the bounding box,
	 to define the total volume in which seats are detected."
	place seatingRange: radius! !

!QGotoButton methodsFor: 'stepping' stamp: 'bgf 9/30/2007 14:21'!
step

	"Fade the button when animating it away after a mouse-leave,
	 and rotate the button if it is spinning.
	 Customized (from TButton) to spin whenever the button is visible, not just on mouseover.
	 We use this for goto buttons that are above 3D destinations like tables."

	| timeLeft |
	persist ifFalse:[
		animateTime ifNotNil: [
			timeLeft := animateTime - self now.
			timeLeft <= 0 ifTrue: [
				stepping := false.
				animateTime := nil. 
				frame solidVisibleTree: false. 
				frame2 ifNotNil:[frame2 solidVisibleTree: false. ].
			] ifFalse: [
				self transparency:(timeLeft asFloat / self fadeDelay)
			].
		].
	].

	spin ifTrue:[ 
		doSpin := doSpin -2. 
		doSpin < -360 ifTrue:[doSpin := 0.].
		self rotationAroundY: doSpin.
	].
	stepping ifTrue:[(self future: 100) step].! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 03:32'!
suggestedPOV
	"One of the names from QUserPreferences ViewpointTable, or empty;
	 suggested point of view that the user will be warped to if they choose this destination;
	 default is nil... no POV modulation."

	^ place suggestedPOV! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 2/24/2009 03:32'!
suggestedPOV: aString

	"One of the pov names from QUserPreferences viewpointTable
	 (or nil),  a POV setting associated with this destination.
	 Default is 'First Person', so the button takes one into first-person view.
	 If this is set to nil, then the destination is not sticky and does not affect the user's camera POV."

	(aString isNil or: [QUserPreferences viewpointTable includesKey: aString]) ifTrue: [
		place suggestedPOV: aString
	] ifFalse: [
		"For persistence of nil..."
		(aString = 'none') 
			ifTrue: [ place suggestedPOV: nil ]
			ifFalse: [
				"Coding-time error, but we don't raise 'error' in replicated space over trivialities..."
				QLogger logError: 'Unknown POV ', aString, ' supplied to GotoButton'
			].
	].! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 4/14/2009 15:25'!
suppressed
	"Bool - is this button suppressed?
	 A suppressed goto button does not appear when its destination is moused-over."
	^ suppressed! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 4/14/2009 15:25'!
suppressed: bSuppressed
	"Bool - is this button suppressed?
	 A suppressed goto button does not appear when its destination is moused-over."
	bSuppressed ifTrue: [
		self persist: false.
		self disable.
	].
	suppressed := bSuppressed.! !

!QGotoButton methodsFor: 'accessing' stamp: 'bgf 1/26/2008 13:38'!
transparency: aNumber
	super transparency: (aNumber min: 0.9)! !

TParticleTxtr subclass: #QParticleTxtr4x4
	instanceVariableNames: 'txtrMatrices txtrIndices count'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Widgets-General'!

!QParticleTxtr4x4 class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QParticleTxtr4x4 methodsFor: 'render' stamp: 'jdm 10/5/2007 11:24'!
doRenderColor: ogl
	| mat yaw rMesh |

	ogl glMultMatrixf: self globalTransform inverseByTranspose transposed.
	mat := Matrix4x4 new.
	billBoard ifTrue:[
		yaw := ogl camera globalYaw.
		mat rotationAroundY: yaw.
		].
	rectangle getDefaultMaterial enable: ogl.
	rMesh := rectangle renderStart: ogl.
	rMesh ifNil:[^ self].
	ogl glMatrixMode: GLTexture;
		glPushMatrix.
	1 to: size do:[:index |
		ogl glMatrixMode:GLTexture;
			glLoadTransposeMatrixf: (txtrMatrices at: (txtrIndices at:index));
			glMatrixMode: GLModelview.
		mat translation: (pPosition at: index).
		ogl glPushMatrix. 
		ogl glMultMatrixf: mat transposed. 
		ogl glColor4fv: (pColor at: index).
		rectangle render: ogl mesh: rMesh.
		ogl glPopMatrix.
	].
	rectangle getDefaultMaterial disable: ogl.
	ogl glMatrixMode: GLTexture;
		glPopMatrix;
		glMatrixMode: GLModelview.
	ogl endMesh.! !

!QParticleTxtr4x4 methodsFor: 'initialize' stamp: 'jdm 8/21/2007 08:17'!
initialize

	| textureMatrix |
	super initialize.
	count := 1.
	txtrMatrices := OrderedCollection new.
	1 to: 4 do:[:i |
		1 to: 4 do:[:j |
			textureMatrix := Matrix4x4 new.
			textureMatrix translation: (Vector3 x:i-1*0.25 y: j-1*0.25 z: 0.0).
			textureMatrix scale: (Vector3 x: 0.25 y: 0.25 z: 1.0).

			txtrMatrices add: textureMatrix.
		]
	].
! !

!QParticleTxtr4x4 methodsFor: 'initialize' stamp: 'das 11/5/2006 22:42'!
initializeWithSize: sz

	super initializeWithSize: sz.
	txtrIndices := Array ofSize: sz.
	1 to: sz do:[ :i | txtrIndices at: i put:16 atRandom].

! !

!QParticleTxtr4x4 methodsFor: 'stepping' stamp: 'das 10/1/2006 17:14'!
step

	super step.
	count := count + 1.
	count = size ifTrue:[count := 1].
! !
QEmbeddedApp initialize!
QEmbeddedMpegPlayer initialize!
